Package: spam
Type: Package
Title: SPArse Matrix
Version: 2.11-4
Date: 2026-05-28
Authors@R: c(person("Reinhard", "Furrer", role = c("aut", "cre"),
                    email = "reinhard.furrer@math.uzh.ch",
                    comment = c(ORCID = "0000-0002-6319-2332")),
             person("Florian", "Gerber", role = c("aut"),
                    email = "florian.gerber@math.uzh.ch",
                    comment = c(ORCID = "0000-0001-8545-5263")),
             person("Roman", "Flury", role = c("aut"),
                    email = "roman.flury@math.uzh.ch",
                    comment = c(ORCID = "0000-0002-0349-8698")),
             person("Daniel", "Gerber", role = "ctb",
                    email = "daniel_gerber_2222@hotmail.com"),
             person("Kaspar", "Moesinger", role = "ctb",
                    email = "kaspar.moesinger@gmail.com"),
             person("Annina", "Cincera", email = "annina.cincera@math.uzh.ch", role = "ctb"),
             person("Youcef", "Saad", role = "ctb",
                    comment = "SPARSEKIT http://www-users.cs.umn.edu/~saad/software/SPARSKIT/"),
             person(c("Esmond", "G."), "Ng", role = "ctb",
                    comment = "Fortran Cholesky routines"),
             person(c("Barry", "W."), "Peyton", role = "ctb",
                    comment = "Fortran Cholesky routines"),
             person(c("Joseph", "W.H."), "Liu", role = "ctb",
                    comment = "Fortran Cholesky routines"),
             person(c("Alan", "D."), "George", role = "ctb",
                    comment = "Fortran Cholesky routines"),
             person(c("Lehoucq", "B."), "Rich", role = "ctb",
                    comment = "ARPACK"),
             person(c("Maschhoff"), "Kristi", role = "ctb",
                    comment = "ARPACK"),
             person(c("Sorensen", "C."), "Danny", role = "ctb",
                    comment = "ARPACK"),
             person(c("Yang"), "Chao", role = "ctb",
                    comment = "ARPACK"))
Depends: R (>= 4.0)
Imports: dotCall64, grid, methods, Rcpp (>= 1.0.8.3)
LinkingTo: Rcpp
Suggests: spam64, fields, Matrix, testthat, R.rsp, truncdist, knitr,
        rmarkdown
VignetteBuilder: R.rsp, knitr
Description: Set of functions for sparse matrix algebra.
    Differences with other sparse matrix packages are:
    (1) we only support (essentially) one sparse matrix format,
    (2) based on transparent and simple structure(s),
    (3) tailored for MCMC calculations within G(M)RF.
    (4) and it is fast and scalable (with the extension package spam64).
    Documentation about 'spam' is provided by vignettes included in this package, see also Furrer and Sain (2010) <doi:10.18637/jss.v036.i10>; see 'citation("spam")' for details.
LazyData: true
License: LGPL-2 | BSD_3_clause + file LICENSE
URL: https://www.math.uzh.ch:443/spam/
BugReports: https://git.math.uzh.ch/reinhard.furrer/spam/-/issues
NeedsCompilation: yes
Packaged: 2026-05-28 14:59:09 UTC; furrer
Author: Reinhard Furrer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6319-2332>),
  Florian Gerber [aut] (ORCID: <https://orcid.org/0000-0001-8545-5263>),
  Roman Flury [aut] (ORCID: <https://orcid.org/0000-0002-0349-8698>),
  Daniel Gerber [ctb],
  Kaspar Moesinger [ctb],
  Annina Cincera [ctb],
  Youcef Saad [ctb] (SPARSEKIT
    http://www-users.cs.umn.edu/~saad/software/SPARSKIT/),
  Esmond G. Ng [ctb] (Fortran Cholesky routines),
  Barry W. Peyton [ctb] (Fortran Cholesky routines),
  Joseph W.H. Liu [ctb] (Fortran Cholesky routines),
  Alan D. George [ctb] (Fortran Cholesky routines),
  Lehoucq B. Rich [ctb] (ARPACK),
  Maschhoff Kristi [ctb] (ARPACK),
  Sorensen C. Danny [ctb] (ARPACK),
  Yang Chao [ctb] (ARPACK)
Maintainer: Reinhard Furrer <reinhard.furrer@math.uzh.ch>
Repository: CRAN
Date/Publication: 2026-05-29 16:50:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-29 17:28:30 UTC; unix
Archs: spam.so.dSYM

useDynLib(spam, .registration = TRUE)
importFrom("Rcpp", "evalCpp")

#importFrom("grid", "grid.layout", "grid.draw", "unit", "gTree", "gList", "gpar",
#                "textGrob", "rectGrob", "linesGrob", "yaxisGrob", "xaxisGrob", "pointsGrob",
#                "editGrob", "grobHeight", "stringHeight",
#                "plotViewport","viewport","dataViewport",
#                "vpStack","vpList")
import("grid")


importFrom("dotCall64", ".C64", "integer_dc", "numeric_dc", "vector_dc")
         # these are the only functions exported by dotCall64, btw.
importFrom("grDevices", "gray", "heat.colors")
importFrom("graphics",
           "axis", "box", "image.default", "par", "points", "polygon", "text")
importFrom("methods",
           "callGeneric", "is", "new", "slot<-",  "validObject", "hasArg")
importFrom("stats", "optim", "rnorm", "runif", "rbinom", "rchisq", "var", "dist")
importFrom("utils", "help", "object.size")

# Next two paragraphs are from Matrix....
# Currently, group generics need to be explicitly imported (Bug?):
importFrom("methods", #Arith,Compare,
           Math, Math2, Summary#, Complex
           )


S3method("all.equal","spam")
#S3method("as.matrix.csr","spam")
S3method("as.matrix","spam")
S3method("as.vector","spam")
S3method("determinant","spam")
S3method("determinant","spam.chol.NgPeyton")
S3method("chol","spam")
S3method("diff","spam")
S3method("dim<-","spam")
S3method("head","spam")
S3method("image","spam")
S3method("isSymmetric","spam")
S3method("plot","spam")
S3method("print","spam")
S3method("print","spam.chol.NgPeyton")
S3method("solve","spam")
#S3method("subset_rows","spam")
S3method("subset","spam")
S3method("summary","spam")
S3method("summary","spam.chol.NgPeyton")
S3method("print","summary.spam")
S3method("print","summary.spam.chol.NgPeyton")
S3method("t","spam")
S3method("tail","spam")
S3method("update","spam.chol.NgPeyton")

S3method("cbind","spam")
S3method("rbind","spam")

# Generic functions
export(
       "spam_random",
       "validspamobject",   # __DEFUNCT__

       "is.spam",

       "spam.version",
       "spam.Version",

       "validate_spam",

       "nearest.dist",
       "spam_rdist",
       "spam_rdist.earth",

       "as.spam.matrix",
       "as.spam.numeric",
       "as.spam.spam",
       "as.spam.dist",
       "as.spam.chol.NgPeyton",
       "as.spam.list",

       "as.vector.spam",
       "as.matrix.spam",
       "spam.list",
       "spam.numeric",

       "diag.of.spam",
       "diag.spam",
       "spam_diag",
       "diag<-.spam",
       "diag.spam<-",
       "rbind.spam",
       "cbind.spam",
       "upper.tri.spam",
       "lower.tri.spam",
       "t.spam",

       "dim<-.spam",
       "pad<-.spam",

       "isSymmetric.spam",
       "all.equal.spam",

       "kronecker.default",
       "kronecker.spam",

       "gmult",

       "diff.spam",

       "circulant.spam",
       "toeplitz.spam",

       "determinant.spam",
       "determinant.spam.chol.NgPeyton",
       "det",# << "identical" as base - but with correct determinant()
       "chol.spam",
       "solve.spam",
       "forwardsolve.spam",
       "backsolve.spam",
       "update.spam.chol.NgPeyton",

       "norm.spam",

       "plot.spam",
       "display.spam",
       "image.spam",

       "print.spam",
       "summary.spam",
       "print.spam.chol.NgPeyton",
       "summary.spam.chol.NgPeyton",

       "apply.spam",

       "rmvnorm",
       "rmvnorm.spam",
       "rmvnorm.canonical",
       "rmvnorm.prec",
       "rmvnorm.const",
       "rmvnorm.canonical.const",
       "rmvnorm.prec.const",
       "rmvnorm.conditional",
       "rmvnorm.cond",

       "rmvt",
       "rmvt.spam",

       "rgrf",

       "precmat",
       "precmat.RW1",
       "precmat.RW2",
       "precmat.RWn",
       "precmat.season",
       "precmat.IGMRFreglat",
       "precmat.IGMRFirreglat",
       "precmat.GMRFreglat",

       "covmat",
       "cov.exp",
       "cov.sph",
       "cor.sph",
       "cov.nug",
       "cov.wu1",
       "cov.wu2",
       "cov.wu3",
       "cov.wend1",
       "cov.wend2",
       "cov.mat",
       "cov.finnmat",
    "cov.mat12",
    "cov.mat32",
    "cov.mat52",


       "rowSums.spam",
       "colSums.spam",
       "rowMeans.spam",
       "colMeans.spam",

       "head.spam",
       "tail.spam",

       "chol2inv.spam",

       "mle",
       "mle.spam",
       "mle.nomean",
       "mle.nomean.spam",
       "neg2loglikelihood",
       "neg2loglikelihood.nomean",
       "neg2loglikelihood.spam",

       "bdiag.spam",

       "var.spam",
       "eigen.spam",
       "eigen_approx",

       "bandwidth",

#       ".spam.matmul.mat",
#       ".spam.matmul",
#       "solve.spam.mat",
#       "solve.spam.dummy",

        "subset.spam",
#       "subset_rows.spam",

        "triplet",
        "as.spam.matrix.csr",
#       "as.matrix.csr.spam",
        "as.dgRMatrix.spam",
        "as.dgCMatrix.spam",
        "as.spam.dgRMatrix",
        "as.spam.dgCMatrix",

        "read.MM",
        "read.HB",

        "powerboost",

        "permutation.spam",

        "crossprod.spam",
        "tcrossprod.spam",

        "map.landkreis",
        "adjacency.landkreis",

        "germany.plot",
#       "germany.info",

        "grid_trace2",
        "grid_zoom",

        "rowpointers<-",
        "entries<-",
        "colindices<-",
        "dimension<-",

        "cleanup",
#       "backsolve"

        "print_nnzpos"
)


# export the two classes
exportClasses("spam",
              "spam.chol.NgPeyton")

exportMethods(
              "Math",
              "Math2",
              "Summary",

         #     "show",
              "print",
              "image",
              "display",

              "spam",
              "as.spam",

              "isSymmetric",
              "all.equal",

              "summary",
              "length",
              "length<-",
              "c",
              "dim",
              "dim<-",
              "pad<-",

              "rbind",
              "cbind",
              "as.spam",
              "spam",
              "as.vector",
              "as.matrix",
              "determinant",
              "t",
              "diag",
              "diag<-",
              "upper.tri",
              "lower.tri",

              "norm",
              "rowSums",
              "rowMeans",
              "colSums",
              "colMeans",

              "head",
              "tail",

              "chol",
              "ordering",
              "forwardsolve",
              "backsolve",
              "solve",
              "chol2inv",

              "kronecker",

              "permutation",

              "crossprod",
              "tcrossprod",


              "[",
              "[<-",

              "%*%",
              "%d*%",
              "%d+%",
              "-",
              "+",
              "*",
              "/",
              "&",
              "|"
)

Versions of `spam` prior to 1.00 included a LICENSE file that
described the license of some Fortran routines incompatible with the
general R philosophy.  The routines are in src/cholmodified.f.  For
the packages `SparseM` and `spam`, Esmond Ng and Barry Peyton have
very kindly given permission to use their routines under an open
source license (2012/03/09 and 2014/08/29).  They have requested that
their code be credited via the following two publications:

   Esmond G. Ng and Barry W. Peyton, "Block sparse Cholesky algorithms
   on advanced uniprocessor computers". SIAM J. Sci. Stat. Comput.
   14  (1993), pp. 1034-1056.

   John R. Gilbert, Esmond G. Ng, and Barry W. Peyton, "An efficient
   algorithm to compute row and column counts for sparse Cholesky
   factorization". SIAM J. Matrix Anal. Appl. 15 (1994), pp. 1075-1091.

The old license file of `spam` versions <1.0 is in ./inst/0LICENCE 

# This demo contains the R code of the example in Section 5.1 of the
# JSS article:
#     "spam: A Sparse Matrix R Package with Emphasis on
#            MCMC Methods for Gaussian Markov Random Fields"
#
# Compared to the R code given in the article, here we give:
# - improved formatting
# - more comments
# - the R code to construct the figures

# SETUP:
library("spam")
options(spam.structurebased=TRUE)
data("UKDriverDeaths")

y <- sqrt(c(UKDriverDeaths))       # square root counts

n <- length(y)                     # n=192
m <- 12                            # We want to predict for one season.
nm <- n+m                          # Total length of s and t


priorshape <-  c(4, 1, 1)          # alpha's, as in Rue & Held (2005)
priorinvscale <- c(4, 0.1, 0.0005) # beta's

# Construct the individual block precisions
# (based on unit precision parameters kappa, denoted with k):

# Qsy, Qty are trivial:
Qsy <- diag.spam(n)
pad(Qsy) <- c(n+m, n)  # previously:  dim(Qsy) <- c(n+m, n)

Qty <- Qsy

Qst <- spam(0, nm, nm)
Qst[cbind(1:n, 1:n)] <- rep(1, n)


# The form of Qss is given by (Rue and Held equation 3.59).
# Qss can be constructed with a loop:
Qss <- spam(0, nm, nm)
for (i in 0:(nm-m)) {
    Qss[i+1:m,i+1:m] <- Qss[i+1:m, i+1:m] + matrix(1,m,m)
#     Qss[i+1:m,i+1:m] <- Qss[i+1:m, i+1:m] + 1 # for older versions of spam
}

# Note that for the final version we need:
# Qss <- k_s * Qss + k_y * diag.spam(nm)


# The form of Qtt is given by (Rue and Held equation 3.40).
# Similar approaches to construct Qtt:

Qtt <- spam(0,nm,nm)
Qtt[cbind(1:(nm-1),2:nm)] <- -c(2,rep(4,nm-3),2)
Qtt[cbind(1:(nm-2),3:nm)] <- rep(1,nm-2)
Qtt <- Qtt + t( Qtt)
diag(Qtt) <- c(1,5,rep(6,nm-4),5,1)


# Create temporary kappa and precision matrix to illustrate
# adjacency matrix and ordering.
k <- c(1,1,1)
Qst_yk <- rbind(cbind(k[2]*Qss + k[1]*diag.spam(nm), k[1]*Qst),
       	        cbind(k[1]*Qst, k[3]*Qtt + k[1]*diag.spam(nm)))

struct <- chol(Qst_yk)

# Figure 6:
display(Qst_yk)
display(struct)

# Note that we do not provide the exactly the same ordering
# algorithms. Hence, the following is sightly different than
# Figure RH4.2.
cholQst_yk <- chol(Qst_yk,pivot="RCM")
P <- ordering(cholQst_yk)
display(Qst_yk[P,P])

# Recall:
# k=( kappa_y, kappa_s, kappa_t)'

# Gibbs sampler
ngibbs <- 500   # Is very fast!
burnin <- 10    # > 0
totalg <- ngibbs+burnin
set.seed(14)

# Initialize parameters:
spost <- tpost <- array(0, c(totalg, nm))
kpost <- array(0, c(totalg, 3))

# Starting values:
kpost[1,] <- c(.5,28,500)
tpost[1,] <- 40

# calculation of a few variables:
postshape <- priorshape + c(	n/2, (n+1)/2, (n+m-2)/2)

# GIBBS' ITERATIONS:
timing <- system.time({
for (ig in 2:totalg) {

  Q <- rbind(cbind(kpost[ig-1,2]*Qss + kpost[ig-1,1]*Qst,
                   kpost[ig-1,1]*Qst),
             cbind(kpost[ig-1,1]*Qst,
                   kpost[ig-1,3]*Qtt + kpost[ig-1,1]*Qst))


  b <- c(kpost[ig-1,1]*Qsy %*% y, kpost[ig-1,1]*Qsy %*% y)

  tmp <- rmvnorm.canonical(1, b, Q, Lstruct=struct)


  spost[ig,] <- tmp[1:nm]

  tpost[ig,] <- tmp[1:nm+nm]


  tmp <- y-spost[ig,1:n]-tpost[ig,1:n]

  postinvscale <- priorinvscale + # prior contribution
    c( sum( tmp^2)/2,     # Qyy_st is the identity
      t(spost[ig,]) %*% (Qss %*% spost[ig,])/2,
      t(tpost[ig,]) %*% (Qtt %*% tpost[ig,])/2)


  kpost[ig,] <- rgamma(3, postshape, postinvscale)

  if( (ig%%10)==0) cat(".")

}

})


# POSTPROCESSING:
cat("\nTotal time:",timing[1],"per iteration:",timing[1]/totalg,"\n")


# Eliminate burn-in:
kpost <- kpost[-c(1:burnin),]
spost <- spost[-c(1:burnin),]
tpost <- tpost[-c(1:burnin),]

print(summary(kpost))

postquant <- apply(spost+tpost, 2, quantile,c(.025,.975))
postmean  <- apply(spost+tpost, 2, mean)
postmedi  <- apply(spost+tpost, 2, median)


######################################################################
# Figure 7:
par(mfcol=c(1,1),mai=c(.6,.8,.01,.01))

plot( y^2, ylim=c(800,2900),xlim=c(0,nm),ylab="Counts")
#lines( postmean^2, col=2)
lines( postmedi^2, col=2)
matlines( t(postquant)^2, col=4,lty=1)

legend("topright",legend=c("Posterior median", "Quantiles of posterior sample",
                    "Quantiles of predictive distribution"),
       bty="n",col=c(2,4,3),lty=1)



# Constructing a predictive distribution:
ypred <- rnorm( ngibbs*nm, c(spost+tpost),sd=rep( 1/sqrt(kpost[,1]), nm))
dim(ypred) <- c(ngibbs,nm)
postpredquant <- apply(ypred, 2, quantile,c(.025,.975))
matlines( t(postpredquant)^2, col=3,lty=1)
points(y^2)


kpostmedian <- apply(kpost,2,median)

par(mfcol=c(1,3),mai=c(.65,.65,.01,.01),cex=.85,mgp=c(2.6,1,0))

matplot( log( kpost), lty=1, type="l",xlab="Index")
abline(h=log(kpostmedian),col=3)
acf( kpost[,3],ylab=expression(kappa[t]))
plot(kpost[,2:3],ylab=expression(kappa[t]),xlab=expression(kappa[s]),cex=.8)
abline(h=kpostmedian[3],v=kpostmedian[2],col=3)


allkappas <- rbind(apply(kpost,2,mean),
                   apply(kpost,2,median),
                   apply(1/kpost,2,mean),
                   apply(1/kpost,2,median))
colnames(allkappas) <- c("kappa_y", "kappa_s", "kappa_t")
rownames(allkappas) <- c("Prec (mean)", "Prec (median)",
                         "Var (mean)", "Var (median) ")
print(allkappas,4)

# This is the MCMC sampler presented in Section 3.1 of the article:
#
# Florian Gerber, Reinhard Furrer (2015). Pitfalls in the Implementation
# of Bayesian Hierarchical Modeling of Areal Count Data: An Illustration
# Using BYM and Leroux Models. Journal of Statistical Software,
# Code Snippets, 63(1), 1-32. URL http://www.jstatsoft.org/v63/c01/.
#
# Note: For illustration we set
# number of generated samples: 5'000
# number of burnin samples:      500
# thinning:                       10
# This takes only 1-2 minutes of computation time.
#
# In the JSS article we used:
# number of generated samples: 300'000
# number of burnin samples:     15'000
# thinning:                         20

invisible(readline(prompt = "Type  <Return>\t to continue : "))

# SETUP:
library("spam")
options(spam.structurebased=TRUE)
library("truncdist")


# Lereux model - one block, no intercept

# load data
data(Oral)
E <- Oral$E
Y <- Oral$Y
n <- 544
A <- adjacency.landkreis(system.file("demodata/germany.adjacency", package="spam"))
A <- as.matrix(A)


# tuning parameters
# (influence acceptance rate)
lambda.proposal.sd <- 0.070992


# sampler length, burnin and thinning
totaln <- 5000
burnin <- 500
thin <- 10

# variable to store samples
upost <- array(NA, c(totaln, n))
kpost <- rep(NA,totaln)
lpost <- rep(NA,totaln)
accept <- array(0, c(totaln, 2),
                list(NULL, c("U", "lambda")))

# initial values
upost[1,] <- rep(c(.1,-.1), 544/2)
kpost[1] <- 15
lpost[1] <- .9
accept[1,] <- 1

# precaluclate some quantities
R <- precmat.IGMRFirreglat(A)
eigenR <- eigen(R); eigenR.value <- eigenR$values;

Q <- (1 - lpost[1]) * diag.spam(544) + lpost[1] * R
Q.det <- sum(log(lpost[1]* eigenR.value + 1 - lpost[1]))
Q.struct <- chol.spam(Q)

postshape <- 0.5 * n - 1

# start sampler
for (i in 2:totaln) {

    ## u
    ## find tilde u
    u.tilde <- upost[i-1,]
    C <- E * exp(u.tilde)
    B <- Y + (u.tilde - 1) * C
    Q.tmp <- diag.spam(C) + kpost[i-1] * Q
    u.tilde <- c(solve.spam(Q.tmp, B))

    C.tilde <- E * exp(u.tilde)
    B.tilde <- Y + (u.tilde - 1) * C.tilde
    Q.tilde <- diag.spam(C.tilde) + kpost[i-1] * Q

    u_ <- c(rmvnorm.canonical(1, B.tilde, Q.tilde, Rstruct = Q.struct))

    u.tilde_ <- u_
    C_ <- E * exp(u.tilde_)
    B_ <- Y + (u.tilde_- 1) * C_
    Q.tmp_ <- diag.spam(C_) + kpost[i-1] * Q
    u.tilde_ <- c(solve.spam(Q.tmp_, B_))

    C.tilde_ <- E * exp(u.tilde_)
    B.tilde_ <- Y + (u.tilde_- 1) * C.tilde_

    log.alpha.u <- sum(Y * u_) - sum(E*exp(u_))  -
      sum(Y * upost[i-1,]) + sum(E*exp(upost[i-1,]) ) +
        sum(upost[i-1,] * B.tilde_) -
          .5 * t(upost[i-1,]) %*% (diag(C.tilde_) %*% upost[i-1,]) -
            sum(u_* B.tilde) + .5 * t(u_) %*% (diag(C.tilde) %*% u_)

    if(exp(log.alpha.u) > runif(1)){
      upost[i,] <- u_
      accept[i,1] <- 1
    } else{
      upost[i,] <- upost[i-1,]
    }

    ## kappa
    kpost[i] <- rgamma(1, shape = postshape,
                       rate = .5 * upost[i,] %*% (Q %*% upost[i,]))

    ## lambda
    lambda_ <- rtrunc(n=1, spec="norm", a=0, b=1,  mean=lpost[i-1], sd=lambda.proposal.sd)
    Q_ <- (1 - lambda_) * diag.spam(544) + lambda_ * R
    Q.det_ <- sum(log(lambda_* eigenR.value + 1 - lambda_))


    alpha.lambda <- exp(.5 * (Q.det_ -
                              kpost[i]* upost[i,] %*% (Q_ %*% upost[i,]) -
                              Q.det +
                              kpost[i]* upost[i,] %*% (Q %*% upost[i,]) ))

    if(alpha.lambda > runif(1)){
      lpost[i] <- lambda_
      Q <- Q_
      Q.det <- Q.det_
      accept[i,2] <- 1
    }else{
      lpost[i] <- lpost[i-1]
    }

    # progress report
    if(i%%500 == 0) cat(paste(i / 50, "%\n", sep = "" ))

}


# remove burnin
kpost <- kpost[-seq(burnin)]
upost <- upost[-seq(burnin), ]
lpost <- lpost[-seq(burnin)]
accept <- accept[-seq(burnin),]

# thinning
index <- c(TRUE, rep(FALSE, (thin - 1)))
kpost <- kpost[index]
upost <- upost[index,]
lpost <- lpost[index]
accept <- accept[index,]

# acceptance rate
apply(accept, 2, mean)
par(mfrow = c(1,2))
plot(accept[,1], yaxt = "n", main = expression(beta))
axis(2, at = c(0,1), label = c("no", "yes"), las = 2)
plot(accept[,2], yaxt = "n", main = expression(lambda))
axis(2, at = c(0,1), label = c("no", "yes"), las = 2)

# trace and mixing plots for the precision parameters
# kappa and lambda
grid.newpage();
grid_trace2(log(kpost), lpost,
            chain1_lab = expression(log~kappa),
            chain2_lab = expression(lambda))

# summary statistics of
# kappa and lambda
summary(kpost)
summary(lpost)


par(mfrow = c(1,2))
## standardized mortality ratios
germany.plot(log(Y/E), main = "SMR")
## estimated relative log-risk
germany.plot(apply(upost, 2, mean), main = "U | Y, hyper-priors")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  # fig.width = 6, fig.height = 6,
  fig.align = "center"
)
options(digits = 3)

## ----loadpkgs, echo=FALSE, eval=TRUE, message=FALSE---------------------------
library("spam")

## ----echo=TRUE, eval=FALSE, message=FALSE-------------------------------------
# install.packages("spam")
# 
# library("spam")

## ----trivial------------------------------------------------------------------
Fmat <- matrix(c(3, 0, 1, 0, 2, 0, 1, 0, 3), nrow = 3, ncol = 3)
Smat <- as.spam(Fmat)

## ----operations---------------------------------------------------------------
Fmat
Smat
Smat %*% t(Smat)
Fmat %*% t(Smat)

## ----nonspam------------------------------------------------------------------
rep(1, 3) %*% Smat

## ----diffbehaviour------------------------------------------------------------
range(Fmat)
range(Smat)

## ----displ1-------------------------------------------------------------------
Smat

## ----displ2-------------------------------------------------------------------
diag.spam(100)

## ----disp3--------------------------------------------------------------------
str(Smat)

## ----disp4--------------------------------------------------------------------
summary(Smat)

## ----disp5, echo=FALSE, warning=FALSE-----------------------------------------
nz <- 2^12
Smat1 <- spam(0, nz, nz)
Smat1[cbind(sample(nz, nz), sample(nz, nz))] <- rnorm(nz)

tmp <- round(summary(Smat1)$density, 3)

## ----disp6, echo=FALSE, warning=FALSE, fig.cap = '\\label{fig:display_spam}Sparsity structure of sparse matrices.'----
par(mfcol=c(1,2),pty='s',mai=c(.8,.8,.2,.2))
display(Smat)
display(Smat1)

## -----------------------------------------------------------------------------
i <- c(2, 4, 4, 5, 5)
j <- c(1, 1, 2, 1, 3)

A <- spam(0, nrow = 5, ncol = 5)
A[cbind(i, j)] <- rep(0.5, length(i))
A <- t(A) + A + diag.spam(5)
A

U <- chol(x = A)

## ----tree, echo=FALSE, results = 'markup', out.width='40%', fig.show='hold', warning=FALSE, fig.cap='On the left side the associated graph to the matrix $\\boldsymbol{A}$ is visualized. The nodes of the graph are labeled according to $\\boldsymbol{A}$ (upright) and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ (italics). On the right side the sparsity structure of $\\boldsymbol{A}$ and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ (top row) and the Cholesky factors $\\boldsymbol{R}$ and $\\boldsymbol{U}$ of $\\boldsymbol{A}$ and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ respectively are given in the bottom row. The dashed lines in $\\boldsymbol{U}$ indicate the supernode partition.'----

knitr::include_graphics(c('figures/tree.png', 'figures/ill.png'))

## ----fillin, echo=FALSE, fig.show='hold', out.width='\\textwidth', warning=FALSE, fig.cap='Sparsity structure of the Cholesky factor with MMD, RCM and no permutation of a precision matrix induced by a second-order neighbor structure of the US counties. The values *nnzR* and *fillin* are the number of non-zero elements in the sparsity structure of the factor and the fill-in, respectively.'----

knitr::include_graphics(c('figures/fig_ch2_factors.png'))

# We illustrate the Cholesky factorization approaches
set.seed(14)

# first start with a full matrix.
xn <- 750
fmat1 <- matrix(rnorm(xn*xn),xn,xn)
fmat1 <- t( fmat1) %*% fmat1
smat1 <- as.spam(fmat1)
smat2 <- smat1 + diag.spam(xn)

# Generic Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, reusing pivoting
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, updating
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# reset to default
options(spam.cholsymmetrycheck=TRUE)

# now create a sparse matrix.
fmat1[fmat1<3] <- 0
smat1 <- as.spam(fmat1)
smat2 <- smat1 + diag.spam(xn)

# Generic Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, reusing pivoting
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, updating
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# reset to default
options(spam.cholsymmetrycheck=TRUE)

# now create an even sparser matrix.
fmat1 <- fmat1+20*diag(xn)
fmat1[fmat1<32] <- 0
smat1 <- as.spam(fmat1)
smat2 <- smat1 + 1* diag.spam(xn)

# Generic Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, reusing pivoting
options(spam.cholsymmetrycheck=FALSE)
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, updating
options(spam.cholsymmetrycheck=FALSE)
tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# reset to default
options(spam.cholsymmetrycheck=TRUE)

# This is the MCMC sampler presented in Section 2.2 of the article:
#
# Florian Gerber, Reinhard Furrer (2015). Pitfalls in the Implementation
# of Bayesian Hierarchical Modeling of Areal Count Data: An Illustration
# Using BYM and Leroux Models. Journal of Statistical Software,
# Code Snippets, 63(1), 1-32. URL http://www.jstatsoft.org/v63/c01/.
#
# Note: For illustration we set
# number of generated samples: 5'000
# number of burnin samples:      500
# thinning:                       10
# This takes 1-2 minutes of computation time.
#
# In the JSS article we used:
# number of generated samples: 300'000
# number of burnin samples:     15'000
# thinning:                         20

invisible(readline(prompt = "Type  <Return>\t to continue : "))

# SETUP:
library("spam")
options(spam.structurebased=TRUE)

# Besag-York-Molie model (BYM)
# load data
data(Oral); attach(Oral)
path <- system.file("demodata/germany.adjacency", package = "spam")
A <- adjacency.landkreis(path); n <- dim(A)[1]

set.seed(2)
# hyper priors
hyperA <- c(1, 1); hyperB <- c(0.5, .01)

# sampler length, burnin and thinning
totalg <- 5000
burnin <- 500
thin <- 10

# variable to store samples
upost <- vpost <- array(0, c(totalg, n))
kpost <- array(NA, c(totalg, 2)); accept <- rep(NA, totalg)

# initial values
upost[1,] <- vpost[1,] <- rep(.001, 544); kpost[1,] <- c(10, 100)

# precalculate some quantities
eta <- upost[1,] + vpost[1,]
C <- exp(eta) * E; diagC <- diag.spam(c(rep(0, n), C))
b <- c( rep(0, n), Y + (eta - 1) * C)
Qu <- R <- precmat.IGMRFirreglat(A); pad(Qu) <- c(2 * n, 2 * n)
Qv <- as.spam(rbind(cbind( diag(n), -diag(n)),
                      cbind(-diag(n),  diag(n))))
Q <- kpost[1,1] * Qu + kpost[1,2] * Qv + diagC

# store symbolic cholesky factorization of Q
struct <- chol(Q, memory = list(nnzcolindices = 6467))

uRuHalf <- t(upost[1,]) %*% (R %*% upost[1,]) / 2
vvHalf <- t(vpost[1,]) %*% vpost[1,] / 2
postshape <- hyperA + c(n - 1, n) / 2

for (i in 2:totalg) {
    # update precision
    kpost[i,] <- rgamma(2, postshape, hyperB + c(uRuHalf, vvHalf))

    # find eta tilde
    etaTilde <- eta
    for(index in 1:2){
        C <- E * exp(etaTilde)
        diagC <- diag.spam(c(rep(0, n), C))
        b <- c(rep(0, 544), Y + (etaTilde - 1) * C)
        Q <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC
        etaTilde <- c(solve.spam(Q, b,
                                 Rstruct = struct))[1:n + n]
    }

    C <- exp(etaTilde) * E; diagC <- diag.spam(c(rep(0, n), C))
    b <- c( rep(0, n), Y + (etaTilde - 1) * C)
    Q <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC

    # simulate proposal
    x_ <- c(rmvnorm.canonical(1, b, Q, Rstruct = struct))
    upost[i,] <- x_[1:n]; eta_ <- x_[1:n + n]; vpost[i,] <- eta_ - upost[i,]
    uRuHalf_ <- t(upost[i,]) %*% (R %*% upost[i,]) / 2
    vvHalf_ <- t(vpost[i,]) %*% vpost[i,] / 2

    # calculate acceptance probability
    etaTilde_ <- eta_
    for(index in 1:2){
       C_ <- E * exp(etaTilde_)
       diagC_ <- diag.spam(c(rep(0, n), C_))
       b_ <- c(rep(0, 544), Y + (etaTilde_ - 1) * C_)
       Q_<- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC_
       etaTilde_ <- c(solve.spam(Q_, b_,
                                 Rstruct = struct))[1:n + n]
    }

    C_ <- exp(etaTilde_) * E; diagC_ <- diag.spam(c(rep(0, n), C_))
    b_ <- c( rep(0, n), Y + (etaTilde_ - 1) * C_)
    Q_ <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC_

    logPost_ <- sum(Y * eta_ - E * exp(eta_)) -
        kpost[i,1] * uRuHalf_ - kpost[i, 2] * vvHalf_
    logPost  <- sum(Y * eta - E * exp(eta)) - kpost[i,1] * uRuHalf -
        kpost[i,2] * vvHalf
    logApproxX_ <- - kpost[i,1] * uRuHalf_ - kpost[i,2] * vvHalf_ -
        sum(.5 * eta_^2 * C) + sum(b * eta_)
    logApproxX  <- - kpost[i,1] * uRuHalf  - kpost[i,2] * vvHalf -
        sum(.5 * eta^2 * C_) + sum(b_ * eta)
    logAlpha <- min(0, logPost_ - logPost + logApproxX - logApproxX_)

    # accept or reject proposal
    if (log(runif(1)) < logAlpha) {
        uRuHalf <- uRuHalf_;  vvHalf <- vvHalf_
        eta <- eta_; b <- b_; C <- C_; accept[i] <- 1
    } else{
        accept[i] <- 0; upost[i,] <- upost[i-1,]; vpost[i,] <- vpost[i-1,]
    }

    # progress report
    if(i%%500 == 0) cat(paste(i / 50, "%\n", sep = "" ))
}



## remove burnin
kpost <- kpost[-seq(burnin), ]
upost <- upost[-seq(burnin), ]
vpost <- vpost[-seq(burnin), ]
accept <- accept[-seq(burnin)]

## thinning
index <- c(TRUE, rep(FALSE, (thin - 1)))
kpost <- kpost[index,]
upost <- upost[index,]
vpost <- vpost[index,]
accept <- accept[index]

## acceptance rate
mean(accept)
plot(accept, yaxt = "n")
axis(2, at = c(0,1), label = c("no", "yes"), las = 2)

## trace and mixing plots for the precision parameters
## kappa_u and kappa_v
grid.newpage()
grid_trace2(kpost, chain1_lab = expression(log~kappa[u]),
            chain2_lab = expression(log~kappa[v]), log = TRUE)

## summary statistics of
## kappa_u and kappa_v
apply(kpost, 2, summary)


par(mfrow = c(1,2))
## standardized mortality ratios
germany.plot(log(Y/E), main = "SMR")
## estimated relative log-risk
germany.plot(apply(upost, 2, mean), main = "U | Y, hyper-priors")

# This is a simple demo, illustrating the functionality of spam.
set.seed(14)

nrow <- 5
ncol <- 7
fmat <- matrix(rnorm(nrow*ncol),nrow)
smat <- as.spam(fmat)


smat[1,]
smat[,1] <- 0
as.spam(smat)


ssmat <- smat %*% t(smat)

b <- c(-2:2)

solve(ssmat,b)

cholssmat <- chol(ssmat)


# works also for large matrices:
set.seed(14)

nz <- 1000
nrow <- 1000
ncol <- 1000
smat <- diag.spam(1,nrow,ncol)

smat[cbind(sample(1:(nrow*ncol),size=nz))] <- runif(nz)

smat <- smat %*% t(smat)


b <- rnorm(nz)

smatinvb <- solve(smat,b)

cholssmat <- chol(smat)


# displaying matrices
opar <- par(no.readonly = TRUE)
par(ask=interactive() &&
            (.Device %in% c("X11","GTK","gnome","windows","quartz")))

display(smat, main="'scatterplot'-type display, very efficient")


options(spam.imagesize=prod(smat@dimension)+1)
display(smat, main="'image'-type display, may be slow and heavy")

par(opar)

# We construct a few large matrices and we compare how much faster (slower)
# we are compared to the full matrix analysis.
# Since all the calculation are also done with full matrices, we do not
# exaggerate with the sizes.

set.seed(14)

compare <- function(expr1,expr2,tag=NULL)
  {
    if( !is.null(tag)) cat( "Comparing: ", tag, fill=TRUE)
    print(data.frame(full=system.time( expr1, TRUE)[1:3],
                     sparse=system.time( expr2, TRUE)[1:3],
                     row.names=c("user","system","elapsed")))
  }

xn <- 1000
xm <- 1200

# first start with a full matrix.
fmat1 <- matrix(rnorm(xn*xm),xn,xm)
smat1 <- as.spam(fmat1)

compare(fmat2 <- t(fmat1), smat2 <- t(smat1), "Transpose")

compare(ffmat <- fmat1 %*% fmat2,
        ssmat <- smat1 %*% smat2, "multiplication")

compare( solve(ffmat),  solve(ssmat), "solving")


compare(rbind(fmat1,fmat1),rbind(smat1,smat1))
compare(cbind(fmat1,fmat1),cbind(smat1,smat1))


# now create a sparse matrix.
fmat1[fmat1<3] <- 0
smat1 <- as.spam(fmat1)

compare(fmat2 <- t(fmat1), smat2 <- t(smat1), "Transpose")

compare(ffmat <- fmat1 %*% fmat2,
        ssmat <- smat1 %*% smat2, "multiplication")

compare(ffmat <- ffmat + diag(xn),
        ssmat <- ssmat + diag.spam(xn), "add identity")

compare(ffmat <- 1:xn %d+% ffmat,
        ssmat <- 1:xn %d+% ssmat, "add identity quicker")

compare( solve(ffmat),  solve(ssmat), "solving")

summary(ssmat)


# compare a few cbind/rbinds

compare(rbind(fmat1,fmat1),rbind(smat1,smat1))
compare(cbind(fmat1,fmat1),cbind(smat1,smat1))

# This demo contains the R code to construct the figures and the table of the
# JSS article:
#     "spam: A Sparse Matrix R Package with Emphasis on
#            MCMC Methods for Gaussian Markov Random Fields"

# The code presented here differs in the following points form the actually used
# one:
# - Very large grid sizes or very high order neighbor structures are not included
#   here;
# - Instead of (100+1) factorization only (50+1) are performed here;
# - No figure fine-tuning is done here.
# - We had a few additional gc(), just to be sure.
# - Minor change due to evolved of 'spam'

# SETUP:
options(spam.structurebased=TRUE)  # just to be sure

######################################################################
# Figure 1:
i <- c( 2,4,4,5,5)
j <- c( 1,1,2,1,3)

A <- spam(0,5,5)
A[cbind(i,j)] <- rep(.5, length(i))
A <- t( A)+A+diag.spam(5)

U <- chol( A)
pivot <- U@pivot
B <- A[pivot,pivot]
R <- chol( B)

U@pivot
U@snmember
U@supernodes

U@entries
U@colindices
U@colpointers
U@rowpointers


display( A)
display( as.spam( chol(as.matrix( A))))
display( B)
display( as.spam(R))
abline( h=-U@supernodes+.5,col=3,lty=2)

######################################################################
# Figure 2:
theta1 <- .1
theta2 <- .01
n <- dim( UScounties.storder)[1]

USmat <- diag.spam(n) + theta1 *  UScounties.storder + theta2 *  UScounties.ndorder


U <- chol( USmat,memory=list(nnzR=146735))
display( as.spam(U))
text(400,-2200,"MMD\nz=146735\nw=30182\ns=1262",adj=0)


U <- chol( USmat, pivot="RCM",memory=list(nnzR=256198,nnzcolindices=140960))
display( as.spam(U))
text(400,-2200,"RCM\nz=256198\nw=140960\ns=1706",adj=0)


U <- chol( USmat, pivot=FALSE,memory=list(nnzR=689615,nnzcolindices=96463))
display( as.spam(U))
text(400,-2200,"no permutation\nz=689615\nw=96463\ns=711",adj=0)


######################################################################
# Figure 3:

# general parameters for the following figures
N <- 50         # would be 100 in the article
stsel <- 1      # user.self
rPsx <- 1       # for function "system.time"
rPsy <- 3       # memory usage
rPint <- .0001  # small interval


theta1 <- .1
theta2 <- .05

xseq <- ceiling(4 + exp(seq(0.5,to=5.5,by=.5))/2)  # would be seq(0,to=6,by=.5) in the article
xseql <- length(xseq)

table <- array(NA,c(xseql,4))
for (ix in 1:xseql) {

  egdx <- expand.grid(1:xseq[ix],1:xseq[ix])
  Cspam <- nearest.dist( egdx, delta=1., upper=NULL)
  Dspam <- nearest.dist( egdx, delta=1.5,upper=NULL)

  mat <- diag.spam(xseq[ix]^2) + theta1 * Cspam + theta2 * Dspam

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[ix,1] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch1 <- chol(mat)}
                             )[stsel]
  Rprof( NULL)
  table[ix,2] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]


  Rprof( memory.profiling=TRUE, interval = rPint)
  table[ix,3] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch2 <- update(ch1,mat) }
                             )[stsel]
  Rprof( NULL)
  table[ix,4] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

}

# Since we have a small N, elements in table might be zero.
table <- pmax(table, 0.0001)

par(mfcol=c(1,2))
plot(xseq, table[,1], type="l", log="xy", ylim=range(table[,c(1,3)]),
     xlab="L (log scale)", ylab="seconds (log scale)")
lines(xseq, table[,3], lty=2)
lines(xseq,table[,1]/table[,3],col=4,lty=3)

plot(xseq, table[,2], type="l", log="xy", ylim=range(table[,c(2,4)]+0.01),
     xlab="L (log scale)", ylab="Mbytes (log scale)")
lines(xseq, table[,4], lty=2)
lines(xseq,table[,2]/table[,4],col=4,lty=3)


######################################################################
# Figure 4:

# general parameters for the following figures
N <- 50         # would be 100 in the article
stsel <- 1      # user.self
rPsx <- 1       # for function "system.time"
rPsy <- 3       # memory usage
rPint <- .0001  # small interval

x <- 50     # was 50 in article
maxnn <- 6  # was 6 in article

egdx <- expand.grid( 1:(maxnn+1), 1:(maxnn+1))
dval <- sort(unique(nearest.dist( egdx, delta=maxnn)@entries))[-1]
dvall <- length( dval)


egdx <- expand.grid( 1:x, 1:x)


table <- array(NA, c(dvall,5))

for (id in 1:dvall) {

  mat <- nearest.dist( egdx, delta=dval[id],upper=NULL)
  mat@entries <- exp(-2*mat@entries)         # arbitrary values to get a spd precision matrix

  table[id,5] <- length(Cspam)

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[id,1] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch1 <- chol(mat)}
                             )[stsel]
  Rprof( NULL)
  table[id,2] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[id,3] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch2 <- update(ch1,mat) }
                             )[stsel]
  Rprof( NULL)
  table[id,4] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

}

# If we have a small N, elements in table might be zero.
table <- pmax(table, 0.0001)

par(mfcol=c(1,2))
plot( dval, table[,1], type="l", log="xy",ylim=range(table[,c(1,3)]),
     xlab="distance (log scale)", ylab="seconds (log scale)")
lines( dval, table[,3],lty=2)
lines( dval, table[,1]/table[,3],col=4,lty=3)

plot( dval, table[,2], type="l", log="xy",ylim=range(table[,c(2,4)]),
     xlab="distance (log scale)", ylab="Mbytes (log scale)")
lines( dval, table[,4],lty=2)
lines( dval, table[,2]/table[,4],col=4,lty=3)


######################################################################
# Figure 5
In <- diag.spam(nrow(UScounties.storder))
struct <- chol(In + .2 * UScounties.storder + .1 * UScounties.ndorder)

len.1 <- 90 # in the article, is set to 180
len.2 <- 50 # in the article, is set to 100
theta.1 <- seq(-.225, to=.515, len=len.1)
theta.2 <- seq(-.09, to=.235, len=len.2)

grid <- array(NA, c(len.1, len.2))
options(spam.cholupdatesingular="null")

for (i in 1:len.1)
  for(j in 1:len.2)
    grid[i,j] <- !is.null(update(struct, In + theta.1[i]*UScounties.storder
                       + theta.2[j]* UScounties.ndorder))

image(theta.1, theta.2, grid, xlab=expression(theta[1]), ylab=expression(theta[2]),
      xlim=c(-.3,.6),ylim=c(-.1,.25),col=c(0,"gray"))
abline(v=0,h=0, lty=2)


######################################################################
# Table 1:
table <- array(NA,c(9,4))

x <- 50    #  was 50 in article
egdx <- expand.grid(1:x,1:x)

# As above hence shortend
gridmat <- diag.spam(x^2) + .2 * nearest.dist( egdx, delta=1.,upper=NULL) +
  .1 * nearest.dist( egdx, delta=1.5,upper=NULL)
# USmat was constructed above.


# Generic call first:
Rprof( memory.profiling=TRUE, interval = rPint)
table[1,1] <- system.time(   for (i in 1:N) ch1 <- chol(gridmat) )[stsel]
Rprof( NULL)
table[1,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[1,3] <- system.time(   for (i in 1:N) ch2 <- chol(USmat)   )[stsel]
Rprof( NULL)
table[1,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Call a chol.spam directly
Rprof( memory.profiling=TRUE, interval = rPint)
table[2,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat))[stsel]
Rprof( NULL)
table[2,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[2,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat)  )[stsel]
Rprof( NULL)
table[2,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Less checking:
options(spam.safemode=c(FALSE, FALSE, FALSE))
Rprof( memory.profiling=TRUE, interval = rPint)
table[3,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat)    )[stsel]
Rprof( NULL)
table[3,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[3,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat)    )[stsel]
Rprof( NULL)
table[3,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
options(spam.safemode=c(TRUE, TRUE, TRUE))

# lesser checking
options(spam.cholsymmetrycheck=FALSE)
Rprof( memory.profiling=TRUE, interval = rPint)
table[4,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat)    )[stsel]
Rprof( NULL)
table[4,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[4,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat)    )[stsel]
Rprof( NULL)
table[4,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
options(spam.cholsymmetrycheck=TRUE)

# Pass optimal memory parameters (from above)
memory1 = summary(ch1)[1:2]
memory2 = summary(ch2)[1:2]
Rprof( memory.profiling=TRUE, interval = rPint)
table[5,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat,memory=memory1)    )[stsel]
Rprof( NULL)
table[5,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[5,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat,memory=memory2)    )[stsel]
Rprof( NULL)
table[5,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]


# All of the above
options(spam.cholsymmetrycheck=FALSE, safemode=c(FALSE,FALSE,FALSE))
Rprof( memory.profiling=TRUE, interval = rPint)
table[6,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,memory=memory1)    )[stsel]
Rprof( NULL)
table[6,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[6,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat,memory=memory2)    )[stsel]
Rprof( NULL)
table[6,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# supply the permutation
pivot1 <- ch1@pivot
pivot2 <- ch2@pivot
Rprof( memory.profiling=TRUE, interval = rPint)
table[7,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,pivot=pivot1,
                                                             memory=memory1)    )[stsel]
Rprof( NULL)
table[7,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[7,3] <- system.time(   for (i in 1:N) ch1 <- chol.spam(USmat,pivot=pivot2,
                                                             memory=memory2)    )[stsel]
Rprof( NULL)
table[7,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Do not check the permutation
options(spam.cholpivotcheck=FALSE)
Rprof( memory.profiling=TRUE, interval = rPint)
table[8,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,pivot=pivot1,
                                                              memory=memory1)    )[stsel]
Rprof( NULL)
table[8,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[8,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat,pivot=pivot2,
                                                              memory=memory2)    )[stsel]
Rprof( NULL)
table[8,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Update only
Rprof( memory.profiling=TRUE, interval = rPint)
table[9,1] <- system.time(   for (i in 1:N) ch1 <- update(ch1,gridmat)   )[stsel]
Rprof( NULL)
table[9,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[9,3] <- system.time(   for (i in 1:N) ch2 <- update(ch2,USmat)   )[stsel]
Rprof( NULL)
table[9,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]


# assemble the table
colnames(table) <- c("grid_time","grid_mem","US_time","US_mem")
rownames(table) <- c("Generic chol","chol.spam","safemode",
                     "symmetrycheck","memory","all","reusing pivot","best cast","update only")


normed.table <- t( round( t(table[-1,])/table[1,],3))
print( t( round( t(table[-1,])/table[1,],3)))

# HEADER ####################################################
# This is file spam/R/norm.R.                               #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

     


########################################################################

#norm <- function(x, type = "sup", ...){
#  typ <- charmatch(tolower(type), c("sup","l1","frobenius","hs"))
#  if (is.na(typ))          stop("undefined norm "",type,"".",call.=FALSE)

#  switch(typ,
#         max(abs(x)),
#         sum(abs(x)),
#         sqrt(sum(x^2)),sqrt(sum(x^2))
#         )
#}
norm.spam <- function(x, type = "m", ...){
  typ <- substr(tolower(type),1,1)

  if (typ %in% c("o", "1")) {
    return( max( colSums(abs(x))))
  }
  if (typ %in% c("i")) {
    return( max( rowSums(abs(x))))
  }
  if (typ %in% c("f", "h")) {
    return( sqrt(sum(x@entries^2)))
  }
  if (typ %in% c("m","s")) {
    return( max(abs(x@entries)) )
  }
  
  stop("undefined norm '",type,"'.",call.=FALSE)

}

setMethod("norm",signature(x="spam",type="character"), 
          function(x, type, ...) norm.spam(x, type))

setMethod("norm",signature(x="spam",type="missing"),
          function(x, type, ...) norm.spam(x, type="O"))

setMethod("norm", signature(x = "numeric", type = "character"),
          function(x, type, ...) base::norm(as.matrix(x), type))

setMethod("norm", signature(x = "numeric", type = "missing"),
          function(x, type, ...) base::norm(as.matrix(x),  type="O"))

setMethod("norm", signature(x = "matrix", type = "character"),
	  function(x, type, ...) base::norm(x, type))
setMethod("norm", signature(x = "matrix", type = "missing"),
	  function(x, type, ...) base::norm(x, type="o"))

# HEADER ####################################################
# This is file spam/R/s3only.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

var.spam <- function(x, ...) {
    inefficiencywarning( "This 'var' operation may be inefficient", prod(dim(x)))
    var(as.matrix(x), ...)
}

# HEADER ####################################################
# This is file spam/tests/demo_article-jss-example2.R.      #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
     


# INITALIZE AND FUNCTIONS:
# JSS article:
#     "spam: A Sparse Matrix R Package with Emphasis on
#            MCMC Methods for Gaussian Markov Random Fields"


# Compared to the R code in the article, here we give:
# - improved formatting
# - more comments, e.g. how to run the code using regular matrices 
# - the code to construct the figures
# - minor modifcations due to evolvement of spam
 

cat("\nThis demo contains the R code of the second example\nin the JSS article. As pointed out by Steve Geinitz\nand Andrea Riebler, the Gibbs sampler is not correct\nand contains several bugs. \n\nI'll post an updated sampler in a future release.\n\n") 


# INITALIZE AND FUNCTIONS:
require("fields", warn.conflict=FALSE)
options(spam.structurebased=TRUE)


# READ DATA:
attach(Oral)



# CONSTRUCT ADJACENCY MATRIX:
loc <- system.file("demodata/germany.adjacency", package="spam")
A <- adjacency.landkreis(loc)
n <- dim(A)[1]
# Verification that we have a symmetric matrix:
# norm(A-t(A)); display(A)


# GIBBS SETUP:
set.seed(14)

# Construct the individual block precisions
# (based on unit precision parameters kappa, denoted with k):

Q1 <- R <- diag.spam( diff(A@rowpointers)) - A   # this is R in (2)
pad(Q1) <- c(2*n,2*n)  # previously:  dim(Q1) <- c(2*n,2*n)

Q2 <-  rbind(cbind( diag.spam(n), -diag.spam(n)),
       	     cbind(-diag.spam(n),  diag.spam(n)))

# Hence the precision Q in (2) is:
# Q <- kappau*Q1 + kappav*Q2

# pre-define
diagC <- as.spam( diag.spam(c(rep(0,n),rep(1,n))))


# Recall:
# k=( kappa_u, kappa_y)'

# hyperparameters
ahyper <- c( 1, 1)
bhyper <- c( .5, .01)


# Gibbs sampler
burnin <- 50
ngibbs <- 150
totalg <- burnin+ngibbs

# Initialize parameters:
upost <- array(0, c(totalg, n))
npost <- array(0, c(totalg, n))
kpost <- array(0, c(totalg, 2)) 

# Starting values:
kpost[1,] <- c(40,500)
upost[1,] <- u <- rnorm(n,sd=.2) *1
npost[1,] <- eta <- u + rnorm(n,sd=.05)*1

uRu <- t(u) %*% (R %*% u)/2
etauetau <- t(eta-u) %*% (eta-u)/2 

postshape <- ahyper + c(n-1,n)/2

accept <- numeric(totalg)

struct <- chol(Q1 + Q2 + diag.spam(2*n),
               memory=list(nnzcolindices=5500))

# struct <- NULL        # If no update steps are wanted

# R <- as.matrix(R)     # If no spam analysis is wanted.
# Q1 <- as.matrix(Q1)
# Q2 <- as.matrix(Q2)


for (ig in 2:totalg) {

  
  kstar <- rgamma(2,postshape, bhyper + c(uRu, etauetau))	

  
  expeta0E <- exp(eta)*E
  expeta0Eeta01 <- expeta0E *(eta-1)
  diagC@entries <- expeta0E
  Q <- kstar[1]*Q1 + kstar[2]*Q2 + diagC
  b <- c( rep(0,n), Y + expeta0Eeta01)

  xstar <- rmvnorm.canonical(1,
                             # vector b:
                             b,
                             # Precision matrix
                             Q,
                             Rstruct=struct)
  

  ustar <- xstar[1:n]
  nstar <- xstar[1:n+n]

  uRustar <- t(ustar) %*% (R %*% ustar)/2
  etauetaustar <- t(nstar-ustar) %*% (nstar-ustar)/2

  
# we work on the log scale:
# logalpha <- min(0, log(ratios))=min(0, expterm+(...)log(kappa)-
  
  exptmp <- sum(expeta0Eeta01*(eta-nstar) - E*(exp(eta)-exp(nstar))) -
    sum( nstar^2*expeta0E)/2   +    sum(eta^2*expeta0E)/2 -
      kstar[1] * uRu           +    kpost[ig-1,1] * uRustar -
        kstar[2] * etauetau    +    kpost[ig-1,2] * etauetaustar
  factmp <- (postshape-1)*(log(kstar)-log(kpost[ig-1,1]))
  
  logalpha <- min(0, exptmp + sum(factmp))
  logU <- log(runif(1))

  if (logU < logalpha) { # ACCEPT draw
    upost[ig,] <- u   <- ustar
    npost[ig,] <- eta <- nstar
    kpost[ig,] <- kstar
    uRu <- uRustar
    etauetau <- etauetaustar
    accept[ig] <- 1
  } else {
    upost[ig,] <- upost[ig-1,]
    npost[ig,] <- npost[ig-1,]
    kpost[ig,] <- kpost[ig-1,]    
  }
                   
  if( (ig%%10)==0) cat('.')

}


if (FALSE) {

# POSTPROCESSING:

accept <- accept[-c(1:burnin)]
cat("\nAcceptance rate:",mean(accept),"\n")

kpost <- kpost[-c(1:burnin),]
upost <- upost[-c(1:burnin),]
npost <- npost[-c(1:burnin),]

kpostmean <- apply(kpost,2,mean)
upostmean <- apply(upost,2,mean)
npostmean <- apply(npost,2,mean)

kpostmedian <- apply(kpost,2,median)
upostmedian <- apply(upost,2,median)
npostmedian <- apply(npost,2,median)

vpost <- npost-upost
vpostmedian <- apply(vpost,2,median)


#





######################################################################
# Figures 
par(mfcol=c(1,3),mai=rep(0,4))
map.landkreis(log(Y))
map.landkreis(Y/E,zlim=c(.1,2.4))
map.landkreis(exp(upostmedian),zlim=c(.1,2.4))


par(mfcol=c(2,4),mai=c(.5,.5,.05,.1),mgp=c(2.3,.8,0))
hist(kpost[,1],main="",xlab=expression(kappa[u]),prob=TRUE)
lines(density(kpost[,1]),col=2)
tmp <- seq(0,to=max(kpost[,1]),l=500)
lines(tmp,dgamma(tmp,ahyper[1],bhyper[1]),col=4)
abline(v=kpostmedian[1],col=3)

hist(kpost[,2],main="",xlab=expression(kappa[y]),prob=TRUE)
lines(density(kpost[,2]),col=2)
tmp <- seq(0,to=max(kpost[,2]),l=500)
lines(tmp,dgamma(tmp,ahyper[2],bhyper[2]),col=4)
abline(v=kpostmedian[2],col=3)

# Trace plots:
plot(kpost[,1],ylab=expression(kappa[u]),type="l")
abline(h=kpostmedian[1],col=3)
plot(kpost[,2],ylab=expression(kappa[y]),type="l")
abline(h=kpostmedian[2],col=3)

# ACF:
acf(kpost[,1],ylab=expression(kappa[u]))
acf(kpost[,2],ylab=expression(kappa[y]))



# scatter plots
plot(kpost[,1],kpost[,2],xlab=expression(kappa[u]),ylab=expression(kappa[y]))
abline(v=kpostmedian[1],h=kpostmedian[2],col=3)


plot(accept+rnorm(ngibbs,sd=.05),pch=".",ylim=c(-1,2),yaxt="n",ylab="")
text(ngibbs/2,1/2,paste("Acceptance rate:",round(mean(accept),3)))
axis(2,at=c(0,1),label=c("Reject","Accept"))

}

detach(Oral)
######################################################################

# HEADER ####################################################
# This is file spam/R/dist.R.                               #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


### in base:
# dist(x, method = "euclidean", diag = FALSE, upper = FALSE, p=2)
#
### in fields
# rdist( x1, x2)
# rdist.earth(x1, x2, miles = TRUE, R = NULL)
# fields.rdist.near( x1, x2, delta, max.points= NULL)
#
### in sp
# spDistsN1(pts, pt, longlat=FALSE)
#
### in amap     ### nbproc  integer, Number of subprocess for parallelization
# Dist(x, method = "euclidean", nbproc = 1, diag = FALSE, upper = FALSE)
#
### in argosfilter
# distance(lat1, lat2, lon1, lon2)  # gc between two pts in km
# distanceTrack(lat,lon)            # gc between pts  in km
#
### in proxy
# dist(x, y = NULL, method = NULL, ..., diag = FALSE, upper = FALSE,
#     pairwise = FALSE, by_rows = TRUE, convert_similarities = TRUE,
#     auto_convert_data_frames = TRUE)
#
### in RFOC
#  GreatDist(LON1, LAT1, LON2, LAT2, EARTHRAD= 6371)
#

nearest.dist <- function( x, y=NULL, method = "euclidean",
                         delta = 1,
                         upper = if(is.null(y)) FALSE else NULL,
                         p = 2, miles=TRUE, R=NULL, fortran = FALSE
#                         eps =  NULL, diag = NULL
                         )
{
  # see help for exact parameter meaning

  # We always include all small distances. Hence, this function
  #   works differently than any other spam functions. An additional
  #   call to an as.spam would eliminate the small values.
#  if (!is.null(diag)) warning("Argument "diag" is deprecated")
#  if (!is.null(eps))  warning("Argument "eps" is deprecated")

  if (!is.na(pmatch(method, "euclidian")))     method <- "euclidean"
  METHODS <- c("euclidean", "maximum", "minkowski", "greatcircle")
  method <- pmatch(method, METHODS)  # result is integer

  if (is.na(method))     stop("invalid distance method")

  if (method == 4) {
    if (is.null(R))
      p <- ifelse( miles,3963.34,6378.388)
    else {
      if (R <= 0)           stop("'R' should be postiive")
      p <- R
    }
    if (abs(delta)>180.1)  stop("'delta' should be smaller than 180 degrees.")
  }

  if (is.null(upper))
    part <- 0L
  else
    part <- ifelse(upper, 1L ,-1L)
  if (is.data.frame(x))  x <- as.matrix(x)
  if (is.list(x)) stop("'x' should be an array or matrix")
  # as.matrix( list() ) does not work
  if (!is.matrix(x)) x <- as.matrix(x)
  nd <- dim(x)[2]
  n1 <- dim(x)[1]

  if (!is.null(y)) {
    # we specify x and y:
    if (is.data.frame(y))  y <- as.matrix(y)
    if (is.list(x)) stop("'x' should be an array or matrix")
    if (!is.matrix(y)) y <- as.matrix(y)
    if (nd!=dim(y)[2]) stop("'x' and 'y' should have the same number of columns.")
    n2 <- dim(y)[1]
    sizemax <- n2*n1
    # there is an as.double just in case that mi (and n1 below) is > 2^16
  } else {
    # x = y, i.e. proper distance matrix
    if (n1==1)         stop("More than a single point in 'x' is required.")
    if (method == 4) {
      p <- -p  # we save one variable...
    }
    y <- x
    n2 <- n1
    sizemax <- n1*(n1+1)/2
  }

  # which code should be used
  force64 <- getOption("spam.force64")
  if (force64 || fortran || sizemax > 2147483647 - 1 ){
    if (!is.null(y)) {
      mi <- min(n1,n2)
      ma <- max(n1,n2)
      nnz <- min(max(getOption("spam.nearestdistnnz")[1],
                     ma*getOption("spam.nearestdistnnz")[2]),
             (as.double(mi)*(mi+1)+(ma-mi)^2)/ ifelse( is.null(upper), 1, 2))
      } else {
        nnz  <- min(max(getOption("spam.nearestdistnnz")[1],
                        ma*getOption("spam.nearestdistnnz")[2]),
                  (as.double(n1)*(n1+1))/ ifelse( is.null(upper), 1, 2))
      }
   # EXPLICIT-STORAGE-FORMAT
    if(max(n1,n2,nnz) > 2147483647 - 1 || force64)
      SS <- .format64()
    else
      SS <- .format32

    repeat {
      d <- NULL # Free the memory allocated by a previous attempt
      d <- .C64("closestdist",
                ##              subroutine closestdist( ncol, x,nrowx, y, nrowy,
                ##  &    part, p, method,
                ##  &    eta, colindices, rowpointers, entries, nnz, iflag)
                SIGNATURE=c(SS$signature, "double", SS$signature, "double", SS$signature,
                            SS$signature, "double", SS$signature, "double", SS$signature,
                            SS$signature, "double", SS$signature, SS$signature),

                nd,
                x,
                n1, #w
                y,

                n2, #w
                part, #arg 6
                p[1],
                method,

                abs( delta[1]),
                colindices=vector(SS$type, nnz),
                rowpointers=vector(SS$type, n1+1), #arg 11
                entries=vector("double",nnz),

                nnz=nnz,
                iflag=0,

                INTENT = c("r", "r", "r", "r",
                           "r", "r", "rw", "r",
                           "r", "w", "w", "w",
                           "rw", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)

     if (d$iflag==0) break else {
       nnz <-  nnz*getOption("spam.nearestdistincreasefactor")*n1/(d$iflag-1)

        # EXPLICIT-STORAGE-FORMAT
        if(max(n1,n2,nnz) > 2147483647 - 1 || force64)
         SS <- .format64()
        else
          SS <- .format32

      # if(nnz > 2147483647) stop("Distance matrix is too dense (2)")

        madens <- d$iflag
        warning(paste("You ask for a 'dense' sparse distance matrix, I require one more iteration.",
                            "\nTo avoid the iteration, increase 'nearestdistnnz' option to something like\n",
                            "'options(spam.nearestdistnnz=c(",d$nnz,",500))'\n(constructed ",madens,
                            " lines out of ",n1,").\n",sep=""), call. = TRUE)

      }
    }

   length(d$entries) <- d$nnz
   length(d$colindices) <- d$nnz
  } else {
    d <- closestdistCpp(x,y,part,p,method,eta=delta)
  }

  if(length(d$entries) == 0) {
    return(.newSpam(
      dimension=c(n1,n2),
      force64 = force64
    ))
  }

  return(.newSpam(
    entries=d$entries,
    colindices=d$colindices,
    rowpointers=d$rowpointers,
    dimension=c(n1,n2),
    force64 = force64
  ))
}

# in fields:
# rdist <- function (x1, x2)

spam_rdist <- function(x1, x2, delta = 1)
     nearest.dist(x1, y=x2,   delta = delta,  upper = NULL)

# in fields:
# rdist.earth <- function (x1, x2, miles = TRUE, R = NULL)
spam_rdist.earth <- function(x1, x2, delta=1, miles = TRUE, R = NULL)
    nearest.dist( x1, y=x2, method = "greatcircle",
                         delta = delta, miles=miles, R=R,  upper = NULL)

# HEADER ####################################################
# This is file spam/tests/testthat/test-permutation.R.      #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())     
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-permutation.R")


########


set.seed(14)
res <- 12.5


grid <- expand.grid(lat=seq(-90+3*res/2,to=90-res,by=res),lon=seq(res/2,to=360,by=res))
dist <- nearest.dist(grid[,2:1],method='gr',upper=NULL, delta=30,R=1)

distm <- as.matrix(dist)

n <- dim(dist)[1]
perm <- sample.int(n,n)

test_that("permutation.spam", {
    spamtest_eq(permutation.spam(dist,P=perm),distm[order(perm),])
    spamtest_eq(permutation.spam(dist,Q=perm),distm[,order(perm)])
    spamtest_eq(permutation.spam(dist,P=perm,ind=T),distm[perm,])
    spamtest_eq(permutation.spam(dist,Q=perm,ind=T),distm[,perm])
    
    spamtest_eq(permutation(dist,P=perm),distm[order(perm),])
    spamtest_eq(permutation(dist,Q=perm),distm[,order(perm)])
    spamtest_eq(permutation(dist,P=perm,ind=T),distm[perm,])
    spamtest_eq(permutation(dist,Q=perm,ind=T),distm[,perm])
    
    spamtest_eq(permutation(distm,P=perm),distm[order(perm),])
    spamtest_eq(permutation(distm,Q=perm),distm[,order(perm)])
    spamtest_eq(permutation(distm,P=perm,ind=T),distm[perm,])
    spamtest_eq(permutation(distm,Q=perm,ind=T),distm[,perm])
    
    spamtest_eq(t(permutation.spam(t(dist),P=perm)),distm[,order(perm)])
    spamtest_eq(t(permutation(t(dist),P=perm)),distm[,order(perm)])
    spamtest_eq(t(permutation(t(distm),P=perm)),distm[,order(perm)])
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-dist.R.             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-dist.R")


distmatrix <- function(x1,x2=NULL,upper=NULL,...)
  {
    if (is.null(x2)) {
      tmp <- as.matrix(dist(x1,...))
      if (is.null(upper)) return(tmp)
      if (upper) tmp[row(tmp)<col(tmp)] <- -1
      if (!upper) tmp[row(tmp)>col(tmp)] <- -1
      return( tmp[tmp>-0.5])
    } else
      return( as.matrix( dist(rbind(x1,x2),...))[1:dim(x1)[1],1:dim(x2)[1]+dim(x1)[1]])
  }


########
## as an aside, comparing nearest.dist with dist, use diag=true, upper=TRUE

test_that("comparing nearest.dist with dist, use diag=true, upper=TRUE",{

    options(spam.printsize=6)


    n1 <- as.integer( 4)
    n2 <- n1
    nd <- as.integer(2)
    set.seed(14)
    x2 <- x1 <- array(runif(n1*nd), c( n1,nd))

    if (F){
                                        # testing the structure
        distmatrix(x1)
        nearest.dist( x1, x1, upper=NULL)

                                        # and all other possibilities (3[upper])
                                        # with x1,x1 and x1, NULL:
        par(mfcol=c(3,2))
        display(   nearest.dist( x1, x1,  upper=NULL))  # default
        display(   nearest.dist( x1, x1,  upper=FALSE))
        display(   nearest.dist( x1, x1,  upper=TRUE))
        display(   nearest.dist( x1,  upper=NULL))
        display(   nearest.dist( x1,  upper=FALSE))
        display(   nearest.dist( x1,  upper=TRUE))
    }


#  nearest.dist( x1) and  nearest.dist( x1,x1) should be identical...
    expect_equal(nearest.dist( x1, x1, upper=NULL),  nearest.dist(x1,  upper=NULL) )
    expect_equal(nearest.dist( x1, x1, upper=FALSE), nearest.dist(x1,  upper=FALSE))
    expect_equal(nearest.dist( x1, x1, upper=TRUE),  nearest.dist(x1,  upper=TRUE) )
# New Rcpp and old Fortran version should be identical...
    expect_equal(nearest.dist(x1, fortran = FALSE), nearest.dist(x1, fortran = TRUE))


# testing  Euclidian
    eta <- 1

    o1 <- nearest.dist( x1, upper=NULL)
    o2 <- distmatrix(x1)

    expect_equal(o2[o2< eta], o1@entries)


    o1 <- nearest.dist( x1, upper=!FALSE)  # is default...
    o3 <- distmatrix(x1, upper=!FALSE)
    expect_equal(o1@entries, o3)



    x2 <- x1 <- array(runif(n1*nd), c( n1,nd))
    o1 <- nearest.dist( x1,x2,upper=NULL)
    o2 <- distmatrix(x1,x2)

    expect_equal(o2[o2< eta], o1@entries)

    ## TODO check test.
    ## o1 <- nearest.dist( x1, upper=!FALSE)
    ## expect_equal(o2[o2< eta & lower.tri(o2)], o1@entries)



# Should cause error:
#    nearest.dist(cbind(1,1))
# this is ok:
    spamtest_eq( nearest.dist(rbind(1,0)), c(0,1,0,0))
    spamtest_eq( nearest.dist(cbind(1,1),cbind(1,0)), 1)



                                        # testing with dist only
    spamtest_eq( as.spam( dist(x1)), nearest.dist(x1,delta=2))


# testing some other norms
    method <- "max"
    p <- 1.0001
    o1 <- nearest.dist( x1,method=method,p=p, upper=TRUE )
    o3 <- distmatrix(x1,method=method,p=p, upper=TRUE)
    expect_equal(o1@entries, o3)

# New Rcpp and old Fortran version should also be identical...
    expect_equal(nearest.dist(x1, method=method, fortran = FALSE), nearest.dist(x1, method = method, fortran = TRUE))



    if (F){ # system.time is not always available...

        n1 <- as.integer( 400)
        set.seed(14)
        x1 <- array(runif(n1*nd), c( n1,nd))

        system.time( o1 <- nearest.dist( x1,method="max",p=p) )
        system.time( o1 <- nearest.dist( x1,method="min",p=1) )
        system.time( o1 <- nearest.dist( x1,method="min",p=1.5) )
        system.time( o1 <- nearest.dist( x1,method="min",p=2) )
        system.time( o1 <- nearest.dist( x1,method="euc",p=1) )
        system.time( o1 <- dist( x1) )
    }

                                        # testing  GC
    n1 <- as.integer( 4)
    n2 <- as.integer(6)
    set.seed(14)
    x1 <- array(runif(n1*2,-90,90), c( n1,2))
    x2 <- array(runif(n2*2,-90,90), c( n2,2))



    if (F){
                                        # structure
        delta <-  180

        par(mfcol=c(3,2))
        display( nearest.dist( x1,    delta=delta,method="gr",  upper=FALSE))
        display( nearest.dist( x1,    delta=delta,method="gr",  upper=TRUE))
        display( nearest.dist( x1,    delta=delta,method="gr",  upper=NULL))
        display( nearest.dist( x1,x1, delta=delta,method="gr",  upper=FALSE))
        display( nearest.dist( x1,x1, delta=delta,method="gr",  upper=TRUE))
        display( nearest.dist( x1,x1, delta=delta,method="gr",  upper=NULL))
    }

                                        #

    if (F){
                                        # if fields would be available, the following can be used as well.
        delta <-  180
        o2 <- rdist.earth(x1)
        o1 <- nearest.dist( x1, method="gr",upper=NULL,delta=delta)
        spamtest_eq(o2- o1@entries)

# New Rcpp and old Fortran version should also be identical...
        expect_equal(nearest.dist(x1, method="gr", fortran = FALSE), nearest.dist(x1, method = "gr", fortran = TRUE))

        o2 <- rdist.earth(x1, R=1)
        o1 <- nearest.dist( x1,  method="gr",upper=NULL,delta=delta,R=1)
        spamtest_eq(o2- o1@entries)


        delta <- 90
        o2 <- rdist.earth(x2,x1,R=1)
        o1 <- nearest.dist( x1,x2, method="gr",upper=NULL,delta=delta,R=1)
        spamtest_eq(o2[o2<delta*pi/180]- o1@entries)

        expect_equal(nearest.dist(x1, method="gr", fortran = FALSE, R=1), nearest.dist(x1, method = "gr", fortran = TRUE, R=1))


    }
## })

## # correct storage mode conversion
## test_that("storage conversion", {
    nx <- 4
    x <- expand.grid(as.double(1:nx),as.double(1:nx))
    expect_equal(nearest.dist( x,delta=nx*2, upper=TRUE)@entries,
                distmatrix(x, upper=TRUE))
    x <- expand.grid(as.integer(1:nx),as.integer(1:nx))
    expect_equal(nearest.dist( x,delta=nx*2, upper=TRUE)@entries,
                 distmatrix(x, upper=TRUE))


# again a bit playing with the parameters:
    if (F){
        rdist.earth(x1, R=1)
        nearest.dist( x1, x1,  method="gr",delta=180,R=1 )
        nearest.dist( x1,      method="gr",delta=180,R=1, upper=FALSE)
        nearest.dist( x1,      method="gr",delta=180,R=1, upper=TRUE)
        nearest.dist( x1,x1,   method="gr",delta=180,R=1, upper=FALSE)
        nearest.dist( x1,x1,   method="gr",delta=180,R=1, upper=TRUE)
    }


# When setting
#     spam.options(nearestdistnnz=c(3,3))
# we do not get a warning because there is a min.
#   nnz <- min(max(.Spam$nearestdistnnz[1], n1 * .Spam$nearestdistnnz[2]),
#              (as.double(n1) * (n1 + 1))/ifelse(is.null(upper),1, 2),
#              2^31 - 2)









#################################################################
# addendum:

    expect_equal( as.spam(dist(0))@entries, 0)
    expect_equal( as.spam(dist(x1)), as.spam(nearest.dist(x1,delta=1000)))

# diag and upper are only processed when displaying
    expect_equal( as.spam(dist(x1)), as.spam(dist(x1,  upper=TRUE)) )

# 'NA/NaN/Inf's are coerced to zero:
      expect_warning( as.spam(dist(c(0, NA, 1)))@entries)

})

# HEADER ####################################################
# This is file spam/R/spam_solve.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# The main function here is very similar to chol.spam().
# We work with different files such that changes can be easily ported via `meld`.
# Both call z <- .spam.chol.basis(). Only, chol() has the verbose argument.

determinant.spam <- function(x, logarithm=TRUE, ...){

    
  argList <- list(...)
  if (!is.null(argList$memory)||!is.null(argList$memory)||!is.null(argList$pivot))
        warning("Arguments 'memory', 'eps' and 'pivot' will become defunct in the future. \nIf you really need these, pass via a Choleski factorization first.")
    
  z <- .chol.spam.basis(x, pivot="MMD", memory=list(), eps=getOption("spam.eps"), verbose=FALSE) 
    
  tmp <- 2* sum( log( z$lnz[ z$xlnz[ -(z$nrow+1)]]))

  logdet <- list()     
  if (logarithm) logdet$modulus <- tmp else logdet$modulus <- exp(tmp)
  attr(logdet$modulus,"logarithm") <- logarithm

  logdet$sign <- 1
  attr(logdet,"class") <- "det"

  return(logdet)
}

determinant.spam.chol.NgPeyton <- function(x, logarithm = TRUE,...)
{
  tmp <- sum( log(x@entries[ x@rowpointers[-(x@dimension[1]+1)]]))
 
  logdet <- list()
  if (logarithm) logdet$modulus <- tmp else logdet$modulus <- exp(tmp)
  attr(logdet$modulus,"logarithm") <- logarithm

  logdet$sign <- 1
  attr(logdet,"class") <- "det"

  return(logdet)
}

setMethod("determinant","spam",               determinant.spam)
setMethod("determinant","spam.chol.NgPeyton", determinant.spam.chol.NgPeyton)

## The ``Right Thing'' to do :
## base::det() calls [base::]determinant();
## our det() should call our determinant() :
det <- base::det
environment(det) <- environment()## == asNamespace("Matrix")
######################################################################
########################################################################

# HEADER ####################################################
# This is file spam/R/spam_solve.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################
########################################################################
#
# Contains routines linked to solving spd linear systems. Namely:
#    chol, solve, backsolve, forwardsolve
#    determinant                (the later because it is based on chol)
#
# As well as associated S4 elements.
#
# The key element is a new class: "spam.chol.NgPeyton", the output
# of 'chol'
#
########################################################################
########################################################################

setClass("spam.chol.NgPeyton",
         representation(entries="numeric",      colindices="integer",
                        colpointers="integer",  rowpointers="integer",
                        dimension="integer",
                        pivot="integer",        invpivot="integer",
                        supernodes="integer",   snmember="integer",
                        memory="integer",       nnzA="integer")
         )

# lindx=  colindices
# xlindx= colpointers
# xlnz=   rowpointers
# snode=snmember
# xsuper=supernodes
# c(... nnztmp,cachesize)= memory


#setClass("spam.chol.NgPeyton",
#         representation(nrow="integer",nnzlindx="integer",
#                        nsuper="integer",lindx="integer",xlindx="integer",nnzl="integer",
#                        lnz="numeric",xlnz="integer",invp="integer",perm="integer",
#                        xsuper="integer"),
         # the prototype corresponds to the cholesky of '1'
#         prototype=prototype(nrow=as.integer(1),nnzlindx=as.integer(1),
#           nsuper=as.integer(1),lindx=as.integer(1),xlindx=as.integer(c(1,2)),
#           nnzl=as.integer(1),lnz=1.0,xlnz=as.integer(c(1,2)),
#           invp=as.integer(1),perm=as.integer(1),xsuper=as.integer(c(1,2))
#           )
#         )


########################################################################

print.spam.chol.NgPeyton <- function(x,...) {
  nrow <- x@dimension[1]
  nnzR <- x@rowpointers[nrow+1]-1
  cat("(Upper) Cholesky factor of dimension ", nrow,
                "x", nrow, " with ",nnzR," nonzero elements.", sep = "", fill=TRUE)
  cat("    (The object is supposed to be used with: 'as.spam', 'backsolve', 'forwardsolve', etc.)\n",
      fill=TRUE)
  cat("Class 'spam.chol.NgPeyton'\n")
  invisible(NULL)
}

setMethod("show","spam.chol.NgPeyton", function(object) {
  nrow <- object@dimension[1]
  nnzR <- object@rowpointers[nrow+1]-1
  cat("(Upper) Cholesky factor of dimension ", nrow,
                "x", nrow, " with ",nnzR," (row-wise) nonzero elements.", sep = "", fill=TRUE)
  cat("    (The object is supposed to be used with: 'as.spam', 'backsolve', 'forwardsolve', etc.)\n",
      fill=TRUE)
  cat("Class 'spam.chol.NgPeyton'\n")
  invisible(NULL)
        })


"diag.of.spam.chol.NgPeyton" <- function(x, nrow, ncol)
  return( x@entries[x@rowpointers[-(x@dimension[1]+1)]])


setMethod("diag",    "spam.chol.NgPeyton", diag.of.spam.chol.NgPeyton)
#setMethod("diag<-",  "spam.chol.NgPeyton", function(x,...) stop("operation not allowed on 'spam.chol.NgPeyton' object"))
setMethod("print",   "spam.chol.NgPeyton", print.spam.chol.NgPeyton)
# setMethod("summary", "spam.chol.NgPeyton", summary.spam.chol.NgPeyton)  # moved to new file
setMethod("dim",     "spam.chol.NgPeyton",function(x) x@dimension)
setMethod("length",  "spam.chol.NgPeyton",function(x) x@rowpointers[x@dimension[1]+1]-1)
setMethod("length<-","spam.chol.NgPeyton",function(x,value) stop("operation not allowed on 'spam.chol.NgPeyton' object") )
setMethod("dim<-",   "spam.chol.NgPeyton",function(x,value) stop("operation not allowed on 'spam.chol.NgPeyton' object") )

setMethod("c","spam.chol.NgPeyton", function(x,...){
    nrow <- x@dimension[1]
    nnzR <- x@rowpointers[nrow+1]-1
    ## newx <- new("spam") ## what is this doing here??? TODO
    nsuper <- as.integer( length(x@supernodes)-1)
    ## xcolindices <- .Fortran('calcja',
    ##                         as.integer(nrow),
    ##                         as.integer(nsuper),
    ##                         as.integer(x@supernodes),
    ##                         as.integer(x@colindices),
    ##                         as.integer(x@colpointers),
    ##                         as.integer(x@rowpointers),
    ##                         xja=vector("integer",nnzR),
    ##                         NAOK = getOption("spam.NAOK"),
    ##                         PACKAGE = "spam")$xja
    if( getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    xcolindices <- .C64('calcja',
                        SIGNATURE = rep(SS$signature, 7),

                        nrow,
                        nsuper,
                        x@supernodes,
                        x@colindices,

                        x@colpointers,
                        x@rowpointers,
                        xja = vector_dc( SS$type, nnzR),

                        INTENT=c("r", "r", "r", "r",
                                 "r", "r", "w"),
                        NAOK = getOption("spam.NAOK"),
                        PACKAGE = SS$package)$xja
    cx <- .C64("spamcsrdns",
               SIGNATURE = c(SS$signature, "double" , SS$signature, SS$signature,
                             "double"),

               nrow = nrow,
               entries = x@entries,
               colindices = xcolindices,
               rowpointers = x@rowpointers,

               res = vector_dc( "double", nrow*nrow),

               INTENT = c("r", "r", "r", "r",
                          "w"),
               NAOK=getOption("spam.NAOK"),
               PACKAGE = SS$package)$res
  if (length( list(...)) < 1)
    return( cx)
  else
    c( cx,c(...))
})

as.spam.chol.NgPeyton <- function(x, eps = getOption("spam.eps")) {
    if( getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

  if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
  nrow <- x@dimension[1]
  nnzR <- x@rowpointers[nrow+1]-1
  nsuper <- length(x@supernodes)-1

  colindices <- .C64('calcja',
                     SIGNATURE=rep(SS$signature, 7),

                     nrow,
                     nsuper,
                     x@supernodes,
                     x@colindices,

                     x@colpointers,
                     x@rowpointers,
                     xja=vector(SS$type, nnzR), #!1#

                     INTENT=c("r", "r", "r", "r", "r", "r", "w"),
                     NAOK = getOption("spam.NAOK"),
                     PACKAGE = SS$package)$xja

  return(.newSpam(
    entries=x@entries,
    colindices=colindices,
    rowpointers=x@rowpointers,
    dimension=x@dimension
  ))
}

setMethod("as.spam","spam.chol.NgPeyton", as.spam.chol.NgPeyton)


##NB setGeneric("backsolve", def = function(r, x, ...) standardGeneric("backsolve"),
#                 useAsDefault= function(r, x,...) base::backsolve(r, x, ...))

# We have some issues here... hence I postpone the proper implementation!!!
# http://r.789695.n4.nabble.com/class-extension-and-documentation-tt4161373.html#none

#"backsolve" <- function(r,x, ...) UseMethod("backsolve")
#"backsolve.default" <- base::backsolve
#setGeneric("backsolve")
#setMethod("backsolve","matrix",base::backsolve)

#"forwardsolve" <- function(l,x, ...) UseMethod("forwardsolve")
#"forwardsolve.default" <- base::forwardsolve
#setGeneric("forwardsolve")
#setMethod("forwardsolve","matrix",base::forwardsolve)


setGeneric("backsolve", def = function(r, x, ...) standardGeneric("backsolve"),
           useAsDefault= function(r, x, ...) base::backsolve(r, x, ...))

setGeneric("forwardsolve", def = function(l, x, ...) standardGeneric("forwardsolve"),
           useAsDefault= function(l, x, ...) base::forwardsolve(l, x, ...))

# adapted from methods
#setGeneric("forwardsolve", function(l, x, k, upper.tri = FALSE, transpose = FALSE, ...)
#           standardGeneric("forwardsolve"),
#           useAsDefault = function(l, x, k = ncol(l), upper.tri = FALSE, transpose = FALSE, ...)
#                  base::forwardsolve(l, x, k = k, upper.tri = upper.tri, transpose = transpose, ... ),
#           signature = c("l", "x"))#, where = where)
##### setGenericImplicit("forwardsolve")#, restore=FALSE)


"ordering.default" <- function(x,inv=FALSE) stop('Operation not defined form this class')

#ordering <- function(x,...) stop('Operation not defined form this class')
#setGeneric("ordering")
setGeneric("ordering",function(x,inv=FALSE)standardGeneric("ordering"))

setMethod("ordering","spam.chol.NgPeyton",function(x,inv=FALSE)
          {
            if (inv) return(x@invpivot) else return(x@pivot) })



setMethod("ordering","matrix",function(x,inv=FALSE)
          {
            if (dim(x)[1]!=dim(x)[2])
              stop("ordering is defined for square matrices only")
            if(inv)return(dim(x)[1]:1) else return(1:dim(x)[1]) })

setMethod("ordering","spam",function(x,inv=FALSE)
          {
            if (dim(x)[1]!=dim(x)[2])
              stop("ordering is defined for square matrices only")
            if(inv)return(dim(x)[1]:1) else return(1:dim(x)[1]) })



update.spam.chol.NgPeyton <- function(object,x,...){
    ## print("update.spam.chol.NgPeyton")
  nrow <- object@dimension[1]
  if (!is.spam(x))
    stop("Covariance should be a 'spam' object.")
  if ((x@rowpointers[nrow+1]-1) != object@nnzA)
    stop("Updated covariance entries do not match length of original one.")

  ## u <- .Fortran("updatefactor",
  ##               as.integer(nrow),
  ##               as.integer(object@nnzA),
  ##               d =  as.double(x@entries),  jd = as.integer(x@colindices),
  ##               id = as.integer(x@rowpointers),
  ##               as.integer(object@invpivot),   as.integer(object@pivot),
  ##               lindx=as.integer(object@colindices),    xlindx=as.integer(object@colpointers),
  ##               nsuper=as.integer( length(object@supernodes)-1),
  ##               entries = vector("double",length(object@entries)), #lnz
  ##               rowpointers = as.integer(object@rowpointers),#xlnz
  ##               snode=as.integer(object@snmember),  xsuper=as.integer(object@supernodes),
  ##               cachesize=as.integer(object@memory[3]),
  ##               ierr = 0L,
    ##               NAOK = getOption("spam.NAOK"),PACKAGE="spam")
    if( getOption("spam.force64") || .format.spam(object)$package != "spam" || .format.spam(x)$package != "spam" )
        SS <- .format64()
    else
        SS <- .format32

    u <- .C64("updatefactor",
     ##          subroutine updatefactor( m,nnzd,
     ## &     d,jd,id, invp,perm,
     ## &                lindx,xlindx, nsuper,lnz,xlnz,
     ## &                snode, xsuper,
     ## &                cachesize,ierr)
              SIGNATURE = c(SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            SS$signature, SS$signature, SS$signature, SS$signature, SS$signature,
                            "double", SS$signature, SS$signature, SS$signature, SS$signature, SS$signature ),
              nrow,
              object@nnzA,

              d =  x@entries,
              jd = x@colindices,
              id = x@rowpointers,

              object@invpivot,
              object@pivot,
              lindx = object@colindices,
              xlindx = object@colpointers,
              nsuper = length(object@supernodes)-1,

              entries = vector_dc( "double", length(object@entries)), #lnz
              rowpointers = object@rowpointers,#xlnz
              snode = object@snmember,
              xsuper = object@supernodes,
              cachesize = object@memory[3],
              ierr = 0,

              ## INTENT = c("r", "r",
              ##            "r", "r", "r",
              ##            "r", "r", "r", "r", "r",
              ##            "rw", "rw", "r", "r", "rw", "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)

  if(u$ierr>1) stop("Internal error in 'update.spam.chol.NgPeyton' code ", u$ierr,call.=FALSE)

  if(u$ierr == 1) {
    if (getOption("spam.cholupdatesingular") == "null")
      return(NULL)
    else if (getOption("spam.cholupdatesingular") == "error")
      stop("Singularity problem when updating a Cholesky Factor.")
    else if (getOption("spam.cholupdatesingular") == "warning")
      warning("Singularity problem when updating a Cholesky Factor.\n'object' not updated.")
    else
      stop("'cholupdatesingular' should be 'error', 'null' or 'warning'.")
  }  else {
    slot(object, "entries", check = FALSE) <- u$entries
  }
  invisible(object)
}



chol.spam <- function(x, pivot = "MMD",
                      method="NgPeyton",
                      memory=list(),
                      eps = getOption("spam.eps"), Rstruct=NULL, ...){

  if (is(Rstruct,"spam.chol.NgPeyton")) invisible( update.spam.chol.NgPeyton(Rstruct,x,...))
    
    force64 <- getOption("spam.force64")
  if(force64 || prod(dim(x)) > 2147483647)
      SS <- .format64()
  else
      SS <- .format32
  #!3#


  if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
  nrow <- x@dimension[1]
  nnzA <- x@rowpointers[nrow+1]-1
  if(nrow!=x@dimension[2]) stop("non-square matrix in 'chol'",call.=FALSE)

  if (any( diag(x, nrow, nrow) < getOption("spam.eps")))
    stop("Input matrix to 'chol' not positive definite (up to eps)",call.=FALSE)
# base rule:
#  `nrow(x) * .Machine$double.neg.eps * max(diag(x)`.


  if(getOption("spam.cholsymmetrycheck")) {
    test <- isSymmetric(x, tol = eps*100)
#  from help of isSymmetric:
#     isSymmetric(object, tol = 100 * .Machine$double.eps, ...)
    if (!isTRUE(test))
      stop("Input matrix to 'chol' not symmetric (up to 100*eps)",call.=FALSE)
  }

  if (method != "NgPeyton")
    warning(gettextf("method = '%s' is not supported. Using 'NgPeyton'",
                     method), domain = NA)

  if (length(pivot)==1) {
    if (pivot==FALSE) {
      doperm <- 0
      pivot <- seq_len(nrow)
    } else if(pivot==TRUE) {
      doperm <- 1
      pivot <- vector(SS$type,nrow)
    } else {
      doperm <- as.integer( switch(match.arg(pivot,c("MMD","RCM")),MMD=1,RCM=2))
      pivot <- vector(SS$type, nrow)
    }
  } else  if (length(pivot)==nrow) {
    doperm <- 0
    if (getOption("spam.cholpivotcheck")) {
      checkpivot(pivot,nrow)
    }
 } else stop("'pivot' should be 'MMD', 'RCM' or a valid permutation")



  ### IMPROVEME get better parameter values
  nnzcfact <- c(5,1,5)
  nnzRfact <- c(5,1,2)
  # nnzcolindices = length of array holding the colindices
  if(is.null(memory$nnzcolindices))  {
    nnzcolindices <- ifelse((nnzA/nrow < 5), # very sparse matrix
                            max(1000,nnzA*(1.05*nnzA/nrow-3.8)),
                            nnzA)*nnzcfact[doperm+1]
    nnzcolindices <- max(nnzcolindices,nnzA)
 }else {
    nnzcolindices <- max(memory$nnzcolindices,nnzA)
    memory$nnzcolindices <- NULL
  }
  # nnzR = length of array holding the nonzero values of the factor
  if(is.null(memory$nnzR)) {
    nnzR <- min(max(4*nnzA,floor(.2*nnzA^1.3))*nnzRfact[doperm+1],nrow*(nrow+1)/2)
  } else {
    nnzR <- memory$nnzR
    memory$nnzR <- NULL
  }

#!4#
  if(is.null(memory$cache))
    cache <- 512
  else {
    cache <- memory$cache
    memory$cache <- NULL
  }

  if (length( memory)>0 )
    warning("The component(s) ", paste("'",names(memory),"'",sep='',collapse=","),
            " of the argument 'memory'\npassed to function 'chol' not meaningful and hence ignored.",call.=FALSE)

  cholstep.intent <- c("r", "r", "r", "r",
                    "r", "r", "rw", "rw",
                    "rw", "r", "rw", "rw",
                    "rw", "rw", "rw", "rw",
                    "rw", "rw", "r", "rw")

  cholstep.signature <- rep(SS$signature, 20)
  cholstep.signature[c(3,15)] <- "double"

  z <- .C64("cholstepwise",
            SIGNATURE=cholstep.signature,

            nrow = nrow,
            nnzA = nnzA,
            d =  x@entries,
            jd = x@colindices,

            id = x@rowpointers, #5
            doperm = doperm,
            invp = vector(SS$type,nrow),
            perm = pivot,

            nnzlindx = 1L, #9
            nnzcolindices = nnzcolindices,
            lindx = vector(SS$type, nnzcolindices),
            xlindx = vector(SS$type, nrow+1),

            nsuper = 1L, #13
            nnzR = nnzR,
            lnz = vector("double",nnzR),
            xlnz = vector(SS$type, nrow+1),

            snode = vector(SS$type, nrow), #17
            xsuper = vector(SS$type, nrow+1),
            cachesize = cache,
            ierr = 0L,

            INTENT=cholstep.intent,
            NAOK = getOption("spam.NAOK"),
            PACKAGE = SS$package)

  if(z$ierr == 1) stop("Singularity problem when calculating the Cholesky factor.")
  if(z$ierr == 6) stop("Inconsitency in the input",call.=FALSE)

  while( z$ierr>1) {
    if(z$ierr == 4) {
      tmp <- ceiling(nnzR*getOption("spam.cholincreasefactor")[1])
      warning("Increased 'nnzR' with 'NgPeyton' method\n",
                    "(currently set to ",tmp," from ",nnzR,")",call.=FALSE)
      nnzR <- tmp
    }
    if(z$ierr == 5) {
      tmp <- ceiling(nnzcolindices*getOption("spam.cholincreasefactor")[2])
      warning("Increased 'nnzcolindices' with 'NgPeyton' method\n",
         "(currently set to ",tmp," from ",nnzcolindices,")",call.=FALSE)
      nnzcolindices <- tmp
    }
    z <- .C64("cholstepwise",
              SIGNATURE=cholstep.signature,

              nrow = nrow,
              nnzA = x@rowpointers[nrow+1]-1,
              d =  x@entries,
              jd = x@colindices,

              id = x@rowpointers,
              doperm = doperm,
              invp = vector(SS$type,nrow),
              perm = pivot,

              nnzlindx = vector(SS$type,1),
              nnzcolindices = nnzcolindices,
              lindx = vector(SS$type, nnzcolindices),
              xlindx = vector(SS$type, nrow+1),     #

              nsuper = 1,          #
              nnzR = nnzR,#
              lnz = vector("double",nnzR),        #
              xlnz = vector(SS$type, nrow+1),     #

              snode = vector(SS$type, nrow),
              xsuper = vector(SS$type, nrow+1),
              cachesize = cache,
              ierr = 0L,

              INTENT=cholstep.intent,
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)

    if(z$ierr == 1) stop("Singularity problem when calculating the Cholesky factor.")
  }
  nnzR <- z$xlnz[length(z$xlnz)]-1

  newx <- new("spam.chol.NgPeyton")
  slot(newx,"entries",check=FALSE) <- z$lnz[1:nnzR]
  slot(newx,"colindices",check=FALSE) <- z$lindx[1:z$nnzlindx]
  slot(newx,"colpointers",check=FALSE) <- z$xlindx[1:(z$nsuper+1)]
  slot(newx,"rowpointers",check=FALSE) <- z$xlnz
  slot(newx,"dimension",check=FALSE) <- c(nrow,nrow)
  slot(newx,"pivot",check=FALSE) <- z$perm
  slot(newx,"invpivot",check=FALSE) <- z$invp
  slot(newx,"supernodes",check=FALSE) <- z$xsuper[1:(z$nsuper+1)]
  slot(newx,"snmember",check=FALSE) <- z$snode
  slot(newx,"memory",check=FALSE) <- c(nnzcolindices,z$nnzR,cache)
  slot(newx,"nnzA",check=FALSE) <- nnzA
  invisible(newx)
}


solve.spam <- function (a, b,  Rstruct = NULL, ...) {
  nrow <- a@dimension[1]
  ncol <- a@dimension[2]
  if (ncol != nrow)      stop("only square matrices can be inverted")

  if (missing(b)) {
    b <- diag(1, ncol)
  }  else {
    if(!is.matrix(b)) b <- as.matrix(b)
  }
  p <- dim(b)[2]
  if(nrow!=dim(b)[1])stop("'b' must be compatible with 'a'")

  # if we have a spam matrix, we calculate the Cholesky factor
  if (is(a,"spam"))
    if (is(Rstruct, "spam.chol.NgPeyton"))
        a <- update.spam.chol.NgPeyton(Rstruct, a, ...)
    else a <- chol.spam(a, ...)


  if (is(a,"spam.chol.NgPeyton")) {
      # The following is a fast way to perform:
                                        #     z <- backsolve(a,forwardsolve( t(a),b))
      ## print("a is spam.chol.NgPeyton in solve.spam")
    nsuper <- as.integer( length(a@supernodes)-1)
    ## z <- .Fortran("backsolves", m = nrow,
    ##               as.integer(nsuper), as.integer(p), as.integer(a@colindices),
    ##               as.integer(a@colpointers), as.double(a@entries),
    ##               as.integer(a@rowpointers), as.integer(a@invpivot), as.integer(a@pivot),
    ##               as.integer(a@supernodes), vector("double",nrow), sol = vector("double",nrow*p),
    ##               as.vector(b,"double"),
    ##               NAOK = getOption("spam.NAOK"),
      ##               PACKAGE = "spam")$sol
      if( getOption("spam.force64") || .format.spam(a)$package != "spam" )
          SS <- .format64()
      else
          SS <- .format32

      z <- .C64("backsolves",
                SIGNATURE = c(rep(SS$signature,5),"double", rep(SS$signature, 4),
                              rep("double",3)),

                m = nrow,
                nsuper,
                p,
                a@colindices,
                a@colpointers,
                a@entries,
                a@rowpointers,
                a@invpivot,
                a@pivot,
                a@supernodes,

                vector_dc("double",nrow),
                sol = vector_dc("double",nrow*p),
                as.vector(b,"double"),

                INTENT = c( rep( "r", 10), rep( "rw", 3)),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package )$sol
  } else z <- backsolve(a, forwardsolve( t(a),b))
    # see the helpfile for a comment about the 't(a)' construct.

  if ( p!=1)    dim(z) <- c(nrow,p)
  return( z)
}

chol2inv.spam <- function (x, ...) {
    ## print("chol2inv.spam")
    nrow <- x@dimension[1]

    if (is(x,"spam.chol.NgPeyton")) {
        ## print("x = spam.chol.NgPeyton")
        y <- vector("double",nrow*nrow)
        y[1L + 0L:(nrow - 1L) * (nrow + 1L)] <- 1.0

        ## z <- .Fortran("backsolves", m = nrow,
        ##               as.integer( length(x@supernodes)-1), as.integer(nrow), as.integer(x@colindices),
        ##               as.integer(x@colpointers), as.double(x@entries),
        ##               as.integer(x@rowpointers), as.integer(x@invpivot), as.integer(x@pivot),
        ##               as.integer(x@supernodes), vector("double",nrow), sol = vector("double",nrow*nrow), as.double(y),
        ##               NAOK = getOption("spam.NAOK"),
        ##               PACKAGE = "spam")$sol

        if( getOption("spam.force64") || .format.spam(x)$package !="spam" )
            SS <- .format64()
        else
            SS <- .format32

        ## print(SS$package)

        z <- .C64("backsolves",
          ##         subroutine backsolves(m,nsuper,nrhs,lindx,xlindx,lnz,
     ## &                   xlnz,invp,perm,xsuper,newrhs,sol,b)
                  SIGNATURE = c( rep( SS$signature, 5), "double", rep( SS$signature, 4),
                                rep( "double", 3)),

                  m = nrow, #r
                  length(x@supernodes)-1, #r
                  nrow, #r
                  x@colindices,
                  x@colpointers,
                  x@entries,
                  x@rowpointers,
                  x@invpivot, #r
                  x@pivot, #r
                  x@supernodes, #r

                  vector_dc("double",nrow), #rw
                  sol = vector_dc("double",nrow*nrow), #w
                  y, #r

                  INTENT = c( rep( "r", 10),
                             "rw", "w", "r"),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = SS$package)$sol

        dim(z) <- c(nrow,nrow)
    } else z <- backsolve.spam(x, forwardsolve.spam( t(x), diag(nrow)))
    return( z)
}

backsolve.spam <- function(r, x,...){#, k = NULL, upper.tri = NULL, transpose = NULL){
# r: spam.chol.NgPeyton structure as returned by chol.spam or a spam object
# x: rhs a vector or a matrix in dense form
# dimensions:  ( m x n) ( n x p)

    if( getOption("spam.force64") || .format.spam(r)$package !="spam" )
        SS <- .format64()
    else
        SS <- .format32


  m <- r@dimension[1]
  if(is.vector(x)) {
    n <- length(x)
    p <- 1
  } else {
    x <- as.matrix(x)
    n <- dim(x)[1]
    p <- dim(x)[2]
  }

  # we separate between "spam.chol.NgPeyton" and "spam"
  if (is(r,"spam.chol.NgPeyton")) {
    if (n!=m) stop("Cholesky factor 'r' not compatible with 'x'")
    nsuper <- length(r@supernodes)-1
    if (!getOption("spam.dopivoting")) {

      z <- .C64("backsolvef",
                SIGNATURE=c(SS$signature, SS$signature, SS$signature, SS$signature,
                            SS$signature, "double", SS$signature, SS$signature,
                            "double"),

                m,
                nsuper,
                p,
                r@colindices,

                r@colpointers,
                r@entries,
                r@rowpointers,
                r@supernodes,

                sol = vector("double",m*p),

                INTENT=c("r", "r", "r", "r", "r", "r", "r", "r", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol

    }else{
      z <- .C64("pivotbacksolve",
                SIGNATURE=c(SS$signature, SS$signature, SS$signature, SS$signature,
                            SS$signature, "double", SS$signature, SS$signature,
                            SS$signature, SS$signature, "double", "double",
                            "double"),

                m,
                nsuper,
                p,
                r@colindices,

                r@colpointers,
                r@entries,
                r@rowpointers,
                r@invpivot,

                r@pivot,
                r@supernodes,
                vector("double",m),
                sol = vector("double",m*p),

                x,

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "r", "r", "r", "w",
                    "r"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol
    }
  } else {
    if (n!=m) stop("Triangular matrix 'r' not compatible with 'x'")
    # solve R sol = x
    z <- .C64("spamback",
              SIGNATURE=c(SS$signature, SS$signature, "double", "double",
                          "double", SS$signature, SS$signature),
              m=m,
              unused=p,
              sol = vector("double",m*p),
              x=x,

              al=r@entries,
              jal=r@colindices,
              ial=r@rowpointers,

              INTENT=c("rw", "rw", "rw", "rw", "rw", "rw", "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE=SS$package)
    if (z$m<0) stop(gettextf("singular matrix in 'backsolve'. Last zero in diagonal [%d]",
            -z$m), domain = NA)
     else z <- z$sol

  }

  if (p>1) dim(z) <- c(m,p)
  return(z)
}

forwardsolve.spam <- function(l, x,...){#, k = NULL, upper.tri = NULL, transpose = NULL){
#  l: spam.chol.NgPeyton structure as returned by chol.spam
#         or an ordinary lower triangular spam matrix
#  x: rhs a vector a matrix in dense form
#  dimensions:  ( m x n) ( n x p)
#  if (!any(is.null(c(upper.tri,k,transpose ))))
#    warning("'k', 'upper.tri' and 'transpose' argument do not have any effect here")
    if( getOption("spam.force64") || .format.spam(l)$package !="spam" )
        SS <- .format64()
    else
        SS <- .format32

  m <- l@dimension[1]
  if(is.vector(x)) {
    n <- length(x)
    p <- 1L
  } else {
    if(!is.matrix(x)) x <- as.matrix(x)
    n <- dim(x)[1]
    p <- dim(x)[2]
  }

  # we separate between "spam.chol.NgPeyton" and "spam"
  if (is(l,"spam.chol.NgPeyton")) {
    if(n!=m) stop("Cholesky factor 'l' not compatible with 'x'")
    nsuper <- length(l@supernodes)-1
    if (!getOption("spam.dopivoting")) {
      z <- .C64("forwardsolvef",
                SIGNATURE=c(SS$signature, SS$signature, SS$signature, SS$signature,
                            SS$signature, "double", SS$signature, SS$signature,
                            "double"),

                m,
                nsuper,
                p,
                l@colindices,

                l@colpointers,
                l@entries,
                l@rowpointers,
                l@supernodes,

                sol = vector("double",m*p),

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol
    }else{
      z <- .C64("pivotforwardsolve",
                SIGNATURE=c(SS$signature, SS$signature, SS$signature, SS$signature,
                            SS$signature, "double", SS$signature, SS$signature,
                            SS$signature, SS$signature, "double", "double",
                            "double"),
                m,
                nsuper,
                p,
                l@colindices,

                l@colpointers,
                l@entries,
                l@rowpointers,
                l@invpivot,

                l@pivot,
                l@supernodes,
                vector("double",m), #!5#
                sol = vector("double",m*p),

                x,

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "r", "r", "r", "w",
                    "r"),
                PACKAGE=SS$package)$sol
    }
  } else {
    if (n!=m) stop("Triangular matrix 'l' not compatible with 'x'")
    # solve L sol = x
    z <- .C64("spamforward",
              SIGNATURE=c(SS$signature, SS$signature, "double", "double",
                          "double", SS$signature, SS$signature),

              m=m,
              p,
              sol = vector("double",m*p),
              x=x,

              al=l@entries,
              jal=l@colindices,
              ial=l@rowpointers,

              INTENT=c("rw", "r", "w", "r", "r", "r", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE=SS$package)
    if (z$m<0) stop(gettextf("singular matrix in 'forwardsolve'. First zero in diagonal [%d]", -z$m), domain = NA)
    else z <- z$sol
  }
  if (p>1)
    dim(z) <- c(m,p)
  return(z)
}


setMethod("chol","spam", chol.spam)
setMethod("solve","spam",solve.spam)
setMethod("chol2inv","spam", chol2inv.spam)
setMethod("chol2inv","spam.chol.NgPeyton", chol2inv.spam)

setMethod("backsolve","spam",#signature(r="spam",x='ANY'),
          backsolve.spam)
setMethod("backsolve","spam.chol.NgPeyton",#signature(r="spam.chol.NgPeyton",x='ANY'),
          backsolve.spam,sealed=TRUE)
#setMethod("backsolve","spam.chol.NgPeyton",    backsolve.spam)
setMethod("forwardsolve","spam",               forwardsolve.spam)
setMethod("forwardsolve","spam.chol.NgPeyton", forwardsolve.spam)

######################################################################
######################################################################

determinant.spam <- function(x, logarithm = TRUE, pivot = "MMD",method="NgPeyton",
                              memory=list(),eps = getOption("spam.eps"), ...){
   # print("determinant.spam")

  if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
  logdet <- list()
 #### start from above
  nrow <- x@dimension[1]
  nnzA <- as.integer( x@rowpointers[nrow+1]-1)
  if(nrow!=x@dimension[2]) stop("non-square matrix in 'chol'",call.=FALSE)

  if(nrow <= 1) {
    return(determinant(as.matrix(x@entries))) }

  if(getOption("spam.cholsymmetrycheck")) {
    test <- isSymmetric.spam(x, tol = eps*100)
    if (!isTRUE(test))
      stop("Input matrix to 'chol' not symmetric (up to 100*eps)",call.=FALSE)
  }

  if (method != "NgPeyton")
    warning(gettextf("method = '%s' is not supported. Using 'NgPeyton'",
                     method), domain = NA)

  if (length(pivot)==nrow) {
    doperm <- 0L
    pivot <- as.vector(pivot,"integer")
    if (getOption("spam.cholpivotcheck")) {
      checkpivot(pivot,nrow)
    }
  } else if (length(pivot)==1) {
    if (pivot==FALSE) {
      doperm <- 0L
      pivot <- seq_len(nrow)
    } else if(pivot==TRUE) {
      doperm <- 1L
      pivot <- vector("integer",nrow)
    } else {
      doperm <- as.integer( switch(match.arg(pivot,c("MMD","RCM")),MMD=1,RCM=2))
      pivot <- vector("integer",nrow)
    }
  } else stop("'pivot' should be 'MMD', 'RCM' or a permutation")


  #!6#
    nnzcfact <- c(5,1,5)
    nnzRfact <- c(5,1,2)

  # nnzcolindices = length of array holding the colindices
  if(is.null(memory$nnzcolindices))  {
    nnzcolindices <- ifelse((nnzA/nrow < 5), # very sparse matrix
                            max(1000,nnzA*(1.05*nnzA/nrow-3.8)),
                            nnzA)*nnzcfact[doperm+1]
    nnzcolindices <- max(nnzcolindices,nnzA)
 }else {
    nnzcolindices <- max(memory$nnzcolindices,nnzA)
    memory$nnzcolindices <- NULL
  }
  # nnzR = length of array holding the nonzero values of the factor
  if(is.null(memory$nnzR))    nnzR <- min(max(4*nnzA,floor(.4*nnzA^1.2))*nnzRfact[doperm+1],nrow*(nrow+1)/2)  else {
    nnzR <- memory$nnzR
    memory$nnzR <- NULL
  }
  if(is.null(memory$cache))    cache <- 64  else {
    cache <- memory$cache
    memory$cache <- NULL
  }

  if (length( memory)>0 )
    warning("The component(s) ", paste("'",names(memory),"'",sep='',collapse=","),
            " of the argument 'memory'\npassed to function 'chol' not meaningful and hence ignored.",call.=FALSE)
  ## print("determinant.spam")
  ## z <- .Fortran("cholstepwise",
  ##               nrow = as.integer(nrow) ,nnzA = as.integer(x@rowpointers[nrow+1]-1),
  ##               d =  as.double(x@entries),jd = as.integer(x@colindices),id = as.integer(x@rowpointers),
  ##               doperm = as.integer(doperm), invp = vector("integer",nrow), perm = as.integer(pivot),
  ##               nnzlindx = vector("integer",1),
  ##               nnzcolindices = as.integer(nnzcolindices),
  ##               lindx = vector("integer",nnzcolindices),
  ##               xlindx = vector("integer",nrow+1),     #
  ##               nsuper = vector("integer",1),          #
  ##               nnzR = as.integer(nnzR),#
  ##               lnz = vector("double",nnzR),        #
  ##               xlnz = vector("integer",nrow+1),     #
  ##               snode = vector("integer",nrow),
  ##               xsuper = vector("integer",nrow+1),
  ##               cachesize = as.integer(cache),
  ##               ierr = 0L,
    ##               NAOK = getOption("spam.NAOK"), PACKAGE = "spam")

    if( getOption("spam.force64") || prod(dim(x)) > 2147483647)
        SS <- .format64()
    else
        SS <- .format32
    cholstep.intent <- c("r", "r", "r", "r",
                    "r", "rw", "rw", "rw",
                    "rw", "rw", "rw", "rw",
                    "rw", "rw", "rw", "rw",
                    "rw", "rw", "rw", "rw")
    cholstep.signature <- rep(SS$signature, 20)
    cholstep.signature[c(3,15)] <- "double"

    z <- .C64("cholstepwise",
      ##             subroutine cholstepwise(m,nnzd,
     ## &     d,jd,id,    doperm,invp,perm,
     ## &                nsub,nsubmax,
     ## &                lindx,xlindx,nsuper,nnzlmax,lnz,xlnz,
     ## &                snode,xsuper,
     ## &                cachsz,ierr)
              SIGNATURE = cholstep.signature,

              nrow = nrow ,
              nnzA = x@rowpointers[nrow+1]-1,
              d =  x@entries,
              jd = x@colindices,

              id = x@rowpointers,
              doperm = doperm,
              invp = vector_dc( SS$type, nrow),
              perm = pivot,

              nnzlindx = vector_dc( SS$type, 1),
              nnzcolindices = nnzcolindices,
              lindx = vector_dc( SS$type, nnzcolindices),
              xlindx = vector_dc( SS$type,nrow+1),     #

              nsuper = vector_dc( SS$type,1),          #
              nnzR = nnzR,#
              lnz = vector_dc( "double", nnzR),        #
              xlnz = vector_dc( SS$type, nrow+1),     #

              snode = vector_dc( SS$type, nrow),
              xsuper = vector_dc( SS$type, nrow+1),
              cachesize = cache,
              ierr = 0,

              ## INTENT = cholstep.intent,
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)


  if(z$ierr == 1) stop("Singularity problem when calculating the Cholesky factor.")
  if(z$ierr == 6) stop("Inconsitency in the input",call.=FALSE)

  while( z$ierr>1) {
    if(z$ierr == 4) {
      warning("Increased 'nnzR' with 'NgPeyton' method\n",
              "(currently set to ",nnzR," from ",ceiling(nnzR*getOption("spam.cholpar")[1]),")",call.=FALSE)
      nnzR <- ceiling(nnzR*getOption("spam.nnzRinc"))
    }
    if(z$ierr == 5) {
      warning("Increased 'nnzcolindices' with 'NgPeyton' method\n",
         "(currently set to ",nnzcolindices," from ",ceiling(nnzcolindices*getOption("spam$cholpar")[2]),")",call.=FALSE)
      nnzcolindices <- ceiling(nnzcolindices*getOption("spam.cholpar")[2])
    }
    print("whileloop in determinant.spam") ##TODO find a case with z$ierr > 1 before migration possible
    z <- .Fortran("cholstepwise",
                  nrow = nrow,nnzA = as.integer(x@rowpointers[nrow+1]-1),
                  d =  as.double(x@entries),jd = x@colindices,id = x@rowpointers,
                  doperm = doperm,invp = vector("integer",nrow), perm = pivot,
                  nnzlindx = vector("integer",1),
                  nnzcolindices = as.integer(nnzcolindices),
                  lindx = vector("integer",nnzcolindices),
                  xlindx = vector("integer",nrow+1),     #
                  nsuper = vector("integer",1),          #
                  nnzR = as.integer(nnzR),#
                  lnz = vector("double",nnzR),        #
                  xlnz = vector("integer",nrow+1),     #
                  snode = vector("integer",nrow),
                  xsuper = vector("integer",nrow+1),
                  cachesize = as.integer(cache),
                  ierr = 0L,
                  NAOK = getOption("spam.NAOK"), PACKAGE = "spam")

  }
 #### end from above
  if(z$ierr == 1) {
                                        # all other errors trapped
      warning("singularity problem or matrix not positive definite",call.=FALSE)
      logdet$modulus <- NA
   } else{
    tmp <- 2* sum( log( z$lnz[ z$xlnz[ -(z$nrow+1)]]))
    if (logarithm) logdet$modulus <- tmp else logdet$modulus <- exp(tmp)
  }

  attr(logdet$modulus,"logarithm") <- logarithm

  logdet$sign <- ifelse(z$ierr == 1,NA,1)
  attr(logdet,"class") <- "det"

  return(logdet)
}

determinant.spam.chol.NgPeyton <- function(x, logarithm = TRUE,...)
{
  logdet <- list()


  tmp <- sum( log(x@entries[ x@rowpointers[-(x@dimension[1]+1)]]))
  if (logarithm) logdet$modulus <- tmp else logdet$modulus <- exp(tmp)

  attr(logdet$modulus,"logarithm") <- logarithm

  logdet$sign <- 1
  attr(logdet,"class") <- "det"

  return(logdet)
}


setMethod("determinant","spam",               determinant.spam)
setMethod("determinant","spam.chol.NgPeyton", determinant.spam.chol.NgPeyton)

## The ``Right Thing'' to do :
## base::det() calls [base::]determinant();
## our det() should call our determinant() :
det <- base::det
environment(det) <- environment()## == asNamespace("Matrix")
######################################################################
########################################################################


"as.matrix.spam.chol.NgPeyton" <- function(x,...){
    ## print("as.matrix.spam.chol.NgPeyton")
    nrow <- x@dimension[1]
    nnzR <- x@rowpointers[nrow+1]-1
    ## newx <- new("spam")
    nsuper <-  length(x@supernodes)-1
    ## xcolindices <- .Fortran('calcja',
    ##                         as.integer(nrow), as.integer(nsuper), as.integer(x@supernodes), as.integer(x@colindices), as.integer(x@colpointers), as.integer(x@rowpointers),
    ##                         xja=vector("integer",nnzR),
    ##                         NAOK = getOption("spam.NAOK"),PACKAGE = "spam")$xja
    if( getOption("spam.force64") || .format.spam(x)$package != "spam" )
        SS <- .format64()
    else
        SS <- .format32

    xcolindices <- .C64('calcja',
                        SIGNATURE = c(rep(SS$signature,7)),
                        nrow,
                        nsuper,
                        x@supernodes,
                        x@colindices,

                        x@colpointers,
                        x@rowpointers,
                        xja = vector_dc( SS$type, nnzR),

                        INTENT=c("r", "r", "r", "r",
                                 "r", "r", "w"),
                        NAOK = getOption("spam.NAOK"),
                        PACKAGE = SS$package)$xja
  ## return(array(.Fortran("spamcsrdns",
  ##                nrow = as.integer(nrow),
  ##                entries = as.double(x@entries),
  ##                colindices = as.integer(xcolindices),
  ##                rowpointers = as.integer(x@rowpointers),
  ##                res = vector("double",nrow*nrow),
  ##                NAOK=getOption("spam.NAOK"),PACKAGE = "spam")$res,
  ##              c(nrow,nrow))      # we preserve dimensions
  ##          )
    return(array(.C64("spamcsrdns",
                      SIGNATURE = c(SS$signature, "double" , SS$signature, SS$signature,
                             "double"),
                 nrow = nrow,
                 entries = x@entries,
                 colindices = xcolindices,
                 rowpointers = x@rowpointers,

                 res = vector_dc( "double", nrow*nrow),

                 INTENT = c("r", "r", "r", "r",
                            "w"),
                 NAOK=getOption("spam.NAOK"),
                 PACKAGE = SS$package)$res,
               c(nrow,nrow))      # we preserve dimensions
         )
}



setMethod("as.matrix","spam.chol.NgPeyton",as.matrix.spam.chol.NgPeyton)
setMethod("as.vector","spam.chol.NgPeyton",
          function(x){
            as.vector.spam(as.spam.chol.NgPeyton(x))
          })


########################################################################
#  force to spam matrices. Would not be required with inheritance

setMethod("image","spam.chol.NgPeyton",
          function(x,cex=NULL,...){
            image.spam(as.spam.chol.NgPeyton(x),cex=cex,...)
          })


setMethod("display","spam.chol.NgPeyton",
          function(x,...){
            display.spam(as.spam.chol.NgPeyton(x),...)
          })

setMethod("t","spam.chol.NgPeyton",
          function(x){
            t.spam(as.spam.chol.NgPeyton(x))
          })

setMethod("chol","spam.chol.NgPeyton",
          function(x){
           x
          })

      subroutine kroneckermult(xnrow,xent,xcol,xrow,
     &     ynrow,yncol,yent,ycol,yrow,
     &     ent, col, row)

      implicit none
      integer xnrow,ynrow,yncol, xcol(*), xrow(*)
      integer   ycol(*), yrow(*), col(*), row(*)

      double precision  xent(*), yent(*), ent(*)

      integer i,k,j,l,n,nr,xdiffi,ydiffk


      n = 1
      nr = 2
      row(1) = 1
      do i = 1,xnrow
         xdiffi = xrow(i+1)-xrow(i)
         do k = 1,ynrow
            ydiffk = yrow(k+1)-yrow(k)
            do j = 1,xdiffi
               do l = 1,ydiffk
                  ent(n) = xent(j+xrow(i)-1)*yent(l+yrow(k)-1)
                  col(n) = ycol(l+yrow(k)-1)+
     &                    (xcol(j+xrow(i)-1)-1)*yncol

                  n=n+1
               enddo
            enddo
            row(nr) = n
            nr = nr+1
         enddo
      enddo

      return
      end


      subroutine kroneckerf(xnrow,xent,xcol,xrow,
     &     ynrow,yncol,yent,ycol,yrow,
     &     ent1, ent2, col, row)

      implicit none
      integer xnrow,ynrow,yncol, xcol(*), xrow(*)
      integer   ycol(*), yrow(*), col(*), row(*)

      double precision  xent(*), yent(*), ent1(*), ent2(*)

      integer i,k,j,l,n,nr,xdiffi,ydiffk


      n = 1
      nr = 2
      row(1) = 1
      do i = 1,xnrow
         xdiffi = xrow(i+1)-xrow(i)-1
         do k = 1,ynrow
            ydiffk = yrow(k+1)-yrow(k)-1
            do j = 0,xdiffi
               do l = 0,ydiffk
                  ent1(n) = xent(j+xrow(i))
                  ent2(n)= yent(l+yrow(k))
                  col(n) = ycol(l+yrow(k))+
     &                    (xcol(j+xrow(i))-1)*yncol

                  n=n+1
               enddo
            enddo
            row(nr) = n
            nr = nr+1
         enddo
      enddo

      return
      end

# HEADER ####################################################
# This is file spam/R/summary.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


summary.spam <- function(object,...) {
    nz <- length(object@entries)
    dens <- nz/prod(object@dimension)*100
    SS <- .format.spam(object, validate = FALSE)
    ans <- list(nnz=nz, density=dens, dim=object@dimension, format=SS)
    class(ans) <- "summary.spam"
    ans
}

summary.spam.chol.NgPeyton <- function(object,...) {
  nrow <- object@dimension[1]
  nnzR <- object@rowpointers[nrow+1]-1
  dens <- nnzR/(nrow^2)
  nnzc <- length(object@colindices)
  nnzA <- object@nnzA
  fill <- nnzR/((nnzA+nrow)/2)
  ans <-  list(nnzR=nnzR, nnzcolindices=nnzc,density=dens,fillin=fill, dim=c(nrow,nrow), nzzA=nnzA)
  class( ans) <- "summary.spam.chol.NgPeyton"
  ans
}

setMethod("summary","spam",summary.spam)
setMethod("summary", "spam.chol.NgPeyton", summary.spam.chol.NgPeyton)

print.summary.spam <- function(x, ...) {
    cat("Matrix object of class 'spam' of dimension ",x$dim[1],"x",
        x$dim[2],",\n",sep="")
    cat("    with ",x$nnz," (row-wise) nonzero elements.\n",sep="")
    cat("    Density of the matrix is ",signif(x$dens,3),"%.\n",sep="")
    cat(paste0("Class 'spam' (", x$format$name, ")\n"))
    invisible(x)
}

print.summary.spam.chol.NgPeyton <- function(x,...) {
  cat("(Upper) Cholesky factor of class 'spam.chol.NgPeyton' of dimension ", x$dim[1],
                "x", x$dim[1], " with ",x$nnzR," (row-wise) nonzero elements.", sep = "", fill=TRUE)
  cat("    Density of the factor is ", signif(x$dens * 100, 3),"%.\n", sep = "")
  cat("    Fill-in ratio is ", signif(x$fill, 3),"\n", sep = "")
  cat("    (Optimal argument for 'chol' is 'memory=list(nnzR=",x$nnzR,
            ifelse(x$nnzA<x$nnzc,paste(",nnzcolindices=",x$nnzc, sep = ""),""),")'.)\n", sep = "")
  cat("Class 'spam.chol.NgPeyton'\n")
  invisible(x)
}

# HEADER ####################################################
# This is file spam/R/xybind.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


########################################################################
"rbind.spam" <-
function(...,deparse.level=0)
{
  force64 <- getOption("spam.force64")

  if (deparse.level!=0) warning("Only 'deparse.level=0' implemented, coerced to zero,")
  addnargs <- ifelse(missing(deparse.level),0,1)

  nargs <- nargs()-addnargs
  if (nargs == 0)     return( NULL)
  args <- list(...)
  if (!is.null( names( args)))  {
    warning("Names of arguments are ignored")
    names( args) <- NULL
  }
  args[which(sapply(args, is.null))] <- NULL

  # nargs needs an update
  nargs <- length(args) -  addnargs


  if (nargs == 0)     return( NULL)
  if (nargs == 1)     return( args[[1]])
  if (nargs == 2) {
    # we distinguish between the cases:
    #    1 spam, spam
    #    2 spam, numeric (scalar, vector, matrix)
    #    3 numeric, spam
    #    4 numeric, numeric

    # Case 1: this is the quick way
    if( is.spam(args[[1]]) & is.spam(args[[2]])) {

      Xdim <- args[[1]]@dimension
      Ydim <- args[[2]]@dimension

      if(Xdim[2]!=Ydim[2])
        stop("Arguments have differing numbers of columns, in rbind.spam()",call.=FALSE)

 #     nrow1 <- args[[1]]@dimension[1]

 #      newx <- new("spam")
 #     newx@entries <- c(args[[1]]@entries, args[[2]]@entries)
 #     newx@colindices <- c(args[[1]]@colindices,  args[[2]]@colindices)
 #     newx@rowpointers <- c(args[[1]]@rowpointers,
 #                           args[[2]]@rowpointers[-1]+args[[1]]@rowpointers[nrow1+1]-as.integer(1))
 #     newx@dimension <- c(nrow1+args[[2]]@dimension[1],args[[1]]@dimension[2])
 #     return(newx)

      XYlen <- as.numeric(args[[1]]@rowpointers[Xdim[1]+1])+as.numeric(args[[2]]@rowpointers[Ydim[1]+1])-2
      if(force64 || XYlen > 2147483647 || .format.spam(args[[1]])$package =="spam64" || .format.spam(args[[2]])$package =="spam64")
          SS <- .format64()
      else
          SS <- .format32

      #xncol,yncol, xlen, ylen,
      #a,ia,ja, b,ib,jb, c,ic,jc
      z <- .C64("rbindf",
                SIGNATURE=c(SS$signature, SS$signature,SS$signature, SS$signature,
                    "double", SS$signature, SS$signature,
                    "double", SS$signature, SS$signature,
                    "double", SS$signature, SS$signature),

                xncol=Xdim[1],
                yncol=Ydim[1],
                xlen=args[[1]]@rowpointers[Xdim[1]+1]-1,
                ylen=args[[2]]@rowpointers[Ydim[1]+1]-1,

                a=args[[1]]@entries,
                ia=args[[1]]@colindices,
                ja=args[[1]]@rowpointers,

                b=args[[2]]@entries,
                ib=args[[2]]@colindices,
                jb=args[[2]]@rowpointers,

                entries=vector_dc("double", XYlen),
                colindices=vector_dc(SS$type, XYlen),
                rowpointers=vector_dc(SS$type, Xdim[1]+Ydim[1]+1),

                INTENT=c("r", "r", "r", "r",
                         "r", "r", "r",
                         "r", "r", "r",
                         "w", "w", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)

      return(.newSpam(
          entries=z$entries,
          colindices=z$colindices,
          rowpointers=z$rowpointers,
          dimension=c(Xdim[1]+Ydim[1], Xdim[2]),
          force64=force64))




    }
    # Case 2:  spam, numeric (scalar, vector, matrix)
    #    if scalar, coherce it first to vector of appropriate length,
    #    if vector, attach dimension.
    if( is.spam(args[[1]]) & is.numeric(args[[2]])) {
      Xdim <- args[[1]]@dimension
      Ylen <- length(args[[2]])
      if (Ylen==1) {
        Xlen <- Xdim[2]
        args[[2]] <- rep( args[[2]], Xlen)
        dim( args[[2]]) <- c(1,Ylen)

      } else   if (is.vector(args[[2]]))
        dim(args[[2]]) <- if( Xdim[1]==1) c(Ylen,1) else c(1,Ylen)
      Ydim <- dim(args[[2]])

      if(Xdim[2]!=Ydim[2])
        stop("Arguments have differing numbers of columns, in rbind.spam()",call.=FALSE)


      newx <- new("spam")
      newx@entries <- c(args[[1]]@entries, as.double(t(args[[2]])))
      newx@colindices <- c(args[[1]]@colindices,  rep.int(as.integer(1:Ydim[2]),Ydim[1]))
      newx@rowpointers <- c(args[[1]]@rowpointers,
                            seq.int(args[[1]]@rowpointers[Xdim[1]+1], by=Ydim[2], length.out=Ydim[1]+1)[-1])
      newx@dimension <- c(Xdim[1]+Ydim[1],Ydim[2])
      return(newx)
    }
    # Case 3:  numeric (scalar, vector, matrix), spam
    #    similar as above
    if( is.numeric(args[[1]]) & is.spam(args[[2]])) {
      Xlen <- length( args[[1]])
      Ydim <- args[[2]]@dimension
      if (Xlen==1) {
        Xlen <- Ydim[2]
        args[[1]] <- rep( args[[1]], Xlen)
        dim( args[[1]]) <- c(1,Xlen)
      } else   if (is.vector(args[[1]]))
        dim(args[[1]]) <- if ( Ydim[1]==1) c(Xlen,1) else  c(1,Xlen)
      Xdim <- dim(args[[1]])

      if(ncol(args[[2]])!=Xdim[2])
        stop("Arguments have differing numbers of columns, in rbind.spam()",call.=FALSE)


      newx <- new("spam")
      newx@entries <- c(as.double(t(args[[1]])), args[[2]]@entries )
      newx@colindices <- c(rep.int(as.integer(1:Xdim[2]),Xdim[1]),
                           args[[2]]@colindices)
      newx@rowpointers <- c(seq.int(1, by=Xdim[2], length.out=Xdim[1]),
                            args[[2]]@rowpointers + Xlen)
      newx@dimension <- c(Ydim[1]+Xdim[1],Ydim[2])
      return(newx)
    }
    # Case 4: numeric,numeric
    #    result is a cleaned spam object.
    if( is.numeric(args[[1]]) & is.numeric(args[[2]]))
      return( as.spam.matrix( rbind(args[[1]],args[[2]]))  )

    stop("Not all argument are of class 'spam' and 'numeric', in rbind.spam()",
         call.=FALSE)
  } else {
    # "recursive" approach only, e.g. no checking
    tmp <- rbind.spam( args[[1]],args[[2]])
    for ( i in 3:nargs)
      tmp <- rbind.spam( tmp, args[[i]])
    return( tmp)
  }
}

cbind.spam <- function(..., deparse.level=0)
{

  force64 <- getOption("spam.force64")

  if (deparse.level!=0) warning("Only 'deparse.level=0' implemented, coerced to zero,")
  addnargs <- ifelse(missing(deparse.level),0,1)

  nargs <- nargs()-addnargs
  if (nargs == 0)     return( NULL)
  args <- list(...)
  if (!is.null( names( args)))  {
    warning("Names of arguments are ignored")
    names( args) <- NULL
  }
  args[which(sapply(args, is.null))] <- NULL
  nargs <- length(args) -  addnargs
  if (nargs == 0)     return( NULL)
  if (nargs == 1)     return( args[[1]])
  if (nargs == 2) {

    Ydim <- if (is.spam(args[[2]])) args[[2]]@dimension  else
          dim(args[[2]])

    if (is.numeric(args[[1]])) {
      # we do _not_ have a spam object
      if (is.vector(args[[1]]))  {
        if (is.null(Ydim)) {
          # if Ydim is NULL, then Y is a vector as well and we need special treatment.
          args[[1]] <- spam.numeric( args[[1]],
                                    max( length(args[[1]]),length(args[[2]])))

        } else {
          # "standard" case, a vector (scalar) with a matrix
          args[[1]] <- spam.numeric( args[[1]], Ydim[1])                  # scalar
        }
      } else {
        # we have a regular matrix
        args[[1]] <- as.spam.matrix(args[[1]])
      }
    } else if (!is.spam(args[[1]])) {
      # we have anything
      stop("Not all argument are of class 'spam' and 'numeric', in cbind.spam()",
           call.=FALSE)
    }

    Xdim <- args[[1]]@dimension

    # now, X=args[[1]] is a spam, the treatment of Y is easier:
    if (is.numeric(args[[2]])) {
      # we do _not_ have a spam object
      if (is.vector(args[[2]]))    args[[2]] <- spam.numeric( args[[2]], Xdim[1])
      else  args[[2]] <- as.spam.matrix(args[[2]])
    } else if (!is.spam(args[[2]])) {
      stop("Not all argument are of class 'spam' and 'numeric', in cbind.spam()",
           call.=FALSE)
    }

    Ydim <- args[[2]]@dimension

    if(Xdim[1]!=Ydim[1])
         stop("Arguments have differing numbers of rows, in cbind.spam()",call.=FALSE)

      XYlen <- as.numeric(args[[1]]@rowpointers[Xdim[1]+1])+as.numeric(args[[2]]@rowpointers[Xdim[1]+1])-2
      if(force64 || XYlen > 2147483647 || .format.spam(args[[1]])$package =="spam64" || .format.spam(args[[2]])$package =="spam64")
          SS <- .format64()
      else
          SS <- .format32

      #xncol,nrow,yncol, clen,
      #a,ia,ja, b,ib,jb, c,ic,jc
      z <- .C64("cbindf",
                SIGNATURE=c(SS$signature, SS$signature,
                    "double", SS$signature, SS$signature,
                    "double", SS$signature, SS$signature,
                    "double", SS$signature, SS$signature),

                xncol=Xdim[2],
                nrow=Xdim[1],

                a=args[[1]]@entries,
                ia=args[[1]]@colindices,
                ja=args[[1]]@rowpointers,

                b=args[[2]]@entries,
                ib=args[[2]]@colindices,
                jb=args[[2]]@rowpointers,

                entries=vector_dc("double", XYlen),
                colindices=vector_dc(SS$type, XYlen),
                rowpointers=vector_dc(SS$type, Xdim[1]+1),

                INTENT=c("r", "r",
                         "r", "r", "r",
                         "r", "r", "r",
                         "w", "w", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)

      return(.newSpam(
          entries=z$entries,
          colindices=z$colindices,
          rowpointers=z$rowpointers,
          dimension=c(Xdim[1], as.numeric(Xdim[2]) + as.numeric(Ydim[2])),
          force64=force64))

  }
  else {
      ## "recursive" approach only
    tmp <- cbind.spam(args[[1]], args[[2]])
    for(i in 3:nargs)
      tmp <- cbind.spam(tmp, args[[i]])
    return(tmp)
  }
}




setMethod("rbind","spam",rbind.spam)
setMethod("cbind","spam",cbind.spam)

# HEADER ####################################################
# This is file spam/tests/testthat/test-xybind.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-xybind.R")



options(spam.printsize = 60)

######################################################################

test_that("rbind", {
    xn <- 3
    xm <- 2
    yn <- 4
    ym <- 2
    set.seed(14)
    
    X <- array(runif(xn*xm), c( xn,xm))
    Y <- array(runif(yn*ym), c( yn,ym))
    
    R <- as.spam(X)
    S <- as.spam(Y)
    
    ## cat("Testing with two matrices:\n")
    spamtest_eq( rbind( X, Y), rbind.spam( X, Y))
    spamtest_eq( rbind( X, Y), rbind.spam( R, Y))
    spamtest_eq( rbind( X, Y), rbind( R, Y))
    spamtest_eq( rbind( X, Y), rbind.spam( X, S))
    spamtest_eq( rbind( X, Y), rbind.spam( R, S))
    spamtest_eq( rbind( X, Y), rbind( R, S))
    spamtest_eq( rbind( X, Y*0), rbind.spam( X, as.spam(Y*0)))
    spamtest_eq( rbind( X*0, Y), rbind.spam( as.spam(X*0), S))
    spamtest_eq( rbind( X*0, Y*0), rbind.spam( as.spam(X*0), as.spam(Y*0)),rel=F)
    spamtest_eq( rbind( X*0, Y*0), rbind.spam( X*0, Y*0),rel=F)
    
    ## cat("Testing with vectors and scalars:\n")
    spamtest_eq( rbind( X, 1:xm), rbind.spam( X, 1:xm))
    spamtest_eq( rbind( X, 1:xm), rbind.spam( R, 1:xm))
    spamtest_eq( rbind( 1:ym, Y), rbind.spam( 1:ym, S))
    spamtest_eq( rbind( 1, Y), rbind.spam( 1, S))
    
    spamtest_eq( rbind( X, 1), rbind.spam( X, 1))
    
    ## cat("Testing with NULL:\n")
    spamtest_eq( rbind( X, NULL), rbind.spam( X, NULL))
    spamtest_eq( rbind( NULL, X, NULL), rbind.spam( NULL, X, NULL))
    spamtest_eq( rbind( NULL, NULL), rbind.spam( NULL,  NULL),rel=F)
})
######################################################################

######################################################################
test_that("cbind", {
    xn <- 3
    xm <- 2
    yn <- 3
    ym <- 4
    set.seed(14)
    
    X <- array(runif(xn*xm), c( xn,xm))
    Y <- array(runif(yn*ym), c( yn,ym))
    
    R <- as.spam(X)
    S <- as.spam(Y)
    
    ## cat("Testing with two matrices:\n")
    spamtest_eq( cbind( X, Y), cbind.spam( X, Y))
    spamtest_eq( cbind( X, Y), cbind.spam( R, Y))
    spamtest_eq( cbind( X, Y), cbind( R, Y))
    spamtest_eq( cbind( X, Y), cbind.spam( X, S))
    spamtest_eq( cbind( X, Y), cbind.spam( R, S))
    spamtest_eq( cbind( X, Y), cbind( R, S))
    spamtest_eq( cbind( X, Y*0), cbind.spam( X, as.spam(Y*0)))
    spamtest_eq( cbind( X*0, Y), cbind.spam( as.spam(X*0), S))
    spamtest_eq( cbind( X*0, Y*0), cbind.spam( as.spam(X*0), as.spam(Y*0)),rel=F)
    spamtest_eq( cbind( X*0, Y*0), cbind.spam( X*0, Y*0),rel=F)
    
    ## cat("Testing with vectors and scalars:\n")
    spamtest_eq( cbind( X, 1:xn), cbind.spam( X, 1:xn))
    spamtest_eq( cbind( X, 1:xn), cbind.spam( R, 1:xn))
    spamtest_eq( cbind( 1:yn, Y), cbind.spam( 1:yn, S))
    spamtest_eq( cbind( 1, Y), cbind.spam( 1, S))
    
    spamtest_eq( cbind( X, 1), cbind.spam( X, 1))
    
    ## cat("Testing with NULL:\n")
    spamtest_eq( cbind( X, NULL), cbind.spam( X, NULL))
    spamtest_eq( cbind( NULL, X, NULL), cbind.spam( NULL, X, NULL))
    spamtest_eq( cbind( NULL, NULL), cbind.spam( NULL,  NULL),rel=F)
})

# HEADER ####################################################
# This is file spam/R/toepliz.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################

"circulant.spam" <- function(x, n=NULL, eps = getOption("spam.eps"))
{
  if (!(is.vector(x)|is.list(x)) )
    stop("'x' is not a vector or a list")

  force64 <- getOption("spam.force64")

  if( is.list(x)) {
    if (!identical(length(x),2L))
      stop("Argument 'x' needs to be a list with two elements")
    if (is.null(n))
      stop("'n' needs to be given")
    ind <- x[[1]]
    x <- x[[2]]
    sel <- (ind <= n)&(abs(x)>eps)
    ind <- ind[sel]
    x <- x[sel]
   
  }else{
    n <- length(x)
    ind <- (1:n)[abs(x) > eps]
    x <- x[ind]
  }
  
  n <- as.integer(n)
  len <- as.integer(length( ind)[1]) # see ?length@value
  if(identical(len,0))
    return(.newSpam(
        # rowpointers = c(1, rep_len64(2, n)), 
        dimension = c(n, n),
        force64 = force64))
#      subroutine circulant(nrow,len, x,j, a,ja,ia)
  nz <- n*len
  ## z <- .Fortran("circulant",
  ##               as.integer(n),
  ##               as.integer(len),
  ##               as.double(x),
  ##               as.integer(ind),
  ##               entries= vector("double", nz),
  ##               colindices = vector("integer", nz),
  ##               rowpointers = vector("integer",  n + 1),
  ##               NAOK = getOption("spam.NAOK"),
  ##               PACKAGE = "spam")
  if(force64 || nz > 2147483647 || n+1 > 2147483647)
      SS <- .format64()
  else
      SS <- .format32
  z <- .C64("circulant",
             SIGNATURE = c(SS$signature, SS$signature, "double", SS$signature,
                 "double", SS$signature, SS$signature),
             
             n,
             len,
             x,
             ind,
            
             entries = vector_dc( "double", nz),
             colindices = vector_dc( SS$type, nz),
             rowpointers = vector_dc( SS$type,  n + 1),

             INTENT = c("r", "r", "r", "r",
                 "rw", "rw", "rw" ),
             NAOK = getOption("spam.NAOK"),
             PACKAGE = SS$package)

                
  ## newx <- new("spam")
  ## slot(newx, "entries", check = FALSE) <- z$entries
  ## slot(newx, "colindices", check = FALSE) <- z$colindices
  ## slot(newx, "rowpointers", check = FALSE) <- z$rowpointers
  ## slot(newx, "dimension", check = FALSE) <- c(n, n)
  ## return(newx)
  return(.newSpam(
      entries = z$entries,
      colindices = z$colindices,
      rowpointers = z$rowpointers,
      dimension = c(n,n),
      force64 = force64
      ))
}

toeplitz.spam <- function(x,y=NULL, eps = getOption("spam.eps"))
{
  force64 <- getOption("spam.force64")

  if (!is.vector(x)) 
    stop("'x' is not a vector")
  n <- length(x)

  if (!is.null(y)){
    if (!identical(length(y),n))
      stop("Length of 'y' and 'x' do not match")
    fullx <- c(rev(y[-1]),x)
  } else {
    fullx <- c(rev(x[-1]),x)
  }

  ind <- (1:(2*n-1))[abs(fullx) > eps]
  fullx <- fullx[ind]

 
  n <- as.integer(n)
  len <- as.integer(length( ind)[1]) # see ?length@value
  if(identical(len,0L)){
      ## print("degenerate")
      return(.newSpam(
          # rowpointers =  c(1, rep_len64(2, n)),
          dimension = c(n, n),
          force64 = force64))
    ## return(new("spam", rowpointers = c(1L, rep.int(2L, n)), 
    ##            dimension = as.integer(c(n, n))))
  }
#      subroutine toeplitz(nrow,len, x,j, a,ja,ia,kk)
  nz <- n*len
    ## z <- .Fortran("toeplitz",
    ##             as.integer(n),
    ##             as.integer(len),
    ##             as.double(fullx),
    ##             as.integer(ind),
    ##             entries= vector("double", nz),
    ##             colindices = vector("integer", nz),
    ##             rowpointers = vector("integer",  n + 1),
    ##             nnz=as.integer(1),
    ##             NAOK = getOption("spam.NAOK"), PACKAGE = "spam")
  if(force64 || n+1 > 2147483647 || nz > 2147483647 )
      SS <- .format64()
  else
      SS <- .format32
  
  z <- .C64("toeplitz",
            SIGNATURE = c(SS$signature, SS$signature, "double", SS$signature,
                "double", SS$signature, SS$signature, SS$signature),

            n,
            len,
            fullx,
            ind,
            
            entries = vector_dc("double", nz),
            colindices = vector_dc( SS$type, nz),
            rowpointers = vector_dc( SS$type,  n + 1),
            nnz = 1,

            INTENT = c("r", "r", "r", "r",
                "w", "w", "w", "rw"),
            NAOK = getOption("spam.NAOK"),
            PACKAGE = SS$package)

                
  ## newx <- new("spam")
  ## slot(newx, "entries", check = FALSE) <- z$entries[1:z$nnz]
  ## slot(newx, "colindices", check = FALSE) <- z$colindices[1:z$nnz]
  ## slot(newx, "rowpointers", check = FALSE) <- z$rowpointers
  ## slot(newx, "dimension", check = FALSE) <- c(n, n)
  ## return(newx)
  return(.newSpam(
      entries = z$entries[1:z$nnz],
      colindices = z$colindices[1:z$nnz],
      rowpointers = z$rowpointers,
      dimension = c(n, n),
      force64 = force64))
}

# HEADER ####################################################
# This is file spam/R/profile.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


.format32 <- list(
    name      = "32-bit",
    type      = "integer", 
    signature = "integer",  
    package   = "spam")

.format64 <- list(
    name      = "64-bit",
    type      = "numeric",
    signature = "int64",
    package   = "spam64")


.format64 <- function(){
    if (!isNamespaceLoaded("spam64")) {
        stop("Large (64-bit) sparse matrices detected. Please load the required package 'spam64' and see the help page '?large_matrix'.")
    }
    list(
        name      = "64-bit",
        type      = "numeric",
        signature = "int64",
        package   = "spam64")
}

.format.spam <- function(x, ... , validate = getOption("spam.validate") ) {
    objects <- c(list(x), list(...))

    if (validate) for(o in objects)  stopifnot(validate_spam(o))
    
    for(o in objects){
        ## If both pointer vectors are of the same type,
        ## use this type to determine the format
        if(identical(typeof(o@colindices), typeof(o@rowpointers))) {
            if(identical(typeof(o@colindices), "double")){
                return(.format64())
            }
            next
        }
        
        ## As fallback use the length of the entries vector and the dimension
        if(nrow(o) > 2147483647 || ncol(o) > 2147483647 || 
           length(o@entries) > 2147483647){
            return(.format64())
        }
    }   
    return(.format32)
}


spam.Version <- function() {
  release <- utils::packageDescription("spam",field="Version")
  date <- utils::packageDescription("spam",field="Date")
  list(status="",
              major=sub("-","",substr(release,1,4)),
              minor=substr(sub("-","",substr(release,5,7)),1,1),
              year=substr(date,1,4),
              month=substr(sub("200.-","",date),1,2),
              day=sub("200.-..-","",date),
              version.string= paste("Spam version ",
                utils::packageDescription("spam",field="Version")," (",
                utils::packageDescription("spam",field="Date"),")",sep="")
              )
}
                
spam.version <- spam.Version()
class(spam.version) <- "simple.list"
  

"inefficiencywarning" <- function(msg,size) {
  maxsize <- if (is.logical(getOption("spam.inefficiencywarning"))) {
    ifelse(getOption("spam.inefficiencywarning"),1,Inf) } else { 
    getOption("spam.inefficiencywarning")
  }
  if (size>maxsize) warning(msg, call. = FALSE)
}
    
".onAttach" <- function (lib, pkg) {
   packageStartupMessage( spam.version$version.string," is loaded.",
       "\nType 'help( Spam)' or 'demo( spam)' for a short introduction ",
       "\nand overview of this package.",
       "\nHelp for individual functions is also obtained by ",
       "adding the\nsuffix '.spam' to the function name, e.g. 'help( chol.spam)'.")
}

.onLoad <- function(libname, pkgname) {
  
  default_options <- list(  spam.eps=.Machine$double.eps,   # smaller than this is considered as zero
     spam.force64=FALSE,
     spam.validate=FALSE,            # validate the spam object before calling a native routine for
                                     # increased stability.
     
     spam.drop=FALSE,                # drop passed to subset functions
     
     spam.printsize=100,             # the max size which we print as regular matrices
     spam.imagesize=10000,           # the max size which we display as regular matrices
     spam.cex=1200,                  # scaling factor for scatter displays

     spam.structurebased=TRUE,      # calculating on nonzero entries only...
     
     spam.inefficiencywarning=1e6,  # tell when something inefficient is done
     
     spam.trivalues=FALSE,           # with upper./lower/.tri return values (TRUE) or only structure?
     spam.listmethod="PE",           # method to be used when using spam.list

     spam.NAOK = FALSE,
     spam.safemodevalidity=TRUE,     # verify while S4 construction
     spam.dopivoting=TRUE,           # what type of back/forwardsolve?
     spam.cholsymmetrycheck=TRUE,     # Should symmetry be tested in the cholesky factorization
     spam.cholpivotcheck=TRUE,        # Should the pivot be tested?
     spam.cholupdatesingular="warning",     # ("error", "warning","NULL")
     spam.cholincreasefactor=c(1.25,1.25),
     spam.nearestdistincreasefactor=1.3,
     spam.nearestdistnnz=c(500^2,500)  )

	# Load the default settings:
	options(default_options)
}

powerboost <- function(flag = "on") {
    if (tolower(flag) %in% c("true","on","an","ein")) {
        options(spam.NAOK = TRUE,
                spam.safemodevalidity = FALSE,
                spam.cholsymmetrycheck = FALSE,
                spam.cholpivotcheck = FALSE,
                spam.eps = 1e-8)
    } else {
         options(spam.NAOK = FALSE,
                spam.safemodevalidity = TRUE,
                spam.cholsymmetrycheck = TRUE,
                spam.cholpivotcheck = TRUE,
                spam.eps = .Machine$double.eps)
    }
    invisible(NULL)
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-profile.R.          #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")

context("test-profile.R")

test_that(".format.spam()",{
    s <- spam(1)
#    s64 <- spam(1, force64 = TRUE)
   
    expect_identical(spam:::.format.spam(s), spam:::.format32)
#    expect_identical(spam:::.format.spam(s64), spam:::.format64)
#    expect_identical(spam:::.format.spam(s, s64), spam:::.format64)
#    expect_identical(spam:::.format.spam(s, s, s, s64), spam:::.format64)
#    expect_identical(spam:::.format.spam(s64, s64, s64), spam:::.format64)
    expect_identical(spam:::.format.spam(s, s, s), spam:::.format32)

    ## test fallback
    s_corrupt <- s; s_corrupt@colindices <- as.double(s_corrupt@colindices)
    expect_identical(spam:::.format.spam(s_corrupt), spam:::.format32,
                     info = "test fallback")

    ## test if validater is called
    a <- spam(1)
    a@entries[1] <- NA
    expect_warning(spam:::.format.spam(a, validate = TRUE),
                   "Slot 'entries' contains non-finite elements.",
                   info = ".format.spam() and validater")
})


## to be added
## test_that(".validate_spam()",{


## })

# HEADER ####################################################
# This is file spam/R/tailhead.R.                           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



head.spam <- function(x, n = 6L, m = n, ...) {
  stopifnot(length(n) == 1L, length(m) == 1L)
   n <- if (n < 0L)
        max(dim(x)[1] + n, 0L)
    else min(n, dim(x)[1])
   m <- if (m < 0L)
        max(dim(x)[2] + m, 0L)
    else min(m, dim(x)[2])
    as.matrix(x[seq_len(n), seq_len(m), drop = FALSE])
}

tail.spam <- function (x, n = 6L, m = n, addrownums = TRUE,  ...)
{
    stopifnot(length(n) == 1L, length(m) == 1L)
    nrx <- dim(x)[1]
    ncx <- dim(x)[2]
    n <- if (n < 0L)
        max(nrx + n, 0L)
    else min(n, nrx)
    m <- if (m < 0L)
        max(ncx + m, 0L)
    else min(m, ncx)
    selr <- seq.int(to = nrx, length.out = n)
    selc <- seq.int(to = ncx, length.out = n)
    ans <- as.matrix( x[selr, selc, drop = FALSE])
    if (addrownums) {
#    if (addrownums && is.null(rownames(x))) { # rownames is null by default
#        rownames(ans) <- paste0("[", selr, ",]") # can be used from R2.15
#        colnames(ans) <- paste0("[,", selc, "]")
        rownames(ans) <- paste("[", selr, ",]", sep = "")
        colnames(ans) <- paste("[,", selc, "]", sep = "")
      }
    ans
}


setMethod("head","spam",head.spam)
setMethod("tail","spam",tail.spam)

setMethod("head","spam.chol.NgPeyton", function(x, n = 6L, m = n, ...) head.spam(as.spam(x), n = 6L, m = n, ...))
setMethod("tail","spam.chol.NgPeyton", function(x, n = 6L, m = n, addrownums = TRUE,  ...) tail.spam(as.spam(x), n = 6L, m = n, addrownums = TRUE,  ...))

# HEADER ####################################################
# This is file spam/R/covmat.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# construct various precision matrices



covmat <- function(h, theta, ... , type="sph") {
  avtype <- c("exponential", "spherical", "nugget",
              "wu1","wu2","wu3","wendland1","wendland2",
              "matern12", "matern32", "matern52", "matern")
  method <- pmatch(tolower(type), avtype)
  if (is.na(method))
     stop("Covariance function not implemented yet. Please ask for.")
  switch(method,
         return(cov.exp(h, theta, ...)),
         return(cov.sph(h, theta, ...)),
         return(cov.nug(h, theta, ...)),
         return(cov.wu1(h, theta, ...)),
         return(cov.wu2(h, theta, ...)),
         return(cov.wu3(h, theta, ...)),
         return(cov.wend1(h, theta, ...)),
         return(cov.wend2(h, theta, ...)),
         return(cov.mat12(h, theta, ...)),
         return(cov.mat32(h, theta, ...)),
         return(cov.mat52(h, theta, ...)),
         return(cov.mat(h, theta, ...)))
}

.par.check.cov <- function(theta,nr=2){
  if (any(theta<0)) {
    warning("Parameters coerced to positive values")
    theta <- abs(theta)
  }
  nt <- length(theta)

  if (nt < nr)
    return( c( theta, rep(1, nr-nt), 0))
  return( c( theta, 0)[1:(nr+1)])
}



cov.sph <- function(h, theta, ..., eps=getOption("spam.eps")) {
# approach: we separate   the cases of h spam versus not.
# we distinguish 'no-nugget'-type and correlation matrix.

  theta <- .par.check.cov(theta)    # checking parameters, range, sill, [nugget]
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
         # range so small, is considered as nugget

  if (is.spam(h)) {
      tmp <- h@entries/theta[1]     # (range==1) is unlikely, we do not test
      if (theta[3] > eps) {         # case of a nugget
          ypos <- tmp <= eps        # nugget cases
          npos <- tmp >= 1          # zero cases
          ypos2 <- !(ypos|npos)     # cov cases

          if (any(ypos)) {
              h@entries[ypos] <- theta[2] + theta[3] }
      }else{
          npos <- tmp >= 1          # zero cases
          ypos2 <- !npos            # cov cases
      }                             # end nugget remaining part remains the same

      if (any(ypos2)) {
          ttmp <- tmp[ypos2]
          h@entries[ypos2] <-  if ( abs(theta[2]-1) > eps) { theta[2]*(1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)
                               }   else { (1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)}
      }
      if (any(npos)) {
          h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
      if (theta[3] > eps) {         # case of a nugget
          ypos <- tmp <= eps        # nugget cases
          npos <- tmp >= 1          # zero cases
          ypos2 <- !(ypos|npos)     # cov cases

          if (any(ypos)) {
              h[ypos] <- theta[2] + theta[3] }
      }else{
          npos <- tmp >= 1          # zero cases
          ypos2 <- !npos            # cov cases
      }                             # end nugget remaining part remains the same

      if (any(ypos2)) {
          ttmp <- tmp[ypos2]
          h[ypos2] <-  if ( abs(theta[2]-1) > eps)  theta[2]*(1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)   else (1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)
       }
      if (any(npos)) {
          h[npos] <- 0 }

  }
  return(h)
}

cor.sph <- function(h, range, ..., eps=getOption("spam.eps")) {
# approach: we separate   the cases of h spam versus not.
# we distinguish 'no-nugget'-type and correlation matrix.

  theta <- .par.check.cov(range,1)  # checking parameters, range
  if (theta[1] < eps) return( cov.nug(h, 1))  # range so small, is considered as nugget
  if (is.spam(h)) {
      tmp <- h@entries/theta[1]     # (range==1) is unlikely, we do not test
      npos <- tmp >= 1          # zero cases
      ypos2 <- !npos            # cor cases

      if (any(ypos2)) {
          ttmp <- tmp[ypos2]
          h@entries[ypos2] <- (1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)
      }
      if (any(npos)) {
          h@entries[npos] <- 0 }

  } else {
      tmp <- c(h)/theta[1]
      npos <- tmp >= 1          # zero cases
      ypos2 <- !npos            # cor cases

      if (any(ypos2)) {
          ttmp <- tmp[ypos2]
          h[ypos2] <-  (1 - 1.5 * ttmp + 0.5 * (ttmp*ttmp)*ttmp)
      }
      if (any(npos)) {
          h[npos] <- 0 }
  }
  return(h)
}

cov.wend1 <- function(h, theta,  ... , eps=getOption("spam.eps")) {
  # is \phi_{3,1} in the 98 paper and \psi_{3,1} in the 95 paper, the latter corresponds also
  # to \phi_{\mu,\kappa} in Bevilacqua.
  # For validity it would only be necessary that \mu >= (d+1)/2 + \kappa. In d=2 we would require here
  # \mu >= 2.5. Most theorem in Bevilacqua require \mu > (d+1)/2 + \kappa +d/2, some  \mu > (d+1)/2 + \kappa + 3
  # here this would mean \mu > 3.5 and > 5.5 !
  theta <- .par.check.cov(theta)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget

  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h@entries[ypos2] <- theta[2]  * ((1 - tmp[ypos2])^4*(4*tmp[ypos2]+1)) }
    if (any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h[ypos2] <- theta[2]  * ((1 - tmp[ypos2])^4*(4*tmp[ypos2]+1)) }
    if (any(npos)) {
      h[npos] <- 0 }
  }

  return(h)
}


cov.wend2 <- function(h, theta,  ..., eps=getOption("spam.eps")) {
  # is \phi_{3,2} in the 98 paper and \psi_{4,2} in the 95 paper
  # See comment for cov.wend1. Here smoothness is increased, k=2 twice mean squared differentiable.
  # Simple add "1" to the values above.
  theta <- .par.check.cov(theta)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
     # range so small, is considered as nugget


  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3]}
    if (any(ypos2)) {
      h@entries[ypos2] <- theta[2] * ((1 - tmp[ypos2])^6*(35*tmp[ypos2]^2+18*tmp[ypos2]+3))/3 }
    if (any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h[ypos2] <- theta[2] * ((1 - tmp[ypos2])^6*(35*tmp[ypos2]^2+18*tmp[ypos2]+3))/3 }
    if (any(npos)) {
      h[npos] <- 0 }
  }

  return(h)
}


cov.wu1 <- function(h, theta, ... ,  eps=getOption("spam.eps")) {
  theta <- .par.check.cov(theta)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget

  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h@entries[ypos2] <- theta[2] * ((1 - tmp[ypos2])^3*(1+3*tmp[ypos2]+tmp[ypos2]^2)) }
    if (any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h[ypos2] <- theta[2] * ((1 - tmp[ypos2])^3*(1+3*tmp[ypos2]+tmp[ypos2]^2)) }
    if (any(npos)) {
      h[npos] <- 0 }
  }

  return( h)
}


cov.wu2 <- function(h, theta,  ... , eps=getOption("spam.eps")) {
  theta <- .par.check.cov(theta)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget

  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h@entries[ypos2] <- theta[2] * ((1 - tmp[ypos2])^4*(4+16*tmp[ypos2]+12*tmp[ypos2]^2+3*tmp[ypos2]^3))/4 }
    if (any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3]}
    if (any(ypos2)) {
      h[ypos2] <- theta[2] * ((1 - tmp[ypos2])^4*(4+16*tmp[ypos2]+12*tmp[ypos2]^2+3*tmp[ypos2]^3))/4 }
    if (any(npos)) {
      h[npos] <-  0 }
  }

  return( h)
}


cov.wu3 <- function(h, theta,  ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget
  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(ypos2)) {
      h@entries[ypos2] <- theta[2] * ((1 - tmp[ypos2])^6*(1+6*tmp[ypos2]+41/3*tmp[ypos2]^2+12*tmp[ypos2]^3+
                                                            5*tmp[ypos2]^4+5/6*tmp[ypos2]^5)) }
    if(any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    ypos2 <- tmp >= eps & tmp < 1
    npos <- tmp >= 1

    if (any(ypos)) {
      h[ypos] <-  theta[2] + theta[3] }
    if (any(ypos2)) {
      h[ypos2] <-  theta[2] * ((1 - tmp[ypos2])^6*(1+6*tmp[ypos2]+41/3*tmp[ypos2]^2+12*tmp[ypos2]^3+
                                                     5*tmp[ypos2]^4+5/6*tmp[ypos2]^5)) }
    if (any(npos)) {
      h[npos] <- 0 }
  }

  return( h)
}


cov.mat <- function(h, theta,  ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta, 3)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget
  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[4] }
    if (any(npos)) {
      h@entries[npos] <- theta[2] * (((2^(-(theta[3] - 1)))/gamma(theta[3])) *
                                       (tmp[npos]^theta[3]) *
                                       besselK(tmp[npos], nu=theta[3])) }
  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[4] }
    if (any(npos)) {
      h[npos] <- theta[2] * (((2^(-(theta[3] - 1)))/gamma(theta[3])) *
                               (tmp[npos]^theta[3]) *
                               besselK(tmp[npos], nu=theta[3])) }
  }

  return(h)
}


cov.finnmat <- function (h, theta, ..., eps = getOption("spam.eps"))
{
  theta <- .par.check.cov(theta, 3)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  if (is.spam(h)) {
    tmp <- sqrt(8 * theta[3])/theta[1] * h@entries
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[4]
    }
    if (any(npos)) {
      h@entries[npos] <- theta[2] * (((2^(-(theta[3] - 1)))/gamma(theta[3])) *
                                       (tmp[npos]^theta[3]) *
                                       besselK(tmp[npos], nu = theta[3]))
    }
  } else {
    tmp <- sqrt(8 * theta[3])/theta[1] * c(h)
    ypos <- tmp < eps
    npos <- !ypos
    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[4]
    }
    if (any(npos)) {
      h[npos] <- theta[2] * (((2^(-(theta[3] - 1)))/gamma(theta[3])) *
                               (tmp[npos]^theta[3]) *
                               besselK(tmp[npos], nu = theta[3])) }
  }

  return(h)
}



cov.exp <- function(h, theta, ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta,2)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget
  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h@entries[npos] <- theta[2] * exp( -tmp[npos]) }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h[npos] <- theta[2] * exp( -tmp[npos])   }
  }

  return(h)
}

cov.mat32 <- function(h, theta, ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta,2)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget
  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h@entries[npos] <- theta[2] * exp( -tmp[npos]) * (1+tmp[npos]) }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h[npos] <- theta[2] * exp( -tmp[npos]) * (1+tmp[npos])  }
  }

  return(h)
}

cov.mat52 <- function(h, theta, ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta,2)
  if (theta[1] < eps) return( cov.nug(h, theta[2] + theta[3]))
  # range so small, is considered as nugget
  if (is.spam(h)) {
    tmp <- h@entries/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h@entries[npos] <- theta[2] * exp( -tmp[npos]) * (1+tmp[npos]+(tmp[npos]^2)/3) }

  } else {
    tmp <- c(h)/theta[1]
    ypos <- tmp < eps
    npos <- !ypos

    if (any(ypos)) {
      h[ypos] <- theta[2] + theta[3] }
    if (any(npos)) {
      h[npos] <- theta[2] * exp( -tmp[npos]) * (1+tmp[npos]+(tmp[npos]^2)/3) }
  }

  return(h)
}

cov.mat12 <- cov.exp

cov.nug <- function(h, theta, ..., eps=getOption("spam.eps")) {

  theta <- .par.check.cov(theta,0)
  if (is.spam(h)) {
    ypos <- h@entries < eps
    npos <- !ypos

    if (any(ypos)) {
      h@entries[ypos] <- theta[1] }
    if (any(npos)) {
      h@entries[npos] <- 0 }

  } else {
    ypos <- c(h) < eps
    npos <- !ypos

    if (any(ypos)) {
      h[ypos] <- theta[1] }
    if (any(npos)) {
      h[npos] <- 0 }
  }

  return(h)
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-covmat.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
     
rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-covmat.R")


test_that("cov.sph", {
    nl <- 10
    range <- .5
    h <- nearest.dist(cbind( runif(nl), runif(nl)),  delta=range, upper=NULL)
    expect_identical( cov.sph(h, c(range, 1, 0)),   cor.sph(h, range) )

    powerboost()
    expect_identical( cov.sph(h, c(range, 1+1e-10 )),     cov.sph(h, range) )
    expect_identical( cov.sph(h, c(range, 1+1e-10 )),     cov.sph(h, range) )
    expect_identical( cov.sph(h, c(range, 1+1e-9, 1e-9)), cov.sph(h, range) )

    powerboost("aus")    
    expect_equal( cov.sph(h, c(range, 1+1e-9, 1e-9)), cov.sph(h, range) )
})


test_that("cov.*", {
    h <- nearest.dist(100*1:10, 100*1:10+1:10, delta=10)
    expect_identical(cov.exp(1:10, 10), cov.exp(h, 10)@entries)
    expect_identical(cov.sph(1:10, 10), cov.sph(h, 10)@entries)
    expect_identical(cov.nug(1:10, 10), cov.nug(h, 10)@entries)
    expect_identical(cov.wu1(1:10, 10), cov.wu1(h, 10)@entries)
    expect_identical(cov.wu2(1:10, 10), cov.wu2(h, 10)@entries)
    expect_identical(cov.wu3(1:10, 10), cov.wu3(h, 10)@entries)
    expect_identical(cov.wend1(1:10, 10), cov.wend1(h, 10)@entries)
    expect_identical(cov.wend2(1:10, 10), cov.wend2(h, 10)@entries)
    expect_identical(cov.mat(1:10, 10), cov.mat(h, 10)@entries)
})

test_that("zero range", {
  set.seed(42)
  n <- 100
  sampleData <- matrix(rnorm(n), nrow = sqrt(n))
  sampleData[4:6, 6:8] <- sampleData[4:6, 6:8] + 5

  x <- 1:10
  locs <- expand.grid(x, x)
  dim1 <- dim2 <- 10
  distmat <- nearest.dist( locs, upper=NULL, delta = 1000) # distance matrix
  out <- spam::mle.nomean.spam(y = (c(sampleData)), distmat = distmat, Covariance =
                          cov.mat,
                        theta0 = c(2, 1, 1, .01),
                        thetalower = c(0, 0.1, 0.01, 0.01),
                        thetaupper = c(5, 5, 5, .2))

  expect_equal(round(out$par,4), c(0, 3.5215, 0.01, 0.2))
})




test_that("cov.matXY*", {
    h <- seq(0, to=10, len=150)
    theta <- c(1.4, 1, 1)
    theta1 <- c(1.4, 1, 0.5, 1)
    theta3 <- c(1.4, 1, 1.5, 1)
    theta5 <- c(1.4, 1, 2.5, 1)
    
    expect_equal(cov.exp(h, theta), cov.mat(h,theta1))
    expect_equal(cov.mat12(h, theta), cov.mat(h,theta1))
    expect_equal(cov.mat32(h, theta), cov.mat(h,theta3))
    expect_equal(cov.mat52(h, theta), cov.mat(h,theta5))
})


theta <- c(1.4, 1, 1)
theta1 <- c(1.4, 1, 0.5, 1)
theta3 <- c(1.4, 1, 1.5, 1)
theta5 <- c(1.4, 1, 2.5, 1)
# h  <- seq(0, to=10, len=10000)
# microbenchmark::microbenchmark(a=cov.exp(h, theta), a2=cov.mat(h,theta1), b=cov.mat32(h, theta), b2=cov.mat(h,theta3), c=cov.mat52(h, theta), c2=cov.mat(h,theta5))

# HEADER ####################################################
# This is file spam/tests/testthat.R.                       #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

library("testthat")
# library("spam")

test_check("spam")

# HEADER ####################################################
# This is file spam/R/mle.R.                                #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

     



neg2loglikelihood.spam <- function(y, X, distmat, Covariance,
                                   beta, theta, Rstruct = NULL, cov.args=NULL,...) {

  Sigma <- do.call(Covariance, c(list(distmat, theta), cov.args))
  if (!is.spam(Sigma)){
    warning("\"Covariance\" should return a spam object. Forced to spam.")
    Sigma <- as.spam(Sigma)
  }
  if (is(Rstruct, "spam.chol.NgPeyton")) 
    cholS <- update.spam.chol.NgPeyton(Rstruct, Sigma, ...)
  else cholS <- chol.spam(Sigma, ...)

  n <- length(y)
  resid <- y-X%*%beta
  return( n * log(2*pi) +
         2*c(determinant.spam.chol.NgPeyton(cholS)$modulus) +
         sum(resid * solve.spam( cholS, resid))
         )
}


neg2loglikelihood <- function(y, X, distmat, Covariance,
                                   beta, theta, cov.args=NULL, ...) {

#   Sigma <- do.call(Covariance,list(distmat,theta))
  Sigma <- do.call(Covariance, c(list(distmat,theta),cov.args))
  cholS <- chol(Sigma, ...)
  logdet <- sum(log(diag(cholS)))
  
  n <- length(y)
  resid <- y-X%*%beta
  
  return( n * log(2*pi) +
         2*logdet +
         sum(resid * backsolve(cholS, forwardsolve(cholS, resid,
            transpose=TRUE, upper.tri=TRUE),n))
         )
}


neg2loglikelihood.nomean <- function (y, distmat, Covariance,
                                      theta, cov.args=NULL, ...) {
# 	Sigma <- do.call(Covariance, list(distmat, theta))
    Sigma <- do.call(Covariance, c(list(distmat,theta),cov.args))
    cholS <- chol(Sigma, ...)
    logdet <- sum(log(diag(cholS)))
    
    n <- length(y)
    return(n * log(2 * pi) + 2 * logdet + sum(y * backsolve(cholS, 
	    forwardsolve(cholS, y, transpose = TRUE, upper.tri = TRUE), n)))
}




mle.spam <- function(y, X, distmat, Covariance,
                     beta0, theta0,
                     thetalower, thetaupper, optim.control=NULL,
                     Rstruct = NULL, hessian = FALSE,cov.args=NULL, ...) {
  
  
  if (!is(Rstruct, "spam.chol.NgPeyton")) {
#    Sigma <- do.call(Covariance,
#                     list(distmat,c(thetaupper[1],theta0[-1])))
    Sigma <- do.call(Covariance,
                     c(list(distmat,c(thetaupper[1],theta0[-1])),cov.args))
    if (!is.spam(Sigma))
      stop("\"Covariance\" should return a spam object.")
   
    Rstruct <- chol.spam(Sigma, ...)
  }
  
  p <- dim(X)[2]
  n <- length(y)
    
  neg2loglikelihood <- function(fulltheta,...) {
#      Sigma <- do.call(Covariance,list(distmat,fulltheta[-(1:p)]))
    Sigma <- do.call(Covariance,c(list(distmat,fulltheta[-(1:p)]),cov.args))
    cholS <- update.spam.chol.NgPeyton(Rstruct, Sigma, ...)

    resid <- y-X%*%fulltheta[1:p]
    return( n * log(2*pi) +
           2*c(determinant.spam.chol.NgPeyton(cholS)$modulus) +
           sum(resid * solve.spam( cholS, resid))
           )
  }

  return(optim(c(beta0,theta0),neg2loglikelihood,
               method = "L-BFGS-B",control = optim.control,
               lower=c(rep(-Inf,p),thetalower),
               upper=c(rep(Inf,p),thetaupper), hessian = hessian))
               
}

mle <- function(y, X, distmat, Covariance,
                beta0, theta0,
                thetalower, thetaupper, optim.control=NULL,
                hessian = FALSE,cov.args=NULL, 
                ...) {
    
  p <- dim(X)[2]
  n <- length(y)
    
  neg2loglikelihood <- function(fulltheta,...) {
#      Sigma <- do.call(Covariance,list(distmat,fulltheta[-(1:p)]))
          Sigma <- do.call(Covariance,c(list(distmat,fulltheta[-(1:p)]),cov.args))
    cholS <- chol(Sigma, ...)
    logdet <- sum(log(diag(cholS)))

    resid <- y-X%*%fulltheta[1:p]
    return( n * log(2*pi) +
           2*logdet +
           sum(resid * backsolve(cholS, forwardsolve(cholS, resid,
              transpose=TRUE, upper.tri=TRUE),n))
           )
  }

  return(optim(c(beta0,theta0),neg2loglikelihood,
               method = "L-BFGS-B",control = optim.control,
               lower=c(rep(-Inf,p),thetalower),
               upper=c(rep(Inf,p),thetaupper), hessian = hessian))
               
}


mle.nomean.spam <- function(y, distmat, Covariance,
                     theta0,
                     thetalower, thetaupper, optim.control = NULL,
                     Rstruct = NULL, hessian = FALSE, cov.args=NULL, ...) {
  
  
  if (!is(Rstruct, "spam.chol.NgPeyton")) {
#    Sigma <- do.call(Covariance,
#                     list(distmat,c(thetaupper[1],theta0[-1])))
    Sigma <- do.call(Covariance,
                     c(list(distmat,c(thetaupper[1],theta0[-1])), cov.args))
    if (!is.spam(Sigma))
      stop("\"Covariance\" should return a spam object.")
   
    Rstruct <- chol.spam(Sigma, ...)
  }
  
  n <- length(y)
    
  neg2loglikelihood <- function(theta,...) {
#    Sigma <- do.call(Covariance,list(distmat,theta))
    Sigma <- do.call(Covariance,
                     c(list(distmat,theta), cov.args))
    cholS <- update.spam.chol.NgPeyton(Rstruct, Sigma, ...)

    return( n * log(2*pi) +
           2*c(determinant.spam.chol.NgPeyton(cholS)$modulus) +
           sum(y * solve.spam( cholS, y))
           )
  }

  return(optim(theta0,neg2loglikelihood,
               method = "L-BFGS-B",control = optim.control,
               lower=thetalower,    upper=thetaupper, hessian = hessian))
               
}




mle.nomean <- function(y, distmat, Covariance,
                       theta0,
                       thetalower, thetaupper, optim.control=NULL,
                       hessian = FALSE, cov.args=NULL,
                       ...) {
    
  n <- length(y)
    
  neg2loglikelihood <- function(theta,...) {
 #   Sigma <- do.call(Covariance,list(distmat,theta))
    Sigma <- do.call(Covariance,
                     c(list(distmat,theta), cov.args))
    cholS <- chol(Sigma, ...)
    logdet <- sum(log(diag(cholS)))

    return( n * log(2*pi) +
           2*logdet +
           sum(y * backsolve(cholS, forwardsolve(cholS, y,
              transpose=TRUE, upper.tri=TRUE),n))
           )
  }

  return(optim(theta0,neg2loglikelihood,
               method = "L-BFGS-B",control = optim.control,
               lower=thetalower,    upper=thetaupper, hessian = hessian))
               
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-mle.R.              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())     
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-mle.R")


################## _DO NOT CHANGE THE PARAMETERS_ #################
# Optimization uses these values for a quick run through!!!


truebeta <- c(1,2,.2)
truetheta <- c(.5,2,.02)

spherical <- function(distmat, theta, eps = 1e-06) {
  Sigma <- distmat
  d <- Sigma@entries/theta[1]
 
  Sigma@entries <- ifelse(d < eps,
                          theta[3]+ theta[2],
                          ifelse(d < 1, theta[2]*(1 - 1.5*d + 0.5*d^3), 0))
  return( Sigma)
}
sphericalmat <- function(distmat, theta, eps = 1e-06) {
  Sigma <- distmat
  d <- Sigma@entries/theta[1]
 
  Sigma@entries <- ifelse(d < eps,
                          theta[3]+ theta[2],
                          ifelse(d < 1, theta[2]*(1 - 1.5*d + 0.5*d^3), 0))
  return( as.matrix(Sigma))
}

xl <- 10
x <- seq(0,1,l=xl)
locs <- expand.grid(x,x)
X <- as.matrix(cbind(1,locs))  # design matrix
cov.sph.mat <- function(...)
  as.matrix(cov.sph(...))     # covariance function

distmat <- nearest.dist(locs,upper=NULL) # distance matrix
Sigma <- cov.sph(distmat,truetheta)    # true covariance matrix


set.seed(15)
y <- c(rmvnorm.spam(1,X%*%truebeta,Sigma)) # construct samples


test_that("mle", {
    spamtest_eq(round(neg2loglikelihood.spam( y, X, distmat, cov.sph,
                                             truebeta, truetheta),2), 262.98)
    
    spamtest_eq(round(neg2loglikelihood( y, X, distmat, cov.sph,
                                        truebeta, truetheta),2), 262.98)
    
    spamtest_eq(round(neg2loglikelihood( y, X, distmat, cov.sph.mat,
                                        truebeta, truetheta),2), 262.98)
    
#    spamtest_eq(round(neg2loglikelihood.spam( y, X, distmat, cov.sph.mat,
#                                             truebeta, truetheta),2), 262.98)

#    expect_warning(round(neg2loglikelihood.spam( y, X, distmat, cov.sph.mat,
#                                             truebeta, truetheta),2))
    
})
# we pass now to the mle:

test_that("mle2", {
# not that we should set:
#    ,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf)
# for quicker testing we use
    res1 <- mle.spam(y, X, distmat, cov.sph,
                     truebeta, truetheta,thetalower=c(0.4,1.5,0.02),thetaupper=c(.6,2.5,.1))
                                        #         truebeta, truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
    
    betahat <- res1$par[1:3]
    spamtest_eq(round(res1$par,2), c(2.35,  1.45, -0.58,  0.50,  1.70,  0.08))
    spamtest_eq(round(res1$val,2), 259.03)
    
    if (F){ # takes too long...
        res2 <- mle(y, X, distmat, cov.sph,
                    truebeta, truetheta,thetalower=c(0.4,1.5,0.02),thetaupper=c(.6,2.5,.1))
                                        #         truebeta, truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
        
        res3 <- mle(y, X, distmat, cov.sph.mat,
                    truebeta, truetheta,thetalower=c(0.4,1.5,0.02),thetaupper=c(.6,2.5,.1))
                                        #         truebeta, truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
        
        
        spamtest_eq(round(res2$par,2), c(2.35,  1.45, -0.58,  0.50,  1.70,  0.08))
        spamtest_eq(round(res2$val,2), 259.03)
        
        spamtest_eq(round(res3$par,2), c(2.35,  1.45, -0.58,  0.50,  1.70,  0.08))
        spamtest_eq(round(res3$val,2), 259.03)
    }
    
    
    res1 <- mle.nomean.spam(y-X%*%betahat, distmat, cov.sph,
                                        #         truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
                            truetheta,thetalower=c(0.4,1,0.02),thetaupper=c(.6,2.5,.1))

    res2 <- mle.nomean(y-X%*%betahat, distmat, cov.sph,
                       truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
    
    res3 <- mle.nomean(y-X%*%betahat, distmat, cov.sph.mat,
                       truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))
    
    spamtest_eq(round(res1$par,2), c(  0.50,  1.70,  0.08))
    spamtest_eq(round(res1$val,2), 259.03)
    
    spamtest_eq(round(res2$par,2), c(  0.50,  1.70,  0.08))
    spamtest_eq(round(res2$val,2), 259.03)
    
    spamtest_eq(round(res3$par,2), c(  0.50,  1.70,  0.08))
    spamtest_eq(round(res3$val,2), 259.03)
})



test_that("cov.args", {

 
    expect_equal(
        neg2loglikelihood.spam( y, X, distmat, cov.sph, truebeta, truetheta, cov.args=list(eps=1e-15)),
        neg2loglikelihood.spam( y, X, distmat, cov.sph, truebeta, truetheta)
        )
    })

# HEADER ####################################################
# This is file spam/tests/demo_cholesky.R.                  #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# We illustrate the Cholesky decompostion approaches
options( echo=FALSE)
library( spam, warn.conflict=FALSE)

set.seed(14)




# first start with a full matrix.
xn <- 50
fmat1 <- matrix(rnorm(xn*xn),xn,xn)
fmat1 <- t( fmat1) %*% fmat1
smat1 <- as.spam(fmat1)
smat2 <- smat1 + diag.spam(xn)

# Generic Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, reusing pivoting
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, updating
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# reset to default
options(spam.cholsymmetrycheck=TRUE)




# now create a sparse matrix.
fmat1[fmat1<3] <- 0
smat1 <- as.spam(fmat1)
smat2 <- smat1 + diag.spam(xn)

# Generic Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, reusing pivoting
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, updating
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# reset to default
options(spam.cholsymmetrycheck=TRUE)



# now create an even sparser matrix.
fmat1 <- fmat1+20*diag(xn)
fmat1[fmat1<32] <- 0
smat1 <- as.spam(fmat1)
smat2 <- smat1 + 1* diag.spam(xn)


# Generic Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( fmat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, direct call
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)

# Sparse Cholesky, without symmetry check
options(spam.cholsymmetrycheck=FALSE)
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, reusing pivoting
options(spam.cholsymmetrycheck=FALSE)
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- chol.spam( smat1,pivot=ch1@pivot)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# Sparse Cholesky, updating
options(spam.cholsymmetrycheck=FALSE)
# tmp <- gc(F);Rprof(memory.profiling=TRUE, interval = 0.01)
ch1 <- update.spam.chol.NgPeyton( ch1, smat2)
# Rprof(NULL);print( summaryRprof(memory="both")$by.total)


# reset to default
options(spam.cholsymmetrycheck=TRUE)


options( echo=TRUE)

# HEADER ####################################################
# This is file spam/R/apply.R.                              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# primitive apply function.

apply.spam <- function(X, MARGIN=NULL, FUN, ...){
  if (!is.spam(X)) stop("\"X\" must be of type \"spam\"")
  if (!is.null(dimnames(X))) warning("dimnames are stripped")
  FUN <- match.fun(FUN)
  d <- dim(X)
  d.ans <- d[MARGIN]
  dn.ans <- NULL
  if (is.null(MARGIN)|| length(MARGIN)==2){
    ans <- FUN(X@entries,...)
    if (length( ans)!=length( X@entries))
      stop("\"FUN\" does not return an appropriate vector")
    if (any(!is.finite(ans))) {
      warning("\"NA/NaN/Inf\" coerced to zero")
      ans[!is.finite(ans)] <- 0
    }

    X@entries <- ans
    return(X)
  }
  ans <- vector("list",d.ans)
  if (MARGIN==1) {
    for (i in 1:d[1])
      ans[[i]] <- FUN(X[i,,drop=F]@entries,...)
  } else   if (MARGIN==2) {
    for (i in 1:d[2])
      ans[[i]] <- FUN(X[,i,drop=F]@entries,...)
  } else stop("\"MARGIN\" must be 1, 2 or c(1,2)")


  # Block very similar to "apply"
  d2 <- prod(d.ans)
  ans.list <- is.recursive(ans[[1]])
  l.ans <- length(ans[[1]])
  ans.names <- names(ans[[1]])

  if (!ans.list){
    ans.list <- any(unlist(lapply(ans, length)) != l.ans)
  }
  if (!ans.list && length(ans.names)) {
    all.same <- sapply(ans, function(x) identical(names(x), ans.names))
    if (!all(all.same))
      ans.names <- NULL
  }
  len.a <- if (ans.list)     d2   else length(ans <- unlist(ans, recursive = FALSE))
  if (length(MARGIN) == 1 && len.a == d2)
    return(ans)
  if (len.a == d2)
    return(array(ans, d.ans))
  if (len.a > 0 && len.a%%d2 == 0) {
    dn.ans <- vector(mode = "list", length(d.ans))
    dn.ans <- c(list(ans.names), dn.ans)
    return(array(ans, c(len.a%/%d2, d.ans), if (!all(sapply(dn.ans,
                                                            is.null))) dn.ans))
  }
  return(ans)


}

c     It has been tested that embedding the loop over the right hand
c     side into the backsolve routine is not faster.

c-----------------------------------------------------------------------
      subroutine getbwd(n,  ja,ia,ml,mu)
c-----------------------------------------------------------------------
c gets the bandwidth of lower part and upper part of A.
c does not assume that A is sorted.
c-----------------------------------------------------------------------
c on entry:
c----------
c n	= integer = the row dimension of the matrix
c ja, ia = matrix in compressed sparse row format.
c
c on return:
c-----------
c ml	= integer. The bandwidth of the strict lower part of A
c mu	= integer. The bandwidth of the strict upper part of A
c
c Notes:
c ===== ml and mu are allowed to be negative or return. This may be
c       useful since it will tell us whether a band is confined
c       in the strict  upper/lower triangular part.
c       indeed the definitions of ml and mu are
c
c       ml = max ( (i-j)  s.t. a(i,j) .ne. 0  )
c       mu = max ( (j-i)  s.t. a(i,j) .ne. 0  )
c----------------------------------------------------------------------c
c Y. Saad, Sep. 21 1989 | simplified by R. Furrer Sept. 2016           c
c----------------------------------------------------------------------c
      implicit none
c     double precision a(*)
      integer n,ja(*),ia(n+1),ml,mu

      integer ldist,i,k

      ml = - n
      mu = - n
      do 3 i=1,n
         do 31 k=ia(i),ia(i+1)-1
            ldist = i-ja(k)
            ml = max(ml,ldist)
            mu = max(mu,-ldist)
 31      continue
 3    continue
      return
c---------------end-of-getbwd ------------------------------------------
c-----------------------------------------------------------------------
      end


c functions slightly modified from sparsekit:
c cperm,rperm,dperm: job argument is eliminated
c-----------------------------------------------------------------------


c-----------------------------------------------------------------------
      subroutine rperm (nrow,a,ja,ia,ao,jao,iao,perm)
      implicit none
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(nrow)
      double precision a(*),ao(*)
c-----------------------------------------------------------------------
c this subroutine permutes the rows of a matrix in CSR format.
c rperm  computes B = P A  where P is a permutation matrix.
c the permutation P is defined through the array perm: for each j,
c perm(j) represents the destination row number of row number j.
c Youcef Saad -- recoded Jan 28, 1991.
c-----------------------------------------------------------------------
c on entry:
c----------
c n 	= dimension of the matrix
c a, ja, ia = input matrix in csr format
c perm 	= integer array of length nrow containing the permutation arrays
c	  for the rows: perm(i) is the destination of row i in the
c         permuted matrix.
c         ---> a(i,j) in the original matrix becomes a(perm(i),j)
c         in the output  matrix.
c
c
c------------
c on return:
c------------
c ao, jao, iao = input matrix in a, ja, ia format
c----------------------------------------------------------------------c
c           Y. Saad, May  2, 1990                                      c
c----------------------------------------------------------------------c

      integer i,j,k,ko,ii
c     determine pointers for output matix.
c
      do 50 j=1,nrow
         i = perm(j)
         iao(i+1) = ia(j+1) - ia(j)
 50   continue
c
c get pointers from lengths
c
      iao(1) = 1
      do 51 j=1,nrow
         iao(j+1)=iao(j+1)+iao(j)
 51   continue
c
c copying
c
      do 100 ii=1,nrow
c
c old row = ii  -- new row = iperm(ii) -- ko = new pointer
c
         ko = iao(perm(ii))
         do 60 k=ia(ii), ia(ii+1)-1
            jao(ko) = ja(k)
            ao(ko) = a(k)
            ko = ko+1
 60      continue
 100  continue
c
      return
c---------end-of-rperm -------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine cperm (nrow,a,ja,ia,ao,jao,iao,perm)
      implicit none
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(*)
      double precision a(*), ao(*)
c-----------------------------------------------------------------------
c this subroutine permutes the columns of a matrix a, ja, ia.
c the result is written in the output matrix  ao, jao, iao.
c cperm computes B = A P, where  P is a permutation matrix
c that maps column j into column perm(j), i.e., on return
c      a(i,j) becomes a(i,perm(j)) in new matrix
c Y. Saad, May 2, 1990 / modified Jan. 28, 1991.
c-----------------------------------------------------------------------
c on entry:
c----------
c nrow 	= row dimension of the matrix
c
c a, ja, ia = input matrix in csr format.
c
c perm	= integer array of length ncol (number of columns of A
c         containing the permutation array  the columns:
c         a(i,j) in the original matrix becomes a(i,perm(j))
c         in the output matrix.
c
c
c------------
c on return:
c------------
c ao, jao, iao = input matrix in a, ja, ia format
c
c Notes:
c-------
c 1. if job=1 then ao, iao are not used.
c 2. This routine is in place: ja, jao can be the same.
c 3. If the matrix is initially sorted (by increasing column number)
c    then ao,jao,iao  may not be on return, hence a call to csort.
c
c----------------------------------------------------------------------c
c local parameters:
      integer k, i, nnz
c
      nnz = ia(nrow+1)-1
      do 100 k=1,nnz
         jao(k) = perm(ja(k))
 100  continue
c
c     done with ja array.
c
      do 1 i=1, nrow+1
         iao(i) = ia(i)
 1    continue
c
      do 2 k=1, nnz
         ao(k) = a(k)
 2    continue
c
      call sortrows(nrow,ao,jao,iao)
c     call csort (nrow,ao,jao,iao,iwork) _does not work_
      return
c---------end-of-cperm--------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine dperm (nrow,a,ja,ia,ao,jao,iao,pperm,qperm)
      implicit none
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),pperm(nrow),
     +        qperm(*)
      double precision a(*),ao(*)
c-----------------------------------------------------------------------
c This routine permutes the rows and columns of a matrix stored in CSR
c format. i.e., it computes P A Q, where P, Q are permutation matrices.
c P maps row i into row perm(i) and Q maps column j into column qperm(j):
c      a(i,j)    becomes   a(pperm(i),qperm(j)) in new matrix
c note that qperm should be of length ncol (number of columns) but this
c is not checked.
c-----------------------------------------------------------------------
c Y. Saad, Sep. 21 1989 / recoded Jan. 28 1991.
c-----------------------------------------------------------------------
c on entry:
c----------
c n 	= dimension of the matrix
c a, ja,
c    ia = input matrix in a, ja, ia format
c pperm 	= integer array of length n containing the permutation arrays
c	  for the rows: pperm(i) is the destination of row i in the
c         permuted matrix
c
c qperm	= same thing for the columns.
c iwork = working array passed to cperm
c
c on return:
c-----------
c ao, jao, iao = input matrix in a, ja, ia format
c
c Notes:
c-------
c  1) algorithm is in place
c----------------------------------------------------------------------c
c local variables
c
c permute rows first
c
      call rperm (nrow,a,ja,ia,   ao,jao,iao,pperm)
c
c then permute columns
c
c
      call cperm (nrow,ao,jao,iao,ao,jao,iao,qperm)
c
      return
c-------end-of-dperm----------------------------------------------------
c-----------------------------------------------------------------------
      end



c-----------------------------------------------------------------------
      subroutine dvperm (n, x, perm)
      implicit none
      integer n
      integer perm(n)
      double precision x(n)
c-----------------------------------------------------------------------
c this subroutine performs an in-place permutation of a real vector x
c according to the permutation array perm(*), i.e., on return,
c the vector x satisfies,
c
c	x(perm(j)) :== x(j), j=1,2,.., n
c
c-----------------------------------------------------------------------
c on entry:
c---------
c n 	= length of vector x.
c perm 	= integer array of length n containing the permutation  array.
c x	= input vector
c
c on return:
c----------
c x	= vector x permuted according to x(perm(*)) :=  x(*)
c
c----------------------------------------------------------------------c
c           Y. Saad, Sep. 21 1989                                      c
c----------------------------------------------------------------------c
c local variables
      integer init,next,k, ii, j
      double precision tmp, tmp1
c
      init      = 1
      tmp       = x(init)
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
c
c loop
c
 6    k = k+1
c
c save the chased element --
c
      tmp1      = x(ii)
      x(ii)     = tmp
      next      = perm(ii)
      if (next .lt. 0 ) goto 65
c
c test for end
c
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next
c
c end loop
c
      goto 6
c
c reinitilaize cycle --
c
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp = x(init)
      ii = perm(init)
      perm(init)=-perm(init)
      goto 6
c
 101  continue
      do 200 j=1, n
            perm(j) = -perm(j)
 200  continue
c
      return
c-------------------end-of-dvperm---------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine ivperm (n, ix, perm)
      implicit none
      integer n, perm(n+1), ix(n)
c-----------------------------------------------------------------------
c this subroutine performs an in-place permutation of an integer vector
c ix according to the permutation array perm(*), i.e., on return,
c the vector x satisfies,
c
c	ix(perm(j)) :== ix(j), j=1,2,.., n
c
c-----------------------------------------------------------------------
c on entry:
c---------
c n 	= length of vector x.
c perm 	= integer array of length n containing the permutation  array.
c ix	= input vector
c
c on return:
c----------
c ix	= vector x permuted according to ix(perm(*)) :=  ix(*)
c
c----------------------------------------------------------------------c
c           Y. Saad, Sep. 21 1989                                      c
c----------------------------------------------------------------------c
c local variables
      integer ii,k,j,next,init,tmp, tmp1
c
      init      = 1
      tmp       = ix(init)
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
c
c loop
c
 6    k = k+1
c
c save the chased element --
c
      tmp1     = ix(ii)
      ix(ii)   = tmp
      next     = perm(ii)
      if (next .lt. 0 ) goto 65
c
c test for end
c
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next
c
c end loop
c
      goto 6
c
c reinitilaize cycle --
c
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp       = ix(init)
      ii        = perm(init)
      perm(init)=-perm(init)
      goto 6
c
 101  continue
      do 200 j=1, n
c         if (perm(j) .lt. 0) then
            perm(j) = -perm(j)
c         endif
 200  continue
c
      return
c-------------------end-of-ivperm---------------------------------------
c-----------------------------------------------------------------------
      end
c
c-----------------------------------------------------------------------
      subroutine aplbdg (nrow,ncol,ja,ia,jb,ib,ndegr,nnz,iw)

      implicit none
      integer nrow, ncol, nnz
      integer ja(*),jb(*),ia(nrow+1),ib(nrow+1),iw(ncol),ndegr(nrow)
c-----------------------------------------------------------------------
c gets the number of nonzero elements in each row of A+B and the total
c number of nonzero elements in A+B.
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format.
c
c b,
c jb,
c ib	=  Matrix B in compressed sparse row format.
c
c iw,nnz,ngegr = zero content
c
c on return:
c----------
c ndegr	= integer array of length nrow containing the degrees (i.e.,
c         the number of nonzeros in  each row of the matrix A + B.
c
c nnz   = total number of nonzero elements found in A * B
c
c work arrays:
c------------
c iw	= integer work array of length equal to ncol.
c
c-----------------------------------------------------------------------
      integer k,j,ii,jr,last,ldg,jc

      do 7 ii=1,nrow
         ldg = 0
c
c    end-of-linked list
c
         last = -1
c
c     row of A
c
         do 5 j = ia(ii),ia(ii+1)-1
            jr = ja(j)
c
c     add element to the linked list
c
            ldg = ldg + 1
            iw(jr) = last
            last = jr
 5       continue
c
c     row of B
c
         do 6 j=ib(ii),ib(ii+1)-1
            jc = jb(j)
            if (iw(jc) .eq. 0) then
c
c     add one element to the linked list
c
               ldg = ldg + 1
               iw(jc) = last
               last = jc
            endif
 6       continue
c     done with row ii.
         ndegr(ii) = ldg
c
c     reset iw to zero
c
         do 61 k=1,ldg
            j = iw(last)
            iw(last) = 0
            last = j
 61      continue
c-----------------------------------------------------------------------
 7    continue
c

      do 8 ii=1, nrow
         nnz = nnz+ndegr(ii)
 8    continue
      return
c----------------end-of-aplbdg -----------------------------------------
c-----------------------------------------------------------------------
      end

# HEADER ####################################################
# This is file spam/tests/demo_spam.R.                      #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# This is a simple demo, wrapping up the functionality of spam.
options( echo=FALSE)
library( spam, warn.conflict=FALSE)


set.seed(14)

nrow <- 5
ncol <- 7
fmat <- matrix(rnorm(nrow*ncol),nrow)
smat <- as.spam(fmat)


smat[1,]
smat[,1] <- 0
as.spam(smat)


ssmat <- smat %*% t(smat)

b <- c(-2:2)

solve(ssmat,b)

cholssmat <- chol(ssmat)


# works also for large matrices:

set.seed(14)

nz <- 100
nrow <- 100
ncol <- 100
smat <- diag.spam(1,nrow,ncol)

smat[cbind(sample(1:(nrow*ncol),size=nz))] <- runif(nz)

smat <- smat %*% t(smat)




b <- rnorm(nz)

smatinvb <- solve(smat,b)

cholssmat <- chol(smat)


# displaying matrices
if (F) {
opar <- par(no.readonly = TRUE)
par(ask=interactive() &&
            (.Device %in% c("X11","GTK","gnome","windows","quartz")))

display(smat, main="'scatterplot'-type display, very efficient")



options(spam.imagesize=prod(smat@dimension)+1)
display(smat, main="'image'-type display, may be slow and heavy")

par(opar)


}
options( echo=TRUE)

# HEADER ####################################################
# This is file spam/R/rep_len64.R.                          #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rep_len64 <- function(x, length.out, NAOK = getOption("spam.NAOK")){
    if(getOption("spam.force64") || length.out > 2147483647){
        .format64()
        return(.C64("rep_len64_c",
                    SIGNATURE = c("double", "int64", "int64", "double"),
                    
                    x = x,
                    lx = length(x),
                    length = length.out,
                    
                    out = numeric_dc(length.out),
                    
                    INTENT = c("r","r","r","w"),
                    NAOK = NAOK,
                    PACKAGE = "spam64"
                    )$out)
    }
    return(rep_len(x = x, length.out = length.out))
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-rep_len64.R.        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())
source("helper.R")

## library("testthat", lib.loc = LIB.LOC)
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = LIB.LOC)


context("test-rep_len64.R")
    

options(spam.printsize=6)
test_that("rep_len64-force64=FALSE", {
    options(spam.force64=FALSE)
    expect_equal(rep_len(1, 10), spam:::rep_len64(1, 10))
    expect_equal(rep_len(1L, 10L), spam:::rep_len64(1L, 10L))
    expect_equal(rep_len(1:2, 10L), spam:::rep_len64(1:2, 10L))
    expect_equal(rep_len(1:3, 16), spam:::rep_len64(1:3, 16))
})

## test_that("rep_len64-force64=TRUE", {
##     options(spam.force64=TRUE)
##     expect_equal(rep_len(1, 10), spam:::rep_len64(1, 10))
##     expect_equal(rep_len(1L, 10L), spam:::rep_len64(1L, 10L))
##     expect_equal(rep_len(1:2, 10L), spam:::rep_len64(1:2, 10L))
##     expect_equal(rep_len(1:3, 16), spam:::rep_len64(1:3, 16))
## })

# HEADER ####################################################
# This is file spam/R/rowcolstats.R.                        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



rowSums.spam <- function(x,...) {
    ## print("1")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(x)
    
    return(.C64("rowsums",
                SIGNATURE=c("double", SS$signature, SS$signature,
                    "double"),

                x@entries,
                x@rowpointers,
                x@dimension[1],

                rs = vector_dc("double", x@dimension[1]),

                INTENT=c("r", "r", "r",
                    "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)$rs )
}

colSums.spam <- function(x,...) {
    ## print("2")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(x)
        
    return(.C64("colsums",
                SIGNATURE=c("double", SS$signature, SS$signature, SS$signature,
                    "double"),

                x@entries,
                x@colindices,
                x@rowpointers,
                x@dimension[1],

                cs = vector_dc("double", x@dimension[2]),

                INTENT=c("r", "r", "r","r",
                    "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)$cs )
}

rowMeans.spam <- function(x,...) {
        ## print("3")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(x)

    return(.C64("rowmeans",
                SIGNATURE=c("double", SS$signature, SS$signature,
                    SS$signature, SS$signature, 
                    "double"),

                x@entries,
                x@rowpointers,
                x@dimension[1],
                x@dimension[2],
                getOption("spam.structurebased"),

                rm = vector_dc("double", x@dimension[1]),

                INTENT=c("r", "r", "r",
                    "r", "r",
                    "rw"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)$rm )
}

colMeans.spam <- function(x,...) {
            ## print("4")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(x)
    
    return(.C64("colmeans",
                SIGNATURE=c("double", SS$signature, SS$signature,
                    SS$signature, SS$signature, SS$signature, 
                    "double", SS$signature),

                x@entries,
                x@colindices,
                x@rowpointers,
                x@dimension[1],
                x@dimension[2],
                getOption("spam.structurebased"),

                cm = vector_dc("double", x@dimension[2]),
                vector_dc(SS$type, x@dimension[2]),

                INTENT=c("r", "r", "r",
                    "r", "r", "r",
                    "rw", "rw"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)$cm )
}



setMethod("rowSums","spam",rowSums.spam)
setMethod("colSums","spam",colSums.spam)
setMethod("rowMeans","spam",rowMeans.spam)
setMethod("colMeans","spam",colMeans.spam)

# HEADER ####################################################
# This is file spam/tests/testthat/test-rowcolstats.R.      #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-rowcolstats.R")


# simple tests:
########################################################################


# construct matrices:
n <- 10
m <- 15

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
tt[tt<0] <- 0

ss <- as.spam(tt)


test_that("rowcolstats", {
    spamtest_eq(rowSums.spam(ss),rowSums(tt)) 
    spamtest_eq(colSums.spam(ss),colSums(tt))
    
    spamtest_eq(rowSums(ss),rowSums(tt)) 
    spamtest_eq(colSums(ss),colSums(tt)) 
    
    
    options(spam.structurebased=FALSE)
    
    spamtest_eq(rowMeans.spam(ss),rowMeans(tt))      
    spamtest_eq(colMeans.spam(ss),colMeans(tt))      
    
    
    spamtest_eq(rowMeans(ss),rowMeans(tt))      
    spamtest_eq(colMeans(ss),colMeans(tt))      
    
    
    
    options(spam.structurebased=TRUE)
    spamtest_eq(rowMeans.spam(ss),rowSums(tt)/apply(tt>0,1,sum))      
    spamtest_eq(colMeans.spam(ss),colSums(tt)/apply(tt>0,2,sum))      

    spamtest_eq(rowMeans(ss),rowSums(tt)/apply(tt>0,1,sum))      
    spamtest_eq(colMeans(ss),colSums(tt)/apply(tt>0,2,sum))      

    
    spamtest_eq(rowMeans.spam(ss),apply.spam(ss,1,mean))      
    spamtest_eq(colMeans.spam(ss),apply.spam(ss,2,mean))          
    
    spamtest_eq(rowMeans(ss),apply.spam(ss,1,mean))      
    spamtest_eq(colMeans(ss),apply.spam(ss,2,mean))      
    
    
    spamtest_eq(rowMeans.spam(spam(0,n,m)),rowMeans(tt*0), relative = FALSE)      
    spamtest_eq(colMeans.spam(spam(0,n,m)),colMeans(tt*0), relative = FALSE)      
    spamtest_eq(rowMeans.spam(as.spam(diag(0,n))),rowMeans(diag(0,n)), relative = FALSE)      
    spamtest_eq(colMeans.spam(as.spam(diag(0,n))),colMeans(diag(0,n)), relative = FALSE)      
    
    spamtest_eq(rowMeans(spam(0,n,m)),rowMeans(tt*0), relative = FALSE)      
    spamtest_eq(colMeans(spam(0,n,m)),colMeans(tt*0), relative = FALSE)      
    spamtest_eq(rowMeans(as.spam(diag(0,n))),rowMeans(diag(0,n)), relative = FALSE)      
    spamtest_eq(colMeans(as.spam(diag(0,n))),colMeans(diag(0,n)), relative = FALSE)      
    
    options(spam.structurebased=TRUE)
})

# HEADER ####################################################
# This is file spam/R/permutation.R.                        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################








checkpivot <- function(pivot, len, type="Pivot") {
  if(is.null(pivot))                       return()
  if(!is.vector(pivot))                        stop(paste(type,"is not a vector."))     
  pivot <- as.vector(pivot,"integer")
  if (length(pivot) !=len)           stop(paste(type,"of wrong length."))
  tmp <- sort.int(pivot)
  if(tmp[1]!=1 ||  any(tmp-seq_len(len)!=0))   stop(paste("Invalid",type))
  return()
}



"permutation.spam" <- function(A, P=NULL, Q=NULL, ind=FALSE, check=TRUE){
  # eliminated .Internal calls as this creates a "Note" on CRAN checks.
                                        # Only 1-2% timing loss, see end of the file.
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(A)
  
    
    nrow <- A@dimension[1]
    ncol <- A@dimension[2]
    
    if (is.null(P)&is.null(Q))
        stop("At least one permutation should be specified")

    nz <- A@rowpointers[nrow+1]-1


    if (check){
        checkpivot(P,nrow,"Permutation")
        checkpivot(Q,ncol,"Permutation")
    }
    
    if (is.null(Q)) {
                                        #      subroutine rperm (nrow,a,ja,ia,ao,jao,iao,perm)
                                        #      B = P A
        P <- as.integer(P)
        if(ind) P <- order(P)
                                        #    if(ind) P <- .Internal(order(T,F,P))
        
        
        z <- .C64("rperm",
                  ## subroutine rperm (nrow,a,ja,ia,ao,jao,iao,perm)
                  SIGNATURE = c( SS$signature, "double", SS$signature, SS$signature,
                                "double", SS$signature, SS$signature,
                                SS$signature),
                  
                  nrow,
                  A@entries,
                  A@colindices,
                  A@rowpointers,
                  
                  entries = vector_dc( "double", nz), 
                  colindices = vector_dc( SS$type, nz),
                  rowpointers = vector_dc( SS$type, nrow + 1),
                  
                  P,

                  INTENT = c("r", "r", "r", "r",
                             "w", "w", "w",
                             "r"),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = SS$package)
    } else {  
        if (is.null(P)){
                                        #      subroutine cperm (nrow,a,ja,ia,ao,jao,iao,perm,iwork) 
                                        #      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(*), iwork(*)
                                        #      double precision a(*), ao(*) 
                                        #      B = A Q 
            Q <- as.integer(Q)
            if(ind) Q <- order(Q)
                                        #       if(ind) Q <- .Internal(order(T,F,Q))
            ## z <- .Fortran("cperm",
            ##               as.integer(nrow),
            ##               A@entries, as.integer(A@colindices),
            ##               as.integer(A@rowpointers),
            ##               entries = vector("double",nz),
            ##               colindices = vector("integer", nz),
            ##               rowpointers = vector("integer", nrow + 1),
            ##               Q,
            ##               NAOK = getOption("spam.NAOK"), PACKAGE = "spam")
            z <- .C64("cperm",
                      SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                                    "double", SS$signature, SS$signature,
                                    SS$signature),
                      
                      nrow,
                      A@entries,
                      A@colindices,
                      A@rowpointers,
                      
                      entries = vector_dc( "double", nz),
                      colindices = vector_dc( SS$type, nz),
                      rowpointers = vector_dc( SS$type, nrow + 1),
                      
                      Q,

                      INTENT = c("r", "r", "r", "r",
                                 "w", "w", "w",
                                 "r"),
                      NAOK = getOption("spam.NAOK"),
                      PACKAGE = SS$package)

        } else {  
                                        #      subroutine dperm (nrow,a,ja,ia,ao,jao,iao,pperm,qperm,iwork)
                                        #      B = P A Q 
            Q <- as.integer(Q)
                                        #        if(ind) Q <- .Internal(order(T,F,Q))
            if(ind) Q <- order(Q)
            P <- as.integer(P)
                                        #        if(ind) P <- .Internal(order(T,F,P))
            if(ind) P <- order(P)
            ## z <- .Fortran("dperm",
            ##               as.integer(nrow),
            ##               A@entries,as.integer(A@colindices),
            ##               as.integer(A@rowpointers),
            ##               entries = vector("double",nz),
            ##               colindices = vector("integer", nz),
            ##               rowpointers = vector("integer", nrow + 1),
            ##               P,Q,
            ##               NAOK = getOption("spam.NAOK"), PACKAGE = "spam")
            z <- .C64("dperm",
                      ## subroutine dperm (nrow,a,ja,ia,ao,jao,iao,pperm,qperm)
                      SIGNATURE = c( SS$signature, "double", SS$signature, SS$signature,
                                    "double", SS$signature, SS$signature,
                                    SS$signature, SS$signature),
                      
                      nrow,
                      A@entries,
                      A@colindices,
                      A@rowpointers,
                      
                      entries = vector_dc( "double", nz),
                      colindices = vector_dc( SS$type, nz),
                      rowpointers = vector_dc( SS$type, nrow + 1),
                      
                      P,Q,

                      INTENT = c("r", "r", "r", "r",
                                 "w", "w", "w",
                                 "r", "r"),
                      NAOK = getOption("spam.NAOK"),
                      PACKAGE = SS$package)
        }   
        
    }
    ## newx <- new("spam")
    ## slot(newx, "entries", check = FALSE) <- z$entries
    ## slot(newx, "colindices", check = FALSE) <- z$colindices
    ## slot(newx, "rowpointers", check = FALSE) <- z$rowpointers
    ## slot(newx, "dimension", check = FALSE) <- c(nrow,ncol)
    return(.newSpam(
        entries = z$entries,
        colindices = z$colindices,
        rowpointers = z$rowpointers,
        dimension = c(nrow,ncol)
    ))
}

permutation.matrix <- function(A, P=NULL, Q=NULL, ind=FALSE, check=TRUE){
  nrow <- dim(A)[1]
  ncol <- dim(A)[1]
  
  if (is.null(P)&is.null(Q))     stop("At least one permutation should be specified")
  
  if (check){
    checkpivot(P,nrow,"Permutation")
    checkpivot(Q,ncol,"Permutation")
  }

  if (ind) {
    if (is.null(Q))     return(A[P,])
    if (is.null(P))     return(A[,Q])
    return(A[P,Q])
  } else {
    if (is.null(Q))     return(A[order(P),])
    if (is.null(P))     return(A[,order(Q)])
    return(A[order(P),order(Q)])
  }
}

setGeneric("permutation",function(A, P=NULL, Q=NULL, ind=FALSE, check=TRUE)standardGeneric("permutation"))
setMethod("permutation","matrix",permutation.matrix)
setMethod("permutation","spam",permutation.spam)


### ss <- sample(1:100000)
### system.time( for( i in 1:1000) tt<-order(ss))
### system.time( for( i in 1:1000) tt<-.Internal(order(T,F,ss)))

# HEADER ####################################################
# This is file spam/R/constructors.R.                       #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



"rowpointers<-" <- function(x, value) {
    dimx <- x@dimension
    nnz1 <- x@rowpointers[dimx[1]+1]
    diffvalue <- diff(value)
    if ( any(!is.finite(value)))
        stop("row pointers should be postive integers.")
    if (!identical( length(x@rowpointers), length(value)))
        stop("wrong length of row pointers in `rowpointers<-`.", call.=FALSE)
    if (any(diffvalue<0))
        stop("row pointers are not monotone increasing in `rowpointers<-`.", call.=FALSE)
    if (any(diffvalue>dimx[2]))
        stop("row pointers have too large leaps in `rowpointers<-`.", call.=FALSE)
    if (value[1]<1)
        stop("first element of row pointers is < 1 in `rowpointers<-`.", call.=FALSE)
    if(value[dimx[1]+1] != nnz1)
        stop("last element of row pointers does not conform in `rowpointers<-`.", call.=FALSE)

    x@rowpointers <- as.integer(value)
    x
}

"colindices<-" <- function(x, value) {
    dimx <- x@dimension
    if ( any(!is.finite(value)))
        stop("column indices should be postive integers in `colindices<-`.", call.=FALSE)

    if ( any(value<1) | any(value> dimx[2]))
        stop("column indices exceed dimension `colindices<-`.", call.=FALSE)
    diffcolindices <- diff(value)     # positive values within each row
    if (all(diff(x@rowpointers)>1) && length(diffcolindices)>0)
                                        # only if we have multiple values
      if (identical( dimx[1], 1L)) {
        if   ( any(diffcolindices<1))
          stop("column indices are not ordered `colindices<-`.", call.=FALSE)
      } else {
        if ( any(diffcolindices[-(x@rowpointers[2:dimx[1]]-1)]<1))
          stop("column indices are not ordered `colindices<-`.", call.=FALSE)
      }
    x@colindices <- as.integer(value)
    x
}

"entries<-" <- function(x, value) {
    if (!identical( length(x@entries), length(value)))
        stop("wrong length in `entries<-`.", call.=FALSE)
    if (!getOption("spam.NAOK")) {
        if (any(!is.finite(value)))
            stop("\"NA/NaN/Inf\" not allowed in `entries<-`.", call.=FALSE)
    }
    if (!is.numeric( value))
        stop("numerical required in `entries<-`.", call.=FALSE)

    x@entries <- as.double(value)
    x
}

"dimension<-" <- function(x, value) {
    stop("modification through dim() or pad()", call.=FALSE)
}

# HEADER ####################################################
# This is file spam/R/plotting.R.                           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################






# Similar to misc3D
if(! exists(".bincode", envir = .BaseNamespaceEnv))
    .bincode <- function(v, breaks, ...) {
        .C("bincode", as.double(v), length(v), as.double(breaks),
           length(breaks), code = integer(length(v)), as.logical(TRUE),
           as.logical(TRUE), nok = TRUE, NAOK = TRUE,
           PACKAGE = "base")$code
    }


image.spam <-
function (x = seq(0, 1, len = nrow(z)), y = seq(0, 1, len = ncol(z)),
    z, zlim = range(z), xlim = range(x), ylim = range(y),
    col = heat.colors(12), add = FALSE, xaxs = "i", yaxs = "i",
    xlab, ylab, breaks, oldstyle = FALSE,cex=NULL, ...)
{

    if (missing(z)) {
      if (!missing(x)) {
        if (is.list(x)) {
          z <- x$z
          y <- x$y
          x <- x$x
        } else {
          if (is.null(dim(x)))
            stop("argument must be matrix-like")
          z <- x
          x <- seq(0, 1, len = dim(z)[1])
        }
        if (missing(xlab))
          xlab <- ""
        if (missing(ylab))
          ylab <- ""
      }
      else stop("no 'z' matrix specified")
    } else if (is.list(x)) {
      xn <- deparse(substitute(x))
      if (missing(xlab))
        xlab <- paste(xn, "x", sep = "$")
      if (missing(ylab))
        ylab <- paste(xn, "y", sep = "$")
      y <- x$y
      x <- x$x
    } else {
      if (missing(xlab))
        xlab <- if (missing(x))
          ""
        else deparse(substitute(x))
      if (missing(ylab))
        ylab <- if (missing(y))
          ""
        else deparse(substitute(y))
    }

    if (any(!is.finite(x)) || any(!is.finite(y)))
      stop("'x' and 'y' values must be finite and non-missing")
    if (any(diff(x) <= 0) || any(diff(y) <= 0))
      stop("increasing 'x' and 'y' values expected")

    spamversion <- (prod(z@dimension) > getOption("spam.imagesize"))
    if (!spamversion){
      image.default(x, y, as.matrix(z),...)
    } else {

      if (!is.spam(z))    stop("'z' must be a matrix")
      xx <- x
      yy <- y
      if (length(x) > 1 && length(x) == dim(z)[1]) {
        dx <- 0.5 * diff(x)
        x <- c(x[1] - dx[1], x[-length(x)] + dx, x[length(x)] +
               dx[length(x) - 1])
      }
      if (length(y) > 1 && length(y) == dim(z)[2]) {
        dy <- 0.5 * diff(y)
        y <- c(y[1] - dy[1], y[-length(y)] + dy, y[length(y)] +
               dy[length(y) - 1])
      }

      zvals <- z@entries
      if (missing(breaks)) {
        nc <- length(col)
        if (!missing(zlim) && (any(!is.finite(zlim)) || diff(zlim) < 0))
          stop("invalid z limits")
        if (diff(zlim) == 0)
          zlim <- if (zlim[1] == 0) {
            c(-1, 1)
          } else zlim[1] + c(-0.4, 0.4) * abs(zlim[1])
        zvals <- (zvals - zlim[1])/diff(zlim)
        zi <- if (oldstyle) {
          floor((nc - 1) * zvals + 0.5)
        } else floor((nc - 1e-05) * zvals + 1e-07)
        zi[zi < 0 | zi >= nc] <- NA
      }    else {
        if (length(breaks) != length(col) + 1)
          stop("must have one more break than colour")
        if (any(!is.finite(breaks)))
          stop("breaks must all be finite")
        # Patch proposed by BR see email.
        zi <- .bincode(zvals, breaks, TRUE, TRUE) - 1

      }
      if (!add)
        plot(NA, NA, xlim = xlim, ylim = ylim, type = "n", xaxs = xaxs,
             yaxs = yaxs, xlab = xlab, ylab = ylab, ...)
      if (length(xx) != dim(z)[1] || length(yy) != dim(z)[2])
        stop("dimensions of z are not length(x) times length(y)")
      if (missing(cex)) {
        warning("default value for 'cex' in 'image' might be a bad choice", call.=FALSE)
        cex <- 1
      }
      points( xx[rep.int((1:dim(z)[1]),diff(z@rowpointers))], yy[z@colindices],
             pch=".", cex=cex*getOption("spam.cex")/sum(z@dimension),
             col=col[zi+1])
    }

    box()
  }

display.spam <- function(x,col=c("gray","white"),xlab="column",ylab="row", cex=NULL,
                       main="",...)
{
  nrow <- x@dimension[1]
  ncol <- x@dimension[2]

# For small matrices, we transform them into regular ones and use the image.default
# routine.
  if (prod(nrow,ncol) < getOption("spam.imagesize")) {
    z <- vector("double", prod(nrow,ncol))
    dim(z) <- c(nrow,ncol)
    z[cbind(rep.int(nrow:1,diff(x@rowpointers)),x@colindices)] <- -1
    image.default(x=1:ncol,y=-(nrow:1),t(z),
                  axes=FALSE, col=col, xlab=xlab, ylab=ylab, ...)
  } else {
    if (missing(cex)) {
      warning("default value for 'cex' in 'display' might not be the optimal choice", call.=FALSE)
      cex <- 1
    }
    plot( x@colindices, rep.int(-(1:nrow),diff(x@rowpointers)), pch=".", cex=cex*getOption("spam.cex")/(ncol+nrow),
         col=col[1],xlab=xlab,ylab=ylab,axes=FALSE,
         ylim=c(-nrow,-0)-.5,xlim=c(0,ncol)+.5,xaxs = "i", yaxs = "i",...)
  }
  # Adjust axes labels.
  axis(1,pretty(1:ncol), ...)
  axis(2,pretty(-(nrow:1)),labels=rev(pretty(1:nrow)), ...)
  box()
}



plot.spam <- function(x,y,xlab=NULL,ylab=NULL,...)
{
  lab <- deparse(substitute(x))
  #only a few cases are considered
  # 1st case, a colum vector only
  if (dim(x)[2]==1) {
    x <- c(x)
    return( plot(x,...))
  }
  # 2nd case a matrix
  tmp <- x[,1:2, drop=FALSE] # extract the first two columns
  plot(c( tmp[,1]), c(tmp[,2]),
       xlab=ifelse(missing(xlab),paste(lab,"[,1]",sep=""),xlab),
       ylab=ifelse(missing(ylab),paste(lab,"[,2]",sep=""),ylab),...)
}

#setGeneric("image", function(x, ...) standardGeneric("image"))
setMethod("image","spam",function(x,cex=NULL,...){image.spam(x,cex=cex,...)})

# the following is unfortunately not possible
#setMethod("image",signature(x="numeric",y="numeric",z="spam"),function(x,y,z,cex=NULL,...){image.spam(x,cex=cex,...)})

setGeneric("display",function(x,...)standardGeneric("display"))
setMethod("display","spam",display.spam)

setMethod("plot", signature(x="spam",y="missing"), plot.spam)
setMethod("plot", signature(x="spam",y="spam"),
          function(x,y,...) {
            warning("'plot' with two 'spam' objects is not implemented",call.=FALSE)
            })

# HEADER ####################################################
# This is file spam/R/eigen.R.                              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



setMode <- function(sMode, symmetric, silent = FALSE){

  if (symmetric) {
    modes <- c("LM", "SM", "LA", "SA") #,"BE")
    if (!(sMode %in% modes)) {
      sMode <- 'LM'
      if (!silent) { warning(paste("the control option \"mode\" is set to", sMode), call. = FALSE) } }
  } else {
    modes <- c("LR", "SR", "LI", "SI", "SM", "LM")
    if (!(sMode %in% modes)) {
      sMode <- 'LR'
      if(!silent) { warning(paste("the control option \"mode\" is set to", sMode), call. = FALSE) } }
  }

  if (sMode == 'LM') { imode <- as.integer(1) }
  if (sMode == 'SM') { imode <- as.integer(2) }
  if (sMode == 'LR') { imode <- as.integer(3) }
  if (sMode == 'SR') { imode <- as.integer(4) }
  if (sMode == 'LI') { imode <- as.integer(5) }
  if (sMode == 'SI') { imode <- as.integer(6) }
  if (sMode == 'LA') { imode <- as.integer(7) }
  if (sMode == 'SA') { imode <- as.integer(8) }
  if (sMode == 'BE') { imode <- as.integer(9) }

  return(imode)
}

mk_cmplxentries <- function(z)
{
  if (dim(z)[2] != 2)
    stop("wrong format from fortran return: dn_eigen_f")

  cmplx <- NULL
  cmplx <- sapply(1:length(z[ , 1]), function(x) { complex(real = z[x, 1], imaginary = z[x, 2]) })

  if (is.null(cmplx))
    stop("error while aggregating fortran return from two vectors to a complex")

  return(cmplx)
}


getEigenval <- function(values, mode, dim, nEig, symmetric){

  if (symmetric) {
    result <- rep(NA, dim)

    orderedInd <- order(values[1:nEig], decreasing = TRUE)
    values <- values[orderedInd]

    # sort works also for complex values
    if (identical(mode %% 2, 0)) { # even modes are SA, SM, SR, SI
      result[(dim - nEig + 1):dim] <- values
    } else {                       # odd modes are LA, LA, LR, LI
      result[1:nEig] <- values }

  } else {
    result <- matrix(NA, nrow = dim, ncol = 2)

    orderedInd <- order(values[1:nEig, 1], decreasing = TRUE)
    values <- values[orderedInd, ]

    if (identical(mode %% 2, 0)) {
      result[(dim - nEig + 1):dim, ] <- values
    } else {
      result[1:nEig, ] <- values }

    result <- mk_cmplxentries(result)
  }

  return(list("eigenvalues" = result , "order" = orderedInd))
}


getEigenvec <- function(v, sym, dimen, nEig, orderind, eigenvalues, cmplxeps){

  if (is.null(v))
    stop("fortran returned NULL eigenvectors")

  if (sym) {

    v <- matrix(v, nrow = dimen, ncol = nEig, byrow = FALSE)
    v <- v[ , orderind, drop = FALSE]

  } else {

    v <- matrix(v[1:(dimen*nEig*2)], nrow = dimen, ncol = nEig*2, byrow = FALSE)

    v_real <- v[ , seq(1, nEig*2, by = 2)]
    v_imag <- v[ , seq(2, nEig*2, by = 2)]

    v_real <- v_real[ , orderind, drop = FALSE]
    v_imag <- v_imag[ , orderind, drop = FALSE]

    rm(v)

    eigenvalues <- stats::na.omit(eigenvalues)
    v <- matrix(NA, nrow = dimen, ncol = nEig)
    v <- sapply(1:nEig, function(x) {
      if (abs(Im(eigenvalues[x])) > cmplxeps) {
          mk_cmplxentries(cbind(v_real[ , x], v_imag[ , x]))
      } else {
        mk_cmplxentries(cbind(v_real[ , x], base::rep.int(0, times = dimen)))
      }
    })

  }

  return(v)
}


eigen_approx <- function(x,
                         nev,
                         ncv,
                         nitr,
                         mode,
                         only.values  = FALSE,
                         verbose      = FALSE,
                         f_routine){

  nz <- x@rowpointers[x@dimension[1]+1]-1  
    if((nz==0) | (norm(x@entries) <= getOption("spam.eps"))){#trap zero matrix
        stop("'x' is an emtpy matrix")
    }
    
  # check & parse arguments
  if (x@dimension[1] <= nev)
    stop("nev: the number of eigenvalues to calculate must be smaller than the matrix dimensions", call. = TRUE)

  if (f_routine != "ds_eigen_f" && f_routine != "dn_eigen_f")
    stop("non valid fortran routine is specified", call. = TRUE)



  f_mode <- setMode(sMode = mode, symmetric = ifelse(identical(f_routine, "ds_eigen_f"), TRUE, FALSE))

  fortran_object <- result <- list(NULL)

  if(getOption("spam.force64") || .format.spam(x)$package != "spam") {
    SS <- .format64()
    f_routine <- paste0(f_routine, "64")
  } else {
    SS <- .format32 }

  # Fortran call: symmetric matrices
  if (identical(f_routine, "ds_eigen_f") || identical(f_routine, "ds_eigen_f64")) {

    # define upperbound for matrices in spam64, since ARPACK routines rely on BLAS/LAPACK functions
    # which have no integer/logical's of kind = 8
    if (max(x@dimension[1], ncv*(ncv + 8))*3 >= 2^31-1) {
      stop("the dimension and of the input matrix and/or the required number of eigenvalues
            is too large for the current eigenvalue decomposition implementation.\n
            Reducing the argument \"nev\" and\"ncv\" in the contorl options may helps.")
    }

    fortran_object <- .C64 (f_routine,
                            SIGNATURE = c("integer", "integer", "integer", "integer",
                                          "integer", "integer", "double",
                                          SS$signature, SS$signature, "double",
                                          "double", "integer"),
                            maxnev    = nev,
                            ncv       = ncv,
                            maxitr    = nitr,
                            n         = x@dimension[1],
                            iwhich    = f_mode,
                            na        = x@dimension[1],
                            a         = x@entries,
                            ja        = x@colindices,
                            ia        = x@rowpointers,
                            v         = vector_dc("double", x@dimension[1]*ncv),
                            d         = vector_dc("double", nev),
                            iparam    = integer_dc(8),
                            INTENT    = c("r", "r", "r", "r", "r", "r", "r", "r", "r",
                                           "rw", "rw", "rw"),
                            NAOK      = getOption("spam.NAOK"),
                            PACKAGE   = SS$package)

    if (is.null(fortran_object)) {
      stop("error while calling fortran routine, check (control) arguments", call. = TRUE) }

    result <-   list ("nEigenVal"     = nev,
                      "Mode"          = f_mode,
                      "Eigenvectors"  = if (!only.values) { fortran_object$v[1:(x@dimension[1]*nev)] } else { NULL },
                      "Eigenvalues"   = fortran_object$d,
                      "nconv"         = fortran_object$iparam[5])
  }

  # Fortran call: nonsymmetric matrices
  if (identical(f_routine, "dn_eigen_f") || identical(f_routine, "dn_eigen_f64")) {

    if (max(x@dimension[1], ncv^2+6*ncv)*3 >= 2^31-1) {
      stop("the dimension and of the input matrix and/or the required number of eigenvalues
            is too large for the current eigenvalue decomposition implementation.\n
            Reducing the argument \"nev\" and\"ncv\" in the contorl options may helps.")
    }


    fortran_object <- .C64 (f_routine,
                            SIGNATURE = c("integer", "integer", "integer", "integer",
                                          "integer", "integer", "double",
                                          SS$signature, SS$signature, "double",
                                          "double", "double", "integer"),
                            maxnev    = nev,
                            ncv       = ncv,
                            maxitr    = nitr,
                            n         = x@dimension[1],
                            iwhich    = f_mode,
                            na        = x@dimension[1],
                            a         = x@entries,
                            ja        = x@colindices,
                            ia        = x@rowpointers,
                            v         = vector_dc("double", x@dimension[1]*ncv),
                            dr        = vector_dc("double", nev+1),
                            di        = vector_dc("double", nev+1),
                            iparam    = integer_dc(8),
                            INTENT    = c("r", "r", "r", "r", "r", "r", "r", "r", "r",
                                          "rw", "rw", "rw", "rw"),
                            NAOK      = getOption("spam.NAOK"),
                            PACKAGE   = SS$package)

    if (is.null(fortran_object)) {
      stop("error while calling fortran routine, check (control) arguments", call. = TRUE) }

    result <-   list ("nEigenVal"     = nev,
                      "Mode"          = f_mode,
                      "Eigenvectors"  = if (!only.values) { fortran_object$v[1:(x@dimension[1]*nev*2)] } else { NULL },
                      "Eigenvalues"   = cbind(fortran_object$dr, fortran_object$di),
                      "nconv"         = fortran_object$iparam[5])
  }

  if (verbose) {
    cat("\n used options/arguments:")
    if (identical(f_routine, "dn_eigen_f") || identical(f_routine, "dn_eigen_f64")) {
      issym <- "non symmetric"
    } else {
      issym <- "symmetric" }
    cat(paste("\n      FORTRAN routine:", issym, "matrices"))
    cat(paste("\n      nitr:", nitr))
    cat(paste("\n      ncv:", ncv, "\n"))

    cat("\n summary FORTRAN return:")
    cat(paste("\n     ", nev, "eigenvalues requested and", result$nconv, "converged\n"))
  }

  if (result$nconv < nev)
    warning(paste("\n only", result$nconv, "instead of", nev ,"eigenvalues converged, try to increase 'control = list(nitr = .., ncv = ..)'"), call. = TRUE)

  if (is.null(result)) {
    stop("unpredicted error, check control options of the eigen.spam function.", call. = TRUE)
  }

  return(result)
}


eigen.spam <- function (x, nev = 10, symmetric, only.values = FALSE, control = list()){

  con <- list(mode     = 'LM',
              nitr     = NULL,
              ncv      = NULL,
              spamflag = FALSE,
              verbose  = FALSE,
              cmplxeps = .Machine$double.eps)

  nmsC <- names(con)
  con[(namc <- names(control))] <- control
  if (length(noNms <- namc[!namc %in% nmsC]))
    warning("unknown names in control: ", paste(noNms, collapse = ", "), call. = TRUE)

  ifelse(!con$verbose, vFlag <- FALSE, vFlag <- TRUE)

  nz <- x@rowpointers[x@dimension[1]+1]-1  
  if((nz==0) | (norm(x@entries) <= getOption("spam.eps"))){#trap zero matrix
      stop("'x' is an emtpy matrix")
  }

  # arpack routines cant handle 'small' matrices
  minDimARPACK <- 100

  resContainer <- list(NULL)

  if (!is.spam(x)) {
    try(x <- as.spam(x)) }
  if (missing(symmetric)){
    symmetric <- isSymmetric.spam(x)
  } else {
    if (!identical(symmetric, isSymmetric.spam(x))) {
      warning(paste("the input matrix is", ifelse(isSymmetric.spam(x), "symmetric", "not symmetric")), call. = TRUE)
    }
  }

  ## dispatching between base::eigen and ARPACK
  if ((!con$spamflag && prod(dim(x)) <= getOption("spam.inefficiencywarning")) || (con$spamflag && prod(dim(x)) <= minDimARPACK)) {
    warning("\n The eigenvalues are calculated with the function 'eigen' from the base package.", call. = TRUE)

    if (con$spamflag) { warning(paste("\n Even 'spamflag = TRUE', since the matrix dimension is smaller than", minDimARPACK), call. = TRUE) }

    resContainer <- base::eigen(x           = x,
                                symmetric   = symmetric,
                                only.values = only.values)

    resEig       <- resContainer$values
    resVec       <- resContainer$vectors

  } else {
  # --- using ARPACK -------------------------------------------------------------- #

    if (!identical(x@dimension[1], x@dimension[2]))
      stop("non-square matrix in 'eigen'")

    if (nev >= x@dimension[1] || nev <= 0)
      stop ("the number of asked eigenvalues is higher or equal the dimension of the input matrix")

    if (symmetric) { ncvMaxMin <- 100 } else {  ncvMaxMin <- 100 }

    if (con$ncv > x@dimension[1] || con$ncv < nev || is.null(con$ncv)) {
      con$ncv <- min(x@dimension[1] + 1, max(2 * nev + 1, ncvMaxMin)) }

    if (is.null(con$nitr)) { con$nitr <- con$ncv + 1000 }

    # calculate eigenvalues and vectors
    resContainer <- eigen_approx(x            = x,
                                 nev          = nev,
                                 ncv          = con$ncv,
                                 nitr         = con$nitr,
                                 mode         = con$mode,
                                 only.values  = only.values,
                                 verbose      = vFlag,
                                 f_routine    = ifelse(symmetric, "ds_eigen_f", "dn_eigen_f"))

    if (is.null(resContainer)) {
      stop("\n FORTRAN return is empty") }

    tmpresEig    <-  getEigenval(values       = resContainer$Eigenvalues,
                                 mode         = setMode(con$mode, symmetric, silent = TRUE),
                                 dim          = x@dimension[1],
                                 nEig         = if (resContainer$nconv > nev) { nev } else { resContainer$nconv },
                                 symmetric    = symmetric)
    resEig <- tmpresEig$eigenvalues

    resVec <- NULL
    if(!only.values) {
      resVec     <-  getEigenvec(v            = resContainer$Eigenvectors,
                                 sym          = symmetric,
                                 dimen        = x@dimension[1],
                                 nEig         = if (resContainer$nconv > nev) { nev } else { resContainer$nconv },
                                 orderind     = tmpresEig$order,
                                 eigenvalues  = resEig,
                                 cmplxeps     = con$cmplxeps) }

  # ------------------------------------------------------------------------------- #
  }

  return (list("values" = resEig, "vectors" = resVec))
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-eigen.R.            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

context("test-eigen.R")
options(spam.force64 = FALSE)

nEV <- 50
prec2020 <- precmat.GMRFreglat(20,20)
dn_gmrf    <- spam::precmat.GMRFreglat(20, 20, c(.4,.3,.2,.1),'m2p4')
dn_gmrf[1:10, 1] <- 100

test_that("check format of return", {
  spameigen <- eigen.spam(prec2020, nev = nEV, control = list(spamflag = TRUE, mode = 'LM'), only.values = FALSE)
  nsys_spameigen <- eigen.spam(dn_gmrf, nev = nEV, control = list(spamflag = TRUE, mode = 'LM'), only.values = FALSE)

  expect_true(is(spameigen,              "list"))
  expect_true(is(spameigen$values,       "numeric"))
  expect_true(is(spameigen$vectors,      "matrix"))
  expect_true(is(nsys_spameigen,         "list"))
  expect_true(is(nsys_spameigen$values,  "complex"))
  expect_true(is(nsys_spameigen$vectors, "matrix"))
})

test_that("eigenvalues", {
  baseeigen <- eigen(prec2020, only.values = TRUE)
  spameigenL <- eigen.spam(prec2020, nev = nEV, control = list(spamflag = T, mode = 'LM'), only.values = TRUE)
  spameigenS <- eigen.spam(prec2020, nev = nEV, control = list(spamflag = T, mode = 'SM'), only.values = TRUE)
  spamtest_eq(head(baseeigen$values, nEV), head(spameigenL$values, nEV), relative = FALSE)
  spamtest_eq(tail(baseeigen$values, nEV), tail(spameigenS$values, nEV), relative = FALSE)

  nsys_baseeigen <- eigen(dn_gmrf, only.values = TRUE)
  nsys_spameigenL <- eigen.spam(dn_gmrf, nev = nEV, control = list(spamflag = T, mode = 'LR', ncv = 400), only.values = TRUE, symmetric = FALSE)
  nsys_spameigenS <- eigen.spam(dn_gmrf, nev = nEV, control = list(spamflag = T, mode = 'SR'), only.values = TRUE)
  # spamtest_eq(head(nsys_baseeigen$values, nEV), head(nsys_spameigenL$values, nEV), relative = FALSE)
  # spamtest_eq(tail(nsys_baseeigen$values, nEV), tail(nsys_spameigenS$values, nEV), relative = TRUE)
})

test_that("eigenvectors", {
  baseeigen <- eigen(prec2020, only.values = FALSE)
  spameigenL <- eigen.spam(prec2020, nev = nEV, control = list(spamflag = T, mode = 'LM'), only.values = FALSE)
  spameigenS <- eigen.spam(prec2020, nev = nEV, control = list(spamflag = T, mode = 'SM'), only.values = FALSE)

  for (i in 1:nEV) {spamtest_eq(sqrt(sum(spameigenL$vectors[, i]^2)), 1) }
  for (i in 1:nEV) {spamtest_eq(sqrt(sum(spameigenS$vectors[, i]^2)), 1) }
  for (i in 1:nEV) {spamtest_eq(as.matrix(prec2020)%*%spameigenL$vectors[, i], spameigenL$vectors[, i]*spameigenL$values[i]) }
})

### rgrf and suchlike
rgrf <- function(  # we can also include dgrf...
 n,          # number of samples
 locs,       # either pass locs or set nx for regular/perturbed grid
 nx, ny=nx,  # at least the first needs to be set.
 xlim=c(0,1), ylim=c(0,1), # size of domain for regular/perturbed grid
 tau=0,      # perturbation if between (0,1/2). 
             # Separation of locations is at least 1-2tau 
 
 Covariance, # similar as with mle
 theta,      # (or theta0)

 beta=0,     # mean if length one, regression coefficients if longer
 X,          # 'design' matrix for mean Xbeta

 method=c('chol'),  # based on Choleski factorization
 method.args=list(sparse=FALSE), # list of arguments that can be passed to the corresponding approach
                     # for chol it can be RStruct, chol.args, cov.args
 eps = getOption("spam.eps"),
 drop = TRUE,
 attributes=TRUE,  # should these been passed back?
 ...)  {

    call <- match.call()
    # construct locations:
    if (missing(locs)) {
        dx <- diff(xlim)/nx
        dy <- diff(ylim)/ny
        locs <- expand.grid(x=seq(from=xlim[1]+dx/2, to=xlim[2]-dx/2, length=nx),
                            y=seq(from=ylim[1]+dx/2, to=ylim[2]-dx/2, length=ny))
        
        Nlocs <- nx * ny
        stopifnot(tau >= 0,  tau < 1/2)
         etau <- tau*c(dx,dy)
       
        if ( tau>eps)  locs <- locs + cbind(runif(Nlocs, -etau,etau), runif(Nlocs, -etau,etau))
         
    } else {
        
        locs <- as.matrix(locs)
        Nlocs <- dim(locs)[1]
        etau <- c(0,0)
    }
    
    # construct mean:
    if (length(beta)>1) {
        X <- as.matrix(X)
        stopifnot( dim(X)[1] == Nlocs,  dim(X)[2] == length(beta))
        mu <- c( X %*% beta ) 
    } else if (length(beta)==1) {
       mu <- if(missing(X)) beta else X * beta
    }

    
    if (Covariance == 'cov.nug') {  # slight exception here...
        out <- sweep( as.matrix( ( array(rnorm(n*Nlocs),c(n,Nlocs)) * theta[1] )), 2, mu, "+")
    } else {
        compact <- if (method.args$sparse) 
            (Covariance %in% c("cov.sph","cov.wend1","cov.wend2","cov.wu1","cov.wu3")) else FALSE
    
    
        if (method=='chol') {
            if (compact) {

                # construct distance and covariance matrix:
                distmat <- nearest.dist(locs, upper=NULL, delta=theta[1])
                Sigma <- do.call(Covariance, c(list(distmat, theta), method.args$cov.args))
               
                
                # now similar as in rmvnorm.spam
                if (is(method.args$Rstruct,"spam.chol.NgPeyton"))
                    cholS <- do.call("update.spam.chol.NgPeyton", list( method.args$Rstruct, Sigma, method.args$chol.args))
                else
                    cholS <- do.call("chol.spam", list( Sigma, method.args$chol.args))
                                        # cholS is the upper triangular part of the permutated matrix Sigma
                iord <- ordering(cholS, inv=TRUE)
                
                R <- as.spam(cholS)
                retval <- as.matrix( ( array(rnorm(n*Nlocs),c(n,Nlocs)) %*% R)[,iord,drop=F])
                                        # It is often better to order the sample than the matrix
                                        # R itself.
                out <- sweep(retval, 2, mu, "+")
            } else {


                if (requireNamespace("pkg", quietly = TRUE)) {
                    distmat <- fields::rdist(locs)
                } else {
                    distmat <- as.matrix(dist(locs))
                }

                Sigma <- do.call(Covariance, c(list(distmat, theta), method.args$cov.args))

                R <-   if (is.null(method.args$chol.args)) chol(Sigma) else
                              do.call("chol", list( Sigma, method.args$chol.args)) 
                N <- dim( Sigma)[1]
                
                out <- sweep( as.matrix( ( array(rnorm(n*Nlocs),c(n,Nlocs)) %*% R)), 2, mu, "+")
            }
            
        } else {
            stop("'method' not implemented yet.")
        }
    }
    if(attributes) {
        attr(out, "locs") <- locs
        attr(out, "mean") <- mu
        attr(out, "call") <- call
        attr(out, "effective.tau") <- etau
    }
    if(drop)   return(drop(out))
    return(out)
    
}

c-----------------------------------------------------------------------
      subroutine amask (nrow,ncol,a,ja,ia,jmask,imask,
     *                  c,jc,ic,nzmax,ierr)
c---------------------------------------------------------------------
      implicit none
      real(8) a(*),c(*)
      integer nrow, ncol
      integer ia(nrow+1),ja(*),jc(*),ic(nrow+1),jmask(*)
      integer imask(nrow+1), nzmax, ierr
      logical iw(ncol)
c-----------------------------------------------------------------------
c futher used variables
      integer k, len, ii, j, k1, k2
c-----------------------------------------------------------------------
c This subroutine builds a sparse matrix from an input matrix by
c extracting only elements in positions defined by the mask jmask, imask
c-----------------------------------------------------------------------
c On entry:
c---------
c nrow  = integer. row dimension of input matrix
c ncol  = integer. Column dimension of input matrix.
c
c a,
c ja,
c ia    = matrix in Compressed Sparse Row format
c
c jmask,
c imask = matrix defining mask (pattern only) stored in compressed
c         sparse row format.
c
c nzmax = length of arrays c and jc. see ierr.
c
c On return:
c-----------
c
c a, ja, ia and jmask, imask are unchanged.
c
c c
c jc,
c ic    = the output matrix in Compressed Sparse Row format.
c
c ierr  = integer. serving as error message.c
c         ierr = 1  means normal return
c         ierr .gt. 1 means that amask stopped when processing
c         row number ierr, because there was not enough space in
c         c, jc according to the value of nzmax.
c
c work arrays:
c-------------
c iw    = logical work array of length ncol.
c
c note:
c------ the  algorithm is in place: c, jc, ic can be the same as
c a, ja, ia in which cas the code will overwrite the matrix c
c on a, ja, ia
c
c-----------------------------------------------------------------------
      ierr = 0
      len = 0
      do 1 j=1, ncol
         iw(j) = .false.
 1    continue
c     unpack the mask for row ii in iw
      do 100 ii=1, nrow
c     save pointer in order to be able to do things in place
         do 2 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .true.
 2       continue
c     add umasked elemnts of row ii
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         ic(ii) = len+1
         do 200 k=k1,k2
            j = ja(k)
            if (iw(j)) then
               len = len+1
               if (len .gt. nzmax) then
                  ierr = ii
                  return
               endif
               jc(len) = j
               c(len) = a(k)
            endif
 200     continue
c
         do 3 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .false.
 3       continue
 100  continue
      ic(nrow+1)=len+1
c
      return
c-----end-of-amask -----------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine aplsb1 (nrow,ncol,a,ja,ia,s,b,jb,ib,c,jc,ic,
     *     nzmax,ierr)
      implicit none
      real(8) a(*), b(*), c(*), s
      integer nrow, ncol
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(nrow+1),ic(nrow+1)
      integer nzmax, ierr
c-----------------------------------------------------------------------
c further used vaeriables
      integer i, j1, j2, ka, kamax, kb, kbmax, kc

c-----------------------------------------------------------------------
c performs the operation C = A+s B for matrices in sorted CSR format.
c the difference with aplsb is that the resulting matrix is such that
c the elements of each row are sorted with increasing column indices in
c each row, provided the original matrices are sorted in the same way.
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow  = integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format with entries sorted
c
c s     = real. scalar factor for B.
c
c b,
c jb,
c ib    =  Matrix B in compressed sparse row format with entries sorted
c        ascendly in each row
c
c nzmax = integer. The  length of the arrays c and jc.
c         amub will stop if the result matrix C  has a number
c         of elements that exceeds exceeds nzmax. See ierr.
c
c on return:
c----------
c c,
c jc,
c ic    = resulting matrix C in compressed sparse row sparse format
c         with entries sorted ascendly in each row.
c
c ierr  = integer. serving as error message.
c         ierr = 0 means normal return,
c         ierr .gt. 0 means that amub stopped while computing the
c         i-th row  of C with i=ierr, because the number
c         of elements in C exceeds nzmax.
c
c Notes:
c-------
c     this will not work if any of the two input matrices is not sorted
c-----------------------------------------------------------------------
      ierr = 0
      kc = 1
      ic(1) = kc
c
c     the following loop does a merge of two sparse rows + adds  them.
c
      do 6 i=1, nrow
         ka = ia(i)
         kb = ib(i)
         kamax = ia(i+1)-1
         kbmax = ib(i+1)-1
 5       continue
c
c     this is a while  -- do loop --
c
         if (ka .le. kamax .or. kb .le. kbmax) then
c
            if (ka .le. kamax) then
               j1 = ja(ka)
            else
c     take j1 large enough  that always j2 .lt. j1
               j1 = ncol+1
            endif
            if (kb .le. kbmax) then
               j2 = jb(kb)
            else
c     similarly take j2 large enough  that always j1 .lt. j2
               j2 = ncol+1
            endif
c
c     three cases
c
            if (j1 .eq. j2) then
               c(kc) = a(ka)+s*b(kb)
               jc(kc) = j1
               ka = ka+1
               kb = kb+1
               kc = kc+1
            else if (j1 .lt. j2) then
               jc(kc) = j1
               c(kc) = a(ka)
               ka = ka+1
               kc = kc+1
            else if (j1 .gt. j2) then
               jc(kc) = j2
               c(kc) = s*b(kb)
               kb = kb+1
               kc = kc+1
            endif
            if (kc .gt. nzmax) goto 999
            goto 5
c
c     end while loop
c
         endif
         ic(i+1) = kc
 6    continue
      return
 999  ierr = i
      return
c------------end-of-aplsb1 ---------------------------------------------
c-----------------------------------------------------------------------
      end
c

      subroutine submat (job,i1,i2,j1,j2,a,ja,ia,nr,nc,ao,jao,iao)
      implicit none
      integer job,i1,i2,j1,j2,nr,nc,ia(*),ja(*),jao(*),iao(*)
      real(8) a(*),ao(*)
c-----------------------------------------------------------------------
c further used variables
      integer i, ii, j, k, k1, k2, klen
c-----------------------------------------------------------------------
c extracts the submatrix A(i1:i2,j1:j2) and puts the result in
c matrix ao,iao,jao
c---- In place: ao,jao,iao may be the same as a,ja,ia.
c--------------
c on input
c---------
c n     = row dimension of the matrix
c i1,i2 = two integers with i2 .ge. i1 indicating the range of rows to be
c          extracted.
c j1,j2 = two integers with j2 .ge. j1 indicating the range of columns
c         to be extracted.
c         * There is no checking whether the input values for i1, i2, j1,
c           j2 are between 1 and n.
c a,
c ja,
c ia    = matrix in compressed sparse row format.
c
c job   = job indicator: if job .ne. 1 then the real values in a are NOT
c         extracted, only the column indices (i.e. data structure) are.
c         otherwise values as well as column indices are extracted...
c
c on output
c--------------
c nr    = number of rows of submatrix
c nc    = number of columns of submatrix
c         * if either of nr or nc is nonpositive the code will quit.
c
c ao,
c jao,iao = extracted matrix in general sparse format with jao containing
c       the column indices,and iao being the pointer to the beginning
c       of the row,in arrays a,ja.
c----------------------------------------------------------------------c
c           Y. Saad, Sep. 21 1989                                      c
c----------------------------------------------------------------------c
      nr = i2-i1+1
      nc = j2-j1+1
c
      if ( nr .le. 0 .or. nc .le. 0) return
c
      klen = 0
c
c     simple procedure. proceeds row-wise...
c
      do 100 i = 1,nr
         ii = i1+i-1
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         iao(i) = klen+1
c-----------------------------------------------------------------------
         do 60 k=k1,k2
            j = ja(k)
            if (j .ge. j1 .and. j .le. j2) then
               klen = klen+1
               if (job .eq. 1) ao(klen) = a(k)
               jao(klen) = j - j1+1
            endif
 60      continue
 100  continue
      iao(nr+1) = klen+1
      return
c------------end-of submat----------------------------------------------
c-----------------------------------------------------------------------
      end
c
      subroutine amux (n, x, y, a,ja,ia)
      implicit none
      real(8)  x(*), y(*), a(*)
      integer n, ja(*), ia(*)
c-----------------------------------------------------------------------
c         A times a vector
c-----------------------------------------------------------------------
c multiplies a matrix by a vector using the dot product form
c Matrix A is stored in compressed sparse row storage.
c
c on entry:
c----------
c n     = row dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length n, containing the product y=Ax
c
c-----------------------------------------------------------------------
c local variables
c
      real(8) t
      integer i, k
c-----------------------------------------------------------------------
      do 100 i = 1,n
c
c     compute the inner product of row i with vector x
c
         t = 0.0d0
         do 99 k=ia(i), ia(i+1)-1
            t = t + a(k)*x(ja(k))
 99      continue
c
c     store result in y(i)
c
         y(i) = t
 100  continue
c
      return
c---------end-of-amux---------------------------------------------------
c-----------------------------------------------------------------------
      end
c
      subroutine amubdg (nrow,ncol,ncolb,ja,ia,jb,ib,ndegr,nnz,iw)
      implicit none
      integer nrow, ncol, ncolb
      integer ja(*),jb(*),ia(nrow+1),ib(ncol+1)
      integer ndegr(nrow),iw(ncolb)
      integer nnz
c-----------------------------------------------------------------------
c further used variables
      integer ii, j, jc, jr, k, last, ldg
c-----------------------------------------------------------------------
c gets the number of nonzero elements in each row of A*B and the total
c number of nonzero elements in A*B.
c-----------------------------------------------------------------------
c on entry:
c --------
c
c nrow  = integer.  row dimension of matrix A
c ncol  = integer.  column dimension of matrix A = row dimension of
c                   matrix B.
c ncolb = integer. the colum dimension of the matrix B.
c
c ja, ia= row structure of input matrix A: ja = column indices of
c         the nonzero elements of A stored by rows.
c         ia = pointer to beginning of each row  in ja.
c
c jb, ib= row structure of input matrix B: jb = column indices of
c         the nonzero elements of A stored by rows.
c         ib = pointer to beginning of each row  in jb.
c
c on return:
c ---------
c ndegr = integer array of length nrow containing the degrees (i.e.,
c         the number of nonzeros in  each row of the matrix A * B
c
c nnz   = total number of nonzero elements found in A * B
c
c work arrays:
c-------------
c iw    = integer work array of length ncolb.
c-----------------------------------------------------------------------
      do 1 k=1, ncolb
         iw(k) = 0
 1    continue

      do 2 k=1, nrow
         ndegr(k) = 0
 2    continue
c
c     method used: Transp(A) * A = sum [over i=1, nrow]  a(i)^T a(i)
c     where a(i) = i-th row of  A. We must be careful not to add  the
c     elements already accounted for.
c
c
      do 7 ii=1,nrow
c
c     for each row of A
c
         ldg = 0
c
c    end-of-linked list
c
         last = -1
         do 6 j = ia(ii),ia(ii+1)-1
c
c     row number to be added:
c
            jr = ja(j)
            do 5 k=ib(jr),ib(jr+1)-1
               jc = jb(k)
               if (iw(jc) .eq. 0) then
c
c     add one element to the linked list
c
                  ldg = ldg + 1
                  iw(jc) = last
                  last = jc
               endif
 5          continue
 6       continue
         ndegr(ii) = ldg
c
c     reset iw to zero
c
         do 61 k=1,ldg
            j = iw(last)
            iw(last) = 0
            last = j
 61      continue
c-----------------------------------------------------------------------
 7    continue
c
      nnz = 0
      do 8 ii=1, nrow
         nnz = nnz+ndegr(ii)
 8    continue
c
      return
c---------------end-of-amubdg ------------------------------------------
c-----------------------------------------------------------------------
      end
c


       subroutine amub (nrow,ncol,job,a,ja,ia,b,jb,ib,
     *                  c,jc,ic,nzmax,iw,ierr)
      implicit none
      real(8) a(*), b(*), c(*)
      integer nrow, ncol, job
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(*),ic(*),iw(ncol)
      integer nzmax, ierr
c-----------------------------------------------------------------------
c other used variables
      integer len, k, ka, kb, jpos, jcol, ii, jj, j
c-----------------------------------------------------------------------
c performs the matrix by matrix product C = A B
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow  = integer. The row dimension of A = row dimension of C
c ncol  = integer. The column dimension of B = column dimension of C
c job   = integer. Job indicator. When job = 0, only the structure
c                  (i.e. the arrays jc, ic) is computed and the
c                  real values are ignored.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format.
c
c b,
c jb,
c ib    =  Matrix B in compressed sparse row format.
c
c nzmax = integer. The  length of the arrays c and jc.
c         amub will stop if the result matrix C  has a number
c         of elements that exceeds exceeds nzmax. See ierr.
c
c on return:
c----------
c c,
c jc,
c ic    = resulting matrix C in compressed sparse row sparse format.
c
c ierr  = integer. serving as error message.
c         ierr = 0 means normal return,
c         ierr .gt. 0 means that amub stopped while computing the
c         i-th row  of C with i=ierr, because the number
c         of elements in C exceeds nzmax.
c
c work arrays:
c------------
c iw    = integer work array of length equal to the number of
c         columns in A.
c Note:
c-------
c   The row dimension of B is not needed. However there is no checking
c   on the condition that ncol(A) = nrow(B).
c
c-----------------------------------------------------------------------
      real(8) scal
      logical values
      values = (job .ne. 0)
c  the following is not necessary... keep [-Wmaybe-uninitialized] quite
      scal = 0.0
      len = 0
      ic(1) = 1
      ierr = 0
c     initialize array iw.
      do 1 j=1, ncol
         iw(j) = 0
 1    continue
c
      do 500 ii=1, nrow
c     row i
         do 200 ka=ia(ii), ia(ii+1)-1
            if (values) scal = a(ka)
            jj   = ja(ka)
            do 100 kb=ib(jj),ib(jj+1)-1
               jcol = jb(kb)
               jpos = iw(jcol)
               if (jpos .eq. 0) then
                  len = len+1
                  if (len .gt. nzmax) then
                     ierr = ii
                     return
                  endif
                  jc(len) = jcol
                  iw(jcol)= len
                  if (values) c(len)  = scal*b(kb)
               else
                  if (values) c(jpos) = c(jpos) + scal*b(kb)
               endif
 100        continue
 200     continue
         do 201 k=ic(ii), len
            iw(jc(k)) = 0
 201     continue
         ic(ii+1) = len+1
 500  continue
      return
c-------------end-of-amub-----------------------------------------------
c-----------------------------------------------------------------------
      end
c
c------------------------------------------------------------------------
      subroutine getl (n,a,ja,ia,ao,jao,iao)
      implicit none
      integer n, ia(*), ja(*), iao(*), jao(*)
      real(8) a(*), ao(*)
c------------------------------------------------------------------------
c this subroutine extracts the lower triangular part of a matrix
c and writes the result ao, jao, iao. The routine is in place in
c that ao, jao, iao can be the same as a, ja, ia if desired.
c-----------
c on input:
c
c n     = dimension of the matrix a.
c a, ja,
c    ia = matrix stored in compressed sparse row format.
c On return:
c ao, jao,
c    iao = lower triangular matrix (lower part of a)
c       stored in a, ja, ia, format
c note: the diagonal element is the last element in each row.
c i.e. in  a(ia(i+1)-1 )
c ao, jao, iao may be the same as a, ja, ia on entry -- in which case
c getl will overwrite the result on a, ja, ia.
c
c------------------------------------------------------------------------
c local variables
      real(8) t
      integer ko, kold, kdiag, k, i
c
c inititialize ko (pointer for output matrix)
c
      ko = 0
      do  7 i=1, n
         kold = ko
         kdiag = 0
         do 71 k = ia(i), ia(i+1) -1
            if (ja(k)  .gt. i) goto 71
            ko = ko+1
            ao(ko) = a(k)
            jao(ko) = ja(k)
            if (ja(k)  .eq. i) kdiag = ko
 71      continue
         if (kdiag .eq. 0 .or. kdiag .eq. ko) goto 72
c
c     exchange
c
         t = ao(kdiag)
         ao(kdiag) = ao(ko)
         ao(ko) = t
c
         k = jao(kdiag)
         jao(kdiag) = jao(ko)
         jao(ko) = k
 72      iao(i) = kold+1
 7    continue
c     redefine iao(n+1)
      iao(n+1) = ko+1
      return
c----------end-of-getl -------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine getu (n,a,ja,ia,ao,jao,iao)
      implicit none
      integer n, ia(*), ja(*), iao(*), jao(*)
      real(8) a(*), ao(*)
c------------------------------------------------------------------------
c this subroutine extracts the upper triangular part of a matrix
c and writes the result ao, jao, iao. The routine is in place in
c that ao, jao, iao can be the same as a, ja, ia if desired.
c-----------
c on input:
c
c n     = dimension of the matrix a.
c a, ja,
c    ia = matrix stored in a, ja, ia, format
c On return:
c ao, jao,
c    iao = upper triangular matrix (upper part of a)
c       stored in compressed sparse row format
c note: the diagonal element is the last element in each row.
c i.e. in  a(ia(i+1)-1 )
c ao, jao, iao may be the same as a, ja, ia on entry -- in which case
c getu will overwrite the result on a, ja, ia.
c
c------------------------------------------------------------------------
c local variables
      real(8) t
      integer ko, k, i, kdiag, kfirst
      ko = 0
      do  7 i=1, n
         kfirst = ko+1
         kdiag = 0
         do 71 k = ia(i), ia(i+1) -1
            if (ja(k)  .lt. i) goto 71
            ko = ko+1
            ao(ko) = a(k)
            jao(ko) = ja(k)
            if (ja(k)  .eq. i) kdiag = ko
 71      continue
         if (kdiag .eq. 0 .or. kdiag .eq. kfirst) goto 72
c     exchange
         t = ao(kdiag)
         ao(kdiag) = ao(kfirst)
         ao(kfirst) = t
c
         k = jao(kdiag)
         jao(kdiag) = jao(kfirst)
         jao(kfirst) = k
 72      iao(i) = kfirst
 7    continue
c     redefine iao(n+1)
      iao(n+1) = ko+1
      return
c----------end-of-getu -------------------------------------------------
c-----------------------------------------------------------------------
      end
c-
      subroutine csrmsr (n,a,ja,ia,ao,jao,wk,iwk)
      implicit none
      integer n
      real(8) a(*),ao(*),wk(n)
      integer ia(n+1),ja(*),jao(*),iwk(n+1)
c-----------------------------------------------------------------------
c other used variables
      integer k, j, iptr, ii, icount, i
c-----------------------------------------------------------------------
c Compressed Sparse Row   to      Modified - Sparse Row
c                                 Sparse row with separate main diagonal
c-----------------------------------------------------------------------
c converts a general sparse matrix a, ja, ia into
c a compressed matrix using a separated diagonal (referred to as
c the bell-labs format as it is used by bell labs semi conductor
c group. We refer to it here as the modified sparse row format.
c Note: this has been coded in such a way that one can overwrite
c the output matrix onto the input matrix if desired by a call of
c the form
c
c     call csrmsr (n, a, ja, ia, a, ja, wk,iwk)
c
c In case ao, jao, are different from a, ja, then one can
c use ao, jao as the work arrays in the calling sequence:
c
c     call csrmsr (n, a, ja, ia, ao, jao, ao,jao)
c
c-----------------------------------------------------------------------
c
c on entry :
c---------
c a, ja, ia = matrix in csr format. note that the
c            algorithm is in place: ao, jao can be the same
c            as a, ja, in which case it will be overwritten on it
c            upon return.
c
c on return :
c-----------
c
c ao, jao  = sparse matrix in modified sparse row storage format:
c          +  ao(1:n) contains the diagonal of the matrix.
c          +  ao(n+2:nnz) contains the nondiagonal elements of the
c             matrix, stored rowwise.
c          +  jao(n+2:nnz) : their column indices
c          +  jao(1:n+1) contains the pointer array for the nondiagonal
c             elements in ao(n+1:nnz) and jao(n+2:nnz).
c             i.e., for i .le. n+1 jao(i) points to beginning of row i
c             in arrays ao, jao.
c              here nnz = number of nonzero elements+1
c work arrays:
c------------
c wk    = real work array of length n
c iwk   = integer work array of length n+1
c
c notes:
c-------
c        Algorithm is in place.  i.e. both:
c
c          call csrmsr (n, a, ja, ia, ao, jao, ao,jao)
c          (in which  ao, jao, are different from a, ja)
c           and
c          call csrmsr (n, a, ja, ia, a, ja, wk,iwk)
c          (in which  wk, jwk, are different from a, ja)
c        are OK.
c--------
c coded by Y. Saad Sep. 1989. Rechecked Feb 27, 1990.
c-----------------------------------------------------------------------
      icount = 0
c
c store away diagonal elements and count nonzero diagonal elements.
c
      do 1 i=1,n
         wk(i) = 0.0d0
         iwk(i+1) = ia(i+1)-ia(i)
         do 2 k=ia(i),ia(i+1)-1
            if (ja(k) .eq. i) then
               wk(i) = a(k)
               icount = icount + 1
               iwk(i+1) = iwk(i+1)-1
            endif
 2       continue
 1    continue
c
c compute total length
c
      iptr = n + ia(n+1) - icount
c
c     copy backwards (to avoid collisions)
c
      do 500 ii=n,1,-1
         do 100 k=ia(ii+1)-1,ia(ii),-1
            j = ja(k)
            if (j .ne. ii) then
               ao(iptr) = a(k)
               jao(iptr) = j
               iptr = iptr-1
            endif
 100     continue
 500  continue
c
c compute pointer values and copy wk(*)
c
      jao(1) = n+2
      do 600 i=1,n
         ao(i) = wk(i)
         jao(i+1) = jao(i)+iwk(i+1)
 600  continue
      return
c------------ end of subroutine csrmsr ---------------------------------
c-----------------------------------------------------------------------
      end
c
      subroutine getdia (nrow,ncol,job,a,ja,ia,len,diag,idiag,ioff)
      implicit none
      real(8) diag(*),a(*)
      integer nrow, ncol, job, len, ioff, ia(*), ja(*), idiag(*)
c-----------------------------------------------------------------------
c further used variables
      integer izero
c-----------------------------------------------------------------------
c this subroutine extracts a given diagonal from a matrix stored in csr
c format. the output matrix may be transformed with the diagonal removed
c from it if desired (as indicated by job.)
c-----------------------------------------------------------------------
c our definition of a diagonal of matrix is a vector of length nrow
c (always) which contains the elements in rows 1 to nrow of
c the matrix that are contained in the diagonal offset by ioff
c with respect to the main diagonal. if the diagonal element
c falls outside the matrix then it is defined as a zero entry.
c thus the proper definition of diag(*) with offset ioff is
c
c     diag(i) = a(i,ioff+i) i=1,2,...,nrow
c     with elements falling outside the matrix being defined as zero.
c
c-----------------------------------------------------------------------
c
c on entry:
c----------
c
c nrow  = integer. the row dimension of the matrix a.
c ncol  = integer. the column dimension of the matrix a.
c job   = integer. job indicator.  if job = 0 then
c         the matrix a, ja, ia, is not altered on return.
c         if job.ne.0  then getdia will remove the entries
c         collected in diag from the original matrix.
c         this is done in place.
c
c a,ja,
c    ia = matrix stored in compressed sparse row a,ja,ia,format
c ioff  = integer,containing the offset of the wanted diagonal
c         the diagonal extracted is the one corresponding to the
c         entries a(i,j) with j-i = ioff.
c         thus ioff = 0 means the main diagonal
c
c on return:
c-----------
c len   = number of nonzero elements found in diag.
c         (len .le. min(nrow,ncol-ioff)-max(1,1-ioff) + 1 )
c
c diag  = real(8) array of length nrow containing the wanted diagonal.
c         diag contains the diagonal (a(i,j),j-i = ioff ) as defined
c         above.
c
c idiag = integer array of  length len, containing the poisitions
c         in the original arrays a and ja of the diagonal elements
c         collected in diag. a zero entry in idiag(i) means that
c         there was no entry found in row i belonging to the diagonal.
c
c a, ja,
c    ia = if job .ne. 0 the matrix is unchanged. otherwise the nonzero
c         diagonal entries collected in diag are removed from the
c         matrix and therefore the arrays a, ja, ia will change.
c         (the matrix a, ja, ia will contain len fewer elements)
c
c----------------------------------------------------------------------c
c     Y. Saad, sep. 21 1989 - modified and retested Feb 17, 1996.      c
c----------------------------------------------------------------------c
c     local variables
      integer istart, max, iend, i, kold, k, kdiag, ko
c
      izero = 0
      istart = max(izero,-ioff)
      iend = min(nrow,ncol-ioff)
      len = 0
      do 1 i=1,nrow
         idiag(i) = 0
         diag(i) = 0.0d0
 1    continue
c
c     extract  diagonal elements
c
      do 6 i=istart+1, iend
         do 51 k= ia(i),ia(i+1) -1
            if (ja(k)-i .eq. ioff) then
               diag(i)= a(k)
               idiag(i) = k
               len = len+1
               goto 6
            endif
 51      continue
 6    continue
      if (job .eq. 0 .or. len .eq.0) return
c
c     remove diagonal elements and rewind structure
c
      ko = 0
      do  7 i=1, nrow
         kold = ko
         kdiag = idiag(i)
         do 71 k= ia(i), ia(i+1)-1
            if (k .ne. kdiag) then
               ko = ko+1
               a(ko) = a(k)
               ja(ko) = ja(k)
            endif
 71      continue
         ia(i) = kold+1
 7    continue
c
c     redefine ia(nrow+1)
c
      ia(nrow+1) = ko+1
      return
c------------end-of-getdia----------------------------------------------
c-----------------------------------------------------------------------
      end
c

# HEADER ####################################################
# This is file spam/R/precmat.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# construct various precision matrices



precmat <- function(n, season=12, m=n, A=NULL, order=1, ... , type="RW1") {
  avtype <- c("rw1", "rw2", "rwn", "season","igmrfreglat","igmrfirreglat","gmrfreglat")
  method <- pmatch(tolower(type), avtype)
  if (is.na(method)) 
     stop("Precision matrix type not implemented yet. Please ask for.")
  switch(method,
         return(precmat.RW1(n)),
         return(precmat.RW2(n)),
         return(precmat.RWn(n, order)),
         return(precmat.season(n,season)),
         return(precmat.IGMRFreglat(n,m,...)),
         return(precmat.IGMRFirreglat(A,...)),
         return(precmat.GMRFreglat(n,m,...)))
}

precmat.IGMRFreglat <- function (n, m, order=1, anisotropy = 1)
{
  if((n<2)|(m<2))
    stop("n and m need to be >1")
  if(order==1) {
    if (anisotropy < 0 | anisotropy > 2) 
        stop("anisotropy parameter needs to be in [0,2]")
    return(kronecker(precmat.RW1(m), diag.spam(2-anisotropy, n)) + 
        kronecker(diag.spam(anisotropy, m), precmat.RW1(n)))
  } else if (order==2) {
    return(kronecker(precmat.RW2(m), diag.spam(n)) + 
        kronecker(diag.spam( m), precmat.RW2(n)))
  }
  if( (order<1)|(order>=min(n,m)))
        stop("order needs to be between 1 and min(n,m)-1")
  Dn <- diff.spam(diag.spam(n), lag = 1, differences = order) 
  Dm <- diff.spam(diag.spam(m), lag = 1, differences = order) 
  return(kronecker(t(Dm)%*%Dm, diag.spam(n)) +             
        kronecker(diag.spam( m), t(Dn)%*%Dn))
}

precmat.IGMRFirreglat <- function(A, eps= getOption("spam.eps")) {
  if(!is.spam(A)) A <- as.spam(A, eps)
  A@entries <- rep.int(1, length(A))
  test <- isSymmetric.spam(A, tol = eps * 100)
  if (!isTRUE(test)) 
    stop("Input matrix  not symmetric (up to 100*eps)",call. = FALSE)
  
  return(diag.spam( diff(A@rowpointers)) - A)
}


precmat.RW1 <- function(n) {
  if(n<2)
    stop("Dimension 'n' should be larger than two")
  Q <- spam(0,n,n)
  Q@entries <- rep(-1,n-1)
  Q@colindices <- as.integer( seq.int(2, to=n,by=1))
  Q@rowpointers <- as.integer(c(seq.int(1,to=n,by=1),n))

  return(Q + t(Q) + diag.spam(c(1, rep.int(2, n-2), 1)))
}

precmat.RW2<- function(n) {
  if(n<4)
    stop("Dimension 'n' should be larger than three")
  Q <- spam(0,n,n)
  Q@entries <- c(-2,1,rep(c(-4,1),n-3),-2)
  Q@colindices <- as.integer( c(rep(2:(n-1),each=2)+c(0,1),n))
  Q@rowpointers <- as.integer(c(seq.int(1,to=2*(n-1),by=2),2*(n-1),2*(n-1)))


  return(Q + t(Q) + diag.spam(c(1,5, rep.int(6, n-4), 5,1)))
}

precmat.RWn <- function(n, order=3)
{
  if( (order<1)|(order>=n))
	stop("order needs to be between 1 and n-1")
  D <- diff.spam(diag.spam(n), lag=1, differences=order)	
  return( t(D)%*%D )
}


precmat.season <- function(n, season=12) {
  if(n<2*season)
    stop("Dimension 'n' should be larger than twice the season")
  Q <- spam(0,n,n)

  m <- season
  first <- rev( outer(1:(m-1), (m-1):1, "pmin"))
  mid <- rep.int((m-1):1, n-2*m+2)
  tmp <- outer(1:(m-1), 1:(m-1), "pmin")
  last <- rev( tmp[upper.tri(tmp)])
  
  Q@entries <- c( first, mid, last)
  Q@rowpointers <- as.integer( c( seq.int( 1, length.out=n-m+1, by=m-1),    # first and mid
                                 (n-m)*(m-1)+1+cumsum((m-1):1),
                                 (m-1)*(n-m/2)+1)  # last
                              )
  Q@colindices <- as.integer( c( rep.int(1:(m-1), n-m+1) +
                                rep.int(1:(n-m+1), rep.int(m-1, n-m+1)), # first and mid
                               n-last+1 )        # last
                             )
  return(Q + t(Q) + diag.spam(c(1:m, rep.int(m, n-2*m), m:1)))

}

"precmat.GMRFreglat" <- function(n,m, par=.1, model = "m1p1",  eps = getOption("spam.eps")){
  if((n<2)|(m<2))
    stop("n and m need to be >1")
  dims <- c(n,m)
  if (model=="m1p1"){
    x <- numeric(dims[1])
    x[1:2] <- c(1,-par[1])
    y <- numeric(prod(dims))
    y[dims[1]+1] <- -par[1] 
    return( kronecker(diag.spam(dims[2]), toeplitz.spam(x,eps=eps))+toeplitz.spam(y,eps=eps))
  }
  
  if (model=="m1p2"){
    x <- numeric(dims[1])
    x[1:2] <- c(1,-par[1])
    y <- numeric(prod(dims))
    y[dims[1]+1] <- -par[2] 
    return( kronecker(diag.spam(dims[2]), toeplitz.spam(x,eps=eps))+toeplitz.spam(y,eps=eps))
  }
  
  if (model=="m2p3"){
    x <- numeric(dims[1])
    x[1:2] <- c(1,-par[1])
    
    y <- numeric(dims[1])
    y[1:2] <- c(-par[2],-par[3])

    z <- numeric(dims[2])
    z[2] <- 1


    p1 <- kronecker( diag.spam(dims[2]), toeplitz.spam(x,eps=eps))
    p2 <- kronecker( toeplitz.spam(z,eps=eps), toeplitz.spam(y,eps=eps))
    return( p1 + p2)
  }
  if (model=="m2p4"){
    x <- numeric(dims[1])
    x[1:2] <- c(1,-par[1])
    
    y <- numeric(dims[1])
    y[1:2] <- c(-par[2],-par[3])
    w <- numeric(dims[1])
    w[1:2] <- c(-par[2],-par[4])

    z <- numeric(dims[2])
    z[2] <- 1


    p1 <- kronecker( diag.spam(dims[2]), toeplitz.spam(x,eps=eps))
    p2 <- kronecker( toeplitz.spam(z,rep(0,dims[2]),eps=eps), toeplitz.spam(y,w,eps=eps))
    p3 <- kronecker( toeplitz.spam(rep(0,dims[2]),z,eps=eps), toeplitz.spam(w,y,eps=eps))
    return( p1 + p2 + p3)
  }
  stop("Model not implemented yet!")

}

# HEADER ####################################################
# This is file spam/R/random.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



spam_random <- function(nrow = 1L, ncol = nrow, density = 0.5, distribution = NULL,
                        digits = NULL, sym = FALSE, spd = FALSE, verbose = FALSE, ...) {

  if (!is.numeric(nrow) || length(nrow) != 1)
    stop("nrow must be a single numeric value.")
  if (!is.numeric(ncol) || length(ncol) != 1)
    stop("ncol must be a single numeric value.")
  if (!is.numeric(density) || length(density) != 1 || density < 0 || density > 1)
    stop("density must be a single numeric value, equal or between 0 and 1.")
  if (!(is.null(distribution) || (is.function(distribution) && any(names(formals(distribution)) == "n"))))
    stop("distribution must be a function which generates random deviates and must have an argument \"n\".")
  if (!(is.null(digits) || (is.numeric(digits) && length(digits) == 1 && digits >= 0)))
    stop("digits must be a single numeric value larger or equal to 0.")
  if (!is.logical(verbose) || length(verbose) != 1)
    stop("verbose must be a single locigal value.")
  if (!is.logical(sym) || length(sym) != 1)
    stop("sym must be a single locigal value.")
  if (!is.logical(spd) || length(spd) != 1)
    stop("spd must be a single locigal value.")
  if (spd && ((ncol != nrow) || density == 0))
    stop("if spd == TRUE, then it must hold that ncol == nrow and density > 0")

  if (density == 1)
    npercol <- rep.int(ncol, nrow)
  else if (density == 0)
    return(spam(0, nrow, ncol))
  else
    npercol <- as.integer(rbinom(n = nrow, size = ncol, prob = density))

  rowp <- c(1L, cumsum(npercol) + 1L)
  n <- sum(npercol)
  if (n == 0) {
    if (sym)
      if (is.null(distribution))
        return(diag.spam(nrow = nrow, ncol = ncol))
    else
      return(diag.spam(x = distribution(min(nrow, ncol), ...),
                       nrow = nrow, ncol = ncol))
    else
      return(spam(0, nrow = nrow, ncol = ncol))
  }

  coli <- integer(n)
  seqcol <- seq_len(ncol)
  for (i in seq_len(nrow)) {
    if (rowp[i] == rowp[i+1L])
      next
    coli[rowp[i]:(rowp[i+1L]-1L)] <- sort(sample(seqcol, npercol[i], replace = FALSE))
  }

  if (is.null(distribution))
    entries <- rep.int(1, n)
  else
    entries <- distribution(n, ...)
  if (!is.null(digits))
    entries <- round(entries, digits)

  rspam <- .newSpam(entries,
                    colindices = coli,
                    rowpointers = rowp,
                    dimension = c(nrow, ncol),
                    force64 = getOption("spam.force64"))

  if (spd) {
    sym <- TRUE
    diag.spam(rspam) <- diag.spam(rspam) + 1 }

  if (sym) {
    s_t <- t.spam(rspam)
    rspam <- s_t[lower.tri(s_t, diag = TRUE)] + rspam[upper.tri(rspam, diag = TRUE)]
  }

  if (spd) {
    st <- rspam
    st@entries <- abs(st@entries)
    st_dia <- diag.spam(st)
    st_rs <- rowSums(st)
    s_diag <- diag(rspam)
    ind <- 2*st_dia <= st_rs
    diag.spam(rspam)[ind] <- ifelse(s_diag >=0, st_rs+1, -st_rs-1)[ind]
  }

  if(verbose)
    message("Density is ", round(n/prod(rspam@dimension), 7), ", specified is ", density,
      " (nnz=",n,").")

  return(rspam)
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-random.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



rm(list = ls())
source("helper.R")

context("test-random.R")

test_that("check arguments", {
  rspam1 <- spam::spam_random(7, digits = 2, distribution = runif, min = -1, sym = TRUE)
  expect_identical(spam::isSymmetric.spam(rspam1), TRUE)

  rspam2 <- spam::spam_random(7, digits = 2, distribution = NULL, sym = FALSE, spd = TRUE)
  cholrspam2 <- chol(rspam2)

  expect_identical(spam::isSymmetric.spam(rspam2), TRUE)
  expect_equivalent(class(cholrspam2), "spam.chol.NgPeyton")
})

# HEADER ####################################################
# This is file spam/R/helper.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################
########################################################################
# a few nice helper functions:


bandwidth <- function(A) {
    if (!is.spam(A)) {
        warning("Matrix not 'spam' object. Coerced to one")
        A <- as.spam(A)
    }
    if( getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(A)
    
    ret <- .C64("getbwd",
                SIGNATURE = c( SS$signature, SS$signature, SS$signature,
                    SS$signature, SS$signature),
                
                A@dimension[1],
                A@colindices,
                A@rowpointers,
                
                low = vector_dc(SS$type, 1),
                upp = vector_dc(SS$type, 1),

                INTENT = c("r", "r", "r",
                    "w", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)
    return(c(ret$low,ret$upp))
}
                  
  

bdiag.spam <- function(...){
  nargs <- nargs()
  if (nargs == 0)     return( NULL)
  args <- list(...)
  args[which(sapply(args, is.null))] <- NULL

  if (nargs == 1)     return( args[[1]])
  if (nargs == 2) {
    # Classical case, concatenate two matrices
    A <- args[[1]]
    B <- args[[2]]
    if(!is.spam(A))
      A <- as.spam(A)
    if(!is.spam(B))
      B <- as.spam(B)
    dimA <- A@dimension
    lenA <- length(A@entries)

    A@entries <- c(A@entries,B@entries)
    A@colindices <- c(A@colindices,B@colindices+dimA[2])
    A@rowpointers <- c(A@rowpointers,B@rowpointers[-1]+lenA) 
    A@dimension <-  dimA+B@dimension
    return(A)
  } else {
    # "recursive" approach only, e.g. no checking
    tmp <- bdiag.spam( args[[1]], args[[2]])
    for ( i in 3:nargs)
      tmp <- bdiag.spam( tmp, args[[i]])
    return( tmp)
  }
}


adjacency.landkreis <- function(loc)
  # this reads the germany graph file provide by
  # loc <- "http://www.math.ntnu.no/~hrue/GMRF-book/germany.graph"
  # or
  # loc <- system.file("demodata/germany.graph", package="INLA")
  # 
  {
    n <- as.numeric( readLines(loc, n=1))

    nnodes <- nodes <- numeric( n)

    adj <- list()
    for (i in 1:n) {
      tmp <- as.numeric(scan(loc, skip=i, nlines=1, quiet=T, what=list(rep("",13)))[[1]])
      nodes[i] <- tmp[1]
      nnodes[i] <- tmp[2]
      adj[[i]] <- tmp[-c(1:2)]
    }
    
    adj <- adj[ order(nodes)]
    nnodes <- nnodes[ order(nodes)]
    A <- spam(0)
    A@colindices <- as.integer( unlist(adj)+1)
    A@rowpointers <- as.integer( c(1,cumsum(lapply(adj, length))+1))
    A@entries <- rep(1, length(unlist(adj)))
    A@dimension <- as.integer( c(n, n))
    return(A)
  }

map.landkreis <- function(data, col=NULL, zlim=range(data), add=FALSE, legendpos=c( 0.88,0.9,0.05,0.4))
# This is a stripped-down version of the function provided by the INLA package.
# Added color argument, changed "append" to "add".
# Legend is tuned for a mai=rep(0,4) call 
{
  npoly <- length(spam::germany)
  ymax <- ymin <- xmax <- xmin <- 1:npoly

  if (length(data)!=npoly)
    stop("data has wrong length")
  
  if (is.null(col)) {
      if (requireNamespace("fields", quietly = TRUE)) {
          col <- fields::tim.colors(64)
      } else {
          col <- gray(seq(.05,to=0.95,length=64))
      }
  }
  ncol <- length(col)
  polycol <- col[round(((data-zlim[1])/diff(zlim)+1e-6)*(ncol-1))+1]
  
  for(i in 1:length(spam::germany)) {
    xmin[i] <- min(spam::germany[[i]][,2],na.rm=T)
    xmax[i] <- max(spam::germany[[i]][,2],na.rm=T)
    ymin[i] <- min(spam::germany[[i]][,3],na.rm=T)
    ymax[i] <- max(spam::germany[[i]][,3],na.rm=T)
  }


  if (!add)
    plot(c(min(xmin),max(xmax)),c(min(ymin),max(ymax)), type="n", axes=F, xlab="", ylab="")
  for(k in npoly:1)
    polygon(spam::germany[[k]][,2],spam::germany[[k]][,3],col=polycol[k])
  if (requireNamespace("fields", quietly = TRUE)) 
    fields::image.plot(as.matrix(data), zlim=zlim, legend.only=T, smallplot=legendpos, cex=.2, col=col)

  invisible()
}



germany.plot <- function(vect,  col=NULL, zlim=range(vect), legend=TRUE,
             main=NULL, cex.axis=1, cex.main=1.5, add=FALSE, ... )
{
  if (length(vect) != spam::germany.info$n)
        stop("data has wrong length")

  if (!add) {
    par(mai=c(.1,.1,.1,.3))
    plot(0,0, xlim=spam::germany.info$xlim, ylim=spam::germany.info$ylim,
         type = "n", axes = F, xlab = "", ylab = "")
  }
  if (is.null(col)) {
      ## from: require(RColorBrewer); col <- colorRampPalette(brewer.pal(9,"Blues"))(100)
    col <- c("#F7FBFF", "#F4F9FE", "#F2F8FD", "#F0F7FD", "#EEF5FC", "#ECF4FB", "#EAF3FB", "#E8F1FA", "#E6F0F9", "#E4EFF9", "#E2EEF8", "#E0ECF7", "#DEEBF7", "#DCEAF6", "#DAE8F5", "#D8E7F5", "#D6E6F4", "#D5E5F4", "#D3E3F3", "#D1E2F2", "#CFE1F2", "#CDDFF1", "#CBDEF0", "#C9DDF0", "#C7DBEF", "#C5DAEE", "#C1D9ED", "#BED7EC", "#BBD6EB", "#B8D5EA", "#B5D3E9", "#B1D2E7", "#AED1E6", "#ABCFE5", "#A8CEE4", "#A4CCE3", "#A1CBE2", "#9ECAE1", "#9AC8E0", "#96C5DF", "#92C3DE", "#8EC1DD", "#89BEDC", "#85BCDB", "#81BADA", "#7DB8DA", "#79B5D9", "#75B3D8", "#71B1D7", "#6DAFD6", "#69ACD5", "#66AAD4", "#62A8D2", "#5FA6D1", "#5CA3D0", "#58A1CE", "#559FCD", "#529DCC", "#4E9ACB", "#4B98C9", "#4896C8", "#4493C7", "#4191C5", "#3E8EC4", "#3C8CC3", "#3989C1", "#3686C0", "#3484BE", "#3181BD", "#2E7EBC", "#2C7CBA", "#2979B9", "#2776B8", "#2474B6", "#2171B5", "#1F6FB3", "#1D6CB1", "#1B69AF", "#1967AD", "#1764AB", "#1562A9", "#135FA7", "#115CA5", "#0F5AA3", "#0D57A1", "#0B559F", "#09529D", "#084F9A", "#084D96", "#084A92", "#08478E", "#08458A", "#084286", "#083F82", "#083D7E", "#083A7A", "#083776", "#083572", "#08326E", "#08306B")
  }
  ncol <- length(col)
  polycol <- (col)[round((((vect) - zlim[1])/diff(zlim) + 1e-06) *
                            (ncol - 1)) + 1]
 
  polygon( spam::germany.poly[17965L:1L,],
          col = (polycol[spam::germany.info$polyid])[599L:1L], ...)

  if (legend&&requireNamespace("fields", quietly = TRUE)){
    legendpos <- c(0.845, 0.89, 0.05, 0.4)
    fields::image.plot(as.matrix(vect), zlim = zlim, legend.only = TRUE,
               smallplot = legendpos, axis.args=list(cex.axis=cex.axis,lwd=0, lwd.ticks=1.3),
               col = col)
  }

  if(!is.null(main))
    text( min(spam::germany.info$xlim), max(spam::germany.info$ylim), main, cex=cex.main, adj=c(0,1))

  invisible()
}



grid_zoom <- function(inputGrob = pointsGrob(runif(200),runif(200)),
                      inputViewport = viewport(name="main"),
                      x = "topleft", y, just,
                      ratio = c(.3,.4), zoom_xlim, zoom_ylim,
                      rect = TRUE, rect_lwd = 1, rect_fill = "gray92",
                      draw =TRUE, zoom_fill = "white", zoom_frame_gp = gpar(lwd = 1),
                      zoom_gp = NULL, zoom_xaxis = xaxisGrob(main = FALSE), zoom_yaxis = NULL) {

## inputGrob <- pointsGrob(runif(100), runif(100), pch=".", gp=gpar(cex=4),default.units="native",name="cc") 
## inputViewort <- viewport(name="main")
## x <- "topleft"
## ratio <- unit(c(.3,.3), "npc")
## zoom_xlim <- c(0.1,.5)
## zoom_ylim <- c(0.1,.5)
## rect <- TRUE
## rect_lwd = 1
## rect_fill = "gray92"
## zoom_fill = "white"
## zoom_frame_gp = gpar(lwd = 1)
## draw = TRUE
## zoom_gp = NULL

#  cat("may not work if other units than \"native\" and if the inputGrob has a complex structure. \n")
  
  if (!identical(length(ratio), 1)) 
    ratio <- c(ratio, ratio)
  if(inherits(x,"character"))  # old: (class(x) == "character")
    switch(x,
           topleft = {x = 0; y = 1; just = c(0, 1)},
           topright = {x = 1; y = 1 ; just = c(1, 1)},
           bottomright = {x = 1; y = 0; just = c(1, 0)},
           bottomleft = {x = 0; y = 0; just = c(0, 0)})

  inputViewport$name <- "main"
  vp <- vpStack(inputViewport,
                vpList(viewport(name="zoom", x = unit(x,"npc"), y = unit(y,"npc"), width=unit(ratio[1],"npc"), height=unit(ratio[2],"npc"), just = just, xscale=zoom_xlim, yscale=zoom_ylim, default.units="native", clip = "on"),
                       viewport(name="zoom_noClip", x = unit(x,"npc"), y = unit(y,"npc"), width=unit(ratio[1],"npc"), height=unit(ratio[2],"npc"), just = just, xscale=zoom_xlim, yscale=zoom_ylim, default.units="native", clip = "off")))
  
  inputGrob <- editGrob(inputGrob, name="main", vp="main")
  zoom_grob <- editGrob(inputGrob, name="zoom", vp="main::zoom")

  if(!is.null(zoom_gp))
    zoom_grob <- editGrob(inputGrob, name="zoom", vp="main::zoom", gp=zoom_gp)
  
  if(!is.null(zoom_xaxis))
    zoom_xaxis <- editGrob(zoom_xaxis, vp="main::zoom_noClip", name = "xaxis")

  if(!is.null(zoom_yaxis))
    zoom_yaxis <- editGrob(zoom_yaxis, vp="main::zoom_noClip", name = "yaxis")

  
  rect <- rectGrob(x = zoom_xlim[1], y = zoom_ylim[1], just = c(0,0),
                    width = diff(zoom_xlim), height = diff(zoom_ylim),
                    default.units = "native", vp = "main", gp = gpar(lwd = rect_lwd, fill=rect_fill))

  rect_frame <- rectGrob(x = zoom_xlim[1], y = zoom_ylim[1], just = c(0,0),
                    width = diff(zoom_xlim), height = diff(zoom_ylim),
                    default.units = "native", vp = "main", gp = gpar(lwd = rect_lwd))
  
  
  gr <- gList(rect, inputGrob, rectGrob(vp="main::zoom", gp=gpar(fill=zoom_fill)), zoom_grob, rectGrob(vp="main::zoom_noClip", gp=zoom_frame_gp), zoom_xaxis, zoom_yaxis, rect_frame)

  out <- gTree(children=gr, childrenvp = vp)

  if (draw)
    grid.draw(out)
  invisible(out)
}


grid_trace2 <- function (chain1, chain2 = NULL,
                         xlim = NULL, ylim1 = NULL, ylim2=NULL,
                         chain1_lab = "", chain2_lab = "", main = "",
                         chain1_yaxis_at = NULL, chain2_yaxis_at = NULL,
                         log = FALSE,
                         cex_points = unit(.2, "mm"),
                         cex_main = unit(1.2, "mm"),
                         lwd_lines = unit(.1, "mm"),
                         lwd_frame = unit(.8, "mm"),
                         draw = TRUE) 
{
    if (is.null(chain2)) {
        chain2 <- chain1[, 2]
        chain1 <- chain1[, 1]
    }
    if (log) {
        chain1 <- log(chain1)
        chain2 <- log(chain2)
    }
    stopifnot(identical(length(chain1), length(chain2)))
    
    n <- length(chain1)
    if(!is.null(xlim)){
      stopifnot(length(xlim)==2)
      chain1.sub <- chain1[xlim[1]:xlim[2]]
      chain2.sub <- chain2[xlim[1]:xlim[2]]
    }else{
      chain1.sub <- chain1
      chain2.sub <- chain2
    }
    if(!is.null(ylim1))
      stopifnot(length(ylim1)==2)
    if(!is.null(ylim2))
      stopifnot(length(ylim2)==2)
       
    vp1 <- plotViewport(unit(c(2.5, 3, 2.5, 2), "cm"), name = "frame")
    vp2 <- viewport(layout = grid.layout(nrow = 1, ncol = 3, 
        respect = rbind(c(0, 0, 1)), widths = unit(c(1, 0.3, 
            1), c("null", "cm", "null"))), name = "lay1")
    vp3 <- viewport(layout.pos.col = 1, name = "traces")
    vp4 <- viewport(layout = grid.layout(nrow = 2, ncol = 1), 
        name = "lay2")
    vp5 <- viewport(layout.pos.row = 1, name = "trace1")
    vp5data <- dataViewport(xData = 1L:n, yData = chain1,
                            xscale = xlim, yscale = ylim1,
                            extension = c(0.02, 
        0.03), name = "trace1data", clip="off")
    vp5data_clip <- dataViewport(xData = 1L:n, yData = chain1,
                            xscale = xlim, yscale = ylim1,
                            extension = c(0.02, 
                              0.03), name = "trace1data_clip", clip="on")

    vp6 <- viewport(layout.pos.row = 2, name = "trace2")
    vp6data_clip <- dataViewport(xData = 1L:n, yData = chain2,
                            xscale = xlim, yscale = ylim2,
                            extension = c(0.02, 
        0.03), name = "trace2data_clip", clip="on")
    vp6data <- dataViewport(xData = 1L:n, yData = chain2,
                            xscale = xlim, yscale = ylim2,
                            extension = c(0.02, 
                              0.03), name = "trace2data", clip="off")

    vp7 <- viewport(layout.pos.col = 3, name = "scatter")
    vp7data_clip <- dataViewport(xData = chain1, yData = chain2,
                            xscale = ylim1, yscale = ylim2,
                            extension = 0.03, 
                                 name = "scatterData_clip", clip="on")
    vp7data <- dataViewport(xData = chain1, yData = chain2,
                            xscale = ylim1, yscale = ylim2,
                            extension = 0.03, 
        name = "scatterData", clip="off")

    vps <- vpStack(vp1, vp2, vpList(vpStack(vp3, vp4, vpList(vpStack(vp5, 
        vp5data, vp5data_clip), vpStack(vp6, vp6data, vp6data_clip))), vpStack(vp7, vp7data, vp7data_clip)))

   
    grs <- gList(rectGrob(vp = "frame::lay1::traces::lay2::trace1", 
        gp = gpar(lwd = lwd_frame), name = "rect_trace1"),
                 linesGrob(x = 1L:n, y = chain1, gp = gpar(lty = 1, lwd = lwd_lines), default.units = "native",  vp = "frame::lay1::traces::lay2::trace1::trace1data::trace1data_clip",  name = "lines_chain1"),
                 yaxisGrob(at = chain1_yaxis_at,
                           vp = "frame::lay1::traces::lay2::trace1::trace1data", 
                           name = "yaxis_chain1"),
                 rectGrob(vp = "frame::lay1::traces::lay2::trace2", 
                          gp = gpar(lwd = lwd_frame), name = "rect_trace2"),
                 linesGrob(x = 1L:n, 
                           y = chain2, gp = gpar(lty = 1, lwd = lwd_lines), default.units = "native", 
                           vp = "frame::lay1::traces::lay2::trace2::trace2data::trace2data_clip", 
                           name = "lines_chain2"),
                 yaxisGrob(at = chain2_yaxis_at, 
                           vp = "frame::lay1::traces::lay2::trace2::trace2data", 
                           name = "yaxis_chain2"),
                 xaxisGrob(vp = "frame::lay1::traces::lay2::trace2::trace2data", 
                           name = "xaxis_chains"),
                 pointsGrob(x = chain1.sub, y = chain2.sub, 
                            pch = 20, gp = gpar(cex = cex_points),
                            default.units = "native", 
                            vp = "frame::lay1::scatter::scatterData::scatterData_clip",
                            name = "points_scatter"), 
                 rectGrob(vp = "frame::lay1::scatter::scatterData",
                          gp = gpar(lwd = lwd_frame),  name = "rect_scatter"),
                 textGrob(chain1_lab, y = unit(-1, "line") - unit(0.2, "cm"),
                          vp = "frame::lay1::scatter", 
                          name = "text_scatter_lab1"),
                 textGrob(chain2_lab, 
                          x = unit(1, "npc") + unit(0.5, "cm"), rot = 90,
                          vp = "frame::lay1::scatter", 
                          name = "text_scatter_lab2"),
                 textGrob(main, 
                          x = unit(.5, "npc") + grobHeight(rectGrob())*.5,
                          y = unit(1, "npc") + max(stringHeight("Fg"),unit(.6, "cm")),
                          vp = "frame::lay1::traces", 
                          name = "title",
                          just=c(.5, .5),
                          gp=gpar(cex=cex_main))
                 )
    out <- gTree(childrenvp = vps, children = grs)
    if (draw) 
        grid.draw(out)
    invisible(out)
}

# HEADER ####################################################
# This is file spam/tests/testthat/test-helper.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-helper.R")



########


set.seed(14)

# bdiag.spam:

A <- spam(rnorm(10),2)
B <- spam(rnorm(16),4)

test_that("bdiag.spam", {
    spamtest_eq( bdiag.spam(A),A)

    spamtest_eq( bdiag.spam(A,B),rbind(cbind(A,rep(0,8)),
                                       cbind(spam(rep(0,20),4),B)))
})

# HEADER ####################################################
# This is file spam/R/defunct.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


validspamobject <- function(...) {
  .Defunct(new = 'validate_spam', package = 'spam', msg = "validspamobject() is defunct. Use validate_spam()")
}

# HEADER ####################################################
# This is file spam/R/foreign.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################




# Contains two sections:
# 1) Routines to transform spam objects to SparseM and Matrix
# 2) Functions to read (and write) MM and HB formats.



# 1a)  spam <-> SparseM
as.spam.matrix.csr <- function(x)
  {
#    if (is.matrix.csr(x)) {
      newx <- new("spam")
      slot(newx,"entries",check=FALSE) <- as.double( x@ra)
      slot(newx,"colindices",check=FALSE) <- x@ja
      slot(newx,"rowpointers",check=FALSE) <- x@ia
      slot(newx,"dimension",check=FALSE) <- x@dimension
      return(newx)
#    } else stop("Wrong object passed to "as.spam.matrix.csr"")
  }


# The following should not be necessary because it is
# as."matrix.csr".spam and not "as.matrix".csr.spam.
# Is there anyway around this?

#as.matrix.csr.spam <- function(x,...) {
#  if (require("SparseM")){
#    newx <- new("matrix.csr")
#    slot(newx,"ra",check=FALSE) <- x@entries
#    slot(newx,"ja",check=FALSE) <- x@colindices
#    slot(newx,"ia",check=FALSE) <- x@rowpointers
#    slot(newx,"dimension",check=FALSE) <- x@dimension
#    return(newx)
#  }

#}


# 1b) spam <-> Matrix

as.dgRMatrix.spam <- function(x) {
    if(.format.spam(x)$package != "spam"){
        stop("dgRMatrix structure is not compatible with numeric/large integer type.")
    }
    if (requireNamespace("Matrix")) {
      newx <- new(p=0:0,"dgRMatrix")
      slot(newx,"x",check=FALSE) <- x@entries
      slot(newx,"j",check=FALSE) <- x@colindices-1L
      slot(newx,"p",check=FALSE) <- x@rowpointers-1L
      slot(newx,"Dim",check=FALSE) <- x@dimension
      return(newx)
    }
  }

as.dgCMatrix.spam <- function(x)  {
    if(.format.spam(x)$package != "spam"){
        stop("dgCMatrix structure is not compatible with numeric/large integer type.")
    }
    if (requireNamespace("Matrix")) {
      dimx <- x@dimension
      nz <- x@rowpointers[dimx[1] + 1] - 1
      z <- .Fortran("transpose", n = as.integer(dimx[1]), m = as.integer(dimx[2]),
                    a = as.double(x@entries),ja = as.integer(x@colindices), ia = as.integer(x@rowpointers),
                    entries = vector("double",nz), colindices = vector("integer", nz),
                    rowpointers = vector("integer", dimx[2] + 1),
                    NAOK = getOption("spam.NAOK"),
                    PACKAGE = "spam")
      ## SS <- .format.spam(x)
      ## z <- .C64("transpose",
      ##           SIGNATURE = c( SS$signature, SS$signature,
      ##                         "double", SS$signature, SS$signature,
      ##                         "double", SS$signature, SS$signature),

      ##           n = dimx[1],
      ##           m = dimx[2],

      ##           a = x@entries,
      ##           ja = x@colindices,
      ##           ia = x@rowpointers,

      ##           entries = vector_dc("double",nz),
      ##           colindices = vector_dc(SS$type, nz),
      ##           rowpointers = vector_dc(SS$type, dimx[2] + 1),

      ##           INTENT = c("r", "r",
      ##                      "r", "r", "r",
      ##                      "w", "w", "w"),
      ##           NAOK = getOption("spam.NAOK"),
      ##           PACKAGE = SS$package)
      newx <- new(p=0:0,"dgCMatrix")
      slot(newx,"x",check=FALSE) <- z$entries
      slot(newx,"i",check=FALSE) <- z$colindices-1L
      slot(newx,"p",check=FALSE) <- z$rowpointers-1L
      slot(newx,"Dim",check=FALSE) <- dimx
      return(newx)
    }
  }


as.spam.dgRMatrix <- function(x)  {

    if (is(x,"dgRMatrix")){
      if (identical(length(x@x),0L))  # zero matrix
        return(new("spam",rowpointers=c(1L,rep.int(2L,x@Dim[1])), dimension=x@Dim))

      newx <- new("spam")
      slot(newx,"entries",check=FALSE) <- x@x
      slot(newx,"colindices",check=FALSE) <- x@j+1L
      slot(newx,"rowpointers",check=FALSE) <- x@p+1L
      slot(newx,"dimension",check=FALSE) <- x@Dim
      return(newx)
    }
    stop("Wrong object passed to 'as.spam.dgRMatrix'")
  }

as.spam.dgCMatrix <- function(x)  {

    if (is(x,"dgCMatrix")){
      if (identical(length(x@x),0L))  # zero matrix
        return(new("spam",rowpointers=c(1L,rep.int(2L,x@Dim[1])), dimension=x@Dim))

      nz <- x@p[x@Dim[2] + 1]
      z <- .Fortran("transpose", n = as.integer(x@Dim[2]), m = as.integer(x@Dim[1]),
                    a = as.double(x@x),ja = as.integer(x@i+1L), ia = as.integer(x@p+1L),
                    entries = vector("double",nz), colindices = vector("integer", nz),
                    rowpointers = vector("integer", x@Dim[1] + 1),
                    NAOK = getOption("spam.NAOK"),
                    PACKAGE = "spam")
      newx <- new("spam")
      slot(newx,"entries",check=FALSE) <- z$entries
      slot(newx,"colindices",check=FALSE) <- z$colindices
      slot(newx,"rowpointers",check=FALSE) <- z$rowpointers
      slot(newx,"dimension",check=FALSE) <- x@Dim
      return(newx)
    }
    stop("Wrong object passed to 'as.spam.dgCMatrix'")
  }



# 2) Import and export
# taken from Matrix 0.999375-10 and adapted for spam
## Utilities for the Harwell-Boeing and MatrixMarket formats


readone <- function(ln, iwd, nper, conv)
# By Bates/Maechler from Matrix 0.999375-10
{
    ln <- gsub("D", "E", ln)
    inds <- seq(0, by = iwd, length = nper + 1)
    (conv)(substring(ln, 1 + inds[-length(inds)], inds[-1]))
}

readmany <- function(conn, nlines, nvals, fmt, conv)
# By Bates/Maechler from Matrix 0.999375-10
{
    if (!grep("[[:digit:]]+[DEFGI][[:digit:]]+", fmt))
	stop("Not a valid format")
    Iind <- regexpr("[DEFGI]", fmt)
    nper <- as.integer(substr(fmt, regexpr("[[:digit:]]+[DEFGI]", fmt), Iind - 1))
    iwd <- as.integer(substr(fmt, Iind + 1, regexpr("[\\.\\)]", fmt) - 1))
    rem <- nvals %% nper
    full <- nvals %/% nper
    ans <- vector("list", nvals %/% nper)
    for (i in seq_len(full))
	ans[[i]] <- readone(readLines(conn, 1, ok = FALSE),
			    iwd, nper, conv)
    if (!rem) return(unlist(ans))
    c(unlist(ans),
      readone(readLines(conn, 1, ok = FALSE), iwd, rem, conv))
}

read.HB <- function(file)
# Adapted from Bates/Maechler Matrix 0.999375-10 version
{
    if (is.character(file))
	file <- if (file == "") stdin() else file(file)
    if (!inherits(file, "connection"))
        stop("'file' must be a character string or connection")
    if (!isOpen(file)) {
        open(file)
        on.exit(close(file))
    }
    hdr <- readLines(file, 4, ok = FALSE)
    Title <- sub("[[:space:]]+$", "", substr(hdr[1], 1, 72))
    Key <- sub("[[:space:]]+$", "", substr(hdr[1], 73, 80))
    totln <- as.integer(substr(hdr[2], 1, 14))
    ptrln <- as.integer(substr(hdr[2], 15, 28))
    indln <- as.integer(substr(hdr[2], 29, 42))
    valln <- as.integer(substr(hdr[2], 43, 56))
    rhsln <- as.integer(substr(hdr[2], 57, 70))
    if (!(t1 <- substr(hdr[3], 1, 1)) %in% c("C", "R", "P"))
        stop(paste("Invalid storage type:", t1))
    if (t1 != "R") stop("Only numeric sparse matrices allowed")
    ## _FIXME: Patterns should also be allowed
    if (!(t2 <- substr(hdr[3], 2, 2)) %in% c("H", "R", "S", "U", "Z"))
        stop(paste("Invalid storage format:", t2))
    if (!(t3 <- substr(hdr[3], 3, 3)) %in% c("A", "E"))
        stop(paste("Invalid assembled indicator:", t3))
    nr <- as.integer(substr(hdr[3], 15, 28))
    nc <- as.integer(substr(hdr[3], 29, 42))
    nz <- as.integer(substr(hdr[3], 43, 56))
    nel <- as.integer(substr(hdr[3], 57, 70))
    ptrfmt <- toupper(sub("[[:space:]]+$", "", substr(hdr[4], 1, 16)))
    indfmt <- toupper(sub("[[:space:]]+$", "", substr(hdr[4], 17, 32)))
    valfmt <- toupper(sub("[[:space:]]+$", "", substr(hdr[4], 33, 52)))
    rhsfmt <- toupper(sub("[[:space:]]+$", "", substr(hdr[4], 53, 72)))
    if (!is.na(rhsln) && rhsln > 0) {
        h5 <- readLines(file, 1, ok = FALSE)
    }
    ptr <- readmany(file, ptrln, nc + 1, ptrfmt, as.integer)
    ind <- readmany(file, indln, nz, indfmt, as.integer)
    vals <- readmany(file, valln, nz, valfmt, as.numeric)

    # Spam related changes:
    if (t3 =="E")
        stop("Only assembled Harwell-Boeing formats implemented")
    z <- .Fortran("transpose", n = as.integer(nc), m = as.integer(nr),
                  a = as.double(vals),ja = as.integer(ind), ia = as.integer(ptr),
                  entries = vector("double",nz), colindices = vector("integer", nz),
                  rowpointers = vector("integer", nr + 1),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = "spam")
    newx <- new("spam")
    slot(newx,"entries",check=FALSE) <- z$entries
    slot(newx,"colindices",check=FALSE) <- z$colindices
    slot(newx,"rowpointers",check=FALSE) <- z$rowpointers
    slot(newx,"dimension",check=FALSE) <- c(nr, nc)
    if (t2 %in% c("H", "S"))
      newx <- newx+t.spam(newx)-diag.spam(diag(newx))
    if (t2 =="Z")
      newx <- newx-t.spam(newx)
    return(newx)
}



# alternatives are implementing
# http://math.nist.gov/MatrixMarket/mmio/f/mmiof77.html
read.MM <- function(file)  {
  if (is.character(file))
    file <- if(file == "") stdin() else file(file)
  if (!inherits(file, "connection"))
    stop("'file' must be a character string or connection")
  if (!isOpen(file)) {
    open(file)
    on.exit(close(file))
  }
  scan1 <- function(what, ...)
    scan(file, nmax = 1, what = what, quiet = TRUE, ...)

  if ((hdr <- tolower(scan1(character()))) != "%%matrixmarket")  # RF: added a to lower
    stop("file is not a MatrixMarket file")
  if (!(typ <- tolower(scan1(character()))) %in% "matrix")
    stop("type '", typ, "' not recognized")
  if (!(repr <- tolower(scan1(character()))) %in% c("coordinate", "array"))
    stop("representation '", repr, "' not recognized")
  elt <- tolower(scan1(character()))
  if (!elt %in% c("real", "complex", "integer", "pattern"))
    stop("element type '", elt, "' not recognized")
  sym <- tolower(scan1(character()))
  if (!sym %in% c("general", "symmetric", "skew-symmetric", "hermitian"))
    stop("symmetry form '", sym, "' not recognized")
  nr <- scan1(integer(), comment.char = "%")
  nc <- scan1(integer())
    # code from now on differs from Matrix one...
  if (repr == "coordinate") {
    nz <- scan1(integer())
    switch(elt,
           "real"    = { what <- list(i= integer(), j= integer(), x= numeric())},
           "integer" = { what <- list(i= integer(), j= integer(), x= numeric())
                         warning("'integer' format coerced to 'double'", call. = FALSE)         },
           "pattern" = { what <-  list(i= integer(), j= integer())
                         warning("matrix elements assumed as 1 ('pattern' format)", call. = FALSE)    },
           "complex" = { what <- list(i= integer(), j= integer(), x= numeric(), y= numeric())
                         warning("retaining only real part of 'complex' format", call. = FALSE) }           )

    z <- scan(file, nmax = nz, quiet = TRUE, what= what)
    newx <- spam.list(list(ind=cbind(z$i,z$j),x= if(elt=="pattern") rep.int(1,nz) else z$x ), nr,nc)

    if (sym %in% c("symmetric", "hermitian"))  {
      dim(newx) <- rep(max(nr,nc),2)
      newx <- newx+t.spam(newx)-diag.spam(diag(newx))
    }
    if (sym=="skew-symmetric") {
      dim(newx) <- rep(max(nr,nc),2)
      newx <- newx-t.spam(newx)
    }
  }  else {
    nz <- nr*nc
    x <- scan(file, nmax = nz, quiet = TRUE, what=numeric())
    z <- .Fortran("spamdnscsr",
                  nrow = as.integer(nr),
                  ncol = as.integer(nc),
                  x = as.numeric(x), as.integer(nr), entries = vector("double",nz),
                  colindices = vector("integer", nz), rowpointers = vector("integer",nr + 1),
                  eps = options("spam.eps"), NAOK = TRUE,
                  PACKAGE = "spam")

    warning("returning a (possibly) dense 'spam' object", call. = FALSE)
    nz <- z$rowpointers[nr+1]-1
    if (identical(nz, 0L))
      return(new("spam",rowpointers=c(1L,rep.int(2L,nr)), dimension=c(nr,nc)))

    newx <- new("spam")
    slot(newx,"entries",check=FALSE) <- z$entries[1:nz]
    slot(newx,"colindices",check=FALSE) <- z$colindices[1:nz]
    slot(newx,"rowpointers",check=FALSE) <- z$rowpointers
    slot(newx,"dimension",check=FALSE) <- c(nr,nc)
  }
  return(newx)
}

c
c dseupd.f
c dsesrt.f
c dsgets.f
c dsaupd.f
c dsaup2.f
c dsconv.f
c dsaitr.f
c dgetv0.f
c dsapps.f
c dseigt.f
c dstqrb.f
c dsortr.f
c
c
c test for equal zero
c
      function eqZERO( a)
      implicit none

      logical eqZERO
      double precision a

      eqZERO = (abs( a) .LE. 1.1*epsilon( 0.0))
      return
      end function
c
c test not equal to zero
c
      function neZERO( a)
      implicit none

      logical neZERO
      double precision a

      neZERO = (abs( a) .GT. 1.1*epsilon( 0.0))
      return
      end function
c
c
c\BeginDoc
c
c\Name: dseupd
c
c\Description:
c
c  This subroutine returns the converged approximations to eigenvalues
c  of A*z = lambda*B*z and (optionally):
c
c      (1) the corresponding approximate eigenvectors,
c
c      (2) an orthonormal (Lanczos) basis for the associated approximate
c          invariant subspace,
c
c      (3) Both.
c
c  There is negligible additional cost to obtain eigenvectors.  An orthonormal
c  (Lanczos) basis is always computed.  There is an additional storage cost
c  of n*nev if both are requested (in this case a separate array Z must be
c  supplied).
c
c  These quantities are obtained from the Lanczos factorization computed
c  by DSAUPD  for the linear operator OP prescribed by the MODE selection
c  (see IPARAM(7) in DSAUPD  documentation.)  DSAUPD  must be called before
c  this routine is called. These approximate eigenvalues and vectors are
c  commonly called Ritz values and Ritz vectors respectively.  They are
c  referred to as such in the comments that follow.   The computed orthonormal
c  basis for the invariant subspace corresponding to these Ritz values is
c  referred to as a Lanczos basis.
c
c  See documentation in the header of the subroutine DSAUPD  for a definition
c  of OP as well as other terms and the relation of computed Ritz values
c  and vectors of OP with respect to the given problem  A*z = lambda*B*z.
c
c  The approximate eigenvalues of the original problem are returned in
c  ascending algebraic order.  The user may elect to call this routine
c  once for each desired Ritz vector and store it peripherally if desired.
c  There is also the option of computing a selected set of these vectors
c  with a single call.
c
c\Usage:
c  call dseupd
c     ( RVEC, HOWMNY, SELECT, D, Z, LDZ, SIGMA, BMAT, N, WHICH, NEV, TOL,
c       RESID, NCV, V, LDV, IPARAM, IPNTR, WORKD, WORKL, LWORKL, INFO )
c
c  RVEC    LOGICAL  (INPUT)
c          Specifies whether Ritz vectors corresponding to the Ritz value
c          approximations to the eigenproblem A*z = lambda*B*z are computed.
c
c             RVEC = .FALSE.     Compute Ritz values only.
c
c             RVEC = .TRUE.      Compute Ritz vectors.
c
c  HOWMNY  Character*1  (INPUT)
c          Specifies how many Ritz vectors are wanted and the form of Z
c          the matrix of Ritz vectors. See remark 1 below.
c          = 'A': compute NEV Ritz vectors;
c          = 'S': compute some of the Ritz vectors, specified
c                 by the logical array SELECT.
c
c  SELECT  Logical array of dimension NCV.  (INPUT/WORKSPACE)
c          If HOWMNY = 'S', SELECT specifies the Ritz vectors to be
c          computed. To select the Ritz vector corresponding to a
c          Ritz value D(j), SELECT(j) must be set to .TRUE..
c          If HOWMNY = 'A' , SELECT is used as a workspace for
c          reordering the Ritz values.
c
c  D       Double precision  array of dimension NEV.  (OUTPUT)
c          On exit, D contains the Ritz value approximations to the
c          eigenvalues of A*z = lambda*B*z. The values are returned
c          in ascending order. If IPARAM(7) = 3,4,5 then D represents
c          the Ritz values of OP computed by dsaupd  transformed to
c          those of the original eigensystem A*z = lambda*B*z. If
c          IPARAM(7) = 1,2 then the Ritz values of OP are the same
c          as the those of A*z = lambda*B*z.
c
c  Z       Double precision  N by NEV array if HOWMNY = 'A'.  (OUTPUT)
c          On exit, Z contains the B-orthonormal Ritz vectors of the
c          eigensystem A*z = lambda*B*z corresponding to the Ritz
c          value approximations.
c          If  RVEC = .FALSE. then Z is not referenced.
c          NOTE: The array Z may be set equal to first NEV columns of the
c          Arnoldi/Lanczos basis array V computed by DSAUPD .
c
c  LDZ     Integer.  (INPUT)
c          The leading dimension of the array Z.  If Ritz vectors are
c          desired, then  LDZ .ge.  max( 1, N ).  In any case,  LDZ .ge. 1.
c
c  SIGMA   Double precision   (INPUT)
c          If IPARAM(7) = 3,4,5 represents the shift. Not referenced if
c          IPARAM(7) = 1 or 2.
c
c
c  **** The remaining arguments MUST be the same as for the   ****
c  **** call to DSAUPD  that was just completed.               ****
c
c  NOTE: The remaining arguments
c
c           BMAT, N, WHICH, NEV, TOL, RESID, NCV, V, LDV, IPARAM, IPNTR,
c           WORKD, WORKL, LWORKL, INFO
c
c         must be passed directly to DSEUPD  following the last call
c         to DSAUPD .  These arguments MUST NOT BE MODIFIED between
c         the the last call to DSAUPD  and the call to DSEUPD .
c
c  Two of these parameters (WORKL, INFO) are also output parameters:
c
c  WORKL   Double precision  work array of length LWORKL.  (OUTPUT/WORKSPACE)
c          WORKL(1:4*ncv) contains information obtained in
c          dsaupd .  They are not changed by dseupd .
c          WORKL(4*ncv+1:ncv*ncv+8*ncv) holds the
c          untransformed Ritz values, the computed error estimates,
c          and the associated eigenvector matrix of H.
c
c          Note: IPNTR(8:10) contains the pointer into WORKL for addresses
c          of the above information computed by dseupd .
c          -------------------------------------------------------------
c          IPNTR(8): pointer to the NCV RITZ values of the original system.
c          IPNTR(9): pointer to the NCV corresponding error bounds.
c          IPNTR(10): pointer to the NCV by NCV matrix of eigenvectors
c                     of the tridiagonal matrix T. Only referenced by
c                     dseupd  if RVEC = .TRUE. See Remarks.
c          -------------------------------------------------------------
c
c  INFO    Integer.  (OUTPUT)
c          Error flag on output.
c          =  0: Normal exit.
c          = -1: N must be positive.
c          = -2: NEV must be positive.
c          = -3: NCV must be greater than NEV and less than or equal to N.
c          = -5: WHICH must be one of 'LM', 'SM', 'LA', 'SA' or 'BE'.
c          = -6: BMAT must be one of 'I' or 'G'.
c          = -7: Length of private work WORKL array is not sufficient.
c          = -8: Error return from trid. eigenvalue calculation;
c                Information error from LAPACK routine dsteqr .
c          = -9: Starting vector is zero.
c          = -10: IPARAM(7) must be 1,2,3,4,5.
c          = -11: IPARAM(7) = 1 and BMAT = 'G' are incompatible.
c          = -12: NEV and WHICH = 'BE' are incompatible.
c          = -14: DSAUPD  did not find any eigenvalues to sufficient
c                 accuracy.
c          = -15: HOWMNY must be one of 'A' or 'S' if RVEC = .true.
c          = -16: HOWMNY = 'S' not yet implemented
c          = -17: DSEUPD  got a different count of the number of converged
c                 Ritz values than DSAUPD  got.  This indicates the user
c                 probably made an error in passing data from DSAUPD  to
c                 DSEUPD  or that the data was modified before entering
c                 DSEUPD .
c
c\BeginLib
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c  3. B.N. Parlett, "The Symmetric Eigenvalue Problem". Prentice-Hall,
c     1980.
c  4. B.N. Parlett, B. Nour-Omid, "Towards a Black Box Lanczos Program",
c     Computer Physics Communications, 53 (1989), pp 169-179.
c  5. B. Nour-Omid, B.N. Parlett, T. Ericson, P.S. Jensen, "How to
c     Implement the Spectral Transformation", Math. Comp., 48 (1987),
c     pp 663-673.
c  6. R.G. Grimes, J.G. Lewis and H.D. Simon, "A Shifted Block Lanczos
c     Algorithm for Solving Sparse Symmetric Generalized Eigenproblems",
c     SIAM J. Matr. Anal. Apps.,  January (1993).
c  7. L. Reichel, W.B. Gragg, "Algorithm 686: FORTRAN Subroutines
c     for Updating the QR decomposition", ACM TOMS, December 1990,
c     Volume 16 Number 4, pp 369-377.
c
c\Remarks
c  1. The converged Ritz values are always returned in increasing
c     (algebraic) order.
c
c  2. Currently only HOWMNY = 'A' is implemented. It is included at this
c     stage for the user who wants to incorporate it.
c
c\Routines called:
c     dsesrt   ARPACK routine that sorts an array X, and applies the
c             corresponding permutation to a matrix A.
c     dsortr   dsortr   ARPACK sorting routine.
c     ivout   ARPACK utility routine that prints integers.
c     dvout    ARPACK utility routine that prints vectors.
c     dgeqr2   LAPACK routine that computes the QR factorization of
c             a matrix.
c     dlacpy   LAPACK matrix copy routine.
c     dlamch   LAPACK routine that determines machine constants.
c     dorm2r   LAPACK routine that applies an orthogonal matrix in
c             factored form.
c     dsteqr   LAPACK routine that computes eigenvalues and eigenvectors
c             of a tridiagonal matrix.
c     dger     Level 2 BLAS rank one update to a matrix.
c     dcopy    Level 1 BLAS that copies one vector to another .
c     dnrm2    Level 1 BLAS that computes the norm of a vector.
c     dscal    Level 1 BLAS that scales a vector.
c     dswap    Level 1 BLAS that swaps the contents of two vectors.

c\Authors
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Chao Yang                    Houston, Texas
c     Dept. of Computational &
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/15/93: Version ' 2.1'
c
c\SCCS Information: @(#)
c FILE: seupd.F   SID: 2.11   DATE OF SID: 04/10/01   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
      subroutine dseupd (rvec  , howmny, select, d    ,
     &                   z     , ldz   , sigma , bmat ,
     &                   n     , which , nev   , tol  ,
     &                   resid , ncv   , v     , ldv  ,
     &                   iparam, ipntr , workd , workl,
     &                   lworkl, info )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat, howmny, which*2
      logical    rvec
      integer    info, ldz, ldv, lworkl, n, ncv, nev
      Double precision
     &           sigma, tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    iparam(7), ipntr(11)
      logical    select(ncv)
      Double precision
     &           d(nev)     , resid(n)  , v(ldv,ncv),
     &           z(ldz, nev), workd(2*n), workl(lworkl)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0 , zero = 0.0D+0 )
      Integer   ione
      parameter (ione = 1)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      character  type*6
      integer    bounds , ierr   , ih    , ihb   , ihd   ,
     &           iq     , iw     , j     , k     , ldh   ,
     &           ldq    , mode   , nconv , next  ,
     &           ritz   , irz    , ibd   , np    , ishift,
     &           leftptr, rghtptr, numcnv, jj
      Double precision
     &           bnorm2 , rnorm, temp, temp1, eps23
      logical    reord, dseupdtrue
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy  , dger   , dgeqr2 , dlacpy , dorm2r , dscal ,
     &           dsesrt , dsteqr , dswap  , dsortr
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dnrm2 , dlamch
      external   dnrm2 , dlamch
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    min
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %------------------------%
c     | Set default parameters |
c     %------------------------%
c
      mode = iparam(7)
      nconv = iparam(5)
      info = 0
      dseupdtrue = .true.
c
c     %--------------%
c     | Quick return |
c     %--------------%
c
      if (nconv .eq. 0) go to 9000
      ierr = 0
c
      if (nconv .le. 0)                        ierr = -14
      if (n .le. 0)                            ierr = -1
      if (nev .le. 0)                          ierr = -2
      if (ncv .le. nev .or.  ncv .gt. n)       ierr = -3
      if (which .ne. 'LM' .and.
     &    which .ne. 'SM' .and.
     &    which .ne. 'LA' .and.
     &    which .ne. 'SA' .and.
     &    which .ne. 'BE')                     ierr = -5
      if (bmat .ne. 'I' .and. bmat .ne. 'G')   ierr = -6
      if ( (howmny .ne. 'A' .and.
     &           howmny .ne. 'P' .and.
     &           howmny .ne. 'S') .and. rvec )
     &                                         ierr = -15
      if (rvec .and. howmny .eq. 'S')          ierr = -16
c
      if (rvec .and. lworkl .lt. ncv**2+8*ncv) ierr = -7
c
      if (mode .eq. 1 .or. mode .eq. 2) then
         type = 'REGULR'
      else if (mode .eq. 3 ) then
         type = 'SHIFTI'
      else if (mode .eq. 4 ) then
         type = 'BUCKLE'
      else if (mode .eq. 5 ) then
         type = 'CAYLEY'
      else
                                               ierr = -10
      end if
      if (mode .eq. 1 .and. bmat .eq. 'G')     ierr = -11
      if (nev .eq. 1 .and. which .eq. 'BE')    ierr = -12
c
c     %------------%
c     | Error Exit |
c     %------------%
c
      if (ierr .ne. 0) then
         info = ierr
         go to 9000
      end if
c
c     %-------------------------------------------------------%
c     | Pointer into WORKL for address of H, RITZ, BOUNDS, Q  |
c     | etc... and the remaining workspace.                   |
c     | Also update pointer to be used on output.             |
c     | Memory is laid out as follows:                        |
c     | workl(1:2*ncv) := generated tridiagonal matrix H      |
c     |       The subdiagonal is stored in workl(2:ncv).      |
c     |       The dead spot is workl(1) but upon exiting      |
c     |       dsaupd  stores the B-norm of the last residual   |
c     |       vector in workl(1). We use this !!!             |
c     | workl(2*ncv+1:2*ncv+ncv) := ritz values               |
c     |       The wanted values are in the first NCONV spots. |
c     | workl(3*ncv+1:3*ncv+ncv) := computed Ritz estimates   |
c     |       The wanted values are in the first NCONV spots. |
c     | NOTE: workl(1:4*ncv) is set by dsaupd  and is not      |
c     |       modified by dseupd .                             |
c     %-------------------------------------------------------%
c
c     %-------------------------------------------------------%
c     | The following is used and set by dseupd .              |
c     | workl(4*ncv+1:4*ncv+ncv) := used as workspace during  |
c     |       computation of the eigenvectors of H. Stores    |
c     |       the diagonal of H. Upon EXIT contains the NCV   |
c     |       Ritz values of the original system. The first   |
c     |       NCONV spots have the wanted values. If MODE =   |
c     |       1 or 2 then will equal workl(2*ncv+1:3*ncv).    |
c     | workl(5*ncv+1:5*ncv+ncv) := used as workspace during  |
c     |       computation of the eigenvectors of H. Stores    |
c     |       the subdiagonal of H. Upon EXIT contains the    |
c     |       NCV corresponding Ritz estimates of the         |
c     |       original system. The first NCONV spots have the |
c     |       wanted values. If MODE = 1,2 then will equal    |
c     |       workl(3*ncv+1:4*ncv).                           |
c     | workl(6*ncv+1:6*ncv+ncv*ncv) := orthogonal Q that is  |
c     |       the eigenvector matrix for H as returned by     |
c     |       dsteqr . Not referenced if RVEC = .False.        |
c     |       Ordering follows that of workl(4*ncv+1:5*ncv)   |
c     | workl(6*ncv+ncv*ncv+1:6*ncv+ncv*ncv+2*ncv) :=         |
c     |       Workspace. Needed by dsteqr  and by dseupd .      |
c     | GRAND total of NCV*(NCV+8) locations.                 |
c     %-------------------------------------------------------%
c
c
      ih     = ipntr(5)
      ritz   = ipntr(6)
      bounds = ipntr(7)
      ldh    = ncv
      ldq    = ncv
      ihd    = bounds + ldh
      ihb    = ihd    + ldh
      iq     = ihb    + ldh
      iw     = iq     + ldh*ncv
      next   = iw     + 2*ncv
      ipntr(4)  = next
      ipntr(8)  = ihd
      ipntr(9)  = ihb
      ipntr(10) = iq
c
c     %----------------------------------------%
c     | irz points to the Ritz values computed |
c     |     by _seigt before exiting _saup2.   |
c     | ibd points to the Ritz estimates       |
c     |     computed by _seigt before exiting  |
c     |     _saup2.                            |
c     %----------------------------------------%
c
      irz = ipntr(11)+ncv
      ibd = irz+ncv
c
c
c     %---------------------------------%
c     | Set machine dependent constant. |
c     %---------------------------------%
c
      eps23 = dlamch ('Epsilon-Machine')
      eps23 = eps23**(2.0D+0  / 3.0D+0 )
c
c     %---------------------------------------%
c     | RNORM is B-norm of the RESID(1:N).    |
c     | BNORM2 is the 2 norm of B*RESID(1:N). |
c     | Upon exit of dsaupd  WORKD(1:N) has    |
c     | B*RESID(1:N).                         |
c     %---------------------------------------%
c
      rnorm = workl(ih)
      if (bmat .eq. 'I') then
         bnorm2 = rnorm
      else if (bmat .eq. 'G') then
         bnorm2 = dnrm2 (n, workd, 1)
      end if
c
c
      if (rvec) then
c
         reord = .false.
c
c        %---------------------------------------------------%
c        | Use the temporary bounds array to store indices   |
c        | These will be used to mark the select array later |
c        %---------------------------------------------------%
c
         do 10 j = 1,ncv
            workl(bounds+j-1) = j
            select(j) = .false.
   10    continue
c
c        %-------------------------------------%
c        | Select the wanted Ritz values.      |
c        | Sort the Ritz values so that the    |
c        | wanted ones appear at the tailing   |
c        | NEV positions of workl(irr) and     |
c        | workl(iri).  Move the corresponding |
c        | error estimates in workl(bound)     |
c        | accordingly.                        |
c        %-------------------------------------%
c
         np     = ncv - nev
         ishift = 0
         call dsgets (ishift, which       , nev          ,
     &                np    , workl(irz)  , workl(bounds),
     &                workl)
c
c
c        %-----------------------------------------------------%
c        | Record indices of the converged wanted Ritz values  |
c        | Mark the select array for possible reordering       |
c        %-----------------------------------------------------%
c
         numcnv = 0
         do 11 j = 1,ncv
            temp1 = max(eps23, abs(workl(irz+ncv-j)) )
            jj = INT(workl(bounds + ncv - j))
            if (numcnv .lt. nconv .and.
     &          workl(ibd+jj-1) .le. tol*temp1) then
               select(jj) = .true.
               numcnv = numcnv + 1
               if (jj .gt. nev) reord = .true.
            endif
   11    continue
c
c        %-----------------------------------------------------------%
c        | Check the count (numcnv) of converged Ritz values with    |
c        | the number (nconv) reported by _saupd.  If these two      |
c        | are different then there has probably been an error       |
c        | caused by incorrect passing of the _saupd data.           |
c        %-----------------------------------------------------------%
c
c
         if (numcnv .ne. nconv) then
            info = -17
            go to 9000
         end if
c
c        %-----------------------------------------------------------%
c        | Call LAPACK routine _steqr to compute the eigenvalues and |
c        | eigenvectors of the final symmetric tridiagonal matrix H. |
c        | Initialize the eigenvector matrix Q to the identity.      |
c        %-----------------------------------------------------------%
c
         call dcopy (ncv-1, workl(ih+1), 1, workl(ihb), 1)
         call dcopy (ncv, workl(ih+ldh), 1, workl(ihd), 1)
c
         call dsteqr ('Identity', ncv, workl(ihd), workl(ihb),
     &                workl(iq) , ldq, workl(iw), ierr)
c
         if (ierr .ne. 0) then
            info = -8
            go to 9000
         end if
c
         if (reord) then
c
c           %---------------------------------------------%
c           | Reordered the eigenvalues and eigenvectors  |
c           | computed by _steqr so that the "converged"  |
c           | eigenvalues appear in the first NCONV       |
c           | positions of workl(ihd), and the associated |
c           | eigenvectors appear in the first NCONV      |
c           | columns.                                    |
c           %---------------------------------------------%
c
            leftptr = 1
            rghtptr = ncv
c
            if (ncv .eq. 1) go to 30
c
 20         if (select(leftptr)) then
c
c              %-------------------------------------------%
c              | Search, from the left, for the first Ritz |
c              | value that has not converged.             |
c              %-------------------------------------------%
c
               leftptr = leftptr + 1
c
            else if ( .not. select(rghtptr)) then
c
c              %----------------------------------------------%
c              | Search, from the right, the first Ritz value |
c              | that has converged.                          |
c              %----------------------------------------------%
c
               rghtptr = rghtptr - 1
c
            else
c
c              %----------------------------------------------%
c              | Swap the Ritz value on the left that has not |
c              | converged with the Ritz value on the right   |
c              | that has converged.  Swap the associated     |
c              | eigenvector of the tridiagonal matrix H as   |
c              | well.                                        |
c              %----------------------------------------------%
c
               temp = workl(ihd+leftptr-1)
               workl(ihd+leftptr-1) = workl(ihd+rghtptr-1)
               workl(ihd+rghtptr-1) = temp
               call dcopy (ncv, workl(iq+ncv*(leftptr-1)), 1,
     &                    workl(iw), 1)
               call dcopy (ncv, workl(iq+ncv*(rghtptr-1)), 1,
     &                    workl(iq+ncv*(leftptr-1)), 1)
               call dcopy (ncv, workl(iw), 1,
     &                    workl(iq+ncv*(rghtptr-1)), 1)
               leftptr = leftptr + 1
               rghtptr = rghtptr - 1
c
            end if
c
            if (leftptr .lt. rghtptr) go to 20
c
 30      end if
c
c
c        %----------------------------------------%
c        | Load the converged Ritz values into D. |
c        %----------------------------------------%
c
         call dcopy (nconv, workl(ihd), 1, d(1), 1)
c
      else
c
c        %-----------------------------------------------------%
c        | Ritz vectors not required. Load Ritz values into D. |
c        %-----------------------------------------------------%
c
         call dcopy (nconv, workl(ritz), 1, d(1), 1)
         call dcopy (ncv, workl(ritz), 1, workl(ihd), 1)
c
      end if
c
c     %------------------------------------------------------------------%
c     | Transform the Ritz values and possibly vectors and corresponding |
c     | Ritz estimates of OP to those of A*x=lambda*B*x. The Ritz values |
c     | (and corresponding data) are returned in ascending order.        |
c     %------------------------------------------------------------------%
c
      if (type .eq. 'REGULR') then
c
c        %---------------------------------------------------------%
c        | Ascending sort of wanted Ritz values, vectors and error |
c        | bounds. Not necessary if only Ritz values are desired.  |
c        %---------------------------------------------------------%
c
         if (rvec) then
            call dsesrt ('LA', rvec , nconv, d, ncv, workl(iq), ldq)
         else
            call dcopy (ncv, workl(bounds), 1, workl(ihb), 1)
         end if
c
      else
c
c        %-------------------------------------------------------------%
c        | *  Make a copy of all the Ritz values.                      |
c        | *  Transform the Ritz values back to the original system.   |
c        |    For TYPE = 'SHIFTI' the transformation is                |
c        |             lambda = 1/theta + sigma                        |
c        |    For TYPE = 'BUCKLE' the transformation is                |
c        |             lambda = sigma * theta / ( theta - 1 )          |
c        |    For TYPE = 'CAYLEY' the transformation is                |
c        |             lambda = sigma * (theta + 1) / (theta - 1 )     |
c        |    where the theta are the Ritz values returned by dsaupd .  |
c        | NOTES:                                                      |
c        | *The Ritz vectors are not affected by the transformation.   |
c        |  They are only reordered.                                   |
c        %-------------------------------------------------------------%
c
         call dcopy  (ncv, workl(ihd), 1, workl(iw), 1)
         if (type .eq. 'SHIFTI') then
            do 40 k=1, ncv
               workl(ihd+k-1) = one / workl(ihd+k-1) + sigma
  40        continue
         else if (type .eq. 'BUCKLE') then
            do 50 k=1, ncv
               workl(ihd+k-1) = sigma * workl(ihd+k-1) /
     &                          (workl(ihd+k-1) - one)
  50        continue
         else if (type .eq. 'CAYLEY') then
            do 60 k=1, ncv
               workl(ihd+k-1) = sigma * (workl(ihd+k-1) + one) /
     &                          (workl(ihd+k-1) - one)
  60        continue
         end if
c
c        %-------------------------------------------------------------%
c        | *  Store the wanted NCONV lambda values into D.             |
c        | *  Sort the NCONV wanted lambda in WORKL(IHD:IHD+NCONV-1)   |
c        |    into ascending order and apply sort to the NCONV theta   |
c        |    values in the transformed system. We will need this to   |
c        |    compute Ritz estimates in the original system.           |
c        | *  Finally sort the lambda`s into ascending order and apply |
c        |    to Ritz vectors if wanted. Else just sort lambda`s into  |
c        |    ascending order.                                         |
c        | NOTES:                                                      |
c        | *workl(iw:iw+ncv-1) contain the theta ordered so that they  |
c        |  match the ordering of the lambda. We`ll use them again for |
c        |  Ritz vector purification.                                  |
c        %-------------------------------------------------------------%
c
         call dcopy (nconv, workl(ihd), 1, d(1), 1)
         call dsortr ('LA', dseupdtrue, nconv, workl(ihd), workl(iw))
         if (rvec) then
            call dsesrt ('LA', rvec , nconv, d, ncv, workl(iq), ldq)
         else
            call dcopy (ncv, workl(bounds), 1, workl(ihb), 1)
            call dscal (ncv, bnorm2/rnorm, workl(ihb), 1)
            call dsortr ('LA', dseupdtrue, nconv, d, workl(ihb))
         end if
c
      end if
c
c     %------------------------------------------------%
c     | Compute the Ritz vectors. Transform the wanted |
c     | eigenvectors of the symmetric tridiagonal H by |
c     | the Lanczos basis matrix V.                    |
c     %------------------------------------------------%
c
      if (rvec .and. howmny .eq. 'A') then
c
c        %----------------------------------------------------------%
c        | Compute the QR factorization of the matrix representing  |
c        | the wanted invariant subspace located in the first NCONV |
c        | columns of workl(iq,ldq).                                |
c        %----------------------------------------------------------%
c
         call dgeqr2 (ncv, nconv        , workl(iq) ,
     &                ldq, workl(iw+ncv), workl(ihb),
     &                ierr)
c
c        %--------------------------------------------------------%
c        | * Postmultiply V by Q.                                 |
c        | * Copy the first NCONV columns of VQ into Z.           |
c        | The N by NCONV matrix Z is now a matrix representation |
c        | of the approximate invariant subspace associated with  |
c        | the Ritz values in workl(ihd).                         |
c        %--------------------------------------------------------%
c
         call dorm2r ('Right', 'Notranspose', n        ,
     &                ncv    , nconv        , workl(iq),
     &                ldq    , workl(iw+ncv), v        ,
     &                ldv    , workd(n+1)   , ierr)
         call dlacpy ('All', n, nconv, v(1,1), ldv, z, ldz)
c
c        %-----------------------------------------------------%
c        | In order to compute the Ritz estimates for the Ritz |
c        | values in both systems, need the last row of the    |
c        | eigenvector matrix. Remember, it`s in factored form |
c        %-----------------------------------------------------%
c
         do 65 j = 1, ncv-1
            workl(ihb+j-1) = zero
  65     continue
         workl(ihb+ncv-1) = one
         call dorm2r ('Left', 'Transpose'  , ncv       ,
     &                ione  , nconv        , workl(iq) ,
     &                ldq   , workl(iw+ncv), workl(ihb),
     &                ncv   , temp         , ierr)
c
      else if (rvec .and. howmny .eq. 'S') then
c
c     Not yet implemented. See remark 2 above.
c
      end if
c
      if (type .eq. 'REGULR' .and. rvec) then
c
            do 70 j=1, ncv
               workl(ihb+j-1) = rnorm * abs( workl(ihb+j-1) )
 70         continue
c
      else if (type .ne. 'REGULR' .and. rvec) then
c
c        %-------------------------------------------------%
c        | *  Determine Ritz estimates of the theta.       |
c        |    If RVEC = .true. then compute Ritz estimates |
c        |               of the theta.                     |
c        |    If RVEC = .false. then copy Ritz estimates   |
c        |              as computed by dsaupd .             |
c        | *  Determine Ritz estimates of the lambda.      |
c        %-------------------------------------------------%
c
         call dscal  (ncv, bnorm2, workl(ihb), 1)
         if (type .eq. 'SHIFTI') then
c
            do 80 k=1, ncv
               workl(ihb+k-1) = abs( workl(ihb+k-1) )
     &                        / workl(iw+k-1)**2
 80         continue
c
         else if (type .eq. 'BUCKLE') then
c
            do 90 k=1, ncv
               workl(ihb+k-1) = sigma * abs( workl(ihb+k-1) )
     &                        / (workl(iw+k-1)-one )**2
 90         continue
c
         else if (type .eq. 'CAYLEY') then
c
            do 100 k=1, ncv
               workl(ihb+k-1) = abs( workl(ihb+k-1)
     &                        / workl(iw+k-1)*(workl(iw+k-1)-one) )
 100        continue
c
         end if
c
      end if
c
c
c     %-------------------------------------------------%
c     | Ritz vector purification step. Formally perform |
c     | one of inverse subspace iteration. Only used    |
c     | for MODE = 3,4,5. See reference 7               |
c     %-------------------------------------------------%
c
      if (rvec .and. (type .eq. 'SHIFTI' .or. type .eq. 'CAYLEY')) then
c
         do 110 k=0, nconv-1
            workl(iw+k) = workl(iq+k*ldq+ncv-1)
     &                  / workl(iw+k)
 110     continue
c
      else if (rvec .and. type .eq. 'BUCKLE') then
c
         do 120 k=0, nconv-1
            workl(iw+k) = workl(iq+k*ldq+ncv-1)
     &                  / (workl(iw+k)-one)
 120     continue
c
      end if
c
      if (type .ne. 'REGULR')
     &   call dger  (n, nconv, one, resid, 1, workl(iw), 1, z, ldz)
c
 9000 continue
c
      return
c
c     %---------------%
c     | End of dseupd |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsesrt
c
c\Description:
c  Sort the array X in the order specified by WHICH and optionally
c  apply the permutation to the columns of the matrix A.
c
c\Usage:
c  call dsesrt
c     ( WHICH, APPLY, N, X, NA, A, LDA)
c
c\Arguments
c  WHICH   Character*2.  (Input)
c          'LM' -> X is sorted into increasing order of magnitude.
c          'SM' -> X is sorted into decreasing order of magnitude.
c          'LA' -> X is sorted into increasing order of algebraic.
c          'SA' -> X is sorted into decreasing order of algebraic.
c
c  APPLY   Logical.  (Input)
c          APPLY = .TRUE.  -> apply the sorted order to A.
c          APPLY = .FALSE. -> do not apply the sorted order to A.
c
c  N       Integer.  (INPUT)
c          Dimension of the array X.
c
c  X      Double precision array of length N.  (INPUT/OUTPUT)
c          The array to be sorted.
c
c  NA      Integer.  (INPUT)
c          Number of rows of the matrix A.
c
c  A      Double precision array of length NA by N.  (INPUT/OUTPUT)
c
c  LDA     Integer.  (INPUT)
c          Leading dimension of A.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Routines
c     dswap  Level 1 BLAS that swaps the contents of two vectors.
c
c\Authors
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/15/93: Version ' 2.1'.
c               Adapted from the sort routine in LANSO and
c               the ARPACK code dsortr
c
c\SCCS Information: @(#)
c FILE: sesrt.F   SID: 2.3   DATE OF SID: 4/19/96   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsesrt (which, apply, n, x, na, a, lda)
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  which*2
      logical    apply
      integer    lda, n, na
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           x(0:n-1), a(lda, 0:n-1)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i, igap, j
      Double precision
     &           temp
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dswap
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      igap = n / 2
c
      if (which .eq. 'SA') then
c
c        X is sorted into decreasing order of algebraic.
c
   10    continue
         if (igap .eq. 0) go to 9000
         do 30 i = igap, n-1
            j = i-igap
   20       continue
c
            if (j.lt.0) go to 30
c
            if (x(j).lt.x(j+igap)) then
               temp = x(j)
               x(j) = x(j+igap)
               x(j+igap) = temp
               if (apply) call dswap( na, a(1, j), 1, a(1,j+igap), 1)
            else
               go to 30
            endif
            j = j-igap
            go to 20
   30    continue
         igap = igap / 2
         go to 10
c
      else if (which .eq. 'SM') then
c
c        X is sorted into decreasing order of magnitude.
c
   40    continue
         if (igap .eq. 0) go to 9000
         do 60 i = igap, n-1
            j = i-igap
   50       continue
c
            if (j.lt.0) go to 60
c
            if (abs(x(j)).lt.abs(x(j+igap))) then
               temp = x(j)
               x(j) = x(j+igap)
               x(j+igap) = temp
               if (apply) call dswap( na, a(1, j), 1, a(1,j+igap), 1)
            else
               go to 60
            endif
            j = j-igap
            go to 50
   60    continue
         igap = igap / 2
         go to 40
c
      else if (which .eq. 'LA') then
c
c        X is sorted into increasing order of algebraic.
c
   70    continue
         if (igap .eq. 0) go to 9000
         do 90 i = igap, n-1
            j = i-igap
   80       continue
c
            if (j.lt.0) go to 90
c
            if (x(j).gt.x(j+igap)) then
               temp = x(j)
               x(j) = x(j+igap)
               x(j+igap) = temp
               if (apply) call dswap( na, a(1, j), 1, a(1,j+igap), 1)
            else
               go to 90
            endif
            j = j-igap
            go to 80
   90    continue
         igap = igap / 2
         go to 70
c
      else if (which .eq. 'LM') then
c
c        X is sorted into increasing order of magnitude.
c
  100    continue
         if (igap .eq. 0) go to 9000
         do 120 i = igap, n-1
            j = i-igap
  110       continue
c
            if (j.lt.0) go to 120
c
            if (abs(x(j)).gt.abs(x(j+igap))) then
               temp = x(j)
               x(j) = x(j+igap)
               x(j+igap) = temp
               if (apply) call dswap( na, a(1, j), 1, a(1,j+igap), 1)
            else
               go to 120
            endif
            j = j-igap
            go to 110
  120    continue
         igap = igap / 2
         go to 100
      end if
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsesrt |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsgets
c
c\Description:
c  Given the eigenvalues of the symmetric tridiagonal matrix H,
c  computes the NP shifts AMU that are zeros of the polynomial of
c  degree NP which filters out components of the unwanted eigenvectors
c  corresponding to the AMU's based on some given criteria.
c
c  NOTE: This is called even in the case of user specified shifts in
c  order to sort the eigenvalues, and error bounds of H for later use.
c
c\Usage:
c  call dsgets
c     ( ISHIFT, WHICH, KEV, NP, RITZ, BOUNDS, SHIFTS )
c
c\Arguments
c  ISHIFT  Integer.  (INPUT)
c          Method for selecting the implicit shifts at each iteration.
c          ISHIFT = 0: user specified shifts
c          ISHIFT = 1: exact shift with respect to the matrix H.
c
c  WHICH   Character*2.  (INPUT)
c          Shift selection criteria.
c          'LM' -> KEV eigenvalues of largest magnitude are retained.
c          'SM' -> KEV eigenvalues of smallest magnitude are retained.
c          'LA' -> KEV eigenvalues of largest value are retained.
c          'SA' -> KEV eigenvalues of smallest value are retained.
c          'BE' -> KEV eigenvalues, half from each end of the spectrum.
c                  If KEV is odd, compute one more from the high end.
c
c  KEV      Integer.  (INPUT)
c          KEV+NP is the size of the matrix H.
c
c  NP      Integer.  (INPUT)
c          Number of implicit shifts to be computed.
c
c  RITZ    Double precision array of length KEV+NP.  (INPUT/OUTPUT)
c          On INPUT, RITZ contains the eigenvalues of H.
c          On OUTPUT, RITZ are sorted so that the unwanted eigenvalues
c          are in the first NP locations and the wanted part is in
c          the last KEV locations.  When exact shifts are selected, the
c          unwanted part corresponds to the shifts to be applied.
c
c  BOUNDS  Double precision array of length KEV+NP.  (INPUT/OUTPUT)
c          Error bounds corresponding to the ordering in RITZ.
c
c  SHIFTS  Double precision array of length NP.  (INPUT/OUTPUT)
c          On INPUT:  contains the user specified shifts if ISHIFT = 0.
c          On OUTPUT: contains the shifts sorted into decreasing order
c          of magnitude with respect to the Ritz estimates contained in
c          BOUNDS. If ISHIFT = 0, SHIFTS is not modified on exit.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dsortr  ARPACK utility sorting routine.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dvout   ARPACK utility routine that prints vectors.
c     dcopy   Level 1 BLAS that copies one vector to another.
c     dswap   Level 1 BLAS that swaps the contents of two vectors.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/93: Version ' 2.1'
c
c\SCCS Information: @(#)
c FILE: sgets.F   SID: 2.4   DATE OF SID: 4/19/96   RELEASE: 2
c
c\Remarks
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsgets ( ishift, which, kev, np, ritz, bounds, shifts )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  which*2
      integer    ishift, kev, np
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           bounds(kev+np), ritz(kev+np), shifts(np)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    kevd2
      logical    dsgetstrue
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dswap, dcopy, dsortr
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    max, min
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
      dsgetstrue = .true.
c
      if (which .eq. 'BE') then
c
c        %-----------------------------------------------------%
c        | Both ends of the spectrum are requested.            |
c        | Sort the eigenvalues into algebraically increasing  |
c        | order first then swap high end of the spectrum next |
c        | to low end in appropriate locations.                |
c        | NOTE: when np < floor(kev/2) be careful not to swap |
c        | overlapping locations.                              |
c        %-----------------------------------------------------%
c
         call dsortr ('LA', dsgetstrue, kev+np, ritz, bounds)
         kevd2 = kev / 2
         if ( kev .gt. 1 ) then
            call dswap ( min(kevd2,np), ritz(1), 1,
     &                   ritz( max(kevd2,np)+1 ), 1)
            call dswap ( min(kevd2,np), bounds(1), 1,
     &                   bounds( max(kevd2,np)+1 ), 1)
         end if
c
      else
c
c        %----------------------------------------------------%
c        | LM, SM, LA, SA case.                               |
c        | Sort the eigenvalues of H into the desired order   |
c        | and apply the resulting order to BOUNDS.           |
c        | The eigenvalues are sorted so that the wanted part |
c        | are always in the last KEV locations.               |
c        %----------------------------------------------------%
c
         call dsortr (which, dsgetstrue, kev+np, ritz, bounds)
      end if
c
      if (ishift .eq. 1 .and. np .gt. 0) then
c
c        %-------------------------------------------------------%
c        | Sort the unwanted Ritz values used as shifts so that  |
c        | the ones with largest Ritz estimates are first.       |
c        | This will tend to minimize the effects of the         |
c        | forward instability of the iteration when the shifts  |
c        | are applied in subroutine dsapps.                     |
c        %-------------------------------------------------------%
c
         call dsortr ('SM', dsgetstrue, np, bounds, ritz)
         call dcopy (np, ritz, 1, shifts(1), 1)
      end if
c
      return
c
c     %---------------%
c     | End of dsgets |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsaupd
c
c\Description:
c
c  Reverse communication interface for the Implicitly Restarted Arnoldi
c  Iteration.  For symmetric problems this reduces to a variant of the Lanczos
c  method.  This method has been designed to compute approximations to a
c  few eigenpairs of a linear operator OP that is real and symmetric
c  with respect to a real positive semi-definite symmetric matrix B,
c  i.e.
c
c       B*OP = (OP`)*B.
c
c  Another way to express this condition is
c
c       < x,OPy > = < OPx,y >  where < z,w > = z`Bw  .
c
c  In the standard eigenproblem B is the identity matrix.
c  ( A` denotes transpose of A)
c
c  The computed approximate eigenvalues are called Ritz values and
c  the corresponding approximate eigenvectors are called Ritz vectors.
c
c  dsaupd  is usually called iteratively to solve one of the
c  following problems:
c
c  Mode 1:  A*x = lambda*x, A symmetric
c           ===> OP = A  and  B = I.
c
c  Mode 2:  A*x = lambda*M*x, A symmetric, M symmetric positive definite
c           ===> OP = inv[M]*A  and  B = M.
c           ===> (If M can be factored see remark 3 below)
c
c  Mode 3:  K*x = lambda*M*x, K symmetric, M symmetric positive semi-definite
c           ===> OP = (inv[K - sigma*M])*M  and  B = M.
c           ===> Shift-and-Invert mode
c
c  Mode 4:  K*x = lambda*KG*x, K symmetric positive semi-definite,
c           KG symmetric indefinite
c           ===> OP = (inv[K - sigma*KG])*K  and  B = K.
c           ===> Buckling mode
c
c  Mode 5:  A*x = lambda*M*x, A symmetric, M symmetric positive semi-definite
c           ===> OP = inv[A - sigma*M]*[A + sigma*M]  and  B = M.
c           ===> Cayley transformed mode
c
c  NOTE: The action of w <- inv[A - sigma*M]*v or w <- inv[M]*v
c        should be accomplished either by a direct method
c        using a sparse matrix factorization and solving
c
c           [A - sigma*M]*w = v  or M*w = v,
c
c        or through an iterative method for solving these
c        systems.  If an iterative method is used, the
c        convergence test must be more stringent than
c        the accuracy requirements for the eigenvalue
c        approximations.
c
c\Usage:
c  call dsaupd
c     ( IDO, BMAT, N, WHICH, NEV, TOL, RESID, NCV, V, LDV, IPARAM,
c       IPNTR, WORKD, WORKL, LWORKL, INFO )
c
c\Arguments
c  IDO     Integer.  (INPUT/OUTPUT)
c          Reverse communication flag.  IDO must be zero on the first
c          call to dsaupd .  IDO will be set internally to
c          indicate the type of operation to be performed.  Control is
c          then given back to the calling routine which has the
c          responsibility to carry out the requested operation and call
c          dsaupd  with the result.  The operand is given in
c          WORKD(IPNTR(1)), the result must be put in WORKD(IPNTR(2)).
c          (If Mode = 2 see remark 5 below)
c          -------------------------------------------------------------
c          IDO =  0: first call to the reverse communication interface
c          IDO = -1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c                    This is for the initialization phase to force the
c                    starting vector into the range of OP.
c          IDO =  1: compute  Y = OP * X where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c                    In mode 3,4 and 5, the vector B * X is already
c                    available in WORKD(ipntr(3)).  It does not
c                    need to be recomputed in forming OP * X.
c          IDO =  2: compute  Y = B * X  where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c          IDO =  3: compute the IPARAM(8) shifts where
c                    IPNTR(11) is the pointer into WORKL for
c                    placing the shifts. See remark 6 below.
c          IDO = 99: done
c          -------------------------------------------------------------
c
c  BMAT    Character*1.  (INPUT)
c          BMAT specifies the type of the matrix B that defines the
c          semi-inner product for the operator OP.
c          B = 'I' -> standard eigenvalue problem A*x = lambda*x
c          B = 'G' -> generalized eigenvalue problem A*x = lambda*B*x
c
c  N       Integer.  (INPUT)
c          Dimension of the eigenproblem.
c
c  WHICH   Character*2.  (INPUT)
c          Specify which of the Ritz values of OP to compute.
c
c          'LA' - compute the NEV largest (algebraic) eigenvalues.
c          'SA' - compute the NEV smallest (algebraic) eigenvalues.
c          'LM' - compute the NEV largest (in magnitude) eigenvalues.
c          'SM' - compute the NEV smallest (in magnitude) eigenvalues.
c          'BE' - compute NEV eigenvalues, half from each end of the
c                 spectrum.  When NEV is odd, compute one more from the
c                 high end than from the low end.
c           (see remark 1 below)
c
c  NEV     Integer.  (INPUT)
c          Number of eigenvalues of OP to be computed. 0 < NEV < N.
c
c  TOL     Double precision  scalar.  (INPUT)
c          Stopping criterion: the relative accuracy of the Ritz value
c          is considered acceptable if BOUNDS(I) .LE. TOL*ABS(RITZ(I)).
c          If TOL .LE. 0. is passed a default is set:
c          DEFAULT = DLAMCH ('EPS')  (machine precision as computed
c                    by the LAPACK auxiliary subroutine DLAMCH ).
c
c  RESID   Double precision  array of length N.  (INPUT/OUTPUT)
c          On INPUT:
c          If INFO .EQ. 0, a random initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          On OUTPUT:
c          RESID contains the final residual vector.
c
c  NCV     Integer.  (INPUT)
c          Number of columns of the matrix V (less than or equal to N).
c          This will indicate how many Lanczos vectors are generated
c          at each iteration.  After the startup phase in which NEV
c          Lanczos vectors are generated, the algorithm generates
c          NCV-NEV Lanczos vectors at each subsequent update iteration.
c          Most of the cost in generating each Lanczos vector is in the
c          matrix-vector product OP*x. (See remark 4 below).
c
c  V       Double precision  N by NCV array.  (OUTPUT)
c          The NCV columns of V contain the Lanczos basis vectors.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  IPARAM  Integer array of length 11.  (INPUT/OUTPUT)
c          IPARAM(1) = ISHIFT: method for selecting the implicit shifts.
c          The shifts selected at each iteration are used to restart
c          the Arnoldi iteration in an implicit fashion.
c          -------------------------------------------------------------
c          ISHIFT = 0: the shifts are provided by the user via
c                      reverse communication.  The NCV eigenvalues of
c                      the current tridiagonal matrix T are returned in
c                      the part of WORKL array corresponding to RITZ.
c                      See remark 6 below.
c          ISHIFT = 1: exact shifts with respect to the reduced
c                      tridiagonal matrix T.  This is equivalent to
c                      restarting the iteration with a starting vector
c                      that is a linear combination of Ritz vectors
c                      associated with the "wanted" Ritz values.
c          -------------------------------------------------------------
c
c          IPARAM(2) = LEVEC
c          No longer referenced. See remark 2 below.
c
c          IPARAM(3) = MXITER
c          On INPUT:  maximum number of Arnoldi update iterations allowed.
c          On OUTPUT: actual number of Arnoldi update iterations taken.
c
c          IPARAM(4) = NB: blocksize to be used in the recurrence.
c          The code currently works only for NB = 1.
c
c          IPARAM(5) = NCONV: number of "converged" Ritz values.
c          This represents the number of Ritz values that satisfy
c          the convergence criterion.
c
c          IPARAM(6) = IUPD
c          No longer referenced. Implicit restarting is ALWAYS used.
c
c          IPARAM(7) = MODE
c          On INPUT determines what type of eigenproblem is being solved.
c          Must be 1,2,3,4,5; See under \Description of dsaupd  for the
c          five modes available.
c
c          IPARAM(8) = NP
c          When ido = 3 and the user provides shifts through reverse
c          communication (IPARAM(1)=0), dsaupd  returns NP, the number
c          of shifts the user is to provide. 0 < NP <=NCV-NEV. See Remark
c          6 below.
c
crc          IPARAM(9) = NUMOP, IPARAM(10) = NUMOPB, IPARAM(11) = NUMREO,
crc          OUTPUT: NUMOP  = total number of OP*x operations,
crc                  NUMOPB = total number of B*x operations if BMAT='G',
crc                  NUMREO = total number of steps of re-orthogonalization.c
c
c  IPNTR   Integer array of length 11.  (OUTPUT)
c          Pointer to mark the starting locations in the WORKD and WORKL
c          arrays for matrices/vectors used by the Lanczos iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X in WORKD.
c          IPNTR(2): pointer to the current result vector Y in WORKD.
c          IPNTR(3): pointer to the vector B * X in WORKD when used in
c                    the shift-and-invert mode.
c          IPNTR(4): pointer to the next available location in WORKL
c                    that is untouched by the program.
c          IPNTR(5): pointer to the NCV by 2 tridiagonal matrix T in WORKL.
c          IPNTR(6): pointer to the NCV RITZ values array in WORKL.
c          IPNTR(7): pointer to the Ritz estimates in array WORKL associated
c                    with the Ritz values located in RITZ in WORKL.
c          IPNTR(11): pointer to the NP shifts in WORKL. See Remark 6 below.
c
c          Note: IPNTR(8:10) is only referenced by dseupd . See Remark 2.
c          IPNTR(8): pointer to the NCV RITZ values of the original system.
c          IPNTR(9): pointer to the NCV corresponding error bounds.
c          IPNTR(10): pointer to the NCV by NCV matrix of eigenvectors
c                     of the tridiagonal matrix T. Only referenced by
c                     dseupd  if RVEC = .TRUE. See Remarks.
c          -------------------------------------------------------------
c
c  WORKD   Double precision  work array of length 3*N.  (REVERSE COMMUNICATION)
c          Distributed array to be used in the basic Arnoldi iteration
c          for reverse communication.  The user should not use WORKD
c          as temporary workspace during the iteration. Upon termination
c          WORKD(1:N) contains B*RESID(1:N). If the Ritz vectors are desired
c          subroutine dseupd  uses this output.
c          See Data Distribution Note below.
c
c  WORKL   Double precision  work array of length LWORKL.  (OUTPUT/WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.  See Data Distribution Note below.
c
c  LWORKL  Integer.  (INPUT)
c          LWORKL must be at least NCV**2 + 8*NCV .
c
c  INFO    Integer.  (INPUT/OUTPUT)
c          If INFO .EQ. 0, a randomly initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          Error flag on output.
c          =  0: Normal exit.
c          =  1: Maximum number of iterations taken.
c                All possible eigenvalues of OP has been found. IPARAM(5)
c                returns the number of wanted converged Ritz values.
c          =  2: No longer an informational error. Deprecated starting
c                with release 2 of ARPACK.
c          =  3: No shifts could be applied during a cycle of the
c                Implicitly restarted Arnoldi iteration. One possibility
c                is to increase the size of NCV relative to NEV.
c                See remark 4 below.
c          = -1: N must be positive.
c          = -2: NEV must be positive.
c          = -3: NCV must be greater than NEV and less than or equal to N.
c          = -4: The maximum number of Arnoldi update iterations allowed
c                must be greater than zero.
c          = -5: WHICH must be one of 'LM', 'SM', 'LA', 'SA' or 'BE'.
c          = -6: BMAT must be one of 'I' or 'G'.
c          = -7: Length of private work array WORKL is not sufficient.
c          = -8: Error return from trid. eigenvalue calculation;
c                Informatinal error from LAPACK routine dsteqr .
c          = -9: Starting vector is zero.
c          = -10: IPARAM(7) must be 1,2,3,4,5.
c          = -11: IPARAM(7) = 1 and BMAT = 'G' are incompatable.
c          = -12: IPARAM(1) must be equal to 0 or 1.
c          = -13: NEV and WHICH = 'BE' are incompatable.
c          = -9999: Could not build an Arnoldi factorization.
c                   IPARAM(5) returns the size of the current Arnoldi
c                   factorization. The user is advised to check that
c                   enough workspace and array storage has been allocated.
c
c
c\Remarks
c  1. The converged Ritz values are always returned in ascending
c     algebraic order.  The computed Ritz values are approximate
c     eigenvalues of OP.  The selection of WHICH should be made
c     with this in mind when Mode = 3,4,5.  After convergence,
c     approximate eigenvalues of the original problem may be obtained
c     with the ARPACK subroutine dseupd .
c
c  2. If the Ritz vectors corresponding to the converged Ritz values
c     are needed, the user must call dseupd  immediately following completion
c     of dsaupd . This is new starting with version 2.1 of ARPACK.
c
c  3. If M can be factored into a Cholesky factorization M = LL`
c     then Mode = 2 should not be selected.  Instead one should use
c     Mode = 1 with  OP = inv(L)*A*inv(L`).  Appropriate triangular
c     linear systems should be solved with L and L` rather
c     than computing inverses.  After convergence, an approximate
c     eigenvector z of the original problem is recovered by solving
c     L`z = x  where x is a Ritz vector of OP.
c
c  4. At present there is no a-priori analysis to guide the selection
c     of NCV relative to NEV.  The only formal requrement is that NCV > NEV.
c     However, it is recommended that NCV .ge. 2*NEV.  If many problems of
c     the same type are to be solved, one should experiment with increasing
c     NCV while keeping NEV fixed for a given test problem.  This will
c     usually decrease the required number of OP*x operations but it
c     also increases the work and storage required to maintain the orthogonal
c     basis vectors.   The optimal "cross-over" with respect to CPU time
c     is problem dependent and must be determined empirically.
c
c  5. If IPARAM(7) = 2 then in the Reverse commuication interface the user
c     must do the following. When IDO = 1, Y = OP * X is to be computed.
c     When IPARAM(7) = 2 OP = inv(B)*A. After computing A*X the user
c     must overwrite X with A*X. Y is then the solution to the linear set
c     of equations B*Y = A*X.
c
c  6. When IPARAM(1) = 0, and IDO = 3, the user needs to provide the
c     NP = IPARAM(8) shifts in locations:
c     1   WORKL(IPNTR(11))
c     2   WORKL(IPNTR(11)+1)
c                        .
c                        .
c                        .
c     NP  WORKL(IPNTR(11)+NP-1).
c
c     The eigenvalues of the current tridiagonal matrix are located in
c     WORKL(IPNTR(6)) through WORKL(IPNTR(6)+NCV-1). They are in the
c     order defined by WHICH. The associated Ritz estimates are located in
c     WORKL(IPNTR(8)), WORKL(IPNTR(8)+1), ... , WORKL(IPNTR(8)+NCV-1).
c
c-----------------------------------------------------------------------
c
c\Data Distribution Note:
c
c  Fortran-D syntax:
c  ================
c  REAL       RESID(N), V(LDV,NCV), WORKD(3*N), WORKL(LWORKL)
c  DECOMPOSE  D1(N), D2(N,NCV)
c  ALIGN      RESID(I) with D1(I)
c  ALIGN      V(I,J)   with D2(I,J)
c  ALIGN      WORKD(I) with D1(I)     range (1:N)
c  ALIGN      WORKD(I) with D1(I-N)   range (N+1:2*N)
c  ALIGN      WORKD(I) with D1(I-2*N) range (2*N+1:3*N)
c  DISTRIBUTE D1(BLOCK), D2(BLOCK,:)
c  REPLICATED WORKL(LWORKL)
c
c  Cray MPP syntax:
c  ===============
c  REAL       RESID(N), V(LDV,NCV), WORKD(N,3), WORKL(LWORKL)
c  SHARED     RESID(BLOCK), V(BLOCK,:), WORKD(BLOCK,:)
c  REPLICATED WORKL(LWORKL)
c
c
c\BeginLib
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c  3. B.N. Parlett, "The Symmetric Eigenvalue Problem". Prentice-Hall,
c     1980.
c  4. B.N. Parlett, B. Nour-Omid, "Towards a Black Box Lanczos Program",
c     Computer Physics Communications, 53 (1989), pp 169-179.
c  5. B. Nour-Omid, B.N. Parlett, T. Ericson, P.S. Jensen, "How to
c     Implement the Spectral Transformation", Math. Comp., 48 (1987),
c     pp 663-673.
c  6. R.G. Grimes, J.G. Lewis and H.D. Simon, "A Shifted Block Lanczos
c     Algorithm for Solving Sparse Symmetric Generalized Eigenproblems",
c     SIAM J. Matr. Anal. Apps.,  January (1993).
c  7. L. Reichel, W.B. Gragg, "Algorithm 686: FORTRAN Subroutines
c     for Updating the QR decomposition", ACM TOMS, December 1990,
c     Volume 16 Number 4, pp 369-377.
c  8. R.B. Lehoucq, D.C. Sorensen, "Implementation of Some Spectral
c     Transformations in a k-Step Arnoldi Method". In Preparation.
c
c\Routines called:
c     dsaup2   ARPACK routine that implements the Implicitly Restarted
c             Arnoldi Iteration.
c     dstats   ARPACK routine that initialize timing and other statistics
c             variables.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dvout    ARPACK utility routine that prints vectors.
c     dlamch   LAPACK routine that determines machine constants.
c
c\Authors
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/15/93: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: saupd.F   SID: 2.8   DATE OF SID: 04/10/01   RELEASE: 2
c
c\Remarks
c     1. None
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsaupd
     &   ( ido, bmat, n, which, nev, tol, resid, ncv, v, ldv, iparam,
     &     ipntr, workd, workl, lworkl, info )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1, which*2
      integer    ido, info, ldv, lworkl, n, ncv, nev
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    iparam(8), ipntr(11)
      Double precision
     &           resid(n), v(ldv,ncv), workd(3*n), workl(lworkl)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           zero
      parameter (zero = 0.0D+0 )
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    bounds, ierr, ih, iq, ishift, iw,
     &           ldh, ldq, mxiter, mode, nb,
     &           nev0, next, np, ritz, j
      save       bounds, ierr, ih, iq, ishift, iw,
     &           ldh, ldq, mxiter, mode, nb,
     &           nev0, next, np, ritz
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dsaup2
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlamch
      external   dlamch
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (ido .eq. 0) then
c
c        %-------------------------------%
c        | Initialize timing statistics  |
c        | & message level for debugging |
c        %-------------------------------%
c
         ierr   = 0
         ishift = iparam(1)
         mxiter = iparam(3)
         nb     = 1
c
c        %--------------------------------------------%
c        | Revision 2 performs only implicit restart. |
c        %--------------------------------------------%
c
         mode   = iparam(7)
c
c        %----------------%
c        | Error checking |
c        %----------------%
c
         if (n .le. 0) then
            ierr = -1
         else if (nev .le. 0) then
            ierr = -2
         else if (ncv .le. nev .or.  ncv .gt. n) then
            ierr = -3
         end if
c
c        %----------------------------------------------%
c        | NP is the number of additional steps to      |
c        | extend the length NEV Lanczos factorization. |
c        %----------------------------------------------%
c
         np     = ncv - nev
c
         if (mxiter .le. 0)                     ierr = -4
         if (which .ne. 'LM' .and.
     &       which .ne. 'SM' .and.
     &       which .ne. 'LA' .and.
     &       which .ne. 'SA' .and.
     &       which .ne. 'BE')                   ierr = -5
         if (bmat .ne. 'I' .and. bmat .ne. 'G') ierr = -6
c
         if (lworkl .lt. ncv**2 + 8*ncv)        ierr = -7
         if (mode .lt. 1 .or. mode .gt. 5) then
                                                ierr = -10
         else if (mode .eq. 1 .and. bmat .eq. 'G') then
                                                ierr = -11
         else if (ishift .lt. 0 .or. ishift .gt. 1) then
                                                ierr = -12
         else if (nev .eq. 1 .and. which .eq. 'BE') then
                                                ierr = -13
         end if
c
c        %------------%
c        | Error Exit |
c        %------------%
c
         if (ierr .ne. 0) then
            info = ierr
            ido  = 99
            go to 9000
         end if
c
c        %------------------------%
c        | Set default parameters |
c        %------------------------%
c
         if (nb .le. 0)                         nb = 1
         if (tol .le. zero)                     tol = dlamch ('EpsMach')
c
c        %----------------------------------------------%
c        | NP is the number of additional steps to      |
c        | extend the length NEV Lanczos factorization. |
c        | NEV0 is the local variable designating the   |
c        | size of the invariant subspace desired.      |
c        %----------------------------------------------%
c
         np     = ncv - nev
         nev0   = nev
c
c        %-----------------------------%
c        | Zero out internal workspace |
c        %-----------------------------%
c
         do 10 j = 1, ncv**2 + 8*ncv
            workl(j) = zero
 10      continue
c
c        %-------------------------------------------------------%
c        | Pointer into WORKL for address of H, RITZ, BOUNDS, Q  |
c        | etc... and the remaining workspace.                   |
c        | Also update pointer to be used on output.             |
c        | Memory is laid out as follows:                        |
c        | workl(1:2*ncv) := generated tridiagonal matrix        |
c        | workl(2*ncv+1:2*ncv+ncv) := ritz values               |
c        | workl(3*ncv+1:3*ncv+ncv) := computed error bounds     |
c        | workl(4*ncv+1:4*ncv+ncv*ncv) := rotation matrix Q     |
c        | workl(4*ncv+ncv*ncv+1:7*ncv+ncv*ncv) := workspace     |
c        %-------------------------------------------------------%
c
         ldh    = ncv
         ldq    = ncv
         ih     = 1
         ritz   = ih     + 2*ldh
         bounds = ritz   + ncv
         iq     = bounds + ncv
         iw     = iq     + ncv**2
         next   = iw     + 3*ncv
c
         ipntr(4) = next
         ipntr(5) = ih
         ipntr(6) = ritz
         ipntr(7) = bounds
         ipntr(11) = iw
      end if
c
c     %-------------------------------------------------------%
c     | Carry out the Implicitly restarted Lanczos Iteration. |
c     %-------------------------------------------------------%
c
      call dsaup2
     &   ( ido, bmat, n, which, nev0, np, tol, resid, mode,
     &     ishift, mxiter, v, ldv, workl(ih), ldh, workl(ritz),
     &     workl(bounds), workl(iq), ldq, workl(iw), ipntr, workd,
     &     info )
c
c     %--------------------------------------------------%
c     | ido .ne. 99 implies use of reverse communication |
c     | to compute operations involving OP or shifts.    |
c     %--------------------------------------------------%
c
      if (ido .eq. 3) iparam(8) = np
      if (ido .ne. 99) go to 9000
c
      iparam(3) = mxiter
      iparam(5) = np
c
c     %------------------------------------%
c     | Exit if there was an informational |
c     | error within dsaup2 .               |
c     %------------------------------------%
c
      if (info .lt. 0) go to 9000
      if (info .eq. 2) info = 3
c
 9000 continue
c
      return
c
c     %---------------%
c     | End of dsaupd  |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsaup2
c
c\Description:
c  Intermediate level interface called by dsaupd.
c
c\Usage:
c  call dsaup2
c     ( IDO, BMAT, N, WHICH, NEV, NP, TOL, RESID, MODE, IUPD,
c       ISHIFT, MXITER, V, LDV, H, LDH, RITZ, BOUNDS, Q, LDQ, WORKL,
c       IPNTR, WORKD, INFO )
c
c\Arguments
c
c  IDO, BMAT, N, WHICH, NEV, TOL, RESID: same as defined in dsaupd.
c  MODE, ISHIFT, MXITER: see the definition of IPARAM in dsaupd.
c
c  NP      Integer.  (INPUT/OUTPUT)
c          Contains the number of implicit shifts to apply during
c          each Arnoldi/Lanczos iteration.
c          If ISHIFT=1, NP is adjusted dynamically at each iteration
c          to accelerate convergence and prevent stagnation.
c          This is also roughly equal to the number of matrix-vector
c          products (involving the operator OP) per Arnoldi iteration.
c          The logic for adjusting is contained within the current
c          subroutine.
c          If ISHIFT=0, NP is the number of shifts the user needs
c          to provide via reverse comunication. 0 < NP < NCV-NEV.
c          NP may be less than NCV-NEV since a leading block of the current
c          upper Tridiagonal matrix has split off and contains "unwanted"
c          Ritz values.
c          Upon termination of the IRA iteration, NP contains the number
c          of "converged" wanted Ritz values.
c
c  IUPD    Integer.  (INPUT)
c          IUPD .EQ. 0: use explicit restart instead implicit update.
c          IUPD .NE. 0: use implicit update.
c          removed since unused argument
c
c  V       Double precision N by (NEV+NP) array.  (INPUT/OUTPUT)
c          The Lanczos basis vectors.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (NEV+NP) by 2 array.  (OUTPUT)
c          H is used to store the generated symmetric tridiagonal matrix
c          The subdiagonal is stored in the first column of H starting
c          at H(2,1).  The main diagonal is stored in the second column
c          of H starting at H(1,2). If dsaup2 converges store the
c          B-norm of the final residual vector in H(1,1).
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  RITZ    Double precision array of length NEV+NP.  (OUTPUT)
c          RITZ(1:NEV) contains the computed Ritz values of OP.
c
c  BOUNDS  Double precision array of length NEV+NP.  (OUTPUT)
c          BOUNDS(1:NEV) contain the error bounds corresponding to RITZ.
c
c  Q       Double precision (NEV+NP) by (NEV+NP) array.  (WORKSPACE)
c          Private (replicated) work array used to accumulate the
c          rotation in the shift application step.
c
c  LDQ     Integer.  (INPUT)
c          Leading dimension of Q exactly as declared in the calling
c          program.
c
c  WORKL   Double precision array of length at least 3*(NEV+NP).  (INPUT/WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.  It is used in the computation of the
c          tridiagonal eigenvalue problem, the calculation and
c          application of the shifts and convergence checking.
c          If ISHIFT .EQ. O and IDO .EQ. 3, the first NP locations
c          of WORKL are used in reverse communication to hold the user
c          supplied shifts.
c
c  IPNTR   Integer array of length 3.  (OUTPUT)
c          Pointer to mark the starting locations in the WORKD for
c          vectors used by the Lanczos iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X.
c          IPNTR(2): pointer to the current result vector Y.
c          IPNTR(3): pointer to the vector B * X when used in one of
c                    the spectral transformation modes.  X is the current
c                    operand.
c          -------------------------------------------------------------
c
c  WORKD   Double precision work array of length 3*N.  (REVERSE COMMUNICATION)
c          Distributed array to be used in the basic Lanczos iteration
c          for reverse communication.  The user should not use WORKD
c          as temporary workspace during the iteration !!!!!!!!!!
c          See Data Distribution Note in dsaupd.
c
c  INFO    Integer.  (INPUT/OUTPUT)
c          If INFO .EQ. 0, a randomly initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          Error flag on output.
c          =     0: Normal return.
c          =     1: All possible eigenvalues of OP has been found.
c                   NP returns the size of the invariant subspace
c                   spanning the operator OP.
c          =     2: No shifts could be applied.
c          =    -8: Error return from trid. eigenvalue calculation;
c                   This should never happen.
c          =    -9: Starting vector is zero.
c          = -9999: Could not build an Lanczos factorization.
c                   Size that was built in returned in NP.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c  3. B.N. Parlett, "The Symmetric Eigenvalue Problem". Prentice-Hall,
c     1980.
c  4. B.N. Parlett, B. Nour-Omid, "Towards a Black Box Lanczos Program",
c     Computer Physics Communications, 53 (1989), pp 169-179.
c  5. B. Nour-Omid, B.N. Parlett, T. Ericson, P.S. Jensen, "How to
c     Implement the Spectral Transformation", Math. Comp., 48 (1987),
c     pp 663-673.
c  6. R.G. Grimes, J.G. Lewis and H.D. Simon, "A Shifted Block Lanczos
c     Algorithm for Solving Sparse Symmetric Generalized Eigenproblems",
c     SIAM J. Matr. Anal. Apps.,  January (1993).
c  7. L. Reichel, W.B. Gragg, "Algorithm 686: FORTRAN Subroutines
c     for Updating the QR decomposition", ACM TOMS, December 1990,
c     Volume 16 Number 4, pp 369-377.
c
c\Routines called:
c     dgetv0  ARPACK initial vector generation routine.
c     dsaitr  ARPACK Lanczos factorization routine.
c     dsapps  ARPACK application of implicit shifts routine.
c     dsconv  ARPACK convergence of Ritz values routine.
c     dseigt  ARPACK compute Ritz values and error bounds routine.
c     dsgets  ARPACK reorder Ritz values and error bounds routine.
c     dsortr  ARPACK sorting routine.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dvout   ARPACK utility routine that prints vectors.
c     dlamch  LAPACK routine that determines machine constants.
c     dcopy   Level 1 BLAS that copies one vector to another.
c     ddot    Level 1 BLAS that computes the scalar product of two vectors.
c     dnrm2   Level 1 BLAS that computes the norm of a vector.
c     dscal   Level 1 BLAS that scales a vector.
c     dswap   Level 1 BLAS that swaps two vectors.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/15/93: Version ' 2.4'
c     xx/xx/95: Version ' 2.4'.  (R.B. Lehoucq)
c
c\SCCS Information: @(#)
c FILE: saup2.F   SID: 2.7   DATE OF SID: 5/19/98   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsaup2
     &   ( ido, bmat, n, which, nev, np, tol, resid, mode,
     &     ishift, mxiter, v, ldv, h, ldh, ritz, bounds,
     &     q, ldq, workl, ipntr, workd, info )

      implicit none
c
c     %----------------------------------------------------%
c     | Include files for debugging and timing information |
c     %----------------------------------------------------%
c
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1, which*2
      integer    ido, info, ishift, ldh, ldq, ldv, mxiter,
     &           n, mode, nev, np
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    ipntr(3)
      Double precision
     &           bounds(nev+np), h(ldh,2), q(ldq,nev+np), resid(n),
     &           ritz(nev+np), v(ldv,nev+np), workd(3*n),
     &           workl(3*(nev+np))
c
c     %------------%
c     | Parameters |
c     %------------%
c
      integer   ione
      parameter (ione=1)
      integer
     &           dsaupzero
      parameter (dsaupzero = 0)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      character  wprime*2
      logical    cnorm, getv0, initv, update, ushift, dsaup2true
      integer    ierr, iter, j, kplusp,  nconv, nevbef, nev0,
     &           np0, nptemp, nevd2, nevm2, dsaup2zero
      Double precision
     &           rnorm, temp, eps23
      save       cnorm, getv0, initv, update, ushift,
     &           iter, kplusp,  nconv, nev0, np0,
     &           rnorm, eps23
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy, dgetv0, dsaitr, dscal, dsconv, dseigt, dsgets,
     &           dsapps, dsortr, dswap
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      logical,  external :: eqZERO

      Double precision
     &           ddot, dnrm2, dlamch
      external   ddot, dnrm2, dlamch
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    min
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      dsaup2true = .true.
      dsaup2zero = 0
c
      if (ido .eq. 0) then
c
c        %---------------------------------%
c        | Set machine dependent constant. |
c        %---------------------------------%
c
         eps23 = dlamch('Epsilon-Machine')
         eps23 = eps23**(2.0D+0/3.0D+0)
c
c        %-------------------------------------%
c        | nev0 and np0 are integer variables  |
c        | hold the initial values of NEV & NP |
c        %-------------------------------------%
c
         nev0   = nev
         np0    = np
c
c        %-------------------------------------%
c        | kplusp is the bound on the largest  |
c        |        Lanczos factorization built. |
c        | nconv is the current number of      |
c        |        "converged" eigenvlues.      |
c        | iter is the counter on the current  |
c        |      iteration step.                |
c        %-------------------------------------%
c
         kplusp = nev0 + np0
         nconv  = 0
         iter   = 0
c
c        %--------------------------------------------%
c        | Set flags for computing the first NEV steps |
c        | of the Lanczos factorization.              |
c        %--------------------------------------------%
c
         getv0    = .true.
         update   = .false.
         ushift   = .false.
         cnorm    = .false.
c
         if (info .ne. 0) then
c
c        %--------------------------------------------%
c        | User provides the initial residual vector. |
c        %--------------------------------------------%
c
            initv = .true.
            info  = 0
         else
            initv = .false.
         end if
      end if
c
c     %---------------------------------------------%
c     | Get a possibly random starting vector and   |
c     | force it into the range of the operator OP. |
c     %---------------------------------------------%
c
c   10 continue
c
      if (getv0) then
         call dgetv0 (ido, bmat, initv, n, ione, v, ldv, resid, rnorm,
     &                ipntr, workd, info)
c
         if (ido .ne. 99) go to 9000
c
         if (eqZERO(rnorm)) then
c
c           %-----------------------------------------%
c           | The initial vector is zero. Error exit. |
c           %-----------------------------------------%
c
            info = -9
            go to 1200
         end if
         getv0 = .false.
         ido  = 0
      end if
c
c     %------------------------------------------------------------%
c     | Back from reverse communication: continue with update step |
c     %------------------------------------------------------------%
c
      if (update) go to 20
c
c     %-------------------------------------------%
c     | Back from computing user specified shifts |
c     %-------------------------------------------%
c
      if (ushift) go to 50
c
c     %-------------------------------------%
c     | Back from computing residual norm   |
c     | at the end of the current iteration |dsaup2true
c     %-------------------------------------%
c
      if (cnorm)  go to 100
c
c     %----------------------------------------------------------%
c     | Compute the first NEV steps of the Lanczos factorization |
c     %----------------------------------------------------------%
c
      call dsaitr (ido, bmat, n, dsaupzero, nev0, mode, resid, rnorm,
     &             v, ldv, h, ldh, ipntr, workd, info)
c
c     %---------------------------------------------------%
c     | ido .ne. 99 implies use of reverse communication  |
c     | to compute operations involving OP and possibly B |
c     %---------------------------------------------------%
c
      if (ido .ne. 99) go to 9000
c
      if (info .gt. 0) then
c
c        %-----------------------------------------------------%
c        | dsaitr was unable to build an Lanczos factorization |
c        | of length NEV0. INFO is returned with the size of   |
c        | the factorization built. Exit main loop.            |
c        %-----------------------------------------------------%
c
         np   = info
         mxiter = iter
         info = -9999
         go to 1200
      end if
c
c     %--------------------------------------------------------------%
c     |                                                              |
c     |           M A I N  LANCZOS  I T E R A T I O N  L O O P       |
c     |           Each iteration implicitly restarts the Lanczos     |
c     |           factorization in place.                            |
c     |                                                              |
c     %--------------------------------------------------------------%
c
 1000 continue
c
         iter = iter + 1
c

c        %------------------------------------------------------------%
c        | Compute NP additional steps of the Lanczos factorization. |
c        %------------------------------------------------------------%
c
         ido = 0
   20    continue
         update = .true.
c
         call dsaitr (ido, bmat, n, nev, np, mode, resid, rnorm, v,
     &                ldv, h, ldh, ipntr, workd, info)
c
c        %---------------------------------------------------%
c        | ido .ne. 99 implies use of reverse communication  |
c        | to compute operations involving OP and possibly B |
c        %---------------------------------------------------%
c
         if (ido .ne. 99) go to 9000
c
         if (info .gt. 0) then
c
c           %-----------------------------------------------------%
c           | dsaitr was unable to build an Lanczos factorization |
c           | of length NEV0+NP0. INFO is returned with the size  |
c           | of the factorization built. Exit main loop.         |
c           %-----------------------------------------------------%
c
            np = info
            mxiter = iter
            info = -9999
            go to 1200
         end if
         update = .false.
c
c        %--------------------------------------------------------%
c        | Compute the eigenvalues and corresponding error bounds |
c        | of the current symmetric tridiagonal matrix.           |
c        %--------------------------------------------------------%
c
         call dseigt (rnorm, kplusp, h, ldh, ritz, bounds, workl, ierr)
c
         if (ierr .ne. 0) then
            info = -8
            go to 1200
         end if
c
c        %----------------------------------------------------%
c        | Make a copy of eigenvalues and corresponding error |
c        | bounds obtained from _seigt.                       |
c        %----------------------------------------------------%
c
         call dcopy(kplusp, ritz, 1, workl(kplusp+1), 1)
         call dcopy(kplusp, bounds, 1, workl(2*kplusp+1), 1)
c
c        %---------------------------------------------------%
c        | Select the wanted Ritz values and their bounds    |
c        | to be used in the convergence test.               |
c        | The selection is based on the requested number of |
c        | eigenvalues instead of the current NEV and NP to  |
c        | prevent possible misconvergence.                  |
c        | * Wanted Ritz values := RITZ(NP+1:NEV+NP)         |
c        | * Shifts := RITZ(1:NP) := WORKL(1:NP)             |
c        %---------------------------------------------------%
c
         nev = nev0
         np = np0
         call dsgets (ishift, which, nev, np, ritz, bounds, workl)
c
c        %-------------------%
c        | Convergence test. |
c        %-------------------%
c
         call dcopy (nev, bounds(np+1), 1, workl(np+1), 1)
         call dsconv (nev, ritz(np+1), workl(np+1), tol, nconv)
c
c        %---------------------------------------------------------%
c        | Count the number of unwanted Ritz values that have zero |
c        | Ritz estimates. If any Ritz estimates are equal to zero |
c        | then a leading block of H of order equal to at least    |
c        | the number of Ritz values with zero Ritz estimates has  |
c        | split off. None of these Ritz values may be removed by  |
c        | shifting. Decrease NP the number of shifts to apply. If |
c        | no shifts may be applied, then prepare to exit          |
c        %---------------------------------------------------------%
c
         nptemp = np
         do 30 j=1, nptemp
            if (eqZERO(bounds(j))) then
               np = np - 1
               nev = nev + 1
            end if
 30      continue
c
         if ( (nconv .ge. nev0) .or.
     &        (iter .gt. mxiter) .or.
     &        (np .eq. 0) ) then
c
c           %------------------------------------------------%
c           | Prepare to exit. Put the converged Ritz values |
c           | and corresponding bounds in RITZ(1:NCONV) and  |
c           | BOUNDS(1:NCONV) respectively. Then sort. Be    |
c           | careful when NCONV > NP since we don't want to |
c           | swap overlapping locations.                    |
c           %------------------------------------------------%
c
            if (which .eq. 'BE') then
c
c              %-----------------------------------------------------%
c              | Both ends of the spectrum are requested.            |
c              | Sort the eigenvalues into algebraically decreasing  |
c              | order first then swap low end of the spectrum next  |
c              | to high end in appropriate locations.               |
c              | NOTE: when np < floor(nev/2) be careful not to swap |
c              | overlapping locations.                              |
c              %-----------------------------------------------------%
c
               wprime = 'SA'
               call dsortr (wprime, dsaup2true, kplusp, ritz, bounds)
               nevd2 = nev0 / 2
               nevm2 = nev0 - nevd2
               if ( nev .gt. 1 ) then
                  call dswap ( min(nevd2,np), ritz(nevm2+1), 1,
     &                 ritz( max(kplusp-nevd2+1,kplusp-np+1) ), 1)
                  call dswap ( min(nevd2,np), bounds(nevm2+1), 1,
     &                 bounds( max(kplusp-nevd2+1,kplusp-np+1)), 1)
               end if
c
            else
c
c              %--------------------------------------------------%
c              | LM, SM, LA, SA case.                             |
c              | Sort the eigenvalues of H into the an order that |
c              | is opposite to WHICH, and apply the resulting    |
c              | order to BOUNDS.  The eigenvalues are sorted so  |
c              | that the wanted part are always within the first |
c              | NEV locations.                                   |
c              %--------------------------------------------------%
c
               if (which .eq. 'LM') wprime = 'SM'
               if (which .eq. 'SM') wprime = 'LM'
               if (which .eq. 'LA') wprime = 'SA'
               if (which .eq. 'SA') wprime = 'LA'
c
               call dsortr (wprime, dsaup2true, kplusp, ritz, bounds)
c
            end if
c
c           %--------------------------------------------------%
c           | Scale the Ritz estimate of each Ritz value       |
c           | by 1 / max(eps23,magnitude of the Ritz value).   |
c           %--------------------------------------------------%
c
            do 35 j = 1, nev0
               temp = max( eps23, abs(ritz(j)) )
               bounds(j) = bounds(j)/temp
 35         continue
c
c           %----------------------------------------------------%
c           | Sort the Ritz values according to the scaled Ritz  |
c           | esitmates.  This will push all the converged ones  |
c           | towards the front of ritzr, ritzi, bounds          |
c           | (in the case when NCONV < NEV.)                    |
c           %----------------------------------------------------%
c
            wprime = 'LA'
            call dsortr(wprime, dsaup2true, nev0, bounds, ritz)
c
c           %----------------------------------------------%
c           | Scale the Ritz estimate back to its original |
c           | value.                                       |
c           %----------------------------------------------%
c
            do 40 j = 1, nev0
                temp = max( eps23, abs(ritz(j)) )
                bounds(j) = bounds(j)*temp
 40         continue
c
c           %--------------------------------------------------%
c           | Sort the "converged" Ritz values again so that   |
c           | the "threshold" values and their associated Ritz |
c           | estimates appear at the appropriate position in  |
c           | ritz and bound.                                  |
c           %--------------------------------------------------%
c
            if (which .eq. 'BE') then
c
c              %------------------------------------------------%
c              | Sort the "converged" Ritz values in increasing |
c              | order.  The "threshold" values are in the      |
c              | middle.                                        |
c              %------------------------------------------------%
c
               wprime = 'LA'
               call dsortr(wprime, dsaup2true, nconv, ritz, bounds)
c
            else
c
c              %----------------------------------------------%
c              | In LM, SM, LA, SA case, sort the "converged" |
c              | Ritz values according to WHICH so that the   |
c              | "threshold" value appears at the front of    |
c              | ritz.                                        |
c              %----------------------------------------------%

               call dsortr(which, dsaup2true, nconv, ritz, bounds)
c
            end if
c
c           %------------------------------------------%
c           |  Use h( 1,1 ) as storage to communicate  |
c           |  rnorm to _seupd if needed               |
c           %------------------------------------------%
c
            h(1,1) = rnorm
c
c
c           %------------------------------------%
c           | Max iterations have been exceeded. |
c           %------------------------------------%
c
            if (iter .gt. mxiter .and. nconv .lt. nev) info = 1
c
c           %---------------------%
c           | No shifts to apply. |
c           %---------------------%
c
            if (np .eq. 0 .and. nconv .lt. nev0) info = 2
c
            np = nconv
            go to 1100
c
         else if (nconv .lt. nev .and. ishift .eq. 1) then
c
c           %---------------------------------------------------%
c           | Do not have all the requested eigenvalues yet.    |
c           | To prevent possible stagnation, adjust the number |
c           | of Ritz values and the shifts.                    |
c           %---------------------------------------------------%
c
            nevbef = nev
            nev = nev + min (nconv, np/2)
            if (nev .eq. 1 .and. kplusp .ge. 6) then
               nev = kplusp / 2
            else if (nev .eq. 1 .and. kplusp .gt. 2) then
               nev = 2
            end if
            np  = kplusp - nev
c
c           %---------------------------------------%
c           | If the size of NEV was just increased |
c           | resort the eigenvalues.               |
c           %---------------------------------------%
c
            if (nevbef .lt. nev)
     &         call dsgets (ishift, which, nev, np, ritz, bounds,
     &              workl)
c
         end if
c
         if (ishift .eq. 0) then
c
c           %-----------------------------------------------------%
c           | User specified shifts: reverse communication to     |
c           | compute the shifts. They are returned in the first  |
c           | NP locations of WORKL.                              |
c           %-----------------------------------------------------%
c
            ushift = .true.
            ido = 3
            go to 9000
         end if
c
   50    continue
c
c        %------------------------------------%
c        | Back from reverse communication;   |
c        | User specified shifts are returned |
c        | in WORKL(1:*NP)                   |
c        %------------------------------------%
c
         ushift = .false.
c
c
c        %---------------------------------------------------------%
c        | Move the NP shifts to the first NP locations of RITZ to |
c        | free up WORKL.  This is for the non-exact shift case;   |
c        | in the exact shift case, dsgets already handles this.   |
c        %---------------------------------------------------------%
c
         if (ishift .eq. 0) call dcopy (np, workl, 1, ritz(1), 1)
cx         if (ishift .eq. 0) call dcopy (np, workl, 1, ritz, 1)
c
c        %---------------------------------------------------------%
c        | Apply the NP0 implicit shifts by QR bulge chasing.      |
c        | Each shift is applied to the entire tridiagonal matrix. |
c        | The first 2*N locations of WORKD are used as workspace. |
c        | After dsapps is done, we have a Lanczos                 |
c        | factorization of length NEV.                            |
c        %---------------------------------------------------------%
c
         call dsapps (n, nev, np, ritz, v, ldv, h, ldh, resid, q, ldq,
     &        workd)
c
c        %---------------------------------------------%
c        | Compute the B-norm of the updated residual. |
c        | Keep B*RESID in WORKD(1:N) to be used in    |
c        | the first step of the next call to dsaitr.  |
c        %---------------------------------------------%
c
         cnorm = .true.
c         call second (t2)
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(n+1), 1)
            ipntr(1) = n + 1
            ipntr(2) = 1
            ido = 2
c
c           %----------------------------------%
c           | Exit in order to compute B*RESID |
c           %----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(1), 1)
         end if
c
  100    continue
c
c        %----------------------------------%
c        | Back from reverse communication; |
c        | WORKD(1:N) := B*RESID            |
c        %----------------------------------%
c
         if (bmat .eq. 'G') then
            rnorm = ddot (n, resid, 1, workd, 1)
            rnorm = sqrt(abs(rnorm))
         else if (bmat .eq. 'I') then
            rnorm = dnrm2(n, resid, 1)
         end if
         cnorm = .false.
c  130    continue
c
c
      go to 1000
c
c     %---------------------------------------------------------------%
c     |                                                               |
c     |  E N D     O F     M A I N     I T E R A T I O N     L O O P  |
c     |                                                               |
c     %---------------------------------------------------------------%
c
 1100 continue
c
      mxiter = iter
      nev = nconv
c
 1200 continue
      ido = 99
c
c     %------------%
c     | Error exit |
c     %------------%
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsaup2 |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsconv
c
c\Description:
c  Convergence testing for the symmetric Arnoldi eigenvalue routine.
c
c\Usage:
c  call dsconv
c     ( N, RITZ, BOUNDS, TOL, NCONV )
c
c\Arguments
c  N       Integer.  (INPUT)
c          Number of Ritz values to check for convergence.
c
c  RITZ    Double precision array of length N.  (INPUT)
c          The Ritz values to be checked for convergence.
c
c  BOUNDS  Double precision array of length N.  (INPUT)
c          Ritz estimates associated with the Ritz values in RITZ.
c
c  TOL     Double precision scalar.  (INPUT)
c          Desired relative accuracy for a Ritz value to be considered
c          "converged".
c
c  NCONV   Integer scalar.  (OUTPUT)
c          Number of "converged" Ritz values.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Routines called:
c     second  ARPACK utility routine for timing.
c     dlamch  LAPACK routine that determines machine constants.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\SCCS Information: @(#)
c FILE: sconv.F   SID: 2.4   DATE OF SID: 4/19/96   RELEASE: 2
c
c\Remarks
c     1. Starting with version 2.4, this routine no longer uses the
c        Parlett strategy using the gap conditions.
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsconv (n, ritz, bounds, tol, nconv)
c
c     %----------------------------------------------------%
c     | Include files for debugging and timing information |
c     %----------------------------------------------------%
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    n, nconv
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           ritz(n), bounds(n)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i
      Double precision
     &           temp, eps23
c
c     %-------------------%
c     | External routines |
c     %-------------------%
c
      Double precision
     &           dlamch
      external   dlamch

c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    abs
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c      call second (t0)
c
      eps23 = dlamch('Epsilon-Machine')
      eps23 = eps23**(2.0D+0 / 3.0D+0)
c
      nconv  = 0
      do 10 i = 1, n
c
c        %-----------------------------------------------------%
c        | The i-th Ritz value is considered "converged"       |
c        | when: bounds(i) .le. TOL*max(eps23, abs(ritz(i)))   |
c        %-----------------------------------------------------%
c
         temp = max( eps23, abs(ritz(i)) )
         if ( bounds(i) .le. tol*temp ) then
            nconv = nconv + 1
         end if
c
   10 continue
c
c      call second (t1)
c
      return
c
c     %---------------%
c     | End of dsconv |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsaitr
c
c\Description:
c  Reverse communication interface for applying NP additional steps to
c  a K step symmetric Arnoldi factorization.
c
c  Input:  OP*V_{k}  -  V_{k}*H = r_{k}*e_{k}^T
c
c          with (V_{k}^T)*B*V_{k} = I, (V_{k}^T)*B*r_{k} = 0.
c
c  Output: OP*V_{k+p}  -  V_{k+p}*H = r_{k+p}*e_{k+p}^T
c
c          with (V_{k+p}^T)*B*V_{k+p} = I, (V_{k+p}^T)*B*r_{k+p} = 0.
c
c  where OP and B are as in dsaupd.  The B-norm of r_{k+p} is also
c  computed and returned.
c
c\Usage:
c  call dsaitr
c     ( IDO, BMAT, N, K, NP, MODE, RESID, RNORM, V, LDV, H, LDH,
c       IPNTR, WORKD, INFO )
c
c\Arguments
c  IDO     Integer.  (INPUT/OUTPUT)
c          Reverse communication flag.
c          -------------------------------------------------------------
c          IDO =  0: first call to the reverse communication interface
c          IDO = -1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y.
c                    This is for the restart phase to force the new
c                    starting vector into the range of OP.
c          IDO =  1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y,
c                    IPNTR(3) is the pointer into WORK for B * X.
c          IDO =  2: compute  Y = B * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y.
c          IDO = 99: done
c          -------------------------------------------------------------
c          When the routine is used in the "shift-and-invert" mode, the
c          vector B * Q is already available and does not need to be
c          recomputed in forming OP * Q.
c
c  BMAT    Character*1.  (INPUT)
c          BMAT specifies the type of matrix B that defines the
c          semi-inner product for the operator OP.  See dsaupd.
c          B = 'I' -> standard eigenvalue problem A*x = lambda*x
c          B = 'G' -> generalized eigenvalue problem A*x = lambda*M*x
c
c  N       Integer.  (INPUT)
c          Dimension of the eigenproblem.
c
c  K       Integer.  (INPUT)
c          Current order of H and the number of columns of V.
c
c  NP      Integer.  (INPUT)
c          Number of additional Arnoldi steps to take.
c
c  MODE    Integer.  (INPUT)
c          Signifies which form for "OP". If MODE=2 then
c          a reduction in the number of B matrix vector multiplies
c          is possible since the B-norm of OP*x is equivalent to
c          the inv(B)-norm of A*x.
c
c  RESID   Double precision array of length N.  (INPUT/OUTPUT)
c          On INPUT:  RESID contains the residual vector r_{k}.
c          On OUTPUT: RESID contains the residual vector r_{k+p}.
c
c  RNORM   Double precision scalar.  (INPUT/OUTPUT)
c          On INPUT the B-norm of r_{k}.
c          On OUTPUT the B-norm of the updated residual r_{k+p}.
c
c  V       Double precision N by K+NP array.  (INPUT/OUTPUT)
c          On INPUT:  V contains the Arnoldi vectors in the first K
c          columns.
c          On OUTPUT: V contains the new NP Arnoldi vectors in the next
c          NP columns.  The first K columns are unchanged.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (K+NP) by 2 array.  (INPUT/OUTPUT)
c          H is used to store the generated symmetric tridiagonal matrix
c          with the subdiagonal in the first column starting at H(2,1)
c          and the main diagonal in the second column.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  IPNTR   Integer array of length 3.  (OUTPUT)
c          Pointer to mark the starting locations in the WORK for
c          vectors used by the Arnoldi iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X.
c          IPNTR(2): pointer to the current result vector Y.
c          IPNTR(3): pointer to the vector B * X when used in the
c                    shift-and-invert mode.  X is the current operand.
c          -------------------------------------------------------------
c
c  WORKD   Double precision work array of length 3*N.  (REVERSE COMMUNICATION)
c          Distributed array to be used in the basic Arnoldi iteration
c          for reverse communication.  The calling program should not
c          use WORKD as temporary workspace during the iteration !!!!!!
c          On INPUT, WORKD(1:N) = B*RESID where RESID is associated
c          with the K step Arnoldi factorization. Used to save some
c          computation at the first step.
c          On OUTPUT, WORKD(1:N) = B*RESID where RESID is associated
c          with the K+NP step Arnoldi factorization.
c
c  INFO    Integer.  (OUTPUT)
c          = 0: Normal exit.
c          > 0: Size of an invariant subspace of OP is found that is
c               less than K + NP.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dgetv0  ARPACK routine to generate the initial vector.
c     ivout   ARPACK utility routine that prints integers.
c     dmout   ARPACK utility routine that prints matrices.
c     dvout   ARPACK utility routine that prints vectors.
c     dlamch  LAPACK routine that determines machine constants.
c     dlascl  LAPACK routine for careful scaling of a matrix.
c     dgemv   Level 2 BLAS routine for matrix vector multiplication.
c     daxpy   Level 1 BLAS that computes a vector triad.
c     dscal   Level 1 BLAS that scales a vector.
c     dcopy   Level 1 BLAS that copies one vector to another .
c     ddot    Level 1 BLAS that computes the scalar product of two vectors.
c     dnrm2   Level 1 BLAS that computes the norm of a vector.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/93: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: saitr.F   SID: 2.6   DATE OF SID: 8/28/96   RELEASE: 2
c
c\Remarks
c  The algorithm implemented is:
c
c  restart = .false.
c  Given V_{k} = [v_{1}, ..., v_{k}], r_{k};
c  r_{k} contains the initial residual vector even for k = 0;
c  Also assume that rnorm = || B*r_{k} || and B*r_{k} are already
c  computed by the calling program.
c
c  betaj = rnorm ; p_{k+1} = B*r_{k} ;
c  For  j = k+1, ..., k+np  Do
c     1) if ( betaj < tol ) stop or restart depending on j.
c        if ( restart ) generate a new starting vector.
c     2) v_{j} = r(j-1)/betaj;  V_{j} = [V_{j-1}, v_{j}];
c        p_{j} = p_{j}/betaj
c     3) r_{j} = OP*v_{j} where OP is defined as in dsaupd
c        For shift-invert mode p_{j} = B*v_{j} is already available.
c        wnorm = || OP*v_{j} ||
c     4) Compute the j-th step residual vector.
c        w_{j} =  V_{j}^T * B * OP * v_{j}
c        r_{j} =  OP*v_{j} - V_{j} * w_{j}
c        alphaj <- j-th component of w_{j}
c        rnorm = || r_{j} ||
c        betaj+1 = rnorm
c        If (rnorm > 0.717*wnorm) accept step and go back to 1)
c     5) Re-orthogonalization step:
c        s = V_{j}'*B*r_{j}
c        r_{j} = r_{j} - V_{j}*s;  rnorm1 = || r_{j} ||
c        alphaj = alphaj + s_{j};
c     6) Iterative refinement step:
c        If (rnorm1 > 0.717*rnorm) then
c           rnorm = rnorm1
c           accept step and go back to 1)
c        Else
c           rnorm = rnorm1
c           If this is the first time in step 6), go to 5)
c           Else r_{j} lies in the span of V_{j} numerically.
c              Set r_{j} = 0 and rnorm = 0; go to 1)
c        EndIf
c  End Do
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsaitr
     &   (ido, bmat, n, k, np, mode, resid, rnorm, v, ldv, h, ldh,
     &     ipntr, workd, info)


      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1
      integer    ido, info, k, ldh, ldv, n, mode, np
      Double precision
     &           rnorm
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    ipntr(3)
      Double precision
     &           h(ldh,2), resid(n), v(ldv,k+np), workd(3*n)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0, zero = 0.0D+0)
      logical fls
      parameter (fls = .false.)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      logical    first, orth1, orth2, rstart, step3, step4
      integer    i, ierr, ipj, irj, ivj, iter, itry, j,
     &           infol, jj
      Double precision
     &           rnorm1, wnorm, safmin, temp1
      save       orth1, orth2, rstart, step3, step4,
     &           ierr, ipj, irj, ivj, iter, itry, j,
     &           rnorm1, safmin, wnorm
c
c     %-----------------------%
c     | Local Array Arguments |
c     %-----------------------%
c
cs      Double precision
cs     &           xtemp(2)
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   daxpy, dcopy, dscal, dgemv, dgetv0,
     &           dlascl
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           ddot, dnrm2, dlamch
      external   ddot, dnrm2, dlamch
c
c     %-----------------%
c     | Data statements |
c     %-----------------%
c
      data      first / .true. /
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (first) then
         first = .false.
c
c        %--------------------------------%
c        | safmin = safe minimum is such  |
c        | that 1/sfmin does not overflow |
c        %--------------------------------%
c
         safmin = dlamch('safmin')
      end if
c
      if (ido .eq. 0) then
c
c        %-------------------------------%
c        | Initialize timing statistics  |
c        | & message level for debugging |
c        %-------------------------------%
c
c         call second (t0)
c
c        %------------------------------%
c        | Initial call to this routine |
c        %------------------------------%
c
         info   = 0
         step3  = .false.
         step4  = .false.
         rstart = .false.
         orth1  = .false.
         orth2  = .false.
c
c        %--------------------------------%
c        | Pointer to the current step of |
c        | the factorization to build     |
c        %--------------------------------%
c
         j      = k + 1
c
c        %------------------------------------------%
c        | Pointers used for reverse communication  |
c        | when using WORKD.                        |
c        %------------------------------------------%
c
         ipj    = 1
         irj    = ipj   + n
         ivj    = irj   + n
      end if
c
c     %-------------------------------------------------%
c     | When in reverse communication mode one of:      |
c     | STEP3, STEP4, ORTH1, ORTH2, RSTART              |
c     | will be .true.                                  |
c     | STEP3: return from computing OP*v_{j}.          |
c     | STEP4: return from computing B-norm of OP*v_{j} |
c     | ORTH1: return from computing B-norm of r_{j+1}  |
c     | ORTH2: return from computing B-norm of          |
c     |        correction to the residual vector.       |
c     | RSTART: return from OP computations needed by   |
c     |         dgetv0.                                 |
c     %-------------------------------------------------%
c
      if (step3)  go to 50
      if (step4)  go to 60
      if (orth1)  go to 70
      if (orth2)  go to 90
      if (rstart) go to 30
c
c     %------------------------------%
c     | Else this is the first step. |
c     %------------------------------%
c
c     %--------------------------------------------------------------%
c     |                                                              |
c     |        A R N O L D I     I T E R A T I O N     L O O P       |
c     |                                                              |
c     | Note:  B*r_{j-1} is already in WORKD(1:N)=WORKD(IPJ:IPJ+N-1) |
c     %--------------------------------------------------------------%
c
 1000 continue
c
c        %---------------------------------------------------------%
c        | Check for exact zero. Equivalent to determing whether a |
c        | j-step Arnoldi factorization is present.                |
c        %---------------------------------------------------------%
c
         if (rnorm .gt. zero) go to 40
c
c           %---------------------------------------------------%
c           | Invariant subspace found, generate a new starting |
c           | vector which is orthogonal to the current Arnoldi |
c           | basis and continue the iteration.                 |
c           %---------------------------------------------------%
c
c           %---------------------------------------------%
c           | ITRY is the loop variable that controls the |
c           | maximum amount of times that a restart is   |
c           | attempted. NRSTRT is used by stat.h         |
c           %---------------------------------------------%
c
            itry   = 1
   20       continue
            rstart = .true.
            ido    = 0
   30       continue
c
c           %--------------------------------------%
c           | If in reverse communication mode and |
c           | RSTART = .true. flow returns here.   |
c           %--------------------------------------%
c
            call dgetv0 (ido, bmat, fls, n, j, v, ldv,
     &                   resid, rnorm, ipntr, workd, ierr)
            if (ido .ne. 99) go to 9000
            if (ierr .lt. 0) then
               itry = itry + 1
               if (itry .le. 3) go to 20
c
c              %------------------------------------------------%
c              | Give up after several restart attempts.        |
c              | Set INFO to the size of the invariant subspace |
c              | which spans OP and exit.                       |
c              %------------------------------------------------%
c
               info = j - 1
               ido = 99
               go to 9000
            end if
c
   40    continue
c
c        %---------------------------------------------------------%
c        | STEP 2:  v_{j} = r_{j-1}/rnorm and p_{j} = p_{j}/rnorm  |
c        | Note that p_{j} = B*r_{j-1}. In order to avoid overflow |
c        | when reciprocating a small RNORM, test against lower    |
c        | machine bound.                                          |
c        %---------------------------------------------------------%
c
         call dcopy (n, resid, 1, v(1,j), 1)
         if (rnorm .ge. safmin) then
             temp1 = one / rnorm
             call dscal (n, temp1, v(1,j), 1)
             call dscal (n, temp1, workd(ipj), 1)
         else
c
c            %-----------------------------------------%
c            | To scale both v_{j} and p_{j} carefully |
c            | use LAPACK routine SLASCL               |
c            %-----------------------------------------%
c
             call dlascl ('General', i, i, rnorm, one, n, 1,
     &                    v(1,j), n, infol)
             call dlascl ('General', i, i, rnorm, one, n, 1,
     &                    workd(ipj), n, infol)
         end if
c
c        %------------------------------------------------------%
c        | STEP 3:  r_{j} = OP*v_{j}; Note that p_{j} = B*v_{j} |
c        | Note that this is not quite yet r_{j}. See STEP 4    |
c        %------------------------------------------------------%
c
         step3 = .true.
         call dcopy (n, v(1,j), 1, workd(ivj), 1)
         ipntr(1) = ivj
         ipntr(2) = irj
         ipntr(3) = ipj
         ido = 1
c
c        %-----------------------------------%
c        | Exit in order to compute OP*v_{j} |
c        %-----------------------------------%
c
         go to 9000
   50    continue
c
c        %-----------------------------------%
c        | Back from reverse communication;  |
c        | WORKD(IRJ:IRJ+N-1) := OP*v_{j}.   |
c        %-----------------------------------%
c
         step3 = .false.
c
c        %------------------------------------------%
c        | Put another copy of OP*v_{j} into RESID. |
c        %------------------------------------------%
c
         call dcopy (n, workd(irj), 1, resid(1), 1)
c
c        %-------------------------------------------%
c        | STEP 4:  Finish extending the symmetric   |
c        |          Arnoldi to length j. If MODE = 2 |
c        |          then B*OP = B*inv(B)*A = A and   |
c        |          we don't need to compute B*OP.   |
c        | NOTE: If MODE = 2 WORKD(IVJ:IVJ+N-1) is   |
c        | assumed to have A*v_{j}.                  |
c        %-------------------------------------------%
c
         if (mode .eq. 2) go to 65
c         call second (t2)
         if (bmat .eq. 'G') then
            step4 = .true.
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %-------------------------------------%
c           | Exit in order to compute B*OP*v_{j} |
c           %-------------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
              call dcopy(n, resid, 1 , workd(ipj), 1)
         end if
   60    continue
c
c        %-----------------------------------%
c        | Back from reverse communication;  |
c        | WORKD(IPJ:IPJ+N-1) := B*OP*v_{j}. |
c        %-----------------------------------%
c
         step4 = .false.
c
c        %-------------------------------------%
c        | The following is needed for STEP 5. |
c        | Compute the B-norm of OP*v_{j}.     |
c        %-------------------------------------%
c
   65    continue
         if (mode .eq. 2) then
c
c           %----------------------------------%
c           | Note that the B-norm of OP*v_{j} |
c           | is the inv(B)-norm of A*v_{j}.   |
c           %----------------------------------%
c
            wnorm = ddot (n, resid, 1, workd(ivj), 1)
            wnorm = sqrt(abs(wnorm))
         else if (bmat .eq. 'G') then
            wnorm = ddot (n, resid, 1, workd(ipj), 1)
            wnorm = sqrt(abs(wnorm))
         else if (bmat .eq. 'I') then
            wnorm = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------%
c        | Compute the j-th residual corresponding |
c        | to the j step factorization.            |
c        | Use Classical Gram Schmidt and compute: |
c        | w_{j} <-  V_{j}^T * B * OP * v_{j}      |
c        | r_{j} <-  OP*v_{j} - V_{j} * w_{j}      |
c        %-----------------------------------------%
c
c
c        %------------------------------------------%
c        | Compute the j Fourier coefficients w_{j} |
c        | WORKD(IPJ:IPJ+N-1) contains B*OP*v_{j}.  |
c        %------------------------------------------%
c
         if (mode .ne. 2 ) then
            call dgemv('T', n, j, one, v, ldv, workd(ipj), 1, zero,
     &                  workd(irj), 1)
         else if (mode .eq. 2) then
            call dgemv('T', n, j, one, v, ldv, workd(ivj), 1, zero,
     &                  workd(irj), 1)
         end if
c
c        %--------------------------------------%
c        | Orthgonalize r_{j} against V_{j}.    |
c        | RESID contains OP*v_{j}. See STEP 3. |
c        %--------------------------------------%
c
         call dgemv('N', n, j, -one, v, ldv, workd(irj), 1, one,
     &               resid(1), 1)
c
c        %--------------------------------------%
c        | Extend H to have j rows and columns. |
c        %--------------------------------------%
c
         h(j,2) = workd(irj + j - 1)
         if (j .eq. 1  .or.  rstart) then
            h(j,1) = zero
         else
            h(j,1) = rnorm
         end if
c         call second (t4)
c
         orth1 = .true.
         iter  = 0
c
c         call second (t2)
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(irj), 1)
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %----------------------------------%
c           | Exit in order to compute B*r_{j} |
c           %----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(ipj), 1)
         end if
   70    continue
c
c        %---------------------------------------------------%
c        | Back from reverse communication if ORTH1 = .true. |
c        | WORKD(IPJ:IPJ+N-1) := B*r_{j}.                    |
c        %---------------------------------------------------%
c
         orth1 = .false.
c
c        %------------------------------%
c        | Compute the B-norm of r_{j}. |
c        %------------------------------%
c
         if (bmat .eq. 'G') then
            rnorm = ddot (n, resid, 1, workd(ipj), 1)
            rnorm = sqrt(abs(rnorm))
         else if (bmat .eq. 'I') then
            rnorm = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------------------------%
c        | STEP 5: Re-orthogonalization / Iterative refinement phase |
c        | Maximum NITER_ITREF tries.                                |
c        |                                                           |
c        |          s      = V_{j}^T * B * r_{j}                     |
c        |          r_{j}  = r_{j} - V_{j}*s                         |
c        |          alphaj = alphaj + s_{j}                          |
c        |                                                           |
c        | The stopping criteria used for iterative refinement is    |
c        | discussed in Parlett's book SEP, page 107 and in Gragg &  |
c        | Reichel ACM TOMS paper; Algorithm 686, Dec. 1990.         |
c        | Determine if we need to correct the residual. The goal is |
c        | to enforce ||v(:,1:j)^T * r_{j}|| .le. eps * || r_{j} ||  |
c        %-----------------------------------------------------------%
c
         if (rnorm .gt. 0.717*wnorm) go to 100
c
c        %---------------------------------------------------%
c        | Enter the Iterative refinement phase. If further  |
c        | refinement is necessary, loop back here. The loop |
c        | variable is ITER. Perform a step of Classical     |
c        | Gram-Schmidt using all the Arnoldi vectors V_{j}  |
c        %---------------------------------------------------%
c
   80    continue
c
c        %----------------------------------------------------%
c        | Compute V_{j}^T * B * r_{j}.                       |
c        | WORKD(IRJ:IRJ+J-1) = v(:,1:J)'*WORKD(IPJ:IPJ+N-1). |
c        %----------------------------------------------------%
c
         call dgemv ('T', n, j, one, v, ldv, workd(ipj), 1,
     &               zero, workd(irj), 1)
c
c        %----------------------------------------------%
c        | Compute the correction to the residual:      |
c        | r_{j} = r_{j} - V_{j} * WORKD(IRJ:IRJ+J-1).  |
c        | The correction to H is v(:,1:J)*H(1:J,1:J) + |
c        | v(:,1:J)*WORKD(IRJ:IRJ+J-1)*e'_j, but only   |
c        | H(j,j) is updated.                           |
c        %----------------------------------------------%
c
         call dgemv ('N', n, j, -one, v, ldv, workd(irj), 1,
     &               one, resid(1), 1)
c
         if (j .eq. 1  .or.  rstart) h(j,1) = zero
         h(j,2) = h(j,2) + workd(irj + j - 1)
c
         orth2 = .true.
c        call second (t2)
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(irj), 1)
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %-----------------------------------%
c           | Exit in order to compute B*r_{j}. |
c           | r_{j} is the corrected residual.  |
c           %-----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(ipj), 1)
         end if
   90    continue
c
c        %---------------------------------------------------%
c        | Back from reverse communication if ORTH2 = .true. |
c        %---------------------------------------------------%
c
c        %-----------------------------------------------------%
c        | Compute the B-norm of the corrected residual r_{j}. |
c        %-----------------------------------------------------%
c
         if (bmat .eq. 'G') then
             rnorm1 = ddot (n, resid, 1, workd(ipj), 1)
             rnorm1 = sqrt(abs(rnorm1))
         else if (bmat .eq. 'I') then
             rnorm1 = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------%
c        | Determine if we need to perform another |
c        | step of re-orthogonalization.           |
c        %-----------------------------------------%
c
         if (rnorm1 .gt. 0.717*rnorm) then
c
c           %--------------------------------%
c           | No need for further refinement |
c           %--------------------------------%
c
            rnorm = rnorm1
c
         else
c
c           %-------------------------------------------%
c           | Another step of iterative refinement step |
c           | is required. NITREF is used by stat.h     |
c           %-------------------------------------------%
c
            rnorm  = rnorm1
            iter   = iter + 1
            if (iter .le. 1) go to 80
c
c           %-------------------------------------------------%
c           | Otherwise RESID is numerically in the span of V |
c           %-------------------------------------------------%
c
            do 95 jj = 1, n
               resid(jj) = zero
  95        continue
            rnorm = zero
         end if
c
c        %----------------------------------------------%
c        | Branch here directly if iterative refinement |
c        | wasn't necessary or after at most NITER_REF  |
c        | steps of iterative refinement.               |
c        %----------------------------------------------%
c
  100    continue
c
         rstart = .false.
         orth2  = .false.
c
c        %----------------------------------------------------------%
c        | Make sure the last off-diagonal element is non negative  |
c        | If not perform a similarity transformation on H(1:j,1:j) |
c        | and scale v(:,j) by -1.                                  |
c        %----------------------------------------------------------%
c
         if (h(j,1) .lt. zero) then
            h(j,1) = -h(j,1)
            if ( j .lt. k+np) then
               call dscal(n, -one, v(1,j+1), 1)
            else
               call dscal(n, -one, resid(1), 1)
            end if
         end if
c
c        %------------------------------------%
c        | STEP 6: Update  j = j+1;  Continue |
c        %------------------------------------%
c
         j = j + 1
         if (j .gt. k+np) then
            ido = 99
c
            go to 9000
         end if
c
c        %--------------------------------------------------------%
c        | Loop back to extend the factorization by another step. |
c        %--------------------------------------------------------%
c
      go to 1000
c
c     %---------------------------------------------------------------%
c     |                                                               |
c     |  E N D     O F     M A I N     I T E R A T I O N     L O O P  |
c     |                                                               |
c     %---------------------------------------------------------------%
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsaitr |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsapps
c
c\Description:
c  Given the Arnoldi factorization
c
c     A*V_{k} - V_{k}*H_{k} = r_{k+p}*e_{k+p}^T,
c
c  apply NP shifts implicitly resulting in
c
c     A*(V_{k}*Q) - (V_{k}*Q)*(Q^T* H_{k}*Q) = r_{k+p}*e_{k+p}^T * Q
c
c  where Q is an orthogonal matrix of order KEV+NP. Q is the product of
c  rotations resulting from the NP bulge chasing sweeps.  The updated Arnoldi
c  factorization becomes:
c
c     A*VNEW_{k} - VNEW_{k}*HNEW_{k} = rnew_{k}*e_{k}^T.
c
c\Usage:
c  call dsapps
c     ( N, KEV, NP, SHIFT, V, LDV, H, LDH, RESID, Q, LDQ, WORKD )
c
c\Arguments
c  N       Integer.  (INPUT)
c          Problem size, i.e. dimension of matrix A.
c
c  KEV     Integer.  (INPUT)
c          INPUT: KEV+NP is the size of the input matrix H.
c          OUTPUT: KEV is the size of the updated matrix HNEW.
c
c  NP      Integer.  (INPUT)
c          Number of implicit shifts to be applied.
c
c  SHIFT   Double precision array of length NP.  (INPUT)
c          The shifts to be applied.
c
c  V       Double precision N by (KEV+NP) array.  (INPUT/OUTPUT)
c          INPUT: V contains the current KEV+NP Arnoldi vectors.
c          OUTPUT: VNEW = V(1:n,1:KEV); the updated Arnoldi vectors
c          are in the first KEV columns of V.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (KEV+NP) by 2 array.  (INPUT/OUTPUT)
c          INPUT: H contains the symmetric tridiagonal matrix of the
c          Arnoldi factorization with the subdiagonal in the 1st column
c          starting at H(2,1) and the main diagonal in the 2nd column.
c          OUTPUT: H contains the updated tridiagonal matrix in the
c          KEV leading submatrix.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  RESID   Double precision array of length (N).  (INPUT/OUTPUT)
c          INPUT: RESID contains the the residual vector r_{k+p}.
c          OUTPUT: RESID is the updated residual vector rnew_{k}.
c
c  Q       Double precision KEV+NP by KEV+NP work array.  (WORKSPACE)
c          Work array used to accumulate the rotations during the bulge
c          chase sweep.
c
c  LDQ     Integer.  (INPUT)
c          Leading dimension of Q exactly as declared in the calling
c          program.
c
c  WORKD   Double precision work array of length 2*N.  (WORKSPACE)
c          Distributed array used in the application of the accumulated
c          orthogonal matrix Q.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c
c\Routines called:
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dvout   ARPACK utility routine that prints vectors.
c     dlamch  LAPACK routine that determines machine constants.
c     dlartg  LAPACK Givens rotation construction routine.
c     dlacpy  LAPACK matrix copy routine.
c     dlaset  LAPACK matrix initialization routine.
c     dgemv   Level 2 BLAS routine for matrix vector multiplication.
c     daxpy   Level 1 BLAS that computes a vector triad.
c     dcopy   Level 1 BLAS that copies one vector to another.
c     dscal   Level 1 BLAS that scales a vector.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/16/93: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: sapps.F   SID: 2.6   DATE OF SID: 3/28/97   RELEASE: 2
c
c\Remarks
c  1. In this version, each shift is applied to all the subblocks of
c     the tridiagonal matrix H and not just to the submatrix that it
c     comes from. This routine assumes that the subdiagonal elements
c     of H that are stored in h(1:kev+np,1) are nonegative upon input
c     and enforce this condition upon output. This version incorporates
c     deflation. See code for documentation.
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsapps
     &   ( n, kev, np, shift, v, ldv, h, ldh, resid, q, ldq, workd )
c
c     %----------------------------------------------------%
c     | Include files for debugging and timing information |
c     %----------------------------------------------------%
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    kev, ldh, ldq, ldv, n, np
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           h(ldh,2), q(ldq,kev+np), resid(n), shift(np),
     &           v(ldv,kev+np), workd(2*n)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0, zero = 0.0D+0)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i, iend, istart, itop, j, jj, kplusp
      logical    first
      Double precision
     &           a1, a2, a3, a4, big, c, epsmch, f, g, r, s
      save       epsmch, first
c
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   daxpy, dcopy, dscal, dlacpy, dlartg, dlaset,
     &           dgemv
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlamch
      external   dlamch
c
c     %----------------------%
c     | Intrinsics Functions |
c     %----------------------%
c
      intrinsic  abs
c
c     %----------------%
c     | Data statments |
c     %----------------%
c
      data       first / .true. /
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (first) then
         epsmch = dlamch('Epsilon-Machine')
         first = .false.
      end if
      itop = 1
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
      kplusp = kev + np
c
c     %----------------------------------------------%
c     | Initialize Q to the identity matrix of order |
c     | kplusp used to accumulate the rotations.     |
c     %----------------------------------------------%
c
      call dlaset ('All', kplusp, kplusp, zero, one, q, ldq)
c
c     %----------------------------------------------%
c     | Quick return if there are no shifts to apply |
c     %----------------------------------------------%
c
      if (np .eq. 0) go to 9000
c
c     %----------------------------------------------------------%
c     | Apply the np shifts implicitly. Apply each shift to the  |
c     | whole matrix and not just to the submatrix from which it |
c     | comes.                                                   |
c     %----------------------------------------------------------%
c
      do 90 jj = 1, np
c
         istart = itop
c
c        %----------------------------------------------------------%
c        | Check for splitting and deflation. Currently we consider |
c        | an off-diagonal element h(i+1,1) negligible if           |
c        |         h(i+1,1) .le. epsmch*( |h(i,2)| + |h(i+1,2)| )   |
c        | for i=1:KEV+NP-1.                                        |
c        | If above condition tests true then we set h(i+1,1) = 0.  |
c        | Note that h(1:KEV+NP,1) are assumed to be non negative.  |
c        %----------------------------------------------------------%
c
   20    continue
c
c        %------------------------------------------------%
c        | The following loop exits early if we encounter |
c        | a negligible off diagonal element.             |
c        %------------------------------------------------%
c
         do 30 i = istart, kplusp-1
            big   = abs(h(i,2)) + abs(h(i+1,2))
            if (h(i+1,1) .le. epsmch*big) then
               h(i+1,1) = zero
               iend = i
               go to 40
            end if
   30    continue
         iend = kplusp
   40    continue
c
         if (istart .lt. iend) then
c
c           %--------------------------------------------------------%
c           | Construct the plane rotation G'(istart,istart+1,theta) |
c           | that attempts to drive h(istart+1,1) to zero.          |
c           %--------------------------------------------------------%
c
             f = h(istart,2) - shift(jj)
             g = h(istart+1,1)
             call dlartg (f, g, c, s, r)
c
c            %-------------------------------------------------------%
c            | Apply rotation to the left and right of H;            |
c            | H <- G' * H * G,  where G = G(istart,istart+1,theta). |
c            | This will create a "bulge".                           |
c            %-------------------------------------------------------%
c
             a1 = c*h(istart,2)   + s*h(istart+1,1)
             a2 = c*h(istart+1,1) + s*h(istart+1,2)
             a4 = c*h(istart+1,2) - s*h(istart+1,1)
             a3 = c*h(istart+1,1) - s*h(istart,2)
             h(istart,2)   = c*a1 + s*a2
             h(istart+1,2) = c*a4 - s*a3
             h(istart+1,1) = c*a3 + s*a4
c
c            %----------------------------------------------------%
c            | Accumulate the rotation in the matrix Q;  Q <- Q*G |
c            %----------------------------------------------------%
c
             do 60 j = 1, min(istart+jj,kplusp)
                a1            =   c*q(j,istart) + s*q(j,istart+1)
                q(j,istart+1) = - s*q(j,istart) + c*q(j,istart+1)
                q(j,istart)   = a1
   60        continue
c
c
c            %----------------------------------------------%
c            | The following loop chases the bulge created. |
c            | Note that the previous rotation may also be  |
c            | done within the following loop. But it is    |
c            | kept separate to make the distinction among  |
c            | the bulge chasing sweeps and the first plane |
c            | rotation designed to drive h(istart+1,1) to  |
c            | zero.                                        |
c            %----------------------------------------------%
c
             do 70 i = istart+1, iend-1
c
c               %----------------------------------------------%
c               | Construct the plane rotation G'(i,i+1,theta) |
c               | that zeros the i-th bulge that was created   |
c               | by G(i-1,i,theta). g represents the bulge.   |
c               %----------------------------------------------%
c
                f = h(i,1)
                g = s*h(i+1,1)
c
c               %----------------------------------%
c               | Final update with G(i-1,i,theta) |
c               %----------------------------------%
c
                h(i+1,1) = c*h(i+1,1)
                call dlartg (f, g, c, s, r)
c
c               %-------------------------------------------%
c               | The following ensures that h(1:iend-1,1), |
c               | the first iend-2 off diagonal of elements |
c               | H, remain non negative.                   |
c               %-------------------------------------------%
c
                if (r .lt. zero) then
                   r = -r
                   c = -c
                   s = -s
                end if
c
c               %--------------------------------------------%
c               | Apply rotation to the left and right of H; |
c               | H <- G * H * G',  where G = G(i,i+1,theta) |
c               %--------------------------------------------%
c
                h(i,1) = r
c
                a1 = c*h(i,2)   + s*h(i+1,1)
                a2 = c*h(i+1,1) + s*h(i+1,2)
                a3 = c*h(i+1,1) - s*h(i,2)
                a4 = c*h(i+1,2) - s*h(i+1,1)
c
                h(i,2)   = c*a1 + s*a2
                h(i+1,2) = c*a4 - s*a3
                h(i+1,1) = c*a3 + s*a4
c
c               %----------------------------------------------------%
c               | Accumulate the rotation in the matrix Q;  Q <- Q*G |
c               %----------------------------------------------------%
c
                do 50 j = 1, min( i+jj, kplusp )
                   a1       =   c*q(j,i) + s*q(j,i+1)
                   q(j,i+1) = - s*q(j,i) + c*q(j,i+1)
                   q(j,i)   = a1
   50           continue
c
   70        continue
c
         end if
c
c        %--------------------------%
c        | Update the block pointer |
c        %--------------------------%
c
         istart = iend + 1
c
c        %------------------------------------------%
c        | Make sure that h(iend,1) is non-negative |
c        | If not then set h(iend,1) <-- -h(iend,1) |
c        | and negate the last column of Q.         |
c        | We have effectively carried out a        |
c        | similarity on transformation H           |
c        %------------------------------------------%
c
         if (h(iend,1) .lt. zero) then
             h(iend,1) = -h(iend,1)
             call dscal(kplusp, -one, q(1,iend), 1)
         end if
c
c        %--------------------------------------------------------%
c        | Apply the same shift to the next block if there is any |
c        %--------------------------------------------------------%
c
         if (iend .lt. kplusp) go to 20
c
c        %-----------------------------------------------------%
c        | Check if we can increase the the start of the block |
c        %-----------------------------------------------------%
c
         do 80 i = itop, kplusp-1
            if (h(i+1,1) .gt. zero) go to 90
            itop  = itop + 1
   80    continue
c
c        %-----------------------------------%
c        | Finished applying the jj-th shift |
c        %-----------------------------------%
c
   90 continue
c
c     %------------------------------------------%
c     | All shifts have been applied. Check for  |
c     | more possible deflation that might occur |
c     | after the last shift is applied.         |
c     %------------------------------------------%
c
      do 100 i = itop, kplusp-1
         big   = abs(h(i,2)) + abs(h(i+1,2))
         if (h(i+1,1) .le. epsmch*big) then
            h(i+1,1) = zero
         end if
 100  continue
c
c     %-------------------------------------------------%
c     | Compute the (kev+1)-st column of (V*Q) and      |
c     | temporarily store the result in WORKD(N+1:2*N). |
c     | This is not necessary if h(kev+1,1) = 0.         |
c     %-------------------------------------------------%
c
      if ( h(kev+1,1) .gt. zero )
     &   call dgemv ('N', n, kplusp, one, v, ldv,
     &                q(1,kev+1), 1, zero, workd(n+1), 1)
c
c     %-------------------------------------------------------%
c     | Compute column 1 to kev of (V*Q) in backward order    |
c     | taking advantage that Q is an upper triangular matrix |
c     | with lower bandwidth np.                              |
c     | Place results in v(:,kplusp-kev:kplusp) temporarily.  |
c     %-------------------------------------------------------%
c
      do 130 i = 1, kev
         call dgemv ('N', n, kplusp-i+1, one, v, ldv,
     &               q(1,kev-i+1), 1, zero, workd(1), 1)
         call dcopy (n, workd, 1, v(1,kplusp-i+1), 1)
  130 continue
c
c     %-------------------------------------------------%
c     |  Move v(:,kplusp-kev+1:kplusp) into v(:,1:kev). |
c     %-------------------------------------------------%
c
      call dlacpy ('All', n, kev, v(1,np+1), ldv, v, ldv)
c
c     %--------------------------------------------%
c     | Copy the (kev+1)-st column of (V*Q) in the |
c     | appropriate place if h(kev+1,1) .ne. zero. |
c     %--------------------------------------------%
c
      if ( h(kev+1,1) .gt. zero )
     &     call dcopy (n, workd(n+1), 1, v(1,kev+1), 1)
c
c     %-------------------------------------%
c     | Update the residual vector:         |
c     |    r <- sigmak*r + betak*v(:,kev+1) |
c     | where                               |
c     |    sigmak = (e_{kev+p}'*Q)*e_{kev}  |
c     |    betak = e_{kev+1}'*H*e_{kev}     |
c     %-------------------------------------%
c
      call dscal (n, q(kplusp,kev), resid(1), 1)
c
      if (h(kev+1,1) .gt. zero)
     &   call daxpy (n, h(kev+1,1), v(1,kev+1), 1, resid, 1)
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsapps |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dseigt
c
c\Description:
c  Compute the eigenvalues of the current symmetric tridiagonal matrix
c  and the corresponding error bounds given the current residual norm.
c
c\Usage:
c  call dseigt
c     ( RNORM, N, H, LDH, EIG, BOUNDS, WORKL, IERR )
c
c\Arguments
c  RNORM   Double precision scalar.  (INPUT)
c          RNORM contains the residual norm corresponding to the current
c          symmetric tridiagonal matrix H.
c
c  N       Integer.  (INPUT)
c          Size of the symmetric tridiagonal matrix H.
c
c  H       Double precision N by 2 array.  (INPUT)
c          H contains the symmetric tridiagonal matrix with the
c          subdiagonal in the first column starting at H(2,1) and the
c          main diagonal in second column.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  EIG     Double precision array of length N.  (OUTPUT)
c          On output, EIG contains the N eigenvalues of H possibly
c          unsorted.  The BOUNDS arrays are returned in the
c          same sorted order as EIG.
c
c  BOUNDS  Double precision array of length N.  (OUTPUT)
c          On output, BOUNDS contains the error estimates corresponding
c          to the eigenvalues EIG.  This is equal to RNORM times the
c          last components of the eigenvectors corresponding to the
c          eigenvalues in EIG.
c
c  WORKL   Double precision work array of length 3*N.  (WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.
c
c  IERR    Integer.  (OUTPUT)
c          Error exit flag from dstqrb.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dstqrb  ARPACK routine that computes the eigenvalues and the
c             last components of the eigenvectors of a symmetric
c             and tridiagonal matrix.
c     second  ARPACK utility routine for timing.
c     dvout   ARPACK utility routine that prints vectors.
c     dcopy   Level 1 BLAS that copies one vector to another.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: seigt.F   SID: 2.4   DATE OF SID: 8/27/96   RELEASE: 2
c
c\Remarks
c     None
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dseigt
     &   ( rnorm, n, h, ldh, eig, bounds, workl, ierr )
c
c     %----------------------------------------------------%
c     | Include files for debugging and timing information |
c     %----------------------------------------------------%
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    ierr, ldh, n
      Double precision
     &           rnorm
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           eig(n), bounds(n), h(ldh,2), workl(3*n)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    k
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy, dstqrb
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
c     call second (t0)
c
c
      call dcopy  (n, h(1,2), 1, eig(1), 1)
      call dcopy  (n-1, h(2,1), 1, workl(1), 1)
      call dstqrb (n, eig, workl, bounds, workl(n+1), ierr)
      if (ierr .ne. 0) go to 9000
c
c     %-----------------------------------------------%
c     | Finally determine the error bounds associated |
c     | with the n Ritz values of H.                  |
c     %-----------------------------------------------%
c
      do 30 k = 1, n
         bounds(k) = rnorm*abs(bounds(k))
   30 continue
c
c     call second (t1)
c
 9000 continue
      return
c
c     %---------------%
c     | End of dseigt |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dstqrb
c
c\Description:
c  Computes all eigenvalues and the last component of the eigenvectors
c  of a symmetric tridiagonal matrix using the implicit QL or QR method.
c
c  This is mostly a modification of the LAPACK routine dsteqr.
c  See Remarks.
c
c\Usage:
c  call dstqrb
c     ( N, D, E, Z, WORK, INFO )
c
c\Arguments
c  N       Integer.  (INPUT)
c          The number of rows and columns in the matrix.  N >= 0.
c
c  D       Double precision array, dimension (N).  (INPUT/OUTPUT)
c          On entry, D contains the diagonal elements of the
c          tridiagonal matrix.
c          On exit, D contains the eigenvalues, in ascending order.
c          If an error exit is made, the eigenvalues are correct
c          for indices 1,2,...,INFO-1, but they are unordered and
c          may not be the smallest eigenvalues of the matrix.
c
c  E       Double precision array, dimension (N-1).  (INPUT/OUTPUT)
c          On entry, E contains the subdiagonal elements of the
c          tridiagonal matrix in positions 1 through N-1.
c          On exit, E has been destroyed.
c
c  Z       Double precision array, dimension (N).  (OUTPUT)
c          On exit, Z contains the last row of the orthonormal
c          eigenvector matrix of the symmetric tridiagonal matrix.
c          If an error exit is made, Z contains the last row of the
c          eigenvector matrix associated with the stored eigenvalues.
c
c  WORK    Double precision array, dimension (max(1,2*N-2)).  (WORKSPACE)
c          Workspace used in accumulating the transformation for
c          computing the last components of the eigenvectors.
c
c  INFO    Integer.  (OUTPUT)
c          = 0:  normal return.
c          < 0:  if INFO = -i, the i-th argument had an illegal value.
c          > 0:  if INFO = +i, the i-th eigenvalue has not converged
c                              after a total of  30*N  iterations.
c
c\Remarks
c  1. None.
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     daxpy   Level 1 BLAS that computes a vector triad.
c     dcopy   Level 1 BLAS that copies one vector to another.
c     dswap   Level 1 BLAS that swaps the contents of two vectors.
c     lsame   LAPACK character comparison routine.
c     dlae2   LAPACK routine that computes the eigenvalues of a 2-by-2
c             symmetric matrix.
c     dlaev2  LAPACK routine that eigendecomposition of a 2-by-2 symmetric
c             matrix.
c     dlamch  LAPACK routine that determines machine constants.
c     dlanst  LAPACK routine that computes the norm of a matrix.
c     dlapy2  LAPACK routine to compute sqrt(x**2+y**2) carefully.
c     dlartg  LAPACK Givens rotation construction routine.
c     dlascl  LAPACK routine for careful scaling of a matrix.
c     dlaset  LAPACK matrix initialization routine.
c     dlasr   LAPACK routine that applies an orthogonal transformation to
c             a matrix.
c     dlasrt  LAPACK sorting routine.
c     dsteqr  LAPACK routine that computes eigenvalues and eigenvectors
c             of a symmetric tridiagonal matrix.
c     xerbla  LAPACK error handler routine.
c
c\Authors
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\SCCS Information: @(#)
c FILE: stqrb.F   SID: 2.5   DATE OF SID: 8/27/96   RELEASE: 2
c
c\Remarks
c     1. Starting with version 2.5, this routine is a modified version
c        of LAPACK version 2.0 subroutine SSTEQR. No lines are deleted,
c        only commeted out and new lines inserted.
c        All lines commented out have "c$$$" at the beginning.
c        Note that the LAPACK version 1.0 subroutine SSTEQR contained
c        bugs.
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dstqrb ( n, d, e, z, work, info )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    info, n
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           d( n ), e( n-1 ), z( n ), work( 2*n-2 )
c
      Double precision
     &                   zero, one, two, three
      parameter          ( zero = 0.0D+0, one = 1.0D+0,
     &                     two = 2.0D+0, three = 3.0D+0 )
c
      integer            izero
      parameter          (izero=0 )
c
      integer            maxit
      parameter          ( maxit = 30 )
c
      integer            i, icompz, ii, iscale, j, jtot, k, l, l1,
     &                   lendm1, lendp1, lendsv, lm1, lsv, m, mm, mm1,
     &                   nm1, nmaxit, lend
      Double precision
     &                   anorm, b, c, eps, eps2, f, g, p, r, rt1, rt2,
     &                   s, safmax, safmin, ssfmax, ssfmin, tst
c     ..
c     .. external functions ..
      logical,  external :: neZERO, eqZERO

      logical            lsame
      Double precision
     &                   dlamch, dlanst, dlapy2
      external           lsame, dlamch, dlanst, dlapy2
c     ..
c     .. external subroutines ..
      external           dlae2, dlaev2, dlartg, dlascl, dlaset, dlasr,
     &                   dlasrt, dswap, xerbla
c     ..
c     .. intrinsic functions ..
      intrinsic          abs, max, sign, sqrt
c     ..
c     .. executable statements ..
c
c     test the input parameters.
c
      info = 0
c
c    *** New starting with version 2.5 ***
c
      icompz = 2
c    *************************************
c
c     quick return if possible
c
      if( n.eq.0 )
     $   return
c
      if( n.eq.1 ) then
         if( icompz.eq.2 )  z( 1 ) = one
         return
      end if
c
c     determine the unit roundoff and over/underflow thresholds.
c
      eps = dlamch( 'e' )
      eps2 = eps**2
      safmin = dlamch( 's' )
      safmax = one / safmin
      ssfmax = sqrt( safmax ) / three
      ssfmin = sqrt( safmin ) / eps2
c
c     compute the eigenvalues and eigenvectors of the tridiagonal
c     matrix.
c
c     *** New starting with version 2.5 ***
c
      if ( icompz .eq. 2 ) then
         do 5 j = 1, n-1
            z(j) = zero
  5      continue
         z( n ) = one
      end if
c     *************************************
c
      nmaxit = n*maxit
      jtot = 0
c
c     determine where the matrix splits and choose ql or qr iteration
c     for each block, according to whether top or bottom diagonal
c     element is smaller.
c
      l1 = 1
      nm1 = n - 1
c
   10 continue
      if( l1.gt.n )
     $   go to 160
      if( l1.gt.1 )
     $   e( l1-1 ) = zero
      if( l1.le.nm1 ) then
         do 20 m = l1, nm1
            tst = abs( e( m ) )
            if (eqZERO( tst))
     $         go to 30
            if( tst.le.( sqrt( abs( d( m ) ) )*sqrt( abs( d( m+
     $          1 ) ) ) )*eps ) then
               e( m ) = zero
               go to 30
            end if
   20    continue
      end if
      m = n
c
   30 continue
      l = l1
      lsv = l
      lend = m
      lendsv = lend
      l1 = m + 1
      if( lend.eq.l )
     $   go to 10
c
c     scale submatrix in rows and columns l to lend
c
      anorm = dlanst( 'i', lend-l+1, d( l ), e( l ) )
      iscale = 0
      if (eqZERO( anorm))
     $   go to 10
      if( anorm.gt.ssfmax ) then
         iscale = 1
         call dlascl( 'g', izero, izero, anorm, ssfmax, lend-l+1, 1,
     $                d( l ), n, info )
         call dlascl( 'g', izero, izero, anorm, ssfmax, lend-l, 1,
     $                 e( l ), n, info )
c
      else if( anorm.lt.ssfmin ) then
         iscale = 2
         call dlascl( 'g', izero, izero, anorm, ssfmin, lend-l+1, 1,
     $                d( l ), n, info )
         call dlascl( 'g', izero, izero, anorm, ssfmin, lend-l, 1,
     $                 e( l ), n, info )
c
      end if
c
c     choose between ql and qr iteration
c
      if( abs( d( lend ) ).lt.abs( d( l ) ) ) then
         lend = lsv
         l = lendsv
      end if
c
      if( lend.gt.l ) then
c
c        ql iteration
c
c        look for small subdiagonal element.
c
   40    continue
         if( l.ne.lend ) then
            lendm1 = lend - 1
            do 50 m = l, lendm1
               tst = abs( e( m ) )**2
               if( tst.le.( eps2*abs( d( m ) ) )*abs( d( m+1 ) )+
     $             safmin )go to 60
   50       continue
         end if
c
         m = lend
c
   60    continue
         if( m.lt.lend )
     $      e( m ) = zero
         p = d( l )
         if( m.eq.l )
     $      go to 80
c
c        if remaining matrix is 2-by-2, use dlae2 or dlaev2
c        to compute its eigensystem.
c
         if( m.eq.l+1 ) then
            if( icompz.gt.0 ) then
               call dlaev2( d( l ), e( l ), d( l+1 ), rt1, rt2, c, s )
               work( l ) = c
               work( n-1+l ) = s
c
c              *** New starting with version 2.5 ***
c
               tst      = z(l+1)
               z(l+1) = c*tst - s*z(l)
               z(l)   = s*tst + c*z(l)
c              *************************************
            else
               call dlae2( d( l ), e( l ), d( l+1 ), rt1, rt2 )
            end if
            d( l ) = rt1
            d( l+1 ) = rt2
            e( l ) = zero
            l = l + 2
            if( l.le.lend )
     $         go to 40
            go to 140
         end if
c
         if( jtot.eq.nmaxit )
     $      go to 140
         jtot = jtot + 1
c
c        form shift.
c
         g = ( d( l+1 )-p ) / ( two*e( l ) )
         r = dlapy2( g, one )
         g = d( m ) - p + ( e( l ) / ( g+sign( r, g ) ) )
c
         s = one
         c = one
         p = zero
c
c        inner loop
c
         mm1 = m - 1
         do 70 i = mm1, l, -1
            f = s*e( i )
            b = c*e( i )
            call dlartg( g, f, c, s, r )
            if( i.ne.m-1 )
     $         e( i+1 ) = r
            g = d( i+1 ) - p
            r = ( d( i )-g )*s + two*c*b
            p = s*r
            d( i+1 ) = g + p
            g = c*r - b
c
c           if eigenvectors are desired, then save rotations.
c
            if( icompz.gt.0 ) then
               work( i ) = c
               work( n-1+i ) = -s
            end if
c
   70    continue
c
c        if eigenvectors are desired, then apply saved rotations.
c
         if( icompz.gt.0 ) then
            mm = m - l + 1
c
c             *** New starting with version 2.5 ***
c
              call dlasr( 'r', 'v', 'b', 1, mm, work( l ),
     &                    work( n-1+l ), z( l ), 1 )
c             *************************************
         end if
c
         d( l ) = d( l ) - p
         e( l ) = g
         go to 40
c
c        eigenvalue found.
c
   80    continue
         d( l ) = p
c
         l = l + 1
         if( l.le.lend )
     $      go to 40
         go to 140
c
      else
c
c        qr iteration
c
c        look for small superdiagonal element.
c
   90    continue
         if( l.ne.lend ) then
            lendp1 = lend + 1
            do 100 m = l, lendp1, -1
               tst = abs( e( m-1 ) )**2
               if( tst.le.( eps2*abs( d( m ) ) )*abs( d( m-1 ) )+
     $             safmin )go to 110
  100       continue
         end if
c
         m = lend
c
  110    continue
         if( m.gt.lend )
     $      e( m-1 ) = zero
         p = d( l )
         if( m.eq.l )
     $      go to 130
c
c        if remaining matrix is 2-by-2, use dlae2 or dlaev2
c        to compute its eigensystem.
c
         if( m.eq.l-1 ) then
            if( icompz.gt.0 ) then
               call dlaev2( d( l-1 ), e( l-1 ), d( l ), rt1, rt2, c, s )
c
c               *** New starting with version 2.5 ***
c
                tst      = z(l)
                z(l)   = c*tst - s*z(l-1)
                z(l-1) = s*tst + c*z(l-1)
c               *************************************
            else
               call dlae2( d( l-1 ), e( l-1 ), d( l ), rt1, rt2 )
            end if
            d( l-1 ) = rt1
            d( l ) = rt2
            e( l-1 ) = zero
            l = l - 2
            if( l.ge.lend )
     $         go to 90
            go to 140
         end if
c
         if( jtot.eq.nmaxit )
     $      go to 140
         jtot = jtot + 1
c
c        form shift.
c
         g = ( d( l-1 )-p ) / ( two*e( l-1 ) )
         r = dlapy2( g, one )
         g = d( m ) - p + ( e( l-1 ) / ( g+sign( r, g ) ) )
c
         s = one
         c = one
         p = zero
c
c        inner loop
c
         lm1 = l - 1
         do 120 i = m, lm1
            f = s*e( i )
            b = c*e( i )
            call dlartg( g, f, c, s, r )
            if( i.ne.m )
     $         e( i-1 ) = r
            g = d( i ) - p
            r = ( d( i+1 )-g )*s + two*c*b
            p = s*r
            d( i ) = g + p
            g = c*r - b
c
c           if eigenvectors are desired, then save rotations.
c
            if( icompz.gt.0 ) then
               work( i ) = c
               work( n-1+i ) = s
            end if
c
  120    continue
c
c        if eigenvectors are desired, then apply saved rotations.
c
         if( icompz.gt.0 ) then
            mm = l - m + 1
c
c           *** New starting with version 2.5 ***
c
            call dlasr( 'r', 'v', 'f', 1, mm, work( m ), work( n-1+m ),
     &                  z( m ), 1 )
c           *************************************
         end if
c
         d( l ) = d( l ) - p
         e( lm1 ) = g
         go to 90
c
c        eigenvalue found.
c
  130    continue
         d( l ) = p
c
         l = l - 1
         if( l.ge.lend )
     $      go to 90
         go to 140
c
      end if
c
c     undo scaling if necessary
c
  140 continue
      if( iscale.eq.1 ) then
         call dlascl( 'g', izero, izero, ssfmax, anorm, lendsv-lsv+1,
     $                 1, d( lsv ), n, info )
         call dlascl( 'g', izero, izero, ssfmax, anorm, lendsv-lsv,
     $                 1, e( lsv ), n, info )
c
      else if( iscale.eq.2 ) then
         call dlascl( 'g', izero, izero, ssfmin, anorm, lendsv-lsv+1,
     $                 1, d( lsv ), n, info )
         call dlascl( 'g', izero, izero, ssfmin, anorm, lendsv-lsv,
     $                 1, e( lsv ), n, info )
c
      end if
c
c     check for no convergence to an eigenvalue after a total
c     of n*maxit iterations.
c
      if( jtot.lt.nmaxit )
     $   go to 10
      do 150 i = 1, n - 1
         if(neZERO( e( i )))
     $      info = info + 1
  150 continue
      go to 190
c
c     order eigenvalues and eigenvectors.
c
  160 continue
      if( icompz.eq.0 ) then
c
c        use quick sort
c
         call dlasrt( 'i', n, d, info )
c
      else
c
c        use selection sort to minimize swaps of eigenvectors
c
         do 180 ii = 2, n
            i = ii - 1
            k = i
            p = d( i )
            do 170 j = ii, n
               if( d( j ).lt.p ) then
                  k = j
                  p = d( j )
               end if
  170       continue
            if( k.ne.i ) then
               d( k ) = d( i )
               d( i ) = p
c
c           *** New starting with version 2.5 ***
c
               p    = z(k)
               z(k) = z(i)
               z(i) = p
c           *************************************
            end if
  180    continue
      end if
c
  190 continue
      return
c
c     %---------------%
c     | End of dstqrb |
c     %---------------%
c
      end
c
c
c\BeginDoc
c
c\Name: dsortr
c
c\Description:
c  Sort the array X1 in the order specified by WHICH and optionally
c  applies the permutation to the array X2.
c
c\Usage:
c  call dsortr
c     ( WHICH, APPLY, N, X1, X2 )
c
c\Arguments
c  WHICH   Character*2.  (Input)
c          'LM' -> X1 is sorted into increasing order of magnitude.
c          'SM' -> X1 is sorted into decreasing order of magnitude.
c          'LA' -> X1 is sorted into increasing order of algebraic.
c          'SA' -> X1 is sorted into decreasing order of algebraic.
c
c  APPLY   Logical.  (Input)
c          APPLY = .TRUE.  -> apply the sorted order to X2.
c          APPLY = .FALSE. -> do not apply the sorted order to X2.
c
c  N       Integer.  (INPUT)
c          Size of the arrays.
c
c  X1      Double precision array of length N.  (INPUT/OUTPUT)
c          The array to be sorted.
c
c  X2      Double precision array of length N.  (INPUT/OUTPUT)
c          Only referenced if APPLY = .TRUE.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/16/93: Version ' 2.1'.
c               Adapted from the sort routine in LANSO.
c
c\SCCS Information: @(#)
c FILE: sortr.F   SID: 2.3   DATE OF SID: 4/19/96   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsortr (which, apply, n, x1, x2)
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  which*2
      logical    apply
      integer    n
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           x1(0:n-1), x2(0:n-1)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i, igap, j
      Double precision
     &           temp
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      igap = n / 2
c
      if (which .eq. 'SA') then
c
c        X1 is sorted into decreasing order of algebraic.
c
   10    continue
         if (igap .eq. 0) go to 9000
         do 30 i = igap, n-1
            j = i-igap
   20       continue
c
            if (j.lt.0) go to 30
c
            if (x1(j).lt.x1(j+igap)) then
               temp = x1(j)
               x1(j) = x1(j+igap)
               x1(j+igap) = temp
               if (apply) then
                  temp = x2(j)
                  x2(j) = x2(j+igap)
                  x2(j+igap) = temp
               end if
            else
               go to 30
            endif
            j = j-igap
            go to 20
   30    continue
         igap = igap / 2
         go to 10
c
      else if (which .eq. 'SM') then
c
c        X1 is sorted into decreasing order of magnitude.
c
   40    continue
         if (igap .eq. 0) go to 9000
         do 60 i = igap, n-1
            j = i-igap
   50       continue
c
            if (j.lt.0) go to 60
c
            if (abs(x1(j)).lt.abs(x1(j+igap))) then
               temp = x1(j)
               x1(j) = x1(j+igap)
               x1(j+igap) = temp
               if (apply) then
                  temp = x2(j)
                  x2(j) = x2(j+igap)
                  x2(j+igap) = temp
               end if
            else
               go to 60
            endif
            j = j-igap
            go to 50
   60    continue
         igap = igap / 2
         go to 40
c
      else if (which .eq. 'LA') then
c
c        X1 is sorted into increasing order of algebraic.
c
   70    continue
         if (igap .eq. 0) go to 9000
         do 90 i = igap, n-1
            j = i-igap
   80       continue
c
            if (j.lt.0) go to 90
c
            if (x1(j).gt.x1(j+igap)) then
               temp = x1(j)
               x1(j) = x1(j+igap)
               x1(j+igap) = temp
               if (apply) then
                  temp = x2(j)
                  x2(j) = x2(j+igap)
                  x2(j+igap) = temp
               end if
            else
               go to 90
            endif
            j = j-igap
            go to 80
   90    continue
         igap = igap / 2
         go to 70
c
      else if (which .eq. 'LM') then
c
c        X1 is sorted into increasing order of magnitude.
c
  100    continue
         if (igap .eq. 0) go to 9000
         do 120 i = igap, n-1
            j = i-igap
  110       continue
c
            if (j.lt.0) go to 120
c
            if (abs(x1(j)).gt.abs(x1(j+igap))) then
               temp = x1(j)
               x1(j) = x1(j+igap)
               x1(j+igap) = temp
               if (apply) then
                  temp = x2(j)
                  x2(j) = x2(j+igap)
                  x2(j+igap) = temp
               end if
            else
               go to 120
            endif
            j = j-igap
            go to 110
  120    continue
         igap = igap / 2
         go to 100
      end if
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsortr |
c     %---------------%
c
      end

c     Changes Oct 2019: to address gcc-10 issues with Rank missmatch:
c     in `call blksb[lf]`: b(1,j) -> b(:,j); newrhs -> newrhs(:)



      subroutine backsolvef(m,nsuper,nrhs,lindx,xlindx,lnz,
     &                   xlnz,xsuper,b)
c see below...
      implicit none

      integer m,nsuper,nrhs,lindx(*),xlindx(m+1),
     &        xlnz(m+1),xsuper(m+1)
      double precision lnz(*),b(m,nrhs)
      integer j
      do j = 1,nrhs
         call blkslb(nsuper,xsuper,xlindx,lindx,xlnz,lnz,b(:,j))
      enddo
      return
      end


      subroutine forwardsolvef(m,nsuper,nrhs,lindx,xlindx,
     &  lnz,xlnz,xsuper,b)
c INPUT:
c     m -- the number of column in the matrix
c     lindx -- an nsub-vector of interger which contains, in
c           column major oder, the row subscripts of the nonzero
c           entries in L in a compressed storage format
c     xlindx -- an nsuper-vector of integer of pointers for lindx
c     lnz -- First contains the non-zero entries of d; later
c            contains the entries of the Cholesky factor
c     xlnz -- column pointer for L stored in lnz
c     xsuper -- array of length m+1 containing the supernode
c               partitioning
c     b -- the rhs of the equality constraint
c OUTPUT:
c     b -- the solution

      implicit none

      integer m,nsuper,nrhs,lindx(*),xlindx(m+1),
     &        xlnz(m+1),xsuper(m+1)
      double precision lnz(*),b(m,nrhs)
      integer j
c
      do j = 1,nrhs
         call blkslf(nsuper,xsuper,xlindx,lindx,xlnz,lnz,b(:,j))
      enddo
      return
      end
C***********************************************************************
C***********************************************************************
C
C   Version:        0.4
C   Last modified:  December 27, 1994
C   Authors:        Esmond G. Ng and Barry W. Peyton
C                   Slight modification by Reinhard Furrer
C
C   Mathematical Sciences Section, Oak Ridge National Laboratory
C
C***********************************************************************
C***********************************************************************
C
C***********************************************************************
      subroutine pivotforwardsolve(m,nsuper,nrhs,lindx,xlindx,lnz,
     &                   xlnz,invp,perm,xsuper,newrhs,sol,b)
c Sparse least squares solver via Ng-Peyton's sparse Cholesky
c    factorization for sparse symmetric positive definite
c INPUT:
c     m -- the number of column in the design matrix X
c     nsubmax -- upper bound of the dimension of lindx
c     lindx -- an nsub-vector of interger which contains, in
c           column major oder, the row subscripts of the nonzero
c           entries in L in a compressed storage format
c     xlindx -- an nsuper-vector of integer of pointers for lindx
c     lnz -- First contains the non-zero entries of d; later
c            contains the entries of the Cholesky factor
c     xlnz -- column pointer for L stored in lnz
c     invp -- an m-vector of integer of inverse permutation
c             vector
c     perm -- an m-vector of integer of permutation vector
c     xsuper -- array of length m+1 containing the supernode
c               partitioning
c     newrhs -- extra work vector for right-hand side and
c               solution
c     sol -- the least squares solution
c     b -- an m-vector, usualy the rhs of the equality constraint
c          X'a = (1-tau)X'e in the rq setting
c OUTPUT:
c     y -- an m-vector of least squares solution
c WORK ARRAYS:
c     b -- an m-vector, usually the rhs of the equality constraint
c          X'a = (1-tau)X'e in the rq setting
      implicit none

      integer m,nsuper,nrhs,lindx(*),xlindx(m+1),
     &        invp(m),perm(m),xlnz(m+1), xsuper(m+1)
      integer i,j
      double precision lnz(*),b(m,nrhs),newrhs(m),sol(m,nrhs)

      do j = 1,nrhs
         do i = 1,m
            newrhs(i) = b(perm(i),j)
         enddo
         call blkslf(nsuper,xsuper,xlindx,lindx,xlnz,lnz,newrhs(:))
         do i = 1,m
            sol(i,j) = newrhs(invp(i))
         enddo
      enddo
      return
      end
C***********************************************************************
      subroutine pivotbacksolve(m,nsuper,nrhs,lindx,xlindx,lnz,
     &                   xlnz,invp,perm,xsuper,newrhs,sol,b)
c     see above
      implicit none

      integer m, nsuper,nrhs,lindx(*),xlindx(m+1),
     &        invp(m),perm(m),xlnz(m+1), xsuper(m+1)
      double precision lnz(*),b(m,nrhs),newrhs(m),sol(m,nrhs)
      integer i,j
      do j = 1,nrhs
         do i = 1,m
            newrhs(i) = b(perm(i),j)
         enddo
         call blkslb(nsuper,xsuper,xlindx,lindx,xlnz,lnz,newrhs)
         do i = 1,m
            sol(i,j) = newrhs(invp(i))
         enddo
      enddo
      return
      end
C***********************************************************************
      subroutine backsolves(m,nsuper,nrhs,lindx,xlindx,lnz,
     &                   xlnz,invp,perm,xsuper,newrhs,sol,b)
c Sparse least squares solver via Ng-Peyton's sparse Cholesky
c    factorization for sparse symmetric positive definite
c INPUT:
c     m -- the number of column in the design matrix X
c     nsubmax -- upper bound of the dimension of lindx
c     lindx -- an nsub-vector of interger which contains, in
c           column major oder, the row subscripts of the nonzero
c           entries in L in a compressed storage format
c     xlindx -- an nsuper-vector of integer of pointers for lindx
c     nnzlmax -- the upper bound of the non-zero entries in
c                L stored in lnz, including the diagonal entries
c     lnz -- First contains the non-zero entries of d; later
c            contains the entries of the Cholesky factor
c     xlnz -- column pointer for L stored in lnz
c     invp -- an m-vector of integer of inverse permutation
c             vector
c     perm -- an m-vector of integer of permutation vector
c     xsuper -- array of length m+1 containing the supernode
c               partitioning
c     newrhs -- extra work vector for right-hand side and
c               solution
c     sol -- the least squares solution
c     b -- an m-vector, usualy the rhs of the equality constraint
c          X'a = (1-tau)X'e in the rq setting
c OUTPUT:
c     y -- an m-vector of least squares solution
c WORK ARRAYS:
c     b -- an m-vector, usually the rhs of the equality constraint
c          X'a = (1-tau)X'e in the rq setting
      implicit none

      integer m,nsuper,nrhs,lindx(*),xlindx(m+1),
     &        invp(m),perm(m),xlnz(m+1), xsuper(m+1)
      double precision lnz(*),b(m,nrhs),newrhs(m),sol(m,nrhs)

      integer i,j
      do j = 1,nrhs
         do i = 1,m
            newrhs(i) = b(perm(i),j)
         enddo
         call blkslv(nsuper,xsuper,xlindx,lindx,xlnz,lnz,newrhs(:))
         do i = 1,m
            sol(i,j) = newrhs(invp(i))
         enddo
      enddo
      return
      end

c
      subroutine d_ope ( n, x, y, a, ja, ia )
c
c ope computes A * x for a sparse matrix A.
c
      implicit none
c
      integer n
c
      double precision  a(*)
      integer i
      integer ia(n+1)
      integer ja(*)
      integer k1
      integer k2
      double precision  x(*)
      double precision  y(*)
c
c spasrse matrix * vector multiplication
c
      do i=1,n
         k1 = ia(i)
         k2 = ia(i+1) -1
         y(i) = dot_product ( a(k1:k2), x(ja(k1:k2)) )
      end do
c
      return
      end
c
c-----------------------------------------------------------------------
c
      subroutine ds_eigen_f (maxnev, ncv, maxitr,
     &                       n, iwhich,
     &                       na, a, ja, ia,
     &                       v, d,
     &                       iparam)
c
      implicit none
c
c     %--------------------%
c     | Input Declarations |
c     %--------------------%
c
      integer          maxnev, ncv, maxitr, n, na, ja(*), ia(na+1),
     &                 iparam(8), iwhich
c
      Double precision
     &                 a(*),
     &                 v(n, ncv), d(maxnev),
     &                 workl(ncv*(ncv+8)), workd(3*n), resid(n)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      Double precision
     &                 tol, sigma
      integer          ipntr(11)
c
      character        bmat*1, which*2
      integer          ido, info, lworkl,
     &                 ishfts, mode1, ierr
      logical          rvec, select(ncv)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &                 zero
      parameter        (zero = 0.0D+0)
c
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %-------------------------------------------------%
c     | The following include statement and assignments |
c     | initiate trace output from the internal         |
c     | actions of ARPACK.  See debug.doc in the        |
c     | DOCUMENTS directory for usage.  Initially, the  |
c     | most useful information will be a breakdown of  |
c     | time spent in the various stages of computation |
c     | given by setting msaupd = 1.                    |
c     %-------------------------------------------------%
c
c     %-------------------------------------------------%
c     | The following sets dimensions for this problem. |
c     %-------------------------------------------------%
c
      bmat  = 'I'
c
      if (iwhich .eq. 1) then
            which = 'LM'
      else if (iwhich .eq. 2) then
            which = 'SM'
      else if (iwhich .eq. 7) then
            which = 'LA'
      else if (iwhich .eq. 8) then
            which = 'SA'
      else if (iwhich .eq. 9) then
            which = 'BE'
      else
c
        goto 9000
      end if
c
c     %-----------------------------------------------------%
c     |                                                     |
c     | Specification of stopping rules and initial         |
c     | conditions before calling DSAUPD                    |
c     |                                                     |
c     | TOL  determines the stopping criterion.             |
c     |                                                     |
c     |      Expect                                         |
c     |           abs(lambdaC - lambdaT) < TOL*abs(lambdaC) |
c     |               computed   true                       |
c     |                                                     |
c     |      If TOL .le. 0,  then TOL <- macheps            |
c     |           (machine precision) is used.              |
c     |                                                     |
c     | IDO  is the REVERSE COMMUNICATION parameter         |
c     |      used to specify actions to be taken on return  |
c     |      from DSAUPD. (See usage below.)                |
c     |                                                     |
c     |      It MUST initially be set to 0 before the first |
c     |      call to DSAUPD.                                |
c     |                                                     |
c     | INFO on entry specifies starting vector information |
c     |      and on return indicates error codes            |
c     |                                                     |
c     |      Initially, setting INFO=0 indicates that a     |
c     |      random starting vector is requested to         |
c     |      start the ARNOLDI iteration.  Setting INFO to  |
c     |      a nonzero value on the initial call is used    |
c     |      if you want to specify your own starting       |
c     |      vector (This vector must be placed in RESID.)  |
c     |                                                     |
c     | The work array WORKL is used in DSAUPD as           |
c     | workspace.  Its dimension LWORKL is set as          |
c     | illustrated below.                                  |
c     |                                                     |
c     %-----------------------------------------------------%
c
      lworkl = ncv * (ncv + 8)
      tol = zero
      info = 0
      ido = 0
c
c     %---------------------------------------------------%
c     | Specification of Algorithm Mode:                  |
c     |                                                   |
c     | This program uses the exact shift strategy        |
c     | (indicated by setting PARAM(1) = 1).              |
c     | IPARAM(3) specifies the maximum number of Arnoldi |
c     | iterations allowed.  Mode 1 of DSAUPD is used     |
c     | (IPARAM(7) = 1). All these options can be changed |
c     | by the user. For details see the documentation in |
c     | DSAUPD.                                           |
c     %---------------------------------------------------%
c
      ishfts = 1
      mode1 = 1
c
      iparam(1) = ishfts
      iparam(3) = maxitr
      iparam(7) = mode1
c
c     %------------------------------------------------%
c     | M A I N   L O O P (Reverse communication loop) |
c     %------------------------------------------------%
c
 10   continue
c
c        %---------------------------------------------%
c        | Repeatedly call the routine DSAUPD and take |
c        | actions indicated by parameter IDO until    |
c        | either convergence is indicated or maxitr   |
c        | has been exceeded.                          |
c        %---------------------------------------------%
c
c
         call dsaupd ( ido, bmat, n, which, maxnev, tol, resid,
     &                 ncv, v, n, iparam, ipntr, workd, workl,
     &                 lworkl, info )
c
         if (ido .eq. -1 .or. ido .eq. 1) then
c
c           %--------------------------------------%
c           | Perform matrix vector multiplication |
c           |              y <--- OP*x             |
c           | The user should supply his/her own   |
c           | matrix vector multiplication routine |
c           | here that takes workd(ipntr(1)) as   |
c           | the input, and return the result to  |
c           | workd(ipntr(2)).                     |
c           %--------------------------------------%
c
         call d_ope (na, workd(ipntr(1)), workd(ipntr(2)),
     &               a, ja, ia)
c
c           %-----------------------------------------%
c           | L O O P   B A C K to call DSAUPD again. |
c           %-----------------------------------------%
c
            go to 10
c
         end if
c
c     %----------------------------------------%
c     | Either we have convergence or there is |
c     | an error.                              |
c     %----------------------------------------%
c
      if ( info .lt. 0 ) then
c
c        %--------------------------%
c        | Error message. Check the |
c        | documentation in DSAUPD. |
c        %--------------------------%
c
         goto 9000
c
      else
c
c        %-------------------------------------------%
c        | No fatal errors occurred.                 |
c        | Post-Process using DSEUPD.                |
c        |                                           |
c        | Computed eigenvalues may be extracted.    |
c        |                                           |
c        | Eigenvectors may be also computed now if  |
c        | desired.  (indicated by rvec = .true.)    |
c        |                                           |
c        | The routine DSEUPD now called to do this  |
c        | post processing (Other modes may require  |
c        | more complicated post processing than     |
c        | mode1.)                                   |
c        |                                           |
c        %-------------------------------------------%
c
          rvec = .true.
c
          call dseupd ( rvec, 'A', select, d, v, n, sigma,
     &         bmat, n, which, maxnev, tol, resid, ncv, v, n,
     &         iparam, ipntr, workd, workl, lworkl, ierr )
c
c         %----------------------------------------------%
c         | Eigenvalues are returned in the first column |
c         | of the two dimensional array D and the       |
c         | corresponding eigenvectors are returned in   |
c         | the first NCONV (=IPARAM(5)) columns of the  |
c         | two dimensional array V if requested.        |
c         | Otherwise, an orthogonal basis for the       |
c         | invariant subspace corresponding to the      |
c         | eigenvalues in D is returned in V.           |
c         %----------------------------------------------%
c
          if ( ierr .ne. 0) then
c
c            %------------------------------------%
c            | Error condition:                   |
c            | Check the documentation of DSEUPD. |
c            %------------------------------------%
c
             goto 9000
c
         end if
c
      end if
c
 9000 continue

      end
c
c

# HEADER ####################################################
# This is file spam/R/math.R.                               #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



# `"Ops"":
#      `"+"", `"-"", `"*"", `"/"", `"^"", `"%%"", `"%/%""
#      `"&"", `"|"", `"!""
#      `"=="", `"!="", `"<"", `"<="", `">="", `">""


#     `Math" `"abs"", `"sign"", `"sqrt"", `"ceiling"", `"floor"",
#          `"trunc"", `"cummax"", `"cummin"", `"cumprod"", `"cumsum"",
#          `"log"", `"log10"", `"log2"", `"log1p"", `"acos"", `"acosh"",
#          `"asin"", `"asinh"", `"atan"", `"atanh"", `"exp"", `"expm1"",
#          `"cos"", `"cosh"", `"cospi"", `"sin"", `"sinh"", `"sinpi"",
#          `"tan"", `"tanh"", `"tanpi"", `"gamma"", `"lgamma"",
#          `"digamma"", `"trigamma""

#     `Math2" `"round"", `"signif""

#     `Summary" `"max"", `"min"", `"range"", `"prod"", `"sum"", `"any"", `"all""


##############
# Unary operators "+", "-" and "!" are handled with e2 missing...
#
# Currently, "+", "-" are handled...
setMethod("!",signature(x="spam"),    function(x){
    if(getOption("spam.structurebased")) {
        x@entries <- as.double(callGeneric(x@entries))
        x
    } else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(x)))
        spam(as.double( callGeneric(as.matrix(x))), nrow=dim(x)[1])
    }
})

setMethod("+",signature(e1="spam",e2="missing"), function(e1) e1 )
setMethod("-",signature(e1="spam",e2="missing"), function(e1) { e1@entries <- -e1@entries; e1} )

#     `Math2" :

setMethod("Math2",signature(x = "spam", digits = "ANY"),
          function(x, digits){ x@entries <- callGeneric(x@entries, digits = digits); x })

#     `Math" :

setMethod("Math","spam", function(x){
    if(getOption("spam.structurebased")) {
        x@entries <- callGeneric(x@entries)
        x
    }else{
        x@entries <- callGeneric(x@entries)
        as.spam.spam( x)
    }
})

#     `Math", where we pass to matrix first...

spam_Math <- function(x) {
    if(getOption("spam.structurebased")) {
        x@entries <- callGeneric(x@entries)
        x
    }else{
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(x)))
        as.spam(callGeneric(as.matrix(x)))
    }}


setMethod("exp","spam", spam_Math )
setMethod("log10","spam", spam_Math )
setMethod("log2","spam", spam_Math )
# from ?log: Do not set S4 methods on `logb" itself.
# special case to set base...
setMethod("log","spam", function(x,...) {
    if(getOption("spam.structurebased")) {
        x@entries <- callGeneric(x@entries,...)
        x
    }else{
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(x)))
        as.spam(callGeneric(as.matrix(x),...))
    }}
          )



setMethod("cos","spam", spam_Math )
#setMethod("cospi","spam", spam_Math )
setMethod("cosh","spam", spam_Math )
setMethod("acosh","spam", spam_Math )
setMethod("acos","spam", spam_Math )

setMethod("gamma","spam", spam_Math )
setMethod("digamma","spam", spam_Math )
setMethod("trigamma","spam", spam_Math )
setMethod("lgamma","spam", spam_Math )

setMethod("cummax","spam", spam_Math )
setMethod("cummin","spam", spam_Math )
setMethod("cumprod","spam", spam_Math )
setMethod("cumsum","spam", spam_Math )


#     `Summary" :
setMethod("Summary","spam", function(x,...,na.rm=FALSE){
    if(getOption("spam.structurebased")) {
        callGeneric(x@entries,...,na.rm=na.rm)
    }else{
        if ( prod( x@dimension) == length( x@entries)) {
            callGeneric(x@entries,...,na.rm=na.rm)
        } else {
            callGeneric(c(0,x@entries),...,na.rm=na.rm)
        }
    }
}
          )

logical_Summary <- function( x,...,na.rm=FALSE){
    if(getOption("spam.structurebased")) {
        callGeneric(as.logical(x@entries),...,na.rm=na.rm)
    }else{
        if ( prod( x@dimension) == length( x@entries)) {
            callGeneric(as.logical(x@entries),...,na.rm=na.rm)
        } else {
            callGeneric(as.logical(c(0,x@entries)),...,na.rm=na.rm)
        }
    }
}


setMethod("any","spam", logical_Summary)
setMethod("all","spam", logical_Summary)


################################################################################################################################################################################################################################################################################################
#     `Ops" `"Arith"", `"Compare"", `"Logic""



#     `Logic" `"&"", `"|"".


"spam_Logic_vectorspam" <- function(e1, e2) {
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e2@entries), prod(e2@dimension))))  # spam is full
            return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
        if(identical(length(e1),1L) | identical(length(e1), length(e2@entries))) {
            e2@entries <- as.double( callGeneric(e1, e2@entries))
            return(e2)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
    }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
        return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
}

"spam_Logic_spamvector" <- function(e1, e2)  {
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e1@entries), prod(e1@dimension))))  # spam is full
            return( as.spam( callGeneric(as.matrix(e1), e2)) )   # 3.0: no as.spam -> matrix
        if(identical(length(e2),1L) | identical(length(e2), length(e1@entries))) {
            e1@entries <- as.double( callGeneric(e1@entries, e2))
            return(e1)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
    }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
        return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
}
"spam_Logic_matrixspam" <- function(e1, e2){
  #    cat("spam_Arith_matrixspam")
  if( identical( e2@dimension, dim(e1))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
      return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}
"spam_Logic_spammatrix" <- function(e1, e2){
  #    cat("spam_Arith_spammatrix")
  if( identical( e1@dimension, dim(e2))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(as.matrix(e1), e2)) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
      return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}


setMethod("|",signature(e1="spam",e2="spam"),
          function(e1,e2){ z <- spam_add(e1,e2);  z@entries <- rep(1,length(z@colindices));z})

setMethod("&",signature(e1="spam",e2="spam"),
          function(e1,e2){ z <- spam_mult(e1,e2); z@entries <- rep(1,length(z@colindices));z})
setMethod("Logic",signature(e1="spam",e2="vector"), spam_Logic_spamvector)
setMethod("Logic",signature(e1="vector",e2="spam"), spam_Logic_vectorspam)
setMethod("Logic",signature(e1="spam",e2="matrix"), spam_Logic_spammatrix)
setMethod("Logic",signature(e1="matrix",e2="spam"), spam_Logic_matrixspam)

##################################################################################################
#     `Compare" `"=="", `">"", `"<"", `"!="", `"<="", `">=""
"spam_Compare" <- function(e1,e2) {
    inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)),  max(prod(dim(e1)), prod(dim(e2))))
    as.spam( callGeneric( as.matrix(e1), as.matrix(e2))  )
}

"spam_Compare_spamvector" <- function(e1, e2){
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e1@entries), prod(e1@dimension))))  # spam is full
            return( as.spam( callGeneric(as.matrix(e1), e2)) )   # 3.0: no as.spam -> matrix
        if(identical(length(e2),1L) | identical(length(e2), length(e1@entries))) {
            e1@entries <- as.double(callGeneric(e1@entries, e2))
            return(e1)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
    }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
        return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
}
"spam_Compare_vectorspam" <- function(e1, e2) {
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e2@entries), prod(e2@dimension))))  # spam is full
            return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
        if(identical(length(e1),1L) | identical(length(e1), length(e2@entries))) {
            e2@entries <- as.double( callGeneric(e1, e2@entries))
            return(e2)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
     }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
        return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
}
"spam_Compare_matrixspam" <- function(e1, e2){
  #    cat("spam_Arith_matrixspam")
  if( identical( e2@dimension, dim(e1))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
      return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}
"spam_Compare_spammatrix" <- function(e1, e2){
  #    cat("spam_Arith_spammatrix")
  if( identical( e1@dimension, dim(e2))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(as.matrix(e1), e2)) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
      return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}


setMethod("Compare",signature(e1="spam",e2="spam"),   spam_Compare )
setMethod("Compare",signature(e1="spam",e2="vector"), spam_Compare_spamvector )
setMethod("Compare",signature(e1="vector",e2="spam"), spam_Compare_vectorspam )
setMethod("Compare",signature(e1="spam",e2="matrix"), spam_Compare_spammatrix )
setMethod("Compare",signature(e1="matrix",e2="spam"), spam_Compare_matrixspam )


##################################################################################################
#     `Arith": `"+"", `"-"", `"*"", `"^"", `"%%"", `"%/%"", `"/""

"spam_Arith_vectorspam" <- function(e1, e2){
#    cat("spam_Arith_vectorspam")
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e2@entries), prod(e2@dimension))))  # spam is full
            return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
        if(identical(length(e1),1L) | identical(length(e1), length(e2@entries))) {
            e2@entries <- callGeneric(e1, e2@entries)
            return(e2)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
    }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
         return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
}
"spam_Arith_spamvector" <- function(e1, e2){
#    cat("spam_Arith_spamvector")
    if(getOption("spam.structurebased")) {
        if( isTRUE(all.equal(length(e1@entries), prod(e1@dimension))))  # spam is full
            return( as.spam( callGeneric(as.matrix(e1), e2)) )          # 3.0: no as.spam -> matrix
        if(identical(length(e2),1L) | identical(length(e2), length(e1@entries))) {
            e1@entries <- callGeneric(e1@entries, e2)
            return(e1)
        }
        stop(gettextf("incompatible lengths for %s operation.", sQuote(.Generic)))
     }  else {
        inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
        return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
}
"spam_Arith_matrixspam" <- function(e1, e2){
  #    cat("spam_Arith_matrixspam")
  if( identical( e2@dimension, dim(e1))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(e1, as.matrix(e2))) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e2)))
      return( as.spam( callGeneric(e1, as.matrix(e2))) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}
"spam_Arith_spammatrix" <- function(e1, e2){
  #    cat("spam_Arith_spammatrix")
  if( identical( e1@dimension, dim(e2))) {
    if(getOption("spam.structurebased")) {
      return( as.spam( callGeneric(as.matrix(e1), e2)) )          # 3.0: no as.spam -> matrix
    }  else {
      inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)), prod(dim(e1)))
      return( as.spam( callGeneric(as.matrix(e1), e2)) )
    }
  }
  stop(gettextf("incompatible dimensions for %s operation.", sQuote(.Generic)))
}
spam_Arith <- function(e1,e2) {
    inefficiencywarning( gettextf("This %s operation may be inefficient",sQuote(.Generic)),  max(prod(dim(e1)), prod(dim(e2))))
    as.spam( callGeneric( as.matrix(e1), as.matrix(e2)))
}



setMethod("Arith",signature(e1="spam",e2="spam"),   spam_Arith )
setMethod("Arith",signature(e1="spam",e2="vector"), spam_Arith_spamvector)
setMethod("Arith",signature(e1="vector",e2="spam"), spam_Arith_vectorspam)
setMethod("Arith",signature(e1="spam",e2="matrix"), spam_Arith_spammatrix)
setMethod("Arith",signature(e1="matrix",e2="spam"), spam_Arith_matrixspam)

setMethod("/",signature(e1="spam",e2="spam"),   function(e1,e2){ "/"(e1,as.matrix(e2)) } )
setMethod("/",signature(e1="matrix",e2="spam"), function(e1,e2){ "/"(e1,as.matrix(e2)) } )
setMethod("^",signature(e1="spam",e2="spam"),   function(e1,e2){ "^"(e1,as.matrix(e2)) } )


######################################################################
# nz <- 128; ln <- nz^2; A <- spam(0,ln,ln); is <- sample(ln,nz); js <- sample(ln,nz);A[cbind(is,js)] <- 1:nz
# nz <- 128; ln <- nz^2; A <- spam(0,ln,ln); is <- sample(ln,ln); js <- sample(ln,ln);A[cbind(is,js)] <- 1:ln
# system.time(   spam:::.spam.addsparsefull(A,1)) ; system.time(   as.matrix.spam(A)+1.5)


#######################################################################
"spam_add" <- function(A, B, s=1)
{

                                        # cat("spam_add")
    nrow <- A@dimension[1]
    ncol <- A@dimension[2]
    if(ncol != B@dimension[2] || nrow != B@dimension[1])
        stop("non-conformable matrices")

    if( getOption("spam.force64") || .format.spam(A)$package == "spam64" || .format.spam(B)$package == "spam64")
        SS <- .format64()
    else
        SS <- .format32

    nzmax <- .C64("aplbdg",
                  ## subroutine aplbdg (nrow,ncol,ja,ia,jb,ib,ndegr,nnz,iw)
                  SIGNATURE = c(SS$signature, SS$signature,
                      SS$signature, SS$signature, SS$signature, SS$signature,
                      SS$signature, SS$signature, SS$signature),

                  nrow,
                  ncol,

                  A@colindices,
                  A@rowpointers,
                  B@colindices,
                  B@rowpointers,

                  vector_dc(SS$type, nrow),
                  nnz = vector_dc(SS$type, 1),
                  vector_dc(SS$type, ncol),

                  INTENT = c("r", "r",
                      "r", "r", "r", "r",
                      "w", "w", "w"),
                  NAOK=getOption("spam.NAOK"),
                  PACKAGE = SS$package)$nnz

    z <- .C64("aplsb1",
              ## subroutine aplsb1 (nrow,ncol,a,ja,ia,s,b,jb,ib,c,jc,ic,nzmax,ierr)
              SIGNATURE = c(SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double",
                  "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  SS$signature, SS$signature),

              nrow,
              ncol,

              A@entries,
              A@colindices,
              A@rowpointers,

              s,

              B@entries,
              B@colindices,
              B@rowpointers,

              entries     = vector_dc("double", nzmax),
              colindices  = vector_dc(SS$type, nzmax),
              rowpointers = vector_dc(SS$type, nrow+1),

              nzmax+1,
              ierr = vector_dc(SS$type, 1),

              INTENT = c("r", "r",
                  "r", "r", "r",
                  "r",
                  "r", "r", "r",
                  "w", "w", "w",
                  "r", "w" ),
              NAOK=getOption("spam.NAOK"),
              PACKAGE = SS$package)

    if(z$ierr != 0) stop("insufficient space for sparse matrix addition")
    nz <- z$rowpointers[nrow+1]-1
    return(.newSpam(
        entries = z$entries[1:nz],
        colindices = z$colindices[1:nz],
        rowpointers = z$rowpointers,
        dimension =  c(nrow,ncol)
        ))
}

setMethod("+",signature(e1="spam",e2="spam"),  function(e1,e2){ spam_add(e1, e2)    })
setMethod("-",signature(e1="spam",e2="spam"),  function(e1,e2){ spam_add(e1, e2, -1)})



###############################################################################

"spam_mult" <- function(e1,e2)
{
#  if(is.vector(e1)) {
#    if(length(e1) == 1){
#      if(e1==0) return( spam(0,nrow(e2),ncol(e2)))
#      else{  # just a scalar
#        e2@entries <- e1*e2@entries
#        return(e2)
#      }
#    }  else if(length(e1) == nrow(e2))
#      return(diag.spam(e1) %*% e2)
#    else # length(e1) == ncol(e2) is not required
#      stop("e1 and e2 not conformable for efficient element-by-element multiplication")
#  }
#  else if(is.vector(e2)) {
#    if(length(e2) == 1){
#      if(e2==0)   return( spam(0,nrow(e1),ncol(e1)))
#      else {
#        e1@entries <- e2*e1@entries
#        return(e1)
#      }
#    }
#    else if(length(e2) == nrow(e1))
#      return(diag.spam(e2) %*% e1)
#    else
#      stop("e1 and e2 not conformable for efficient element-by-element multiplication")
#  }
#  if(is.matrix(e1))
#    e1 <- as.spam(e1)
#  else if(is.matrix(e2))
#    e2 <- as.spam(e2)
#  if(!(is.spam(e1) && is.spam(e2)))
#    stop("Arguments must be of class:  vector, matrix or spam")

  e1row <- e1@dimension[1]
  e1col <- e1@dimension[2]
  if(e1col != e2@dimension[2] | e1row != e2@dimension[1])
    stop("non-conformable matrices")
  nnzmax <- length(intersect(e1@colindices+e1col*(rep(1:e1row,diff(e1@rowpointers))-1),
                             e2@colindices+e2@dimension[2]*(rep(1:e2@dimension[1],diff(e2@rowpointers))-1)))+1

  if( getOption("spam.force64") || .format.spam(e1)$package == "spam64" || .format.spam(e2)$package == "spam64")
      SS <- .format64()
  else
      SS <- .format32

  z <- .C64("aemub",
     ##        subroutine aemub (nrow,ncol,a,ja,ia,amask,jmask,imask,
     ## *                  c,jc,ic,nzmax,ierr)
            SIGNATURE = c(SS$signature, SS$signature,
                "double", SS$signature, SS$signature,
                "double", SS$signature, SS$signature,
                "double", SS$signature, SS$signature,
                SS$signature, SS$signature),

            e1row,
            e1col,

            e1@entries, e1@colindices, e1@rowpointers,

            e2@entries, e2@colindices, e2@rowpointers,

            entries     = vector_dc("double", nnzmax),
            colindices  = vector_dc(SS$type, nnzmax),
            rowpointers = vector_dc(SS$type, e1row+1),

            nnzmax,
            ierr = vector_dc(SS$type,1),

            INTENT = c("r","r",
                "r", "r", "r",
                "r", "r", "r",
                "w", "w", "w",
                "r", "w"),
            NAOK=getOption("spam.NAOK"),
            PACKAGE = SS$package)
  ## z <- .Fortran("aemub",
  ##               as.integer(e1row),
  ##               as.integer(e1col),
  ##               as.double(e1@entries), as.integer(e1@colindices),  as.integer(e1@rowpointers),
  ##               as.double(e2@entries), as.integer(e2@colindices),  as.integer(e2@rowpointers),
  ##               entries     = vector("double",nnzmax),
  ##               colindices  = vector("integer",nnzmax),
  ##               rowpointers = vector("integer",e1row+1),
  ##               ## integer(e1col),
  ##               double(e1col),
  ##               as.integer(nnzmax),
  ##               ierr = vector("integer",1),
  ##               NAOK=getOption("spam.NAOK"),  PACKAGE = "spam")

    if(z$ierr != 0)      stop("insufficient space for element-wise sparse matrix multiplication")

    nnz <- z$rowpointers[e1row+1]-1

    if(identical(z$entries,0L)){#trap zero matrix
        z$colindices <- 1L
        z$rowpointers <- c(1L,rep(2L,e1row))
    }

    return(.newSpam(
        entries=z$entries[1:nnz],
        colindices=z$colindices[1:nnz],
        rowpointers=z$rowpointers,
        dimension=c(e1row,e1col)))
}


setMethod("*",signature(e1="spam",e2="spam"), spam_mult)


##########################################################################################################################################################################################################################################################################################################################################################################################################################################################################################################
"matrix_add_spammatrix" <- function(A,B){
                                        #    cat("matrix_add_spammatrix")
                                        # A is sparse, B is full
                                        #  if (missing(B)) return(A)
                                        #  if (!is.numeric(B)) stop("numeric argument expected")
    nrow <- A@dimension[1]
    ncol <- A@dimension[2]
    pdim <- prod(nrow,ncol)
                                        #  if (is.matrix(B)) {
    if(ncol != dim(B)[2] || nrow != dim(B)[1])
        stop("non-conformable matrices")
                                        #  } else {
                                        #    if(pdim%%length(B)!=0) {
                                        #      stop("longer object length
                                        #        is not a multiple of shorter object length")
                                        #    } else  B <- rep(B,pdim %/% length(B))
                                        #  }
    ## print("addsparsefull")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(A)

    return(array(
        .C64("addsparsefull",
                 SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                               "double"),

                 nrow,
                 A@entries,
                 A@colindices,
                 A@rowpointers,
                 b = c(B),

                 INTENT = c("r", "r", "r", "r",
                            "rw"),
                 NAOK=getOption("spam.NAOK"),
                 PACKAGE = SS$package )$b,
        c(nrow,ncol)))
 }

"matrix_sub_spammatrix" <- function(A,B){
#    cat("matrix_sub_spammatrix")
  # A is sparse, B is full
#  if (missing(B)) {
#    A@entries <- -A@entries
#    return(A)
#  }
#  if (!is.numeric(B)) stop("numeric argument expected")
  nrow <- A@dimension[1]
  ncol <- A@dimension[2]
#  pdim <- prod(nrow,ncol)
#  if (is.matrix(B)) {
    if(ncol != dim(B)[2] || nrow != dim(B)[1])
      stop("non-conformable matrices")
#  } else {
#    if(pdim %% length(B)!=0) {
#      stop("longer object length
#        is not a multiple of shorter object length")
#    } else  B <- rep(B,pdim %/% length(B))
#  }
                                        #  if (!is.double(B[1]))  B <- as.double(B)
    ## print("subfullsparse")
    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(A)

    return(array(
        .C64("subfullsparse",
             SIGNATURE = c( SS$signature, SS$signature,
                           "double", SS$signature, SS$signature,
                           "double"),

             nrow,
             ncol,

             A@entries,
             A@colindices,
             A@rowpointers,

             b = c(B),

             INTENT = c("r", "r",
                        "r", "r", "r",
                        "rw"),
             NAOK=getOption("spam.NAOK"),
             PACKAGE = SS$package)$b,
        c(nrow,ncol)))
}

"matrix_sub_matrixspam" <- function(B,A){
#    cat("matrix_sub_spammatrix")
  # A is sparse, B is full
  if (!is.numeric(B)) stop("numeric argument expected")
  nrow <- A@dimension[1]
  ncol <- A@dimension[2]
#  pdim <- prod(nrow,ncol)
#  if (is.matrix(B)) {
    if(ncol != dim(B)[2] || nrow != dim(B)[1])
      stop("non-conformable matrices")
#  } else {
#    if(pdim %% length(B)!=0) {
#      stop("longer object length
#        is not a multiple of shorter object length")
#    } else  B <- rep(B,pdim %/% length(B))
#  }
#  if (!is.double(B[1]))  B <- as.double(B)

    if(  getOption("spam.force64") )
        SS <- .format64()
    else
        SS <- .format.spam(A)

    return(array(
        .C64("subsparsefull",
             SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                           "double"),

             nrow,
             A@entries,
             A@colindices,
             A@rowpointers,

             b = c(B),

             INTENT = c("r", "r", "r", "r",
                        "rw"),
             NAOK=getOption("spam.NAOK"),
             PACKAGE = SS$package )$b,
        c(nrow,ncol)))
}
#
setMethod("+",signature(e1="spam",   e2="matrix"), function(e1,e2){ matrix_add_spammatrix(e1,e2)})
setMethod("+",signature(e1="matrix",   e2="spam"), function(e1,e2){ matrix_add_spammatrix(e2,e1)})
setMethod("-",signature(e1="matrix",   e2="spam"), function(e1,e2){ matrix_sub_matrixspam(e1,e2)})
setMethod("-",signature(e1="spam",   e2="matrix"), function(e1,e2){ matrix_sub_spammatrix(e1,e2)})


#"spam_division" <- function(e1,e2) { # Element-wise matrix division of two spams
#  if(is.numeric(e1) && length(e1) == 1)
#  { e2@entries <- e1/e2@entries
#    return(e2)
# } else if(is.numeric(e2) && length(e2) == 1) {
#    e1@entries <- e1@entries/e2
#    return(e1)
#  }
#  else if(is.spam(e1) || is.spam(e2) || is.matrix(e1) || is.matrix(e2)){
#        if(is.matrix(e1)) e1 <- as.spam(e1)
#        if(is.matrix(e2)) e2 <- as.spam(e2)
#        nrow <- e1@dimension[1]
#        ncol <- e1@dimension[2]
#        if(ncol != e2@dimension[2] | nrow != e2@dimension[1])
#          stop("matrices not conformable for element-by-element division")
#	nzmax <- length(unique(c(e1@colindices+ncol*(rep(1:nrow,diff(e1@rowpointers))-1),
#                                 e2@colindices+e2@dimension[2]*(rep(1:e2@dimension[1],diff(e2@rowpointers))-1))))+1
#        z <- .Fortran("_aedib_",   # does not order the colindicies upon return!
#                      nrow,
#                      ncol,
#                      as.integer(1),
#                      as.double(e1@entries), e1@colindices,  e1@rowpointers,
#                      as.double(e2@entries), e2@colindices,  e2@rowpointers,
#                      entries     = vector("double",nzmax),
#                      colindices  = vector("integer",nzmax),
#                      rowpointers = vector("integer",nrow+1),
#                      as.integer(nzmax),
#                      integer(ncol),
#                      double(ncol),
#                      ierr = vector("integer",1),
#                      NAOK=getOption("spam.NAOK"),PACKAGE = "spam"
#                      )
#        if(z$ierr != 0) stop("insufficient space for element-wise sparse matrix division")
#        nz <- z$rowpointers[nrow+1]-1
#        return(new("spam",entries=z$entries[1:nz],colindices=z$colindices[1:nz],rowpointers=z$rowpointers,
#                   dimension=c(nrow,ncol)))
#    }



##"spam_exponent" <- function(e1, e2)
#{
#    nrow <- e1@dimension[1]
#    ncol <- e1@dimension[2]
#    if(ncol != e2@dimension[2] | nrow != e2@dimension[1])
#        stop("matrices not conformable for element-wise exponentiation ")
#    nzmax <- length(unique(c(e1@colindices+ncol*(rep(1:nrow,diff(e1@rowpointers))-1),
#                             e2@colindices+e2@dimension[2]*(rep(1:e2@dimension[1],diff(e2@rowpointers))-1))))+1
#    z <- .Fortran("_aeexpb_", does not reorder col indices!
#                  as.integer(nrow), as.integer(ncol),
#                  1L,
#                  as.double(e1@entries),  as.integer(e1@colindices),  as.integer(e1@rowpointers),
#                  as.double(e2@entries),  as.integer(e2@colindices),  as.integer(e2@rowpointers),
#                  entries     = vector("double",nzmax),
#                  colindices  = vector("integer",nzmax),
#                  rowpointers = vector("integer",nrow+1),
#                  as.integer(nzmax),
#                  integer(ncol),       double(ncol),
#                  ierr = vector("integer",1),
#                  NAOK=getOption("spam.NAOK"),PACKAGE = "spam"
#                  )
#    if(z$ierr != 0) stop("insufficient space for element-wise exponentiation")
#    nz <- z$rowpointers[nrow+1]-1
#    return(new("spam",entries=z$entries[1:nz],colindices=z$colindices[1:nz],rowpointers=z$rowpointers,
#               dimension=c(nrow,ncol)))
#}

#############################
#getClass("numeric")
#  Extends: "vector"
#getClass("matrix")
#Extends: Class "vector", by class "array", distance 3, with explicit coerce
# Hence we use use vector, especially, to include the NA case that is not of type numeric!

# HEADER ####################################################
# This is file spam/R/diff.R.                               #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

     


########################################################################
diff.spam <- 
function (x, lag = 1, differences = 1, ...) 
{
    xlen <-   dim(x)[1L]
    if (length(lag) > 1L || length(differences) > 1L || lag < 
        1L || differences < 1L) 
        stop("'lag' and 'differences' must be integers >= 1")
    if (lag * differences >= xlen) {        
        return( as.spam( matrix( 0, 0L, dim(x)[2L])))
#       return( numeric( 0))        
# https://bugs.r-project.org/show_bug.cgi?id=18972
# -    return(x[0L]) # empty, but of proper mode
# +    return(if (ismat) x[0L, , drop = FALSE] else x[0L]) # empty, but of proper mode
# https://bugs.r-project.org/attachment.cgi?id=3569&action=edit
    }    
    for (i in 1L:differences){
      x <- x[(1L+lag):xlen,, drop = FALSE] - x[1L:(xlen-lag),, drop = FALSE] 
      xlen <- xlen - lag
    }
    return( x)
}

setMethod("diff","spam",diff.spam)

# HEADER ####################################################
# This is file spam/R/tcrossprod.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

     

crossprod.spam <- function(x, y=NULL, ...) {
    dimx <- dim(x)
    if( is.null(y)) {
        if(!is.spam(x)) return(crossprod(x))

        if(dimx[2]==1L) return(matrix( sum(x@entries^2)))

        return( t.spam(x) %*% x)
    }
    if( (!is.spam(x)) & (!is.spam(y))) return(crossprod(x,y))
    
    return( t(x) %*% y)
    
}
tcrossprod.spam <- function(x, y=NULL, ...) {
    dimx <- dim(x)
    if( is.null(y)) {
        if(!is.spam(x)) return(tcrossprod(x))

        if(dimx[2]==1L) return(matrix( sum(x@entries^2)))

        return( x %*% t.spam(x))
    }
    if( (!is.spam(x)) & (!is.spam(y))) return(tcrossprod(x,y))
    
    return( x %*% t(y))
    
}


setMethod("crossprod",signature(x="spam",y="missing"), crossprod.spam)
setMethod("crossprod",signature(x="spam",y="spam"), crossprod.spam)
setMethod("crossprod",signature(x="spam",y="ANY"), crossprod.spam)
setMethod("crossprod",signature(x="ANY",y="spam"), crossprod.spam)
setMethod("tcrossprod",signature(x="spam",y="missing"), tcrossprod.spam)
setMethod("tcrossprod",signature(x="spam",y="spam"), tcrossprod.spam)
setMethod("tcrossprod",signature(x="spam",y="ANY"), tcrossprod.spam)
setMethod("tcrossprod",signature(x="ANY",y="spam"), tcrossprod.spam)

# HEADER ####################################################
# This is file spam/tests/testthat/test-crossprod.R.        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-crossprod.R")
     


options(spam.printsize=60)

######################################################################

test_that("crossprod n=1", {
    set.seed(1)
    
    xf <- rnorm(10)
    xf[xf<0] <- 0
    xs <- as.spam(xf)
    
    yf <- rnorm(10)
    yf[yf<0] <- 0
    ys <- as.spam(yf)
    
    spamtest_eq( crossprod( xf), crossprod.spam( xs))
    spamtest_eq( crossprod( xf, yf), crossprod.spam( xs, ys))
    spamtest_eq( crossprod( xf, yf), crossprod.spam( xs, yf))

    # now dispatching
    spamtest_eq( crossprod( xf), crossprod( xs))
    spamtest_eq( crossprod( xf, yf), crossprod( xs, ys))
    spamtest_eq( crossprod( xf, yf), crossprod( xs, yf))
    
    
    dim(xf) <- c(2,5)
    dim(yf) <- c(2,5)
    ys <- as.spam(yf)
    xs <- as.spam(xf)
    
    spamtest_eq( crossprod( xf, yf), crossprod.spam( xs, ys))
    # now dispatching
    spamtest_eq( crossprod( xf, yf), crossprod( xs, ys))
})

c-----------------------------------------------------------------
      subroutine gri(i,rpin,rindex)
      implicit none
      integer j, i, rpin(*), rindex
c-----------------------------------------------------------------
c     purpose:
c     --------
c     get row index for the i-th entrie of the sparse matrix
c on entry:
c----------
c     i      = number of the element
c     rpin   = input rowpointers vector
c on return:
c-----------
c     rindex  = output row index
c-----------------------------------------------------------------
      j=1
      do while (rpin(j) <= i)
         j=j+1
      end do
      rindex=j-1
      return
      end
c--------end-of-gri---------------------------------------------

c-----------------------------------------------------------------
      subroutine gfact(i,j,splits,fact,nfact,out)
      implicit none
      integer j, i, nfact, splits(nfact+1)
      double precision fact(nfact,nfact), out
c-----------------------------------------------------------------
c     purpose:
c     --------
c     get fact for coordinates i,j,
c on entry:
c----------
c     i,j    = indices of element
c     splits = splits of the matrix
c     fact   = values to be returned
c     nfact  = ncol(fact)=nrow(fact)
c return:
c-----------
c     out    = the correct value of fact
c----------------------------------------------------------------
      integer ii, jj
      if (i >= splits(nfact+1) .OR. j >= splits(nfact+1)) then
         goto 9000
c         stop 'i,j are out of larger than (nfact+1)'
      end if
      ii=1
      do while (splits(ii+1) <= i)
         ii=ii+1
      end do
      jj=1
      do while (splits(jj+1) <= j)
         jj=jj+1
      end do
      out=fact(ii,jj)

 9000 continue

      end
c--------end-of-gfact---------------------------------------------


c-----------------------------------------------------------------
      subroutine gmult_f(a, ia, ja, na, splits, fact, nfact, out)
      implicit none
      integer ia(*), ja(*), na, nfact, splits(nfact+1)
      double precision a(na), fact(nfact,nfact), out(na)
c-----------------------------------------------------------------
c     purpose:
c     --------
c     block multipli the entries of a sparse matrix
c on entry:
c----------
c     a      = entries
c     ia     = colindices
c     ja     = rowpointer
c     splits = splits, length is nfact+1
c     fact   = matrix(nfact, nfact)
c     nfact  = ncol(fact)=nrow(fact)
c return:
c-----------
c     out    = modified entries
c----------------------------------------------------------------
      integer ii, ri
      double precision f
      do ii = 1, na
         call gri(ii, ja, ri)
         call gfact(ri, ia(ii), splits, fact, nfact, f)
         out(ii) = a(ii) * f
      end do
      return
      end
c--------end-of-gmult---------------------------------------------

# HEADER ####################################################
# This is file spam/R/definitions.R.                        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# We check that
# 1.  #colindices equals #entries
# 2.	class(entries) is "numeric"
# 3.	rowpointers[1] = 1 and rowpointers[MAX] = #entries + 1
# 4.	dimension[1] ==  #rowpointers - 1
#
# --- Test in C ---
# c1.  rowpointers must be increasing
# c2.  No NaN/NA/Inf in rowpointer, colindices or entries
# c3.  colindices are in domain [1, dimension[2]]
#
# Warning: if class(colindices) != class(rowpointers)
validate_spam <- function(object) {
  # 1
  if(length(object@entries) != length(object@colindices))
    return("Length of slot 'entries' does not equal to length of slot 'colindices'.")

  # 2
  if(!is.double(object@entries))
    return("Slot 'entries' must be of type double (not integer).")

  # 3
  if(object@rowpointers[1] != 1)
    return("Slot 'rowpointers' must be have 1 on the first position.")
  if(object@rowpointers[length(object@rowpointers)] != length(object@entries) + 1)
    return("Last value of rowpointers is incorrect.")

  # 4
  if(length(object@rowpointers) != object@dimension[1] + 1)
    return("Slot 'rowpointers' has not expected length.")


  # --- For performance issues, the following tests should be implemented in C ---

  # c1
  result <- any(head(object@rowpointers, n=-1) > object@rowpointers[-1], na.rm=TRUE)
  if(result == TRUE)
    return("Slot 'rowpointers' is not increasing.")

  # c2
  if(any(!is.finite(object@rowpointers)))
    return("Slot 'rowpointers' contains non-finite elements.")
  if(any(!is.finite(object@colindices)))
    return("Slot 'colindices' contains non-finite elements.")

  # raise a warning if any non-finite element is in entries:
  if(!getOption("spam.NAOK") && any(!is.finite(object@entries)))
    warning("Slot 'entries' contains non-finite elements.")

  if(any(object@colindices < 1) | any(object@colindices > object@dimension[2]))
    return("Slot 'colindices' contains element outside the domain [1, ncol].")

  return(TRUE)
}



# Start with the two classes. Remaining part contains `spam` only material.
setClass(
  Class = "spam",
  slots = c(entries="numeric",
            colindices="numeric",
            rowpointers="numeric",
            dimension="numeric"
    ),
  prototype = prototype(entries=0,
                        colindices=1,
                        rowpointers=c(1,2),
                        dimension=c(1,1)
    ),
  validity = validate_spam
)

setClass("spam.chol.NgPeyton",
         representation(entries="numeric",      colindices="integer",
                        colpointers="integer",  rowpointers="integer",
                        dimension="integer",
                        pivot="integer",        invpivot="integer",
                        supernodes="integer",   snmember="integer",
                        memory="integer",       nnzA="integer")
         )

# lindx=  colindices
# xlindx= colpointers
# xlnz=   rowpointers
# snode=snmember
# xsuper=supernodes
# c(... nnztmp,cachesize)= memory
########################################################################



# Internal function to create new spam or spam64 object, depending on the size.
.newSpam <- function(entries=0,
                     colindices=1,
                     rowpointers=NULL,
                     dimension=c(1,1),
                     force64=getOption("spam.force64")
                     )
{
    if(is.null(rowpointers)) {
        rowpointers <- c(1,rep_len64(2,dimension[1]))
    }

    if(force64 || length(colindices) > 2147483647 ||
       length(colindices) > 2147483647 || max(dimension) > 2147483647 ){
        newx <- new("spam")
        slot(newx,"entries",check=FALSE) <- entries
        entries <- NULL
        slot(newx,"colindices",check=FALSE) <- as.numeric(colindices)
        colindices <- NULL
        slot(newx,"rowpointers",check=FALSE) <- as.numeric(rowpointers)
        rowpointers <- NULL
        slot(newx,"dimension",check=FALSE) <- as.numeric(dimension)
        return(newx)
    } else {
        newx <- new("spam")
        slot(newx,"entries",check=FALSE) <- entries
        entries <- NULL
        slot(newx,"colindices",check=FALSE) <- as.integer(colindices)
        colindices <- NULL
        slot(newx,"rowpointers",check=FALSE) <- as.integer(rowpointers)
        rowpointers <- NULL
        slot(newx,"dimension",check=FALSE) <- as.integer(dimension)
        return(newx)
    }
}


# Create a string containing the matrix size with appropriate unit.
# not planned to be exported.
# If length(x)==1, this represents the number of doubles (or other size).
printSize <- function(x, size=8, digits=2L) {
    mat <- if (length(x)==1) 8*as.numeric(x) else  as.numeric( object.size(x))
    unitpower <- max(min( floor( log(mat, 1024)),3),1)
    paste(round(mat/1024^unitpower, digits=digits), c("Kb","Mb","Gb")[unitpower])
}


.print.spam <- function(x, digits = 2, rowpointer = FALSE, minimal = TRUE, # show row pointers in left margin
              zerosymbol = "'", ...){

    if (minimal) {
      print(x@entries, ...)
      return(invisible(NULL))
    }

    # never true if used via print.spam
    if(is.matrix(x)) {
      x <- as.spam(x) }
    digits <- round(digits)

    stopifnot(is.spam(x), is.numeric(digits), length(digits)==1,
              !rowpointer || (rowpointer && (is.spam(x) || is(x, "spam.chol.NgPeyton"))), # rowpointes are only available for spam
              is.character(zerosymbol), nchar(zerosymbol) == 1)

    if (length(x@entries) == 0 || ((length(x@entries) == 1) && (x@entries[1] == 0))) {
        cat("zero-matrix. class: spam (", .format.spam(x)$name, "). dim: ",  x@dimension[1],
              "x", x@dimension[2], ".", sep = "", fill = TRUE)

        return(invisible(NULL))
    }

    tmp <- round(x@entries, digits = digits)
    x@entries <- ifelse(tmp == 0 & x@entries < 0, 0, tmp) # because sprintf() prints "-0"

    if(digits == 0 || any(grep("\\.", format(x@entries, scientific = FALSE)))) {
        digits <- min(digits, max(nchar(gsub("(.*)(\\.)|([0]*$)","", x@entries))))
    } else {
        digits <- 0 }

    width <- max(nchar(as.character(round(abs(x@entries))))) #left side of comma
    if(any(round(x@entries,digits = digits) < 0)) { # minus sign
        width <- width+1 }
    if(digits > 0) {
        width <- width+1+digits }
    if(width < 3 && ncol(x) >= 10) { #minus sign or two digits for col number
        width <- width+1 }

    if(nrow(x) > getOption("spam.printsize") || ncol(x)*(width+1) > getOption("width")[1]) {
        cat("class: spam (", .format.spam(x)$name, "). dim: ", x@dimension[1], "x",
            x@dimension[2], ". density: ", signif(100*length(x@entries)/prod(x@dimension),3),
            "%.\nrow-wise nonzero elements:",   sep = "", fill = TRUE)
        if(length(x@entries)>10)
            cat(round(x@entries[1:10], digits), "...\n")
        else
            cat(round(x@entries, digits), "\n")

        return(invisible(NULL))
    }

    headerstring <- paste("Matrix of dimension ",x@dimension[1],"x",
                          x@dimension[2], " with (row-wise) nonzero elements", sep ="")
    if (rowpointer)
      headerstring <- paste(headerstring, ",\n number left of the vertical bars denote the respective rowpointer:\n", sep = "")
    else
      headerstring <- paste(headerstring, ":\n", sep = "")

    cat(headerstring)
    cat("    ", formatC(1:ncol(x), width = width), "\n", sep = " ")
    cat("    ", rep("-", ncol(x)*(width+1)), "\n" , sep = "")
    zerostring <- paste(c(rep(" ",width-1), zerosymbol), collapse="")
    rowtmp <- rep(zerostring, x@dimension[2])
    for(i in 1:x@dimension[1]){
        if(rowpointer)
            cat(formatC(x@rowpointers[i], width = 3), "| ", sep = "")
        else
            cat(formatC(i, width = 3), "| ", sep = "")
        if(x@rowpointers[i]==x@rowpointers[i+1]){
            cat(rowtmp, "\n")
            next
        }
        row <- rowtmp
        coli <- x@colindices[x@rowpointers[i]:(x@rowpointers[i+1]-1)]
        row[coli] <- sprintf(paste0("%", width, ".", digits,"f", collapse=""),
                             x@entries[x@rowpointers[i]:(x@rowpointers[i+1]-1)],
                             width = digits+2)
        cat(row, "\n")
    }
    cat(paste0("class: spam (", .format.spam(x)$name, ")\n"))
}


print_nnzpos <- function(x, ...){
  stopifnot(is.spam(x) || is(x, "spam.chol.NgPeyton"))
  if(prod(x@dimension) < getOption("spam.printsize"))
    stop("dim too large for nnzpos=TRUE")
  x@entries <- seq_along(x@entries)
  spam::print.spam(x, digits = 0, minimal = FALSE, ...)
}


print.spam <- function(x, ...) {
  SS <- .format.spam(x, validate = FALSE)
  if (prod(x@dimension) < getOption("spam.printsize")) {
    print(as.matrix.spam(x), ...)
  } else {
    if ( (length(x@entries)==1) && (x@entries[1]==0)) {
      cat("Zero matrix of dimension ",x@dimension[1],"x",
                x@dimension[2],".\n",sep="", fill=TRUE)
    } else {
      .print.spam(x, ...)
    }
  }

  cat(paste0("Class 'spam' (", SS$name, ")\n"))
  invisible(NULL)
}


setMethod("show","spam",  function(object) print.spam(x = object))
setMethod("print","spam", print.spam)

setMethod("length<-","spam",function(x,value) stop("operation not allowed on 'spam' object") )
setMethod("length","spam",function(x) length(x@entries) )
## equivalent to x@rowpointers[x@dimension[1]+1]-1
## however this resulted in a strange error:
## R version 3.2.1 (2015-06-18) -- "World-Famous Astronaut"
## Platform: x86_64-unknown-linux-gnu (64-bit)
## > require(spam)
## Spam version 1.4-0 (2016-08-29) is loaded.
## > gctorture(TRUE)
## > length(spam(1))
## Error: unimplemented type 'integer' in 'coerceToInteger'


setMethod("c","spam", function(x,...){
    if(getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    dimx <- x@dimension

    cx <- .C64("spamcsrdns",
               SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                             "double"),

               nrow = x@dimension,
               entries = x@entries,
               colindices = x@colindices,
               rowpointers = x@rowpointers,

               res = vector_dc("double", prod(dimx)),

               INTENT = c("r", "r", "r", "r",
                   "rw"),
               NAOK = getOption("spam.NAOK"),
               PACKAGE = SS$package
               )$res

    if (length( list(...)) < 1){
        return( cx)
    }else{
        return( c( cx, c(...)))
    }
})

########################################################################
# diag and derivatives
"diag.spam" <- function(x=1, nrow, ncol)  {
  ## structure of function according an older version of base:::diag
  if (is.spam(x)) return( diag.of.spam( x, nrow, ncol))

  if (is.array(x) && length(dim(x)) != 1)
    stop("first argument is array, but not matrix.")

  if (missing(x))
    n <- as.integer(nrow)
  else if (length(x) == 1 && missing(nrow) && missing(ncol)) {
    n <- as.integer(x)
    x <- 1
  }
  else n <- length(x)
  if (!missing(nrow))
    n <- as.integer(nrow)
  if(missing(ncol))
    ncol <- n
  p <- as.integer(ncol)

  m <- min(n, p)

  ###CheckMe: are the following two lines efficient?
  entries <- vector("double", m)
  entries[1:m] <- x
  return(.newSpam(
      entries=entries,
        colindices=1:m,
        rowpointers=c(1:m,rep(m+1,n+1-m)),
        dimension=c(n,p)
    ))
}

spam_diag <- function(x=1, nrow, ncol)  diag.spam(x=x, nrow, ncol)

"diag<-.spam" <-  function(x, value) {
    if(getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    nrow <- x@dimension[1]
    minrc <- min( x@dimension)
    if (length(value)==1)
        value <- rep(value,minrc)
    else if (length(value)!=minrc)
        stop("replacement diagonal has wrong length")

                                        #!7#
    if(2147483647 < minrc + length(x@entries))
        SS <- .format64()

    z <- .C64("setdiagmat",
              SIGNATURE=c(SS$signature, SS$signature, "double", SS$signature,
                          SS$signature, "double", SS$signature),

              nrow = nrow,
              n = minrc,
              a = c(x@entries, vector("double", minrc)),
              ja = c(x@colindices, vector(SS$type, minrc)),
              ia = x@rowpointers,
              diag = value,
              iw = vector_dc(SS$type, nrow), #!8#

              INTENT=c("r", "r", "rw", "rw", "rw", "r", "r"),
              NAOK= getOption("spam.NAOK"),
              PACKAGE = SS$package)


    nz <- z$ia[nrow+1]-1
    entries <- z$a
    colindices <- z$ja
    rowpointers <- z$ia
    rm(z)

    length(entries) <- nz
    length(colindices) <- nz

    return(.newSpam(
        entries=entries,
        colindices=colindices,
        rowpointers=rowpointers,
        dimension=x@dimension
    ))
}

"diag.spam<-" <- get("diag<-.spam")

diag.of.spam <- function(x, nrow, ncol) {
    if(getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    len <- min(x@dimension)

    result <- .C64("getdiag",
                   SIGNATURE=c("double", SS$signature, SS$signature, SS$signature, "double"),

                   a = x@entries,
                   colindices =  x@colindices,
                   rowpointers = x@rowpointers,
                   len = len,
                   diag = vector_dc("double",len),

                   INTENT=c("r", "r", "r", "r", "w"),
                   NAOK = getOption("spam.NAOK"),
                   PACKAGE = SS$package
                   )
    return(result$diag)
}


setMethod("diag","spam",diag.of.spam)
setMethod("diag<-","spam",get("diag<-.spam"))


########################################################################

t.spam <- function(x){
# TODO: Check if we need to switch to 64-bit (in case of too many rows/cols)
    if(getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    dimx <- x@dimension
    nz <- x@rowpointers[dimx[1]+1]-1
    z <- .C64("transpose",
              SIGNATURE=c(SS$signature, SS$signature,
                          "double", SS$signature, SS$signature,
                          "double", SS$signature, SS$signature),
              n = dimx[1],
              m = dimx[2],

              a = x@entries,
              ja = x@colindices,
              ia = x@rowpointers,

              entries = vector_dc("double",nz),
              colindices = vector_dc(SS$type,nz),
              rowpointers = vector_dc(SS$type,dimx[2]+1),

              INTENT = c("r", "r",
                         "r", "r", "r",
                         "w", "w", "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)


    return(.newSpam(
        entries=z$entries,
        colindices=z$colindices,
        rowpointers=z$rowpointers,
        dimension=dimx[2:1]
    ))
}

setMethod("t","spam",t.spam)


########################################################################


"is.spam" <- function(x) is(x,"spam")

"as.spam" <- function(x,  eps = getOption("spam.eps"))
    stop("coercion not defined form this class")

"spam" <- function(x, nrow = 1, ncol = 1, eps = getOption("spam.eps"))
    stop("argument 'x' should be of mode 'numeric' (or 'spam')")

as.spam.spam <- function(x, eps = getOption("spam.eps"))  {
    force64 <- getOption("spam.force64")

    if(force64)
        SS <- .format64()
    else
        SS <- .format.spam(x)

    if (eps < .Machine$double.eps)
        stop("'eps' should not be smaller than machine precision", call.=FALSE)
    dimx <- x@dimension

    z <- .C64("cleanspam",
              SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature, "double"),

              nrow=dimx[1],
              entries=x@entries,
              colindices=x@colindices,
              rowpointers=x@rowpointers,
              eps=eps,

              INTENT=c("r", "rw", "rw", "rw", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package
              )
    nz <- z$rowpointers[dimx[1]+1]-1
    if(nz==0) {
        return(.newSpam(
            dimension=dimx,
            force64 = force64
        ))
    }

    return(.newSpam(
        entries=z$entries[1:nz],
        colindices=z$colindices[1:nz],
        rowpointers=z$rowpointers[1:(dimx[1]+1)],
        dimension=dimx,
        force64=force64
    ))
}

"cleanup" <- function(x, eps = getOption("spam.eps")) {
  if (is.spam(x)) as.spam.spam(x,eps) else x
}

as.spam.matrix <- function(x, eps = getOption("spam.eps")) {
    force64 <- getOption("spam.force64")
    if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)

    dimx <- dim(x)
    nonz <- abs(x) > eps
    nz <- sum(nonz, na.rm = TRUE) + sum(!is.finite(nonz))

    #only zero entries
    if(nz==0) {
        return(.newSpam(
            dimension=dimx,
            force64=force64
        ))
    }

    # EXPLICIT-STORAGE-FORMAT: Depending on the length of x, use 32 or 64-bit:
    SS <- if( force64 || nz > 2147483647 || max(dimx) > 2147483647 ){
              .format64()
          }else{
              .format32
          }

    z <- .C64("spamdnscsr",
              SIGNATURE=c(SS$signature, SS$signature, "double", SS$signature,
                          "double", SS$signature, SS$signature, "double"),

              nrow = dimx[1],
              ncol = dimx[2],
              x = x,
              dimx[1],

              entries = vector_dc("double",nz),
              colindices = vector_dc(SS$type,nz),
              rowpointers = vector_dc(SS$type,dimx[1]+1),
              eps = eps,

              INTENT=c("r", "r", "r", "r",
                       "w", "w", "w", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package
              )

    return(.newSpam( entries=z$entries,
                    colindices=z$colindices,
                    rowpointers=z$rowpointers,
                    dimension=dimx,
                    force64 = force64))
}

"as.spam.numeric" <- function(x, eps = getOption("spam.eps")) {
    if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)

    force64 <- getOption("spam.force64")

    poselements <- (abs(x)>eps)
    if (any(!is.finite(x))) {
        poselements[!is.finite(x)] <- TRUE
    }
    lx <- length(x)
    nz <- sum(as.numeric(poselements))

    # empty matrix
    if (identical(nz,0)){
        return(
            .newSpam(
                # rowpointers = c(1,rep_len64(2,lx)),
                dimension = c(lx,1),
                force64 = force64
            )
        )
    }

    return(
        .newSpam(
            entries = as.double(x[poselements]),
            colindices = rep_len64(1, nz),
            rowpointers = cumsum(c(1, poselements)),
            dimension = c(lx,1),
            force64 = force64
        )
    )
}

as.spam.dist <- function(x, eps = getOption("spam.eps")) {
    if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
    if (any(!is.finite(x))) {
        warning("'NA/NaN/Inf' coerced to zero")
        x[!is.finite(x)] <- 0
    }
    dimx <- attr(x,"Size")
    size <- dimx*(dimx-1)/2

    force64 <- getOption("spam.force64")

    if(force64 || dimx^2 > 2147483647)
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("disttospam",
              SIGNATURE = c(SS$signature, "double",
                            "double", SS$signature, SS$signature,
                            "double"),

              nrow=dimx, #r
              x=as.vector(x,"double"), #r

              entries=vector_dc("double",size), #w
              colindices=vector_dc("integer",size), #w
              rowpointers=vector_dc("integer",dimx+1), #w

              eps=eps, #r

              INTENT = c("r", "r",
                         "w", "w", "w",
                         "r"),
              NAOK=getOption("spam.NAOK"),
              PACKAGE = SS$package
              )
    nz <- z$rowpointers[dimx+1]-1
    if(nz==0) return( .newSpam(
                              # rowpointers = c( 1, rep( 2, dimx)),
                              dimension = c( dimx, dimx),
                              force64 = force64))


    return(.newSpam(
        entries=z$entries[1:nz],
        colindices=z$colindices[1:nz],
        rowpointers=z$rowpointers[1:(dimx+1)],
        dimension=c(dimx,dimx),
        force64 = force64
    ))
}

"as.spam.list" <-  function(x, eps = getOption("spam.eps"))
    spam.list(x,eps)

spam.numeric <- function(x, nrow = 1, ncol = 1, eps = getOption("spam.eps"))  {
    force64 <- getOption("spam.force64")
    if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
    if (any(!is.finite(x))) {
        warning("'NA/NaN/Inf' coerced to zero")
        x[!is.finite(x)] <- 0
    }
    lenx <- length( x)
    if (missing(nrow))
        nrow <- ceiling( lenx/ncol)
    else if (missing(ncol))
        ncol <- ceiling( lenx/nrow)
    dimx <- c(nrow, ncol)
    if (lenx != prod(nrow,  ncol)) {
        if(lenx==1 && abs(x)<eps) {
            return(.newSpam(
                dimension=dimx,
                force64=force64
            ))
        }
        else if(prod(nrow,ncol)%%lenx!=0)
            warning("ncol*nrow indivisable by length(x)")

        x <- rep_len64(x, prod(nrow,ncol))
    }


    dimx <- c(nrow, ncol)

    nz <- sum(as.numeric(abs(x) > eps))

    if(is.na(nz) | is.nan(nz))
        stop("NA or NaN in matrix.")

    if(nz==0) {
        return(.newSpam(
            dimension=dimx,
            force64=force64
        ))
    }

    # EXPLICIT-STORAGE-FORMAT: Depending on the length of x, use 32 or 64-bit:
    if(force64 || nz > 2147483647 || max(dimx) > 2147483647 )
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("spamdnscsr",
              SIGNATURE=c(SS$signature, SS$signature, "double", SS$signature,
                          "double", SS$signature, SS$signature, "double"),

              nrow = dimx[1],
              ncol = dimx[2],
              x = x,
              dimx[1],

              entries = vector_dc("double",nz),
              colindices = vector_dc(SS$type,nz),
              rowpointers = vector_dc(SS$type,dimx[1]+1),
              eps = eps,

              INTENT=c("r", "r", "r", "r",
                       "w", "w", "w", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package
              )

    entries <- z$entries
    colindices <- z$colindices
    rowpointers <- z$rowpointers
    z <- NULL

    return( .newSpam( entries = entries,
                     colindices = colindices,
                     rowpointers = rowpointers,
                     dimension = dimx,
                     force64 = force64))
}

setOldClass(c("dist", "numeric"))


setGeneric("as.spam")
setMethod("as.spam","spam",   as.spam.spam)
setMethod("as.spam","matrix", as.spam.matrix)
setMethod("as.spam","numeric",as.spam.numeric)
setMethod("as.spam","dist",   as.spam.dist)

setGeneric("spam")
setMethod("spam","numeric",spam.numeric)
setMethod("spam","spam",as.spam)



triplet <- function(x, tri=FALSE){
# inverse of spam.list
  dimx <- dim(x)
  if (length(dimx)!=2) stop("'x' should be a matrix like object of dim 2")
  if (is.spam(x)) {
    return(c({ if (tri) list(i=rep(1:dimx[1],diff(x@rowpointers)),
                  j= x@colindices) else list(indices=cbind(rep(1:dimx[1],diff(x@rowpointers)),
                  x@colindices) ) }, list(values=x@entries)
             )
           )
  } else {
    return(c({ if (tri) list(i=rep.int(1:dimx[1],dimx[2]),
                             j=rep.int(1:dimx[2],rep.int(dimx[1],dimx[2]))) else
              list(indices=cbind(rep.int(1:dimx[1],dimx[2]),
                  rep.int(1:dimx[2],rep.int(dimx[1],dimx[2]))))
             } , list(values=c(x))
             )
           )
  }
}

########################################################################

as.matrix.spam <- function(x, ...) {
    if( getOption("spam.force64") || .format.spam(x)$package == "spam64")
        SS <- .format64()
    else
        SS <- .format32

    dimx <- x@dimension
    m <- .C64("spamcsrdns",
              SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                            "double"),

              nrow = dimx[1],
              entries = x@entries,
              colindices = x@colindices,
              rowpointers = x@rowpointers,

              res = vector_dc("double", prod(dimx)),

              INTENT = c("r", "r", "r", "r",
                         "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)$res

    dim(m) <- dimx
    return(m)
}


setMethod("as.matrix","spam",as.matrix.spam)

"as.vector.spam" <- function(x, mode="any"){
    if(getOption("spam.structurebased")) {
        return( as.vector( x@entries, mode=mode))
    } else {
        inefficiencywarning( gettextf("This `as.vector` coercion operation may be inefficient"), prod(dim(x)))
        return( as.vector( as.matrix(x), mode=mode))
    }
}

setMethod("as.vector","spam",as.vector.spam)


########################################################################


"complement.spam" <- function(x){
# this function returns the structure of the zeros of the spam object x.
    force64 <- getOption("spam.force64")

    nrow <- x@dimension[1]
    ncol <- x@dimension[2]
    nnz <- x@rowpointers[nrow+1]-1
    nz <- prod(nrow,ncol) - nnz

                                        # we work through special cases
    if(nz == 0){
        ## print("1")
        return( spam(0, nrow, ncol))
    }
    if(nnz == 1 && x@entries == 0){
        ## print("2")
        return( spam(1, nrow, ncol))
    }

                                        # normal case, proceed to efficient function
    ## print("3")
    if( force64 || .format.spam(x)$package == "spam64" || nz > 2147483647 || nrow > 2147483646)
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("notzero",
              SIGNATURE=c(SS$signature, SS$signature,
                          SS$signature, SS$signature, SS$signature, SS$signature,
                          SS$signature, SS$signature),

              ja = x@colindices,
              ia = x@rowpointers,

              nrow = nrow,
              ncol = ncol,
              nnz = nnz,
              nz = nz,

              colindices = vector_dc( SS$type, nz),
              rowpointers = vector_dc( SS$type, nrow+1),

              INTENT=c("r", "r",
                       "r", "r", "r", "r",
                       "w", "w"),
              NAOK=getOption("spam.NAOK"),
              PACKAGE= SS$package
              )
    ## newx <- new("spam")
    ## slot(newx,"entries",check=FALSE) <- rep.int(1.0,nz)
    ## slot(newx,"colindices",check=FALSE) <- z$colindices
    ## slot(newx,"rowpointers",check=FALSE) <- z$rowpointers
    ## slot(newx,"dimension",check=FALSE) <- x@dimension
    return(.newSpam(
        entries = rep_len64(1.0, nz),
        colindices = z$colindices,
        rowpointers = z$rowpointers,
        dimension = x@dimension,
        force64 = force64
    ))
}


# ASSIGNING:
##########################################################################################

# as S3subsetting causes problems, we eliminate this...
#"[<-.spam" <- function (x, rw, cl,value) {#cat("qq");
#                                          assign.spam(x,rw,cl,value) }


setReplaceMethod("[", signature(x = "spam",
			 i = "missing", j = "missing", value = "ANY"),
	  function (x, i, j, value) {#cat("mm");
                                     assign.spam(x,1:x@dimension[1],1:x@dimension[2],value)})


setMethod("[<-",signature(x="spam",i="vector",j="vector", value = "spam"),
	  function (x, i, j, value)
          {#### FIXME Highly inefficient!
            inefficiencywarning( "Performing inefficient replacement...", prod(dim(value)))
            as.spam(assign.spam(x,i,j,as.matrix(value)))} )
setMethod("[<-",signature(x="spam",i="missing",j="vector", value = "spam"),
	  function (x, i, j, value)
          {#### FIXME Highly inefficient!
            inefficiencywarning( "Performing inefficient replacement...", prod(dim(value)))
            as.spam(assign.spam(x,1:x@dimension[1],j,as.matrix(value)))} )
setMethod("[<-",signature(x="spam",i="vector",j="missing", value = "spam"),
	  function (x, i, j, value)
          {#### FIXME Highly inefficient!
            inefficiencywarning( "Performing inefficient replacement...", prod(dim(value)))
            as.spam(assign.spam(x,i,1:x@dimension[2],as.matrix(value)))} )


setMethod("[<-",signature(x="spam",i="vector",j="missing", value = "ANY"),
	  function (x, i, j, value) {#cat(i);
            assign.spam(x,i,1:x@dimension[2],value)} )

setMethod("[<-",signature(x="spam",i="vector",j="vector", value = "ANY"),
	  function (x, i, j, value) {#cat("vv");
            assign.spam(x,i,j,value)} )

setMethod("[<-",signature(x="spam",i="missing",j="vector", value = "ANY"),
	  function (x, i, j, value) {#cat(j);
            assign.spam(x,1:x@dimension[1],j,value)} )

setMethod("[<-",signature(x="spam",i="matrix",j="missing", value = "ANY"),
	  function (x, i, j, value) {#cat("Mm");
            assign.spam(x,i,NULL,value) })

setMethod("[<-",signature(x="spam",i="matrix",j="matrix",value = "ANY"),
	  function (x, i, j, value) {#cat("MM");
            assign.spam(x,cbind(c(i),c(j)),NULL,value) })

setMethod("[<-",signature(x="spam",i="spam",j="missing", value = "ANY"),
	  function (x, i, j, value)
{
    dimx <- x@dimension
    nrow <- dimx[1]
    ncol <- dimx[2]
    if ( i@dimension[1]>nrow | i@dimension[2]>ncol)
        stop("subscript out of bounds",call.=FALSE)
    if ( ( (i@rowpointers[i@dimension[1]+1]-1) %%length(value))!= 0)
        stop("number of items to replace is not a multiple of replacement length")

    nzmax <- as.integer(min(prod(nrow,ncol), i@rowpointers[i@dimension[1]+1]+x@rowpointers[nrow+1]-2))

    if (length(value)!=  (i@rowpointers[i@dimension[1]+1]-1) )
        value <- rep(value, (i@rowpointers[i@dimension[1]+1]-1) %/%length(value))

    ## z <- .Fortran("subass",
    ##               as.integer(nrow), as.integer(ncol),
    ##               as.double(x@entries),      as.integer(x@colindices),    as.integer(x@rowpointers),
    ##               b=as.double(value),  bj=as.integer(i@colindices), bi=as.integer(i@rowpointers),
    ##               c=vector("double",nzmax),jc=vector("integer",nzmax),ic=vector("integer",nrow+1),
    ##               nzmax=as.integer(nzmax),
    ##               PACKAGE="spam")
    if(getOption("spam.force64") || .format.spam(x)$package != "spam" || .format.spam(i)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("subass",
              ## subroutine subass(nrow,ncol,a,ja,ia,b,jb,ib,c,jc,ic,nzmax)
              SIGNATURE = c(SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            SS$signature ),

              nrow,
              ncol,

              x@entries,
              x@colindices,
              x@rowpointers,

              b = value,
              bj = i@colindices,
              bi = i@rowpointers,

              c = vector_dc( "double", nzmax),
              jc = vector_dc( SS$type, nzmax),
              ic = vector_dc( SS$type, nrow+1),

              nzmax = nzmax,

              INTENT = c("r", "r",
                         "r", "r", "r",
                         "r", "r", "r",
                         "w", "w", "w",
                         "r" ),
              PACKAGE = SS$package)

    cnz <- z$ic[nrow+1]-1

    return(.newSpam(
        entries = z$c[1:cnz],
        colindices = z$jc[1:cnz],
        rowpointers = z$ic[1:(nrow+1)],
        dimension = c(nrow,ncol)))
}      )

setMethod("[<-", signature(x = "spam", i = "ANY", j = "ANY", value = "ANY"),
	  function(x,i,j, value){#  cat(value,class(value))
          stop("invalid or not-yet-implemented 'spam' subassigning")})



"assign.spam" <-
function (x, rw, cl,value)
{
  # we separate into cases where:
  # (A) rw matrix:
  #     1: logical: transformation to spam and extract structure
  #     2: two column matrix: extract (i,j) as given by the lines.
  #     3: all else extract   x[ c( rw)]
  # (B) rw and cl one element: ((i,j)
  # (C) rw and cl vectors:  (i1:i2,j1:j2)               [i1<=i2, j1<=j2]
  #                         (c(i1,...,ii),c(j1,...,jj)) [arbitrary block]

  if (!is.numeric(value)) stop(paste("Assignment of numeric structures only, here",class(value)))

  dimx <- x@dimension
  nrow <- dimx[1]
  ncol <- dimx[2]

  if (is.matrix(rw)) {
    if (is.logical(rw)) {
      return( x[as.spam(rw)] <- value)
    }
    if (dim(rw)[2]==2) {
      ir <- rw[,1]
      jr <- rw[,2]
    } else  {
      ir <- c(rw-1) %% nrow + 1
      jr <- c(rw-1) %/% nrow + 1
      rw <- cbind(ir,jr)
    }
    if ( (min(ir)<1)|(max(ir)>x@dimension[1])|(min(jr)<1)|(max(jr)>x@dimension[2]))
      stop("subscript out of bounds",call.=FALSE)
    if (any(duplicated(cbind(ir,jr))))
      stop("only unique index for subassigning",call.=FALSE)
    nir <- length(ir)
    if ( (nir%%length(value))!= 0)
      stop("number of items to replace is not a multiple of replacement length")
    value <- rep(value, nir%/%length(value))


    ord <- order(ir,jr)
    rw <- rw[ord,,drop=F]
    ## print("1")
    ## bia <- .Fortran("constructia",
    ##                 as.integer(nrow),as.integer(nir),
    ##                 rowpointers=vector("integer",nrow+1),
    ##                 ir=as.integer(c(rw[,1],0)),
    ##                 PACKAGE="spam")$rowpointers
    if( getOption("spam.force64") || nrow+1 > 2147483647 )
        SS <- .format64()
    else
        SS <- .format32

    bia <- .C64("constructia",
                SIGNATURE = c( SS$signature, SS$signature, SS$signature, SS$signature),

                nrow,
                nir,
                rowpointers = vector_dc( SS$type, nrow+1),
                ir = c(rw[,1],0),

                INTENT = c("r", "r", "w", "r"),
                PACKAGE = SS$package)$rowpointers
    nzmax <- as.integer(min(prod(nrow,ncol), nir+x@rowpointers[nrow+1])+2)
    ## z <- .Fortran("subass",
    ##               as.integer(nrow),as.integer(ncol),
    ##               as.double(x@entries), as.integer(x@colindices), as.integer(x@rowpointers),
    ##               b=as.vector(value[ord],"double"),
    ##               bj=as.vector(rw[,2],"integer"),  bi=as.integer(bia),
    ##               entries=vector("double",nzmax),
    ##               colindices=vector("integer",nzmax),
    ##               rowpointers=vector("integer",nrow+1),
    ##               nzmax=as.integer(nzmax),
    ##               NAOK=getOption("spam.NAOK"),PACKAGE="spam")
    if( getOption("spam.force64") ||  .format.spam(x)$package != "spam" || length( bia) >  2147483647 )
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("subass",
                  SIGNATURE = c(SS$signature, SS$signature,
                      "double", SS$signature, SS$signature,
                      "double", SS$signature, SS$signature,
                      "double", SS$signature, SS$signature,
                      SS$signature ),

                  nrow,
                  ncol,

                  x@entries,
                  x@colindices,
                  x@rowpointers,

                  b = as.vector( value[ord], "double" ),
                  bj = as.vector( rw[,2], "integer" ),
                  bi = bia,

                  entries = vector_dc( "double", nzmax),
                  colindices = vector_dc( SS$type, nzmax),
                  rowpointers = vector_dc( SS$type, nrow+1),

                  nzmax = nzmax,

                  INTENT = c("r", "r",
                      "r", "r", "r",
                      "r", "r", "r",
                      "w", "w", "w",
                      "r" ),
                  NAOK=getOption("spam.NAOK"),
                  PACKAGE = SS$package)
    cnz <- z$rowpointers[nrow+1]-1
    if (cnz<0) {
      cat("Negative cnz in subassigning, forced to one. Please report.")
      return( spam(0))
    }

    return(.newSpam(
        entries =  z$entries[1:cnz],
        colindices = z$colindices[1:cnz],
        rowpointers = z$rowpointers,
        dimension = c(nrow,ncol)
        ))

  }
  # negative subsetting:
  if ( max(rw)<0 )    rw <- seq_len( nrow)[rw]
  if ( max(cl)<0 )    cl <- seq_len( ncol)[cl]

  # logical
  if (is.logical(rw))    rw <- seq_len( nrow)[rw]
  if (is.logical(cl))    cl <- seq_len( ncol)[cl]

  # sanity check
  if (length(rw)==0) stop("You should assign at least one element for the rows",call.=FALSE)
  if (length(cl)==0) stop("You should assign at least one element for the columns",call.=FALSE)


  if ( (min(rw)<1)|(max(rw)>x@dimension[1])|(min(cl)<1)|(max(cl)>x@dimension[2]))
    stop("subscript out of bounds",call.=FALSE)

  if (is.vector(rw) && is.vector(cl)) {
    if (any(duplicated(rw))||any(duplicated(cl)))
      stop("only unique index for subassigning",call.=FALSE)

    nrw <- length(rw)   # length returns an integer, so is a product therof
    ncl <- length(cl)
    bnz <- nrw*ncl

    if ( (bnz%%length(value))!= 0)
      stop("number of items to replace is not a multiple of replacement length")

    # we pack the value into a vector _row by row_
    value <- c(t(array(as.double(value),c(nrw,ncl))[order(rw),order(cl)]))

    bia <- vector("integer",nrow)  # bia has size of nrow + 1
    bia[rw] <- ncl        # in each row we have ncl new objects
    bia <- as.integer(c(1,cumsum(bia)+1))

    # we construct now a sparse matrix containing the "value" at positions rw and cl.
    # then we use the subassign function.
    nzmax <- as.integer(min(prod(nrow,ncol), bnz+x@rowpointers[nrow+1])+2)

    ## print("2")
    ## z <- .Fortran("subass",
    ##               as.integer(nrow),as.integer(ncol),
    ##               as.double(x@entries), as.integer(x@colindices) ,as.integer(x@rowpointers),
    ##               b=as.double(value),
    ##               bj=as.integer(rep(sort(as.integer(cl)),nrw)),
    ##               bi=as.integer(bia),
    ##               entries=vector("double",nzmax),colindices=vector("integer",nzmax),
    ##               rowpointers=vector("integer",nrow+1),
    ##               nzmax=as.integer(nzmax),
    ##               NAOK=getOption("spam.NAOK"),PACKAGE="spam")
     if( getOption("spam.force64") ||  .format.spam(x)$package != "spam" || length( bia) >  2147483647 )
        SS <- .format64()
    else
        SS <- .format32
    z <- .C64("subass",
              SIGNATURE = c(SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  SS$signature ),

              nrow,
              ncol,

              x@entries,
              x@colindices,
              x@rowpointers,

              b = value,
              bj = rep_len64(sort(as.integer(cl)),nrw*length(cl)),
              bi = bia,

              entries = vector_dc( "double", nzmax),
              colindices = vector_dc( SS$type, nzmax),
              rowpointers = vector_dc( SS$type, nrow+1),

              nzmax = nzmax,

              INTENT = c("r", "r",
                  "r", "r", "r",
                  "r", "r", "r",
                  "w", "w", "w",
                  "r" ),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)
    cnz <- z$rowpointers[nrow+1]-1

    return(.newSpam(
        entries = z$entries[1:cnz],
        colindices = z$colindices[1:cnz],
        rowpointers = z$rowpointers,
        dimension = c(nrow,ncol)
        ))
  }
  stop("invalid or not-yet-implemented 'spam' subsetting")
}


#!8#
# x:spam, y:matrix, returns matrix
.spam.matmul.mat <- function(x,y)
{
    nrow <- x@dimension[1]
    ncol <- x@dimension[2]
    yrow <- dim(y)[1]
    ycol <- dim(y)[2]

    # EXPLICIT-STORAGE-FORMAT
    SS <- .format.spam(x)
    if( getOption("spam.force64")|| as.numeric(nrow)*as.numeric(ycol) > 2147483647)
        SS <- .format64()

    if(yrow != ncol) stop("not conformable for multiplication")
    z <- .C64("amuxmat",
              SIGNATURE=c(SS$signature, SS$signature, SS$signature, "double",
                  "double", "double", SS$signature, SS$signature),

              nrow,
              yrow,
              ycol,
              y,

              y = vector_dc("double",nrow*ycol),
              x@entries,
              x@colindices,
              x@rowpointers,

              INTENT=c("r", "r", "r", "r",
                       "w", "r", "r", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)$y
    dim(z) <- c(nrow,ycol)
    return(z)
}

.spam.matmul <- function(x,y) {
    # --- CHANGED ---
    # Refactored. -> See .spam.matmul.vector

  if (is.matrix(y)) y <- as.spam(y)
  if (is.matrix(x)) x <- as.spam(x)

  #matrix multiply two sparse spam matrices
  xn <- x@dimension[1]
  xm <- x@dimension[2]
  yl <- y@dimension[2]
  if(xm != y@dimension[1])
    stop("matrices not conformable for multiplication")

  # EXPLICIT-STORAGE-FORMAT
  SS <- .format.spam(x, y)
  if( getOption("spam.force64") || as.numeric(xn)*as.numeric(yl) > 2147483647)
        SS <- .format64()

  z <- .C64("amubdg",
            SIGNATURE = rep(SS$signature, 10),

            xn,
            xm,
            yl,
            x@colindices,

            x@rowpointers,
            y@colindices,
            y@rowpointers,
            unused2=vector_dc(SS$type, xn),

            nz = vector_dc(SS$type,1),
            unused=vector_dc(SS$type, yl),

            INTENT=c("r", "r", "r", "r",
                     "r", "r", "r", "w",
                     "w", "w"),
            NAOK = getOption("spam.NAOK"),
            PACKAGE = SS$package)
  nzmax <- z$nz
  z <- NULL

  # EXPLICIT-STORAGE-FORMAT
  # As we now know the number of entries, we can decide again if we need 64-bit
  SS <- .format.spam(x, y)
  if( getOption("spam.force64") || nzmax > 2147483647)
        SS <- .format64()

  z <- .C64("amub",
            SIGNATURE=c(SS$signature, SS$signature, SS$signature, "double",
                        SS$signature, SS$signature, "double", SS$signature,
                        SS$signature, "double", SS$signature, SS$signature,
                        SS$signature, SS$signature, SS$signature),

            xn,
            yl,
            1L,
            x@entries,

            x@colindices,
            x@rowpointers,
            y@entries,
            y@colindices,

            y@rowpointers,
            entries = vector_dc("double",nzmax),
            colindices = vector_dc(SS$type,nzmax),
            rowpointers = vector_dc(SS$type,xn+1),

            nzmax,
                                        # TODO: Check if we can get rid of this argument and allocate memory in Fortran:
            vector_dc(SS$type, yl),
            ierr = vector_dc(SS$type, 1),

            INTENT=c("r", "r", "r", "r",
                     "r", "r", "r", "r",
                     "r", "w", "w", "w",
                     "r", "r", "w"),
            NAOK = getOption("spam.NAOK"),
            PACKAGE = SS$package)
  nz <- z$rowpointers[xn+1]-1
  if(z$ierr != 0) stop("insufficient space for sparse matrix multiplication")


  if(nz==0){#trap zero matrix
    return(.newSpam(
      dimension=c(xn,yl)
    ))
  }

   entries <- z$entries
   colindices <- z$colindices
   rowpointers <- z$rowpointers
   z <- NULL

   length(entries) <- nz
   length(colindices) <- nz

   z <- .C64("sortrows",
             SIGNATURE=c(SS$signature, "double", SS$signature, SS$signature),

             xn,
             entries=entries,
             colindices=colindices,
             rowpointers=rowpointers,

             INTENT=c("r", "rw", "rw", "rw"),
             NAOK = getOption("spam.NAOK"),
             PACKAGE = SS$package)

  return(.newSpam(
    entries=z$entries,
    colindices=z$colindices,
    rowpointers=z$rowpointers,
    dimension=c(xn,yl)
  ))
}


.spam.matmul.vector <- function(x,y)
{
    # --- CHANGED ---
    # Refactored the function .spam.matmul into the functions
    # .spam.matmul and .spam.matmul.vector

    # if we have x*Y, we calculate t(t(Y)*x)
    # Use "is.spam(y)" instead of "is.vector(x)" because the later might be
    # misleading in case that x has any attributes attached.
    if(is.spam(y)) {
        A <- t(y)
        b <- x
    } else {
        A <- x
        b <- y
    }
    SS <- .format.spam(A)

    nrow <- A@dimension[1]
    ncol <- A@dimension[2]
    if(length(b) != ncol)  stop("not conformable for multiplication")
    z <- .C64("amux",
              NAOK = getOption("spam.NAOK"),
              SIGNATURE=c(SS$signature, "double", "double", "double",
                          SS$signature, SS$signature),

          nrow,
          b,
          y = vector_dc("double",nrow),
          A@entries,

          A@colindices,
          A@rowpointers,

          INTENT=c("r", "r", "w", "r", "r", "r"),
          PACKAGE = SS$package)$y

    if(is.spam(y))
        dim(z) <- c(1,nrow)
    else
        dim(z) <- c(nrow,1)
    return(z)
}

setMethod("%*%",signature(x="spam",y="spam"),    .spam.matmul)
setMethod("%*%",signature(x="spam",y="matrix"),  .spam.matmul.mat)
setMethod("%*%",signature(x="spam",y="numeric"), .spam.matmul.vector)
setMethod("%*%",signature(x="matrix",y="spam"),  .spam.matmul)
setMethod("%*%",signature(x="numeric",y="spam"), .spam.matmul.vector)

upper.tri.spam <- function(x,diag=FALSE)
  {
    dimx <- x@dimension
    nrow <- dimx[1]

    SS <- .format.spam(x)
    z <- .C64("getu",
              SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature),
              nrow,
              x@entries,
              x@colindices,
              x@rowpointers,

              entries = x@entries,
              colindices = x@colindices,
              rowpointers = x@rowpointers,

              INTENT = c("r", "r", "r", "r",
                  "w", "w", "w"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package )

    nz <- z$rowpointers[dimx[1]+1]-1

    if (!diag) {
        z <- .C64("getdia",
                  SIGNATURE = c(SS$signature, SS$signature, SS$signature,
                                    "double", SS$signature, SS$signature,
                      SS$signature, "double", SS$signature, SS$signature),

                  n = nrow,
                  m = nrow,
                  job = 1,

                  entries = z$entries[1:nz],
                  colindices = z$colindices[1:nz],
                  rowpointers = z$rowpointers,

                  len = nrow,
                  diag = vector_dc("double", nrow),
                  idiag = vector_dc(SS$type, nrow),
                  ioff = 0,

                  INTENT = c("r", "r", "r",
                      "rw", "rw", "rw",
                      "w", "w", "w", "r"),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = SS$package
                  )
        nz <- z$rowpointers[nrow+1]-1
    }
    if(getOption("spam.trivalues")){
        return(.newSpam(
            entries = z$entries[1:nz],
            colindices = z$colindices[1:nz],
            rowpointers = z$rowpointers,
            dimension = dimx,
            ))
    }else{
        return(.newSpam(
            entries = rep_len64(1,nz),
            colindices=z$colindices[1:nz],
            rowpointers=z$rowpointers,
            dimension=dimx
            ))
    }
  }

lower.tri.spam <- function(x,diag=FALSE)
{
    dimx <- x@dimension
    nrow <- dimx[1]
    SS <- .format.spam(x)
    z <- .C64("getl",
              SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature),
              nrow,
              x@entries,
              x@colindices,
              x@rowpointers,

              entries = x@entries,
              colindices = x@colindices,
              rowpointers = x@rowpointers,

              INTENT = c("r", "r", "r", "r",
                  "w", "w", "w"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package )
    nz <- z$rowpointers[nrow+1]-1

    if (!diag) {
        z <- .C64("getdia",
                  SIGNATURE = c(SS$signature, SS$signature, SS$signature,
                      "double", SS$signature, SS$signature,
                      SS$signature, "double", SS$signature, SS$signature),

                  n = nrow,
                  m = nrow,
                  job = 1,

                  entries = z$entries[1:nz],
                  colindices = z$colindices[1:nz],
                  rowpointers = z$rowpointers,

                  len = nrow,
                  diag = vector_dc("double", nrow),
                  idiag = vector_dc(SS$type, nrow),
                  ioff = 0,

                  INTENT = c("r", "r", "r",
                      "rw", "rw", "rw",
                      "w", "w", "w", "r"),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = SS$package
                  )
        nz <- z$rowpointers[nrow+1]-1
    }

    if(getOption("spam.trivalues")){
        return(.newSpam(
            entries = z$entries[1:nz],
            colindices = z$colindices[1:nz],
            rowpointers = z$rowpointers,
            dimension = dimx,
            ))
    }else{
        return(.newSpam(
            entries = rep_len64(1,nz),
            colindices=z$colindices[1:nz],
            rowpointers=z$rowpointers,
            dimension=dimx
            ))
    }
}



setGeneric("upper.tri")
setMethod("upper.tri","spam",upper.tri.spam)
setGeneric("lower.tri")
setMethod("lower.tri","spam",lower.tri.spam)


# fields uses the construct of vector representation for a diagonal matrix.

# Create a special matrix multiply for diagonal matrices.
# Diagonal matrix assumed to be just a vector.
# NOTE: this is not a symmetric operation:
#  when a left vector is given it is a diagonal matrix
#  when a right vector is given it is a vector.
#
.spam.diagmulmat <- function(x,y){
    nrow <- y@dimension[1]
    if(length(x) != nrow)
        stop("not conformable for multiplication")

    SS <- .format.spam(y)

    z <- .C64("diagmua",
              SIGNATURE = c(SS$signature,
                  "double", SS$signature, "double"),

              nrow,

              entries = y@entries,
              y@rowpointers,
              as.vector(x,"double"),

              INTENT = c("r",
                  "rw", "r", "r"),
              NAOK=getOption("spam.NAOK"),
              PACKAGE = SS$package)$entries

    y@entries <- z
    return(y)
}

.spam.diagaddmat <- function(x,y){
      ## subroutine diagaddmat (nrow, n, a, ja, ia, diag, iw)
    nrow <- y@dimension[1]
    minrc <- min( y@dimension)
    if(length(x) != minrc) stop("not conformable for addition")

    SS <- .format.spam(y)

    z <- .C64("diagaddmat",
              SIGNATURE = c( SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double", SS$signature),

              nrow = nrow,
              n = minrc,

              a = c(y@entries, rep_len64(0,minrc)),
              ja = c(y@colindices, rep_len64(0,minrc)),
              ia = y@rowpointers,

              diag = x,
              iw = vector_dc(SS$type, nrow),

              INTENT = c("r", "r",
                  "rw", "rw", "rw",
                  "r", "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package )

    nz <- z$ia[nrow+1]-1

    return(.newSpam(
        entries = z$a[1:nz],
        colindices = z$ja[1:nz],
        rowpointers = z$ia,
        dimension = y@dimension))
}

setGeneric("%d*%",function(x,y,...)standardGeneric("%d*%"))

setMethod("%d*%",signature(x="matrix",y="ANY"),       function(x,y){x%*%y} )
setMethod("%d*%",signature(x="numeric",y="matrix"),   function(x,y){x*y} )
setMethod("%d*%",signature(x="numeric",y="numeric"),  function(x,y){cbind(x*y)} )

setMethod("%d*%",signature(x="spam",y="spam"),    .spam.matmul )
setMethod("%d*%",signature(x="spam",y="ANY"),     .spam.matmul )
setMethod("%d*%",signature(x="numeric",y="spam"), .spam.diagmulmat )


setGeneric("%d+%",function(x,y,...)standardGeneric("%d+%"))
setMethod("%d+%",signature(x="matrix",y="ANY"),      function(x,y){ x+y } )
setMethod("%d+%",signature(x="numeric",y="matrix"),  function(x,y){ diag(x)+y} )
setMethod("%d+%",signature(x="numeric",y="numeric"), function(x,y){ diag(x)+y} )

setMethod("%d+%",signature(x="spam",y="spam"),     function(x,y){ x+y})
setMethod("%d+%",signature(x="spam",y="ANY"),      function(x,y){ x+y})
setMethod("%d+%",signature(x="numeric",y="spam"),  .spam.diagaddmat )


all.equal.spam <- function (target, current, tolerance = .Machine$double.eps^0.5,
    scale = NULL, check.attributes = FALSE,...)
{
    if (check.attributes)
        warning("attributes are not supported for 'spam' objects. Ignoring 'check.attributes' argument")
    if (!is.spam(target)) stop("'target' should be of class 'spam'")
    if (!is.spam(current)) {
        return(paste("target is spam, current is ", data.class(current), sep = ""))
    }
    msg <- NULL
    lt <- length(target)
    lc <- length(current)
    if (lt != lc) {
      return(paste("Lengths (", lt, ", ", lc, ") differ", sep = ""))
    }
    dt <- target@dimension
    dc <- current@dimension
    if ( !all( dt == dc ))
      return(paste("Dimensions ([",dt[1],",",dt[2],"], [",
                    dc[1],",",dc[2], "]) differ", sep = ""))
    # --- CHANGED ---
    # Add suppressWarnings
    tmp <- suppressWarnings(sum(target@colindices != current@colindices))
    if ( tmp > 0)
      msg <- c(msg,paste("Column-sparsity structure differ (at least",
                    tmp,"instance(s))"))

    tmp <- suppressWarnings(sum(target@rowpointers != current@rowpointers))
    if ( tmp > 0)
      msg <- c(msg,paste("Row-sparsity structure differ (at least",
                    tmp,"instance(s))"))

    what <- if (is.null(scale)) {
      scale <- mean(abs(target@entries))
      if (is.finite(scale) && scale > tolerance) {
        "relative"
      } else {
        scale <- 1
        "absolute"
      }
    } else {
        "scaled"
    }
    xy <- suppressWarnings(mean(abs(target@entries - current@entries) / scale))
    if (is.na(xy) || xy > tolerance)
        msg <- c(msg,paste("Mean", what, "difference:", format(xy)))
    if (is.null(msg))
        TRUE
    else msg
}


isSymmetric.spam <- function(object, tol = 100 * .Machine$double.eps, ...)
{
  # very similar to is.Symmetric.matrix
  test <-  all.equal.spam(object, t.spam(object), tolerance = tol, ...)

  # Possibility that structure is different but not contents

  if (!isTRUE(test)) {
    object <- as.spam.spam(object)
    test <-  all.equal.spam(object, t.spam(object), tolerance = tol, ...)
  }
  isTRUE(test)

}


setMethod("all.equal",signature(target="spam",current="spam"), all.equal.spam )
setMethod("all.equal",signature(target="matrix",current="spam"),
          function (target, current, tolerance = .Machine$double.eps^0.5,
                    scale = NULL, check.attributes = FALSE,eps = getOption("spam.eps"),...)
{
    if (check.attributes)
      warning("attributes are not supported for 'spam' objects. Ignoring 'check.attributes' argument")
    msg <- NULL

    dimx <- dim(target)
    nz <- length(target)
     ## z <- .Fortran("spamdnscsr", nrow = as.integer(dimx[1]), ncol = as.integer(dimx[2]),
     ##    x = as.double(target), as.integer(dimx[1]), entries = vector("double",
     ##        nz), colindices = vector("integer", nz), rowpointers = vector("integer",
     ##        dimx[1] + 1), eps = as.double(eps),
     ##              NAOK = getOption("spam.NAOK"), PACKAGE = "spam")
    if( getOption("spam.force64") || .format.spam(current)$package != "spam" || prod(dimx) > 2147483647)
        SS <- .format64()
    else
        SS <- .format32
    z <- .C64("spamdnscsr",
              SIGNATURE = c(SS$signature, SS$signature, "double", SS$signature,
                  "double", SS$signature, SS$signature, "double"),

              nrow = dimx[1],
              ncol = dimx[2],
              x = target,
              dimx[1],
              entries = vector_dc( "double", nz),
              colindices = vector_dc( SS$type, nz),
              rowpointers = vector_dc( SS$type, dimx[1] + 1),
              eps = eps,

              INTENT = c("r", "r", "r", "r",
                  "w", "w", "w", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)

    lt <- z$rowpointers[dimx[1] + 1] - 1
    lc <- length(current)

    if (lt != lc) {
        return(paste("Lengths (", lt, ", ", lc, ") differ", sep = ""))
    }
    dt <- dim(target)
    dc <- current@dimension
    if ( !all( dt == dc ))
        return(paste("Dimensions ([",dt[1],",",dt[2],"], [",
                     dc[1],",",dc[2], "]) differ", sep = ""))
    tmp <- sum(z$colindices[1:lt] != current@colindices)
    if ( tmp>0)
        msg <- c(msg,paste("Column-sparsity structure differ (at least",
                           tmp,"instance(s))"))

    tmp <- sum(z$rowpointers != current@rowpointers)
    if ( tmp>0)
        msg <- c(msg,paste("Row-sparsity structure differ (at least",
                           tmp,"instance(s))"))

    xy <- mean(abs(z$entries[1:lt] - current@entries))
    what <- if (is.null(scale)) {
        xn <- mean(abs(z$entries))
        if (is.finite(xn) && xn > tolerance) {
            xy <- xy/xn
            "relative"
        }
        else "absolute"
    }
    else {
        xy <- xy/scale
        "scaled"
    }
    if (is.na(xy) || xy > tolerance)
        msg <- c(msg,paste("Mean", what, "difference:",
            format(xy)))
    if (is.null(msg))
        TRUE
    else msg

}
 )
setMethod("all.equal",signature(target="spam",current="matrix"), function(target, current, ...)
     { all.equal(current, target, ...) })
setMethod("isSymmetric","spam", isSymmetric.spam)

# HEADER ####################################################
# This is file spam/tests/testthat/test-constructors.R.     #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

  # source("tests/testthat/helper.R")#####
## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-constructors.R")

options(spam.printsize = 60)

set.seed(14)
n <- 7
ln <- 50
A <- spam(0,n,n)
is <- sample(n, ln, replace = TRUE)
js <- sample(n, ln, replace = TRUE)

A[ unique( cbind(is,js))[1:16,] ] <- 1:8


re <- A@rowpointers
rowpointers(A) <- re

# following will case error, thus the `try`
test_that("test for error messages", {
    r <- re; r[1:2] <- rev(r[1:2]);
    expect_error(rowpointers(A) <- r,
                 "row pointers are not monotone increasing")
    r <- re; r[n+1] <- 2;
    expect_error(rowpointers(A) <- r,
                 "row pointers are not monotone increasing")
    r <- re; r[1] <- 0;
    expect_error(rowpointers(A) <- r,
                 "first element of row pointers is < 1")
    r <- re; r[n+1] <- 20;
    expect_error( rowpointers(A) <- r  )
    r <- c(rep(1,n),n+1);
    expect_error( rowpointers(A) <- r  )


    ce <- A@colindices
    colindices(A) <- ce

    ## TODO expect error or not ?? currrently does not return an error
    ## r <- ce; r[1:4] <- rev(r[1:4]);
    ## expect_error( colindices(A) <- r  )
    r <- ce; r[1] <- 0;
    expect_error( colindices(A) <- r  )
    r <- ce; r[1] <- 20;
    expect_error( colindices(A) <- r  )

    entries(A) <- A@entries
    expect_error( entries(A) <- as.logical(A@entries))
    expect_error( entries(A) <- c(r,1))


    expect_error( entries(A) <- r[-1])
    expect_error( dimension(A) <- c(1,2))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-definitions.R.      #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-definitions.R")


################## PR 18272 #################
# https://bugs.r-project.org/show_bug.cgi?id=18272
# Patch by Mike Chirico

Sys.setenv("_R_CHECK_LENGTH_1_LOGIC2_" = "true")
# all.equal(c(1, 1), c(1.01, 1.01), scale = c(.01, .01))  # patch 2022.01.03
all.equal.spam(diag.spam(2), 1.01*diag.spam(2), scale = c(.01, .01))  # patch 2022.01.03


Sys.unsetenv("_R_CHECK_LENGTH_1_LOGIC2_")
# all.equal(c(1, 1), c(1.01, 1.01), scale = c(.01, .01)) # patch 2022.01.03
all.equal.spam(diag.spam(2), 1.01*diag.spam(2), scale = c(.01, .01))  # patch 2022.01.03




###################################
# general stuff, mainly from help of all.equal.spam
test_that("all.equal.spam", {
  obj <- diag.spam(2)
  obj[1,2] <- .Machine$double.eps

  expect_equivalent( all.equal( diag.spam(2), obj),"Lengths (2, 3) differ")
  expect_equivalent( all.equal( t(obj), obj), c("Column-sparsity structure differ (at least 1 instance(s))",
      "Row-sparsity structure differ (at least 1 instance(s))"))
  expect_equivalent( all.equal( t(obj), obj*1.1), c("Column-sparsity structure differ (at least 1 instance(s))",
      "Row-sparsity structure differ (at least 1 instance(s))", "Mean relative difference: 0.1" ))

  # We can compare a spam to a matrix
  expect_equivalent( all.equal(diag(2),diag.spam(2)), TRUE)


 })

# HEADER ####################################################
# This is file spam/tests/testthat/test-dim.R.              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-dim.R")


# simple tests:
########################################################################


# construct matrices:
n <- 10
m <- 15

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
tt[tt<0] <- 0

ss <- as.spam(tt)



test_that("dim", {
    spamtest_eq(ss,tt)
    
    dim(ss) <- c(m,n)
    dim(tt) <- c(m,n)
    spamtest_eq(ss,tt)
    
    dim(ss) <- c(m*n,1)
    dim(tt) <- c(m*n,1)
    spamtest_eq(ss,tt)
    
    dim(ss) <- c(1, m*n)
    dim(tt) <- c(1, m*n)
    spamtest_eq(ss,tt)
    
    expect_error( dim(ss) <- c(-1, -m*n),
                 "Indices need to be positive")
    expect_error( dim(ss) <- c(1, m, n),
                 "dims should be of length 1 or 2")
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-diff.R.             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-diff.R")
     

options(spam.structurebased=FALSE) # test for equivalence!

n <- 10
x <- array(rnorm(n^2),c(n,n))

test_that("diff", {
    spamtest_eq(diff(x), diff(as.spam(x)))
    spamtest_eq(diff(x,d=2), diff(as.spam(x), d=2))
    spamtest_eq(diff(x,d=4), diff(as.spam(x), d=4))
    spamtest_eq(diff(x,2, d=2), diff(as.spam(x),2, d=2))
# The following works on R >=4.6
#    expect_equal(diff(x,4, d=4), as.matrix(diff(as.spam(x),4, d=4)))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-solve.R.            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################
rm(list = ls())
source("helper.R")

library("testthat")
library("spam", lib.loc = LIB.LOC)

context("test-solve.R")


# construct spd matrices (should be at least 3x3):
n <- 10


set.seed(11)
tt <- matrix(rnorm(n*n),n,n)
tt <- t(tt) %*% tt
tt[tt<0] <- 0
# I have seen that with R version 2.4.0 Patched (2006-11-25 r39997)
# on i486-pc-linux-gnu, tt is not symmetric...
tt <- tt-(tt-t(tt))/2


ss <- as.spam(tt)

css <- chol(ss)
ctt <- chol(tt[ordering(css),ordering(css)])

# solving system
test_that("'solve' and derivatives", {
    b <- rnorm(n)

    spamtest_eq(solve(ss),solve(tt))
    spamtest_eq(solve(ss,b),solve(tt,b))

    spamtest_eq(t(as.spam(css))%*%as.spam(css), t(ctt)%*%ctt)
    spamtest_eq(t(as.spam(css))%*%as.spam(css), tt[ordering(css),ordering(css)])
    spamtest_eq((t(as.spam(css))%*%as.spam(css))[ordering(css,inv=T),ordering(css,inv=T)], tt)

    spamtest_eq(backsolve(css,forwardsolve(css,b[ordering(css,inv=T)]))[ordering(css)],
                backsolve(ctt,forwardsolve(t(ctt),b),n))
#### ,n as patch

    spamtest_eq(backsolve(css,b[ordering(css,inv=T)])[ordering(css)],
                backsolve(ctt,b,n))
#### ,n as patch

    spamtest_eq(forwardsolve(css,b[ordering(css,inv=T)])[ordering(css)],
                forwardsolve(t(ctt),b))
    spamtest_eq(forwardsolve(css,b)[ordering(css)],
                forwardsolve(t(ctt),b[ordering(css)]))

    spamtest_eq(forwardsolve(css,tt[ordering(css,inv=T),])[ordering(css),],
                forwardsolve(t(ctt),tt))
})

test_that("option 'chol.update'", {
    ss1 <- ss+diag.spam(dim(ss)[1])
    spamtest_eq( chol(ss), update.spam.chol.NgPeyton(css, ss))

    spamtest_eq( chol(ss, Rstruct=css), update.spam.chol.NgPeyton(css, ss))

    sel <- which(ss[1,,drop=TRUE]!=0)
    ss1[1,sel[-1]] <- 0
    ss2 <- ss
    ss2[n,1] <- .1
    options(spam.cholsymmetrycheck=FALSE)
    spamtest_eq(as.spam(update.spam.chol.NgPeyton(css,ss1)), as.spam( chol(ss1)))
    spamtest_diff(as.spam(update.spam.chol.NgPeyton(css,ss1)), as.spam( chol(ss2)))
})

# spam.options(trivalues=TRUE)
# spam.options(trivalues=FALSE)

options(spam.cholsymmetrycheck=TRUE)

# methods for spam.chol.NgPeyton
test_that("'spam.chol.NgPeyton'", {
    spamtest_eq(as.spam(css), ctt)
    spamtest_eq(as.matrix(css), as.matrix(ctt))
    spamtest_eq(diag(css), diag(ctt))
    spamtest_eq(length(css), length(ctt[ctt!=0]))
    spamtest_eq(dim(css), dim(ctt))
    spamtest_eq(c(css), c(ctt))
})


test_that("option 'cholupdatesingular'\n", {
    ss3 <- spam(rep(1,4),2)
    ch3 <- chol( ss3+diag.spam(2))
    options(spam.cholupdatesingular="null")
    spamtest_eq(is.null(update(ch3, ss3)),TRUE)
    options(spam.cholupdatesingular="warning")
    options(warn=1)
    expect_warning(update(ch3, ss3),
                   "Singularity problem")
    options(spam.cholupdatesingular="error")
    expect_error(update(ch3, ss3),
                 "Singularity problem")
    options(spam.cholupdatesingular="NULL")
    expect_error(update(ch3, ss3),
                 "'cholupdatesingular' should be 'error', 'null' or 'warning'")
})




# determinants
test_that("'det' and derivatives", {
    spamtest_eq(det(ss),det(tt))
    spamtest_eq(det(ss,log=T),det(tt,log=T))
    spamtest_eq(determinant(ss)$mod,determinant(tt)$mod)
    spamtest_eq(determinant(ss,log=F)$mod,determinant(tt,log=F)$mod)
    spamtest_eq(det(chol(ss)),det(chol(tt)))
    spamtest_eq(2*sum(log(diag(css))), determinant(tt)$modulus)

    expect_error(det(spam(1,2,1)), "non-square")
    # det now gives an error if neg definite! (2.8)
    expect_error(det(spam(-1)), "not positive definit")

    # since 2.8: warning when passing arguments to det
    expect_warning(det(spam(1), pivot=FALSE), "will become defunct in the future")

})

# orderings and derivatives
test_that("'ordering' and derivatives", {
    tt5 <- matrix(c( 2,0,2,0,4,0,2,0,3),3)
    ss5 <- spam(  c( 2,0,2,0,4,0,2,0,3),3)
    spamtest_eq(ordering(tt5),1:3)
    spamtest_eq(ordering(ss5),1:3)
    spamtest_eq(ordering(tt5,inv=T),3:1)
    spamtest_eq(ordering(ss5,inv=T),3:1)
    spamtest_eq(ordering(chol(ss5)),c(2,3,1))
    spamtest_eq(ordering(chol(ss5),inv=T),c(3,1,2))
})



# spam triangular solves
test_that("triangular solves", {
    ## We need to generate a upper triangular matrix first.
    ctt <- chol(tt)
    css <- as.spam(ctt)
    b <- rnorm(nrow(tt))

    # a good way to test that we do not use the lower triangular part:
    ctt[2:10,1] <- NA

                                        # Recall:
    spamtest_eq(backsolve(ctt,forwardsolve(t(ctt),b),n),
                solve(tt,b))

                                        # Now do testing:
    spamtest_eq(forwardsolve(t(css),b), forwardsolve(t(ctt),b))

    spamtest_eq(forwardsolve(ss,b), forwardsolve(tt,b))

    cs <- ss
    cs[upper.tri(cs)] <- 0
    spamtest_eq(forwardsolve( cs ,b), forwardsolve( ss,b))


#### ,n as patch
    spamtest_eq(backsolve(css,b), backsolve(ctt,b, n))
    spamtest_eq(backsolve(ss,b), backsolve(tt,b, n))
    spamtest_eq(backsolve(t(cs),b), backsolve(tt,b, n))

    spamtest_eq(backsolve(css,forwardsolve(t(css),b)),
                backsolve(ctt,forwardsolve(t(ctt),b), n))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-overall.R.          #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam", lib.loc = LIB.LOC)


context("test-overall.R")




options(spam.structurebased = FALSE)

# construct matrices (should be at least 3x5, with n<m):
n <- 10
m <- 15

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
tt[tt<0] <- 0
tt2 <- matrix(rnorm(m*n),m,n)
tt2[tt2<0] <- 0
tt3 <- t(tt2)
tt4 <- tt  %*% t(tt)

ss <- as.spam(tt)
ss2 <- as.spam(tt2)
ss3 <- as.spam(tt3)
ss4 <- ss  %*% t(ss)


test_that("simple math", {
    kf <- matrix(rnorm(m*n),n,m)
    ks <- as.spam(kf)
    expect_equal(max(ks),max(kf))
    expect_equal(range(ks),range(kf))
    spamtest_eq(log(ks-2*min(ks)),log(kf-2*min(kf)))
    spamtest_eq(cos(ks),cos(kf))
    spamtest_eq(round(ks,2),round(kf,2))
})

test_that("t()", {
    rvec <- 1:n
    spamtest_eq( t(ss), t(tt))
    spamtest_eq( t(matrix(rvec,nrow=1)),t(spam(rvec,nrow=1)))
    spamtest_eq( t(matrix(rvec,ncol=1)),t(spam(rvec,ncol=1)))
})

test_that("add/subtract", {
    spamtest_eq(ss+ss,tt+tt)
    spamtest_eq(ss-ss,tt-tt,rel=FALSE)
    spamtest_eq(ss+tt,tt+tt)
    spamtest_eq(tt+ss,tt+tt)
    spamtest_eq(ss-tt,tt-tt,rel=FALSE)
    spamtest_eq(tt-ss,tt-tt,rel=FALSE)

    spamtest_eq(ss+ss3,tt+tt3)
    spamtest_eq(ss-ss3,tt-tt3)
    spamtest_eq(ss+tt3,tt+tt3)
    
    spamtest_eq(ss+3,tt+3)
    spamtest_eq(ss+ss[,3,drop=T],tt+tt[,3])
    spamtest_eq(ss+tt[,3],tt+tt[,3])
    spamtest_eq(ss+tt[,5],tt+tt[,5])
    
    spamtest_eq(3-ss,3-tt)
    
    spamtest_eq(ss-tt[,3],tt-tt[,3])
    spamtest_eq(ss-tt[,5],tt-tt[,5])
    spamtest_eq(tt[,3]-ss,tt[,3]-tt)
    spamtest_eq(tt[,5]-ss,tt[,5]-tt)
    
    spamtest_eq(-ss,-tt)
    spamtest_eq(+ss,+tt)
})


test_that("multiplication/Division", {
    spamtest_eq(ss*ss,tt*tt)
    spamtest_eq(ss*tt,tt*tt)
    spamtest_eq(tt*ss,tt*tt)
        
    spamtest_eq(ss*ss3,tt*tt3)
    spamtest_eq(ss*ss3,tt*tt3)
    spamtest_eq(ss*tt3,tt*tt3)
    spamtest_eq(tt*ss3,tt*tt3)
    
    spamtest_eq(ss*ss,tt*tt)
    spamtest_eq(ss*tt,tt*tt)
    spamtest_eq(tt*ss,tt*tt)

    spamtest_eq(ss*ss3,tt*tt3)
    spamtest_eq(ss*ss3,tt*tt3)
    spamtest_eq(ss*tt3,tt*tt3)
    spamtest_eq(tt*ss3,tt*tt3)
})

test_that("&,|", { 
    spamtest_eq(ss&ss3,tt&tt3)
    spamtest_eq(ss|ss3,tt|tt3)
    
    ## recall, spam has to be first element
    spamtest_eq(ss&tt3,tt&tt3)
    spamtest_eq(ss|tt3,tt|tt3)
    
    spamtest_eq(ss3&1,tt3&1)
    spamtest_eq(ss3|1,tt3|1)
    spamtest_eq(ss3|0,tt3|0)
})
          
test_that("assign", {
    rw <- 1:3
    cl <- c(1,3)
    
    ss[1,] <- 1;tt[1,] <- 1
    spamtest_eq(ss,tt)
    ss[1,2] <- 1;tt[1,2] <- 1
    spamtest_eq(ss,tt)
    ss[1,] <- 1:m;tt[1,] <- 1:m
    spamtest_eq(ss,tt)
    ss[3:1,] <- 1:m;tt[3:1,] <- 1:m
    spamtest_eq(ss,tt)
    
    set.seed(14)
    rw <- sample(c(T,F),nrow(tt),rep=T)
    cl <- sample(c(T,F),ncol(tt),rep=T)
    ass <- rnorm(sum(rw)*sum(cl))
    ss[rw,cl] <- ass
    tt[rw,cl] <- ass
    
    spamtest_eq(ss[rw,cl],tt[rw,cl])
    
    ss[cbind(1,1)] <- 4;tt[cbind(1,1)] <- 4
    spamtest_eq(ss,tt)
    ss[rbind(dim(ss))] <- 4;tt[rbind(dim(tt))] <- 4
    spamtest_eq(ss,tt)
    
    ss[rbind(c(1,1),dim(ss))] <- c(0.1,0.1);tt[rbind(c(1,1),dim(tt))] <- c(0.1,0.1)
    spamtest_eq(ss,tt)
    
    
    rw <- c(1,3);cl <- 1:3;
    ss[rw,cl] <- 1:3;tt[rw,cl] <- 1:3
    spamtest_eq(ss,tt)
    rw <- c(3,1);cl <- 1:3; 
    ss[rw,cl] <- 1:3;tt[rw,cl] <- 1:3
    spamtest_eq(ss,tt)
    
    nn <- length(tt[-rw,-cl])
    spamtest_eq(ss[-rw,-cl] <- 1:nn,tt[-rw,-cl] <- 1:nn)
    nn <- length(tt[-rw,cl])
    spamtest_eq(ss[-rw,cl] <- 1:nn,tt[-rw,cl] <- 1:nn)
    nn <- length(tt[rw,-cl])
    spamtest_eq(ss[rw,-cl] <- 1:nn,tt[rw,-cl] <- 1:nn)
})
    

test_that("mathstuff", {
    spamtest_eq(sqrt(ss),sqrt(tt))
    
    spamtest_eq(lower.tri.spam(ss),lower.tri(tt)&tt!=0)
    spamtest_eq(lower.tri.spam(ss,F),lower.tri(tt,F)&tt!=0)
    spamtest_eq(upper.tri.spam(ss),upper.tri(tt)&tt!=0)
    spamtest_eq(upper.tri.spam(ss,F),upper.tri(tt,F)&tt!=0)
    
    spamtest_eq(lower.tri(ss),lower.tri(tt)&tt!=0)
    spamtest_eq(lower.tri(ss,F),lower.tri(tt,F)&tt!=0)
    spamtest_eq(upper.tri(ss),upper.tri(tt)&tt!=0)
    spamtest_eq(upper.tri(ss,F),upper.tri(tt,F)&tt!=0)
    

    if (F) {# only works for full matrices
        spamtest_eq(ss/tt,tt/tt)
        spamtest_eq(ss/ss,tt/tt)
        
        kk <- tt/tt
        kk[is.na(kk)] <- 0
        spamtest_eq(ss/tt,kk)
        spamtest_eq(ss/ss,kk)
        
        spamtest_eq(ss^tt,tt^tt)
        spamtest_eq(ss^ss,tt^tt)
    }
    
    ## maybe not all of them make sense
    if (F) { # this need to be discussed
        spamtest_eq(ss2/tt2,tt2/tt2)
        spamtest_eq(ss2^tt2,tt2^tt2)
        spamtest_eq(ss2/ss2,tt2/tt2)
        spamtest_eq(ss2^ss2,tt2^tt2)
    }
})
                                  
test_that("rbind(), cbind()", {
    spamtest_eq(rbind(tt,t(tt2)), rbind.spam(ss,t(ss2)))
    spamtest_eq(rbind(tt,tt,t(tt2),1:ncol(tt)), rbind.spam(ss,ss,t(ss2),t(spam(1:ncol(tt)))))
    
    spamtest_eq(rbind(tt,t(tt2)), rbind(ss,t(ss2)))
    spamtest_eq(rbind(tt,tt,t(tt2),1:ncol(tt)), rbind(ss,ss,t(ss2),t(spam(1:ncol(tt)))))

    
    spamtest_eq(cbind(tt,t(tt2)), cbind.spam(ss,t(ss2)))
    spamtest_eq(cbind(tt,tt,t(tt2),1:nrow(tt)), cbind.spam(ss,ss,t(ss2),spam(1:nrow(tt))))
    
    spamtest_eq(cbind(tt,t(tt2)), cbind(ss,t(ss2)))
    spamtest_eq(cbind(tt,tt,t(tt2),1:nrow(tt)), cbind(ss,ss,t(ss2),spam(1:nrow(tt))))

    expect_error(rbind(ss,ss2))
    expect_error(cbind(ss,ss2))
    if (F) {
                                        # dummy testing
        rbind.spam()
        cbind.spam()
        rbind.spam(deparse.level=0)
        cbind.spam(deparse.level=0)
        
                                        # the following should produce warnings:
        rbind(a=ss)
        cbind(b=ss)
        rbind.spam(deparse.level=1)
        cbind.spam(deparse.level=1)
                                        # the following should produces errors:
     }
})

# testing diag:
test_that("diag() and similar", {
    spamtest_eq(diag(tt),diag(ss))
    spamtest_eq(diag.spam(ss),diag(ss))
    
    spamtest_eq(diag.spam(1:4),diag(1:4))
    spamtest_eq(diag.spam(1,2,3),diag(1,2,3))
    
    spamtest_eq(diag.spam(1:4,4,6),diag(1:4,4,6))
    spamtest_eq(diag.spam(1:4,12),diag(1:4,12))
    
    diag(tt) <- diag(ss) <- 1:n
    spamtest_eq(tt, ss)
    diag(tt) <- diag(ss) <- 2
    spamtest_eq(tt, ss)
    
    diag(tt) <- diag(ss) <- 0
    diag(tt) <- diag(ss) <- 1:n
    spamtest_eq(tt, ss)
    
    spamtest_eq(diag(tt[,5]),diag(ss[,5]))
    spamtest_eq(diag(tt[,5,drop=T]),diag(ss[,5,drop=T]))
    spamtest_eq(diag(tt[,5,drop=F]),diag(ss[,5,drop=F]),rel=F)
})


# testing as.spam
test_that("as.spam() and derivatives", {
    spamtest_eq(as.spam(0), as.matrix(0),rel=FALSE)
    b <- rnorm(n)
    spamtest_eq(as.spam(b), b )
    spamtest_eq(as.spam(-abs(b)), -abs(b) )
    spamtest_eq(-as.spam(abs(b)), -abs(b) )
    spamtest_eq(as.spam(tt), tt)
    
                                        # testing spam
    spamtest_eq(spam(0,1000,1000), matrix(0,1000,1000),rel=FALSE)
    spamtest_eq(spam(1,12,12),matrix(1,12,12))
})

# no NA, NaN, Inf handling
## if (F) { #not implemented
##   ss[2,] <- NA

##   tt[4,] <- NA
##   tt[,4] <- NaN
##   tt[5,] <- Inf
##   as.spam(tt)
## }



# transformation from list and else
test_that("as.spam.list() and triplet()", {
    spamtest_eq( as.spam.list(triplet(ss)), ss)
    spamtest_eq( as.spam.list(triplet(ss, tri=TRUE)), ss)
    
    spamtest_eq( as.spam.list(triplet(tt)), ss)
    spamtest_eq( as.spam.list(triplet(tt, tri=TRUE)), ss)

    spamtest_eq( as.spam(triplet(ss)), ss)
    spamtest_eq( as.spam(triplet(ss, tri=TRUE)), ss)
    
    spamtest_eq( as.spam(triplet(tt)), ss)
    spamtest_eq( as.spam(triplet(tt, tri=TRUE)), ss)

  expect_error(spam.list( list(i=1, 2)))
  expect_error(spam.list( list(ind=1, 2)))
  expect_error(spam.list( list(ind=1, 1, 2)))
  expect_error(spam.list( list(ind=1, j=0, 2)))
  #spam.list( list(ind=numeric(0), j=numeric(0), numeric(0))))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-kronecker.R.        #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat", lib.loc = LIB.LOC)
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = LIB.LOC)


context("test-kronecker.R")
    

options(spam.printsize=6)

xn <- 3
xm <- 2
yn <- 4
ym <- 2
set.seed(14)

X <- array(runif(xn*xm), c( xn,xm))
Y <- array(runif(yn*ym), c( yn,ym))

R <- as.spam(X)
S <- as.spam(Y)

b <- rnorm(5)

# with matrices

test_that("two matrices:", {
    spamtest_eq( kronecker( X, Y), kronecker.spam( X, Y) )
    spamtest_eq( kronecker( X, Y), kronecker.spam( R, S) )
    
    
    spamtest_eq( kronecker( X, Y), kronecker( R, S) )
    spamtest_eq( kronecker( X, Y), kronecker( R, Y) )
    spamtest_eq( kronecker( X, Y), kronecker( X, S) )
})

test_that("matrix vector", {
    spamtest_eq( kronecker( X, b), kronecker.spam( X, b) )
    spamtest_eq( kronecker( b, Y), kronecker.spam( b, S) )
    
    spamtest_eq( kronecker( X, b), kronecker( R, b) )
    spamtest_eq( kronecker( b, Y), kronecker( b, S) )
})

    
test_that("degenerate cases", {
    spamtest_eq( kronecker( X, 0), kronecker.spam( X, 0),rel=FALSE )
    spamtest_eq( kronecker( 0, 0), kronecker.spam( 0, 0),rel=FALSE  )
    spamtest_eq( kronecker( 0, Y), kronecker( spam(0), Y),rel=FALSE  )
}) 
    
test_that("different operators", {
    spamtest_eq( kronecker(X,Y,FUN="+"),kronecker(R,S,FUN="+"))
    spamtest_eq( kronecker(X,b,FUN="+"),kronecker(R,b,FUN="+"))
    spamtest_eq( kronecker(c(0,1,0),Y,FUN="+"),kronecker(c(0,1,0),S,FUN="+"))
    expect_warning(tmp <- kronecker(diag.spam(2),S,FUN="+"),
                   "Sparseness structure of ")
    spamtest_diff(tmp, kronecker(diag(2),Y,FUN="+"))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-math.R.             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## source("tests/testthat/helper.R")
## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = LIB.LOC)


context("test-math.R")



# see Matrix::rsparsematrix
spam_random <- function(n, m=n, nnz=min(m-1,4)*n, fill=rnorm, seed=NULL, ...)
    {
        if (!is.null(seed)) set.seed(seed)
        ind <- sample.int((n*m), size=nnz)
#
#        as.spam( list(i=(ind %% m)+1, j=(ind %/% n)+1, fill(length(ind), ...)))
        tmp <- matrix(0,n,m)
        tmp[ind] <- fill(length( ind), ...)
        as.spam(tmp)
    }

# construct matrices:
n <- 10
m <- 5

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
tt[tt<0] <- 0

ss <- as.spam(tt)
options(spam.structurebased=FALSE) # test for equivalence!


#     ‘Math’ ‘"abs"’, ‘"sign"’, ‘"sqrt"’, ‘"ceiling"’, ‘"floor"’,
#          ‘"trunc"’, ‘"cummax"’, ‘"cummin"’, ‘"cumprod"’, ‘"cumsum"’,
#          ‘"log"’, ‘"log10"’, ‘"log2"’, ‘"log1p"’, ‘"acos"’, ‘"acosh"’,
#          ‘"asin"’, ‘"asinh"’, ‘"atan"’, ‘"atanh"’, ‘"exp"’, ‘"expm1"’,
#          ‘"cos"’, ‘"cosh"’, ‘"cospi"’, ‘"sin"’, ‘"sinh"’, ‘"sinpi"’,
#          ‘"tan"’, ‘"tanh"’, ‘"tanpi"’, ‘"gamma"’, ‘"lgamma"’,
#          ‘"digamma"’, ‘"trigamma"’

#     ‘Math2’ ‘"round"’, ‘"signif"’

#     ‘Summary’ ‘"max"’, ‘"min"’, ‘"range"’, ‘"prod"’, ‘"sum"’, ‘"any"’, ‘"all"’

#
# !

test_that("math", {
    A <- diag.spam(4)   ; B <- diag(4)
    spamtest_eq(A, B)
    spamtest_eq(!A, !B)
    diag(A)=0  ; diag(B) <- 0
    spamtest_eq(!A, !B)
                                        # str(A) # is what needs to be expected...,
                                        # different to spam:::complement.spam(A)
})











test_that("math Summary", {
    expect_equal(max(ss), max(tt))
    expect_equal(min(ss), min(tt))
    expect_equal(range(ss), range(tt))
    expect_equal(prod(ss), prod(tt))
    expect_equal(sum(ss), sum(tt))
    expect_warning(anytt <- any(tt),
                   "coercing argument")
    expect_equal(any(ss), anytt)
    expect_warning(alltt <- all(tt),
                   "coercing argument")
    expect_equal(all(ss), alltt)
})

test_that("Math2", {
    spamtest_eq(round(ss), round(tt))
    spamtest_eq(signif(ss), signif(tt))

#     ‘Math’ ‘"abs"’, ‘"sign"’, ‘"sqrt"’, ‘"ceiling"’, ‘"floor"’,
#          ‘"trunc"’, ‘"log1p"’
#          ‘"asin"’, ‘"asinh"’, ‘"atan"’, ‘"atanh"’, ‘"expm1"’,
#           ‘"sin"’, ‘"sinh"’, ‘"sinpi"’,
#          ‘"tan"’, ‘"tanh"’, ‘"tanpi"’,

#          ‘"cummax"’, ‘"cummin"’, ‘"cumprod"’, ‘"cumsum"’,
#          ‘"log"’, ‘"log10"’, ‘"log2"’, ‘"acos"’, ‘"acosh"’,
#          , ‘"exp"’, ‘"cos"’, ‘"cosh"’, ‘"cospi"’
#           ‘"gamma"’, ‘"lgamma"’,   ‘"digamma"’, ‘"trigamma"’


    spamtest_eq(abs(ss), abs(tt))
    spamtest_eq(cos(ss), cos(tt))
    spamtest_eq(cosh(ss), cosh(tt))


    options(spam.NAOK=TRUE) # test for equivalence!

    expect_equal(suppressWarnings(c(gamma(tt))), suppressWarnings(c(gamma(ss))), rel = FALSE)  #
    expect_equal(c(suppressWarnings(digamma(ss))), suppressWarnings(c(digamma(tt)))) #
    expect_equal(c(trigamma(ss)), c(trigamma(tt)))
    spamtest_eq(exp(ss), exp(tt))
    spamtest_eq(expm1(ss), expm1(tt))


    expect_equal(c(log(ss)), c(log(tt)))
    spamtest_eq(cummax(ss), cummax(tt))

    ## TODO
    ## for (f in getGroupMembers("Math"))
    ##     spamtest_eq(do.call(f, list(ss)),
    ##                  do.call(f, list(tt)), relative = FALSE)




    for ( i in c(TRUE, FALSE)) {
        options(spam.structurebased=i)

        set.seed(122)
        S1 <- spam_random(5, nnz=10)
        S2 <- spam_random(5, nnz=10)
        F1 <- as.matrix(S1)
        F2 <- as.matrix(S2)

        spamtest_eq(S1+S2, F1+F2)
        spamtest_eq(S1+F2, F1+F2)
        spamtest_eq(F1+S2, F1+F2)

        spamtest_eq(S1-S2, F1-F2)
        spamtest_eq(S1-F2, F1-F2)
        spamtest_eq(F1-S2, F1-F2)

        S1 <- spam_random(5, nnz=0)
        S2 <- spam_random(5, nnz=10)
        F1 <- as.matrix(S1)
        F2 <- as.matrix(S2)

        spamtest_eq(S1+S2, F1+F2)
        spamtest_eq(S1+F2, F1+F2)
        spamtest_eq(F1+S2, F1+F2)

        spamtest_eq(S1-S2, F1-F2)
        spamtest_eq(S1-F2, F1-F2)
        spamtest_eq(F1-S2, F1-F2)
    }
})

test_that("Bug 2.9.2", {

e1 <- matrix( 1:12, 3)
e2 <- -spam( 1:12,3)
e3 <- spam_random(3,4)

     spamtest_eq( e1+ e2, e1+ as.matrix(e2), relative=FALSE) # covered previously
     spamtest_eq( e1- e2, e1- as.matrix(e2), relative=FALSE) # covered previously
     spamtest_eq( e1+ e3, e1+ as.matrix(e3), relative=FALSE) # covered previously
     spamtest_eq( e1- e3, e1- as.matrix(e3), relative=FALSE) # covered previously


     spamtest_eq( e1* e2, e1* as.matrix(e2), relative=FALSE)
     spamtest_eq( e1/ e2, e1/ as.matrix(e2), relative=FALSE)
     spamtest_eq( e1^ e2, e1^ as.matrix(e2), relative=FALSE)
     spamtest_eq( e1& e2, e1& as.matrix(e2), relative=FALSE)
     spamtest_eq( e1| e2, e1| as.matrix(e2), relative=FALSE)

     spamtest_eq( e1* e3, e1* as.matrix(e3), relative=FALSE)
#     spamtest_eq( e1/ e3, e1/ as.matrix(e3), relative=FALSE)  # will fly in 3.
      spamtest_eq( e3/e1, as.matrix(e3)/e1, relative=FALSE)
    spamtest_eq( e1^ e3, e1^ as.matrix(e3), relative=FALSE)
     spamtest_eq( e1& e3, e1& as.matrix(e3), relative=FALSE)
     spamtest_eq( e1| e3, e1| as.matrix(e3), relative=FALSE)



})


options(spam.structurebased=TRUE) # test for equivalence!

test_that("Bug 2.10-0", {

  set.seed(3)
  e1 <- matrix( 1:12, 3)
  e2 <- -spam( 1:12,3)
  e3 <- spam_random(3,4)

  spamtest_eq( e1+ e2, e1+ as.matrix(e2), relative=FALSE)
  spamtest_eq( e1- e2, e1- as.matrix(e2), relative=FALSE)
  spamtest_eq( e1+ e3, e1+ as.matrix(e3), relative=FALSE)
  spamtest_eq( e1- e3, e1- as.matrix(e3), relative=FALSE)


  spamtest_eq( e1* e2, e1* as.matrix(e2), relative=FALSE)
  spamtest_eq( e1/ e2, e1/ as.matrix(e2), relative=FALSE)
  spamtest_eq( e2/ e1, e2/ as.matrix(e1), relative=FALSE)
  spamtest_eq( e1^ e2, e1^ as.matrix(e2), relative=FALSE)

  spamtest_eq( e1& e2, e1& as.matrix(e2), relative=FALSE)
  spamtest_eq( e1| e2, e1| as.matrix(e2), relative=FALSE)
  spamtest_eq( e2& e1, e2& as.matrix(e1), relative=FALSE)
  spamtest_eq( e2| e1, e2| as.matrix(e1), relative=FALSE)

  spamtest_eq( e1* e3, e1* as.matrix(e3), relative=FALSE)
  spamtest_eq( e1/ e3, e1/ as.matrix(e3), relative=FALSE)
  spamtest_eq( e3/ e1, as.matrix(e3)/ e1, relative=FALSE)
  spamtest_eq( e1^ e3, e1^ as.matrix(e3), relative=FALSE)
  spamtest_eq( e1& e3, e1& as.matrix(e3), relative=FALSE)
  spamtest_eq( e1| e3, e1| as.matrix(e3), relative=FALSE)

  spamtest_eq( e3* e1, e3* as.matrix(e1), relative=FALSE)
  spamtest_eq( e3/ e1, e3/ as.matrix(e1), relative=FALSE)
  spamtest_eq( e1/ e3, as.matrix(e1)/ e3, relative=FALSE)
  spamtest_eq( e3^ e1, e3^ as.matrix(e1), relative=FALSE)
  spamtest_eq( e3& e1, e3& as.matrix(e1), relative=FALSE)
  spamtest_eq( e3| e1, e3| as.matrix(e1), relative=FALSE)



})

# HEADER ####################################################
# This is file spam/tests/testthat/test-ops.R.              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-ops.R")


# construct matrices:
n <- 10
m <- 5

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
rr <- matrix(rnorm(m*n),n,m)

tt[tt<0] <- 0
rr[rr>0] <- 0

ss <- as.spam(tt)
qq <- as.spam(rr)
options(spam.structurebased=FALSE) # test for equivalence!

options(spam.NAOK=TRUE) # test for equivalence!


test_that("ops", {
    for (f in rev(getGroupMembers("Arith")))
        spamtest_eq( do.call(f, list(ss,qq)), do.call(f, list(tt,rr)))
    
    for (f in getGroupMembers("Compare"))
        spamtest_eq( do.call(f, list(ss,qq)), do.call(f, list(tt,rr)), relative = FALSE)
    
    for (f in getGroupMembers("Logic"))
        spamtest_eq( do.call(f, list(ss,qq)), do.call(f, list(tt,rr)), relative = FALSE)
    
    
    tv <- sv <- ss@entries
    qv <- qq@entries
    options(spam.structurebased=TRUE)
    
    
    for (g in getGroupMembers("Ops")) {
        for (f in getGroupMembers(g)) {
            expect_equal(do.call(f, list(ss,sv))@entries,
                         as.numeric(do.call(f, list(tv,sv))))
            expect_equal(do.call(f, list(sv,ss))@entries,
                         as.numeric(do.call(f, list(sv,tv))))
            expect_equal(do.call(f, list(ss,4))@entries,
                         as.numeric(do.call(f, list(tv,4))))
        }
    }

    expect_error(do.call(f, list(ss,1:2)))
})
#####################################################################################

test_that("??", {
    options(spam.inefficiencywarning=TRUE)
    options(spam.structurebased=FALSE)
    
    expect_equal(diag(2)+diag.spam(2), diag(rep(2,2)))
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-spamlist.R.         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-spamlist.R")


########

test_that("spam.list", {
    spamtest_eq(spam( list(ind=numeric(0), j=numeric(0), numeric(0)),nrow=4,ncol=3),
                spam(0,4,3),rel=FALSE)

    i <- c(1,2,3,4,5)
    j <- c(5,4,3,2,1)
    ss3 <- spam(0,5,5)
    ss3[cbind(i,j)] <- i/j
    spamtest_eq(spam.list(list(i=i,j=j,i/j)), ss3)
    pad(ss3) <- c(13,13)
    spamtest_eq(spam.list(list(i=i,j=j,i/j),13,13), ss3)
    pad(ss3) <- c(3,3)
    spamtest_eq(spam.list(list(i=i,j=j,i/j),3,3), ss3)
    pad(ss3) <- c(2,2)
    spamtest_eq(spam.list(list(i=i,j=j,i/j),2,2), ss3,rel=F)
    
    
    spamtest_eq({options(spam.listmethod='EP');
                 spam.list(list(i=i,j=j,i/j),ncol=3)},
                {options(spam.listmethod='BS');
                 method='BS';spam.list(list(i=i,j=j,i/j),ncol=3)})
    spamtest_eq({options(spam.listmethod='EP');
                 spam.list(list(i=i,j=j,i/j),ncol=3,nrow=4)},
                {options(spam.listmethod='BS');
                 spam.list(list(i=i,j=j,i/j),ncol=3,nrow=4)})
    
    spamtest_eq(spam.list(list(i=i,j=j,i/j),ncol=1,nrow=1),
                0,rel=F)
  
    
    
    set.seed(2011)
    m = 1000
    rmax = 30
    cmax = 40
    i = floor(runif(m) * rmax) + 1
    j = floor(runif(m) * cmax) + 1
    val = floor(10 * runif(m)) + 1
    
    
    options(spam.listmethod='EP')
    ss1 <- spam.list(list(i=i,j=j,val))
    options(spam.listmethod='BS')
    ss2 <- spam.list(list(i=i,j=j,val))
    
    spamtest_eq(ss1,ss2,rel=F)
})

test_that("spam with list", {
    spamtest_eq(spam( list(ind=numeric(0), j=numeric(0), numeric(0)),nrow=4,ncol=3),
                spam(0,4,3),rel=FALSE)

    i <- c(1,2,3,4,5)
    j <- c(5,4,3,2,1)
    ss3 <- spam(0,5,5)
    ss3[cbind(i,j)] <- i/j
    spamtest_eq(spam(list(i=i,j=j,i/j)), ss3)
    pad(ss3) <- c(13,13)
    spamtest_eq(spam(list(i=i,j=j,i/j),13,13), ss3)
    pad(ss3) <- c(3,3)
    spamtest_eq(spam(list(i=i,j=j,i/j),3,3), ss3)
    pad(ss3) <- c(2,2)
    spamtest_eq(spam(list(i=i,j=j,i/j),2,2), ss3,rel=F)
    
    
    spamtest_eq({options(spam.listmethod='EP');
                 spam(list(i=i,j=j,i/j),ncol=3)},
                {options(spam.listmethod='BS');
                 method='BS';spam(list(i=i,j=j,i/j),ncol=3)})
    spamtest_eq({options(spam.listmethod='EP');
                 spam(list(i=i,j=j,i/j),ncol=3,nrow=4)},
                {options(spam.listmethod='BS');
                 spam(list(i=i,j=j,i/j),ncol=3,nrow=4)})
    
    spamtest_eq(spam(list(i=i,j=j,i/j),ncol=1,nrow=1),
                0,rel=F)
  
    
    
    set.seed(2011)
    m = 1000
    rmax = 30
    cmax = 40
    i = floor(runif(m) * rmax) + 1
    j = floor(runif(m) * cmax) + 1
    val = floor(10 * runif(m)) + 1
    
    
    options(spam.listmethod='EP')
    ss1 <- spam(list(i=i,j=j,val))
    options(spam.listmethod='BS')
    ss2 <- spam(list(i=i,j=j,val))
    
    spamtest_eq(ss1,ss2,rel=F)
})

# HEADER ####################################################
# This is file spam/tests/testthat/test-subset.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")


## library("testthat")                             
## library("spam", lib.loc = LIB.LOC)

context("test-subset.R")




# subsetting:
########################################################################


# construct matrices (should be at least 3x5, with n<m):
n <- 10
m <- 15

set.seed(14)
tt <- matrix(rnorm(m*n),n,m)
tt[tt<0] <- 0

ss <- as.spam(tt)

test_that("subset", {
    spamtest_eq(ss[],tt[])      # ok
    spamtest_eq(ss[,],tt[,])    # ok
    spamtest_eq(ss[1,],tt[1,])  # ok
    spamtest_eq(ss[,2],tt[,2])  # ok
    spamtest_eq(ss[1,3],tt[1,3], relative = FALSE)# ok
    spamtest_eq(ss[3:1,],tt[3:1,])# ok
    
    
    rw <- sample(c(T,F),nrow(tt),rep=T)
    cl <- sample(c(T,F),ncol(tt),rep=T)
    spamtest_eq(ss[rw,cl],tt[rw,cl])
    spamtest_eq(ss[rw],tt[rw])
    
    
    
    
    rw <- c(1,3);cl <- 1:3;
    spamtest_eq(ss[rw,cl],tt[rw,cl])
    spamtest_eq(ss[-rw,cl],tt[-rw,cl])
    spamtest_eq(ss[-rw,-cl],tt[-rw,-cl])
    rw <- c(3,1);cl <- 1:3; spamtest_eq(ss[rw,cl],tt[rw,cl])
    rw <- c(3,1,2,1);cl <- 1:3; spamtest_eq(ss[rw,cl],tt[rw,cl])
    
    tmp <- cbind(sample(1:3,24,rep=T),sample(1:5,24,rep=T))
    spamtest_eq(ss[tmp],tt[tmp])
    
    
    spamtest_eq(diag(10)[1:2,9:10],diag.spam(10)[1:2,9:10],rel=F)
    
    rs <- sample(c(0,1:(2*n)),2*m,replace=T)
    spamtest_eq(ss[rs],tt[rs])
                                        # NAs simply work!
    rs <- sample(c(0,1:(2*n),NA),2*m,replace=T)
    spamtest_eq(ss[rs],tt[rs])
    
    rs <- sample(c(T,F,NA),2*m,replace=T)
    expect_equal(ss[rs],tt[rs])
    
                                        # stuff from 0.31:
    
    tt <- array(1:36,c(6,6))
    ss <- as.spam( tt)
    
    for (i in 1:4) {
        rs <- cbind(rep(1:i,each=i),rep(1:i,i))
        spamtest_eq(ss[rs],tt[rs])
        spamtest_eq(ss[rs+1],tt[rs+1])
        spamtest_eq(ss[rs+2],tt[rs+2])
        
        ti <- array(1:(i^2),c(i,i))
        si <- as.spam( ti)
        spamtest_eq(si[rs],ti[rs])  
        
        si <- spam(0,i,i)
        ti <- as.matrix(si)
        spamtest_eq(si[rs],ti[rs], relative = FALSE)
        
        si <- diag.spam(i)
        ti <- diag(i)
        spamtest_eq(si[rs],ti[rs])
    }
})

## if (F) {
## # large timing example

##   n <- 100
##   m <- 15000

##   set.seed(14)
##   tt <- matrix(rnorm(m*n,mean=-1),n,m)
##   tt[tt<0] <- 0

##   ss <- as.spam(tt)

##   set.seed(14)
##   system.time(for (i in 1:100) {
##     spam:::subset.rows.spam(ss,sample(1:n,10)) })
##   set.seed(14)
##   system.time(for (i in 1:100) {
##     spam:::subset.spam(ss,sample(1:n,10),1:m) })

## }

# HEADER ####################################################
# This is file spam/tests/jss_areal_counts.R.               #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# JSS article: 
#     "Pitfalls in the implementation of Bayesian
#      hierarchical modeling of areal count data.
#      An illustration using BYM and Leroux models."
#
# test the MCMC sampler from the paper with 30 iterations.


# SETUP:
library("spam")
options(spam.structurebased=TRUE)

# BYM ---------------------------------------------
data(Oral); attach(Oral) 
path <- system.file("demodata/germany.adjacency", package = "spam")
A <- adjacency.landkreis(path); n <- dim(A)[1]

set.seed(2)
hyperA <- c(1, 1); hyperB <- c(0.5, .01)  
totalg <- 30

upost <- vpost <- array(0, c(totalg, n))  
kpost <- array(NA, c(totalg, 2)); accept <- rep(NA, totalg)
upost[1,] <- vpost[1,] <- rep(.001, 544); kpost[1,] <- c(10, 100) 

eta <- upost[1,] + vpost[1,]
C <- exp(eta) * E; diagC <- diag.spam(c(rep(0, n), C))
b <- c( rep(0, n), Y + (eta - 1) * C)
Qu <- R <- precmat.IGMRFirreglat(A); pad(Qu) <- c(2 * n, 2 * n)
Qv <- as.spam(rbind(cbind( diag(n), -diag(n)),
                      cbind(-diag(n),  diag(n))))
Q <- kpost[1,1] * Qu + kpost[1,2] * Qv + diagC
struct <- chol(Q, memory = list(nnzcolindices = 6467))
uRuHalf <- t(upost[1,]) %*% (R %*% upost[1,]) / 2
vvHalf <- t(vpost[1,]) %*% vpost[1,] / 2
postshape <- hyperA + c(n - 1, n) / 2

for (i in 2:totalg) {
     kpost[i,] <- rgamma(2, postshape, hyperB + c(uRuHalf, vvHalf))
 
     etaTilde <- eta
     for(index in 1:2){
         C <- E * exp(etaTilde)
         diagC <- diag.spam(c(rep(0, n), C))
         b <- c(rep(0, 544), Y + (etaTilde - 1) * C)
         Q <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC
         etaTilde <- c(solve.spam(Q, b,
                                  Rstruct = struct))[1:n + n]
     }
 
     C <- exp(etaTilde) * E; diagC <- diag.spam(c(rep(0, n), C))
     b <- c( rep(0, n), Y + (etaTilde - 1) * C)
     Q <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC

     x_ <- c(rmvnorm.canonical(1, b, Q, Rstruct = struct))
     upost[i,] <- x_[1:n]; eta_ <- x_[1:n + n]; vpost[i,] <- eta_ - upost[i,]
     uRuHalf_ <- t(upost[i,]) %*% (R %*% upost[i,]) / 2
     vvHalf_ <- t(vpost[i,]) %*% vpost[i,] / 2

     etaTilde_ <- eta_
     for(index in 1:2){
       C_ <- E * exp(etaTilde_)
       diagC_ <- diag.spam(c(rep(0, n), C_))
       b_ <- c(rep(0, 544), Y + (etaTilde_ - 1) * C_)
       Q_<- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC_
       etaTilde_ <- c(solve.spam(Q_, b_,
                                Rstruct = struct))[1:n + n]
     }
     
     C_ <- exp(etaTilde_) * E; diagC_ <- diag.spam(c(rep(0, n), C_))
     b_ <- c( rep(0, n), Y + (etaTilde_ - 1) * C_)
     Q_ <- kpost[i,1] * Qu + kpost[i,2] * Qv + diagC_

     logPost_ <- sum(Y * eta_ - E * exp(eta_)) -
         kpost[i,1] * uRuHalf_ - kpost[i, 2] * vvHalf_
     logPost  <- sum(Y * eta - E * exp(eta)) - kpost[i,1] * uRuHalf -
         kpost[i,2] * vvHalf
     logApproxX_ <- - kpost[i,1] * uRuHalf_ - kpost[i,2] * vvHalf_ -
         sum(.5 * eta_^2 * C) + sum(b * eta_)
     logApproxX  <- - kpost[i,1] * uRuHalf  - kpost[i,2] * vvHalf -
         sum(.5 * eta^2 * C_) + sum(b_ * eta)
     logAlpha <- min(0, logPost_ - logPost + logApproxX - logApproxX_)
 
     if (log(runif(1)) < logAlpha) {    
         uRuHalf <- uRuHalf_;  vvHalf <- vvHalf_
         eta <- eta_; b <- b_; C <- C_; accept[i] <- 1
     } else{                            
       accept[i] <- 0; upost[i,] <- upost[i-1,]; vpost[i,] <- vpost[i-1,]} 
 }
 
# values of  30th iteration
head(eta)
tail(b)
head(C)
tail(accept)
tail(upost[30,])
tail(vpost[30,])
sum(accept[-1])
sum(upost)

# HEADER ####################################################
# This is file spam/tests/demo_jss10-figures-table.R.       #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# This demo contains the R code to construct the figures and the table of the
# article:
#     "spam: A Sparse Matrix R Package with Emphasis on
#            MCMC Methods for Gaussian Markov Random Fields"
# submitted to JSS.


# The code presented here differs in the following points form the actually used
# one:
# - Very large grid sizes or very high order neighbor structures are not included
#   here;
# - Instead of (100+1) factorizations only (10+1) are performed here;
# - No figure fine-tuning is done here.
# - We had a few additional gc(), just to be sure.



# The following are tests specific. Not all computers run with profiling. Instead
# of commenting, we define dummies.
options( echo=FALSE)
library( spam, warn.conflict=FALSE)


Rprof <- function(memory.profiling=TRUE, interval=0.1)
  return()
summaryRprof <- function(memory="both")
  return(list(by.total=rbind(1:4)))









# Figure 1:
i <- c( 2,4,4,5,5)
j <- c( 1,1,2,1,3)

A <- spam(0,5,5)
A[cbind(i,j)] <- rep(.5, length(i))
A <- t( A)+A+diag.spam(5)

U <- chol( A)
pivot <- U@pivot
B <- A[pivot,pivot]
R <- chol( B)

U@pivot
U@snmember
U@supernodes

U@entries
U@colindices
U@colpointers
U@rowpointers

if (F){
  display( A)
  display( as.spam( chol(as.matrix( A))))
  display( B)
  display( as.spam(R))
  abline( h=-U@supernodes+.5,col=3,lty=2)
}

# Figure 2:
theta1 <- .1
theta2 <- .01
n <- dim( UScounties.storder)[1]

USmat <- diag.spam(n) + theta1 *  UScounties.storder + theta2 *  UScounties.ndorder


U <- chol( USmat,memory=list(nnzR=146735))
if (F) {
  display( as.spam(U))
  text(400,-2200,"MMD\nz=146735\nw=30182\ns=1262",adj=0)
}

U <- chol( USmat, pivot="RCM",memory=list(nnzR=256198,nnzcolindices=140960))
if (F) {
  display( as.spam(U))
  text(400,-2200,"RCM\nz=256198\nw=140960\ns=1706",adj=0)
}

U <- chol( USmat, pivot=FALSE,memory=list(nnzR=689615,nnzcolindices=96463))
if (F) {
  display( as.spam(U))
  text(400,-2200,"no permutation\nz=689615\nw=96463\ns=711",adj=0)
}

# general parameters for the following
N <- 10         # would be 100 in the article
stsel <- 1      # user.self
rPsx <- 1       # for function "system.time"
rPsy <- 3       # memory usage
rPint <- .0001  # small interval


# Figure 3:
theta1 <- .1
theta2 <- .05

xseq <- ceiling(4 + exp(seq(0,to=4,by=1))/2)  # would be seq(0.5,to=6,by=.5) in the article
xseql <- length(xseq)

table <- array(NA,c(xseql,4))
for (ix in 1:xseql) {

  egdx <- expand.grid(1:xseq[ix],1:xseq[ix])
  Cspam <- nearest.dist( egdx, delta=1., upper=NULL)
  Dspam <- nearest.dist( egdx, delta=1.5,upper=NULL)

  mat <- diag.spam(xseq[ix]^2) + theta1 * Cspam + theta2 * Dspam

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[ix,1] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch1 <- chol(mat)}
                             )[stsel]
  Rprof( NULL)
  table[ix,2] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]


  Rprof( memory.profiling=TRUE, interval = rPint)
  table[ix,3] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch2 <- update(ch1,mat) }
                             )[stsel]
  Rprof( NULL)
  table[ix,4] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

}

if (F) {
  # Since we have a small N, elements in table might be zero.
  table <- pmax(table, 0.0001)

  par(mfcol=c(1,2))
  plot(xseq, table[,1], type='l', log='xy', ylim=range(table[,c(1,3)]),
       xlab="L (log scale)", ylab="seconds (log scale)")
  lines(xseq, table[,3], lty=2)

  plot(xseq, table[,2], type='l', log='xy', ylim=range(table[,c(2,4)]+0.01),
       xlab="L (log scale)", ylab="Mbytes (log scale)")
  lines(xseq, table[,4], lty=2)
}


# Figure 4:

x <- 20     # was 50 in article
maxnn <- 3  # was 6 in article

egdx <- expand.grid( 1:(maxnn+1), 1:(maxnn+1))
dval <- sort(unique(nearest.dist( egdx, delta=maxnn)@entries))
dvall <- length( dval)


egdx <- expand.grid( 1:x, 1:x)


table <- array(NA, c(dvall,5))

for (id in 1:dvall) {

  mat <- nearest.dist( egdx, delta=dval[id],upper=NULL)
  mat@entries <- exp(-2*mat@entries)         # arbitrary values to get a spd precision matrix

  table[id,5] <- length(Cspam)

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[id,1] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch1 <- chol(mat)}
                             )[stsel]
  Rprof( NULL)
  table[id,2] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

  Rprof( memory.profiling=TRUE, interval = rPint)
  table[id,3] <- system.time( { ch1 <- chol(mat);
                                for (i in 1:N) ch2 <- update(ch1,mat) }
                             )[stsel]
  Rprof( NULL)
  table[id,4] <- summaryRprof( memory="both")$by.total[rPsx,rPsy]

}


if (F) {
  # Since we have a small N, elements in table might be zero.
  table <- pmax(table, 0.0001)

  par(mfcol=c(1,2))
  plot( dval, table[,1], type='l', log='xy',ylim=range(table[,c(1,3)]),
       xlab="distance (log scale)", ylab="seconds (log scale)")
  lines( dval, table[,3],lty=2)

  plot( dval, table[,2], type='l', log='xy',ylim=range(table[,c(2,4)]),
       xlab="distance (log scale)", ylab="Mbytes (log scale)")
  lines( dval, table[,4],lty=2)
}



# Table 1:
table <- array(NA,c(9,4))

x <- 10    #  was 50 in article
egdx <- expand.grid(1:x,1:x)

# As above hence shortend
gridmat <- diag.spam(x^2) + .2 * nearest.dist( egdx, delta=1.,upper=NULL) +
  .1 * nearest.dist( egdx, delta=1.5,upper=NULL)
# USmat was constructed above.


# Generic call first:
Rprof( memory.profiling=TRUE, interval = rPint)
table[1,1] <- system.time(   for (i in 1:N) ch1 <- chol(gridmat) )[stsel]
Rprof( NULL)
table[1,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[1,3] <- system.time(   for (i in 1:N) ch2 <- chol(USmat)   )[stsel]
Rprof( NULL)
table[1,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Call a chol.spam directly
Rprof( memory.profiling=TRUE, interval = rPint)
table[2,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat))[stsel]
Rprof( NULL)
table[2,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[2,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat)  )[stsel]
Rprof( NULL)
table[2,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Less checking:
options(spam.safemode=c(FALSE, FALSE, FALSE))
Rprof( memory.profiling=TRUE, interval = rPint)
table[3,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat)    )[stsel]
Rprof( NULL)
table[3,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[3,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat)    )[stsel]
Rprof( NULL)
table[3,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
options(spam.safemode=c(TRUE, TRUE, TRUE))


# lesser checking
options(spam.cholsymmetrycheck=FALSE)
Rprof( memory.profiling=TRUE, interval = rPint)
table[4,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat)    )[stsel]
Rprof( NULL)
table[4,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[4,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat)    )[stsel]
Rprof( NULL)
table[4,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
options(spam.cholsymmetrycheck=TRUE)

# Pass optimal memory parameters (from above
memory1 = summary(ch1)[1:2]
memory2 = summary(ch2)[1:2]
Rprof( memory.profiling=TRUE, interval = rPint)
table[5,1] <- system.time(   for (i in 1:N) ch1 <- chol( gridmat,memory=memory1)    )[stsel]
Rprof( NULL)
table[5,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[5,3] <- system.time(   for (i in 1:N) ch2 <- chol( USmat,memory=memory2)    )[stsel]
Rprof( NULL)
table[5,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# All of the above
options(spam.cholsymmetrycheck=FALSE, safemode=c(FALSE,FALSE,FALSE))
Rprof( memory.profiling=TRUE, interval = rPint)
table[6,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,memory=memory1)    )[stsel]
Rprof( NULL)
table[6,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[6,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat,memory=memory2)    )[stsel]
Rprof( NULL)
table[6,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# supply the permutation
pivot1 <- ch1@pivot
pivot2 <- ch2@pivot
Rprof( memory.profiling=TRUE, interval = rPint)
table[7,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,pivot=pivot1,
                                                             memory=memory1)    )[stsel]
Rprof( NULL)
table[7,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]
Rprof( memory.profiling=TRUE, interval = rPint)
table[7,3] <- system.time(   for (i in 1:N) ch1 <- chol.spam(USmat,pivot=pivot2,
                                                             memory=memory2)    )[stsel]
Rprof( NULL)
table[7,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Do not check the permutation
options(spam.cholpivotcheck=FALSE)
Rprof( memory.profiling=TRUE, interval = rPint)
table[8,1] <- system.time(   for (i in 1:N) ch1 <- chol.spam(gridmat,pivot=pivot1,
                                                              memory=memory1)    )[stsel]
Rprof( NULL)
table[8,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[8,3] <- system.time(   for (i in 1:N) ch2 <- chol.spam(USmat,pivot=pivot2,
                                                              memory=memory2)    )[stsel]
Rprof( NULL)
table[8,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

# Update only
Rprof( memory.profiling=TRUE, interval = rPint)
table[9,1] <- system.time(   for (i in 1:N) ch1 <- update(ch1,gridmat)   )[stsel]
Rprof( NULL)
table[9,2] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]

Rprof( memory.profiling=TRUE, interval = rPint)
table[9,3] <- system.time(   for (i in 1:N) ch2 <- update(ch2,USmat)   )[stsel]
Rprof( NULL)
table[9,4] <- summaryRprof(memory="both")$by.total[rPsx,rPsy]


# assemble the table
colnames(table) <- c("grid_time","grid_mem","US_time","US_mem")
rownames(table) <- c("Generic chol","chol.spam","safemode",
                     "symmetrycheck","memory","all","reusing pivot","best cast","update only")


normed.table <- t( round( t(table[-1,])/table[1,],3))

if (F) {
 print( t( round( t(table[-1,])/table[1,],3)))
}




# Figure 5
In <- diag.spam(nrow(UScounties.storder))
struct <- chol(In + .2 * UScounties.storder + .1 * UScounties.ndorder)

len.1 <- 10 # in the article, is set to 180
len.2 <- 5 # in the article, is set to 100
theta.1 <- seq(-.225, to=.515, len=len.1)
theta.2 <- seq(-.09, to=.235, len=len.2)

grid <- array(NA, c(len.1, len.2))
options(spam.cholupdatesingular='null')

for (i in 1:len.1)
  for(j in 1:len.2)
    grid[i,j] <- !is.null(update(struct, In + theta.1[i]*UScounties.storder
                       + theta.2[j]* UScounties.ndorder))


options( echo=TRUE)

# HEADER ####################################################
# This is file spam/R/spamlist.R.                           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



as.spam.list <- function(x, eps = getOption("spam.eps")) {
    spam.list( x,  eps=eps)
}

spam.list <-  function(x, nrow, ncol, eps = getOption("spam.eps")) {
    force64 <- getOption("spam.force64")
    
    if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
    if (!is.list(x)|(length(x)<2)|(length(x)>3))
        stop("Argument 'x' needs to be a list with two or three elements")
                                        # two cases: list of length
                                        # -  two (matrix with two columns called ind* and the elements)
                                        # -  three (each one column called i*, j*.   

    if (identical(length(x),2L)) {
        indnr <- pmatch("ind",names(x)) 
        if (is.na(indnr)) stop("Argument 'x' needs an element called 'indices'")
        elenr <- ifelse( identical( indnr,1L), 2L, 1L)
        
        nz <- length( x[[elenr]])

        dimx <- dim(x[[indnr]])
        if (is.null(dimx)||(dimx[2] != 2))  stop("Indices should have two columns")
        if (dimx[1] != nz) stop("Number of indices does not match with number of elements")
        
        ir <- as.integer(x[[indnr]][,1])
        jc <- as.integer(x[[indnr]][,2])

        if(force64 || length(x[[elenr]]) > 2147483646)
            SS <- .format64()
        else
            SS <- .format32
        
    } else {
        inr <- pmatch("i",names(x)) 
        jnr <- pmatch("j",names(x))
        
        if (is.na(inr)||is.na(jnr)) stop("Argument 'x' needs elements called 'i' and 'j'")
        elenr <- c(1:3)[-c(inr,jnr)]
        nz <- length( x[[elenr]])
        
        ir <- as.integer(x[[inr]])
        jc <- as.integer(x[[jnr]])

        if ((length(ir) != nz)||(length(jc) != nz))
            stop("Number of indices does not match with number of elements")

        if(force64 || length(x[[elenr]]) > 2147483646)
            SS <- .format64()
        else
            SS <- .format32
    }
    
    if (nz == 0)
        return(.newSpam(
            # rowpointers = c(1,rep_len64(2, nrow)),
            dimension = c(nrow,ncol),
            force64 = force64))
    if (any( ir <= 0) || any( jc <= 0))
        stop("Indices need to be positive")
    if (any(!is.finite(x[[elenr]]))) {
        warning("'NA/NaN/Inf' coerced to zero")
        x[[elenr]][!is.finite(x[[elenr]])] <- 0
    }
    nrow <- as.integer(ifelse(missing(nrow),max(ir),nrow))
    ncol <- as.integer(ifelse(missing(ncol),max(jc),ncol))
    ## z <- .Fortran(ifelse(toupper(getOption("spam.listmethod")=="PE"),"triplet3csr","triplet2csr"),
    ##               nrow=as.integer(nrow), ncol=as.integer(ncol),
    ##               nz=as.integer(nz),
    ##               as.double(x[[elenr]]),as.integer(ir),as.integer(jc),
    ##               entries=vector("double",nz),
    ##               colindices=vector("integer",nz),
    ##               rowpointers=vector("integer",nrow+1), as.double(eps),
    ##               NAOK=TRUE, PACKAGE = "spam"
    ##               )
    
    z <- .C64(ifelse(toupper(getOption("spam.listmethod")=="PE"),"triplet3csr","triplet2csr"),
              ## subroutine triplet3csr(nrow,ncol,nnz,a,ir,jc,ao,jao,iao,eps)
              ## subroutine triplet2csr(nrow,ncol,nnz,a,ir,jc,ao,jao,iao,eps)
              SIGNATURE = c( SS$signature, SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double", SS$signature, SS$signature,
                  "double"),
              
              nrow = nrow,
              ncol = ncol,
              nz = nz,
              
              x[[elenr]],
              ir,
              jc,
              
              entries = vector_dc( "double", nz),
              colindices = vector_dc( SS$type, nz),
              rowpointers = vector_dc( SS$type, nrow+1),
              
              eps,

              INTENT = c("r", "r", "rw",
                  "rw", "rw", "rw",
                  "rw", "rw", "rw",
                  "r"),
              NAOK=TRUE,
              PACKAGE = SS$package )

    

                                        #  print(z)
    if (z$nz == 0){
    ## if (identical(z$nz, 0)){
        ## print("special case")
        return(.newSpam(
            # rowpointers = c(1, rep_len64(2,nrow)),
            dimension = c(nrow, ncol),
            force64 = force64))
         ## return(new("spam",rowpointers=c(1L,rep.int(2L,nrow)), dimension=c(nrow,ncol)))
    }
   
    ## newx <- new("spam")
    ## slot(newx,"entries",check=FALSE) <- z$entries[1:z$nz]
    ## slot(newx,"colindices",check=FALSE) <- z$colindices[1:z$nz]
    ## slot(newx,"rowpointers",check=FALSE) <- z$rowpointers
    ## slot(newx,"dimension",check=FALSE) <- c(nrow,ncol)
    ## return(newx)
    return(.newSpam(
        entries = z$entries[1:z$nz],
        colindices = z$colindices[1:z$nz],
        rowpointers = z$rowpointers,
        dimension = c(nrow,ncol),
        force64 = force64))
}

setMethod("as.spam", "list", as.spam.list) #  { function(x,eps) spam.list(x,eps=eps)})
setMethod("spam", "list", spam.list)

c     system("R CMD SHLIB ../src/xybind.f")

      subroutine cbindf(xncol,nrow, a,ia,ja, b,ib,jb,
     &     c,ic,jc)

      implicit none
      integer xncol, nrow
      integer ia(*), ja(*), ib(*), jb(*), ic(*), jc(*)

      double precision  a(*), b(*), c(*)

      integer j,j1,i,k

      k=1
      do j = 1,nrow
         jc(j)=ja(j)+jb(j)-1

         j1=j+1
         if (ja(j) .lt. ja(j1)) then
            do i=ja(j),ja(j1)-1
               c(k)=a(i)
               ic(k)=ia(i)
               k=k+1
c               if (k.gt.clen)  return
            enddo
         endif
         if (jb(j) .lt. jb(j1)) then
            do i=jb(j),jb(j1)-1
               c(k)=b(i)
               ic(k)=ib(i)+xncol
               k=k+1
c               if (k.gt.clen)  return
            enddo
         endif
      enddo
      j=nrow+1
      jc(j)=ja(j)+jb(j)-1


      return
      end

      subroutine rbindf(xnrow,ynrow, xlen, ylen, a,ia,ja, b,ib,jb,
     &     c,ic,jc)

      implicit none
      integer xnrow, ynrow, xlen, ylen
      integer ia(*), ja(*), ib(*), jb(*), ic(*), jc(*)

      double precision  a(*), b(*), c(*)

      integer j

c     fill row pointers first      
      do j = 1,xnrow+1
         jc(j)=ja(j)
      enddo
      do j = 2,ynrow+1
         jc(j+xnrow)=ja(xnrow+1)+jb(j)-1
      enddo
      

c     now will colpointers and entries concatenation only...      
      do j = 1,xlen
         ic(j) = ia(j) 
         c(j) = a(j)
      enddo
      do j = 1, ylen
         ic(j+xlen) = ib(j) 
         c(j+xlen) = b(j)
      enddo   

      return
      end

c
      subroutine dn_eigen_f(maxnev, ncv, maxitr,
     &                      n, iwhich,
     &                      na, a, ja, ia,
     &                      v, dr, di, iparam)
c
      implicit none
c
      integer           maxnev, ncv, n, na,
     &                  iwhich, maxitr
c     %--------------%
c     | Local Arrays |
c     %--------------%
c
      integer           iparam(8), ipntr(14),
     &                  ja(*), ia(na+1)
c
      logical           select(ncv)
c
      Double precision
     &                  dr(maxnev+1), di(maxnev+1), resid(n),
     &                  v(n, ncv), workd(3*n),
     &                  workev(3*ncv),
     &                  workl(3*ncv**2+6*ncv),
     &                  a(*)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      character         bmat*1, which*2
      integer           ido, lworkl, info,
     &                  ierr, ishfts, mode
      Double precision
     &                  tol, sigmar, sigmai
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &                  zero
      parameter         (zero = 0.0D+0)
c
      bmat   = 'I'
c
      lworkl = 3*ncv*ncv+6*ncv
      tol    = zero
      ido    = 0
      info   = 0
c
      ishfts = 1
      mode   = 1
c
      iparam(1) = ishfts
      iparam(3) = maxitr
      iparam(7) = mode
c
      if (iwhich .eq. 1) then
            which = 'LM'
      else if (iwhich .eq. 2) then
            which = 'SM'
      else if (iwhich .eq. 3) then
            which = 'LR'
      else if (iwhich .eq. 4) then
            which = 'SR'
      else if (iwhich .eq. 5) then
            which = 'LI'
      else if (iwhich .eq. 6) then
            which = 'SI'
      else
c
        goto 9000
      end if
c
c
 10   continue
c
c        %---------------------------------------------%
c        | Repeatedly call the routine DNAUPD and take |
c        | actions indicated by parameter IDO until    |
c        | either convergence is indicated or maxitr   |
c        | has been exceeded.                          |
c        %---------------------------------------------%
c
         call dnaupd ( ido, bmat, n, which, maxnev, tol, resid,
     &        ncv, v, n, iparam, ipntr, workd, workl, lworkl,
     &        info )
c
      if (ido .eq. -1 .or. ido .eq. 1) then
c
            call d_ope (na, workd(ipntr(1)), workd(ipntr(2)),
     &                  a, ja, ia)
c
            go to 10
c
      end if
c
      if ( info .lt. 0 ) then
c
         goto 9000
c
      else
c
         call dneupd ( .true., 'A', select, dr, di, v, n,
     &         sigmar, sigmai, workev, bmat, n, which, maxnev, tol,
     &         resid, ncv, v, n, iparam, ipntr, workd, workl,
     &         lworkl, ierr )
c
         if ( ierr .lt. 0 ) then
c
            goto 9000
c
         end if
c
      endif
c
 9000 continue
c
      end
c

c     Modifications:
c     2023-10-17: eliminated all kind=4 occurences.
c
c      
c dneupd.f
c dnaupd.f
c dnaup2.f
c dnapps.f
c dnconv.f
c dsortc.f
c dneigh.f
c dlaqrb.f
c dngets.f
c dgetv0.f
c dnaitr.f
c


c\BeginDoc
c
c\Name: dneupd
c
c\Description:
c
c  This subroutine returns the converged approximations to eigenvalues
c  of A*z = lambda*B*z and (optionally):
c
c      (1) The corresponding approximate eigenvectors;
c
c      (2) An orthonormal basis for the associated approximate
c          invariant subspace;
c
c      (3) Both.
c
c  There is negligible additional cost to obtain eigenvectors.  An orthonormal
c  basis is always computed.  There is an additional storage cost of n*nev
c  if both are requested (in this case a separate array Z must be supplied).
c
c  The approximate eigenvalues and eigenvectors of  A*z = lambda*B*z
c  are derived from approximate eigenvalues and eigenvectors of
c  of the linear operator OP prescribed by the MODE selection in the
c  call to DNAUPD .  DNAUPD  must be called before this routine is called.
c  These approximate eigenvalues and vectors are commonly called Ritz
c  values and Ritz vectors respectively.  They are referred to as such
c  in the comments that follow.  The computed orthonormal basis for the
c  invariant subspace corresponding to these Ritz values is referred to as a
c  Schur basis.
c
c  See documentation in the header of the subroutine DNAUPD  for
c  definition of OP as well as other terms and the relation of computed
c  Ritz values and Ritz vectors of OP with respect to the given problem
c  A*z = lambda*B*z.  For a brief description, see definitions of
c  IPARAM(7), MODE and WHICH in the documentation of DNAUPD .
c
c\Usage:
c  call dneupd
c     ( RVEC, HOWMNY, SELECT, DR, DI, Z, LDZ, SIGMAR, SIGMAI, WORKEV, BMAT,
c       N, WHICH, NEV, TOL, RESID, NCV, V, LDV, IPARAM, IPNTR, WORKD, WORKL,
c       LWORKL, INFO )
c
c\Arguments:
c  RVEC    LOGICAL  (INPUT)
c          Specifies whether a basis for the invariant subspace corresponding
c          to the converged Ritz value approximations for the eigenproblem
c          A*z = lambda*B*z is computed.
c
c             RVEC = .FALSE.     Compute Ritz values only.
c
c             RVEC = .TRUE.      Compute the Ritz vectors or Schur vectors.
c                                See Remarks below.
c
c  HOWMNY  Character*1  (INPUT)
c          Specifies the form of the basis for the invariant subspace
c          corresponding to the converged Ritz values that is to be computed.
c
c          = 'A': Compute NEV Ritz vectors;
c          = 'P': Compute NEV Schur vectors;
c          = 'S': compute some of the Ritz vectors, specified
c                 by the logical array SELECT.
c
c  SELECT  Logical array of dimension NCV.  (INPUT)
c          If HOWMNY = 'S', SELECT specifies the Ritz vectors to be
c          computed. To select the Ritz vector corresponding to a
c          Ritz value (DR(j), DI(j)), SELECT(j) must be set to .TRUE..
c          If HOWMNY = 'A' or 'P', SELECT is used as internal workspace.
c
c  DR      Double precision  array of dimension NEV+1.  (OUTPUT)
c          If IPARAM(7) = 1,2 or 3 and SIGMAI=0.0  then on exit: DR contains
c          the real part of the Ritz  approximations to the eigenvalues of
c          A*z = lambda*B*z.
c          If IPARAM(7) = 3, 4 and SIGMAI is not equal to zero, then on exit:
c          DR contains the real part of the Ritz values of OP computed by
c          DNAUPD . A further computation must be performed by the user
c          to transform the Ritz values computed for OP by DNAUPD  to those
c          of the original system A*z = lambda*B*z. See remark 3 below.
c
c  DI      Double precision  array of dimension NEV+1.  (OUTPUT)
c          On exit, DI contains the imaginary part of the Ritz value
c          approximations to the eigenvalues of A*z = lambda*B*z associated
c          with DR.
c
c          NOTE: When Ritz values are complex, they will come in complex
c                conjugate pairs.  If eigenvectors are requested, the
c                corresponding Ritz vectors will also come in conjugate
c                pairs and the real and imaginary parts of these are
c                represented in two consecutive columns of the array Z
c                (see below).
c
c  Z       Double precision  N by NEV+1 array if RVEC = .TRUE. and HOWMNY = 'A'. (OUTPUT)
c          On exit, if RVEC = .TRUE. and HOWMNY = 'A', then the columns of
c          Z represent approximate eigenvectors (Ritz vectors) corresponding
c          to the NCONV=IPARAM(5) Ritz values for eigensystem
c          A*z = lambda*B*z.
c
c          The complex Ritz vector associated with the Ritz value
c          with positive imaginary part is stored in two consecutive
c          columns.  The first column holds the real part of the Ritz
c          vector and the second column holds the imaginary part.  The
c          Ritz vector associated with the Ritz value with negative
c          imaginary part is simply the complex conjugate of the Ritz vector
c          associated with the positive imaginary part.
c
c          If  RVEC = .FALSE. or HOWMNY = 'P', then Z is not referenced.
c
c          NOTE: If if RVEC = .TRUE. and a Schur basis is not required,
c          the array Z may be set equal to first NEV+1 columns of the Arnoldi
c          basis array V computed by DNAUPD .  In this case the Arnoldi basis
c          will be destroyed and overwritten with the eigenvector basis.
c
c  LDZ     Integer.  (INPUT)
c          The leading dimension of the array Z.  If Ritz vectors are
c          desired, then  LDZ >= max( 1, N ).  In any case,  LDZ >= 1.
c
c  SIGMAR  Double precision   (INPUT)
c          If IPARAM(7) = 3 or 4, represents the real part of the shift.
c          Not referenced if IPARAM(7) = 1 or 2.
c
c  SIGMAI  Double precision   (INPUT)
c          If IPARAM(7) = 3 or 4, represents the imaginary part of the shift.
c          Not referenced if IPARAM(7) = 1 or 2. See remark 3 below.
c
c  WORKEV  Double precision  work array of dimension 3*NCV.  (WORKSPACE)
c
c  **** The remaining arguments MUST be the same as for the   ****
c  **** call to DNAUPD  that was just completed.               ****
c
c  NOTE: The remaining arguments
c
c           BMAT, N, WHICH, NEV, TOL, RESID, NCV, V, LDV, IPARAM, IPNTR,
c           WORKD, WORKL, LWORKL, INFO
c
c         must be passed directly to DNEUPD  following the last call
c         to DNAUPD .  These arguments MUST NOT BE MODIFIED between
c         the the last call to DNAUPD  and the call to DNEUPD .
c
c  Three of these parameters (V, WORKL, INFO) are also output parameters:
c
c  V       Double precision  N by NCV array.  (INPUT/OUTPUT)
c
c          Upon INPUT: the NCV columns of V contain the Arnoldi basis
c                      vectors for OP as constructed by DNAUPD  .
c
c          Upon OUTPUT: If RVEC = .TRUE. the first NCONV=IPARAM(5) columns
c                       contain approximate Schur vectors that span the
c                       desired invariant subspace.  See Remark 2 below.
c
c          NOTE: If the array Z has been set equal to first NEV+1 columns
c          of the array V and RVEC=.TRUE. and HOWMNY= 'A', then the
c          Arnoldi basis held by V has been overwritten by the desired
c          Ritz vectors.  If a separate array Z has been passed then
c          the first NCONV=IPARAM(5) columns of V will contain approximate
c          Schur vectors that span the desired invariant subspace.
c
c  WORKL   Double precision  work array of length LWORKL.  (OUTPUT/WORKSPACE)
c          WORKL(1:ncv*ncv+3*ncv) contains information obtained in
c          dnaupd .  They are not changed by dneupd .
c          WORKL(ncv*ncv+3*ncv+1:3*ncv*ncv+6*ncv) holds the
c          real and imaginary part of the untransformed Ritz values,
c          the upper quasi-triangular matrix for H, and the
c          associated matrix representation of the invariant subspace for H.
c
c          Note: IPNTR(9:13) contains the pointer into WORKL for addresses
c          of the above information computed by dneupd .
c          -------------------------------------------------------------
c          IPNTR(9):  pointer to the real part of the NCV RITZ values of the
c                     original system.
c          IPNTR(10): pointer to the imaginary part of the NCV RITZ values of
c                     the original system.
c          IPNTR(11): pointer to the NCV corresponding error bounds.
c          IPNTR(12): pointer to the NCV by NCV upper quasi-triangular
c                     Schur matrix for H.
c          IPNTR(13): pointer to the NCV by NCV matrix of eigenvectors
c                     of the upper Hessenberg matrix H. Only referenced by
c                     dneupd  if RVEC = .TRUE. See Remark 2 below.
c          -------------------------------------------------------------
c
c  INFO    Integer.  (OUTPUT)
c          Error flag on output.
c
c          =  0: Normal exit.
c
c          =  1: The Schur form computed by LAPACK routine dlahqr
c                could not be reordered by LAPACK routine dtrsen .
c                Re-enter subroutine dneupd  with IPARAM(5)=NCV and
c                increase the size of the arrays DR and DI to have
c                dimension at least dimension NCV and allocate at least NCV
c                columns for Z. NOTE: Not necessary if Z and V share
c                the same space. Please notify the authors if this error
c                occurs.
c
c          = -1: N must be positive.
c          = -2: NEV must be positive.
c          = -3: NCV-NEV >= 2 and less than or equal to N.
c          = -5: WHICH must be one of 'LM', 'SM', 'LR', 'SR', 'LI', 'SI'
c          = -6: BMAT must be one of 'I' or 'G'.
c          = -7: Length of private work WORKL array is not sufficient.
c          = -8: Error return from calculation of a real Schur form.
c                Informational error from LAPACK routine dlahqr .
c          = -9: Error return from calculation of eigenvectors.
c                Informational error from LAPACK routine dtrevc .
c          = -10: IPARAM(7) must be 1,2,3,4.
c          = -11: IPARAM(7) = 1 and BMAT = 'G' are incompatible.
c          = -12: HOWMNY = 'S' not yet implemented
c          = -13: HOWMNY must be one of 'A' or 'P' if RVEC = .true.
c          = -14: DNAUPD  did not find any eigenvalues to sufficient
c                 accuracy.
c          = -15: DNEUPD got a different count of the number of converged
c                 Ritz values than DNAUPD got.  This indicates the user
c                 probably made an error in passing data from DNAUPD to
c                 DNEUPD or that the data was modified before entering
c                 DNEUPD
c
c\BeginLib
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c  3. B.N. Parlett & Y. Saad, "Complex Shift and Invert Strategies for
c     Real Matrices", Linear Algebra and its Applications, vol 88/89,
c     pp 575-595, (1987).
c
c\Routines called:
c     ivout   ARPACK utility routine that prints integers.
c     dmout    ARPACK utility routine that prints matrices
c     dvout    ARPACK utility routine that prints vectors.
c     dgeqr2   LAPACK routine that computes the QR factorization of
c             a matrix.
c     dlacpy   LAPACK matrix copy routine.
c     dlahqr   LAPACK routine to compute the real Schur form of an
c             upper Hessenberg matrix.
c     dlamch   LAPACK routine that determines machine constants.
c     dlapy2   LAPACK routine to compute sqrt(x**2+y**2) carefully.
c     dlaset   LAPACK matrix initialization routine.
c     dorm2r   LAPACK routine that applies an orthogonal matrix in
c             factored form.
c     dtrevc   LAPACK routine to compute the eigenvectors of a matrix
c             in upper quasi-triangular form.
c     dtrsen   LAPACK routine that re-orders the Schur form.
c     dtrmm    Level 3 BLAS matrix times an upper triangular matrix.
c     dger     Level 2 BLAS rank one update to a matrix.
c     dcopy    Level 1 BLAS that copies one vector to another .
c     ddot     Level 1 BLAS that computes the scalar product of two vectors.
c     dnrm2    Level 1 BLAS that computes the norm of a vector.
c     dscal    Level 1 BLAS that scales a vector.
c
c\Remarks
c
c  1. Currently only HOWMNY = 'A' and 'P' are implemented.
c
c     Let trans(X) denote the transpose of X.
c
c  2. Schur vectors are an orthogonal representation for the basis of
c     Ritz vectors. Thus, their numerical properties are often superior.
c     If RVEC = .TRUE. then the relationship
c             A * V(:,1:IPARAM(5)) = V(:,1:IPARAM(5)) * T, and
c     trans(V(:,1:IPARAM(5))) * V(:,1:IPARAM(5)) = I are approximately
c     satisfied. Here T is the leading submatrix of order IPARAM(5) of the
c     real upper quasi-triangular matrix stored workl(ipntr(12)). That is,
c     T is block upper triangular with 1-by-1 and 2-by-2 diagonal blocks;
c     each 2-by-2 diagonal block has its diagonal elements equal and its
c     off-diagonal elements of opposite sign.  Corresponding to each 2-by-2
c     diagonal block is a complex conjugate pair of Ritz values. The real
c     Ritz values are stored on the diagonal of T.
c
c  3. If IPARAM(7) = 3 or 4 and SIGMAI is not equal zero, then the user must
c     form the IPARAM(5) Rayleigh quotients in order to transform the Ritz
c     values computed by DNAUPD  for OP to those of A*z = lambda*B*z.
c     Set RVEC = .true. and HOWMNY = 'A', and
c     compute
c           trans(Z(:,I)) * A * Z(:,I) if DI(I) = 0.
c     If DI(I) is not equal to zero and DI(I+1) = - D(I),
c     then the desired real and imaginary parts of the Ritz value are
c           trans(Z(:,I)) * A * Z(:,I) +  trans(Z(:,I+1)) * A * Z(:,I+1),
c           trans(Z(:,I)) * A * Z(:,I+1) -  trans(Z(:,I+1)) * A * Z(:,I),
c     respectively.
c     Another possibility is to set RVEC = .true. and HOWMNY = 'P' and
c     compute trans(V(:,1:IPARAM(5))) * A * V(:,1:IPARAM(5)) and then an upper
c     quasi-triangular matrix of order IPARAM(5) is computed. See remark
c     2 above.
c
c\Authors
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Chao Yang                    Houston, Texas
c     Dept. of Computational &
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\SCCS Information: @(#)
c FILE: neupd.F   SID: 2.7   DATE OF SID: 09/20/00   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
      subroutine dneupd (rvec , howmny, select, dr    , di,
     &                   z    , ldz   , sigmar, sigmai, workev,
     &                   bmat , n     , which , nev   , tol,
     &                   resid, ncv   , v     , ldv   , iparam,
     &                   ipntr, workd , workl , lworkl, info)
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat, howmny, which*2
      logical    rvec
      integer    info, ldz, ldv, lworkl, n, ncv, nev
      Double precision
     &           sigmar, sigmai, tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    iparam(8), ipntr(14)
      logical    select(ncv)
      Double precision
     &           dr(nev+1)    , di(nev+1), resid(n)  ,
     &           v(ldv,ncv)   , z(ldz,*) , workd(3*n),
     &           workl(lworkl), workev(3*ncv)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0 , zero = 0.0D-5 )
c
      integer(4)
     &           ione
      parameter (ione = 1)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      character  type*6
      integer    bounds, ierr  , ih    , ihbds   ,
     &           iheigr, iheigi, iconj , nconv   ,
     &           invsub, iuptri, iwev  , iwork(1),
     &           j     , k     , ldh   , ldq     ,
     &           mode  , outncv, ritzr   ,
     &           ritzi , wri   , wrr   , irr     ,
     &           iri   , ibd   , ishift, numcnv  ,
     &           np    , jj
      logical    reord
      Double precision
     &           conds  , rnorm, sep  , temp,
     &           vl(1,1), temp1, eps23
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy  , dger   , dgeqr2 , dlacpy ,
     &           dlahqr , dlaset , dorm2r ,
     &           dtrevc , dtrmm  , dtrsen , dscal
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlapy2 , dnrm2 , dlamch , ddot
      external   dlapy2 , dnrm2 , dlamch , ddot
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    abs, min, sqrt
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %------------------------%
c     | Set default parameters |
c     %------------------------%
c
      mode = iparam(7)
      nconv = iparam(5)
      info = 0
c
c     %---------------------------------%
c     | Get machine dependent constant. |
c     %---------------------------------%
c
      eps23 = dlamch ('Epsilon-Machine')
      eps23 = eps23**(2.0D+0  / 3.0D+0 )
c
c     %--------------%
c     | Quick return |
c     %--------------%
c
      ierr = 0
c
      if (nconv .le. 0) then
         ierr = -14
      else if (n .le. 0) then
         ierr = -1
      else if (nev .le. 0) then
         ierr = -2
      else if (ncv .le. nev+1 .or.  ncv .gt. n) then
         ierr = -3
      else if (which .ne. 'LM' .and.
     &        which .ne. 'SM' .and.
     &        which .ne. 'LR' .and.
     &        which .ne. 'SR' .and.
     &        which .ne. 'LI' .and.
     &        which .ne. 'SI') then
         ierr = -5
      else if (bmat .ne. 'I' .and. bmat .ne. 'G') then
         ierr = -6
      else if (lworkl .lt. 3*ncv**2 + 6*ncv) then
         ierr = -7
      else if ( (howmny .ne. 'A' .and.
     &           howmny .ne. 'P' .and.
     &           howmny .ne. 'S') .and. rvec ) then
         ierr = -13
      else if (howmny .eq. 'S' ) then
         ierr = -12
      end if
c
      if (mode .eq. 1 .or. mode .eq. 2) then
         type = 'REGULR'
      else if (mode .eq. 3 .and. (abs(sigmai) .le. zero)) then
         type = 'SHIFTI'
      else if (mode .eq. 3 ) then
         type = 'REALPT'
      else if (mode .eq. 4 ) then
         type = 'IMAGPT'
      else
                                              ierr = -10
      end if
      if (mode .eq. 1 .and. bmat .eq. 'G')    ierr = -11
c
c     %------------%
c     | Error Exit |
c     %------------%
c
      if (ierr .ne. 0) then
         info = ierr
         go to 9000
      end if
c
c     %--------------------------------------------------------%
c     | Pointer into WORKL for address of H, RITZ, BOUNDS, Q   |
c     | etc... and the remaining workspace.                    |
c     | Also update pointer to be used on output.              |
c     | Memory is laid out as follows:                         |
c     | workl(1:ncv*ncv) := generated Hessenberg matrix        |
c     | workl(ncv*ncv+1:ncv*ncv+2*ncv) := real and imaginary   |
c     |                                   parts of ritz values |
c     | workl(ncv*ncv+2*ncv+1:ncv*ncv+3*ncv) := error bounds   |
c     %--------------------------------------------------------%
c
c     %-----------------------------------------------------------%
c     | The following is used and set by DNEUPD .                  |
c     | workl(ncv*ncv+3*ncv+1:ncv*ncv+4*ncv) := The untransformed |
c     |                             real part of the Ritz values. |
c     | workl(ncv*ncv+4*ncv+1:ncv*ncv+5*ncv) := The untransformed |
c     |                        imaginary part of the Ritz values. |
c     | workl(ncv*ncv+5*ncv+1:ncv*ncv+6*ncv) := The untransformed |
c     |                           error bounds of the Ritz values |
c     | workl(ncv*ncv+6*ncv+1:2*ncv*ncv+6*ncv) := Holds the upper |
c     |                             quasi-triangular matrix for H |
c     | workl(2*ncv*ncv+6*ncv+1: 3*ncv*ncv+6*ncv) := Holds the    |
c     |       associated matrix representation of the invariant   |
c     |       subspace for H.                                     |
c     | GRAND total of NCV * ( 3 * NCV + 6 ) locations.           |
c     %-----------------------------------------------------------%
c
      ih     = ipntr(5)
      ritzr  = ipntr(6)
      ritzi  = ipntr(7)
      bounds = ipntr(8)
      ldh    = ncv
      ldq    = ncv
      iheigr = bounds + ldh
      iheigi = iheigr + ldh
      ihbds  = iheigi + ldh
      iuptri = ihbds  + ldh
      invsub = iuptri + ldh*ncv
      ipntr(9)  = iheigr
      ipntr(10) = iheigi
      ipntr(11) = ihbds
      ipntr(12) = iuptri
      ipntr(13) = invsub
      wrr = 1
      wri = ncv + 1
      iwev = wri + ncv
c
c     %-----------------------------------------%
c     | irr points to the REAL part of the Ritz |
c     |     values computed by _neigh before    |
c     |     exiting _naup2.                     |
c     | iri points to the IMAGINARY part of the |
c     |     Ritz values computed by _neigh      |
c     |     before exiting _naup2.              |
c     | ibd points to the Ritz estimates        |
c     |     computed by _neigh before exiting   |
c     |     _naup2.                             |
c     %-----------------------------------------%
c
      irr = ipntr(14)+ncv*ncv
      iri = irr+ncv
      ibd = iri+ncv
c
c     %------------------------------------%
c     | RNORM is B-norm of the RESID(1:N). |
c     %------------------------------------%
c
      rnorm = workl(ih+2)
      workl(ih+2) = zero
c
c
      if (rvec) then
c
         reord = .false.
c
c        %---------------------------------------------------%
c        | Use the temporary bounds array to store indices   |
c        | These will be used to mark the select array later |
c        %---------------------------------------------------%
c
         do 10 j = 1,ncv
            workl(bounds+j-1) = j
            select(j) = .false.
   10    continue
c
c        %-------------------------------------%
c        | Select the wanted Ritz values.      |
c        | Sort the Ritz values so that the    |
c        | wanted ones appear at the tailing   |
c        | NEV positions of workl(irr) and     |
c        | workl(iri).  Move the corresponding |
c        | error estimates in workl(bound)     |
c        | accordingly.                        |
c        %-------------------------------------%
c
         np     = ncv - nev
         ishift = 0
         call dngets (ishift       , which     , nev       ,
     &                np           , workl(irr), workl(iri),
     &                workl(bounds))
c
c
c        %-----------------------------------------------------%
c        | Record indices of the converged wanted Ritz values  |
c        | Mark the select array for possible reordering       |
c        %-----------------------------------------------------%
c
         numcnv = 0
         do 11 j = 1,ncv
            temp1 = max(eps23,
     &                 dlapy2 ( workl(irr+ncv-j), workl(iri+ncv-j) ))
            jj = INT(workl(bounds + ncv - j))
            if (numcnv .lt. nconv .and.
     &          workl(ibd+jj-1) .le. tol*temp1) then
               select(jj) = .true.
               numcnv = numcnv + 1
               if (jj .gt. nev) reord = .true.
            endif
   11    continue
c
c        %-----------------------------------------------------------%
c        | Check the count (numcnv) of converged Ritz values with    |
c        | the number (nconv) reported by dnaupd.  If these two      |
c        | are different then there has probably been an error       |
c        | caused by incorrect passing of the dnaupd data.           |
c        %-----------------------------------------------------------%
c
c
         if (numcnv .ne. nconv) then
            info = -15
            go to 9000
         end if
c
c        %-----------------------------------------------------------%
c        | Call LAPACK routine dlahqr  to compute the real Schur form |
c        | of the upper Hessenberg matrix returned by DNAUPD .        |
c        | Make a copy of the upper Hessenberg matrix.               |
c        | Initialize the Schur vector matrix Q to the identity.     |
c        %-----------------------------------------------------------%
c
         call dcopy (ldh*ncv, workl(ih), ione,
     &               workl(iuptri), ione)
         call dlaset ('All', ncv, ncv,
     &                zero , one, workl(invsub),
     &                ldq)
         call dlahqr (.true., .true.       , ncv,
     &                ione, ncv, workl(iuptri),
     &                ldh   , workl(iheigr), workl(iheigi),
     &                ione, ncv, workl(invsub),
     &                ldq   , ierr)
         call dcopy (ncv, workl(invsub+ncv-1), ldq,
     &               workl(ihbds), 1)
c
         if (ierr .ne. 0) then
            info = -8
            go to 9000
         end if
c
         if (reord) then
c
c           %-----------------------------------------------------%
c           | Reorder the computed upper quasi-triangular matrix. |
c           %-----------------------------------------------------%
c
            call dtrsen ('None'       , 'V'          ,
     &                   select       , ncv          ,
     &                   workl(iuptri), ldh          ,
     &                   workl(invsub), ldq          ,
     &                   workl(iheigr), workl(iheigi),
     &                   nconv        , conds        ,
     &                   sep          , workl(ihbds) ,
     &                   ncv          , iwork        ,
     &                   1            , ierr)
c
            if (ierr .eq. 1) then
               info = 1
               go to 9000
            end if
c
         end if
c
c        %---------------------------------------%
c        | Copy the last row of the Schur vector |
c        | into workl(ihbds).  This will be used |
c        | to compute the Ritz estimates of      |
c        | converged Ritz values.                |
c        %---------------------------------------%
c
         call dcopy (ncv, workl(invsub+ncv-1), ldq, 
     &               workl(ihbds), ione)
c
c        %----------------------------------------------------%
c        | Place the computed eigenvalues of H into DR and DI |
c        | if a spectral transformation was not used.         |
c        %----------------------------------------------------%
c
         if (type .eq. 'REGULR') then
            call dcopy (nconv, workl(iheigr), 1, dr(ione), 1)
            call dcopy (nconv, workl(iheigi), 1, di(ione), 1)
         end if
c
c        %----------------------------------------------------------%
c        | Compute the QR factorization of the matrix representing  |
c        | the wanted invariant subspace located in the first NCONV |
c        | columns of workl(invsub,ldq).                            |
c        %----------------------------------------------------------%
c
         call dgeqr2 (ncv, nconv , workl(invsub),
     &               ldq, workev, workev(ncv+1),
     &               ierr)
c
c        %---------------------------------------------------------%
c        | * Postmultiply V by Q using dorm2r .                     |
c        | * Copy the first NCONV columns of VQ into Z.            |
c        | * Postmultiply Z by R.                                  |
c        | The N by NCONV matrix Z is now a matrix representation  |
c        | of the approximate invariant subspace associated with   |
c        | the Ritz values in workl(iheigr) and workl(iheigi)      |
c        | The first NCONV columns of V are now approximate Schur  |
c        | vectors associated with the real upper quasi-triangular |
c        | matrix of order NCONV in workl(iuptri)                  |
c        %---------------------------------------------------------%
c
         call dorm2r ('Right', 'Notranspose', n            ,
     &                ncv   , nconv        , workl(invsub),
     &                ldq   , workev       , v            ,
     &                ldv   , workd(n+1)   , ierr)
         call dlacpy ('All', n, nconv, v(1,1), ldv, z, ldz)
c
         do 20 j=1, nconv
c
c           %---------------------------------------------------%
c           | Perform both a column and row scaling if the      |
c           | diagonal element of workl(invsub,ldq) is negative |
c           | I'm lazy and don't take advantage of the upper    |
c           | quasi-triangular form of workl(iuptri,ldq)        |
c           | Note that since Q is orthogonal, R is a diagonal  |
c           | matrix consisting of plus or minus ones           |
c           %---------------------------------------------------%
c
            if (workl(invsub+(j-1)*ldq+j-1) .lt. zero) then
               call dscal (nconv, -one, workl(iuptri+j-1),
     &                     ldq)
               call dscal (nconv, -one, workl(iuptri+(j-1)*ldq),
     &                     ione)
            end if
c
 20      continue
c
         if (howmny .eq. 'A') then
c
c           %--------------------------------------------%
c           | Compute the NCONV wanted eigenvectors of T |
c           | located in workl(iuptri,ldq).              |
c           %--------------------------------------------%
c
            do 30 j=1, ncv
               if (j .le. nconv) then
                  select(j) = .true.
               else
                  select(j) = .false.
               end if
 30         continue
c
            call dtrevc ('Right', 'Select'     , select       ,
     &                   ncv    , workl(iuptri), ldq          ,
     &                   vl, ione, workl(invsub),
     &                   ldq    , ncv          , outncv       ,
     &                   workev(1) , ierr)
c
            if (ierr .ne. 0) then
                info = -9
                go to 9000
            end if
c
c           %------------------------------------------------%
c           | Scale the returning eigenvectors so that their |
c           | Euclidean norms are all one. LAPACK subroutine |
c           | dtrevc  returns each eigenvector normalized so  |
c           | that the element of largest magnitude has      |
c           | magnitude 1;                                   |
c           %------------------------------------------------%
c
            iconj = 0
            do 40 j=1, nconv
c
               if ( abs( workl(iheigi+j-1) ) .le. zero ) then
c
c                 %----------------------%
c                 | real eigenvalue case |
c                 %----------------------%
c
                  temp = dnrm2 ( ncv, workl(invsub+(j-1)*ldq),
     &                          ione)
                  call dscal ( ncv, one / temp,
     &                 workl(invsub+(j-1)*ldq), ione )
c
               else
c
c                 %-------------------------------------------%
c                 | Complex conjugate pair case. Note that    |
c                 | since the real and imaginary part of      |
c                 | the eigenvector are stored in consecutive |
c                 | columns, we further normalize by the      |
c                 | square root of two.                       |
c                 %-------------------------------------------%
c
                  if (iconj .eq. 0) then
                     temp = dlapy2 (dnrm2 (ncv,
     &                                   workl(invsub+(j-1)*ldq),
     &                                   1),
     &                             dnrm2 (ncv,
     &                                   workl(invsub+j*ldq),
     &                                   1))
                     call dscal (ncv, one/temp,
     &                           workl(invsub+(j-1)*ldq),
     &                           ione)
                     call dscal (ncv, one/temp,
     &                           workl(invsub+j*ldq),
     &                           ione )
                     iconj = 1
                  else
                     iconj = 0
                  end if
c
               end if
c
 40         continue
c
            call dgemv ('T', ncv, nconv, one, workl(invsub),
     &               ldq, workl(ihbds), ione,
     &               zero, workev(ione), 1)
c
            iconj = 0
            do 45 j=1, nconv
               if (workl(iheigi+j-1) .le. zero) then
c
c                 %-------------------------------------------%
c                 | Complex conjugate pair case. Note that    |
c                 | since the real and imaginary part of      |
c                 | the eigenvector are stored in consecutive |
c                 %-------------------------------------------%
c
                  if (iconj .eq. 0) then
                     workev(j) = dlapy2 (workev(j), workev(j+1))
                     workev(j+1) = workev(j)
                     iconj = 1
                  else
                     iconj = 0
                  end if
               end if
 45         continue
c
c           %---------------------------------------%
c           | Copy Ritz estimates into workl(ihbds) |
c           %---------------------------------------%
c
            call dcopy (nconv, workev, 1, workl(ihbds), 1)
c
c           %---------------------------------------------------------%
c           | Compute the QR factorization of the eigenvector matrix  |
c           | associated with leading portion of T in the first NCONV |
c           | columns of workl(invsub,ldq).                           |
c           %---------------------------------------------------------%
c
            call dgeqr2 (ncv, nconv , workl(invsub),
     &                   ldq, workev, workev(ncv+1),
     &                   ierr)
c
c           %----------------------------------------------%
c           | * Postmultiply Z by Q.                       |
c           | * Postmultiply Z by R.                       |
c           | The N by NCONV matrix Z is now contains the  |
c           | Ritz vectors associated with the Ritz values |
c           | in workl(iheigr) and workl(iheigi).          |
c           %----------------------------------------------%
c
            call dorm2r ('Right', 'Notranspose', n            ,
     &                   ncv  , nconv        , workl(invsub),
     &                   ldq  , workev       , z            ,
     &                   ldz  , workd(n+1)   , ierr)
c
            call dtrmm ('Right'   , 'Upper'       , 'No transpose',
     &                  'Non-unit', n            , nconv         ,
     &                  one       , workl(invsub), ldq           ,
     &                  z         , ldz)
c
         end if
c
      else
c
c        %------------------------------------------------------%
c        | An approximate invariant subspace is not needed.     |
c        | Place the Ritz values computed DNAUPD  into DR and DI |
c        %------------------------------------------------------%
c
         call dcopy (nconv, workl(ritzr), 1, dr(ione), 1)
         call dcopy (nconv, workl(ritzi), 1, di(ione), 1)
         call dcopy (nconv, workl(ritzr), 1, workl(iheigr), 1)
         call dcopy (nconv, workl(ritzi), 1, workl(iheigi), 1)
         call dcopy (nconv, workl(bounds), 1, workl(ihbds), 1)
      end if
c
c     %------------------------------------------------%
c     | Transform the Ritz values and possibly vectors |
c     | and corresponding error bounds of OP to those  |
c     | of A*x = lambda*B*x.                           |
c     %------------------------------------------------%
c
      if (type .eq. 'REGULR') then
c
         if (rvec)
     &      call dscal (ncv, rnorm, workl(ihbds), ione)
c
      else
c
c        %---------------------------------------%
c        |   A spectral transformation was used. |
c        | * Determine the Ritz estimates of the |
c        |   Ritz values in the original system. |
c        %---------------------------------------%
c
         if (type .eq. 'SHIFTI') then
c
            if (rvec)
     &         call dscal (ncv, rnorm, workl(ihbds), ione)
c
            do 50 k=1, ncv
               temp = dlapy2 ( workl(iheigr+k-1),
     &                        workl(iheigi+k-1) )
               workl(ihbds+k-1) = abs( workl(ihbds+k-1) )
     &                          / temp / temp
 50         continue
c
         else if (type .eq. 'REALPT') then
c
            do 60 k=1, ncv
 60         continue
c
         else if (type .eq. 'IMAGPT') then
c
            do 70 k=1, ncv
 70         continue
c
         end if
c
c        %-----------------------------------------------------------%
c        | *  Transform the Ritz values back to the original system. |
c        |    For TYPE = 'SHIFTI' the transformation is              |
c        |             lambda = 1/theta + sigma                      |
c        |    For TYPE = 'REALPT' or 'IMAGPT' the user must from     |
c        |    Rayleigh quotients or a projection. See remark 3 above.|
c        | NOTES:                                                    |
c        | *The Ritz vectors are not affected by the transformation. |
c        %-----------------------------------------------------------%
c
         if (type .eq. 'SHIFTI') then
c
            do 80 k=1, ncv
               temp = dlapy2 ( workl(iheigr+k-1),
     &                        workl(iheigi+k-1) )
               workl(iheigr+k-1) = workl(iheigr+k-1)/temp/temp
     &                           + sigmar
               workl(iheigi+k-1) = -workl(iheigi+k-1)/temp/temp
     &                           + sigmai
 80         continue
c
            call dcopy (nconv, workl(iheigr), 1, dr(ione), 1)
            call dcopy (nconv, workl(iheigi), 1, di(ione), 1)
c
         else if (type .eq. 'REALPT' .or. type .eq. 'IMAGPT') then
c
            call dcopy (nconv, workl(iheigr), 1, dr(ione), 1)
            call dcopy (nconv, workl(iheigi), 1, di(ione), 1)
c
         end if
c
      end if
c
c     %-------------------------------------------------%
c     | Eigenvector Purification step. Formally perform |
c     | one of inverse subspace iteration. Only used    |
c     | for MODE = 2.                                   |
c     %-------------------------------------------------%
c
      if (rvec .and. howmny .eq. 'A' .and. type .eq. 'SHIFTI') then
c
c        %------------------------------------------------%
c        | Purify the computed Ritz vectors by adding a   |
c        | little bit of the residual vector:             |
c        |                      T                         |
c        |          resid(:)*( e    s ) / theta           |
c        |                      NCV                       |
c        | where H s = s theta. Remember that when theta  |
c        | has nonzero imaginary part, the corresponding  |
c        | Ritz vector is stored across two columns of Z. |
c        %------------------------------------------------%
c
         iconj = 0
         do 110 j=1, nconv
c            if (workl(iheigi+j-1) .eq. zero) then
            if (abs(workl(iheigi+j-1)) .le. zero) then
               workev(j) =  workl(invsub+(j-1)*ldq+ncv-1) /
     &                      workl(iheigr+j-1)
            else if (iconj .eq. 0) then
               temp = dlapy2 ( workl(iheigr+j-1), workl(iheigi+j-1) )
               workev(j) = ( workl(invsub+(j-1)*ldq+ncv-1) *
     &                       workl(iheigr+j-1) +
     &                       workl(invsub+j*ldq+ncv-1) *
     &                       workl(iheigi+j-1) ) / temp / temp
               workev(j+1) = ( workl(invsub+j*ldq+ncv-1) *
     &                         workl(iheigr+j-1) -
     &                         workl(invsub+(j-1)*ldq+ncv-1) *
     &                         workl(iheigi+j-1) ) / temp / temp
               iconj = 1
            else
               iconj = 0
            end if
 110     continue
c
c        %---------------------------------------%
c        | Perform a rank one update to Z and    |
c        | purify all the Ritz vectors together. |
c        %---------------------------------------%
c
         call dger (n, nconv, one, resid, 1, workev, 1, z, ldz)
c
      end if
c
 9000 continue
c
      return
c
c     %---------------%
c     | End of DNEUPD  |
c     %---------------%
c
      end
c\BeginDoc
c
c\Name: dnaupd
c
c\Description:
c  Reverse communication interface for the Implicitly Restarted Arnoldi
c  iteration. This subroutine computes approximations to a few eigenpairs
c  of a linear operator "OP" with respect to a semi-inner product defined by
c  a symmetric positive semi-definite real matrix B. B may be the identity
c  matrix. NOTE: If the linear operator "OP" is real and symmetric
c  with respect to the real positive semi-definite symmetric matrix B,
c  i.e. B*OP = (OP`)*B, then subroutine dsaupd should be used instead.
c
c  The computed approximate eigenvalues are called Ritz values and
c  the corresponding approximate eigenvectors are called Ritz vectors.
c
c  dnaupd is usually called iteratively to solve one of the
c  following problems:
c
c  Mode 1:  A*x = lambda*x.
c           ===> OP = A  and  B = I.
c
c  Mode 2:  A*x = lambda*M*x, M symmetric positive definite
c           ===> OP = inv[M]*A  and  B = M.
c           ===> (If M can be factored see remark 3 below)
c
c  Mode 3:  A*x = lambda*M*x, M symmetric semi-definite
c           ===> OP = Real_Part{ inv[A - sigma*M]*M }  and  B = M.
c           ===> shift-and-invert mode (in real arithmetic)
c           If OP*x = amu*x, then
c           amu = 1/2 * [ 1/(lambda-sigma) + 1/(lambda-conjg(sigma)) ].
c           Note: If sigma is real, i.e. imaginary part of sigma is zero;
c                 Real_Part{ inv[A - sigma*M]*M } == inv[A - sigma*M]*M
c                 amu == 1/(lambda-sigma).
c
c  Mode 4:  A*x = lambda*M*x, M symmetric semi-definite
c           ===> OP = Imaginary_Part{ inv[A - sigma*M]*M }  and  B = M.
c           ===> shift-and-invert mode (in real arithmetic)
c           If OP*x = amu*x, then
c           amu = 1/2i * [ 1/(lambda-sigma) - 1/(lambda-conjg(sigma)) ].
c
c  Both mode 3 and 4 give the same enhancement to eigenvalues close to
c  the (complex) shift sigma.  However, as lambda goes to infinity,
c  the operator OP in mode 4 dampens the eigenvalues more strongly than
c  does OP defined in mode 3.
c
c  NOTE: The action of w <- inv[A - sigma*M]*v or w <- inv[M]*v
c        should be accomplished either by a direct method
c        using a sparse matrix factorization and solving
c
c           [A - sigma*M]*w = v  or M*w = v,
c
c        or through an iterative method for solving these
c        systems.  If an iterative method is used, the
c        convergence test must be more stringent than
c        the accuracy requirements for the eigenvalue
c        approximations.
c
c\Usage:
c  call dnaupd
c     ( IDO, BMAT, N, WHICH, NEV, TOL, RESID, NCV, V, LDV, IPARAM,
c       IPNTR, WORKD, WORKL, LWORKL, INFO )
c
c\Arguments
c  IDO     Integer.  (INPUT/OUTPUT)
c          Reverse communication flag.  IDO must be zero on the first
c          call to dnaupd.  IDO will be set internally to
c          indicate the type of operation to be performed.  Control is
c          then given back to the calling routine which has the
c          responsibility to carry out the requested operation and call
c          dnaupd with the result.  The operand is given in
c          WORKD(IPNTR(1)), the result must be put in WORKD(IPNTR(2)).
c          -------------------------------------------------------------
c          IDO =  0: first call to the reverse communication interface
c          IDO = -1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c                    This is for the initialization phase to force the
c                    starting vector into the range of OP.
c          IDO =  1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c                    In mode 3 and 4, the vector B * X is already
c                    available in WORKD(ipntr(3)).  It does not
c                    need to be recomputed in forming OP * X.
c          IDO =  2: compute  Y = B * X  where
c                    IPNTR(1) is the pointer into WORKD for X,
c                    IPNTR(2) is the pointer into WORKD for Y.
c          IDO =  3: compute the IPARAM(8) real and imaginary parts
c                    of the shifts where INPTR(14) is the pointer
c                    into WORKL for placing the shifts. See Remark
c                    5 below.
c          IDO = 99: done
c          -------------------------------------------------------------
c
c  BMAT    Character*1.  (INPUT)
c          BMAT specifies the type of the matrix B that defines the
c          semi-inner product for the operator OP.
c          BMAT = 'I' -> standard eigenvalue problem A*x = lambda*x
c          BMAT = 'G' -> generalized eigenvalue problem A*x = lambda*B*x
c
c  N       Integer.  (INPUT)
c          Dimension of the eigenproblem.
c
c  WHICH   Character*2.  (INPUT)
c          'LM' -> want the NEV eigenvalues of largest magnitude.
c          'SM' -> want the NEV eigenvalues of smallest magnitude.
c          'LR' -> want the NEV eigenvalues of largest real part.
c          'SR' -> want the NEV eigenvalues of smallest real part.
c          'LI' -> want the NEV eigenvalues of largest imaginary part.
c          'SI' -> want the NEV eigenvalues of smallest imaginary part.
c
c  NEV     Integer.  (INPUT/OUTPUT)
c          Number of eigenvalues of OP to be computed. 0 < NEV < N-1.
c
c  TOL     Double precision scalar.  (INPUT)
c          Stopping criterion: the relative accuracy of the Ritz value
c          is considered acceptable if BOUNDS(I) .LE. TOL*ABS(RITZ(I))
c          where ABS(RITZ(I)) is the magnitude when RITZ(I) is complex.
c          DEFAULT = DLAMCH('EPS')  (machine precision as computed
c                    by the LAPACK auxiliary subroutine DLAMCH).
c
c  RESID   Double precision array of length N.  (INPUT/OUTPUT)
c          On INPUT:
c          If INFO .EQ. 0, a random initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          On OUTPUT:
c          RESID contains the final residual vector.
c
c  NCV     Integer.  (INPUT)
c          Number of columns of the matrix V. NCV must satisfy the two
c          inequalities 2 <= NCV-NEV and NCV <= N.
c          This will indicate how many Arnoldi vectors are generated
c          at each iteration.  After the startup phase in which NEV
c          Arnoldi vectors are generated, the algorithm generates
c          approximately NCV-NEV Arnoldi vectors at each subsequent update
c          iteration. Most of the cost in generating each Arnoldi vector is
c          in the matrix-vector operation OP*x.
c          NOTE: 2 <= NCV-NEV in order that complex conjugate pairs of Ritz
c          values are kept together. (See remark 4 below)
c
c  V       Double precision array N by NCV.  (OUTPUT)
c          Contains the final set of Arnoldi basis vectors.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling program.
c
c  IPARAM  Integer array of length 11.  (INPUT/OUTPUT)
c          IPARAM(1) = ISHIFT: method for selecting the implicit shifts.
c          The shifts selected at each iteration are used to restart
c          the Arnoldi iteration in an implicit fashion.
c          -------------------------------------------------------------
c          ISHIFT = 0: the shifts are provided by the user via
c                      reverse communication.  The real and imaginary
c                      parts of the NCV eigenvalues of the Hessenberg
c                      matrix H are returned in the part of the WORKL
c                      array corresponding to RITZR and RITZI. See remark
c                      5 below.
c          ISHIFT = 1: exact shifts with respect to the current
c                      Hessenberg matrix H.  This is equivalent to
c                      restarting the iteration with a starting vector
c                      that is a linear combination of approximate Schur
c                      vectors associated with the "wanted" Ritz values.
c          -------------------------------------------------------------
c
c          IPARAM(2) = No longer referenced.
c          IPARAM(3) = MXITER
c          On INPUT:  maximum number of Arnoldi update iterations allowed.
c          On OUTPUT: actual number of Arnoldi update iterations taken.
c
c          IPARAM(4) = NB: blocksize to be used in the recurrence.
c          The code currently works only for NB = 1.
c
c          IPARAM(5) = NCONV: number of "converged" Ritz values.
c          This represents the number of Ritz values that satisfy
c          the convergence criterion.
c
c          IPARAM(6) = IUPD
c          No longer referenced. Implicit restarting is ALWAYS used.
c
c          IPARAM(7) = MODE
c          On INPUT determines what type of eigenproblem is being solved.
c          Must be 1,2,3,4; See under \Description of dnaupd for the
c          four modes available.
c
c          IPARAM(8) = NP
c          When ido = 3 and the user provides shifts through reverse
c          communication (IPARAM(1)=0), dnaupd returns NP, the number
c          of shifts the user is to provide. 0 < NP <=NCV-NEV. See Remark
c          5 below.
c
crc          IPARAM(9) = NUMOP, IPARAM(10) = NUMOPB, IPARAM(11) = NUMREO,
crc          OUTPUT: NUMOP  = total number of OP*x operations,
crc                  NUMOPB = total number of B*x operations if BMAT='G',
crc                  NUMREO = total number of steps of re-orthogonalization.
c
c  IPNTR   Integer array of length 14.  (OUTPUT)
c          Pointer to mark the starting locations in the WORKD and WORKL
c          arrays for matrices/vectors used by the Arnoldi iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X in WORKD.
c          IPNTR(2): pointer to the current result vector Y in WORKD.
c          IPNTR(3): pointer to the vector B * X in WORKD when used in
c                    the shift-and-invert mode.
c          IPNTR(4): pointer to the next available location in WORKL
c                    that is untouched by the program.
c          IPNTR(5): pointer to the NCV by NCV upper Hessenberg matrix
c                    H in WORKL.
c          IPNTR(6): pointer to the real part of the ritz value array
c                    RITZR in WORKL.
c          IPNTR(7): pointer to the imaginary part of the ritz value array
c                    RITZI in WORKL.
c          IPNTR(8): pointer to the Ritz estimates in array WORKL associated
c                    with the Ritz values located in RITZR and RITZI in WORKL.
c
c          IPNTR(14): pointer to the NP shifts in WORKL. See Remark 5 below.
c
c          Note: IPNTR(9:13) is only referenced by dneupd. See Remark 2 below.
c
c          IPNTR(9):  pointer to the real part of the NCV RITZ values of the
c                     original system.
c          IPNTR(10): pointer to the imaginary part of the NCV RITZ values of
c                     the original system.
c          IPNTR(11): pointer to the NCV corresponding error bounds.
c          IPNTR(12): pointer to the NCV by NCV upper quasi-triangular
c                     Schur matrix for H.
c          IPNTR(13): pointer to the NCV by NCV matrix of eigenvectors
c                     of the upper Hessenberg matrix H. Only referenced by
c                     dneupd if RVEC = .TRUE. See Remark 2 below.
c          -------------------------------------------------------------
c
c  WORKD   Double precision work array of length 3*N.  (REVERSE COMMUNICATION)
c          Distributed array to be used in the basic Arnoldi iteration
c          for reverse communication.  The user should not use WORKD
c          as temporary workspace during the iteration. Upon termination
c          WORKD(1:N) contains B*RESID(1:N). If an invariant subspace
c          associated with the converged Ritz values is desired, see remark
c          2 below, subroutine dneupd uses this output.
c          See Data Distribution Note below.
c
c  WORKL   Double precision work array of length LWORKL.  (OUTPUT/WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.  See Data Distribution Note below.
c
c  LWORKL  Integer.  (INPUT)
c          LWORKL must be at least 3*NCV**2 + 6*NCV.
c
c  INFO    Integer.  (INPUT/OUTPUT)
c          If INFO .EQ. 0, a randomly initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          Error flag on output.
c          =  0: Normal exit.
c          =  1: Maximum number of iterations taken.
c                All possible eigenvalues of OP has been found. IPARAM(5)
c                returns the number of wanted converged Ritz values.
c          =  2: No longer an informational error. Deprecated starting
c                with release 2 of ARPACK.
c          =  3: No shifts could be applied during a cycle of the
c                Implicitly restarted Arnoldi iteration. One possibility
c                is to increase the size of NCV relative to NEV.
c                See remark 4 below.
c          = -1: N must be positive.
c          = -2: NEV must be positive.
c          = -3: NCV-NEV >= 2 and less than or equal to N.
c          = -4: The maximum number of Arnoldi update iteration
c                must be greater than zero.
c          = -5: WHICH must be one of 'LM', 'SM', 'LR', 'SR', 'LI', 'SI'
c          = -6: BMAT must be one of 'I' or 'G'.
c          = -7: Length of private work array is not sufficient.
c          = -8: Error return from LAPACK eigenvalue calculation;
c          = -9: Starting vector is zero.
c          = -10: IPARAM(7) must be 1,2,3,4.
c          = -11: IPARAM(7) = 1 and BMAT = 'G' are incompatable.
c          = -12: IPARAM(1) must be equal to 0 or 1.
c          = -9999: Could not build an Arnoldi factorization.
c                   IPARAM(5) returns the size of the current Arnoldi
c                   factorization.
c
c\Remarks
c  1. The computed Ritz values are approximate eigenvalues of OP. The
c     selection of WHICH should be made with this in mind when
c     Mode = 3 and 4.  After convergence, approximate eigenvalues of the
c     original problem may be obtained with the ARPACK subroutine dneupd.
c
c  2. If a basis for the invariant subspace corresponding to the converged Ritz
c     values is needed, the user must call dneupd immediately following
c     completion of dnaupd. This is new starting with release 2 of ARPACK.
c
c  3. If M can be factored into a Cholesky factorization M = LL`
c     then Mode = 2 should not be selected.  Instead one should use
c     Mode = 1 with  OP = inv(L)*A*inv(L`).  Appropriate triangular
c     linear systems should be solved with L and L` rather
c     than computing inverses.  After convergence, an approximate
c     eigenvector z of the original problem is recovered by solving
c     L`z = x  where x is a Ritz vector of OP.
c
c  4. At present there is no a-priori analysis to guide the selection
c     of NCV relative to NEV.  The only formal requrement is that NCV > NEV + 2.
c     However, it is recommended that NCV .ge. 2*NEV+1.  If many problems of
c     the same type are to be solved, one should experiment with increasing
c     NCV while keeping NEV fixed for a given test problem.  This will
c     usually decrease the required number of OP*x operations but it
c     also increases the work and storage required to maintain the orthogonal
c     basis vectors.  The optimal "cross-over" with respect to CPU time
c     is problem dependent and must be determined empirically.
c     See Chapter 8 of Reference 2 for further information.
c
c  5. When IPARAM(1) = 0, and IDO = 3, the user needs to provide the
c     NP = IPARAM(8) real and imaginary parts of the shifts in locations
c         real part                  imaginary part
c         -----------------------    --------------
c     1   WORKL(IPNTR(14))           WORKL(IPNTR(14)+NP)
c     2   WORKL(IPNTR(14)+1)         WORKL(IPNTR(14)+NP+1)
c                        .                          .
c                        .                          .
c                        .                          .
c     NP  WORKL(IPNTR(14)+NP-1)      WORKL(IPNTR(14)+2*NP-1).
c
c     Only complex conjugate pairs of shifts may be applied and the pairs
c     must be placed in consecutive locations. The real part of the
c     eigenvalues of the current upper Hessenberg matrix are located in
c     WORKL(IPNTR(6)) through WORKL(IPNTR(6)+NCV-1) and the imaginary part
c     in WORKL(IPNTR(7)) through WORKL(IPNTR(7)+NCV-1). They are ordered
c     according to the order defined by WHICH. The complex conjugate
c     pairs are kept together and the associated Ritz estimates are located in
c     WORKL(IPNTR(8)), WORKL(IPNTR(8)+1), ... , WORKL(IPNTR(8)+NCV-1).
c
c-----------------------------------------------------------------------
c
c\Data Distribution Note:
c
c  Fortran-D syntax:
c  ================
c  Double precision resid(n), v(ldv,ncv), workd(3*n), workl(lworkl)
c  decompose  d1(n), d2(n,ncv)
c  align      resid(i) with d1(i)
c  align      v(i,j)   with d2(i,j)
c  align      workd(i) with d1(i)     range (1:n)
c  align      workd(i) with d1(i-n)   range (n+1:2*n)
c  align      workd(i) with d1(i-2*n) range (2*n+1:3*n)
c  distribute d1(block), d2(block,:)
c  replicated workl(lworkl)
c
c  Cray MPP syntax:
c  ===============
c  Double precision  resid(n), v(ldv,ncv), workd(n,3), workl(lworkl)
c  shared     resid(block), v(block,:), workd(block,:)
c  replicated workl(lworkl)
c
c  CM2/CM5 syntax:
c  ==============
c
c-----------------------------------------------------------------------
c
c     include   'ex-nonsym.doc'
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c  3. B.N. Parlett & Y. Saad, "Complex Shift and Invert Strategies for
c     Real Matrices", Linear Algebra and its Applications, vol 88/89,
c     pp 575-595, (1987).
c
c\Routines called:
c     dnaup2  ARPACK routine that implements the Implicitly Restarted
c             Arnoldi Iteration.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dvout   ARPACK utility routine that prints vectors.
c     dlamch  LAPACK routine that determines machine constants.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     12/16/93: Version '1.1'
c
c\SCCS Information: @(#)
c FILE: naupd.F   SID: 2.10   DATE OF SID: 08/23/02   RELEASE: 2
c
c\Remarks
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dnaupd
     &   ( ido, bmat, n, which, nev, tol, resid, ncv, v, ldv, iparam,
     &     ipntr, workd, workl, lworkl, info )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1, which*2
      integer    ido, info, ldv, lworkl, n, ncv, nev
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    iparam(8), ipntr(14)
      Double precision
     &           resid(n), v(ldv,ncv), workd(3*n), workl(lworkl)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           zero
      parameter (zero = 0.0D+0)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    bounds, ierr, ih, iq, ishift, iupd, iw,
     &           ldh, ldq, levec, mode, mxiter, nb,
     &           nev0, next, np, ritzi, ritzr, j
      save       bounds, ih, iq, ishift, iupd, iw, ldh, ldq,
     &           levec, mode,  mxiter, nb, nev0, next,
     &           np, ritzi, ritzr
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dnaup2
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlamch
      external   dlamch
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (levec .gt. 10000000) then
        goto 9000
      end if
c
      if (ido .eq. 0) then
c
c        %-------------------------------%
c        | Initialize timing statistics  |
c        | & message level for debugging |
c        %-------------------------------%
c
c        %----------------%
c        | Error checking |
c        %----------------%
c
         ierr   = 0
         ishift = iparam(1)
c         levec  = iparam(2)
         mxiter = iparam(3)
c         nb     = iparam(4)
         nb     = 1
c
c        %--------------------------------------------%
c        | Revision 2 performs only implicit restart. |
c        %--------------------------------------------%
c
         iupd   = 1
         mode   = iparam(7)
c
         if (n .le. 0) then
            ierr = -1
         else if (nev .le. 0) then
            ierr = -2
         else if (ncv .le. nev+1 .or.  ncv .gt. n) then
            ierr = -3
         else if (mxiter .le.          0) then
            ierr = 4
         else if (which .ne. 'LM' .and.
     &       which .ne. 'SM' .and.
     &       which .ne. 'LR' .and.
     &       which .ne. 'SR' .and.
     &       which .ne. 'LI' .and.
     &       which .ne. 'SI') then
            ierr = -5
         else if (bmat .ne. 'I' .and. bmat .ne. 'G') then
            ierr = -6
         else if (lworkl .lt. 3*ncv**2 + 6*ncv) then
            ierr = -7
         else if (mode .lt. 1 .or. mode .gt. 4) then
            ierr = -10
         else if (mode .eq. 1 .and. bmat .eq. 'G') then
            ierr = -11
         else if (ishift .lt. 0 .or. ishift .gt. 1) then
            ierr = -12
         end if
c
c        %------------%
c        | Error Exit |
c        %------------%
c
         if (ierr .ne. 0) then
            info = ierr
            ido  = 99
            go to 9000
         end if
c
c        %------------------------%
c        | Set default parameters |
c        %------------------------%
c
         if (nb .le. 0)           nb = 1
         if (tol .le. zero)       tol = dlamch('EpsMach')
c
c        %----------------------------------------------%
c        | NP is the number of additional steps to      |
c        | extend the length NEV Lanczos factorization. |
c        | NEV0 is the local variable designating the   |
c        | size of the invariant subspace desired.      |
c        %----------------------------------------------%
c
         np     = ncv - nev
         nev0   = nev
c
c        %-----------------------------%
c        | Zero out internal workspace |
c        %-----------------------------%
c
         do 10 j = 1, 3*ncv**2 + 6*ncv
            workl(j) = zero
  10     continue
c
c        %-------------------------------------------------------------%
c        | Pointer into WORKL for address of H, RITZ, BOUNDS, Q        |
c        | etc... and the remaining workspace.                         |
c        | Also update pointer to be used on output.                   |
c        | Memory is laid out as follows:                              |
c        | workl(1:ncv*ncv) := generated Hessenberg matrix             |
c        | workl(ncv*ncv+1:ncv*ncv+2*ncv) := real and imaginary        |
c        |                                   parts of ritz values      |
c        | workl(ncv*ncv+2*ncv+1:ncv*ncv+3*ncv) := error bounds        |
c        | workl(ncv*ncv+3*ncv+1:2*ncv*ncv+3*ncv) := rotation matrix Q |
c        | workl(2*ncv*ncv+3*ncv+1:3*ncv*ncv+6*ncv) := workspace       |
c        | The final workspace is needed by subroutine dneigh called   |
c        | by dnaup2. Subroutine dneigh calls LAPACK routines for      |
c        | calculating eigenvalues and the last row of the eigenvector |
c        | matrix.                                                     |
c        %-------------------------------------------------------------%
c
         ldh    = ncv
         ldq    = ncv
         ih     = 1
         ritzr  = ih     + ldh*ncv
         ritzi  = ritzr  + ncv
         bounds = ritzi  + ncv
         iq     = bounds + ncv
         iw     = iq     + ldq*ncv
         next   = iw     + ncv**2 + 3*ncv
c
         ipntr(4) = next
         ipntr(5) = ih
         ipntr(6) = ritzr
         ipntr(7) = ritzi
         ipntr(8) = bounds
         ipntr(14) = iw
c
      end if
c
c     %-------------------------------------------------------%
c     | Carry out the Implicitly restarted Arnoldi Iteration. |
c     %-------------------------------------------------------%
c
      call dnaup2
     &   ( ido, bmat, n, which, nev0, np, tol, resid, mode,
     &     ishift, mxiter, v, ldv, workl(ih), ldh, workl(ritzr),
     &     workl(ritzi), workl(bounds), workl(iq), ldq, workl(iw),
     &     ipntr, workd, info )
c
c     %--------------------------------------------------%
c     | ido .ne. 99 implies use of reverse communication |
c     | to compute operations involving OP or shifts.    |
c     %--------------------------------------------------%
c
      if (ido .eq. 3) iparam(8) = np
      if (ido .ne. 99) go to 9000
c
      iparam(3) = mxiter
      iparam(5) = np
c
c     %------------------------------------%
c     | Exit if there was an informational |
c     | error within dnaup2.               |
c     %------------------------------------%
c
      if (info .lt. 0) go to 9000
      if (info .eq. 2) info = 3
c
c        %--------------------------------------------------------%
c        | Version Number & Version Date are defined in version.h |
c        %--------------------------------------------------------%
c
 9000 continue
c
      return
c
c     %---------------%
c     | End of dnaupd |
c     %---------------%
c
      end
c\BeginDoc
c
c\Name: dnaup2
c
c\Description:
c  Intermediate level interface called by dnaupd.
c
c\Usage:
c  call dnaup2
c     ( IDO, BMAT, N, WHICH, NEV, NP, TOL, RESID, MODE, IUPD,
c       ISHIFT, MXITER, V, LDV, H, LDH, RITZR, RITZI, BOUNDS,
c       Q, LDQ, WORKL, IPNTR, WORKD, INFO )
c
c\Arguments
c
c  IDO, BMAT, N, WHICH, NEV, TOL, RESID: same as defined in dnaupd.
c  MODE, ISHIFT, MXITER: see the definition of IPARAM in dnaupd.
c
c  NP      Integer.  (INPUT/OUTPUT)
c          Contains the number of implicit shifts to apply during
c          each Arnoldi iteration.
c          If ISHIFT=1, NP is adjusted dynamically at each iteration
c          to accelerate convergence and prevent stagnation.
c          This is also roughly equal to the number of matrix-vector
c          products (involving the operator OP) per Arnoldi iteration.
c          The logic for adjusting is contained within the current
c          subroutine.
c          If ISHIFT=0, NP is the number of shifts the user needs
c          to provide via reverse comunication. 0 < NP < NCV-NEV.
c          NP may be less than NCV-NEV for two reasons. The first, is
c          to keep complex conjugate pairs of "wanted" Ritz values
c          together. The second, is that a leading block of the current
c          upper Hessenberg matrix has split off and contains "unwanted"
c          Ritz values.
c          Upon termination of the IRA iteration, NP contains the number
c          of "converged" wanted Ritz values.
c
c  IUPD    Integer.  (INPUT)
c          IUPD .EQ. 0: use explicit restart instead implicit update.
c          IUPD .NE. 0: use implicit update.
c
c  V       Double precision N by (NEV+NP) array.  (INPUT/OUTPUT)
c          The Arnoldi basis vectors are returned in the first NEV
c          columns of V.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (NEV+NP) by (NEV+NP) array.  (OUTPUT)
c          H is used to store the generated upper Hessenberg matrix
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  RITZR,  Double precision arrays of length NEV+NP.  (OUTPUT)
c  RITZI   RITZR(1:NEV) (resp. RITZI(1:NEV)) contains the real (resp.
c          imaginary) part of the computed Ritz values of OP.
c
c  BOUNDS  Double precision array of length NEV+NP.  (OUTPUT)
c          BOUNDS(1:NEV) contain the error bounds corresponding to
c          the computed Ritz values.
c
c  Q       Double precision (NEV+NP) by (NEV+NP) array.  (WORKSPACE)
c          Private (replicated) work array used to accumulate the
c          rotation in the shift application step.
c
c  LDQ     Integer.  (INPUT)
c          Leading dimension of Q exactly as declared in the calling
c          program.
c
c  WORKL   Double precision work array of length at least
c          (NEV+NP)**2 + 3*(NEV+NP).  (INPUT/WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.  It is used in shifts calculation, shifts
c          application and convergence checking.
c
c          On exit, the last 3*(NEV+NP) locations of WORKL contain
c          the Ritz values (real,imaginary) and associated Ritz
c          estimates of the current Hessenberg matrix.  They are
c          listed in the same order as returned from dneigh.
c
c          If ISHIFT .EQ. O and IDO .EQ. 3, the first 2*NP locations
c          of WORKL are used in reverse communication to hold the user
c          supplied shifts.
c
c  IPNTR   Integer array of length 3.  (OUTPUT)
c          Pointer to mark the starting locations in the WORKD for
c          vectors used by the Arnoldi iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X.
c          IPNTR(2): pointer to the current result vector Y.
c          IPNTR(3): pointer to the vector B * X when used in the
c                    shift-and-invert mode.  X is the current operand.
c          -------------------------------------------------------------
c
c  WORKD   Double precision work array of length 3*N.  (WORKSPACE)
c          Distributed array to be used in the basic Arnoldi iteration
c          for reverse communication.  The user should not use WORKD
c          as temporary workspace during the iteration !!!!!!!!!!
c          See Data Distribution Note in DNAUPD.
c
c  INFO    Integer.  (INPUT/OUTPUT)
c          If INFO .EQ. 0, a randomly initial residual vector is used.
c          If INFO .NE. 0, RESID contains the initial residual vector,
c                          possibly from a previous run.
c          Error flag on output.
c          =     0: Normal return.
c          =     1: Maximum number of iterations taken.
c                   All possible eigenvalues of OP has been found.
c                   NP returns the number of converged Ritz values.
c          =     2: No shifts could be applied.
c          =    -8: Error return from LAPACK eigenvalue calculation;
c                   This should never happen.
c          =    -9: Starting vector is zero.
c          = -9999: Could not build an Arnoldi factorization.
c                   Size that was built in returned in NP.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c
c\Routines called:
c     dgetv0  ARPACK initial vector generation routine.
c     dnaitr  ARPACK Arnoldi factorization routine.
c     dnapps  ARPACK application of implicit shifts routine.
c     dnconv  ARPACK convergence of Ritz values routine.
c     dneigh  ARPACK compute Ritz values and error bounds routine.
c     dngets  ARPACK reorder Ritz values and error bounds routine.
c     dsortc  ARPACK sorting routine.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dmout   ARPACK utility routine that prints matrices
c     dvout   ARPACK utility routine that prints vectors.
c     dlamch  LAPACK routine that determines machine constants.
c     dlapy2  LAPACK routine to compute sqrt(x**2+y**2) carefully.
c     dcopy   Level 1 BLAS that copies one vector to another .
c     ddot    Level 1 BLAS that computes the scalar product of two vectors.
c     dnrm2   Level 1 BLAS that computes the norm of a vector.
c     dswap   Level 1 BLAS that swaps two vectors.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\SCCS Information: @(#)
c FILE: naup2.F   SID: 2.8   DATE OF SID: 10/17/00   RELEASE: 2
c
c\Remarks
c     1. None
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dnaup2
     &   ( ido, bmat, n, which, nev, np, tol, resid, mode,
     &     ishift, mxiter, v, ldv, h, ldh, ritzr, ritzi, bounds,
     &     q, ldq, workl, ipntr, workd, info )
c
c     %----------------------------------------------------%
c     | Include files for debugging and timing information |
c     %----------------------------------------------------%
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1, which*2
      integer    ido, info, ishift, mode, ldh, ldq, ldv, mxiter,
     &           n, nev, np
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    ipntr(13)
      Double precision
     &           bounds(nev+np), h(ldh,nev+np), q(ldq,nev+np), resid(n),
     &           ritzi(nev+np), ritzr(nev+np), v(ldv,nev+np),
     &           workd(3*n), workl( (nev+np)*(nev+np+3) )
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           zero
      parameter (zero = 0.0D+0)
      integer    dnazero
      integer(4) ione
      parameter (dnazero = 0, ione = 1)

c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      character  wprime*2
      logical    cnorm , getv0, initv, update, ushift, dnatrue
      integer    ierr  , iter , j    , kplusp, nconv,
     &           nevbef, nev0 , np0  , nptemp, numcnv
      Double precision
     &           rnorm , temp , eps23
      save       cnorm , getv0, initv, update, ushift,
     &           rnorm , iter , eps23, kplusp,  nconv ,
     &           nevbef, nev0 , np0  , numcnv
c
c     %-----------------------%
c     | Local array arguments |
c     %-----------------------%
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy , dgetv0, dnaitr, dnconv, dneigh,
     &           dngets, dnapps
c     &           dngets, dnapps, second
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
c
      Double precision
     &           ddot, dnrm2, dlapy2, dlamch
      external   ddot, dnrm2, dlapy2, dlamch
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    min, max, abs, sqrt
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      dnatrue = .true.
c
      if (ido .eq. 0) then
c
c        %-------------------------------------%
c        | Get the machine dependent constant. |
c        %-------------------------------------%
c
         eps23 = dlamch('Epsilon-Machine')
         eps23 = eps23**(2.0D+0 / 3.0D+0)
c
         nev0   = nev
         np0    = np
c
c        %-------------------------------------%
c        | kplusp is the bound on the largest  |
c        |        Lanczos factorization built. |
c        | nconv is the current number of      |
c        |        "converged" eigenvlues.      |
c        | iter is the counter on the current  |
c        |      iteration step.                |
c        %-------------------------------------%
c
         kplusp = nev + np
         nconv  = 0
         iter   = 0
c
c        %---------------------------------------%
c        | Set flags for computing the first NEV |
c        | steps of the Arnoldi factorization.   |
c        %---------------------------------------%
c
         getv0    = .true.
         update   = .false.
         ushift   = .false.
         cnorm    = .false.
c
         if (info .ne. 0) then
c
c           %--------------------------------------------%
c           | User provides the initial residual vector. |
c           %--------------------------------------------%
c
            initv = .true.
            info  = 0
         else
            initv = .false.
         end if
      end if
c
c     %---------------------------------------------%
c     | Get a possibly random starting vector and   |
c     | force it into the range of the operator OP. |
c     %---------------------------------------------%
c
c   10 continue
c
      if (getv0) then
         call dgetv0 (ido, bmat, initv, n, ione, v, ldv, resid, rnorm,
     &                ipntr, workd, info)
c
         if (ido .ne. 99) go to 9000
c
         if (abs(rnorm) .le. zero) then
c         if (rnorm .eq. zero) then
c
c           %-----------------------------------------%
c           | The initial vector is zero. Error exit. |
c           %-----------------------------------------%
c
            info = -9
            go to 1100
         end if
         getv0 = .false.
         ido  = 0
      end if
c
c     %-----------------------------------%
c     | Back from reverse communication : |
c     | continue with update step         |
c     %-----------------------------------%
c
      if (update) go to 20
c
c     %-------------------------------------------%
c     | Back from computing user specified shifts |
c     %-------------------------------------------%
c
      if (ushift) go to 50
c
c     %-------------------------------------%
c     | Back from computing residual norm   |
c     | at the end of the current iteration |
c     %-------------------------------------%
c
      if (cnorm)  go to 100
c
c     %----------------------------------------------------------%
c     | Compute the first NEV steps of the Arnoldi factorization |
c     %----------------------------------------------------------%
c
      call dnaitr (ido, bmat, n, dnazero, nev, mode, resid, rnorm,
     &             v, ldv, h, ldh, ipntr, workd, info)
c
c     %---------------------------------------------------%
c     | ido .ne. 99 implies use of reverse communication  |
c     | to compute operations involving OP and possibly B |
c     %---------------------------------------------------%
c
      if (ido .ne. 99) go to 9000
c
      if (info .gt. 0) then
         np   = info
         mxiter = iter
         info = -9999
         go to 1200
      end if
c
c     %--------------------------------------------------------------%
c     |                                                              |
c     |           M A I N  ARNOLDI  I T E R A T I O N  L O O P       |
c     |           Each iteration implicitly restarts the Arnoldi     |
c     |           factorization in place.                            |
c     |                                                              |
c     %--------------------------------------------------------------%
c
 1000 continue
c
         iter = iter + 1
c
c        %-----------------------------------------------------------%
c        | Compute NP additional steps of the Arnoldi factorization. |
c        | Adjust NP since NEV might have been updated by last call  |
c        | to the shift application routine dnapps.                  |
c        %-----------------------------------------------------------%
c
         np  = kplusp - nev
c
c        %-----------------------------------------------------------%
c        | Compute NP additional steps of the Arnoldi factorization. |
c        %-----------------------------------------------------------%
c
         ido = 0
   20    continue
         update = .true.
c
         call dnaitr (ido  , bmat, n  , nev, np , mode , resid,
     &                rnorm, v   , ldv, h  , ldh, ipntr, workd,
     &                info)
c
c        %---------------------------------------------------%
c        | ido .ne. 99 implies use of reverse communication  |
c        | to compute operations involving OP and possibly B |
c        %---------------------------------------------------%
c
         if (ido .ne. 99) go to 9000
c
         if (info .gt. 0) then
            np = info
            mxiter = iter
            info = -9999
            go to 1200
         end if
         update = .false.
c
c        %--------------------------------------------------------%
c        | Compute the eigenvalues and corresponding error bounds |
c        | of the current upper Hessenberg matrix.                |
c        %--------------------------------------------------------%
c
         call dneigh (rnorm, kplusp, h, ldh, ritzr, ritzi, bounds,
     &                q, ldq, workl, ierr)
c
         if (ierr .ne. 0) then
            info = -8
            go to 1200
         end if
c
c        %----------------------------------------------------%
c        | Make a copy of eigenvalues and corresponding error |
c        | bounds obtained from dneigh.                       |
c        %----------------------------------------------------%
c
         call dcopy(kplusp, ritzr, 1, workl(kplusp**2+1), 1)
         call dcopy(kplusp, ritzi, 1, workl(kplusp**2+kplusp+1), 1)
         call dcopy(kplusp, bounds, 1, workl(kplusp**2+2*kplusp+1), 1)
c
c        %---------------------------------------------------%
c        | Select the wanted Ritz values and their bounds    |
c        | to be used in the convergence test.               |
c        | The wanted part of the spectrum and corresponding |
c        | error bounds are in the last NEV loc. of RITZR,   |
c        | RITZI and BOUNDS respectively. The variables NEV  |
c        | and NP may be updated if the NEV-th wanted Ritz   |
c        | value has a non zero imaginary part. In this case |
c        | NEV is increased by one and NP decreased by one.  |
c        | NOTE: The last two arguments of dngets are no     |
c        | longer used as of version 2.1.                    |
c        %---------------------------------------------------%
c
         nev = nev0
         np = np0
         numcnv = nev
         call dngets (ishift, which, nev, np, ritzr, ritzi,
     &                bounds)
         if (nev .eq. nev0+1) numcnv = nev0+1
c
c        %-------------------%
c        | Convergence test. |
c        %-------------------%
c
         call dcopy (nev, bounds(np+1), 1, workl(2*np+1), 1)
         call dnconv (nev, ritzr(np+1), ritzi(np+1), workl(2*np+1),
     &        tol, nconv)
c
c        %---------------------------------------------------------%
c        | Count the number of unwanted Ritz values that have zero |
c        | Ritz estimates. If any Ritz estimates are equal to zero |
c        | then a leading block of H of order equal to at least    |
c        | the number of Ritz values with zero Ritz estimates has  |
c        | split off. None of these Ritz values may be removed by  |
c        | shifting. Decrease NP the number of shifts to apply. If |
c        | no shifts may be applied, then prepare to exit          |
c        %---------------------------------------------------------%
c
         nptemp = np
         do 30 j=1, nptemp
c            if (bounds(j) .eq. zero) then
            if (abs( bounds(j) ) .le. zero) then
               np = np - 1
               nev = nev + 1
            end if
 30      continue
c
         if ( (nconv .ge. numcnv) .or.
     &        (iter .gt. mxiter) .or.
     &        (np .eq. 0) ) then
c
c
c           %------------------------------------------------%
c           | Prepare to exit. Put the converged Ritz values |
c           | and corresponding bounds in RITZ(1:NCONV) and  |
c           | BOUNDS(1:NCONV) respectively. Then sort. Be    |
c           | careful when NCONV > NP                        |
c           %------------------------------------------------%
c
c           %------------------------------------------%
c           |  Use h( 3,1 ) as storage to communicate  |
c           |  rnorm to _neupd if needed               |
c           %------------------------------------------%

            h(3,1) = rnorm
c
c           %----------------------------------------------%
c           | To be consistent with dngets, we first do a  |
c           | pre-processing sort in order to keep complex |
c           | conjugate pairs together.  This is similar   |
c           | to the pre-processing sort used in dngets    |
c           | except that the sort is done in the opposite |
c           | order.                                       |
c           %----------------------------------------------%
c
            if (which .eq. 'LM') wprime = 'SR'
            if (which .eq. 'SM') wprime = 'LR'
            if (which .eq. 'LR') wprime = 'SM'
            if (which .eq. 'SR') wprime = 'LM'
            if (which .eq. 'LI') wprime = 'SM'
            if (which .eq. 'SI') wprime = 'LM'
c
            call dsortc (wprime, dnatrue, kplusp, ritzr, ritzi, bounds)
c
c           %----------------------------------------------%
c           | Now sort Ritz values so that converged Ritz  |
c           | values appear within the first NEV locations |
c           | of ritzr, ritzi and bounds, and the most     |
c           | desired one appears at the front.            |
c           %----------------------------------------------%
c
            if (which .eq. 'LM') wprime = 'SM'
            if (which .eq. 'SM') wprime = 'LM'
            if (which .eq. 'LR') wprime = 'SR'
            if (which .eq. 'SR') wprime = 'LR'
            if (which .eq. 'LI') wprime = 'SI'
            if (which .eq. 'SI') wprime = 'LI'
c
            call dsortc(wprime, dnatrue, kplusp, ritzr, ritzi, bounds)
c
c           %--------------------------------------------------%
c           | Scale the Ritz estimate of each Ritz value       |
c           | by 1 / max(eps23,magnitude of the Ritz value).   |
c           %--------------------------------------------------%
c
            do 35 j = 1, numcnv
                temp = max(eps23,dlapy2(ritzr(j),
     &                                   ritzi(j)))
                bounds(j) = bounds(j)/temp
 35         continue
c
c           %----------------------------------------------------%
c           | Sort the Ritz values according to the scaled Ritz  |
c           | esitmates.  This will push all the converged ones  |
c           | towards the front of ritzr, ritzi, bounds          |
c           | (in the case when NCONV < NEV.)                    |
c           %----------------------------------------------------%
c
            wprime = 'LR'
            call dsortc(wprime, dnatrue, numcnv, bounds, ritzr, ritzi)
c
c           %----------------------------------------------%
c           | Scale the Ritz estimate back to its original |
c           | value.                                       |
c           %----------------------------------------------%
c
            do 40 j = 1, numcnv
                temp = max(eps23, dlapy2(ritzr(j),
     &                                   ritzi(j)))
                bounds(j) = bounds(j)*temp
 40         continue
c
c           %------------------------------------------------%
c           | Sort the converged Ritz values again so that   |
c           | the "threshold" value appears at the front of  |
c           | ritzr, ritzi and bound.                        |
c           %------------------------------------------------%
c
            call dsortc(which, dnatrue, nconv, ritzr, ritzi, bounds)
c
c
c           %------------------------------------%
c           | Max iterations have been exceeded. |
c           %------------------------------------%
c
            if (iter .gt. mxiter .and. nconv .lt. numcnv) info = 1
c
c           %---------------------%
c           | No shifts to apply. |
c           %---------------------%
c
            if (np .eq. 0 .and. nconv .lt. numcnv) info = 2
c
            np = nconv
            go to 1100
c
         else if ( (nconv .lt. numcnv) .and. (ishift .eq. 1) ) then
c
c           %-------------------------------------------------%
c           | Do not have all the requested eigenvalues yet.  |
c           | To prevent possible stagnation, adjust the size |
c           | of NEV.                                         |
c           %-------------------------------------------------%
c
            nevbef = nev
            nev = nev + min(nconv, np/2)
            if (nev .eq. 1 .and. kplusp .ge. 6) then
               nev = kplusp / 2
            else if (nev .eq. 1 .and. kplusp .gt. 3) then
               nev = 2
            end if
            np = kplusp - nev
c
c           %---------------------------------------%
c           | If the size of NEV was just increased |
c           | resort the eigenvalues.               |
c           %---------------------------------------%
c
            if (nevbef .lt. nev)
     &         call dngets (ishift, which, nev, np, ritzr, ritzi,
     &              bounds)
c
         end if
c
         if (ishift .eq. 0) then
c
c           %-------------------------------------------------------%
c           | User specified shifts: reverse comminucation to       |
c           | compute the shifts. They are returned in the first    |
c           | 2*NP locations of WORKL.                              |
c           %-------------------------------------------------------%
c
            ushift = .true.
            ido = 3
            go to 9000
         end if
c
   50    continue
c
c        %------------------------------------%
c        | Back from reverse communication;   |
c        | User specified shifts are returned |
c        | in WORKL(1:2*NP)                   |
c        %------------------------------------%
c
         ushift = .false.
c
         if ( ishift .eq. 0 ) then
c
c            %----------------------------------%
c            | Move the NP shifts from WORKL to |
c            | RITZR, RITZI to free up WORKL    |
c            | for non-exact shift case.        |
c            %----------------------------------%
c
             call dcopy (np, workl,       1, ritzr(1), 1)
             call dcopy (np, workl(np+1), 1, ritzi(1), 1)
         end if
c
c
c        %---------------------------------------------------------%
c        | Apply the NP implicit shifts by QR bulge chasing.       |
c        | Each shift is applied to the whole upper Hessenberg     |
c        | matrix H.                                               |
c        | The first 2*N locations of WORKD are used as workspace. |
c        %---------------------------------------------------------%
c
         call dnapps (n, nev, np, ritzr, ritzi, v, ldv,
     &                h, ldh, resid, q, ldq, workl, workd)
c
c        %---------------------------------------------%
c        | Compute the B-norm of the updated residual. |
c        | Keep B*RESID in WORKD(1:N) to be used in    |
c        | the first step of the next call to dnaitr.  |
c        %---------------------------------------------%
c
         cnorm = .true.
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(n+1), 1)
            ipntr(1) = n + 1
            ipntr(2) = 1
            ido = 2
c
c           %----------------------------------%
c           | Exit in order to compute B*RESID |
c           %----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(1), 1)
         end if
c
  100    continue
c
c        %----------------------------------%
c        | Back from reverse communication; |
c        | WORKD(1:N) := B*RESID            |
c        %----------------------------------%
c
         if (bmat .eq. 'G') then
            rnorm = ddot (n, resid, 1, workd, 1)
            rnorm = sqrt(abs(rnorm))
         else if (bmat .eq. 'I') then
            rnorm = dnrm2(n, resid, 1)
         end if
         cnorm = .false.
c
      go to 1000
c
c     %---------------------------------------------------------------%
c     |                                                               |
c     |  E N D     O F     M A I N     I T E R A T I O N     L O O P  |
c     |                                                               |
c     %---------------------------------------------------------------%
c
 1100 continue
c
      mxiter = iter
      nev = numcnv
c
 1200 continue
      ido = 99
c
c     %------------%
c     | Error Exit |
c     %------------%
c
 9000 continue
c
c     %---------------%
c     | End of dnaup2 |
c     %---------------%
c
      return
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dnapps
c
c\Description:
c  Given the Arnoldi factorization
c
c     A*V_{k} - V_{k}*H_{k} = r_{k+p}*e_{k+p}^T,
c
c  apply NP implicit shifts resulting in
c
c     A*(V_{k}*Q) - (V_{k}*Q)*(Q^T* H_{k}*Q) = r_{k+p}*e_{k+p}^T * Q
c
c  where Q is an orthogonal matrix which is the product of rotations
c  and reflections resulting from the NP bulge chage sweeps.
c  The updated Arnoldi factorization becomes:
c
c     A*VNEW_{k} - VNEW_{k}*HNEW_{k} = rnew_{k}*e_{k}^T.
c
c\Usage:
c  call dnapps
c     ( N, KEV, NP, SHIFTR, SHIFTI, V, LDV, H, LDH, RESID, Q, LDQ,
c       WORKL, WORKD )
c
c\Arguments
c  N       Integer.  (INPUT)
c          Problem size, i.e. size of matrix A.
c
c  KEV     Integer.  (INPUT/OUTPUT)
c          KEV+NP is the size of the input matrix H.
c          KEV is the size of the updated matrix HNEW.  KEV is only
c          updated on ouput when fewer than NP shifts are applied in
c          order to keep the conjugate pair together.
c
c  NP      Integer.  (INPUT)
c          Number of implicit shifts to be applied.
c
c  SHIFTR, Double precision array of length NP.  (INPUT)
c  SHIFTI  Real and imaginary part of the shifts to be applied.
c          Upon, entry to dnapps, the shifts must be sorted so that the
c          conjugate pairs are in consecutive locations.
c
c  V       Double precision N by (KEV+NP) array.  (INPUT/OUTPUT)
c          On INPUT, V contains the current KEV+NP Arnoldi vectors.
c          On OUTPUT, V contains the updated KEV Arnoldi vectors
c          in the first KEV columns of V.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (KEV+NP) by (KEV+NP) array.  (INPUT/OUTPUT)
c          On INPUT, H contains the current KEV+NP by KEV+NP upper
c          Hessenber matrix of the Arnoldi factorization.
c          On OUTPUT, H contains the updated KEV by KEV upper Hessenberg
c          matrix in the KEV leading submatrix.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  RESID   Double precision array of length N.  (INPUT/OUTPUT)
c          On INPUT, RESID contains the the residual vector r_{k+p}.
c          On OUTPUT, RESID is the update residual vector rnew_{k}
c          in the first KEV locations.
c
c  Q       Double precision KEV+NP by KEV+NP work array.  (WORKSPACE)
c          Work array used to accumulate the rotations and reflections
c          during the bulge chase sweep.
c
c  LDQ     Integer.  (INPUT)
c          Leading dimension of Q exactly as declared in the calling
c          program.
c
c  WORKL   Double precision work array of length (KEV+NP).  (WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.
c
c  WORKD   Double precision work array of length 2*N.  (WORKSPACE)
c          Distributed array used in the application of the accumulated
c          orthogonal matrix Q.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c
c\Routines called:
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dmout   ARPACK utility routine that prints matrices.
c     dvout   ARPACK utility routine that prints vectors.
c     dlabad  LAPACK routine that computes machine constants.
c     dlacpy  LAPACK matrix copy routine.
c     dlamch  LAPACK routine that determines machine constants.
c     dlanhs  LAPACK routine that computes various norms of a matrix.
c     dlapy2  LAPACK routine to compute sqrt(x**2+y**2) carefully.
c     dlarf   LAPACK routine that applies Householder reflection to
c             a matrix.
c     dlarfg  LAPACK Householder reflection construction routine.
c     dlartg  LAPACK Givens rotation construction routine.
c     dlaset  LAPACK matrix initialization routine.
c     dgemv   Level 2 BLAS routine for matrix vector multiplication.
c     daxpy   Level 1 BLAS that computes a vector triad.
c     dcopy   Level 1 BLAS that copies one vector to another .
c     dscal   Level 1 BLAS that scales a vector.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: napps.F   SID: 2.4   DATE OF SID: 3/28/97   RELEASE: 2
c
c\Remarks
c  1. In this version, each shift is applied to all the sublocks of
c     the Hessenberg matrix H and not just to the submatrix that it
c     comes from. Deflation as in LAPACK routine dlahqr (QR algorithm
c     for upper Hessenberg matrices ) is used.
c     The subdiagonals of H are enforced to be non-negative.
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dnapps
     &   ( n, kev, np, shiftr, shifti, v, ldv, h, ldh, resid, q, ldq,
     &     workl, workd )

      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    kev, ldh, ldq, ldv, n, np
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           h(ldh,kev+np), resid(n), shifti(np), shiftr(np),
     &           v(ldv,kev+np), q(ldq,kev+np), workd(2*n), workl(kev+np)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0, zero = 0.0D+0)
      integer(4)  ione
      parameter (ione = 1)
c
c     %------------------------%
c     | Local Scalars & Arrays |
c     %------------------------%
c
      integer    i, iend, ir, istart, j, jj, kplusp, nr
      logical    cconj, first
      Double precision
     &           c, f, g, h11, h12, h21, h22, h32, ovfl, r, s, sigmai,
     &           sigmar, smlnum, ulp, unfl, u(3), t, tau, tst1
      save       first, ovfl, smlnum, ulp, unfl
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   daxpy, dcopy, dscal, dlacpy, dlarfg, dlarf,
     &           dlaset, dlabad, dlartg
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
c
      Double precision
     &           dlamch, dlanhs, dlapy2
      external   dlamch, dlanhs, dlapy2
c
c     %----------------------%
c     | Intrinsics Functions |
c     %----------------------%
c
      intrinsic  abs, max, min
c
c     %----------------%
c     | Data statments |
c     %----------------%
c
      data       first / .true. /
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (first) then
c
c        %-----------------------------------------------%
c        | Set machine-dependent constants for the       |
c        | stopping criterion. If norm(H) <= sqrt(OVFL), |
c        | overflow should not occur.                    |
c        | REFERENCE: LAPACK subroutine dlahqr           |
c        %-----------------------------------------------%
c
         unfl = dlamch( 'safe minimum' )
         ovfl = one / unfl
         call dlabad( unfl, ovfl )
         ulp = dlamch( 'precision' )
         smlnum = unfl*( n / ulp )
         first = .false.
      end if
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
c      call second (t0)
      kplusp = kev + np
c
c     %--------------------------------------------%
c     | Initialize Q to the identity to accumulate |
c     | the rotations and reflections              |
c     %--------------------------------------------%
c
      call dlaset ('All', kplusp, kplusp, zero, one, q, ldq)
c
c     %----------------------------------------------%
c     | Quick return if there are no shifts to apply |
c     %----------------------------------------------%
c
      if (np .eq. 0) go to 9000
c
c     %----------------------------------------------%
c     | Chase the bulge with the application of each |
c     | implicit shift. Each shift is applied to the |
c     | whole matrix including each block.           |
c     %----------------------------------------------%
c
      cconj = .false.
      do 110 jj = 1, np
         sigmar = shiftr(jj)
         sigmai = shifti(jj)
c
c        %-------------------------------------------------%
c        | The following set of conditionals is necessary  |
c        | in order that complex conjugate pairs of shifts |
c        | are applied together or not at all.             |
c        %-------------------------------------------------%
c
         if ( cconj ) then
c
c           %-----------------------------------------%
c           | cconj = .true. means the previous shift |
c           | had non-zero imaginary part.            |
c           %-----------------------------------------%
c
            cconj = .false.
            go to 110
         else if ( jj .lt. np .and. abs( sigmai ) .gt. zero ) then
c
c           %------------------------------------%
c           | Start of a complex conjugate pair. |
c           %------------------------------------%
c
            cconj = .true.
         else if ( jj .eq. np .and. abs( sigmai ) .gt. zero ) then
c
c           %----------------------------------------------%
c           | The last shift has a nonzero imaginary part. |
c           | Don't apply it; thus the order of the        |
c           | compressed H is order KEV+1 since only np-1  |
c           | were applied.                                |
c           %----------------------------------------------%
c
            kev = kev + 1
            go to 110
         end if
         istart = 1
   20    continue
c
c        %--------------------------------------------------%
c        | if sigmai = 0 then                               |
c        |    Apply the jj-th shift ...                     |
c        | else                                             |
c        |    Apply the jj-th and (jj+1)-th together ...    |
c        |    (Note that jj < np at this point in the code) |
c        | end                                              |
c        | to the current block of H. The next do loop      |
c        | determines the current block ;                   |
c        %--------------------------------------------------%
c
         do 30 i = istart, kplusp-1
c
c           %----------------------------------------%
c           | Check for splitting and deflation. Use |
c           | a standard test as in the QR algorithm |
c           | REFERENCE: LAPACK subroutine dlahqr    |
c           %----------------------------------------%
c
            tst1 = abs( h( i, i ) ) + abs( h( i+1, i+1 ) )
c            if( tst1.eq.zero )
            if( abs(tst1) .le. zero )
     &         tst1 = dlanhs( '1', kplusp-jj+1, h, ldh, workl )
            if( abs( h( i+1,i ) ).le.max( ulp*tst1, smlnum ) ) then
               iend = i
               h(i+1,i) = zero
               go to 40
            end if
   30    continue
         iend = kplusp
   40    continue
c
c        %------------------------------------------------%
c        | No reason to apply a shift to block of order 1 |
c        %------------------------------------------------%
c
         if ( istart .eq. iend ) go to 100
c
c        %------------------------------------------------------%
c        | If istart + 1 = iend then no reason to apply a       |
c        | complex conjugate pair of shifts on a 2 by 2 matrix. |
c        %------------------------------------------------------%
c
         if ( istart + 1 .eq. iend .and. abs( sigmai ) .gt. zero )
     &      go to 100
c
         h11 = h(istart,istart)
         h21 = h(istart+1,istart)
         if ( abs( sigmai ) .le. zero ) then
c
c           %---------------------------------------------%
c           | Real-valued shift ==> apply single shift QR |
c           %---------------------------------------------%
c
            f = h11 - sigmar
            g = h21
c
            do 80 i = istart, iend-1
c
c              %-----------------------------------------------------%
c              | Contruct the plane rotation G to zero out the bulge |
c              %-----------------------------------------------------%
c
               call dlartg (f, g, c, s, r)
               if (i .gt. istart) then
c
c                 %-------------------------------------------%
c                 | The following ensures that h(1:iend-1,1), |
c                 | the first iend-2 off diagonal of elements |
c                 | H, remain non negative.                   |
c                 %-------------------------------------------%
c
                  if (r .lt. zero) then
                     r = -r
                     c = -c
                     s = -s
                  end if
                  h(i,i-1) = r
                  h(i+1,i-1) = zero
               end if
c
c              %---------------------------------------------%
c              | Apply rotation to the left of H;  H <- G'*H |
c              %---------------------------------------------%
c
               do 50 j = i, kplusp
                  t        =  c*h(i,j) + s*h(i+1,j)
                  h(i+1,j) = -s*h(i,j) + c*h(i+1,j)
                  h(i,j)   = t
   50          continue
c
c              %---------------------------------------------%
c              | Apply rotation to the right of H;  H <- H*G |
c              %---------------------------------------------%
c
               do 60 j = 1, min(i+2,iend)
                  t        =  c*h(j,i) + s*h(j,i+1)
                  h(j,i+1) = -s*h(j,i) + c*h(j,i+1)
                  h(j,i)   = t
   60          continue
c
c              %----------------------------------------------------%
c              | Accumulate the rotation in the matrix Q;  Q <- Q*G |
c              %----------------------------------------------------%
c
               do 70 j = 1, min( i+jj, kplusp )
                  t        =   c*q(j,i) + s*q(j,i+1)
                  q(j,i+1) = - s*q(j,i) + c*q(j,i+1)
                  q(j,i)   = t
   70          continue
c
c              %---------------------------%
c              | Prepare for next rotation |
c              %---------------------------%
c
               if (i .lt. iend-1) then
                  f = h(i+1,i)
                  g = h(i+2,i)
               end if
   80       continue
c
c           %-----------------------------------%
c           | Finished applying the real shift. |
c           %-----------------------------------%
c
         else
c
c           %----------------------------------------------------%
c           | Complex conjugate shifts ==> apply double shift QR |
c           %----------------------------------------------------%
c
            h12 = h(istart,istart+1)
            h22 = h(istart+1,istart+1)
            h32 = h(istart+2,istart+1)
c
c           %---------------------------------------------------------%
c           | Compute 1st column of (H - shift*I)*(H - conj(shift)*I) |
c           %---------------------------------------------------------%
c
            s    = 2.0*sigmar
            t = dlapy2 ( sigmar, sigmai )
            u(1) = ( h11 * (h11 - s) + t * t ) / h21 + h12
            u(2) = h11 + h22 - s
            u(3) = h32
c
            do 90 i = istart, iend-1
c
               nr = min ( 3, iend-i+1 )
c
c              %-----------------------------------------------------%
c              | Construct Householder reflector G to zero out u(1). |
c              | G is of the form I - tau*( 1 u )' * ( 1 u' ).       |
c              %-----------------------------------------------------%
c
               call dlarfg ( nr, u(1), u(2), 1, tau )
c
               if (i .gt. istart) then
                  h(i,i-1)   = u(1)
                  h(i+1,i-1) = zero
                  if (i .lt. iend-1) h(i+2,i-1) = zero
               end if
               u(1) = one
c
c              %--------------------------------------%
c              | Apply the reflector to the left of H |
c              %--------------------------------------%
c
               call dlarf ('Left', nr, kplusp-i+1, u, 1, tau,
     &                     h(i,i), ldh, workl)
c
c              %---------------------------------------%
c              | Apply the reflector to the right of H |
c              %---------------------------------------%
c
               ir = min ( i+3, iend )
               call dlarf ('Right', ir, nr, u, 1, tau,
     &                     h(1,i), ldh, workl)
c
c              %-----------------------------------------------------%
c              | Accumulate the reflector in the matrix Q;  Q <- Q*G |
c              %-----------------------------------------------------%
c
               call dlarf ('Right', kplusp, nr, u, 1, tau,
     &                     q(1,i), ldq, workl)
c
c              %----------------------------%
c              | Prepare for next reflector |
c              %----------------------------%
c
               if (i .lt. iend-1) then
                  u(1) = h(i+1,i)
                  u(2) = h(i+2,i)
                  if (i .lt. iend-2) u(3) = h(i+3,i)
               end if
c
   90       continue
c
c           %--------------------------------------------%
c           | Finished applying a complex pair of shifts |
c           | to the current block                       |
c           %--------------------------------------------%
c
         end if
c
  100    continue
c
c        %---------------------------------------------------------%
c        | Apply the same shift to the next block if there is any. |
c        %---------------------------------------------------------%
c
         istart = iend + 1
         if (iend .lt. kplusp) go to 20
c
c        %---------------------------------------------%
c        | Loop back to the top to get the next shift. |
c        %---------------------------------------------%
c
  110 continue
c
c     %--------------------------------------------------%
c     | Perform a similarity transformation that makes   |
c     | sure that H will have non negative sub diagonals |
c     %--------------------------------------------------%
c
      do 120 j=1,kev
         if ( h(j+1,j) .lt. zero ) then
            call dscal( kplusp-j+1, -one, h(j+1,j), ldh)
            call dscal( min(j+2, kplusp), -one, h(1,j+1),
     &                  ione)
            call dscal( min(j+np+1,kplusp), -one, q(1,j+1),
     &                  ione)
cv              call dscal( min(j+2, kplusp), -one, h(1,j+1), 1 )
cv              call dscal( min(j+np+1,kplusp), -one, q(1,j+1), 1 )
         end if
 120  continue
c
      do 130 i = 1, kev
c
c        %--------------------------------------------%
c        | Final check for splitting and deflation.   |
c        | Use a standard test as in the QR algorithm |
c        | REFERENCE: LAPACK subroutine dlahqr        |
c        %--------------------------------------------%
c
         tst1 = abs( h( i, i ) ) + abs( h( i+1, i+1 ) )
c         if( tst1.eq.zero )
         if( abs(tst1) .le. zero)
     &       tst1 = dlanhs( '1', kev, h, ldh, workl )
         if( h( i+1,i ) .le. max( ulp*tst1, smlnum ) )
     &       h(i+1,i) = zero
 130  continue
c
c     %-------------------------------------------------%
c     | Compute the (kev+1)-st column of (V*Q) and      |
c     | temporarily store the result in WORKD(N+1:2*N). |
c     | This is needed in the residual update since we  |
c     | cannot GUARANTEE that the corresponding entry   |
c     | of H would be zero as in exact arithmetic.      |
c     %-------------------------------------------------%
c
      if (h(kev+1,kev) .gt. zero)
     &    call dgemv ('N', n, kplusp, one, v, ldv, q(1,kev+1), 1, zero,
     &                workd(n+1), 1)
c
c     %----------------------------------------------------------%
c     | Compute column 1 to kev of (V*Q) in backward order       |
c     | taking advantage of the upper Hessenberg structure of Q. |
c     %----------------------------------------------------------%
c
      do 140 i = 1, kev
         call dgemv ('N', n, kplusp-i+1, one, v, ldv,
     &        q(1,kev-i+1), 1, zero, workd(1), 1)
         call dcopy (n, workd, 1, v(1,kplusp-i+1), 1)
  140 continue
c
c     %-------------------------------------------------%
c     |  Move v(:,kplusp-kev+1:kplusp) into v(:,1:kev). |
c     %-------------------------------------------------%
c
      call dlacpy ('A', n, kev, v(1,kplusp-kev+1), ldv, v, ldv)
c
c     %--------------------------------------------------------------%
c     | Copy the (kev+1)-st column of (V*Q) in the appropriate place |
c     %--------------------------------------------------------------%
c
      if (h(kev+1,kev) .gt. zero)
     &   call dcopy (n, workd(n+1), 1, v(1,kev+1), 1)
c
c     %-------------------------------------%
c     | Update the residual vector:         |
c     |    r <- sigmak*r + betak*v(:,kev+1) |
c     | where                               |
c     |    sigmak = (e_{kplusp}'*Q)*e_{kev} |
c     |    betak = e_{kev+1}'*H*e_{kev}     |
c     %-------------------------------------%
c
      call dscal (n, q(kplusp,kev), resid(1), ione)
      if (h(kev+1,kev) .gt. zero)
     &   call daxpy (n, h(kev+1,kev), v(1,kev+1), 1, resid, 1)
c
 9000 continue
c
      return
c
c     %---------------%
c     | End of dnapps |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dnconv
c
c\Description:
c  Convergence testing for the nonsymmetric Arnoldi eigenvalue routine.
c
c\Usage:
c  call dnconv
c     ( N, RITZR, RITZI, BOUNDS, TOL, NCONV )
c
c\Arguments
c  N       Integer.  (INPUT)
c          Number of Ritz values to check for convergence.
c
c  RITZR,  Double precision arrays of length N.  (INPUT)
c  RITZI   Real and imaginary parts of the Ritz values to be checked
c          for convergence.

c  BOUNDS  Double precision array of length N.  (INPUT)
c          Ritz estimates for the Ritz values in RITZR and RITZI.
c
c  TOL     Double precision scalar.  (INPUT)
c          Desired backward error for a Ritz value to be considered
c          "converged".
c
c  NCONV   Integer scalar.  (OUTPUT)
c          Number of "converged" Ritz values.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     second  ARPACK utility routine for timing.
c     dlamch  LAPACK routine that determines machine constants.
c     dlapy2  LAPACK routine to compute sqrt(x**2+y**2) carefully.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.1'
c
c\SCCS Information: @(#)
c FILE: nconv.F   SID: 2.3   DATE OF SID: 4/20/96   RELEASE: 2
c
c\Remarks
c     1. xxxx
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dnconv (n, ritzr, ritzi, bounds, tol, nconv)

      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    n, nconv
      Double precision
     &           tol
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%

      Double precision
     &           ritzr(n), ritzi(n), bounds(n)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i
      Double precision
     &           temp, eps23
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlapy2, dlamch
      external   dlapy2, dlamch

c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %-------------------------------------------------------------%
c     | Convergence test: unlike in the symmetric code, I am not    |
c     | using things like refined error bounds and gap condition    |
c     | because I don't know the exact equivalent concept.          |
c     |                                                             |
c     | Instead the i-th Ritz value is considered "converged" when: |
c     |                                                             |
c     |     bounds(i) .le. ( TOL * | ritz | )                       |
c     |                                                             |
c     | for some appropriate choice of norm.                        |
c     %-------------------------------------------------------------%
c
c     %---------------------------------%
c     | Get machine dependent constant. |
c     %---------------------------------%
c
      eps23 = dlamch('Epsilon-Machine')
      eps23 = eps23**(2.0D+0 / 3.0D+0)
c
      nconv  = 0
      do 20 i = 1, n
         temp = max( eps23, dlapy2( ritzr(i), ritzi(i) ) )
         if (bounds(i) .le. tol*temp)   nconv = nconv + 1
   20 continue
c
      return
c
c     %---------------%
c     | End of dnconv |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dsortc
c
c\Description:
c  Sorts the complex array in XREAL and XIMAG into the order
c  specified by WHICH and optionally applies the permutation to the
c  real array Y. It is assumed that if an element of XIMAG is
c  nonzero, then its negative is also an element. In other words,
c  both members of a complex conjugate pair are to be sorted and the
c  pairs are kept adjacent to each other.
c
c\Usage:
c  call dsortc
c     ( WHICH, APPLY, N, XREAL, XIMAG, Y )
c
c\Arguments
c  WHICH   Character*2.  (Input)
c          'LM' -> sort XREAL,XIMAG into increasing order of magnitude.
c          'SM' -> sort XREAL,XIMAG into decreasing order of magnitude.
c          'LR' -> sort XREAL into increasing order of algebraic.
c          'SR' -> sort XREAL into decreasing order of algebraic.
c          'LI' -> sort XIMAG into increasing order of magnitude.
c          'SI' -> sort XIMAG into decreasing order of magnitude.
c          NOTE: If an element of XIMAG is non-zero, then its negative
c                is also an element.
c
c  APPLY   Logical.  (Input)
c          APPLY = .TRUE.  -> apply the sorted order to array Y.
c          APPLY = .FALSE. -> do not apply the sorted order to array Y.
c
c  N       Integer.  (INPUT)
c          Size of the arrays.
c
c  XREAL,  Double precision array of length N.  (INPUT/OUTPUT)
c  XIMAG   Real and imaginary part of the array to be sorted.
c
c  Y       Double precision array of length N.  (INPUT/OUTPUT)
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.1'
c               Adapted from the sort routine in LANSO.
c
c\SCCS Information: @(#)
c FILE: sortc.F   SID: 2.3   DATE OF SID: 4/20/96   RELEASE: 2
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dsortc (which, apply, n, xreal, ximag, y)
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  which*2
      logical    apply
      integer    n
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           xreal(0:n-1), ximag(0:n-1), y(0:n-1)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      integer    i, igap, j
      Double precision
     &           temp, temp1, temp2
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlapy2
      external   dlapy2
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      igap = n / 2
c
      if (which .eq. 'LM') then
c
c        %------------------------------------------------------%
c        | Sort XREAL,XIMAG into increasing order of magnitude. |
c        %------------------------------------------------------%
c
   10    continue
         if (igap .eq. 0) go to 9000
c
         do 30 i = igap, n-1
            j = i-igap
   20       continue
c
            if (j.lt.0) go to 30
c
            temp1 = dlapy2(xreal(j),ximag(j))
            temp2 = dlapy2(xreal(j+igap),ximag(j+igap))
c
            if (temp1.gt.temp2) then
                temp = xreal(j)
                xreal(j) = xreal(j+igap)
                xreal(j+igap) = temp
c
                temp = ximag(j)
                ximag(j) = ximag(j+igap)
                ximag(j+igap) = temp
c
                if (apply) then
                    temp = y(j)
                    y(j) = y(j+igap)
                    y(j+igap) = temp
                end if
            else
                go to 30
            end if
            j = j-igap
            go to 20
   30    continue
         igap = igap / 2
         go to 10
c
      else if (which .eq. 'SM') then
c
c        %------------------------------------------------------%
c        | Sort XREAL,XIMAG into decreasing order of magnitude. |
c        %------------------------------------------------------%
c
   40    continue
         if (igap .eq. 0) go to 9000
c
         do 60 i = igap, n-1
            j = i-igap
   50       continue
c
            if (j .lt. 0) go to 60
c
            temp1 = dlapy2(xreal(j),ximag(j))
            temp2 = dlapy2(xreal(j+igap),ximag(j+igap))
c
            if (temp1.lt.temp2) then
               temp = xreal(j)
               xreal(j) = xreal(j+igap)
               xreal(j+igap) = temp
c
               temp = ximag(j)
               ximag(j) = ximag(j+igap)
               ximag(j+igap) = temp
c
               if (apply) then
                  temp = y(j)
                  y(j) = y(j+igap)
                  y(j+igap) = temp
               end if
            else
               go to 60
            endif
            j = j-igap
            go to 50
   60    continue
         igap = igap / 2
         go to 40
c
      else if (which .eq. 'LR') then
c
c        %------------------------------------------------%
c        | Sort XREAL into increasing order of algebraic. |
c        %------------------------------------------------%
c
   70    continue
         if (igap .eq. 0) go to 9000
c
         do 90 i = igap, n-1
            j = i-igap
   80       continue
c
            if (j.lt.0) go to 90
c
            if (xreal(j).gt.xreal(j+igap)) then
               temp = xreal(j)
               xreal(j) = xreal(j+igap)
               xreal(j+igap) = temp
c
               temp = ximag(j)
               ximag(j) = ximag(j+igap)
               ximag(j+igap) = temp
c
               if (apply) then
                  temp = y(j)
                  y(j) = y(j+igap)
                  y(j+igap) = temp
               end if
            else
               go to 90
            endif
            j = j-igap
            go to 80
   90    continue
         igap = igap / 2
         go to 70
c
      else if (which .eq. 'SR') then
c
c        %------------------------------------------------%
c        | Sort XREAL into decreasing order of algebraic. |
c        %------------------------------------------------%
c
  100    continue
         if (igap .eq. 0) go to 9000
         do 120 i = igap, n-1
            j = i-igap
  110       continue
c
            if (j.lt.0) go to 120
c
            if (xreal(j).lt.xreal(j+igap)) then
               temp = xreal(j)
               xreal(j) = xreal(j+igap)
               xreal(j+igap) = temp
c
               temp = ximag(j)
               ximag(j) = ximag(j+igap)
               ximag(j+igap) = temp
c
               if (apply) then
                  temp = y(j)
                  y(j) = y(j+igap)
                  y(j+igap) = temp
               end if
            else
               go to 120
            endif
            j = j-igap
            go to 110
  120    continue
         igap = igap / 2
         go to 100
c
      else if (which .eq. 'LI') then
c
c        %------------------------------------------------%
c        | Sort XIMAG into increasing order of magnitude. |
c        %------------------------------------------------%
c
  130    continue
         if (igap .eq. 0) go to 9000
         do 150 i = igap, n-1
            j = i-igap
  140       continue
c
            if (j.lt.0) go to 150
c
            if (abs(ximag(j)).gt.abs(ximag(j+igap))) then
               temp = xreal(j)
               xreal(j) = xreal(j+igap)
               xreal(j+igap) = temp
c
               temp = ximag(j)
               ximag(j) = ximag(j+igap)
               ximag(j+igap) = temp
c
               if (apply) then
                  temp = y(j)
                  y(j) = y(j+igap)
                  y(j+igap) = temp
               end if
            else
               go to 150
            endif
            j = j-igap
            go to 140
  150    continue
         igap = igap / 2
         go to 130
c
      else if (which .eq. 'SI') then
c
c        %------------------------------------------------%
c        | Sort XIMAG into decreasing order of magnitude. |
c        %------------------------------------------------%
c
  160    continue
         if (igap .eq. 0) go to 9000
         do 180 i = igap, n-1
            j = i-igap
  170       continue
c
            if (j.lt.0) go to 180
c
            if (abs(ximag(j)).lt.abs(ximag(j+igap))) then
               temp = xreal(j)
               xreal(j) = xreal(j+igap)
               xreal(j+igap) = temp
c
               temp = ximag(j)
               ximag(j) = ximag(j+igap)
               ximag(j+igap) = temp
c
               if (apply) then
                  temp = y(j)
                  y(j) = y(j+igap)
                  y(j+igap) = temp
               end if
            else
               go to 180
            endif
            j = j-igap
            go to 170
  180    continue
         igap = igap / 2
         go to 160
      end if
c
 9000 continue
      return
c
c     %---------------%
c     | End of dsortc |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dneigh
c
c\Description:
c  Compute the eigenvalues of the current upper Hessenberg matrix
c  and the corresponding Ritz estimates given the current residual norm.
c
c\Usage:
c  call dneigh
c     ( RNORM, N, H, LDH, RITZR, RITZI, BOUNDS, Q, LDQ, WORKL, IERR )
c
c\Arguments
c  RNORM   Double precision scalar.  (INPUT)
c          Residual norm corresponding to the current upper Hessenberg
c          matrix H.
c
c  N       Integer.  (INPUT)
c          Size of the matrix H.
c
c  H       Double precision N by N array.  (INPUT)
c          H contains the current upper Hessenberg matrix.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  RITZR,  Double precision arrays of length N.  (OUTPUT)
c  RITZI   On output, RITZR(1:N) (resp. RITZI(1:N)) contains the real
c          (respectively imaginary) parts of the eigenvalues of H.
c
c  BOUNDS  Double precision array of length N.  (OUTPUT)
c          On output, BOUNDS contains the Ritz estimates associated with
c          the eigenvalues RITZR and RITZI.  This is equal to RNORM
c          times the last components of the eigenvectors corresponding
c          to the eigenvalues in RITZR and RITZI.
c
c  Q       Double precision N by N array.  (WORKSPACE)
c          Workspace needed to store the eigenvectors of H.
c
c  LDQ     Integer.  (INPUT)
c          Leading dimension of Q exactly as declared in the calling
c          program.
c
c  WORKL   Double precision work array of length N**2 + 3*N.  (WORKSPACE)
c          Private (replicated) array on each PE or array allocated on
c          the front end.  This is needed to keep the full Schur form
c          of H and also in the calculation of the eigenvectors of H.
c
c  IERR    Integer.  (OUTPUT)
c          Error exit flag from dlaqrb or dtrevc.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dlaqrb  ARPACK routine to compute the real Schur form of an
c             upper Hessenberg matrix and last row of the Schur vectors.
c     second  ARPACK utility routine for timing.
c     dmout   ARPACK utility routine that prints matrices
c     dvout   ARPACK utility routine that prints vectors.
c     dlacpy  LAPACK matrix copy routine.
c     dlapy2  LAPACK routine to compute sqrt(x**2+y**2) carefully.
c     dtrevc  LAPACK routine to compute the eigenvectors of a matrix
c             in upper quasi-triangular form
c     dgemv   Level 2 BLAS routine for matrix vector multiplication.
c     dcopy   Level 1 BLAS that copies one vector to another .
c     dnrm2   Level 1 BLAS that computes the norm of a vector.
c     dscal   Level 1 BLAS that scales a vector.
c
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.1'
c
c\SCCS Information: @(#)
c FILE: neigh.F   SID: 2.3   DATE OF SID: 4/20/96   RELEASE: 2
c
c\Remarks
c     None
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dneigh (rnorm, n, h, ldh, ritzr, ritzi, bounds,
     &     q, ldq, workl, ierr)

      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      integer    ierr, n, ldh, ldq
      Double precision
     &           rnorm
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           bounds(n), h(ldh,n), q(ldq,n), ritzi(n), ritzr(n),
     &           workl(n*(n+3))
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0, zero = 0.0D+0)
c
      integer(4)
     &           ione
      parameter (ione = 1)

c
c     %------------------------%
c     | Local Scalars & Arrays |
c     %------------------------%
c
c     % dneightrue = .true. to adress logical(4)/(8)
c     % initialization for spam/spam64
c     % analogue for dneighone
      logical    select(1), dneightrue
      integer    i, iconj,  dneighone
      Double precision
     &           temp, vl(1)
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy, dlacpy, dlaqrb, dtrevc
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           dlapy2, dnrm2
      external   dlapy2, dnrm2
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic  abs
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
      dneightrue = .true.
      dneighone = 1
c
c
c     %-----------------------------------------------------------%
c     | 1. Compute the eigenvalues, the last components of the    |
c     |    corresponding Schur vectors and the full Schur form T  |
c     |    of the current upper Hessenberg matrix H.              |
c     | dlaqrb returns the full Schur form of H in WORKL(1:N**2)  |
c     | and the last components of the Schur vectors in BOUNDS.   |
c     %-----------------------------------------------------------%
c
cx      call dlacpy ('All', n, n, h, ldh, workl, n)
      call dlacpy ('All', n, n, h(1,1), ldh, workl, n)
      call dlaqrb (dneightrue, n, dneighone, n, workl, n, ritzr, ritzi,
     &             bounds, ierr)
      if (ierr .ne. 0) go to 9000
c
c
c     %-----------------------------------------------------------%
c     | 2. Compute the eigenvectors of the full Schur form T and  |
c     |    apply the last components of the Schur vectors to get  |
c     |    the last components of the corresponding eigenvectors. |
c     | Remember that if the i-th and (i+1)-st eigenvalues are    |
c     | complex conjugate pairs, then the real & imaginary part   |
c     | of the eigenvector components are split across adjacent   |
c     | columns of Q.                                             |
c     %-----------------------------------------------------------%
c
      call dtrevc ('R', 'A', select, n, workl, n, vl, n, q, ldq,
     &             n, n, workl(n*n+1), ierr)
c
      if (ierr .ne. 0) go to 9000
c
c     %------------------------------------------------%
c     | Scale the returning eigenvectors so that their |
c     | euclidean norms are all one. LAPACK subroutine |
c     | dtrevc returns each eigenvector normalized so  |
c     | that the element of largest magnitude has      |
c     | magnitude 1; here the magnitude of a complex   |
c     | number (x,y) is taken to be |x| + |y|.         |
c     %------------------------------------------------%
c
      iconj = 0
      do 10 i=1, n
         if ( abs( ritzi(i) ) .le. zero ) then
c
c           %----------------------%
c           | Real eigenvalue case |
c           %----------------------%
c
            temp = dnrm2( n, q(1,i), 1 )
            call dscal ( n, one / temp, q(1,i), ione )
         else
c
c           %-------------------------------------------%
c           | Complex conjugate pair case. Note that    |
c           | since the real and imaginary part of      |
c           | the eigenvector are stored in consecutive |
c           | columns, we further normalize by the      |
c           | square root of two.                       |
c           %-------------------------------------------%
c
            if (iconj .eq. 0) then
               temp = dlapy2( dnrm2( n, q(1,i), 1 ),
     &                        dnrm2( n, q(1,i+1), 1 ) )
               call dscal ( n, one / temp, q(1,i),
     &                      ione )
               call dscal ( n, one / temp, q(1,i+1),
     &                      ione )
               iconj = 1
            else
               iconj = 0
            end if
         end if
   10 continue
c
      call dgemv('T', n, n, one, q, ldq, bounds(1), 1, zero,
     &         workl(1), 1)
c
c
c     %----------------------------%
c     | Compute the Ritz estimates |
c     %----------------------------%
c
      iconj = 0
      do 20 i = 1, n
         if ( abs( ritzi(i) ) .le. zero ) then
c
c           %----------------------%
c           | Real eigenvalue case |
c           %----------------------%
c
            bounds(i) = rnorm * abs( workl(i) )
         else
c
c           %-------------------------------------------%
c           | Complex conjugate pair case. Note that    |
c           | since the real and imaginary part of      |
c           | the eigenvector are stored in consecutive |
c           | columns, we need to take the magnitude    |
c           | of the last components of the two vectors |
c           %-------------------------------------------%
c
            if (iconj .eq. 0) then
               bounds(i) = rnorm * dlapy2( workl(i), workl(i+1) )
               bounds(i+1) = bounds(i)
               iconj = 1
            else
               iconj = 0
            end if
         end if
   20 continue
c
 9000 continue
      return
c
c     %---------------%
c     | End of dneigh |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dlaqrb
c
c\Description:
c  Compute the eigenvalues and the Schur decomposition of an upper
c  Hessenberg submatrix in rows and columns ILO to IHI.  Only the
c  last component of the Schur vectors are computed.
c
c  This is mostly a modification of the LAPACK routine dlahqr.
c
c\Usage:
c  call dlaqrb
c     ( WANTT, N, ILO, IHI, H, LDH, WR, WI,  Z, INFO )
c
c\Arguments
c  WANTT   Logical variable.  (INPUT)
c          = .TRUE. : the full Schur form T is required;
c          = .FALSE.: only eigenvalues are required.
c
c  N       Integer.  (INPUT)
c          The order of the matrix H.  N >= 0.
c
c  ILO     Integer.  (INPUT)
c  IHI     Integer.  (INPUT)
c          It is assumed that H is already upper quasi-triangular in
c          rows and columns IHI+1:N, and that H(ILO,ILO-1) = 0 (unless
c          ILO = 1). SLAQRB works primarily with the Hessenberg
c          submatrix in rows and columns ILO to IHI, but applies
c          transformations to all of H if WANTT is .TRUE..
c          1 <= ILO <= max(1,IHI); IHI <= N.
c
c  H       Double precision array, dimension (LDH,N).  (INPUT/OUTPUT)
c          On entry, the upper Hessenberg matrix H.
c          On exit, if WANTT is .TRUE., H is upper quasi-triangular in
c          rows and columns ILO:IHI, with any 2-by-2 diagonal blocks in
c          standard form. If WANTT is .FALSE., the contents of H are
c          unspecified on exit.
c
c  LDH     Integer.  (INPUT)
c          The leading dimension of the array H. LDH >= max(1,N).
c
c  WR      Double precision array, dimension (N).  (OUTPUT)
c  WI      Double precision array, dimension (N).  (OUTPUT)
c          The real and imaginary parts, respectively, of the computed
c          eigenvalues ILO to IHI are stored in the corresponding
c          elements of WR and WI. If two eigenvalues are computed as a
c          complex conjugate pair, they are stored in consecutive
c          elements of WR and WI, say the i-th and (i+1)th, with
c          WI(i) > 0 and WI(i+1) < 0. If WANTT is .TRUE., the
c          eigenvalues are stored in the same order as on the diagonal
c          of the Schur form returned in H, with WR(i) = H(i,i), and, if
c          H(i:i+1,i:i+1) is a 2-by-2 diagonal block,
c          WI(i) = sqrt(H(i+1,i)*H(i,i+1)) and WI(i+1) = -WI(i).
c
c  Z       Double precision array, dimension (N).  (OUTPUT)
c          On exit Z contains the last components of the Schur vectors.
c
c  INFO    Integer.  (OUPUT)
c          = 0: successful exit
c          > 0: SLAQRB failed to compute all the eigenvalues ILO to IHI
c               in a total of 30*(IHI-ILO+1) iterations; if INFO = i,
c               elements i+1:ihi of WR and WI contain those eigenvalues
c               which have been successfully computed.
c
c\Remarks
c  1. None.
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dlabad  LAPACK routine that computes machine constants.
c     dlamch  LAPACK routine that determines machine constants.
c     dlanhs  LAPACK routine that computes various norms of a matrix.
c     dlanv2  LAPACK routine that computes the Schur factorization of
c             2 by 2 nonsymmetric matrix in standard form.
c     dlarfg  LAPACK Householder reflection construction routine.
c     dcopy   Level 1 BLAS that copies one vector to another.
c     drot    Level 1 BLAS that applies a rotation to a 2 by 2 matrix.

c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.4'
c               Modified from the LAPACK routine dlahqr so that only the
c               last component of the Schur vectors are computed.
c
c\SCCS Information: @(#)
c FILE: laqrb.F   SID: 2.2   DATE OF SID: 8/27/96   RELEASE: 2
c
c\Remarks
c     1. None
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dlaqrb ( wantt, n, ilo, ihi, h, ldh, wr, wi,
     &                    z, info )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      logical    wantt
      integer    ihi, ilo, info, ldh, n
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           h( ldh, * ), wi( * ), wr( * ), z( * )
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           zero, one, dat1, dat2
      parameter (zero = 0.0D+0, one = 1.0D+0, dat1 = 7.5D-1,
     &           dat2 = -4.375D-1)
c
      integer(4) ione
      parameter (ione = 1)
c
c     %------------------------%
c     | Local Scalars & Arrays |
c     %------------------------%
c
      integer    i, i1, i2, itn, its, j, k, l, m, nh, nr
      integer    dlqthree
     &
      parameter (dlqthree = 3)
      Double precision
     &           cs, h00, h10, h11, h12, h21, h22, h33, h33s,
     &           h43h34, h44, h44s, ovfl, s, smlnum, sn, sum,
     &           t1, t2, t3, tst1, ulp, unfl, v1, v2, v3
      Double precision
     &           v( 3 ), work( 1 )
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
c
      Double precision
     &           dlamch, dlanhs
      external   dlamch, dlanhs
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy, dlabad, dlanv2, dlarfg, drot
c
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      info = 0
c
c     % if i2 not initialized; Warning: ‘i2’ may be used uninitialized
c     % in this function [-Wmaybe-uninitialized]
      i2 = 0
c
c     %--------------------------%
c     | Quick return if possible |
c     %--------------------------%
c
      if( n.eq.0 )
     &   return
      if( ilo.eq.ihi ) then
         wr( ilo ) = h( ilo, ilo )
         wi( ilo ) = zero
         return
      end if
c
c     %---------------------------------------------%
c     | Initialize the vector of last components of |
c     | the Schur vectors for accumulation.         |
c     %---------------------------------------------%
c
      do 5 j = 1, n-1
         z(j) = zero
  5   continue
      z(n) = one
c
      nh = ihi - ilo + 1
c
c     %-------------------------------------------------------------%
c     | Set machine-dependent constants for the stopping criterion. |
c     | If norm(H) <= sqrt(OVFL), overflow should not occur.        |
c     %-------------------------------------------------------------%
c
      unfl = dlamch( 'safe minimum' )
      ovfl = one / unfl
      call dlabad( unfl, ovfl )
      ulp = dlamch( 'precision' )
      smlnum = unfl*( nh / ulp )
c
c     %---------------------------------------------------------------%
c     | I1 and I2 are the indices of the first row and last column    |
c     | of H to which transformations must be applied. If eigenvalues |
c     | only are computed, I1 and I2 are set inside the main loop.    |
c     | Zero out H(J+2,J) = ZERO for J=1:N if WANTT = .TRUE.          |
c     | else H(J+2,J) for J=ILO:IHI-ILO-1 if WANTT = .FALSE.          |
c     %---------------------------------------------------------------%
c
      if( wantt ) then
         i1 = 1
         i2 = n
         do 8 i=1,i2-2
            h(i1+i+1,i) = zero
 8       continue
      else
         do 9 i=1, ihi-ilo-1
            h(ilo+i+1,ilo+i-1) = zero
 9       continue
      end if
c
c     %---------------------------------------------------%
c     | ITN is the total number of QR iterations allowed. |
c     %---------------------------------------------------%
c
      itn = 30*nh
c
c     ------------------------------------------------------------------
c     The main loop begins here. I is the loop index and decreases from
c     IHI to ILO in steps of 1 or 2. Each iteration of the loop works
c     with the active submatrix in rows and columns L to I.
c     Eigenvalues I+1 to IHI have already converged. Either L = ILO or
c     H(L,L-1) is negligible so that the matrix splits.
c     ------------------------------------------------------------------
c
      i = ihi
   10 continue
      l = ilo
      if( i.lt.ilo )
     &   go to 150

c     %--------------------------------------------------------------%
c     | Perform QR iterations on rows and columns ILO to I until a   |
c     | submatrix of order 1 or 2 splits off at the bottom because a |
c     | subdiagonal element has become negligible.                   |
c     %--------------------------------------------------------------%

      do 130 its = 0, itn
c
c        %----------------------------------------------%
c        | Look for a single small subdiagonal element. |
c        %----------------------------------------------%
c
         do 20 k = i, l + 1, -1
            tst1 = abs( h( k-1, k-1 ) ) + abs( h( k, k ) )
            if( abs(tst1) .le. zero)
     &         tst1 = dlanhs( '1', i-l+1, h( l, l ), ldh, work )
            if( abs( h( k, k-1 ) ).le.max( ulp*tst1, smlnum ) )
     &         go to 30
   20    continue
   30    continue
         l = k
         if( l.gt.ilo ) then
c
c           %------------------------%
c           | H(L,L-1) is negligible |
c           %------------------------%
c
            h( l, l-1 ) = zero
         end if
c
c        %-------------------------------------------------------------%
c        | Exit from loop if a submatrix of order 1 or 2 has split off |
c        %-------------------------------------------------------------%
c
         if( l.ge.i-1 )
     &      go to 140
c
c        %---------------------------------------------------------%
c        | Now the active submatrix is in rows and columns L to I. |
c        | If eigenvalues only are being computed, only the active |
c        | submatrix need be transformed.                          |
c        %---------------------------------------------------------%
c
         if( .not.wantt ) then
            i1 = l
            i2 = i
         end if
c
         if( its.eq.10 .or. its.eq.20 ) then
c
c           %-------------------%
c           | Exceptional shift |
c           %-------------------%
c
            s = abs( h( i, i-1 ) ) + abs( h( i-1, i-2 ) )
            h44 = dat1*s
            h33 = h44
            h43h34 = dat2*s*s
c
         else
c
c           %-----------------------------------------%
c           | Prepare to use Wilkinson's double shift |
c           %-----------------------------------------%
c
            h44 = h( i, i )
            h33 = h( i-1, i-1 )
            h43h34 = h( i, i-1 )*h( i-1, i )
         end if
c
c        %-----------------------------------------------------%
c        | Look for two consecutive small subdiagonal elements |
c        %-----------------------------------------------------%
c
         do 40 m = i - 2, l, -1
c
c           %---------------------------------------------------------%
c           | Determine the effect of starting the double-shift QR    |
c           | iteration at row M, and see if this would make H(M,M-1) |
c           | negligible.                                             |
c           %---------------------------------------------------------%
c
            h11 = h( m, m )
            h22 = h( m+1, m+1 )
            h21 = h( m+1, m )
            h12 = h( m, m+1 )
            h44s = h44 - h11
            h33s = h33 - h11
            v1 = ( h33s*h44s-h43h34 ) / h21 + h12
            v2 = h22 - h11 - h33s - h44s
            v3 = h( m+2, m+1 )
            s = abs( v1 ) + abs( v2 ) + abs( v3 )
            v1 = v1 / s
            v2 = v2 / s
            v3 = v3 / s
            v( 1 ) = v1
            v( 2 ) = v2
            v( 3 ) = v3
            if( m.eq.l )
     &         go to 50
            h00 = h( m-1, m-1 )
            h10 = h( m, m-1 )
            tst1 = abs( v1 )*( abs( h00 )+abs( h11 )+abs( h22 ) )
            if( abs( h10 )*( abs( v2 )+abs( v3 ) ).le.ulp*tst1 )
     &         go to 50
   40    continue
   50    continue
c
c        %----------------------%
c        | Double-shift QR step |
c        %----------------------%
c
         do 120 k = m, i - 1
c
c           ------------------------------------------------------------
c           The first iteration of this loop determines a reflection G
c           from the vector V and applies it from left and right to H,
c           thus creating a nonzero bulge below the subdiagonal.
c
c           Each subsequent iteration determines a reflection G to
c           restore the Hessenberg form in the (K-1)th column, and thus
c           chases the bulge one step toward the bottom of the active
c           submatrix. NR is the order of G.
c           ------------------------------------------------------------
c
            nr = min( dlqthree, i-k+1 )
            if( k.gt.m )
     &         call dcopy( nr, h( k, k-1 ), 1, v(1), 1 )
            call dlarfg( nr, v( 1 ), v( 2 ), 1, t1 )
            if( k.gt.m ) then
               h( k, k-1 ) = v( 1 )
               h( k+1, k-1 ) = zero
               if( k.lt.i-1 )
     &            h( k+2, k-1 ) = zero
            else if( m.gt.l ) then
               h( k, k-1 ) = -h( k, k-1 )
            end if
            v2 = v( 2 )
            t2 = t1*v2
            if( nr.eq.3 ) then
               v3 = v( 3 )
               t3 = t1*v3
c
c              %------------------------------------------------%
c              | Apply G from the left to transform the rows of |
c              | the matrix in columns K to I2.                 |
c              %------------------------------------------------%
c
               do 60 j = k, i2
                  sum = h( k, j ) + v2*h( k+1, j ) + v3*h( k+2, j )
                  h( k, j ) = h( k, j ) - sum*t1
                  h( k+1, j ) = h( k+1, j ) - sum*t2
                  h( k+2, j ) = h( k+2, j ) - sum*t3
   60          continue
c
c              %----------------------------------------------------%
c              | Apply G from the right to transform the columns of |
c              | the matrix in rows I1 to min(K+3,I).               |
c              %----------------------------------------------------%
c
               do 70 j = i1, min( k+3, i )
                  sum = h( j, k ) + v2*h( j, k+1 ) + v3*h( j, k+2 )
                  h( j, k ) = h( j, k ) - sum*t1
                  h( j, k+1 ) = h( j, k+1 ) - sum*t2
                  h( j, k+2 ) = h( j, k+2 ) - sum*t3
   70          continue
c
c              %----------------------------------%
c              | Accumulate transformations for Z |
c              %----------------------------------%
c
               sum      = z( k ) + v2*z( k+1 ) + v3*z( k+2 )
               z( k )   = z( k ) - sum*t1
               z( k+1 ) = z( k+1 ) - sum*t2
               z( k+2 ) = z( k+2 ) - sum*t3

            else if( nr.eq.2 ) then
c
c              %------------------------------------------------%
c              | Apply G from the left to transform the rows of |
c              | the matrix in columns K to I2.                 |
c              %------------------------------------------------%
c
               do 90 j = k, i2
                  sum = h( k, j ) + v2*h( k+1, j )
                  h( k, j ) = h( k, j ) - sum*t1
                  h( k+1, j ) = h( k+1, j ) - sum*t2
   90          continue
c
c              %----------------------------------------------------%
c              | Apply G from the right to transform the columns of |
c              | the matrix in rows I1 to min(K+3,I).               |
c              %----------------------------------------------------%
c
               do 100 j = i1, i
                  sum = h( j, k ) + v2*h( j, k+1 )
                  h( j, k ) = h( j, k ) - sum*t1
                  h( j, k+1 ) = h( j, k+1 ) - sum*t2
  100          continue
c
c              %----------------------------------%
c              | Accumulate transformations for Z |
c              %----------------------------------%
c
               sum      = z( k ) + v2*z( k+1 )
               z( k )   = z( k ) - sum*t1
               z( k+1 ) = z( k+1 ) - sum*t2
            end if
  120    continue

  130 continue
c
c     %-------------------------------------------------------%
c     | Failure to converge in remaining number of iterations |
c     %-------------------------------------------------------%
c
      info = i
      return

  140 continue

      if( l.eq.i ) then
c
c        %------------------------------------------------------%
c        | H(I,I-1) is negligible: one eigenvalue has converged |
c        %------------------------------------------------------%
c
         wr( i ) = h( i, i )
         wi( i ) = zero

      else if( l.eq.i-1 ) then
c
c        %--------------------------------------------------------%
c        | H(I-1,I-2) is negligible;                              |
c        | a pair of eigenvalues have converged.                  |
c        |                                                        |
c        | Transform the 2-by-2 submatrix to standard Schur form, |
c        | and compute and store the eigenvalues.                 |
c        %--------------------------------------------------------%
c
         call dlanv2( h( i-1, i-1 ), h( i-1, i ), h( i, i-1 ),
     &                h( i, i ), wr( i-1 ), wi( i-1 ), wr( i ), wi( i ),
     &                cs, sn )

         if( wantt ) then
c
c           %-----------------------------------------------------%
c           | Apply the transformation to the rest of H and to Z, |
c           | as required.                                        |
c           %-----------------------------------------------------%
c
            if( i2.gt.i )
     &         call drot( i2-i, h( i-1, i+1 ), ldh, h( i, i+1 ), ldh,
     &                    cs, sn )
            call drot( i-i1-1, h( i1, i-1 ), ione, h( i1, i ), ione,
     &                 cs, sn )
            sum      = cs*z( i-1 ) + sn*z( i )
            z( i )   = cs*z( i )   - sn*z( i-1 )
            z( i-1 ) = sum
         end if
      end if
c
c     %---------------------------------------------------------%
c     | Decrement number of remaining iterations, and return to |
c     | start of the main loop with new value of I.             |
c     %---------------------------------------------------------%
c
      itn = itn - its
      i = l - 1
      go to 10

  150 continue
      return
c
c     %---------------%
c     | End of dlaqrb |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dngets
c
c\Description:
c  Given the eigenvalues of the upper Hessenberg matrix H,
c  computes the NP shifts AMU that are zeros of the polynomial of
c  degree NP which filters out components of the unwanted eigenvectors
c  corresponding to the AMU's based on some given criteria.
c
c  NOTE: call this even in the case of user specified shifts in order
c  to sort the eigenvalues, and error bounds of H for later use.
c
c\Usage:
c  call dngets
c     ( ISHIFT, WHICH, KEV, NP, RITZR, RITZI, BOUNDS, SHIFTR, SHIFTI )
c
c\Arguments
c  ISHIFT  Integer.  (INPUT)
c          Method for selecting the implicit shifts at each iteration.
c          ISHIFT = 0: user specified shifts
c          ISHIFT = 1: exact shift with respect to the matrix H.
c
c  WHICH   Character*2.  (INPUT)
c          Shift selection criteria.
c          'LM' -> want the KEV eigenvalues of largest magnitude.
c          'SM' -> want the KEV eigenvalues of smallest magnitude.
c          'LR' -> want the KEV eigenvalues of largest real part.
c          'SR' -> want the KEV eigenvalues of smallest real part.
c          'LI' -> want the KEV eigenvalues of largest imaginary part.
c          'SI' -> want the KEV eigenvalues of smallest imaginary part.
c
c  KEV      Integer.  (INPUT/OUTPUT)
c           INPUT: KEV+NP is the size of the matrix H.
c           OUTPUT: Possibly increases KEV by one to keep complex conjugate
c           pairs together.
c
c  NP       Integer.  (INPUT/OUTPUT)
c           Number of implicit shifts to be computed.
c           OUTPUT: Possibly decreases NP by one to keep complex conjugate
c           pairs together.
c
c  RITZR,  Double precision array of length KEV+NP.  (INPUT/OUTPUT)
c  RITZI   On INPUT, RITZR and RITZI contain the real and imaginary
c          parts of the eigenvalues of H.
c          On OUTPUT, RITZR and RITZI are sorted so that the unwanted
c          eigenvalues are in the first NP locations and the wanted
c          portion is in the last KEV locations.  When exact shifts are
c          selected, the unwanted part corresponds to the shifts to
c          be applied. Also, if ISHIFT .eq. 1, the unwanted eigenvalues
c          are further sorted so that the ones with largest Ritz values
c          are first.
c
c  BOUNDS  Double precision array of length KEV+NP.  (INPUT/OUTPUT)
c          Error bounds corresponding to the ordering in RITZ.
c
c  SHIFTR, SHIFTI  *** USE deprecated as of version 2.1. ***
c
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\Routines called:
c     dsortc  ARPACK sorting routine.
c     dcopy   Level 1 BLAS that copies one vector to another .
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.1'
c
c\SCCS Information: @(#)
c FILE: ngets.F   SID: 2.3   DATE OF SID: 4/20/96   RELEASE: 2
c
c\Remarks
c     1. xxxx
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dngets ( ishift, which, kev, np, ritzr, ritzi, bounds)
c     &                    shiftr, shifti )
c
      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  which*2
      integer    ishift, kev, np
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      Double precision
     &           bounds(kev+np), ritzr(kev+np), ritzi(kev+np)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           zero
      parameter (zero = 0.0)
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      logical    dngetstrue
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   dcopy, dsortc
c
c     %----------------------%
c     | Intrinsics Functions |
c     %----------------------%
c
      intrinsic  abs
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
c     %-------------------------------%
c     | Initialize timing statistics  |
c     | & message level for debugging |
c     %-------------------------------%
c
      dngetstrue = .true.
c
c
c     %----------------------------------------------------%
c     | LM, SM, LR, SR, LI, SI case.                       |
c     | Sort the eigenvalues of H into the desired order   |
c     | and apply the resulting order to BOUNDS.           |
c     | The eigenvalues are sorted so that the wanted part |
c     | are always in the last KEV locations.              |
c     | We first do a pre-processing sort in order to keep |
c     | complex conjugate pairs together                   |
c     %----------------------------------------------------%
c
      if (which .eq. 'LM') then
         call dsortc ('LR', dngetstrue, kev+np, ritzr, ritzi, bounds)
      else if (which .eq. 'SM') then
         call dsortc ('SR', dngetstrue, kev+np, ritzr, ritzi, bounds)
      else if (which .eq. 'LR') then
         call dsortc ('LM', dngetstrue, kev+np, ritzr, ritzi, bounds)
      else if (which .eq. 'SR') then
         call dsortc ('SM', dngetstrue, kev+np, ritzr, ritzi, bounds)
      else if (which .eq. 'LI') then
         call dsortc ('LM', dngetstrue, kev+np, ritzr, ritzi, bounds)
      else if (which .eq. 'SI') then
         call dsortc ('SM', dngetstrue, kev+np, ritzr, ritzi, bounds)
      end if
c
      call dsortc (which, dngetstrue, kev+np, ritzr, ritzi, bounds)
c
c     %-------------------------------------------------------%
c     | Increase KEV by one if the ( ritzr(np),ritzi(np) )    |
c     | = ( ritzr(np+1),-ritzi(np+1) ) and ritz(np) .ne. zero |
c     | Accordingly decrease NP by one. In other words keep   |
c     | complex conjugate pairs together.                     |
c     %-------------------------------------------------------%
c
      if (       (  abs(ritzr(np+1) - ritzr(np)) .le. zero)
     &     .and. (  abs(ritzi(np+1) + ritzi(np)) .le. zero) ) then

         np = np - 1
         kev = kev + 1
      end if
c
      if ( ishift .eq. 1 ) then
c
c        %-------------------------------------------------------%
c        | Sort the unwanted Ritz values used as shifts so that  |
c        | the ones with largest Ritz estimates are first        |
c        | This will tend to minimize the effects of the         |
c        | forward instability of the iteration when they shifts |
c        | are applied in subroutine dnapps.                     |
c        | Be careful and use 'SR' since we want to sort BOUNDS! |
c        %-------------------------------------------------------%
c
         call dsortc ( 'SR', dngetstrue, np, bounds, ritzr, ritzi )
      end if
c
      return
c
c     %---------------%
c     | End of dngets |
c     %---------------%
c
      end
c-----------------------------------------------------------------------
c\BeginDoc
c
c\Name: dnaitr
c
c\Description:
c  Reverse communication interface for applying NP additional steps to
c  a K step nonsymmetric Arnoldi factorization.
c
c  Input:  OP*V_{k}  -  V_{k}*H = r_{k}*e_{k}^T
c
c          with (V_{k}^T)*B*V_{k} = I, (V_{k}^T)*B*r_{k} = 0.
c
c  Output: OP*V_{k+p}  -  V_{k+p}*H = r_{k+p}*e_{k+p}^T
c
c          with (V_{k+p}^T)*B*V_{k+p} = I, (V_{k+p}^T)*B*r_{k+p} = 0.
c
c  where OP and B are as in dnaupd.  The B-norm of r_{k+p} is also
c  computed and returned.
c
c\Usage:
c  call dnaitr
c     ( IDO, BMAT, N, K, NP, NB, RESID, RNORM, V, LDV, H, LDH,
c       IPNTR, WORKD, INFO )
c
c\Arguments
c  IDO     Integer.  (INPUT/OUTPUT)
c          Reverse communication flag.
c          -------------------------------------------------------------
c          IDO =  0: first call to the reverse communication interface
c          IDO = -1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y.
c                    This is for the restart phase to force the new
c                    starting vector into the range of OP.
c          IDO =  1: compute  Y = OP * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y,
c                    IPNTR(3) is the pointer into WORK for B * X.
c          IDO =  2: compute  Y = B * X  where
c                    IPNTR(1) is the pointer into WORK for X,
c                    IPNTR(2) is the pointer into WORK for Y.
c          IDO = 99: done
c          -------------------------------------------------------------
c          When the routine is used in the "shift-and-invert" mode, the
c          vector B * Q is already available and do not need to be
c          recompute in forming OP * Q.
c
c  BMAT    Character*1.  (INPUT)
c          BMAT specifies the type of the matrix B that defines the
c          semi-inner product for the operator OP.  See dnaupd.
c          B = 'I' -> standard eigenvalue problem A*x = lambda*x
c          B = 'G' -> generalized eigenvalue problem A*x = lambda*M**x
c
c  N       Integer.  (INPUT)
c          Dimension of the eigenproblem.
c
c  K       Integer.  (INPUT)
c          Current size of V and H.
c
c  NP      Integer.  (INPUT)
c          Number of additional Arnoldi steps to take.
c
c  NB      Integer.  (INPUT)
c          Blocksize to be used in the recurrence.
c          Only work for NB = 1 right now.  The goal is to have a
c          program that implement both the block and non-block method.
c
c  RESID   Double precision array of length N.  (INPUT/OUTPUT)
c          On INPUT:  RESID contains the residual vector r_{k}.
c          On OUTPUT: RESID contains the residual vector r_{k+p}.
c
c  RNORM   Double precision scalar.  (INPUT/OUTPUT)
c          B-norm of the starting residual on input.
c          B-norm of the updated residual r_{k+p} on output.
c
c  V       Double precision N by K+NP array.  (INPUT/OUTPUT)
c          On INPUT:  V contains the Arnoldi vectors in the first K
c          columns.
c          On OUTPUT: V contains the new NP Arnoldi vectors in the next
c          NP columns.  The first K columns are unchanged.
c
c  LDV     Integer.  (INPUT)
c          Leading dimension of V exactly as declared in the calling
c          program.
c
c  H       Double precision (K+NP) by (K+NP) array.  (INPUT/OUTPUT)
c          H is used to store the generated upper Hessenberg matrix.
c
c  LDH     Integer.  (INPUT)
c          Leading dimension of H exactly as declared in the calling
c          program.
c
c  IPNTR   Integer array of length 3.  (OUTPUT)
c          Pointer to mark the starting locations in the WORK for
c          vectors used by the Arnoldi iteration.
c          -------------------------------------------------------------
c          IPNTR(1): pointer to the current operand vector X.
c          IPNTR(2): pointer to the current result vector Y.
c          IPNTR(3): pointer to the vector B * X when used in the
c                    shift-and-invert mode.  X is the current operand.
c          -------------------------------------------------------------
c
c  WORKD   Double precision work array of length 3*N.  (REVERSE COMMUNICATION)
c          Distributed array to be used in the basic Arnoldi iteration
c          for reverse communication.  The calling program should not
c          use WORKD as temporary workspace during the iteration !!!!!!
c          On input, WORKD(1:N) = B*RESID and is used to save some
c          computation at the first step.
c
c  INFO    Integer.  (OUTPUT)
c          = 0: Normal exit.
c          > 0: Size of the spanning invariant subspace of OP found.
c
c\EndDoc
c
c-----------------------------------------------------------------------
c
c\BeginLib
c
c\Local variables:
c     xxxxxx  real
c
c\References:
c  1. D.C. Sorensen, "Implicit Application of Polynomial Filters in
c     a k-Step Arnoldi Method", SIAM J. Matr. Anal. Apps., 13 (1992),
c     pp 357-385.
c  2. R.B. Lehoucq, "Analysis and Implementation of an Implicitly
c     Restarted Arnoldi Iteration", Rice University Technical Report
c     TR95-13, Department of Computational and Applied Mathematics.
c
c\Routines called:
c     dgetv0  ARPACK routine to generate the initial vector.
c     ivout   ARPACK utility routine that prints integers.
c     second  ARPACK utility routine for timing.
c     dmout   ARPACK utility routine that prints matrices
c     dvout   ARPACK utility routine that prints vectors.
c     dlabad  LAPACK routine that computes machine constants.
c     dlamch  LAPACK routine that determines machine constants.
c     dlascl  LAPACK routine for careful scaling of a matrix.
c     dlanhs  LAPACK routine that computes various norms of a matrix.
c     dgemv   Level 2 BLAS routine for matrix vector multiplication.
c     daxpy   Level 1 BLAS that computes a vector triad.
c     dscal   Level 1 BLAS that scales a vector.
c     dcopy   Level 1 BLAS that copies one vector to another .
c     ddot    Level 1 BLAS that computes the scalar product of two vectors.
c     dnrm2   Level 1 BLAS that computes the norm of a vector.
c
c\Author
c     Danny Sorensen               Phuong Vu
c     Richard Lehoucq              CRPC / Rice University
c     Dept. of Computational &     Houston, Texas
c     Applied Mathematics
c     Rice University
c     Houston, Texas
c
c\Revision history:
c     xx/xx/92: Version ' 2.4'
c
c\SCCS Information: @(#)
c FILE: naitr.F   SID: 2.4   DATE OF SID: 8/27/96   RELEASE: 2
c
c\Remarks
c  The algorithm implemented is:
c
c  restart = .false.
c  Given V_{k} = [v_{1}, ..., v_{k}], r_{k};
c  r_{k} contains the initial residual vector even for k = 0;
c  Also assume that rnorm = || B*r_{k} || and B*r_{k} are already
c  computed by the calling program.
c
c  betaj = rnorm ; p_{k+1} = B*r_{k} ;
c  For  j = k+1, ..., k+np  Do
c     1) if ( betaj < tol ) stop or restart depending on j.
c        ( At present tol is zero )
c        if ( restart ) generate a new starting vector.
c     2) v_{j} = r(j-1)/betaj;  V_{j} = [V_{j-1}, v_{j}];
c        p_{j} = p_{j}/betaj
c     3) r_{j} = OP*v_{j} where OP is defined as in dnaupd
c        For shift-invert mode p_{j} = B*v_{j} is already available.
c        wnorm = || OP*v_{j} ||
c     4) Compute the j-th step residual vector.
c        w_{j} =  V_{j}^T * B * OP * v_{j}
c        r_{j} =  OP*v_{j} - V_{j} * w_{j}
c        H(:,j) = w_{j};
c        H(j,j-1) = rnorm
c        rnorm = || r_(j) ||
c        If (rnorm > 0.717*wnorm) accept step and go back to 1)
c     5) Re-orthogonalization step:
c        s = V_{j}'*B*r_{j}
c        r_{j} = r_{j} - V_{j}*s;  rnorm1 = || r_{j} ||
c        alphaj = alphaj + s_{j};
c     6) Iterative refinement step:
c        If (rnorm1 > 0.717*rnorm) then
c           rnorm = rnorm1
c           accept step and go back to 1)
c        Else
c           rnorm = rnorm1
c           If this is the first time in step 6), go to 5)
c           Else r_{j} lies in the span of V_{j} numerically.
c              Set r_{j} = 0 and rnorm = 0; go to 1)
c        EndIf
c  End Do
c
c\EndLib
c
c-----------------------------------------------------------------------
c
      subroutine dnaitr
     &   (ido, bmat, n, k, np, nb, resid, rnorm, v, ldv, h, ldh,
     &     ipntr, workd, info)

      implicit none
c
c     %------------------%
c     | Scalar Arguments |
c     %------------------%
c
      character  bmat*1
      integer    ido, info, k, ldh, ldv, n, nb, np
      Double precision
     &           rnorm
c
c     %-----------------%
c     | Array Arguments |
c     %-----------------%
c
      integer    ipntr(3)
      Double precision
     &           h(ldh,k+np), resid(n), v(ldv,k+np), workd(3*n)
c
c     %------------%
c     | Parameters |
c     %------------%
c
      Double precision
     &           one, zero
      parameter (one = 1.0D+0, zero = 0.0D+0)
      logical    fls
      parameter (fls = .false. )
c
c     %---------------%
c     | Local Scalars |
c     %---------------%
c
      logical    first, orth1, orth2, rstart, step3, step4
      integer    ierr, i, infol, ipj, irj, ivj, iter, itry, j,
     &            jj
      integer    dnaitrone
      parameter (dnaitrone = 1)
c
      integer(4)
     &           ione
      parameter (ione = 1)
c
      Double precision
     &           betaj, ovfl, temp1, rnorm1, smlnum, tst1, ulp, unfl,
     &           wnorm
      save       first, orth1, orth2, rstart, step3, step4,
     &           ierr, ipj, irj, ivj, iter, itry, j,  ovfl,
     &           betaj, rnorm1, smlnum, ulp, unfl, wnorm
c
c     %----------------------%
c     | External Subroutines |
c     %----------------------%
c
      external   daxpy, dcopy, dscal, dgemv, dgetv0, dlabad
c
c     %--------------------%
c     | External Functions |
c     %--------------------%
c
      Double precision
     &           ddot, dnrm2, dlanhs, dlamch
      external   ddot, dnrm2, dlanhs, dlamch
c
c     %---------------------%
c     | Intrinsic Functions |
c     %---------------------%
c
      intrinsic    abs, sqrt
c
c     %-----------------%
c     | Data statements |
c     %-----------------%
c
      data      first / .true. /
c
c     %-----------------------%
c     | Executable Statements |
c     %-----------------------%
c
      if (nb .gt. 1000) then
        goto 9000
      end if
c
      if (first) then
c
c        %-----------------------------------------%
c        | Set machine-dependent constants for the |
c        | the splitting and deflation criterion.  |
c        | If norm(H) <= sqrt(OVFL),               |
c        | overflow should not occur.              |
c        | REFERENCE: LAPACK subroutine dlahqr     |
c        %-----------------------------------------%
c
         unfl = dlamch( 'safe minimum' )
         ovfl = one / unfl
         call dlabad( unfl, ovfl )
         ulp = dlamch( 'precision' )
         smlnum = unfl*( n / ulp )
         first = .false.
      end if
c
      if (ido .eq. 0) then
c
c        %-------------------------------%
c        | Initialize timing statistics  |
c        | & message level for debugging |
c        %-------------------------------%
c
c         call second (t0)
cm         msglvl = mnaitr
c
c        %------------------------------%
c        | Initial call to this routine |
c        %------------------------------%
c
         info   = 0
         step3  = .false.
         step4  = .false.
         rstart = .false.
         orth1  = .false.
         orth2  = .false.
         j      = k + 1
         ipj    = 1
         irj    = ipj   + n
         ivj    = irj   + n
      end if
c
c     %-------------------------------------------------%
c     | When in reverse communication mode one of:      |
c     | STEP3, STEP4, ORTH1, ORTH2, RSTART              |
c     | will be .true. when ....                        |
c     | STEP3: return from computing OP*v_{j}.          |
c     | STEP4: return from computing B-norm of OP*v_{j} |
c     | ORTH1: return from computing B-norm of r_{j+1}  |
c     | ORTH2: return from computing B-norm of          |
c     |        correction to the residual vector.       |
c     | RSTART: return from OP computations needed by   |
c     |         dgetv0.                                 |
c     %-------------------------------------------------%
c
      if (step3)  go to 50
      if (step4)  go to 60
      if (orth1)  go to 70
      if (orth2)  go to 90
      if (rstart) go to 30
c
c     %-----------------------------%
c     | Else this is the first step |
c     %-----------------------------%
c
c     %--------------------------------------------------------------%
c     |                                                              |
c     |        A R N O L D I     I T E R A T I O N     L O O P       |
c     |                                                              |
c     | Note:  B*r_{j-1} is already in WORKD(1:N)=WORKD(IPJ:IPJ+N-1) |
c     %--------------------------------------------------------------%

 1000 continue
c
c        %---------------------------------------------------%
c        | STEP 1: Check if the B norm of j-th residual      |
c        | vector is zero. Equivalent to determing whether   |
c        | an exact j-step Arnoldi factorization is present. |
c        %---------------------------------------------------%
c
         betaj = rnorm
         if (rnorm .gt. zero) go to 40
c
c           %---------------------------------------------------%
c           | Invariant subspace found, generate a new starting |
c           | vector which is orthogonal to the current Arnoldi |
c           | basis and continue the iteration.                 |
c           %---------------------------------------------------%
c           %---------------------------------------------%
c           | ITRY is the loop variable that controls the |
c           | maximum amount of times that a restart is   |
c           | attempted. NRSTRT is used by stat.h         |
c           %---------------------------------------------%
c
            betaj  = zero
cp            nrstrt = nrstrt + 1
            itry   = 1
   20       continue
            rstart = .true.
            ido    = 0
   30       continue
c
c           %--------------------------------------%
c           | If in reverse communication mode and |
c           | RSTART = .true. flow returns here.   |
c           %--------------------------------------%
c
            call dgetv0 (ido, bmat, fls, n, j, v, ldv,
     &                   resid, rnorm, ipntr, workd, ierr)
            if (ido .ne. 99) go to 9000
            if (ierr .lt. 0) then
               itry = itry + 1
               if (itry .le. 3) go to 20
c
c              %------------------------------------------------%
c              | Give up after several restart attempts.        |
c              | Set INFO to the size of the invariant subspace |
c              | which spans OP and exit.                       |
c              %------------------------------------------------%
c
               info = j - 1
               ido = 99
               go to 9000
            end if
c
   40    continue
c
c        %---------------------------------------------------------%
c        | STEP 2:  v_{j} = r_{j-1}/rnorm and p_{j} = p_{j}/rnorm  |
c        | Note that p_{j} = B*r_{j-1}. In order to avoid overflow |
c        | when reciprocating a small RNORM, test against lower    |
c        | machine bound.                                          |
c        %---------------------------------------------------------%
c
         call dcopy (n, resid, 1, v(1,j), 1)
         if (rnorm .ge. unfl) then
             temp1 = one / rnorm
             call dscal (n, temp1, v(1,j), ione)
             call dscal (n, temp1, workd(ipj), ione)
         else
c
c            %-----------------------------------------%
c            | To scale both v_{j} and p_{j} carefully |
c            | use LAPACK routine SLASCL               |
c            %-----------------------------------------%
c
             call dlascl ('General', i, i, rnorm, one, n, 1,
     &                    v(1,j), n, infol)
             call dlascl ('General', i, i, rnorm, one, n, 1,
     &                    workd(ipj), n, infol)
         end if
c
c        %------------------------------------------------------%
c        | STEP 3:  r_{j} = OP*v_{j}; Note that p_{j} = B*v_{j} |
c        | Note that this is not quite yet r_{j}. See STEP 4    |
c        %------------------------------------------------------%
c
         step3 = .true.
         call dcopy (n, v(1,j), 1, workd(ivj), 1)
         ipntr(1) = ivj
         ipntr(2) = irj
         ipntr(3) = ipj
         ido = 1
c
c        %-----------------------------------%
c        | Exit in order to compute OP*v_{j} |
c        %-----------------------------------%
c
         go to 9000
   50    continue
c
c        %----------------------------------%
c        | Back from reverse communication; |
c        | WORKD(IRJ:IRJ+N-1) := OP*v_{j}   |
c        | if step3 = .true.                |
c        %----------------------------------%
c
         step3 = .false.
c
c        %------------------------------------------%
c        | Put another copy of OP*v_{j} into RESID. |
c        %------------------------------------------%
c
         call dcopy (n, workd(irj), 1, resid(1), 1)
c
c        %---------------------------------------%
c        | STEP 4:  Finish extending the Arnoldi |
c        |          factorization to length j.   |
c        %---------------------------------------%
c
         if (bmat .eq. 'G') then
            step4 = .true.
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %-------------------------------------%
c           | Exit in order to compute B*OP*v_{j} |
c           %-------------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(ipj), 1)
         end if
   60    continue
c
c        %----------------------------------%
c        | Back from reverse communication; |
c        | WORKD(IPJ:IPJ+N-1) := B*OP*v_{j} |
c        | if step4 = .true.                |
c        %----------------------------------%
c
         step4 = .false.
c
c        %-------------------------------------%
c        | The following is needed for STEP 5. |
c        | Compute the B-norm of OP*v_{j}.     |
c        %-------------------------------------%
c
         if (bmat .eq. 'G') then
             wnorm = ddot (n, resid, 1, workd(ipj), 1)
             wnorm = sqrt(abs(wnorm))
         else if (bmat .eq. 'I') then
            wnorm = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------%
c        | Compute the j-th residual corresponding |
c        | to the j step factorization.            |
c        | Use Classical Gram Schmidt and compute: |
c        | w_{j} <-  V_{j}^T * B * OP * v_{j}      |
c        | r_{j} <-  OP*v_{j} - V_{j} * w_{j}      |
c        %-----------------------------------------%
c
c
c        %------------------------------------------%
c        | Compute the j Fourier coefficients w_{j} |
c        | WORKD(IPJ:IPJ+N-1) contains B*OP*v_{j}.  |
c        %------------------------------------------%
c
         call dgemv ('T', n, j, one, v, ldv, workd(ipj), 1,
     &               zero, h(1,j), 1)
c
c        %--------------------------------------%
c        | Orthogonalize r_{j} against V_{j}.   |
c        | RESID contains OP*v_{j}. See STEP 3. |
c        %--------------------------------------%
c
         call dgemv ('N', n, j, -one, v, ldv, h(1,j), 1,
     &               one, resid(1), 1)
c
         if (j .gt. 1) h(j,j-1) = betaj
c
         orth1 = .true.
c
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(irj), 1)
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %----------------------------------%
c           | Exit in order to compute B*r_{j} |
c           %----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(ipj), 1)
         end if
   70    continue
c
c        %---------------------------------------------------%
c        | Back from reverse communication if ORTH1 = .true. |
c        | WORKD(IPJ:IPJ+N-1) := B*r_{j}.                    |
c        %---------------------------------------------------%
c
         orth1 = .false.
c
c        %------------------------------%
c        | Compute the B-norm of r_{j}. |
c        %------------------------------%
c
         if (bmat .eq. 'G') then
            rnorm = ddot (n, resid, 1, workd(ipj), 1)
            rnorm = sqrt(abs(rnorm))
         else if (bmat .eq. 'I') then
            rnorm = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------------------------%
c        | STEP 5: Re-orthogonalization / Iterative refinement phase |
c        | Maximum NITER_ITREF tries.                                |
c        |                                                           |
c        |          s      = V_{j}^T * B * r_{j}                     |
c        |          r_{j}  = r_{j} - V_{j}*s                         |
c        |          alphaj = alphaj + s_{j}                          |
c        |                                                           |
c        | The stopping criteria used for iterative refinement is    |
c        | discussed in Parlett's book SEP, page 107 and in Gragg &  |
c        | Reichel ACM TOMS paper; Algorithm 686, Dec. 1990.         |
c        | Determine if we need to correct the residual. The goal is |
c        | to enforce ||v(:,1:j)^T * r_{j}|| .le. eps * || r_{j} ||  |
c        | The following test determines whether the sine of the     |
c        | angle between  OP*x and the computed residual is less     |
c        | than or equal to 0.717.                                   |
c        %-----------------------------------------------------------%
c
         if (rnorm .gt. 0.717*wnorm) go to 100
         iter  = 0
c
c        %---------------------------------------------------%
c        | Enter the Iterative refinement phase. If further  |
c        | refinement is necessary, loop back here. The loop |
c        | variable is ITER. Perform a step of Classical     |
c        | Gram-Schmidt using all the Arnoldi vectors V_{j}  |
c        %---------------------------------------------------%
c
   80    continue
c
c        %----------------------------------------------------%
c        | Compute V_{j}^T * B * r_{j}.                       |
c        | WORKD(IRJ:IRJ+J-1) = v(:,1:J)'*WORKD(IPJ:IPJ+N-1). |
c        %----------------------------------------------------%
c
         call dgemv ('T', n, j, one, v, ldv, workd(ipj), 1,
     &               zero, workd(irj), 1)
c
c        %---------------------------------------------%
c        | Compute the correction to the residual:     |
c        | r_{j} = r_{j} - V_{j} * WORKD(IRJ:IRJ+J-1). |
c        | The correction to H is v(:,1:J)*H(1:J,1:J)  |
c        | + v(:,1:J)*WORKD(IRJ:IRJ+J-1)*e'_j.         |
c        %---------------------------------------------%
c
         call dgemv ('N', n, j, -one, v, ldv, workd(irj), 1,
     &               one, resid(1), 1)
         call daxpy (j, one, workd(irj), 1, h(1,j), 1)
c
         orth2 = .true.
         if (bmat .eq. 'G') then
            call dcopy (n, resid, 1, workd(irj), 1)
            ipntr(1) = irj
            ipntr(2) = ipj
            ido = 2
c
c           %-----------------------------------%
c           | Exit in order to compute B*r_{j}. |
c           | r_{j} is the corrected residual.  |
c           %-----------------------------------%
c
            go to 9000
         else if (bmat .eq. 'I') then
            call dcopy (n, resid, 1, workd(ipj), 1)
         end if
   90    continue
c
c        %---------------------------------------------------%
c        | Back from reverse communication if ORTH2 = .true. |
c        %---------------------------------------------------%
c
c        %-----------------------------------------------------%
c        | Compute the B-norm of the corrected residual r_{j}. |
c        %-----------------------------------------------------%
c
         if (bmat .eq. 'G') then
             rnorm1 = ddot (n, resid, 1, workd(ipj), 1)
             rnorm1 = sqrt(abs(rnorm1))
         else if (bmat .eq. 'I') then
             rnorm1 = dnrm2(n, resid, 1)
         end if
c
c        %-----------------------------------------%
c        | Determine if we need to perform another |
c        | step of re-orthogonalization.           |
c        %-----------------------------------------%
c
         if (rnorm1 .gt. 0.717*rnorm) then
c
c           %---------------------------------------%
c           | No need for further refinement.       |
c           | The cosine of the angle between the   |
c           | corrected residual vector and the old |
c           | residual vector is greater than 0.717 |
c           | In other words the corrected residual |
c           | and the old residual vector share an  |
c           | angle of less than arcCOS(0.717)      |
c           %---------------------------------------%
c
            rnorm = rnorm1
c
         else
c
c           %-------------------------------------------%
c           | Another step of iterative refinement step |
c           | is required. NITREF is used by stat.h     |
c           %-------------------------------------------%
c
            rnorm  = rnorm1
            iter   = iter + 1
            if (iter .le. 1) go to 80
c
c           %-------------------------------------------------%
c           | Otherwise RESID is numerically in the span of V |
c           %-------------------------------------------------%
c
            do 95 jj = 1, n
               resid(jj) = zero
  95        continue
            rnorm = zero
         end if
c
c        %----------------------------------------------%
c        | Branch here directly if iterative refinement |
c        | wasn't necessary or after at most NITER_REF  |
c        | steps of iterative refinement.               |
c        %----------------------------------------------%
c
  100    continue
c
         rstart = .false.
         orth2  = .false.
c
c        %------------------------------------%
c        | STEP 6: Update  j = j+1;  Continue |
c        %------------------------------------%
c
         j = j + 1
         if (j .gt. k+np) then
            ido = 99
            do 110 i = max(dnaitrone,k), k+np-1
c
c              %--------------------------------------------%
c              | Check for splitting and deflation.         |
c              | Use a standard test as in the QR algorithm |
c              | REFERENCE: LAPACK subroutine dlahqr        |
c              %--------------------------------------------%
c
               tst1 = abs( h( i, i ) ) + abs( h( i+1, i+1 ) )
               if( abs(tst1) .le. zero )
     &              tst1 = dlanhs( '1', k+np, h, ldh, workd(n+1) )
               if( abs( h( i+1,i ) ).le.max( ulp*tst1, smlnum ) )
     &              h(i+1,i) = zero
 110        continue
c
            go to 9000
         end if
c
c        %--------------------------------------------------------%
c        | Loop back to extend the factorization by another step. |
c        %--------------------------------------------------------%
c
      go to 1000
c
c     %---------------------------------------------------------------%
c     |                                                               |
c     |  E N D     O F     M A I N     I T E R A T I O N     L O O P  |
c     |                                                               |
c     %---------------------------------------------------------------%
c
 9000 continue
      return
c
c     %---------------%
c     | End of dnaitr |
c     %---------------%
c
      end
c

# HEADER ####################################################
# This is file spam/R/spam_solve.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################
########################################################################
#
# Contains routines linked to solving spd linear systems. Namely:
#    chol, solve, backsolve, forwardsolve
#    determinant                (the later because it is based on chol)
#
# As well as associated S4 elements.
#
# The key element is a new class: "spam.chol.NgPeyton", the output
# of 'chol'
#
########################################################################
########################################################################


# lindx=  colindices
# xlindx= colpointers
# xlnz=   rowpointers
# snode=snmember
# xsuper=supernodes
# c(... nnztmp,cachesize)= memory
#### Summary of memory usage in cholstepwise:
# m
# nsuper
# nnzlmax
#
# Internally, we additionally assign several arrays (of length):
#  m: adj, colcnt
#  7*m+3: iwork

# iwork: is often used, anytime:


########################################################################

print.spam.chol.NgPeyton <- function(x,...) {
  nrow <- x@dimension[1]
  nnzR <- x@rowpointers[nrow+1]-1
  cat("(Upper) Cholesky factor of dimension ", nrow,
                "x", nrow, " with ",nnzR," (row-wise) nonzero elements (",printSize(x),").", sep = "", fill=TRUE)
  cat("    (The object is supposed to be used with: 'as.spam', 'backsolve', 'forwardsolve', etc.)\n",
      fill=TRUE)
  cat("Class 'spam.chol.NgPeyton'\n")
  invisible(NULL)
}

setMethod("show","spam.chol.NgPeyton", function(object) {
  nrow <- object@dimension[1]
  nnzR <- object@rowpointers[nrow+1]-1
  cat("(Upper) Cholesky factor of dimension ", nrow,
                "x", nrow, " with ",nnzR," (row-wise) nonzero elements (",printSize(object),").", sep = "", fill=TRUE)
  cat("    (The object is supposed to be used with: 'as.spam', 'backsolve', 'forwardsolve', etc.)\n",
      fill=TRUE)
  cat("Class 'spam.chol.NgPeyton'\n")
  invisible(NULL)
        })


"diag.of.spam.chol.NgPeyton" <- function(x, nrow, ncol)
  return( x@entries[x@rowpointers[-(x@dimension[1]+1)]])


setMethod("diag",    "spam.chol.NgPeyton", diag.of.spam.chol.NgPeyton)
setMethod("diag<-",  "spam.chol.NgPeyton", function(x) stop("diagonal cannot be changed on 'spam.chol.NgPeyton' object"))

setMethod("print",   "spam.chol.NgPeyton", print.spam.chol.NgPeyton)
setMethod("length",  "spam.chol.NgPeyton",function(x) x@rowpointers[x@dimension[1]+1]-1)
setMethod("length<-","spam.chol.NgPeyton",function(x,value) stop("length cannot be changed on 'spam.chol.NgPeyton' object") )
setMethod("dim",     "spam.chol.NgPeyton",function(x) x@dimension)
setMethod("dim<-",   "spam.chol.NgPeyton",function(x,value) stop("dimension cannot be altered on 'spam.chol.NgPeyton' object") )

# summary.spam.chol.NgPeyton is in file `summary.R`

setMethod("c","spam.chol.NgPeyton", function(x,...){
    nrow <- x@dimension[1]
    nnzR <- x@rowpointers[nrow+1]-1
    nsuper <- as.integer( length(x@supernodes)-1)
    if( getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

    xcolindices <- .C64('calcja',
                        SIGNATURE = rep(SS$signature, 7),

                        nrow,
                        nsuper,
                        x@supernodes,
                        x@colindices,

                        x@colpointers,
                        x@rowpointers,
                        xja = vector_dc( SS$type, nnzR),

                        INTENT=c("r", "r", "r", "r",
                                 "r", "r", "w"),
                        NAOK = getOption("spam.NAOK"),
                        PACKAGE = SS$package)$xja
    cx <- .C64("spamcsrdns",
               SIGNATURE = c(SS$signature, "double" , SS$signature, SS$signature, "double"),

               nrow = nrow,
               entries = x@entries,
               colindices = xcolindices,
               rowpointers = x@rowpointers,

               res = vector_dc( "double", nrow*nrow),

               INTENT = c("r", "r", "r", "r",
                          "w"),
               NAOK=getOption("spam.NAOK"),
               PACKAGE = SS$package)$res
  if (length( list(...)) < 1)
    return( cx)
  else
    c( cx,c(...))
})

as.spam.chol.NgPeyton <- function(x, eps = getOption("spam.eps")) {
    if( getOption("spam.force64") || .format.spam(x)$package != "spam")
        SS <- .format64()
    else
        SS <- .format32

  if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision",call.=FALSE)
  nrow <- x@dimension[1]
  nnzR <- x@rowpointers[nrow+1]-1
  nsuper <- length(x@supernodes)-1

  colindices <- .C64('calcja',
                     SIGNATURE=rep(SS$signature, 7),

                     nrow,
                     nsuper,
                     x@supernodes,
                     x@colindices,

                     x@colpointers,
                     x@rowpointers,
                     xja=vector(SS$type, nnzR), #!1#

                     INTENT=c("r", "r", "r", "r", "r", "r", "w"),
                     NAOK = getOption("spam.NAOK"),
                     PACKAGE = SS$package)$xja

  return(.newSpam(
    entries=x@entries,
    colindices=colindices,
    rowpointers=x@rowpointers,
    dimension=x@dimension
  ))
}



"as.matrix.spam.chol.NgPeyton" <- function(x,...){
    ## print("as.matrix.spam.chol.NgPeyton")
    nrow <- x@dimension[1]
    nnzR <- x@rowpointers[nrow+1]-1L
    ## newx <- new("spam")
    nsuper <-  length(x@supernodes)-1L
    ## xcolindices <- .Fortran('calcja',
    ##                         as.integer(nrow), as.integer(nsuper), as.integer(x@supernodes), as.integer(x@colindices), as.integer(x@colpointers), as.integer(x@rowpointers),
    ##                         xja=vector("integer",nnzR),
    ##                         NAOK = getOption("spam.NAOK"),PACKAGE = "spam")$xja
    if( getOption("spam.force64") || .format.spam(x)$package != "spam" )
        SS <- .format64()
    else
        SS <- .format32

    xcolindices <- .C64('calcja',
                        SIGNATURE = c(rep(SS$signature,7)),
                        nrow,
                        nsuper,
                        x@supernodes,
                        x@colindices,

                        x@colpointers,
                        x@rowpointers,
                        xja = vector_dc( SS$type, nnzR),

                        INTENT=c("r", "r", "r", "r",
                                 "r", "r", "w"),
                        NAOK = getOption("spam.NAOK"),
                        PACKAGE = SS$package)$xja
    return(array(.C64("spamcsrdns",
                      SIGNATURE = c(SS$signature, "double" , SS$signature, SS$signature,
                             "double"),
                 nrow = nrow,
                 entries = x@entries,
                 colindices = xcolindices,
                 rowpointers = x@rowpointers,

                 res = vector_dc( "double", nrow*nrow),  # dotCall takes care of size!

                 INTENT = c("r", "r", "r", "r",
                            "w"),
                 NAOK=getOption("spam.NAOK"),
                 PACKAGE = SS$package)$res,
               c(nrow,nrow))      # we preserve dimensions
         )
}

setMethod("as.spam","spam.chol.NgPeyton", as.spam.chol.NgPeyton)
setMethod("as.matrix","spam.chol.NgPeyton",as.matrix.spam.chol.NgPeyton)
setMethod("as.vector","spam.chol.NgPeyton",
          function(x){
            as.vector.spam(as.spam.chol.NgPeyton(x))
          })

########################################################################


setGeneric("backsolve", def = function(r, x, ...) standardGeneric("backsolve"),
           useAsDefault= function(r, x, ...) base::backsolve(r, x, ...))

setGeneric("forwardsolve", def = function(l, x, ...) standardGeneric("forwardsolve"),
           useAsDefault= function(l, x, ...) base::forwardsolve(l, x, ...))

# adapted from methods
#setGeneric("forwardsolve", function(l, x, k, upper.tri = FALSE, transpose = FALSE, ...)
#           standardGeneric("forwardsolve"),
#           useAsDefault = function(l, x, k = ncol(l), upper.tri = FALSE, transpose = FALSE, ...)
#                  base::forwardsolve(l, x, k = k, upper.tri = upper.tri, transpose = transpose, ... ),
#           signature = c("l", "x"))#, where = where)
##### setGenericImplicit("forwardsolve")#, restore=FALSE)


setMethod("chol","spam", chol.spam)
setMethod("solve","spam",solve.spam)
setMethod("chol2inv","spam", chol2inv.spam)
setMethod("chol2inv","spam.chol.NgPeyton", chol2inv.spam)

setMethod("backsolve","spam",              #signature(r="spam",x='ANY'),
          backsolve.spam)
setMethod("backsolve","spam.chol.NgPeyton",#signature(r="spam.chol.NgPeyton",x='ANY'),
          backsolve.spam, sealed=TRUE)
setMethod("forwardsolve","spam",               forwardsolve.spam)
setMethod("forwardsolve","spam.chol.NgPeyton", forwardsolve.spam)

# ?? sealed=TRUE ??  from help(setMethod):
#  'sealed' prevents the method being redefined, but should never be needed
#    when the method is defined in the source code of a package.

######################################################################


"ordering.default" <- function(x, inv=FALSE) stop('Operation not defined form this class')

setGeneric("ordering", function(x, inv=FALSE) standardGeneric("ordering"))

setMethod("ordering","spam.chol.NgPeyton",function(x,inv=FALSE)
          {
            if (inv) return(x@invpivot) else return(x@pivot) })

setMethod("ordering","matrix",function(x,inv=FALSE)
          {
            if (dim(x)[1]!=dim(x)[2])
              stop("ordering is defined for square matrices only")
            if(inv)return(dim(x)[1]:1) else return(1:dim(x)[1]) })

setMethod("ordering","spam",function(x,inv=FALSE)
          {
            if (dim(x)[1]!=dim(x)[2])
              stop("ordering is defined for square matrices only")
            if(inv) return(dim(x)[1]:1) else return(1:dim(x)[1]) })



########################################################################
#  force to spam matrices. Would not be required with inheritance

setMethod("image","spam.chol.NgPeyton",
          function(x,cex=NULL,...){
            image.spam(as.spam.chol.NgPeyton(x),cex=cex,...)
          })


setMethod("display","spam.chol.NgPeyton",
          function(x,...){
            display.spam(as.spam.chol.NgPeyton(x),...)
          })

setMethod("t","spam.chol.NgPeyton",
          function(x){
            t.spam(as.spam.chol.NgPeyton(x))
          })

setMethod("chol","spam.chol.NgPeyton",
          function(x){
           x
          })

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

closestdistCpp <- function(x, y, part, p, method, eta) {
    .Call(`_spam_closestdistCpp`, x, y, part, p, method, eta)
}

# HEADER ####################################################
# This is file spam/R/gmult.R.                              #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

gmult <- function(x, splits, fact) {
  if(!is.spam(x)) { # preserve zero entries
    x <- as.spam(x) }

  splits <- sort.int(unique(as.integer(splits)))
  if(splits[length(splits)] < max(dim(x))+1) {
    splits <- c(splits, max(dim(x))+1) }

  fact <- as.matrix(fact)
  stopifnot(dim(fact) == (length(splits)-1))

  ll <- length(x@entries)

  if(.format.spam(x)$package == "spam64" || getOption("spam.force64")) {
    SS <- .format64()
  } else {
    SS <- .format32
  }

  x@entries <- dotCall64::.C64("gmult_f",
                               SIGNATURE = c("double", SS$signature, SS$signature, SS$signature, SS$signature,
                                             "double", SS$signature, "double"),
                               a = x@entries,
                               ia = x@colindices,
                               ja = x@rowpointers,
                               na = ll,
                               splits = splits,
                               fact = fact,
                               nfact = dim(fact)[2],
                               out = dotCall64::numeric_dc(ll),
                               INTENT = c( "r", "r", "r", "r",
                                           "r", "r", "r", "rw"),
                               NAOK=getOption("spam.NAOK"),
                               PACKAGE = SS$package)$out

  return(x)
}

# HEADER ####################################################
# This is file spam/R/spam_solve.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################
########################################################################
#
# Contains routines linked to solving spd linear systems. Namely:
#    chol.spam, chol.update and triangular solves.
# Associated S4 elements and determinant are in different files.
#
#
########################################################################
########################################################################






# TODO below, check with solve.spam and triangular solves for use of vector_dc
# determinant.spam from "rw" to "w" signature.


chol.spam <- function(x, pivot="MMD", method="NgPeyton", memory=list(),
                      eps=getOption("spam.eps"), Rstruct=NULL, ..., verbose=FALSE){

  if (is(Rstruct,"spam.chol.NgPeyton")) invisible( update.spam.chol.NgPeyton(Rstruct,x,...))

    if (verbose) {
        timeused <- proc.time()
        cat("\nStarting Cholesky factorization with method '",method,"'.\n")
    }
  if (method != "NgPeyton")
      warning(gettextf("method = '%s' is not supported. Using 'NgPeyton'",
                     method), domain = NA)

  z <- .chol.spam.basis(x, pivot=pivot, memory=memory, eps=eps, verbose=verbose)

  nnzRfinal <- z$xlnz[length(z$xlnz)]-1  # reduce to final size

  newx <- new("spam.chol.NgPeyton")
  slot(newx, "entries", check=FALSE) <- z$lnz[1:nnzRfinal]
  slot(newx, "colindices", check=FALSE) <- z$lindx[1:z$nnzlindx]
  slot(newx, "colpointers", check=FALSE) <- z$xlindx[1:(z$nsuper+1)]
  slot(newx, "rowpointers", check=FALSE) <- z$xlnz
  slot(newx, "dimension", check=FALSE) <- c(z$nrow, z$nrow)
  slot(newx, "pivot", check=FALSE) <- z$perm
  slot(newx, "invpivot", check=FALSE) <- z$invp
  slot(newx, "supernodes", check=FALSE) <- z$xsuper[1:(z$nsuper+1)]
  slot(newx, "snmember", check=FALSE) <- z$snode
  slot(newx, "memory", check=FALSE) <- c(z$nnzcolindices, z$nnzR, z$cachesize)
  slot(newx, "nnzA", check=FALSE) <- z$nnzA

  if (verbose) {
      cat("Final size of Cholesky factor (of class 'spam.chol.NgPeyton'): ",
          printSize(newx), ".\n")
      cat("Finished factorization (total time used:", (proc.time()-timeused)[1], "s).\n\n")
  }

  invisible(newx)
}



########################################################################


update.spam.chol.NgPeyton <- function(object,x,...){
    ## print("update.spam.chol.NgPeyton")
  nrow <- object@dimension[1]
  if (!is.spam(x))
    stop("Covariance should be a 'spam' object.")
  if ((x@rowpointers[nrow+1]-1) != object@nnzA)
    stop("Updated covariance entries do not match length of original one.")

    if( getOption("spam.force64") || .format.spam(object)$package != "spam" || .format.spam(x)$package != "spam" )
        SS <- .format64()
    else
        SS <- .format32

    u <- .C64("updatefactor",
     ##          subroutine updatefactor( m,nnzd,
     ## &     d,jd,id, invp,perm,
     ## &                lindx,xlindx, nsuper,lnz,xlnz,
     ## &                snode, xsuper,
     ## &                cachesize,ierr)
              SIGNATURE = c(SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            SS$signature, SS$signature, SS$signature, SS$signature, SS$signature,
                            "double", SS$signature, SS$signature, SS$signature, SS$signature, SS$signature ),
              nrow,
              object@nnzA,

              d =  x@entries,
              jd = x@colindices,
              id = x@rowpointers,

              object@invpivot,
              object@pivot,
              lindx = object@colindices,
              xlindx = object@colpointers,
              nsuper = length(object@supernodes)-1,

              entries = vector_dc( "double", length(object@entries)), #lnz
              rowpointers = object@rowpointers,#xlnz
              snode = object@snmember,
              xsuper = object@supernodes,
              cachesize = object@memory[3],
              ierr = 0,

              ## INTENT = c("r", "r",
              ##            "r", "r", "r",
              ##            "r", "r", "r", "r", "r",
              ##            "rw", "rw", "r", "r", "rw", "rw"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)

  if(u$ierr>1) stop("Internal error in 'update.spam.chol.NgPeyton' code ", u$ierr,call.=FALSE)

  if(u$ierr == 1) {
    if (getOption("spam.cholupdatesingular") == "null")
      return(NULL)
    else if (getOption("spam.cholupdatesingular") == "error")
      stop("Singularity problem when updating a Cholesky Factor.")
    else if (getOption("spam.cholupdatesingular") == "warning")
      warning("Singularity problem when updating a Cholesky Factor.\n'object' not updated.")
    else
      stop("'cholupdatesingular' should be 'error', 'null' or 'warning'.")
  }  else {
    slot(object, "entries", check=FALSE) <- u$entries
  }
  invisible(object)
}




solve.spam <- function (a, b,  Rstruct=NULL, ...) {
  nrow <- a@dimension[1]
  ncol <- a@dimension[2]
  if (ncol != nrow)      stop("only square matrices can be inverted")

  if (missing(b)) {
    b <- diag(1, ncol)
  }  else {
    if(!is.matrix(b)) b <- as.matrix(b)
  }
  p <- dim(b)[2]
  if(nrow!=dim(b)[1])stop("'b' must be compatible with 'a'")

  # if we have a spam matrix, we calculate the Cholesky factor
  if (is(a,"spam"))
    if (is(Rstruct, "spam.chol.NgPeyton"))
        a <- update.spam.chol.NgPeyton(Rstruct, a, ...)
    else a <- chol.spam(a, ...)


  if (is(a,"spam.chol.NgPeyton")) {
      # The following is a fast way to perform:
      #     z <- backsolve(a,forwardsolve( t(a),b))
    nsuper <- as.integer( length(a@supernodes)-1)
      if( getOption("spam.force64") || .format.spam(a)$package != "spam" )
          SS <- .format64()
      else
          SS <- .format32

      z <- .C64("backsolves",
                SIGNATURE = c(rep(SS$signature,5),"double", rep(SS$signature, 4),
                              rep("double",3)),

                m = nrow,
                nsuper,
                p,
                a@colindices,
                a@colpointers,
                a@entries,  # "double"
                a@rowpointers,
                a@invpivot,
                a@pivot,
                a@supernodes,

                vector_dc("double",nrow),
                sol = vector_dc("double",nrow*p),
                as.vector(b,"double"),

                INTENT = c( rep( "r", 10), rep( "rw", 3)),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package )$sol
  } else z <- backsolve(a, forwardsolve( t(a),b))
    # see the helpfile for a comment about the 't(a)' construct.

  if ( p!=1)    dim(z) <- c(nrow,p)
  return( z)
}

chol2inv.spam <- function (x, ...) {
    ## print("chol2inv.spam")
    nrow <- x@dimension[1]

    if (is(x,"spam.chol.NgPeyton")) {
        y <- vector("double",nrow*nrow)
        y[1L + 0L:(nrow - 1L) * (nrow + 1L)] <- 1.0

        if( getOption("spam.force64") || .format.spam(x)$package !="spam" )
            SS <- .format64()
        else
            SS <- .format32


        z <- .C64("backsolves",
          ##         subroutine backsolves(m,nsuper,nrhs,lindx,xlindx,lnz,
     ## &                   xlnz,invp,perm,xsuper,newrhs,sol,b)
                  SIGNATURE = c( rep( SS$signature, 5), "double", rep( SS$signature, 4),
                                rep( "double", 3)),

                  m = nrow, #r
                  length(x@supernodes)-1, #r
                  nrow, #r
                  x@colindices,
                  x@colpointers,
                  x@entries,   # "double"
                  x@rowpointers,
                  x@invpivot, #r
                  x@pivot, #r
                  x@supernodes, #r

                  vector_dc("double",nrow), #rw
                  sol = vector_dc("double",nrow*nrow), #w
                  y, #r

                  INTENT = c( rep( "r", 10),
                             "rw", "w", "r"),
                  NAOK = getOption("spam.NAOK"),
                  PACKAGE = SS$package)$sol

        dim(z) <- c(nrow,nrow)
    } else z <- backsolve.spam(x, forwardsolve.spam( t(x), diag(nrow)))
    return( z)
}

backsolve.spam <- function(r, x,...){#, k = NULL, upper.tri = NULL, transpose = NULL){
# r: spam.chol.NgPeyton structure as returned by chol.spam or a spam object
# x: rhs a vector or a matrix in dense form
# dimensions:  ( m x n) ( n x p)

    if( getOption("spam.force64") || .format.spam(r)$package !="spam" )
        SS <- .format64()
    else
        SS <- .format32


  m <- r@dimension[1]
  if(is.vector(x)) {
    n <- length(x)
    p <- 1
  } else {
    x <- as.matrix(x)
    n <- dim(x)[1]
    p <- dim(x)[2]
  }

  # we separate between "spam.chol.NgPeyton" and "spam"
  if (is(r,"spam.chol.NgPeyton")) {
    if (n!=m) stop("Cholesky factor 'r' not compatible with 'x'")
    nsuper <- length(r@supernodes)-1
    if (!getOption("spam.dopivoting")) {

      z <- .C64("backsolvef",
                SIGNATURE=c(rep(SS$signature, 5), "double", SS$signature, SS$signature,
                            "double"),

                m,
                nsuper,
                p,
                r@colindices,

                r@colpointers,
                r@entries,   # "double"
                r@rowpointers,
                r@supernodes,

                sol = vector_dc("double",m*p),

                INTENT=c("r", "r", "r", "r", "r", "r", "r", "r", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol

    }else{
      z <- .C64("pivotbacksolve",
                SIGNATURE=c(rep(SS$signature, 5), "double", rep(SS$signature, 4),
                            "double", "double", "double"),

                m,
                nsuper,
                p,
                r@colindices,

                r@colpointers,
                r@entries, #  "double"
                r@rowpointers,
                r@invpivot,

                r@pivot,
                r@supernodes,
                vector("double",m),
                sol = vector_dc("double",m*p),
                x,            #  "double"

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "r", "r", "r", "w",
                    "r"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol
    }
  } else {
    if (n!=m) stop("Triangular matrix 'r' not compatible with 'x'")
    # solve R sol = x
    z <- .C64("spamback",
              SIGNATURE=c(SS$signature, SS$signature, "double", "double",
                          "double", SS$signature, SS$signature),
              m=m,
              unused=p,
              sol = vector_dc("double",m*p),
              x=x,

              al=r@entries,
              jal=r@colindices,
              ial=r@rowpointers,

              INTENT=c("rw", "r", "w", "r", "r", "r", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE=SS$package)
    if (z$m<0) stop(gettextf("singular matrix in 'backsolve'. Last zero in diagonal [%d]",
            -z$m), domain = NA)
     else z <- z$sol

  }

  if (p>1) dim(z) <- c(m,p)
  return(z)
}

forwardsolve.spam <- function(l, x,...){#, k = NULL, upper.tri = NULL, transpose = NULL){
#  l: spam.chol.NgPeyton structure as returned by chol.spam
#         or an ordinary lower triangular spam matrix
#  x: rhs a vector a matrix in dense form
#  dimensions:  ( m x n) ( n x p)
#  if (!any(is.null(c(upper.tri,k,transpose ))))
#    warning("'k', 'upper.tri' and 'transpose' argument do not have any effect here")
    if( getOption("spam.force64") || .format.spam(l)$package !="spam" )
        SS <- .format64()
    else
        SS <- .format32

  m <- l@dimension[1]
  if(is.vector(x)) {
    n <- length(x)
    p <- 1L
  } else {
    if(!is.matrix(x)) x <- as.matrix(x)
    n <- dim(x)[1]
    p <- dim(x)[2]
  }

  # we separate between "spam.chol.NgPeyton" and "spam"
  if (is(l,"spam.chol.NgPeyton")) {
    if(n!=m) stop("Cholesky factor 'l' not compatible with 'x'")
    nsuper <- length(l@supernodes)-1
    if (!getOption("spam.dopivoting")) {
      z <- .C64("forwardsolvef",
                SIGNATURE=c(SS$signature, SS$signature, SS$signature, SS$signature,
                            SS$signature, "double", SS$signature, SS$signature,
                            "double"),

                m,
                nsuper,
                p,
                l@colindices,

                l@colpointers,
                l@entries,
                l@rowpointers,
                l@supernodes,

                sol = vector("double",m*p),

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE=SS$package)$sol
    }else{
      z <- .C64("pivotforwardsolve",
                SIGNATURE=c(rep(SS$signature,5), "double", rep(SS$signature, 4),
                            "double", "double", "double"),
                m,
                nsuper,
                p,
                l@colindices,

                l@colpointers,
                l@entries,     # "double"
                l@rowpointers,
                l@invpivot,

                l@pivot,
                l@supernodes,
                vector("double",m),
                sol = vector_dc("double",m*p),

                x,

                INTENT=c("r", "r", "r", "r",
                    "r", "r", "r", "r",
                    "r", "r", "r", "w",
                    "r"),
                PACKAGE=SS$package)$sol
    }
  } else {
    if (n!=m) stop("Triangular matrix 'l' not compatible with 'x'")
    # solve L sol = x
    z <- .C64("spamforward",
              SIGNATURE=c(SS$signature, SS$signature, "double", "double",
                          "double", SS$signature, SS$signature),

              m=m,
              p,
              sol = vector_dc("double",m*p),
              x=x,

              al=l@entries,
              jal=l@colindices,
              ial=l@rowpointers,

              INTENT=c("rw", "r", "w", "r", "r", "r", "r"),
              NAOK = getOption("spam.NAOK"),
              PACKAGE=SS$package)
    if (z$m<0) stop(gettextf("singular matrix in 'forwardsolve'. First zero in diagonal [%d]", -z$m), domain = NA)
    else z <- z$sol
  }
  if (p>1)
    dim(z) <- c(m,p)
  return(z)
}




########################################################################

C     closestdistXY,  distance between x and x or between x and y
C
C
C     We have four distances implemented:
C       c("euclidean", "maximum", "minkowski", "greatcircle")
C
c     In case we need the distance matrix between x and x, then the
c     following parameters are used as well:
c     if part=-1, lower tri, part=0 the entire matrix
c        part= 1, upper tri only.
c     only values smaller than eta are considered.
c     p power for minkowski



      double precision function euclid(x,y,p)
      implicit none
      double precision x,y,p
      p=p ! to avoid: Warning: Unused dummy argument 'p' at (1)
      euclid=(x-y)**2
      return
      end



      double precision function minkowski(x,y,p)
      implicit none
      double precision x,y,p
      minkowski=abs(x-y)**p
      return
      end



      subroutine closestdist( ncol, x,nrowx, y, nrowy,
     &    part, p, method,
     &    eta, colindices, rowpointers, entries, nnz, iflag)

      implicit none


      double precision euclid, minkowski
      external euclid, minkowski

      integer ncol,nrowx, nrowy, nnz, method,  part,  iflag
      integer colindices(nnz), rowpointers(nrowx+1)
      double precision p, x(nrowx,ncol),y(nrowy,ncol)
      double precision eta, entries(nnz)




      if (method.eq.1) then
         p=2.0
         call closestEdistXY( ncol, x,nrowx, y, nrowy,
     &        part, p, euclid,
     &        eta, colindices, rowpointers, entries, nnz, iflag)
      endif
      if (method.eq.2) then
         p=1.0
         call closestMAXdistXY( ncol, x,nrowx, y, nrowy,
     &        part,
     &        eta, colindices, rowpointers, entries, nnz, iflag)
      endif
      if (method.eq.3) then
         call closestEdistXY( ncol, x,nrowx, y, nrowy,
     &        part, p, minkowski,
     &        eta, colindices, rowpointers, entries, nnz, iflag)
      endif
      if (method.eq.4) then
         call closestGCdistXY( x,nrowx, y, nrowy,
     &        part, p,
     &        eta, colindices, rowpointers, entries, nnz, iflag)
      endif
      return
      end



      subroutine closestEdistXY( ncol, x,xnrow, y, ynrow,
     &    part, p, distfcn,
     &    eta, colindices, rowpointers, entries, nnz, iflag)

      implicit none

      double precision distfcn
      external distfcn

      integer ncol,xnrow, ynrow, nnz,  part,iflag
      integer colindices(nnz),rowpointers(xnrow+1)
      double precision p,x(xnrow,ncol), y(ynrow,ncol)
      double precision eta, entries(nnz)

c     local variables
      integer jja, i,j,k, jfrom, jto
      double precision etap, tmp,pinv


      etap=eta**p
      pinv=1/p

      jja=1

      rowpointers(1)=1
      jfrom = 1
      jto = ynrow


c cycle over all rows of x (independent of part)

      do i= 1,xnrow

         if (part .lt. 0) then
            jto = i
         endif
         if (part .gt. 0) then
            jfrom = i
         endif

         do 10 j = jfrom,jto

c Start calculating the distance (until delta is exceeded)
            tmp = 0.0
            do  k = 1, ncol
               tmp = tmp + distfcn(x(i,k),y(j,k),p)
               if( tmp.gt.etap) goto 10
            enddo
c Delta is not exceeded.

c     in case nnz was too small, recall line to get a better estimate
            if( jja .gt. nnz) then
               iflag = i
               goto 20
            endif

            colindices(jja) = j
            if (abs(p-2) .le. 0.D0) then
               entries(jja) = sqrt(tmp)
            else
               if (abs(p-1) .le. 0.D0) then
                  entries(jja) = tmp
               else
                  entries(jja) = tmp**pinv
               endif
            endif
            jja = jja + 1


 10      continue
         rowpointers(i+1)=jja
      enddo


      if (part.gt.0) then
         rowpointers(xnrow+1)=jja
      endif
      nnz=jja-1
 20   continue

      return
      end

      subroutine closestMAXdistXY( ncol, x,xnrow, y, ynrow,
     &    part,
     &    eta, colindices, rowpointers, entries, nnz, iflag)

      implicit none

      integer ncol,xnrow, ynrow, nnz,  part,iflag
      integer colindices(nnz),rowpointers(xnrow+1)
      double precision x(xnrow,ncol), y(ynrow,ncol)
      double precision eta, entries(nnz)

c     local variables
      integer jja, i,j,k, jfrom, jto
      double precision  tmp



      jja=1

      rowpointers(1)=1
      jfrom = 1
      jto = ynrow


      do i= 1,xnrow

         if (part .lt. 0) then
            jto = i
         endif
         if (part .gt. 0) then
            jfrom = i
         endif

         do 10 j = jfrom,jto

c Start calculating the distance
            tmp = 0.0
            do  k = 1, ncol
               tmp = max(tmp, abs(x(i,k)-y(j,k)))
               if( tmp.gt.eta) goto 10
            enddo

c Delta is not exceeded.
c     (i,j) has a distance smaller than eta.



c     in case nnz was too small, recall line to get a better estimate
            if( jja .gt. nnz) then
               iflag = i
               goto 20
            endif

            colindices(jja) = j
            entries(jja) = tmp
            jja = jja + 1


 10      continue
         rowpointers(i+1)=jja
      enddo


      if (part.gt.0) then
         rowpointers(xnrow+1)=jja
      endif
      nnz=jja-1
 20   continue

      return
      end



      subroutine closestGCdistXY( x,nx, y, ny,
     &    part,p,
     &    eta, colindices, rowpointers, entries, nnz, iflag)

      implicit none

      integer nx, ny, nnz, colindices(nnz),rowpointers(nx+1)
      integer part, iflag
      double precision x(nx,2), y(ny,2), p, eta, entries(nnz)

c     local variables
      logical equi
      integer jja, i,j, jfrom, jto
      double precision etap, tmp, tmp1, tmp2
      double precision rad, thres
      double precision scy12(ny), ccy12(ny), sy2(ny)
      double precision scx12,     ccx12,     sx2

      parameter (rad = 0.01745329251994329)
      parameter (thres = 0.99999999999)


c     Great savings if we know that x=y.
c     Changes for archaic d... to ... trigonometric fcn
      if (p .lt. 0) then
         equi=.TRUE.
         p=-p
      else
         equi= .FALSE.
      endif

      jja=1

      etap=cos(eta*rad)
      rowpointers(1)=1
      jfrom = 1
      jto = ny


      DO j=1,ny
         tmp1=y(j,1)*rad
         tmp2=y(j,2)*rad
         ccy12(j)=cos(tmp1)*cos(tmp2)
         scy12(j)=sin(tmp1)*cos(tmp2)
         sy2(j)=sin(tmp2)
      ENDDO



      do i= 1,nx

c     x2 is missing if equi=.TRUE. and we reuse the y stuff
         if (equi .eqv. .TRUE.) then
            ccx12=ccy12(i)
            scx12=scy12(i)
            sx2=sy2(i)
         else
            tmp1=x(i,1)*rad
            tmp2=x(i,2)*rad
            ccx12=cos(tmp1)*cos(tmp2)
            scx12=sin(tmp1)*cos(tmp2)
            sx2=sin(tmp2)
         endif

         if (part .lt. 0) then
            jto = i
         endif
         if (part .gt. 0) then
            jfrom = i
         endif

         do 10 j = jfrom,jto


c     Start calculating the distance
            tmp = ccx12 * ccy12(j) + scx12 * scy12(j)  + sx2*sy2(j)


            if (tmp .lt. etap) goto 10
c     Delta is not exceeded.

c     Due to numerical instabilities, we need the following... 0.15-2:
c     Patch suggested at code clinics.
            if  (tmp .ge. thres) then
               tmp = 0.0
            else
               tmp = acos( tmp)
            endif

c     (i,j) has a distance smaller than eta.

c     In case nnz was too small, recall line to get a better estimate
            if( jja .gt. nnz) then
               iflag = i
               goto 20
            endif

            colindices(jja) = j
            entries(jja) = tmp*p
            jja = jja + 1


 10      continue
         rowpointers(i+1)=jja
      enddo


      if (part.gt.0) then
         rowpointers(nx+1)=jja
      endif
      nnz=jja-1
 20   continue

      return
      end

# HEADER ####################################################
# This is file spam/R/dim.R.                                #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# This is the actual dim...

"dim<-.spam" <- function(x, value) {
    if (is.spam(x)) {

        dimx <- x@dimension
        pdim <- prod(dimx)
        vlen <- prod(value)
        if( !identical(pdim,vlen))
            stop( sprintf("dims [product %d] do not match the length of object [%d]. Do you want `pad`",
                          pdim,vlen))

        if (length(value)>2)
            stop("dims should be of length 1 or 2")
        if (identical(length(value),1L))
            return( c(x) )

        if(any(dimx<1))
            stop("the dims contain negative values")

        tmp <- cbind(st=rep(1:dim(x)[1],diff(x@rowpointers)), nd=x@colindices)
        ind <- tmp[,1]+(tmp[,2]-1)*dimx[1] - 1

        slist <- list(i = ind%%value[1]   +1,
                       j = ind%/%value[1] +1,
                       x@entries)

        return( spam.list( slist, nrow=value[1], ncol=value[2],
                          eps = .Machine$double.eps))


    } else  {
        dim(x) <- value
        x
    }
}


########################################################################
# dim and derivatives

"pad<-.spam" <- function(x,value) {

    force64 <- getOption("spam.force64")

    # check if value is valid
    if ( (min(value)<1 ) || any(!is.finite(value)))
        stop("dims should be postive integers.")
    if (!identical( length(value), 2L))
        stop("dims should be of length 2.")

    dimx <- x@dimension
    last <- value[1]+1

  # In three steps:
  #  1) Address col truncation
            # to safe time, we also take into account if we have fewer or equal rows
  #  2) Augment rows
  #  3) if fewer rows and more columns, truncate
  # In any case, dimensions are fixed at the end.


  # If fewer cols required, we run reducedim
  if (dimx[2]>value[2]){

#     subroutine reducedim(a,ja,ia,eps,bnrow,bncol,k,b,jb,ib)

      if( force64 || .format.spam(x)$package == "spam64")
          SS <- .format64()
      else
          SS <- .format32

      z <- .C64("reducedim",
                SIGNATURE=c("double", SS$signature, SS$signature,
                    "double", SS$signature, SS$signature, SS$signature,
                    "double", SS$signature, SS$signature),
                oldra = x@entries,
                oldja = x@colindices,
                oldia = x@rowpointers,

                eps = getOption("spam.eps"),
                min(value[1],dimx[1]),
                value[2],
                nz = 1,

                entries=vector_dc("double",length(x@entries)),
                colindices=vector_dc(SS$type,length(x@entries)),
                rowpointers=vector_dc(SS$type,last),

                INTENT=c("r", "r", "r",
                         "r", "r", "r", "w",
                         "w", "w", "w"),
                NAOK = getOption("spam.NAOK"),
                PACKAGE = SS$package)

    if (z$nz==1 ){ #was identical( z$nz,1L)
        ## print("2")
            return(
                .newSpam(
                    # entries=x@entries,
                    # colindices=x@colindices,
                    # rowpointers=c(1,rep_len64(2,value[1])),
                    dimension=value,
                    force64=force64
                    )
                )
    }
      nz <- z$nz-1
      x <- .newSpam(
          entries=z$entries[1:nz],
          colindices=z$colindices[1:nz],
          rowpointers=z$rowpointers[1:min(last,dimx[1]+1)],
          dimension=value, #actually here dim 2 = value 2 but dim1 maybe not yet
          force64=force64
          )
  }
    # augment rows
  if  (dimx[1]<value[1]){
      ## print("3")
      x <- .newSpam(
              entries=x@entries,
              colindices=x@colindices,
              rowpointers= c( x@rowpointers,
                  rep_len64( x@rowpointers[length(x@rowpointers)],value[1]-dimx[1])),
              dimension=value,
              force64=force64
              )
  }

    # special case: fewer rows and more columns, truncate
  if((dimx[1]>=value[1])&(dimx[2]<=value[2])) { ## added =, think about it again
      ## print("4")
      lastelement <- (x@rowpointers[last]-1)

      x <- .newSpam(
          entries= x@entries[1:lastelement],
          colindices= x@colindices[1:lastelement],
          rowpointers= x@rowpointers[1:last],
          dimension=value,
          force64=force64
          )
  }
    #before dim x = value x was here with slot option
  return(x)

}



setMethod("dim",   "spam", function(x) x@dimension )
setMethod("dim<-",   "spam", get("dim<-.spam"))

setGeneric("pad<-", function(x, value) standardGeneric("pad<-"))
setMethod("pad<-",   "spam", get("pad<-.spam"))
setMethod("pad<-",   "matrix",
          function(x, value) {
              if (!identical( length(value), 2L)) stop("dims should be of length 2.")
              tmp <- matrix(0, value)
              mr <- 1:min(value[1], dim(x)[1])
              mc <- 1:min(value[2], dim(x)[2])
              tmp[mr,mc] <- x[mr,mc]
              return(tmp)
          })

# HEADER ####################################################
# This is file spam/tests/demo_timing.R.                    #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


# We construct a few large matrices and we compare how much faster (slower)
# we are compared to the full matrix analysis.
# Since all the calculation are also done with full matrices, we do not
# exagerate with the sizes.

options( echo=FALSE)
library( spam, warn.conflict=FALSE)

set.seed(14)


# In the test function, we do not print out the actual times
# We would get too many differences pointed out!
compare <- function(expr1,expr2,tag=NULL)
  {
    if( !is.null(tag)) cat( "Comparing: ", tag, fill=TRUE)
    invisible(data.frame(full=system.time( expr1, TRUE)[1:3],
                     sparse=system.time( expr2, TRUE)[1:3],
                     row.names=c("user","system","elapsed")))
  }

xn <- 10
xm <- 12

# first start with a full matrix.
fmat1 <- matrix(rnorm(xn*xm),xn,xm)
smat1 <- as.spam(fmat1)

compare(fmat2 <- t(fmat1), smat2 <- t(smat1), "Transpose")

compare(ffmat <- fmat1 %*% fmat2,
        ssmat <- smat1 %*% smat2, "multiplication")

compare( solve(ffmat),  solve(ssmat), "solving")


compare(rbind(fmat1,fmat1),rbind(smat1,smat1))
compare(cbind(fmat1,fmat1),cbind(smat1,smat1))





# now create a sparse matrix.
fmat1[fmat1<3] <- 0
smat1 <- as.spam(fmat1)



compare(fmat2 <- t(fmat1), smat2 <- t(smat1), "Transpose")

compare(ffmat <- fmat1 %*% fmat2,
        ssmat <- smat1 %*% smat2, "multiplication")

compare(ffmat <- ffmat + diag(xn),
        ssmat <- ssmat + diag.spam(xn), "add identity")

compare(ffmat <- 1:xn %d+% ffmat,
        ssmat <- 1:xn %d+% ssmat, "add identity quicker")

compare( solve(ffmat),  solve(ssmat), "solving")

summary(ssmat)


# compare a few cbind/rbinds

compare(rbind(fmat1,fmat1),rbind(smat1,smat1))
compare(cbind(fmat1,fmat1),cbind(smat1,smat1))


options( echo=TRUE)





# illustrate the new spam x matrix multiply:
if (F){
  n <- 1000
  
  A <- spam(0,n,n)
  A[cbind(1:(n-1),2:n)] <- -c(2,rep(4,n-3),2)
  A[cbind(1:(n-2),3:n)] <- rep(1,n-2)
  A <- A + t( A)
  diag(A) <- c(1,5,rep(6,n-4),5,1)


  B <- array(rnorm(n*n),c(n,n))

  system.time(C1 <- .spam.matmul.mat(A,B))
  system.time(C2 <- .spam.matmul(A,B))
  norm(C1-C2)


}

      subroutine amuxmat (n,m,p, x, y, a,ja,ia)
      implicit none
      integer n, m, p, ja(*), ia(*)
      double precision  x(m,p), y(n,p), a(*)
c-----------------------------------------------------------------------
c Multiplies a sparse matrix by a full matrix using consecutive dot
c products, cf. subroutine amux from sparse kit.
c Matrix A is stored in compressed sparse row storage.
c
c on entry:
c----------
c n     = row dimension of A
c p     = column dimension of x
c x     = array of dimension mxp, m column dimension of A.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = array of dimension nxp, containing the product y=Ax
c
c Reinhard Furrer
c-----------------------------------------------------------------------
c local variables
c
      double precision t
      integer j, i, k
c-----------------------------------------------------------------------
      do j = 1,p
         do i = 1,n
c
c     compute the inner product of row i with vector x
c
            t = 0.0d0
            do  k=ia(i), ia(i+1)-1
               t = t + a(k)*x(ja(k),j)
            enddo
c
            y(i,j) = t
         enddo
      enddo
c
      return
c---------end-of-amuxmat------------------------------------------------
c-----------------------------------------------------------------------
      end
c

      subroutine notzero (ja,ia,nrow,ncol,nnz,nz,jao,iao)
c Return the structure of the zero entries in ra,ja,ia, in
c  compressed sparse row format via rao, jao, iao.
c INPUT:
c     ja, ia -- sparse structure of the matrix A
c     nrow -- number of rows in `a'
c     ncol -- number of columns in `a'
c     nnz -- number of non-zero elements
c     nz -- number of zero elements
c OUTPUT:
c     jao, iao --  sparse structure of the zero entries
c WORK ARRAY:
c     colmn -- logical vector of length ncol

      implicit none
      integer  nrow,ncol,nnz,nz,inz,
     &       ja(nnz),ia(nrow+1),jao(nz),iao(nrow+1)
      logical colmn(ncol)
      integer i,j,k
      inz = 0
      iao(1) = 1
      do i = 1,nrow
         iao(i+1) = iao(i)
         do k = 1,ncol
            colmn(k) = .true.
         enddo
         do j = ia(i),ia(i+1)-1
            colmn(ja(j)) = .false.
         enddo
         do k = 1,ncol
            if(colmn(k)) then
               inz = inz + 1
               jao(inz) = k
               iao(i+1) = iao(i+1) + 1
            endif
         enddo
      enddo
      return
      end


      subroutine setdiagmat (nrow, n, a, ja, ia, diag, iw)
      implicit none
      integer nrow, n
      double precision a(*),  diag(n)
      integer ja(*), ia(nrow+1), iw(nrow)
c-----------------------------------------------------------------------
c Sets the diagonal entries of a sparse matrix
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A
c n = integer. Smallest dimension of A
c
c a, ja, ia   = Matrix A in compressed sparse row format. Sorted.
c diag = diagonal matrix stored as a vector diag(1:n)
c iw   = n vector of zeros.
c
c on return:
c----------
c updated matrix A
c iw    = iw contains the positions of the diagonal entries in the
c         output matrix. (i.e., a(iw(k)), ja(iw(k)), k=1,...n,
c         are the values/column indices of the diagonal elements
c         of the output matrix. ).
c
c Reinhard Furrer
c-----------------------------------------------------------------
      logical insert
      integer i,j, k, k1, k2, icount

c
c     get positions of diagonal elements in data structure.
c
      do  11 i=1,n
         do 21 j= ia(i),ia(i+1)-1
            if (ja(j) .ge. i) then
               if (ja(j) .eq. i) then
                  iw(i) = j
               endif
               goto 11
            endif
 21      continue
 11   continue
c
c     count number of holes in diagonal and add diag(*) elements to
c     valid diagonal entries.
c
      icount = 0
      do 31 i=1, n
         if (iw(i) .eq. 0) then
            icount = icount+1
         else
            a(iw(i)) = diag(i)
         endif
 31      continue
c
c     if no diagonal elements to insert return
c
      if (icount .eq. 0) return
c
c     shift the nonzero elements if needed, to allow for created
c     diagonal elements.
c
c
c     copy rows backward
c
      do 5 i=nrow, 1, -1
c
c     go through  row ii
c
         k1 = ia(i)
         k2 = ia(i+1)-1

         ia(i+1) = ia(i+1)+icount

         if ((i .gt. n) .or. (iw(i) .gt. 0)) then
c     iw(ii) equal to 0, means no diagonal element in a, we need to insert it
c     test is thus true.

c     no fill-in, only copying
            do 4 k = k2,k1,-1
               ja(k+icount)=ja(k)
               a(k+icount)=a(k)
 4          continue
            iw(i)=-i
         else
            insert=.TRUE.
            if (k2.lt.k1) then
               ja(k2+icount)=i
               a(k2+icount)=diag(i)
               iw(i)=k2+icount
               icount=icount-1
               insert = .FALSE.
               if (icount .eq. 0) return
            else
               do 6 k = k2,k1,-1
                  if (ja(k).gt. i) then
                     ja(k+icount)=ja(k)
                     a(k+icount)=a(k)
                  else  if  (insert) then
                     ja(k+icount)=i
                     a(k+icount)=diag(i)
                     iw(i)=k+icount
                     icount=icount-1
                     insert = .FALSE.
                     if (icount .eq. 0) return
                  endif
                  if (ja(k).lt. i) then
                     ja(k+icount)=ja(k)
                     a(k+icount)=a(k)
                  endif
 6             continue
c     in case there is only one element, larger than i, we still need to
c     add the diagonal element
               if  (insert) then
                   ja(k+icount)=i
                   a(k+icount)=diag(i)
                   iw(i)=k+icount
                   icount=icount-1
                   insert = .FALSE.
                   if (icount .eq. 0) return
                endif
            endif
         endif
 5    continue
      return
c-----------------------------------------------------------------------
c------------end-of-diagaddmat------------------------------------------
      end
      subroutine diagaddmat (nrow, n, a, ja, ia, diag, iw)
      implicit none
      integer nrow, n
      double precision a(*),  diag(n)
      integer ja(*), ia(nrow+1), iw(nrow)
c-----------------------------------------------------------------------
c Adds a diagonal matrix to a sparse matrix:  A = Diag + A
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A
c n = integer. Smallest dimension of A
c
c a, ja, ia   = Matrix A in compressed sparse row format. Sorted.
c diag = diagonal matrix stored as a vector diag(1:n)
c iw   = n vector of zeros.
c
c on return:
c----------
c updated matrix A
c iw    = iw contains the positions of the diagonal entries in the
c         output matrix. (i.e., a(iw(k)), ja(iw(k)), k=1,...n,
c         are the values/column indices of the diagonal elements
c         of the output matrix. ).
c
c Reinhard Furrer
c-----------------------------------------------------------------
      logical insert
      integer i,j, k, k1, k2, icount

c
c     get positions of diagonal elements in data structure.
c
      do  11 i=1,n
         do 21 j= ia(i),ia(i+1)-1
            if (ja(j) .ge. i) then
               if (ja(j) .eq. i) then
                  iw(i) = j
               endif
               goto 11
            endif
 21      continue
 11   continue
c
c     count number of holes in diagonal and add diag(*) elements to
c     valid diagonal entries.
c
      icount = 0
      do 31 i=1, n
         if (iw(i) .eq. 0) then
            icount = icount+1
         else
            a(iw(i)) = a(iw(i)) + diag(i)
         endif
 31      continue
c
c     if no diagonal elements to insert return
c
      if (icount .eq. 0) return
c
c     shift the nonzero elements if needed, to allow for created
c     diagonal elements.
c
c
c     copy rows backward
c
      do 5 i=nrow, 1, -1
c
c     go through  row ii
c
         k1 = ia(i)
         k2 = ia(i+1)-1

         ia(i+1) = ia(i+1)+icount

         if ((i .gt. n) .or. (iw(i) .gt. 0)) then
c     iw(ii) equal to 0, means no diagonal element in a, we need to insert it
c     test is thus true.

c     no fill-in, only copying
            do 4 k = k2,k1,-1
               ja(k+icount)=ja(k)
               a(k+icount)=a(k)
 4          continue
            iw(i)=-i
         else
            insert=.TRUE.
            if (k2.lt.k1) then
               ja(k2+icount)=i
               a(k2+icount)=diag(i)
               iw(i)=k2+icount
               icount=icount-1
               insert = .FALSE.
               if (icount .eq. 0) return
            else
               do 6 k = k2,k1,-1
                  if (ja(k).gt. i) then
                     ja(k+icount)=ja(k)
                     a(k+icount)=a(k)
                  else  if  (insert) then
                     ja(k+icount)=i
                     a(k+icount)=diag(i)
                     iw(i)=k+icount
                     icount=icount-1
                     insert = .FALSE.
                     if (icount .eq. 0) return
                  endif
                  if (ja(k).lt. i) then
                     ja(k+icount)=ja(k)
                     a(k+icount)=a(k)
                  endif
 6             continue
c     in case there is only one element, larger than i, we still need to
c     add the diagonal element
               if  (insert) then
                   ja(k+icount)=i
                   a(k+icount)=diag(i)
                   iw(i)=k+icount
                   icount=icount-1
                   insert = .FALSE.
                   if (icount .eq. 0) return
                endif
            endif
         endif
 5    continue
      return
c-----------------------------------------------------------------------
c------------end-of-setdiagmat------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine diagmua (nrow, a, ia, diag)
      implicit none
      integer          nrow, ia(nrow+1)
      double precision a(*),  diag(nrow), scal
c-----------------------------------------------------------------------
c performs the matrix by matrix product A = Diag * A  (in place)
c (diamua from sparsekit provides more functionality)
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A
c a, ia   = Matrix A in compressed sparse row format.
c           (ja is not needed)
c
c diag = diagonal matrix stored as a vector diag(1:n)
c
c on return:
c----------
c a, 	= resulting matrix A in compressed sparse row sparse format.
c
c Notes:
c-------
c     Reinhard Furrer 2007-06-21
c
c-----------------------------------------------------------------
c     local variables
      integer          ii, k, k1, k2

      do 1 ii=1,nrow
c
c     normalize each row
c
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         scal = diag(ii)
         do 2 k=k1, k2
            a(k) = a(k)*scal
 2       continue
 1    continue
c
      return
c----------end-of-diagmua------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------

c-----------------------------------------------------------------------
      subroutine getdiag (a,ja,ia,len,diag)

      implicit none
      double precision diag(*),a(*)
      integer len, ia(*), ja(*)
c-----------------------------------------------------------------------
c This subroutine extracts the main diagonal.
c (getdia from sparsekit provides more functionality)
c-----------------------------------------------------------------------
c
c on entry:
c----------
c
c len= min(nrow, ncol) = min dimension of the matrix a.
c a,ja,ia = matrix stored in sorted compressed sparse row a,ja,ia,format
c diag  = array of zeros.
c
c on return:
c-----------
c diag  = array of length containing the wanted diagonal.
c
c Notes:
c-------
c     Reinhard Furrer 2006-11-02
c----------------------------------------------------------------------c
c     local variables
      integer i, k
c
c     extract  diagonal elements
c
      do 1 i=1, len
         do k= ia(i),ia(i+1) -1
            if (ja(k) .ge. i) then
c     we are at or beyond the diagonal.
               if (ja(k) .eq. i) then
                  diag(i)= a(k)
               endif
               goto 1
            endif
         enddo
 1    continue
      return
c------------end-of-getdiag----------------------------------------------
c-----------------------------------------------------------------------
      end




c Functions that are new or  modified.

      subroutine subsparsefull(nrow,a,ja,ia,b)
c
c     subtracts a sparse matrix from a full one
c     algorithm is in-place, i.e. b is changed
c
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-21
c-----------------------------------------------------------------------

      implicit none
      integer nrow,ja(*),ia(nrow+1)
      double precision a(*), b(nrow,*)

      integer i,k


      do i=1,nrow
         do k=ia(i),ia(i+1)-1
            b(i,ja(k)) = b(i,ja(k))-a(k)
         enddo
      enddo
      return
      end

      subroutine subfullsparse(nrow,ncol,a,ja,ia,b)
c
c     subtracts a full matrix from a sparse one
c     algorithm is in-place, i.e. b is changed
c
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-21
c-----------------------------------------------------------------------

      implicit none
      integer nrow,ncol,ja(*),ia(nrow+1)
      double precision a(*), b(nrow,*)

      integer i,j,k


      do i=1,nrow
         do j=1,ncol
            b(i,j) = -b(i,j)
         enddo
         do k=ia(i),ia(i+1)-1
            b(i,ja(k)) = b(i,ja(k))+a(k)
         enddo
      enddo
      return
      end

      subroutine addsparsefull(nrow,a,ja,ia,b)
c
c     adds a sparse matrix to a full one
c     algorithm is in-place, i.e. b is changed
c
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-21
c-----------------------------------------------------------------------

      implicit none
      integer nrow,ja(*),ia(nrow+1)
      double precision a(*), b(nrow,*)

      integer i,k


      do i=1,nrow
         do k=ia(i),ia(i+1)-1
            b(i,ja(k)) = b(i,ja(k))+a(k)
         enddo
      enddo
      return
      end

      subroutine constructia(nrow,nir,ia,ir)
c
c     constructs from a regular row index vector a sparse ia vector.
c     note that a regular column index vector corresponds to the
c     sparse ja vector. for example:
c         A[ir,jc] =>  A@ja = jc, A@ia = constructia(nrow,nir,ia,ir)$ia
c
c     nrow: row dimension of A
c     nir:  length of ir
c     ir:   array of length nir+1!!!
c
c Notes:
c-------
c     _*Row indices have to be ordered!*_
c
c     Reinhard Furrer 2006-09-13
c-----------------------------------------------------------------------

      implicit none
      integer nrow,nir
      integer ia(nrow+1),ir(*)

      integer i,k

      k=1
      ia(1)=1
      do i=1,nrow
 5       continue
         if (ir(k) .eq. i) then
            k=k+1
            if (k .le. nir) goto 5
         endif
         ia(i+1)=k
      enddo

      ia(nrow+1)=nir+1

      return
      end





      subroutine disttospam(nrow,x,a,ja,ia,eps)

      implicit none
      integer nrow, ia(nrow+1), ja(*)
      double precision x(*), a(*), eps
c
c     Convertion of an R dist object (removes zero entries as well).
c
c On entry:
c----------
c     nrow    -- row dimension of the matrix
c     x       -- elements of the dist object (is lower diagonal)
c                        n*(i-1) - i*(i-1)/2 + j-i  for i < j
c
c     a,ja,ia -- input matrix in CSR format
c
c On return:
c-----------
c     a,ja,ia -- cleaned matrix
c
c Notes:
c-------
c     Reinhard Furrer 2008-08-13
c-----------------------------------------------------------------------
c
c     Local
      integer i,j,k, tmp


      ia(1) = 1
      k = 1
      do i = 2, nrow
         ia(i) = k
         do j=1 , i-1
            tmp = nrow*(j-1)-j*(j-1)/2+i-j
            if (.not.(dabs(x(tmp)) .le. eps)) then
               ja(k) = j
               a(k) = x(tmp)
               k = k + 1
            endif

         enddo
      enddo

      ia(nrow+1) = k
      return

c---- end of disttospam -------------------------------------------------
c-----------------------------------------------------------------------
      end

      subroutine setdiaold (nrow,ncol,a,ja,ia,c,jc,ic,cmax,diag,eps)

      implicit none

      double precision  a(*),c(*),diag(*),eps
      integer nrow, ncol, ia(*), ja(*), ic(*), jc(*), cmax
c
c     this routine sets the diagonal entries of a matix, provided they
c     are non-zero.
c
c On entry:
c----------
c     nrow,ncol    --  dimensions of the matrix
c     a,ja,ia -- input matrix in CSR format
c     c,jc,ic -- input matrix in CSR format with enough space, see below
c     diag -- diagonal values to set
c     eps  -- what is smaller than zero?
c
c On return:
c-----------
c     c,jc,ic -- matrix with modified diag in CSR format
c
c Notes:
c-------
c     Reinhard Furrer 2006-10-30
c-----------------------------------------------------------------------
c
c     Local
      double precision b(nrow)
      integer i,k, len, ib(nrow+1), jb(nrow)

c
      len=0
      ib(1)=1
      do i=1,nrow
         jb(i)=0
      enddo


      do 10 i=1,nrow
         do 15 k= ia(i),ia(i+1) -1
            if (ja(k) .eq. i) then
               a(k)=diag(i)
               c(k)=diag(i)
               ib(i+1)=ib(i)
               goto 10
            endif
            if (ja(k) .gt. i) then
               if (diag(i).gt.eps) then
                  len=len+1
                  jb(len)=i
                  ib(i+1)=ib(i)+1
                  b(len)=diag(i)
               else
                  ib(i+1)=ib(i)
               endif
               goto 10
            endif
 15      continue
 10   continue

      if (len .eq. 0) return
c
c     set nonexisiting elements.
c

      call subass(nrow,ncol,a,ja,ia,b,jb,ib,c,jc,ic,cmax)

      return
c------------end of setdia----------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------


c
c-----------------------------------------------------------------------
      subroutine subass(nrow,ncol,a,ja,ia,b,jb,ib,c,jc,ic,nzmax)
      implicit none
      integer nrow,ncol,nzmax
      integer ja(*),jb(*),jc(*),ia(*),ib(*),ic(*)
      double precision a(*), b(*), c(*)

c-----------------------------------------------------------------------
c replaces the elements of A with those of B for matrices in sorted CSR
c format. we assume that each row is sorted with increasing column
c indices.
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c
c a,ja,ia,
c b,jb,ib = Matrices A and B in compressed sparse row format with column
c           entries sorted ascendly in each row
c
c nzmax	= integer. The max length of the arrays c and jc.
c
c on return:
c----------
c c,jc,ic = resulting matrix C in compressed sparse row sparse format
c           with entries sorted ascendly in each row.
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-13, based on sparsekit2 subroutine aplb1
c-----------------------------------------------------------------------

c     local variables
      integer i,j1,j2,ka,kb,kc,kamax,kbmax

      kc = 1
      ic(1) = kc
c
c     looping over the rows:
      do 6 i=1, nrow
         ka = ia(i)
         kb = ib(i)
         kamax = ia(i+1)-1
         kbmax = ib(i+1)-1
 5       continue

c     If we have one or more entries then ka <= kamax
c     If we do not have any entries in both A and B
c     we will not enter the if clause. In which case
c     we repeatedly copy ic(i+1) <- ic(i).
         if (ka .le. kamax .or. kb .le. kbmax) then

c     j1 and j2 are left hand pointers of the first entry
c     of A and B. If no entry, they are set to ncol+1
            if (ka .le. kamax) then
               j1 = ja(ka)
            else
               j1 = ncol+1
            endif
            if (kb .le. kbmax) then
               j2 = jb(kb)
            else
               j2 = ncol+1
            endif
c
c     Three cases:
c       j1=j2: copy element of b in c, incr. all three pointers
c       j1<j2: copy element of a in c, incr. a and c pointers
c       j1>j2: copy element of b in c, incr. b and c pointers
            if (j1 .eq. j2) then
               c(kc) = b(kb)
               jc(kc) = j1
               ka = ka+1
               kb = kb+1
               kc = kc+1
            else if (j1 .lt. j2) then
               jc(kc) = j1
               c(kc) = a(ka)
               ka = ka+1
               kc = kc+1
            else if (j1 .gt. j2) then
               jc(kc) = j2
               c(kc) = b(kb)
               kb = kb+1
               kc = kc+1
            endif

C     the next four lines should not be required...
            if (kc .gt. nzmax+1) then
c               write (*,*) "exceeding array capacities...",i,nzmax,
c     & ka,kb,kc,j1,j2,kamax,kbmax,ncol,jb(kb)
               return
            endif
            goto 5
         endif

         ic(i+1) = kc
 6    continue
      return
c------------end-of-subass----------------------------------------------
c-----------------------------------------------------------------------
      end

      subroutine spamcsrdns(nrow,a,ja,ia,dns)

      implicit none
      integer i,k
      integer nrow,ja(*),ia(*)
      double precision dns(nrow,*),a(*)

c-----------------------------------------------------------------------
c Compressed Sparse Row    to    Dense
c-----------------------------------------------------------------------
c
c converts a row-stored sparse matrix into a densely stored one
c
c On entry:
c----------
c
c nrow  = row-dimension of a
c a,
c ja,
c ia    = input matrix in compressed sparse row format.
c         (a=value array, ja=column array, ia=pointer array)
c dns   = array where to store dense matrix
c
c on return:
c-----------
c dns   = the sparse matrix a, ja, ia has been stored in dns(nrow,*)
c
c changes:
c---------
c eliminated the ierr
c eliminated the filling of zeros: all done with
c-----------------------------------------------------------------------
      do i=1,nrow
         do  k=ia(i),ia(i+1)-1
            dns(i,ja(k)) = a(k)
         enddo
      enddo
      return
c---- end of csrdns ----------------------------------------------------
c-----------------------------------------------------------------------
      end



c-----------------------------------------------------------------------
      subroutine spamdnscsr(nrow,ncol,dns,ndns,a,ja,ia,eps)

      implicit none
      integer i,j,next
      integer nrow,ncol,ndns,ia(*),ja(*)
      double precision dns(ndns,*),a(*),eps
c-----------------------------------------------------------------------
c Converts a densely stored matrix into a CSR  sparse matrix.
c-----------------------------------------------------------------------
c on entry:
c---------
c
c nrow  = row-dimension of a
c ncol  = column dimension of a
c nzmax = maximum number of nonzero elements allowed. This
c         should be set to be the lengths of the arrays a and ja.
c dns   = input nrow x ncol (dense) matrix.
c ndns  = first dimension of dns.
c
c on return:
c----------
c
c a, ja, ia = value, column, pointer  arrays for output matrix
c
c changes:
c---------
c eliminated the ierr
c introduced epsilon
c-----------------------------------------------------------------------
      next = 1
      ia(1) = 1
      do  i=1,nrow
         do  j=1, ncol
            if (.not.(dabs(dns(i,j)) .le. eps)) then
               ja(next) = j
               a(next) = dns(i,j)
               next = next+1
            endif
         enddo
         ia(i+1) = next
      enddo
      return
c---- end of dnscsr ----------------------------------------------------
c-----------------------------------------------------------------------
      end

c-----------------------------------------------------------------------
      subroutine getmask(nrow,nnz,ir,jc,jao,iao)
c-----------------------------------------------------------------------
      implicit none
      integer nrow,nnz,ir(*),jc(*),jao(*),iao(*)
      integer k,k0,j,i,iad
c-----------------------------------------------------------------------
c  Gets Compressed Sparse Row indices from Coordinate ones
c-----------------------------------------------------------------------
c  Loosely based on coocsr from Sparsekit.
c
c on entry:
c---------
c nrow	= dimension of the matrix
c nnz	= number of nonzero elements in matrix
c ir,
c jc    = matrix in coordinate format. ir(k), jc(k) store the nnz
c         nonzero index. The order of the elements is arbitrary.
c iao   = vector of 0 of size nrow+1
c
c on return:
c-----------
c ir 	is destroyed
c
c jao, iao = matrix index in general sparse matrix format with
c       jao containing the column indices,
c	and iao being the pointer to the beginning of the row
c
c------------------------------------------------------------------------

c determine row-lengths.
      do 2 k=1, nnz
         iao(ir(k)) = iao(ir(k))+1
 2    continue
c starting position of each row..
      k = 1
      do 3 j=1,nrow+1
         k0 = iao(j)
         iao(j) = k
         k = k+k0
 3    continue
c go through the structure  once more. Fill in output matrix.
      do 4 k=1, nnz
         i = ir(k)
         j = jc(k)
         iad = iao(i)
         jao(iad) = j
         iao(i) = iad+1
 4    continue
c shift back iao
      do 5 j=nrow,1,-1
         iao(j+1) = iao(j)
 5    continue
      iao(1) = 1
      return
c-----------------------------------------------------------------------
      end


c-----------------------------------------------------------------------
      subroutine getblock(a,ja,ia, nrw, rw, ncl, cl, bnz, b,jb,ib)
c-----------------------------------------------------------------------
c     purpose:
c     --------
c     this function returns the elements a(rw,cl) of a matrix a,
c     for any index vector rw and cl. the matrix is assumed to be stored
c     in compressed sparse row (csr) format.
c
c
c     Reinhard Furrer 2006-09-12
c-----------------------------------------------------------------------
c     parameters:
c     -----------
c on entry:
c----------
c     a,ja,ia = the matrix a in compressed sparse row format (input).
c     nrw,rw
c     ncl,cl  = length of and the vector containing the rows and columns
c               to extract
c
c on return:
c-----------
c     bnz     = nonzero elements of b
c     b,jb,ib = the matrix a(rw,cl) in compressed sparse row format.
c
c note:
c------
c     no error testing is done. It is assumed that b has enough space
c     allocated.
c-----------------------------------------------------------------------
      implicit none

      integer nrw,rw(*),  ncl, cl(*)
      integer bnz, ia(*),ja(*), ib(*),jb(*)
      double precision a(*),b(*)
c
c     local variables.
c
      integer irw, jcl, jja
c
c      write(*,*) cl(1),cl(2)
      bnz = 1
      ib(1) = 1
      do irw = 1,nrw
         do jcl = 1,ncl

            do jja = ia(rw(irw)),ia(rw(irw)+1)-1

               if (cl(jcl) .eq. ja(jja)) then
c     we've found one...
                  b(bnz)  = a(jja)
                  jb(bnz) = jcl
                  bnz = bnz + 1
               endif
            enddo
         enddo
         ib(irw+1) = bnz
c     end irw, we've cycled over all lines
      enddo
      bnz = bnz - 1
c      write(*,*) cl(1),cl(2)

      return
c--------end-of-getblock------------------------------------------------
c-----------------------------------------------------------------------
      end


c-----------------------------------------------------------------------
      subroutine getlines(a,ja,ia, nrw, rw, bnz, b,jb,ib)
c-----------------------------------------------------------------------
c     purpose:
c     --------
c     this function returns the lines rw of a matrix a.
c     the matrix is assumed to be stored
c     in compressed sparse row (csr) format.
c
c
c     Reinhard Furrer 2012-04-04
c-----------------------------------------------------------------------
c     parameters:
c     -----------
c on entry:
c----------
c     a,ja,ia = the matrix a in compressed sparse row format (input).
c     nrw,rw  = length of and the vector containing the rows and columns
c               to extract
c
c on return:
c-----------
c     bnz     = nonzero elements of b
c     b,jb,ib = the matrix a(rw,cl) in compressed sparse row format.
c
c note:
c------
c     no error testing is done. It is assumed that b has enough space
c     allocated.
c-----------------------------------------------------------------------
      implicit none

      integer nrw,rw(*)
      integer bnz, ia(*),ja(*), ib(*),jb(*)
      double precision a(*),b(*)
c
c     local variables.
c
      integer irw, jja
c
      bnz = 1
      ib(1) = 1
      do irw = 1,nrw
         do jja = ia(rw(irw)),ia(rw(irw)+1)-1

            b(bnz)  = a(jja)
            jb(bnz) = ja(jja)
            bnz = bnz + 1
         enddo
         ib(irw+1) = bnz
c     end irw, we've cycled over all lines
      enddo
      bnz = bnz - 1

      return
c--------end-of-getlines------------------------------------------------
c-----------------------------------------------------------------------
      end



c-----------------------------------------------------------------------
      subroutine getelem(i,j,a,ja,ia,iadd,elem)
c-----------------------------------------------------------------------
c     purpose:
c     --------
c     this function returns the element a(i,j) of a matrix a,
c     for any pair (i,j).  the matrix is assumed to be stored
c     in compressed sparse row (csr) format. getelem performs a
c     binary search.
c     also returns (in iadd) the address of the element a(i,j) in
c     arrays a and ja when the search is successsful (zero if not).
c-----------------------------------------------------------------------
c     parameters:
c     -----------
c on entry:
c----------
c     i      = the row index of the element sought (input).
c     j      = the column index of the element sought (input).
c     a      = the matrix a in compressed sparse row format (input).
c     ja     = the array of column indices (input).
c     ia     = the array of pointers to the rows' data (input).
c on return:
c-----------
c     elem = value of a(i,j).
c     iadd   = address of element a(i,j) in arrays a, ja if found,
c              zero if not found. (output)
c
c     note: the inputs i and j are not checked for validity.
c-----------------------------------------------------------------------
c     noel m. nachtigal october 28, 1990 -- youcef saad jan 20, 1991.
c
c     Reinhard Furrer: converted to subroutine and eliminated sorted
c         many manipulations... last for 0.31; Sept 13
c-----------------------------------------------------------------------
      implicit none

      integer i, ia(*), iadd, j, ja(*)
      double precision a(*),elem
c
c     local variables.
c
      integer ibeg, iend, imid
c
c     initialization
c
      iadd = 0
      ibeg = ia(i)
      iend = ia(i+1)-1

c     empty line! test at beginning
 10   if (iend .lt. ibeg) return

c
c     begin binary search:
c     test of bounds
      if (ja(ibeg).gt.j) return
      if (ja(iend).lt.j) return

      if (ja(ibeg).eq.j) then
         iadd = ibeg
         goto 20
      endif
      if (ja(iend).eq.j) then
         iadd = iend
         goto 20
      endif

c     compute the middle index and test if found
      imid = ( ibeg + iend ) / 2
      if (ja(imid).eq.j) then
         iadd = imid
         goto 20
      endif

c     update the interval bounds.
      if (ja(imid).gt.j) then
         iend = imid -1
      else
         ibeg = imid +1
      endif

      goto 10
c
c     set iadd and elem before returning
 20   elem = a(iadd)
      return
c--------end-of-getelem-------------------------------------------------
c-----------------------------------------------------------------------
      end

      subroutine getallelem(nir,ir,jr,a,ja,ia,alliadd,allelem)
c-----------------------------------------------------------------------
c     purpose:
c     --------
c     wrapper to getelem to retrieve several elements.
c-----------------------------------------------------------------------
c     Reinhard Furrer 2006-09-12
c-----------------------------------------------------------------------
      implicit none

      integer nir,ir(nir),jr(nir),ja(*),ia(*),alliadd(nir)
      double precision a(*),allelem(nir)
c     local vars
      integer i
      do i = 1,nir
         call getelem(ir(i),jr(i),a,ja,ia,alliadd(i),allelem(i))
      enddo
      return
c--------end-of-allgetelem----------------------------------------------
c-----------------------------------------------------------------------
      end


c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c-
c- Modified by P. T. Ng from sparsekit
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c      subroutine aemub (nrow,ncol,a,ja,ia,amask,jmask,imask,
c     *                  c,jc,ic,iw,aw,nzmax,ierr)
       subroutine aemub (nrow,ncol,a,ja,ia,amask,jmask,imask,
     *                  c,jc,ic,nzmax,ierr)
c---------------------------------------------------------------------
      implicit none
      integer nrow, ncol
      real(8) a(*),c(*),amask(*),aw(ncol)
      integer ia(nrow+1),ja(*),jc(*),ic(nrow+1),jmask(*)
      integer imask(nrow+1), nzmax, ierr
      logical iw(ncol)
c-----------------------------------------------------------------------
c further used variables
      integer len, k, k1, k2, ii, j
c-----------------------------------------------------------------------
c Modified from amask by Pin T. Ng on 2/27/03 to perform
c element-wise multiplication
c-----------------------------------------------------------------------
c On entry:
c---------
c nrow  = integer. row dimension of input matrix
c ncol  = integer. Column dimension of input matrix.
c
c a,
c ja,
c ia    = the A matrix in Compressed Sparse Row format
c
c amask,
c jmask,
c imask = matrix defining mask stored in compressed
c         sparse row format. (This is the B matrix)
c
c nzmax = length of arrays c and jc. see ierr.
c
c On return:
c-----------
c
c a, ja, ia and amask, jmask, imask are unchanged.
c
c c
c jc,
c ic    = the output matrix in Compressed Sparse Row format.
c
c ierr  = integer. serving as error message.c
c         ierr = 1  means normal return
c         ierr .gt. 1 means that amask stopped when processing
c         row number ierr, because there was not enough space in
c         c, jc according to the value of nzmax.
c
c work arrays:
c-------------
c iw    = logical work array of length ncol.
c aw    = real work array of length ncol.
c
c note:
c------ the  algorithm is in place: c, jc, ic can be the same as
c a, ja, ia in which cas the code will overwrite the matrix c
c on a, ja, ia
c
c-----------------------------------------------------------------------
      ierr = 0
      len = 0
      do 1 j=1, ncol
         iw(j) = .false.
         aw(j) = 0.0
 1    continue
c     unpack the mask for row ii in iw
      do 100 ii=1, nrow
c     save pointer and value in order to be able to do things in place
         do 2 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .true.
            aw(jmask(k)) = amask(k)
 2       continue
c     add umasked elemnts of row ii
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         ic(ii) = len+1
         do 200 k=k1,k2
            j = ja(k)
            if (iw(j)) then
               len = len+1
               if (len .gt. nzmax) then
                  ierr = ii
                  return
               endif
               jc(len) = j
               c(len) = a(k)*aw(j)
            endif
 200     continue
c
         do 3 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .false.
            aw(jmask(k)) = 0.0
 3       continue
 100  continue
      ic(nrow+1)=len+1
c
      return
c-----end-of-aemub -----------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine aemub1 (nrow,ncol,a,ja,ia,b,jb,ib,c,jc,ic,
     *     nzmax,ierr)
      implicit none
      integer nrow, ncol, nzmax, ierr
      real(8) a(*), b(*), c(*)
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(nrow+1),ic(nrow+1)
c-----------------------------------------------------------------------
c A modification of aplsb by Pin Ng on 6/12/02 to
c perform the element-wise operation C = A*B for matrices in
c sorted CSR format.
c the difference with aplsb is that the resulting matrix is such that
c the elements of each row are sorted with increasing column indices in
c each row, provided the original matrices are sorted in the same way.
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format with entries sorted
c
c b,
c jb,
c ib	=  Matrix B in compressed sparse row format with entries sorted
c        ascendly in each row
c
c nzmax	= integer. The  length of the arrays c and jc.
c         amub will stop if the result matrix C  has a number
c         of elements that exceeds exceeds nzmax. See ierr.
c
c on return:
c----------
c c,
c jc,
c ic	= resulting matrix C in compressed sparse row sparse format
c         with entries sorted ascendly in each row.
c
c ierr	= integer. serving as error message.
c         ierr = 0 means normal return,
c         ierr .gt. 0 means that amub stopped while computing the
c         i-th row  of C with i=ierr, because the number
c         of elements in C exceeds nzmax.
c
c Notes:
c-------
c     this will not work if any of the two input matrices is not sorted
c-----------------------------------------------------------------------
      integer i, ka, kb, kc, kamax, kbmax, j1, j2
      ierr = 0
      kc = 1
      ic(1) = kc
c
c     the following loop does a merge of two sparse rows and
c     multiplies  them.
c
      do 6 i=1, nrow
         ka = ia(i)
         kb = ib(i)
         kamax = ia(i+1)-1
         kbmax = ib(i+1)-1
 5       continue
c
c     this is a while  -- do loop --
c
         if (ka .le. kamax .or. kb .le. kbmax) then
c
            if (ka .le. kamax) then
               j1 = ja(ka)
            else
c     take j1 large enough  that always j2 .lt. j1
               j1 = ncol+1
            endif
            if (kb .le. kbmax) then
               j2 = jb(kb)
            else
c     similarly take j2 large enough  that always j1 .lt. j2
               j2 = ncol+1
            endif
c
c     three cases
c
            if (j1 .eq. j2) then
               c(kc) = a(ka)*b(kb)
               jc(kc) = j1
               ka = ka+1
               kb = kb+1
               kc = kc+1
            else if (j1 .lt. j2) then
               ka = ka+1
            else if (j1 .gt. j2) then
               kb = kb+1
            endif
            if (kc .gt. nzmax) goto 999
            goto 5
c
c     end while loop
c
         endif
         ic(i+1) = kc
 6    continue
      return
 999  ierr = i
      return
c------------end-of-aemub1 ---------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine aedib (nrow,ncol,job,a,ja,ia,b,jb,ib,
     *     c,jc,ic,nzmax,iw,aw,ierr)
      implicit none
      integer nrow, ncol, job, nzmax, ierr
      real(8) a(*), b(*), c(*), aw(ncol)
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(nrow+1),ic(nrow+1),
     *     iw(ncol)
c-----------------------------------------------------------------------
c performs the element-wise matrix division  C = A/B.
c Modified from aplsb by Pin Ng on 2/27/03
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c job   = integer. Job indicator. When job = 0, only the structure
c                  (i.e. the arrays jc, ic) is computed and the
c                  real values are ignored.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format.
c
c b,
c jb,
c ib	=  Matrix B in compressed sparse row format.
c
c nzmax	= integer. The  length of the arrays c and jc.
c         amub will stop if the result matrix C  has a number
c         of elements that exceeds exceeds nzmax. See ierr.
c
c on return:
c----------
c c,
c jc,
c ic	= resulting matrix C in compressed sparse row sparse format.
c
c ierr	= integer. serving as error message.
c         ierr = 0 means normal return,
c         ierr .gt. 0 means that amub stopped while computing the
c         i-th row  of C with i=ierr, because the number
c         of elements in C exceeds nzmax.
c
c work arrays:
c------------
c iw	= integer work array of length equal to the number of
c         columns in A.
c aw	= real work array of length equal to the number of
c         columns in A.
c
c-----------------------------------------------------------------------
      logical values
      integer len, j, ii, ka, kb, k, jpos, jcol
      values = (job .ne. 0)
      ierr = 0
      len = 0
      ic(1) = 1
      do 1 j=1, ncol
         iw(j) = 0
 1    continue
c
      do 500 ii=1, nrow
c     row i
         do 200 ka=ia(ii), ia(ii+1)-1
            len = len+1
            jcol    = ja(ka)
            if (len .gt. nzmax) goto 999
            jc(len) = jcol
            if (values) c(len)  = a(ka)/0.0
            iw(jcol)= len
            aw(jcol) = a(ka)
 200     continue
c
         do 300 kb=ib(ii),ib(ii+1)-1
            jcol = jb(kb)
            jpos = iw(jcol)
            if (jpos .eq. 0) then
               len = len+1
               if (len .gt. nzmax) goto 999
               jc(len) = jcol
               if (values) c(len)  = 0.0
               iw(jcol)= len
            else
               if (values) c(jpos) = aw(jcol)/b(kb)
            endif
 300     continue
         do 301 k=ic(ii), len
            iw(jc(k)) = 0
 301     continue
         ic(ii+1) = len+1
 500  continue
      return
 999  ierr = ii
      return
c------------end of aedib -----------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine aeexpb (nrow,ncol,job,a,ja,ia,b,jb,ib,
     *     c,jc,ic,nzmax,iw,aw,ierr)
      implicit none
      integer nrow, ncol, job, nzmax, ierr
      real(8) a(*), b(*), c(*), aw(ncol)
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(nrow+1),ic(nrow+1),
     *     iw(ncol)
c-----------------------------------------------------------------------
c performs the element-wise matrix division  C = A/B.
c Modified from aplsb by Pin Ng on 2/27/03
c-----------------------------------------------------------------------
c on entry:
c ---------
c nrow	= integer. The row dimension of A and B
c ncol  = integer. The column dimension of A and B.
c job   = integer. Job indicator. When job = 0, only the structure
c                  (i.e. the arrays jc, ic) is computed and the
c                  real values are ignored.
c
c a,
c ja,
c ia   = Matrix A in compressed sparse row format.
c
c b,
c jb,
c ib	=  Matrix B in compressed sparse row format.
c
c nzmax	= integer. The  length of the arrays c and jc.
c         amub will stop if the result matrix C  has a number
c         of elements that exceeds exceeds nzmax. See ierr.
c
c on return:
c----------
c c,
c jc,
c ic	= resulting matrix C in compressed sparse row sparse format.
c
c ierr	= integer. serving as error message.
c         ierr = 0 means normal return,
c         ierr .gt. 0 means that amub stopped while computing the
c         i-th row  of C with i=ierr, because the number
c         of elements in C exceeds nzmax.
c
c work arrays:
c------------
c iw	= integer work array of length equal to the number of
c         columns in A.
c aw	= real work array of length equal to the number of
c         columns in A.
c
c-----------------------------------------------------------------------
      logical values
      integer len, kb, ka, k, jpos, jcol, j, ii
      values = (job .ne. 0)
      ierr = 0
      len = 0
      ic(1) = 1
      do 1 j=1, ncol
         iw(j) = 0
 1    continue
c
      do 500 ii=1, nrow
c     row i
         do 200 ka=ia(ii), ia(ii+1)-1
            len = len+1
            jcol    = ja(ka)
            if (len .gt. nzmax) goto 999
            jc(len) = jcol
            if (values) c(len)  = 1.0
            iw(jcol)= len
            aw(jcol) = a(ka)
 200     continue
c
         do 300 kb=ib(ii),ib(ii+1)-1
            jcol = jb(kb)
            jpos = iw(jcol)
            if (jpos .eq. 0) then
               len = len+1
               if (len .gt. nzmax) goto 999
               jc(len) = jcol
               if (values) c(len)  = 0.0**b(kb)
               iw(jcol)= len
            else
               if (values) c(jpos) = aw(jcol)**b(kb)
            endif
 300     continue
         do 301 k=ic(ii), len
            iw(jc(k)) = 0
 301     continue
         ic(ii+1) = len+1
 500  continue
      return
 999  ierr = ii
      return
c------------end of aeexpb -----------------------------------------------
c-----------------------------------------------------------------------
      end



      SUBROUTINE CALCJA(nrow,nsuper,
     %     xsuper,lindx,xlindx,xlnz,
     %     cholcja)
c     small function to calculate ja for the cholesky factor
c     as they use a condensed format. GRATULIERU LIT!


c     INPUT:
c     nrow (integer)           number of rows
c     nsuper (integer)         number of supernodes
c     xsuper (integer)         supernode partition
c     xlindx,lindx  (integer)  row indices for each supernode
c     xlnz (integer)           ia for cholesky factor
c
c     OUTPUT:
c     cholcja (integer)        ja for cholesky factor

      IMPLICIT NONE

      INTEGER nrow,nsuper
      INTEGER xsuper(nrow),lindx(*),xlindx(nrow+1),xlnz(nrow+1)
      INTEGER cholcja(*)

      INTEGER k, i, j, m, n

      k=1
      m=1
      DO i=1,nsuper
         DO j=1,( xsuper(i+1)-xsuper(i))
            DO n=1,(xlnz(k+1)-xlnz(k))
               cholcja(m)=lindx( xlindx(i)+j-2  + n)
               m=m+1
            ENDDO
            k=k+1
         ENDDO
      ENDDO

      RETURN
      END






      subroutine transpose(n,m,a,ja,ia,ao,jao,iao)

      implicit none
      integer n,m,ia(n+1),iao(m+1),ja(*),jao(*)
      double precision  a(*),ao(*)


      integer i,j,k,next
c-----------------------------------------------------------------------
c     Transposition
c     similar to csrcsc from sparsekit
c-----------------------------------------------------------------------
c on entry:
c----------
c n	= number of rows of CSR matrix.
c m    = number of columns of CSC matrix.
c a	= real array of length nnz (nnz=number of nonzero elements in input
c         matrix) containing the nonzero elements.
c ja	= integer array of length nnz containing the column positions
c 	  of the corresponding elements in a.
c ia	= integer of size n+1. ia(k) contains the position in a, ja of
c	  the beginning of the k-th row.
c
c on return:
c ----------
c ao	= real array of size nzz containing the "a" part of the transpose
c jao	= integer array of size nnz containing the column indices.
c iao	= integer array of size n+1 containing the "ia" index array of
c	  the transpose.
c
c-----------------------------------------------------------------------
c----------------- compute lengths of rows of transp(A) ----------------
      do  i=1, n
         do  k=ia(i), ia(i+1)-1
            j = ja(k)+1
            iao(j) = iao(j)+1
         enddo
      enddo
c---------- compute pointers from lengths ------------------------------
      iao(1) = 1
      do  i=1,m
         iao(i+1) = iao(i) + iao(i+1)
      enddo
c--------------- now do the actual copying -----------------------------
      do  i=1,n
         do  k=ia(i),ia(i+1)-1
            j = ja(k)
            next = iao(j)
            ao(next) = a(k)
            jao(next) = i
            iao(j) = next+1
         enddo
      enddo
c-------------------------- reshift iao and leave ----------------------
      do  i=m,1,-1
         iao(i+1) = iao(i)
      enddo
      iao(1) = 1
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------

c-----------------------------------------------------------------------



      subroutine reducedim(a,ja,ia,eps,bnrow,bncol,k,b,jb,ib)

      implicit none
      double precision a(*),b(*),eps
      integer bnrow, bncol,k
      integer ia(*),ja(*),ib(*),jb(*)

      integer   i, j, jaj

c-----------------------------------------------------------------------
c  Reduces the dimension of A to (,bnrow,bncol) by copying it to B.
c  (Hence not in place - for R purposes).
c  Only elements smaller than eps are copied.
c-----------------------------------------------------------------------
c on entry:
c---------
c
c------------------------------------------------------------------------

      k=1
      do i=1,bnrow
         ib(i)=k
         do j=ia(i), ia(i+1)-1
            jaj=ja(j)
            if (jaj .le.bncol) then
               if (abs( a(j)) .gt. eps) then
                  b(k)=a(j)
                  jb(k)=jaj
                  k=k+1
               endif
            endif
         enddo
      enddo
      ib(bnrow+1)=k
      return
c-----------------------------------------------------------------------
      end

c-----------------------------------------------------------------------
c Currently not used...
      subroutine reducediminplace(eps,nrow,ncol,k,a,ja,ia)

      implicit none
      double precision a(*),eps
      integer nrow, ncol,k
      integer ia(*),ja(*)

      integer   i, j, jj, itmp

c-----------------------------------------------------------------------
c  Reduces the dimension of A to (nrow,ncol) _in place_
c  Only elements smaller than eps are copied.
c-----------------------------------------------------------------------
c  Reinhard Furrer, June 2008
c------------------------------------------------------------------------

      k=1
      do i=1,nrow
         itmp = ia(i)
         ia(i)=k
         do j=itmp, ia(i+1)-1
            jj=ja(j)
            if (jj .le. ncol) then
               if (abs( a(jj)) .gt. eps) then
                  a(k)=a(jj)
                  ja(k)=jj
                  k=k+1
               endif
            endif
         enddo
      enddo
      ia(nrow+1)=k
      return
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------



c----------------------------------------------------------------------c
c      T R I A N G U L A R    S Y S T E M    S O L U T I O N S         c
c                                                                      c
c      spamforward and spamback                                        c
c----------------------------------------------------------------------c

      subroutine spamforward (n,p,x,b,l,jl,il)
      implicit none

      integer n, p, jl(*),il(n+1)
      double precision  x(n,p), b(n,p), l(*)

      integer i, j, k
      double precision  t

c-----------------------------------------------------------------------
c   solves    L x = y ; L = lower triang. /  CSR format
c                        sequential forward elimination
c-----------------------------------------------------------------------
c
c On entry:
c----------
c n,p      = integer. dimensions of problem.
c b      = real array containg the right side.
c
c l, jl, il,    = Lower triangular matrix stored in CSR format.
c
c On return:
c-----------
c       x  = The solution of  L x  = b.
c--------------------------------------------------------------------
c     Reinhard Furrer June 2008, April 2012, Sept 2016, Nov 2019


c     initialize k for next setting
      k=0
c     if first diagonal element is zero, break
      if (abs(l(1)) .le. 0.D0) goto 5

c     cycle over all columns of b
      do i=1,p

c     first row has one element then cycle over all rows
         x(1,i) = b(1,i) / l(1)
         do 3 k = 2, n
            t = b(k,i)
            do 1 j = il(k), il(k+1)-1
               if (jl(j) .lt. k) then
                  t = t-l(j)*x(jl(j),i)
               else
                  if (jl(j) .eq. k) then
                     if (abs(l(j)) .le. 0.D0) goto 5
c     diagonal element is not zero, hence we divide and leave the loop
                     x(k,i) = t / l(j)
                     goto 3
                  endif
               endif
 1          continue
 3       continue
      enddo

      return
 5    n = -k
      return
      end
c-----------------------------------------------------------------------
      subroutine spamback (n,p,x,b,r,jr,ir)
      implicit none

      integer n, p, jr(*),ir(n+1)
      double precision  x(n,p), b(n,p), r(*)

      integer l, j, k
      double precision  t
c-----------------------------------------------------------------------
c             Solves   R x = b    R = upper triangular.
c-----------------------------------------------------------------------
c
c On entry:
c----------
c n,p      = integers. dimension of problem.
c b        = real array containg the right side.
c
c r, jr, ir,    = Upper triangular matrix stored in CSR format.
c
c On return:
c-----------
c       x = The solution of  R x = b .
c--------------------------------------------------------------------
c     Reinhard Furrer June 2008, April 2012, Sept 2016, Nov 2019

      k = n+1
      if (abs( r(ir(n+1)-1) ) .le. 0.D0) goto 5
      do l=1,p
         x(n,l) = b(n,l) / r(ir(n+1)-1)
         do 3 k = n-1,1,-1
            t = b(k,l)
            do 1  j = ir(k+1)-1,ir(k),-1
               if (jr(j) .gt. k) then
                  t = t - r(j)*x(jr(j),l)
               else
                  if (jr(j) .eq. k) then
                     if (abs(r(j)) .le. 0.D0) goto 5
c     diagonal element is not zero, hence we divide and leave the loop
                     x(k,l) = t / r(j)
                     goto 3
                  endif
               endif
 1          continue
 3       continue
      enddo
      return
 5    n = -k
      return
      end
c-----------------------------------------------------------------------

# HEADER ####################################################
# This is file spam/R/s4coerce.R.                           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



# a few coercions that make sense...

#      showMethods(coerce)

setAs("spam","logical", def=function(from) {
    if(getOption("spam.structurebased")) {
        return( as.logical(from@entries))     
    }else{
        inefficiencywarning( gettextf("This operation may be inefficient"), prod(dim(from)))
        return( as.logical(as.matrix(from)))
    }})

setAs("spam","vector", def=function(from) {
    if(getOption("spam.structurebased")) {
        return( as.vector(from@entries))     
    }else{
        inefficiencywarning( gettextf("This operation may be inefficient"), prod(dim(from)))
        return( as.vector(as.matrix(from)))
    }})

setAs("spam","integer", def=function(from) {
    if(getOption("spam.structurebased")) {
        return( as.integer(from@entries))     
    }else{
        inefficiencywarning( gettextf("This operation may be inefficient"), prod(dim(from)))
        return( as.integer(as.matrix(from)))
    }})

setAs("spam","matrix", def=function(from) {
    inefficiencywarning( gettextf("This operation may be inefficient"), prod(dim(from)))
    return( as.logical(as.matrix(from)))
})
setAs("spam","list", def=function(from) {
    return( triplet(from))
})

      subroutine triplet3csr(nrow,ncol,nnz,a,ir,jc,ao,jao,iao,eps)

      implicit none
      double precision a(*),ao(*),eps
      integer nrow,ncol,nnz,ir(*),jc(*),jao(*),iao(*)

      integer kk,k,i,j,tmpi, cr(nrow), ig(nrow+1), g(nnz)
      integer st(nrow+1)
      double precision tmpa(ncol)

C     We assume that we have the correct dimensions.

c     in case we need to determine the max and min
c     also clean up the vectors containing the elements

c     provide empty arrays:
      do kk = 1,nnz
         g(kk) = 0
      enddo
      do kk = 1,nrow
         cr(kk) = 0
      enddo


c     row need to be determined
      k=0
      do kk=1, nnz
         if ((jc(kk).le.ncol).and.(ir(kk).le.nrow)) then
            k=k+1
            if (k.lt.kk) then
               jc(k)=jc(kk)
               ir(k)=ir(kk)
               a(k)=a(kk)
            endif
         endif
      enddo
      nnz=k

      do kk = 1,nnz
         cr(ir(kk)) = cr(ir(kk)) + 1
c         jao(ir(kk)) = cr(ir(kk))
      enddo


c      return
      st(1) = 1
      do kk = 1, nrow
         st(kk+1) = st(kk) + cr(kk)
      enddo

      do kk = 1, nrow
         ig(kk) = st(kk)
      enddo
      do k=1,nnz
         kk = ir(k)
         g(ig(kk)) = k
         ig(kk) = ig(kk) +1
      enddo
c      return


      kk = 0
      iao(1)=1
      do i = 1,nrow
         do j=1,ncol
            tmpa(j)=0.0
         enddo
         do j=1,cr(i)
            tmpi = g(st(i) + j - 1)
            tmpa(jc(tmpi)) = tmpa(jc(tmpi)) + a(tmpi)
         enddo
         do j=1,ncol
            if( abs(tmpa(j)).gt.eps) then
               kk = kk + 1
               ao(kk) = tmpa(j)
               jao(kk) = j
            endif
         enddo
         iao(i+1)= kk+1
      enddo

      nnz = kk
      return
      end

c-----------------------------------------------------------------------
      subroutine triplet2csr(nrow,ncol,nnz,a,ir,jc,ao,jao,iao,eps)

      implicit none
      double precision a(*),ao(*),eps
      integer nrow,ncol,nnz,ir(*),jc(*),jao(*),iao(*)

      integer     newnnz,   ipos,   k,  i, j, tmp1, tmp2
      double precision tmp
c-----------------------------------------------------------------------
c  Triplet representation     to   Compressed Sparse Row
c  Similar to coocsr from sparsekit
c-----------------------------------------------------------------------
c converts a matrix that is stored in coordinate format
c  a, ir, jc into a row general sparse ao, jao, iao format.
c
c on entry:
c---------
c nrow  = row dimension of matrix
c nrow  = col dimension of matrix
c nnz   = number of nonzero elements in matrix
c a,
c ir,
c jc    = matrix in coordinate format. a(k), ir(k), jc(k) store the nnz
c         nonzero elements of the matrix with a(k) = actual real value of
c         the elements, ir(k) = its row number and jc(k) = its column
c         number. The order of the elements is arbitrary.
c
c on return:
c-----------
c nnz   = number of nonzero elements in matrix
c ao, jao, iao = matrix in general sparse matrix format with ao
c       continung the real values, jao containing the column indices,
c       and iao being the pointer to the beginning of the row,
c       in arrays ao, jao.
c
c------------------------------------------------------------------------

c     cycle over all entries and count the number of elements in each row
c     skip if larger than nrow and ncol. newnnz is actual number within
c     matrix(nrow,ncol).
      newnnz = 0
      do  k=1, nnz
         tmp1 = ir(k)
         if (tmp1 .le. nrow) then
            tmp2 = jc(k)
            if (tmp2 .le. ncol) then
               if (abs(a(k)) .gt. eps) then
                  iao(tmp1) = iao(tmp1)+1
                  newnnz = newnnz + 1
                  if (newnnz.lt.k) then
                    jc(newnnz) = tmp2
                    ir(newnnz) = tmp1
                    a(newnnz) = a(k)
                  endif
               endif
            endif
         endif
      enddo
c Starting position of each row, essentially a cumsum of iao
      k = 1
      do j=1,nrow+1
         tmp1 = iao(j)
         iao(j) = k
         k = k + tmp1
      enddo
c Go through the structure  once more. Fill in output matrix.
c iao is miss used.
      do  k=1, newnnz
         i = ir(k)
         tmp1 = iao(i)
         ao(tmp1) = a(k)
         jao(tmp1) = jc(k)
         iao(i) = tmp1+1
      enddo
c Shift back iao
      do j=nrow,1,-1
         iao(j+1) = iao(j)
      enddo
      iao(1) = 1

c Sort the individual rows
      do i = 1, nrow
         do  ipos = iao(i), iao(i+1)-1
            do  j = iao(i+1)-1, ipos+1, -1
               k = j - 1
               if (jao(k).eq.jao(j)) then
                  ao(k) = ao(k)+ao(j)
                  ao(j) = 0.0
               else
                  if (jao(k).gt.jao(j)) then
                     tmp1 = jao(k)
                     jao(k) = jao(j)
                     jao(j) = tmp1
                     tmp = ao(k)
                     ao(k) = ao(j)
                     ao(j) = tmp
                  endif
               endif
            enddo
         enddo
      enddo


      call cleanspam(nrow,ao,jao,iao,eps)
      nnz = iao(nrow+1)-1
      return
c-----------------------------------------------------------------------
      end



      subroutine cleanspam(nrow,a,ja,ia,eps)

      implicit none
      integer nrow, ia(nrow+1), ja(*)
      double precision  a(*), eps
c
c     this routine removes zero entries. for more complicated cleaning
c     use the sparsekit2 subroutine clncsr.
c
c On entry:
c----------
c     nrow    -- row dimension of the matrix
c     a,ja,ia -- input matrix in CSR format
c
c On return:
c-----------
c     a,ja,ia -- cleaned matrix
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-13
c-----------------------------------------------------------------------
c
c     Local
      integer i,j,k, oldia(nrow+1)

      do  i = 1, nrow+1
         oldia(i) = ia(i)
      enddo

      k = 1
      do i = 1, nrow
         ia(i) = k
         do j=oldia(i),oldia(i+1)-1
            if (.not.(dabs(a(j)) .le. eps)) then

               ja(k) = ja(j)
               a(k) = a(j)
               k = k + 1
            endif

         enddo
      enddo

      ia(nrow+1) = k
      return

c---- end of cleanspam -------------------------------------------------
c-----------------------------------------------------------------------
      end



      subroutine circulant(nrow,len, x,j, a,ja,ia)

      implicit none
      integer nrow, len, ia(nrow+1), ja(*), j(len)
      double precision  a(*), x(len)
c
c
c
c On entry:
c----------
c     nrow    -- row dimension of the matrix
c     len     -- #nnz per line
c     x,j     -- input values and indicies
c
c On return:
c-----------
c     a,ja,ia -- cleaned circulant matrix
c
c Notes:
c-------
c     Reinhard Furrer 2011-08-03
c-----------------------------------------------------------------------
c
c     Local
      integer i,k, kk

      kk = 1
      ia(1) = 1

      do  i = 1, nrow
         ia(i+1) = ia(i)+len

         do k = 1, len
            ja(kk) = mod( j(k) +i-2, nrow)+1
            a(kk)  = x(k)
            kk = kk+1
         enddo
      enddo
      call sortrows(nrow,a, ja, ia)
      return

c---- end of circulant -------------------------------------------------
c-----------------------------------------------------------------------
      end

      subroutine toeplitz(nrow,len, x,j, a,ja,ia,kk)

      implicit none
      integer nrow, len, ia(nrow+1), ja(*), j(len), kk
      double precision  a(*), x(len)
c
c
c
c On entry:
c----------
c     nrow    -- row dimension of the matrix
c     len     -- total #nnz per line and column
c     x,j     -- input values and indicies (indices are shifted!)
c
c On return:
c-----------
c     a,ja,ia -- toeplitz matrix
c     kk     -- nonzero elements
c
c Notes:
c-------
c     Reinhard Furrer 2011-08-03
c-----------------------------------------------------------------------
c
c     Local
      integer i,k, newj

      kk = 1
      ia(1) = 1

      do  i = 1, nrow

         do k = 1, len
            newj = j(k) + i - nrow
            if ((newj.ge.1).and.(newj.le.nrow)) then
               ja(kk) = newj
               a(kk)  = x(k)
               kk = kk+1
            endif

         enddo
         ia(i+1) = kk
      enddo
      kk = kk - 1
      return

c---- end of toeplitz -------------------------------------------------
c-----------------------------------------------------------------------
      end

c-----------------------------------------------------------------------
      subroutine sortrows(nrow,a,ja,ia)

      implicit none
      integer nrow
      integer ia(nrow+1),ja(*)
      double precision  a(*)
c
c     sorts the rows according to column entries
c
c On entry:
c----------
c     nrow    -- row dimension of the matrix
c     a,ja,ia -- input matrix in sparse format
c
c On return:
c-----------
c     a,ja,ia -- cleaned matrix
c
c Notes:
c-------
c     Reinhard Furrer 2006-09-13
c-----------------------------------------------------------------------
c     Local variables
      integer i,j,k,ko,ipos
      double precision  tmp
c
c

c     .. order the entries according to column indices
c     burble-sort is used
c
      do 190 i = 1, nrow
         do 160 ipos = ia(i), ia(i+1)-1
            do 150 j = ia(i+1)-1, ipos+1, -1
               k = j - 1
               if (ja(k).gt.ja(j)) then
                  ko = ja(k)
                  ja(k) = ja(j)
                  ja(j) = ko
                  tmp = a(k)
                  a(k) = a(j)
                  a(j) = tmp
               endif
 150        continue
 160     continue
 190  continue
      return
c---- end of sortrows --------------------------------------------------
c-----------------------------------------------------------------------
      end

# HEADER ####################################################
# This is file spam/R/spam_solve.R.                         #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



########################################################################
########################################################################
#
# Contains common routine for chol.spam and determinant.spam 
# The lattre and Associated S4 elements are in different files.
#
#
########################################################################
########################################################################


# lindx=  colindices       of length nsub   = nnzR
# xlindx= colpointers                nsuper = nnzcolindices
# xlnz=   rowpointers
# snode=  snmember
# xsuper= supernodes
# c(... nnztmp,cachesize)= memory

#### Summary of memory usage in cholstepwise:
# Fortran      R
# m            nrow,ncol
# nsuper                     nnzcolindices
# nnzlmax                    nnzR
#
# Internally, we additionally assign several arrays (of length):
#  m: adj, colcnt
#  7*m+3: iwork
#  iwork: but 7*m+3 is used!

.chol.spam.basis <- function(x,  pivot,  memory,  eps,  verbose){

  if (eps<.Machine$double.eps) stop("'eps' should not be smaller than machine precision", call.=FALSE)

  nrow <- x@dimension[1]
  nnzA <- x@rowpointers[nrow+1]-1L
  if(nrow!=x@dimension[2]) stop("non-square matrix in 'chol'", call.=FALSE)

  if (any( diag.of.spam(x, nrow, nrow) < eps))
    stop("Input matrix to 'chol' not positive definite (up to eps)", call.=FALSE)
# base rule:
#  `nrow(x) * .Machine$double.neg.eps * max(diag(x)`.

  if (length(pivot)==1) {
        if (pivot==FALSE) {
            doperm <- 0
         } else if(pivot==TRUE) {
            doperm <- 1
        } else {
            doperm <- as.integer( switch(match.arg(pivot,c("MMD","RCM")), MMD=1, RCM=2))
        }
  } else  if (length(pivot)==nrow) {
        doperm <- 0
        if (getOption("spam.cholpivotcheck")) {
            checkpivot(pivot,nrow)
        }
  } else stop("'pivot' should be 'MMD', 'RCM' or a valid permutation")

  ### IMPROVEME get better parameter values
  nnzcfact <- c(5,1,5)
  nnzRfact <- c(5,1,2)
  # nnzcolindices = length of array holding the colindices
  if(is.null(memory$nnzcolindices))  {
    nnzcolindices <- ifelse((nnzA/nrow < 5), # very sparse matrix
                           max(1000,nnzA*(1.05*nnzA/nrow-3.8)),
                           nnzA)*nnzcfact[doperm+1]
    nnzcolindices <- max(nnzcolindices,nnzA)
    if (verbose) cat("Guessing upper bound of column indicies as ", nnzcolindices, " (",
            printSize(nnzcolindices),").\n", sep='')
  } else {
    nnzcolindices <- max(memory$nnzcolindices,nnzA)
    memory$nnzcolindices <- NULL
  }
  # nnzR = length of array holding the nonzero values of the factor
  if(is.null(memory$nnzR)) {
    nnzR <- min(max(4*nnzA,floor(.2*nnzA^1.3))*nnzRfact[doperm+1],nrow*(nrow+1)/2)
    if (verbose) cat("Guessing upper bound of factor size.\n")
  } else {
    nnzR <- memory$nnzR
    memory$nnzR <- NULL
  }


  if(is.null(memory$cache))     cache <- 512  else {
    cache <- memory$cache
    memory$cache <- NULL
  }

  if (length( memory)>0 )
    warning("The component(s) ", paste("'",names(memory),"'",sep='',collapse=","),
            " of the argument 'memory'\npassed to function 'chol' not meaningful and hence ignored.",call.=FALSE)


  if(getOption("spam.cholsymmetrycheck")) {
    if (verbose) cat("Checking symmetry.")    
    test <- isSymmetric(x, tol = eps*100)
#  from help of isSymmetric:
#     isSymmetric(object, tol = 100 * .Machine$double.eps, ...)
    if (!isTRUE(test))
      stop("Input matrix to 'chol' not symmetric (up to 100*eps)",call.=FALSE)
  }


    if (verbose) {
        cat("Factorizing matrix of dimension ", nrow, " with ", length(x@entries)," entries (",
            printSize(x),  "). \nReserved memory for factor with ", nnzR, " entries (",
            printSize(nnzR),").\n", sep='')
        }

    force64 <- getOption("spam.force64")
    # we elaborate on the sizes and check for 2^31-2 to allow one additional iteration per counter
    if(force64 || 7*nrow+3  > 2147483646 || nnzcolindices  > 2147483646 || nnzR  > 2147483646) {
        SS <- .format64()
    } else {
        SS <- .format32
    }


    if (length(pivot)==1) {
        if (pivot==FALSE) {
            pivot <- as.vector( seq_len(nrow), SS$type)
        } else if(pivot==TRUE) {
            pivot <- vector(SS$type,nrow)
        } else {
           pivot <- vector(SS$type, nrow)
        }
    } else  pivot <- as.vector( pivot, SS$type)



    if (verbose) {
        cat("\nWorking with",SS$name,"spam matrix")
        if (SS$type != class(x@dimension)) cat(", casting original matrix")

        ## note that object.size(x)  is not required as we have reading mode
        m1 <- object.size(x) +
            6*object.size(pivot)+
            (nnzcolindices +     # length lindx
             2*nrow +nnzA + 7*nrow+3)*   # work arrays {colcnt, split}, {adj, adjncy}  iwork
            ifelse(SS$signature=='double',8,4) +      # depending on 32/64-bit
            nnzR*8    # final entries
        cat(".\nApproximate amount of memory needed: ", printSize( m1),".\n",
            "Starting now the heavy calculation (this might take some time ...\n",sep='')
                                        # gestimate for total use is:
#        m2 <- (8+6)*object.size(pivot)+
#            nnzcolindices*ifelse(SS$signature=='double',8,4) + nnzR*8
    }
  cholstep.intent <- c("rw", "rw", "r", "r",  # we are passing back a few of the arguments 
                    "r", "r", "rw", "rw",
                    "rw", "rw", "w", "w",    # 'w' for lindx xlindx  # _FOR SURE_
                    "rw", "rw", "w", "w",   # 'w' for lnz  xlnz
                    "rw", "rw", "rw", "rw")

  cholstep.signature <- rep(SS$signature, 20)
  cholstep.signature[c(3,15)] <- "double"
  z <- .C64("cholstepwise",                # 11 huge objects...
            SIGNATURE=cholstep.signature,

            nrow = as.vector(nrow, mode=SS$type),
            nnzA = as.vector(nnzA, SS$type),
            d =  x@entries,                          #Size _D_nrow
            jd = as.vector(x@colindices, SS$type),   #Size nrow

            id = as.vector(x@rowpointers, SS$type),#5#Size nrow
            doperm = as.vector(doperm,    SS$type),
            invp = vector(SS$type,nrow),             #Size nrow
            perm = pivot,                            #Size nrow

            nnzlindx = vector(SS$type, 1), #9
            nnzcolindices = as.vector(nnzcolindices, SS$type),
            lindx = vector_dc(SS$type, nnzcolindices), #Size nnzcolindices
            xlindx = vector_dc(SS$type, nrow+1),       #Size nrow

            nsuper = vector(SS$type, 1), #13
            nnzR = as.vector(nnzR, SS$type),
            lnz = vector_dc( "double", nnzR),      #Size _D_nnzR
            xlnz = vector_dc( SS$type, nrow+1),    #Size nrow

            snode = vector(SS$type, nrow), #17     #Size nrow
            xsuper = vector(SS$type, nrow+1),      #Size nrow
            cachesize = as.vector(cache, SS$type),
            ierr = vector(SS$type, 1),

            INTENT=cholstep.intent,
            NAOK = getOption("spam.NAOK"),
            PACKAGE = SS$package)

  if (verbose) cat("... . Size of R-Fortran transfered object: ", printSize( z),").\n", sep='')

  if(z$ierr == 1) stop("Singularity problem when calculating the Cholesky factor.")
  if(z$ierr == 6) stop("Inconsitency in the input",call.=FALSE)

  while( z$ierr>1) {
    if (verbose) cat("Not enough memory allocated. (Starting another round ...\n")    
    if(z$ierr == 4) {
      tmp <- ceiling(nnzR*getOption("spam.cholincreasefactor")[1])
      warning("Increased 'nnzR' with 'NgPeyton' method\n",
                    "(currently set to ",tmp," from ",nnzR,")",call.=FALSE)
      nnzR <- tmp
    }
    if(z$ierr == 5) {
      tmp <- ceiling(nnzcolindices*getOption("spam.cholincreasefactor")[2])
      warning("Increased 'nnzcolindices' with 'NgPeyton' method\n",
         "(currently set to ",tmp," from ",nnzcolindices,")",call.=FALSE)
      nnzcolindices <- tmp
    }
    if(force64 || nnzcolindices  > 2147483647) {
        SS <- .format64()
    } else {
        SS <- .format32
    }
    cholstep.signature <- rep(SS$signature, 20)
    cholstep.signature[c(3,15)] <- "double"


    z <- .C64("cholstepwise",
              SIGNATURE=cholstep.signature,

              nrow = as.vector(nrow, mode=SS$type),
              nnzA = as.vector(nnzA, SS$type),
              d =  x@entries,
              jd = as.vector(x@colindices, SS$type),

              id = as.vector(x@rowpointers, SS$type),
              doperm = as.vector(doperm,    SS$type),
              invp = vector(SS$type, nrow),
              perm = pivot,

              nnzlindx = vector(SS$type,1),   # length of colindices
              nnzcolindices = as.vector(nnzcolindices, SS$type),
              lindx = vector(SS$type, nnzcolindices),  # one backup!!
              xlindx = vector(SS$type, nrow+1),   # col pointers. will be shorter

              nsuper = vector(SS$type, 1),        #  number of supernodes
              nnzR = as.vector(nnzR, SS$type),    #  allocation for elements in the factor
              lnz = vector("double",nnzR),        # factor 
              xlnz = vector(SS$type, nrow+1),     # row pointers of factor

              snode = vector(SS$type, nrow),
              xsuper = vector(SS$type, nrow+1),
              cachesize = as.vector(cache, SS$type),
              ierr = vector(SS$type, 1),

              INTENT=cholstep.intent,
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package)

    if (verbose) cat("... . Size of R-Fortran transfered object: ", printSize( z),").\n", sep='')
    if(z$ierr == 1) stop("Singularity problem or matrix not positive definite", call.=FALSE)
  }
  z  
}


########################################################################

# HEADER ####################################################
# This is file spam/tests/testthat/helper.R.                #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################


require('spam')
# For efficient primary testing:
#    require('spam',lib='lib') ; setwd( 'spam/tests/testthat/') ;require('testthat')
#    for (i in rev(system('ls', intern=T) )) source(i)


spamtest_eq <- function( xtest, xtrue, tol= 1.0e-1, relative=TRUE, label="", ...){
    cxtest <- c(xtest); cxtrue <- c(xtrue)
    if(relative)
        denom <- mean(abs(cxtrue), na.rm = TRUE)
    else
        denom <- 1
    test_value <- sum(abs(cxtest - cxtrue), na.rm = TRUE) / denom
    expect_lt(test_value, tol, label = paste(  "[spamtest_eg failed]", label), ...)
}
spamtest_diff <- function( xtest, xtrue, tol= 1.0e-6, relative=TRUE, label="", ...){
    cxtest <- c(xtest); cxtrue <- c(xtrue)
    if(relative)
        denom <- mean(abs(cxtrue), na.rm = TRUE)
    else
        denom <- 1
    test_value <- sum(abs(cxtest - cxtrue), na.rm = TRUE) / denom
    expect_gt(test_value, tol, label = paste( "[spamtest_diff failed]", label), ...)
}

LIB.LOC <- "../../../lib"

## reset to inital settings
default_options <- list(  spam.eps=.Machine$double.eps,   # smaller than this is considered as zero
                        spam.force64=FALSE,
                        spam.validate=FALSE,            # validate the spam object before calling a native routine for
                                        # increased stability.
                        
                        spam.drop=FALSE,                # drop passed to subset functions
                        
                        spam.printsize=100,             # the max size which we print as regular matrices
                        spam.imagesize=10000,           # the max size which we display as regular matrices
                        spam.cex=1200,                  # scaling factor for scatter displays
                        
                        spam.structurebased=TRUE,      # calculating on nonzero entries only...
                        
                        spam.inefficiencywarning=1e6,  # tell when something inefficient is done
                        
                        spam.trivalues=FALSE,           # with upper./lower/.tri return values (TRUE) or only structure?
                        spam.listmethod="PE",           # method to be used when using spam.list
                        
                        spam.NAOK = FALSE,
                        safemodevalidity=TRUE,  # verify while S4 construction
                        spam.dopivoting=TRUE,           # what type of back/forwardsolve?
                        spam.cholsymmetrycheck=TRUE,     # Should symmetry be tested in the cholesky factorization
                        spam.cholpivotcheck=TRUE,        # Should the pivot be tested?
                        spam.cholupdatesingular="warning",     # ("error", "warning","NULL")
                        spam.cholincreasefactor=c(1.25,1.25),
                        spam.nearestdistincreasefactor=1.3,
                        spam.nearestdistnnz=c(500^2,500)  )

                                        # Load the default settings:
options(default_options)

# HEADER ####################################################
# This is file spam/R/subset.R.                             #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################




# SUBSETTING
##########################################################################################

# notice the drop catch...
#   I don"t know the best and official way, but it works as it is here...

setMethod("[", signature(x = "spam",
			 i = "missing", j = "missing", drop = "ANY"),
	  function (x, i, j,..., drop) { # cat("missmiss")
           x})

setMethod("[",signature(x="spam",i="vector",j="missing", drop = "logical"),
	  function (x, i, j,..., drop) {  #cat("   log call was", deparse(match.call()), "\n")
            if (nargs()==3) {
              subset_rows.spam(x, i,drop=drop)
            } else {
              subset_rows.spam(x, i,,drop=drop)
            }}
          )

setMethod("[",signature(x="spam",i="vector",j="missing", drop = "missing"),
	  function (x, i, j,..., drop) { #cat("    mis call was", deparse(match.call()), "\n")
            if (nargs()==2) {
              subset_rows.spam(x, i)
            } else {
              subset_rows.spam(x, i,)
              }})

setMethod("[",signature(x="spam",i="vector",j="vector", drop = "ANY"),
	  function (x, i, j,..., drop) { # cat("vecvec")
            subset.spam(x,rw=i,cl=j,drop=drop)} )

setMethod("[",signature(x="spam",i="missing",j="vector", drop = "ANY"),
	  function (x, i, j,...,drop) { # cat("missvec")
            subset.spam(x,rw=1:x@dimension[1],cl=j,drop=drop)} )

setMethod("[",signature(x="spam",i="matrix",j="missing", drop = "missing"),
	  function (x, i, j,..., drop) {subset.spam(x,rw=i) })

setMethod("[",signature(x="spam",i="matrix",j="missing", drop = "logical"),
	  function (x, i, j,..., drop) {subset.spam(x,rw=i,drop=drop) })

setMethod("[",signature(x="spam",i="matrix",j="matrix", drop = "ANY"),
	  function (x, i, j,..., drop) {subset.spam(x,rw=cbind(c(i),c(j)),drop=drop) })

setMethod("[",signature(x="spam",i="spam",j="missing", drop = "ANY"),
	  function (x, i, j,..., drop=getOption("spam.drop"))
{
                                        # drop is not implemented yet
    ## print("subset spam spam")
  dimx <- x@dimension
  nrow <- dimx[1]
  ncol <- dimx[2]
  if ( i@dimension[1]>nrow | i@dimension[2]>ncol)
    stop("subscript out of bounds",call.=FALSE)
  ## z <- .Fortran("amask",
  ##               nrow=as.integer(nrow),
  ##               ncol=as.integer(ncol),
  ##               a=as.double(x@entries),
  ##               colindices=as.integer(x@colindices),
  ##               rowpointers=as.integer(x@rowpointers),
  ##               jmask=as.integer(i@colindices),
  ##               imask=as.integer(c(i@rowpointers,rep(i@rowpointers[length(i@rowpointers)],nrow+1-length(i@rowpointers)))),
  ##               c=as.double(x@entries),
  ##               jc=as.integer(x@colindices),
  ##               ic=as.integer(x@rowpointers),
  ##               iw=logical(ncol),
  ##               nzmax=as.integer(length(i@colindices)),
  ##               ierr=0L,
    ##               NAOK=getOption("spam.NAOK"),PACKAGE="spam") # some copying is required!!!!
    if(.format.spam(x)$package == "spam64")
        SS <- .format64()
    else
        SS <- .format32

    z <- .C64("amask",
             ##     subroutine amask (nrow,ncol,a,ja,ia,jmask,imask,
     ## *                  c,jc,ic,nzmax,ierr)
              SIGNATURE = c( SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                            SS$signature, SS$signature,
                            "double", SS$signature, SS$signature,
                             SS$signature, SS$signature),

              nrow = nrow,
              ncol = ncol,

              a = x@entries,
              colindices = x@colindices,
              rowpointers = x@rowpointers,

              jmask = i@colindices,
              imask = c(i@rowpointers, rep_len64(i@rowpointers[length(i@rowpointers)], nrow+1-length(i@rowpointers))),

              c = x@entries,
              jc = x@colindices,
              ic = x@rowpointers,

              nzmax = length( i@colindices),
              ierr = 0,

              INTENT = c( "rw", "rw",
                         "r", "r", "r",
                         "r", "r",
                         "w", "w", "w",
                         "r", "w" ),
              NAOK = getOption("spam.NAOK"),
              PACKAGE = SS$package) # some copying is required!!!!

  nz <- z$ic[nrow+1]-1
    if (nz==0) {
        ## print("nz = 0")
        return( numeric(0))
    }
    if (drop) {
        ## print("drop")
    ic <- unique( z$ic[1:(z$nrow+1)])
    dimx <- as.integer(c(length(ic)-1,max(z$jc[1:nz])))
    } else {
        ## print("notdrop")
    ic <-z$ic[1:(z$nrow+1)]
  }
  ## return(new("spam",entries=z$c[1:nz],colindices=z$jc[1:nz],rowpointers=ic,
    ##              dimension=dimx))
    return(.newSpam(
        entries = z$c[1:nz],
        colindices = z$jc[1:nz],
        rowpointers = ic,
        dimension = dimx))
}      )

setMethod("[", signature(x = "spam", i = "ANY", j = "ANY", drop = "ANY"),
	  function(x,i,j, drop)
          stop("Invalid or not-yet-implemented 'spam' subsetting"))

# the proper S3 subsetting causes problems...
#  "[.spam" <- function (x, rw, cl,drop=getOption("spam.drop")) {subset.spam(x,rw=rw,cl=cl,drop) }


"subset_rows.spam" <-
function (x, i, ..., drop=getOption("spam.drop"))
  # approach: we extract rows (nargs=2) or elements (nargs=3)
  #  i is a vector of integers or logical!
  # nargs idea from Matrix!
{
  nA <- nargs()+missing(drop)
#  cat("subset_rows.spam call was", deparse(match.call())," ",nargs(), " " , nA, "\n")
 dimx <- x@dimension
  nrow <- dimx[1]
  ncol <- dimx[2]
  mini <- min(i, na.rm=TRUE)
  maxi <- max(i, na.rm=TRUE)
  if (mini<0 & maxi>0)
    stop("Negative and positive subscripts mixed")

  SS <- .format.spam(x)

  if(nA==3) { # extract elements
    if (is.logical(i)) {
      inefficiencywarning( "Logical subsetting may be inefficient, is this really what you want?",
                          prod(dimx))
      return(.C64("spamcsrdns",
                  SIGNATURE = c(SS$signature,
                                "double", SS$signature, SS$signature,
                                "double"),
                  nrow,

                  entries = x@entries,
                  colindices = x@colindices,
                  rowpointers = x@rowpointers,

                  res = vector_dc("double", prod(dimx)),
 #                 INTENT = c("r","r","r","r","w"), # not checked   !!!
                  NAOK=getOption("spam.NAOK"),PACKAGE = SS$package)$res[i])
    }

    if (mini<0) {
      inefficiencywarning( "Negative subsetting may be inefficient, is this really what you want?",
                          prod(dimx))
      return(.C64("spamcsrdns",
                  SIGNATURE = c(SS$signature,
                                "double", SS$signature, SS$signature,
                                "double"),
                  nrow,

                  entries=x@entries,
                  colindices=x@colindices,
                  rowpointers=x@rowpointers,

                  res=vector_dc("double",prod(dimx)),
#                  INTENT = c("r","r","r","r","w"),  # not checked   !!!
                  NAOK=getOption("spam.NAOK"),PACKAGE = SS$package)$res[i])
    }
    # eliminate zeros,
    # force too large to NA, keep NAs
    i <- i[i>0]
    ind <- !(is.na(i)|(i> (nrow*ncol)))
    ii <- i[ind]-1
    i <- ii %% nrow+1
    j <- ii %/% nrow+1
    nir <- length(i)
    z <- vector("double",length(ind))
    z[!ind] <- NA
    z[ind] <- .C64("getallelem",
                   SIGNATURE = c(SS$signature, SS$signature, SS$signature,
                                 "double", SS$signature, SS$signature,
                                 SS$signature, "double"),
                   nir,
                   i,
                   j,

                   x@entries,
                   x@colindices,
                   x@rowpointers,

                   iadd = vector_dc(SS$type,nir),
                   allelem = vector_dc("double",nir),

                   ## INTENT = c("r","r","r",
                   ##            "r","r","r",
                   ##            "w","w"),   # not checked   !!!
                   NAOK=getOption("spam.NAOK"),PACKAGE=SS$package)$allelem
# getallelem(nir,ir,jr,a,ja,ia,alliadd,allelem)
    return(z)
  }
  if(nA==4) {

    if (is.logical(i)) {  # logical
      if( length(i) > nrow)  stop("(subscript) logical subscript too long",call.=FALSE)

      i <- seq_len( nrow)[i]
    }  else {

      i <- i[i!=0]     # eliminate zero lines

    if (maxi>x@dimension[1])
      stop("subscript out of bounds",call.=FALSE)

      # negative values:
      if ( maxi <= 0 )       i <- seq_len( nrow)[i]
    }


    ni <- length(i)
    if (ni==0) return(numeric(0))   # zero elements...

    if (any(is.na(i))) {
      stop("'NA's in subsetting vector have been eliminated.")
#      i <- i[!is.na(i)]
    }


    nz <- (sum(x@rowpointers[i+1]-x@rowpointers[i]))
    if (nz==0) {#trap zero matrix
      if (drop==TRUE && (ni==1 || ncol==1)) return( vector("double",max(ni,ncol)))
      else
        return(new("spam",rowpointers=c(1L,rep.int(2L,ni )),
                   dimension = c(ni,ncol)))
    }  else {
 #          subroutine getlines(a,ja,ia, ni, i, bnz, b,jb,ib)
     z <- .C64("getlines",
               SIGNATURE = c("double", SS$signature, SS$signature,
                             SS$signature,  SS$signature, SS$signature,
                             "double", SS$signature, SS$signature),
               x@entries, x@colindices, x@rowpointers,

               ni,
               i,
               newnz=nz,

               entries=vector_dc("double", nz),
               colindices=vector_dc(SS$type, nz),
               rowpointers=vector_dc(SS$typ, ni+1),

               NAOK=getOption("spam.NAOK"),
               PACKAGE=SS$package)
  #    print(c(nz,z$newni,is.integer(nz), is.integer(z$newni),z$newni!=ni))
     if(z$newnz!=nz) stop(gettextf("Subsetting error, please report %d, %d",z$newnz,nz))
    }
#    print(c(drop,ni,ncol,(drop==TRUE && (ni==1 || ncol==1) )))
    if (drop==TRUE && (ni==1 || ncol==1))
      # this is essentially a c() call
      return(.C64("spamcsrdns",
                  SIGNATURE = c(SS$signature,
                                "double", SS$signature, SS$signature,
                                "double"),
                  nrow=ni,

                  entries=z$entries,
                  colindices=z$colindices,
                  rowpointers=z$rowpointers,

                  res=vector_dc("double",prod(ni,ncol)),
                  NAOK=getOption("spam.NAOK"),
                  PACKAGE = SS$package)$res)
    else {
        return(.newSpam(
            entries=z$entries,
            colindices=z$colindices,
            rowpointers=z$rowpointers,
            dimension=c(ni,ncol),
            force64=getOption("spam.force64")))
    }

  }
  stop("incorrect number of dimensions")
}

## subset_rows.spam <- function(x, i, drop){

##   dimx <- x@dimension
##   nrow <- dimx[1]
##   ncol <- dimx[2]

##     ni <- length(i)
##     if (ni==0) return(numeric(0))   # zero elements...

##     if (any(is.na(i))) {
##       stop(""NA"s in subsetting vector not allowed.")
##     }


##     nz <- sum(x@rowpointers[i+1]-x@rowpointers[i])
##     if (nz==0) { #trap zero matrix
##       if (drop==TRUE && (ni==1 || ncol==1))
##         return( vector("double",max(ni,ncol)))
##       else
##         return(new("spam",rowpointers=c(1,rep(2L,ni )),
##                    dimension = c(ni,ncol)))
##     }

##     SS <- .format.spam(x)

##      z <- .C64("getlines",
##                SIGNATURE=c("double", SS$signature, SS$signature, SS$signature,
##                    SS$signature, SS$signature, "double", SS$signature,
##                    SS$signature),

##                x@entries,
##                x@colindices,
##                x@rowpointers,
##                ni,

##                i, #int64
##                newnz=nz,
##                entries=vector_dc("double",nz),
##                colindices=vector_dc(SS$type,nz),

##                rowpointers=vector_dc(SS$type,ni+1),

##                INTENT=c("r", "r", "r", "r",
##                    "r", "rw", "w", "w",
##                    "w"),
##                NAOK = getOption("spam.NAOK"),
##                PACKAGE=SS$package)
##      if(z$newnz!=nz) stop(gettextf("Subsetting error, please report %d, %d",z$newnz,nz))

##     # --- CHANGED ---

##   newx <- .newSpam(entries=z$entries,
##                    colindices=z$colindices,
##                    rowpointers=z$rowpointers,
##                    dimension=c(ni,ncol))

##     if (drop==TRUE && (ni==1 || ncol==1)) {

##         x <- newx
##         SS <- .format.spam(x)

##         dimx <- x@dimension

##         result <- .C64("spamcsrdns",
##                        SIGNATURE=c(SS$signature, "double", SS$signature, SS$signature, "double"),
##                        nrow=dimx[1],
##                        entries=x@entries,
##                        colindices=x@colindices,
##                        rowpointers=x@rowpointers,
##                        res=vector("double",prod(dimx)),  # TODO: Expects zeros, because it
##                                         # only overwrites the non-zero elements! (use RW)

##                        INTENT=c("r", "r", "r", "r", "rw"),
##                        NAOK = getOption("spam.NAOK"),
##                        PACKAGE = SS$package
##                        )
##         return(result$res)
##     }else {
##       return(newx)
##     }

##   stop("incorrect number of dimensions")
## }

"subset.spam" <- function (x,rw,cl,...,drop=getOption("spam.drop"))
{
  # we separate into cases where:
  # (A) rw matrix:
  #     1: logical: transformation to spam and extract structure
  #     2: two column matrix: extract (i,j) as given by the lines.
  #     3: all else extract   x[ c( rw)]
  # (B) rw and cl one element: ((i,j)
  # (C) rw and cl vectors:  (i1:i2,j1:j2)               [i1<=i2, j1<=j2]
  #                         (c(i1,...,ii),c(j1,...,jj)) [arbitrary block]
#  if (missing(drop)) drop <- getOption("spam.drop")
#  print(drop)
  dimx <- x@dimension
  nrow <- dimx[1]
  ncol <- dimx[2]

  SS <- .format.spam(x)

  if (is.matrix(rw)) {
    if (is.logical(rw)) {
      return( x[as.spam.matrix(rw)] )
    }
    if (dim(rw)[2]==2) {
      ir <- rw[,1]
      jr <- rw[,2]
    } else  {
      ir <- c(rw-1) %% nrow + 1
      jr <- c(rw-1) %/% nrow + 1
    }
    if ( (min(ir)<1)|(max(ir)>x@dimension[1])|(min(jr)<1)|(max(jr)>x@dimension[2]))
      stop("subscript out of bounds",call.=FALSE)
    nir <- length(ir)
    return(.C64("getallelem",
                SIGNATURE = c(SS$signature, SS$signature, SS$signature,
                              "double", SS$signature, SS$signature,
                              SS$signature,"double"),
                nir,
                ir,
                jr,

                x@entries, x@colindices, x@rowpointers,

                vector_dc(SS$type, nir),
                allelem=vector_dc("double",nir),
                NAOK=getOption("spam.NAOK"), PACKAGE=SS$package)$allelem)

  }
  # negative values:
  if ( max(rw)<0 )       rw <- seq_len( nrow)[rw]
  if ( max(cl)<0 )       cl <- seq_len( ncol)[cl]

  # logical
  if (is.logical(rw))    rw <- seq_len( nrow)[rw]
  if (is.logical(cl))    cl <- seq_len( ncol)[cl]

  if (length(cl)==0) stop("You should subset at least one element for the columns",call.=FALSE)
  if (length(rw)==0) stop("You should subset at least one element for the rows",call.=FALSE)

  if ( (min(rw)<1)|(max(rw)>x@dimension[1])|(min(cl)<1)|(max(cl)>x@dimension[2]))
    stop("subscript out of bounds",call.=FALSE)

  if (length(rw)==1 & length(cl)==1){
                                        # function to extract only one element
    return(.C64("getelem",
                SIGNATURE = c(SS$signature, SS$signature,
                              "double", SS$signature, SS$signature,
                              SS$signature, "double"),
                rw,
                cl,

                x@entries, x@colindices, x@rowpointers,

                iadd=vector(SS$type, 1),
                elem=vector("double", 1),
                PACKAGE=SS$package)$elem)
  }
  if (is.vector(rw) && is.vector(cl)) {
    nrw <- length(rw)   # length returns an integer, so is a product therof
    ncl <- length(cl)
    diffrw <- diff(rw)
    diffcl <- diff(cl)
    nz <- ( min( (1+sum(diff(sort(rw))==0))*(1+sum(diff(sort(cl))==0))*
                length(x@entries), prod(nrw,ncl)))  # very pessimistic
    if (all(diffrw==1) & all(diffcl==1)) {
      z <- .C64("submat",
                SIGNATURE = c(SS$signature,
                              SS$signature, SS$signature, SS$signature, SS$signature,
                              "double", SS$signature, SS$signature,
                              SS$signature, SS$signature,
                              "double", SS$signature, SS$signature),
                job=1, # need values as well

                i1=rw[1],
                i2=rw[nrw],
                j1=cl[1],
                j2=cl[ncl],

                x@entries, x@colindices, x@rowpointers,

                nr=0,
                nc=0,

                entries=vector_dc("double",nz),
                colindices=vector_dc(SS$type,nz),
                rowpointers=vector_dc(SS$type,nrw+1),

                NAOK=getOption("spam.NAOK"),PACKAGE = SS$package)
      nz <- z$rowpointers[z$nr+1]-1
    } else {
      z <- .C64("getblock",
                SIGNATURE = c("double", SS$signature, SS$signature,
                              SS$signature, SS$signature, SS$signature, SS$signature, SS$signature,
                              "double", SS$signature, SS$signature),
                x@entries, x@colindices, x@rowpointers,

                nr=nrw,
                rw,
                nc=ncl,
                cl,
                nz=nz,

                entries=vector("double",nz),
                colindices=vector(SS$type,nz),rowpointers=vector(SS$type,nrw+1),
                NAOK=getOption("spam.NAOK"),
                PACKAGE = SS$package)
      nz <- z$nz
    }
    if (nz==0) {#trap zero matrix
      if (drop==TRUE && (z$nr==1 || z$nc==1)) return( vector("double",max(z$nr,z$nc)))
      else
        return(new("spam",rowpointers=c(1L,rep.int(2L,z$nr )),
                   dimension = c(z$nr,z$nc)))
    }

    if (drop==TRUE && (z$nr==1 || z$nc==1))
      # this is essentially a c() call
      return(.C64("spamcsrdns",
                  SIGNATURE = c(SS$signature, "double",
                                SS$signature, SS$signature, "double"),
                  nrow=z$nr,
                  entries=z$entries[1:nz],

                  colindices=z$colindices[1:nz],
                  rowpointers=z$rowpointers[1:(z$nr+1)],

                  res=vector_dc("double", prod(z$nr,z$nc)),
                  NAOK=getOption("spam.NAOK"), PACKAGE = SS$package)$res)
    else {
        return(.newSpam(
            entries=z$entries[1:nz],
            colindices=z$colindices[1:nz],
            rowpointers=z$rowpointers[1:(z$nr+1)],
            dimension=c(z$nr,z$nc),
            force64=getOption("spam.force64")))
    }

  }
  stop("invalid or not-yet-implemented 'spam' subsetting")
}



#subset.rows.spam <- function(x, i, ..., drop=getOption("spam.drop")) {
subset.rows.spam <- function(...) {
    .Defunct('spam:::subset_rows.spam', package = 'spam', msg = "'subset.rows.spam' is defunct. Use 'spam:::subset_rows.spam' instead.\n")
}

# HEADER ####################################################
# This is file spam/R/rmvnorm.R.                            #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################



# Draw from a multivariate normal:
# (Algorithm 2.3 from Rue and Held, 2005)
rmvnorm <- function(n, mu = rep.int(0, dim(Sigma)[1]), Sigma, ..., mean, sigma) {

  if (!missing(mean))   mu <- mean
  if (!missing(sigma))  Sigma <- sigma

  if (is( Sigma, "spam")) return( rmvnorm.spam( n, mu, Sigma, ...))

  R <- chol( Sigma, ...)
  N <- dim( Sigma)[1]

  return(sweep( as.matrix( ( array(rnorm(n*N),c(n,N)) %*% R)), 2, mu, "+"))
}

rmvnorm.spam <- function(n, mu = rep.int(0, dim(Sigma)[1]), Sigma, Rstruct = NULL, ..., mean, sigma) {
  # taken from ?chol.spam
  if (!missing(mean))   mu <- mean
  if (!missing(sigma))  Sigma <- sigma

  if (!is(Sigma, "spam"))  return(  rmvnorm(n, mu, Sigma, ...))  # just in case!

  if (is(Rstruct,"spam.chol.NgPeyton"))
    cholS <- update.spam.chol.NgPeyton( Rstruct, Sigma, ...)
  else
    cholS <- chol.spam( Sigma,...)
  # cholS is the upper triangular part of the permutated matrix Sigma
  iord <- ordering(cholS, inv=TRUE)

  N <- dim(Sigma)[1]

  R <- as.spam(cholS)
  retval <- as.matrix( ( array(rnorm(n*N),c(n,N)) %*% R)[,iord,drop=F])
     # It is often better to order the sample than the matrix
     # R itself.
  return(sweep(retval, 2, mu, "+"))
}


rmvt <- function(n, Sigma, df = 1, delta = rep(0, nrow(Sigma)),
                 type = c("shifted", "Kshirsagar"), ..., sigma) {

  if (!missing(sigma))  Sigma <- sigma

  return( rmvt.spam(n = n, Sigma, df, delta, type, ...))
}


rmvt.spam <- function (n, Sigma, df = 1, delta = rep(0, nrow(Sigma)),
    type = c("shifted", "Kshirsagar"), ..., sigma) {

    if (!missing(sigma))  Sigma <- sigma
    if (length(delta) != nrow(Sigma))
        stop("'delta' and 'Sigma' have non-conforming size")
    if (hasArg(mean))
        stop("Providing 'mean' does *not* sample from a multivariate t-distribution!")
    if (df == 0 || ( (df > 0) & is.infinite(df)))
        return(rmvnorm.spam(n, mean = delta, Sigma = Sigma, ...))
    type <- match.arg(type)
    switch(type, Kshirsagar = {
        return(rmvnorm.spam(n, mean = delta, Sigma = Sigma, ...)/sqrt(rchisq(n,
            df)/df))
    }, shifted = {
        return( sweep( rmvnorm.spam(n, Sigma = Sigma, ...)/sqrt(rchisq(n,
            df)/df), 2, delta, "+"))
    }, stop("wrong 'type'"))
}






# Draw from a multivariate normal given a precision matrix:
# (Algorithm 2.4 from Rue and Held, 2005)
rmvnorm.prec <- function(n, mu = rep.int(0, dim(Q)[1]), Q, Rstruct = NULL, ...) {

  if (is(Rstruct,"spam.chol.NgPeyton"))
    R <- update.spam.chol.NgPeyton( Rstruct, Q, ...)
  else
    R <- chol(Q, ...)
  # R is the upper triangular part of the permutated matrix Sigma

  N <- dim(Q)[1]
  nu <- backsolve(R, array(rnorm(n*N), c(N, n)), k = N)
  return(t(nu + mu))
}


# Draw from the canonical representation of a multivariate normal:
# (Algorithm 2.5 from Rue and Held, 2005)
rmvnorm.canonical <- function(n, b, Q, Rstruct=NULL, ...) {
  N=dim(Q)[1]
  if (is(Rstruct,"spam.chol.NgPeyton"))
    R <- update.spam.chol.NgPeyton( Rstruct, Q, ...)
  else
    R <- chol(Q,...)

  if(is(R,"spam.chol.NgPeyton")){
     mu <- drop(solve.spam( R, b))
  } else {
     mu <- backsolve( R, forwardsolve( t(R), b), k=N)
  }
  nu <- backsolve(R, array( rnorm(n*N), c(N, n)), k=N)
  return(t(nu+mu))
}



rmvnorm.const <- function (n, mu = rep.int(0, dim(Sigma)[1]), Sigma,
                                Rstruct = NULL,
                                A = array(1, c(1,dim(Sigma)[1])), a=0, U=NULL,  ...)
{
    N <- dim(Sigma)[1]
    if (!identical(dim(A)[2], N)) stop("Incorrect constraint specification")

    if (is(Rstruct, "spam.chol.NgPeyton"))
        cholS <- update.spam.chol.NgPeyton(Rstruct, Sigma, ...)
    else cholS <- chol.spam(Sigma, ...)
    iord <- ordering(cholS, inv = TRUE)
    N <- dim(Sigma)[1]
    R <- as.spam(cholS)
    x <- sweep( (array(rnorm(n * N), c(n, N)) %*% R)[, iord, drop = F], 2, mu, "+")

    if (is.null(U)){
      V <- backsolve( R, forwardsolve( R, t(A)), k=N)
      W <- A %*% V
      U <- solve(W, t(V))
    }
    correct <- A %*% t(x) - a
    return(x - t( t(U)%*% correct))
}


rmvnorm.prec.const <- function (n, mu = rep.int(0, dim(Q)[1]), Q,
                                Rstruct = NULL,
                                A = array(1, c(1,dim(Q)[1])), a=0, U=NULL,  ...)
{
    N = dim(Q)[1]
    if (!identical(dim(A)[2], N)) stop("Incorrect constraint specification")

    if (is(Rstruct, "spam.chol.NgPeyton"))
        R <- update.spam.chol.NgPeyton(Rstruct, Q, ...)
    else R <- chol(Q, ...)
    x <- backsolve(R, array(rnorm(n * N), c(N, n)), k=N) + mu


    if (is.null(U)){
        tV <- t( backsolve( R, forwardsolve( R, t(A)), k=N))
        W <- tcrossprod(A, tV)
        U <- solve(W, tV)
    }
    correct <- A %*% x - a
    return(t(x- t(U) %*% correct))
}



rmvnorm.canonical.const <- function (n, b, Q, Rstruct = NULL,
                                     A = array(1, c(1,dim(Q)[1])), a=0, U=NULL, ...)
{
    N = dim(Q)[1]
    if (!identical(dim(A)[2], N)) stop("Incorrect constraint specification")

    if (is(Rstruct, "spam.chol.NgPeyton"))
        R <- update.spam.chol.NgPeyton(Rstruct, Q, ...)
    else R <- chol(Q, ...)
    if (is(R, "spam.chol.NgPeyton")) {
        mu <- drop(solve.spam(R, b))
    }
    else {
        mu <- backsolve(R, forwardsolve(t(R), b))
    }
    x <- backsolve(R, array(rnorm(n * N), c(N, n)), k=N) + mu


    if (is.null(U)){
        tV <- t( backsolve( R, forwardsolve( R, t(A)), k=N))
        W <- tcrossprod(A, tV)
        U <- solve(W, tV)
    }
    correct <- A %*% x - a
    return(t(x- t(U) %*% correct))


}



### conditional simulation
rmvnorm.conditional <- function(n, y, mu = rep.int(0, dim(SigmaXX)[1]+dim(SigmaYY)[1]),
                                   SigmaXX, SigmaYY, SigmaXY,
                                   noise,
                                   RstructYY = NULL, ...) {

  if (!is(SigmaXX, "spam"))  stop("Implemented only for 'spam' covariance matrices")

  nx <- dim(SigmaXX)[1]
  if (missing(SigmaYY)) {
    if (missing(noise)) stop("Either 'noise' or 'SigmaYY' are required")
    SigmaYY <- SigmaXX+diag.spam(noise, nx, nx)
  } else if (!missing(noise)) stop("Either 'noise' or 'SigmaYY' are required, not both")

  ny <- dim(SigmaYY)[1]
  if (length(y)!=ny)  stop("Length of 'y' does not match dimension of 'SigmaYY'")

  if (missing(SigmaXY)) {  # save one transpose
    if (nx!=ny) stop("'SigmaXY' needs to be specified")
    Sigma <- rbind.spam(cbind.spam(SigmaXX, SigmaXX), cbind(SigmaXX, SigmaYY))
    SigmaXY <- SigmaXX   # alternative would be to have another if below...
  } else {
    if (!all(dim(SigmaXY) == c(nx,ny))) stop("Wrong dimension for 'SigmaXY'")
    Sigma <- rbind.spam(cbind.spam(SigmaXX, SigmaXY), cbind(t(SigmaXY), SigmaYY))
  }
  #  Rstruct <- chol.spam(Sigma)
  rsample <- t(rmvnorm.spam(n = n, mu = mu, Sigma = Sigma)) #, Rstruct = Rstruct)

  t(rsample[1:nx,] + SigmaXY %*%
      solve.spam(a = SigmaYY, b = y-rsample[(nx+1):(nx+ny),],
                 Rstruct = RstructYY))
}
rmvnorm.cond <- rmvnorm.conditional

# HEADER ####################################################
# This is file spam/tests/testthat/test-helper.R.           #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################

rm(list = ls())
source("helper.R")

## library("testthat")
## library("spam64", lib.loc = LIB.LOC)
## library("spam", lib.loc = "../../../lib/")


context("test-rmvnorm.R")



########
verbose <- FALSE

set.seed(14)


n <- 5
Sigma <- .25^abs(outer(1:n,1:n,'-'))
Q <- as.spam(solve(Sigma))
b <- 1:n

struct <- chol(Q, verbose=verbose)

test_that("rmvnorm", {

    set.seed(14)
    tmp1 <- rmvnorm(10,  Sigma=Sigma, verbose=verbose)
    set.seed(14)
    tmp2 <- rmvnorm.spam(10, Sigma=Sigma)
    expect_equal( tmp1, tmp2)

    set.seed(14)
    tmp1 <- rmvnorm(10,  Sigma=Q)
    set.seed(14)
    tmp2 <- rmvnorm.spam(10, Sigma=Q)
    expect_equal( tmp1, tmp2)



    set.seed(14)
    tmp1 <- rmvnorm.canonical(10, b, Q)
    set.seed(14)
    spamtest_eq( rmvnorm.canonical(10, b, Q, Lstruct=struct), tmp1 )
    set.seed(14)
    spamtest_eq( rmvnorm.prec(10, solve(Q,b), Q), tmp1 )
    set.seed(14)
    spamtest_eq( rmvnorm.prec(10, solve(Q,b), Q, Lstruct=struct), tmp1 )
    set.seed(14)
    ## cat("For rmvnorm.canonical:\n- comparing sample mean with truth:\n")
    ## for (i in 10^(1:4))
    ##     cat('    sample size n=',i,' yields  Frobenius-norm:',
    ##         norm( apply(rmvnorm.canonical(i, b, Q, Lstruct=struct), 2,mean)- solve(Q,b),'f'),'\n')
    ## cat("- comparing sample variance with truth:\n")
    ## for (i in 10^(1:4)){
    ##     cat('    sample size n=',i,' yields Frobenius-norm:',
    ##         norm( var( rmvnorm.canonical(i, b, Q=Q, Lstruct=struct))- Sigma,'f'),'\n')
    ## set.seed(14)
    ## cat("For rmvnorm.prec:\n- comparing sample mean with truth:\n")
    ## for (i in 10^(1:4))
    ##     cat('    sample size n=',i,' yields  Frobenius-norm:',
    ##         norm( apply(rmvnorm.prec(i, b, Q, Lstruct=struct), 2,mean)- b,'f'),'\n')
    ## cat("- comparing sample variance with truth:\n")
    ## for (i in 10^(1:4)){
    ##     cat('    sample size n=',i,' yields Frobenius-norm:',
    ##         norm( var( rmvnorm.prec(i, Q=Q, Lstruct=struct))- Sigma,'f'),'\n')
    ## }
    ## set.seed(14)
    ## cat("For rmvnorm.spam:\n- comparing sample mean with truth:\n")
    ## for (i in 10^(1:4))
    ##   cat('    sample size n=',i,' yields  Frobenius-norm:',
    ##       norm( apply(rmvnorm.spam(i, b, as.spam(Sigma), Lstruct=struct), 2,mean)- b,'f'),'\n')
    ## cat("- comparing sample variance with truth:\n")
    ## for (i in 10^(1:4)){
    ##   cat('    sample size n=',i,' yields Frobenius-norm:',
    ##       norm( var( rmvnorm.spam(i, b, as.spam(Sigma), Lstruct=struct))- Sigma,'f'),'\n')
    ## }
})


test_that("rmvt", {

  set.seed(14)
  tmp1 <- rmvt.spam(10,  Sigma=Sigma, df = 0)
  set.seed(14)
  tmp2 <- rmvnorm.spam(10, Sigma = Sigma)
  expect_equal( tmp1, tmp2)

  set.seed(14)
  tmp1 <- rmvt.spam(10, Sigma = Sigma, df = 4)
  set.seed(14)
  tmp2 <- rmvnorm.spam(10, Sigma = Sigma)/sqrt(rchisq(10, 4)/4)
  expect_equal( tmp1, tmp2)

  set.seed(14)
  tmp1 <- rmvt.spam(10, Sigma = Sigma, df = 4, type = "shifted")
  set.seed(14)
  tmp2 <- sweep(rmvnorm.spam(10, Sigma = Sigma)/sqrt(rchisq(10, df = 4)/4), 2, rep(0, nrow(Sigma)), "+")
  expect_equal( tmp1, tmp2)

  set.seed(14)
  tmp1 <- rmvt(10, Sigma = Sigma, df = 4, type = "shifted")
  set.seed(14)
  tmp2 <- rmvt.spam(10, Sigma = Sigma, df = 4, type = "shifted")
  expect_equal( tmp1, tmp2)
})



require('fields')

context("test-rgrf.R: locations")


tau <- 0
nx <- 49
ny <- nx
field <- rgrf(1, nx=nx, ny=ny, tau=tau, Covariance="cov.nug", theta=1)
quilt.plot(cbind(attr(field,"locs"),z=field), nx=nx, ny=ny)
points(attr(field,"locs"))

expect_error(rgrf(1, nx=3, tau=.5, Covariance="cov.nug", theta=1))

tau <- 0.499
nx <- 4
ny <- nx
field <- rgrf(1, nx=nx, ny=ny, tau=tau, Covariance="cov.nug", theta=c(.2, 1, 1.5))
plot(attr(field,"locs"))
quilt.plot(cbind(attr(field,"locs"),z=field), nx=nx, ny=ny)
points(attr(field,"locs"))

# unit grid:
nx <- 12
ny <- 7
field <- rgrf(1, nx=nx, ny=ny, xlim=c(0,nx), ylim=c(0,ny),  Covariance="cov.nug", theta=1)
quilt.plot(cbind(attr(field,"locs"),z=field), nx=nx, ny=ny)
points(attr(field,"locs"))

# effective tau:
tau <- 0.25
nx <- 12
ny <- 7
field <- rgrf(1, nx=nx, ny=ny, xlim=c(0,nx), ylim=c(0,ny), tau=tau,  Covariance="cov.nug", theta=1)
quilt.plot(cbind(attr(field,"locs"),z=field), nx=nx, ny=ny)
points(attr(field,"locs"))
attr(field,"eff")

tau <- 0.25
nx <- 12
ny <- 7
field <- rgrf(1, nx=nx, ny=ny, xlim=c(0,nx), ylim=c(0,ny), tau=tau,  Covariance="cov.nug", theta=1)
quilt.plot(cbind(round(attr(field,"locs")+.499)-.49,z=field), nx=nx, ny=ny)
points(attr(field,"locs"))
attr(field,"eff")

# min distance checking:  (within 0,1)
tau <- 0.499
nx <- 40
ny <- nx
field <- rgrf(1, nx=nx, ny=ny, tau=tau, Covariance="cov.nug", theta=c(.2, 1, 1.5))

tt <- rdist(attr(field,"locs"))
etau <- attr(field,"effective.tau")[1]

expect_lt( (1/nx-2*etau), min( tt[tt>0]))
expect_lt( max( tt[tt>0]), sqrt(2)*((nx-1)/nx+2*etau))



context("test-rgrf.R: mean")
tau <- 0.499
nx <- 20
ny <- nx
field <- rgrf(1, nx=nx, ny=ny, tau=tau, Covariance="cov.nug", theta=c(.2, 1, 1.5))

locs <- attr(field,"locs")
field <- rgrf(1, nx=nx, ny=ny, tau=tau, X=cbind(1,locs), beta=c(4,1,1), Covariance="cov.nug", theta=c(.2, 1, 1.5))

quilt.plot(cbind(attr(field,"locs"),z=field))



context("test-rgrf.R: samples")


nx <- 12
ny <- nx
set.seed(12)
field <- rgrf(1, nx=nx, ny=ny, Covariance="cov.sph", theta=c(1,.2,.2))

distmat <- rdist( attr(field,"locs"))

set.seed(12)
out <- rmvnorm(1, Sigma=cov.sph(distmat,theta=c(1,.2,.2)))

expect_equal( norm(field-out), 0)


expect_error( rgrf(1, nx=nx, ny=ny, Covariance="cov.sphd", theta=c(1,.2,.2)))
expect_error( rgrf(1, nx=nx, ny=ny, Covariance="cov.sphd", theta=c(1,.2,.2)))
expect_error( rgrf(1, nx=nx, ny=ny, Covariance="cov.sphd", theta=c(1,.2,.2)))
expect_error( rgrf(1, nx=nx, ny=ny, Covariance="cov.sphd", theta=c(1,.2,.2)))




if (F) {
    n <- 2000
    locs <- cbind(runif(n), runif(n))

    # rdist just beats the shit out of it...
#    microbenchmark::microbenchmark(
#                        a=rdist(locs),
#                        b=as.matrix(dist(locs)),
#                        c=nearest.dist(locs, upper=NULL, delta=.5),
#                        d=nearest.dist(locs, upper=NULL, delta=sqrt(2)),
#                        times=10)

    }

# HEADER ####################################################
# This is file spam/R/kronecker.R.                          #
# It is part of the R package spam,                         #
#  --> https://CRAN.R-project.org/package=spam              #
#  --> https://CRAN.R-project.org/package=spam64            #
#  --> https://git.math.uzh.ch/reinhard.furrer/spam         #
# by Reinhard Furrer [aut, cre], Florian Gerber [aut],      #
#    Roman Flury [aut], Daniel Gerber [ctb],                #
#    Kaspar Moesinger [ctb]                                 #
# HEADER END ################################################






kronecker.spam <- function(X,Y,FUN = "*", make.dimnames = FALSE, ...)
{
    ## print("1")
    if(getOption("spam.force64"))
        SS <- .format64()
    else
        SS <- .format32

  if (make.dimnames)
    warning("dimnames not supported within sparse matrices")
  lenx <- length(X)
  leny <- length(Y)
  kronlen <- lenx*leny

  if(is.spam(X)){
    Xdim <- X@dimension
    Xentries <- X@entries
    Xcol <- X@colindices
    Xrow <- X@rowpointers
    if(.format.spam(X)$package == "spam64")
        SS <- .format64()
  }else if (is.vector(X)){
    Xentries <- X
    Xcol <- rep.int(as.integer(1),lenx)
    Xdim <- as.integer(c(lenx,1))
    Xrow <- seq_len(lenx+1)
  } else {
    Xentries <- as.double(t(X))
    Xdim <- dim(X)
    Xcol <- rep.int(as.integer(1:Xdim[2]),Xdim[1])
    Xrow <- seq.int(1, by=Xdim[2], length.out=Xdim[1]+1)
  }

  if(is.spam(Y)){
    Ydim <- Y@dimension
    Yentries <- Y@entries
    Ycol <- Y@colindices
    Yrow <- Y@rowpointers
    if(.format.spam(Y)$package == "spam64")
        SS <- .format64()
  } else if (is.vector(Y)){
    Yentries <- Y
    Ycol <- rep.int(as.integer(1),leny)
    Ydim <- as.integer(c(leny,1))
    Yrow <- seq_len(leny+1)
  } else {
    Yentries <- as.double(t(Y))
    Ydim <- dim(Y)
    Ycol <- rep.int(as.integer(1:Ydim[2]),Ydim[1])
    Yrow <- seq.int(1, by=Ydim[2], length.out=Ydim[1]+1)
  }
  ## kronxy <- new("spam")

    if (FUN=="*") {
      z <- .C64("kroneckermult",
                SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                              SS$signature, SS$signature, "double", SS$signature, SS$signature,
                              "double", SS$signature, SS$signature),

                Xdim[1],
                Xentries,
                Xcol,
                Xrow,

                Ydim[1],
                Ydim[2],
                Yentries,
                Ycol,
                Yrow,

                entries = vector_dc( "double", kronlen),
                colindices = vector_dc( SS$type, kronlen),
                rowpointers = vector_dc( SS$type, Xdim[1]*Ydim[1]+1),

                INTENT = c( "r", "r", "r", "r",
                           "r", "r", "r", "r", "r",
                           "w", "w", "w"),
                NAOK=getOption("spam.NAOK"),
                PACKAGE = SS$package)
        kronxy <- .newSpam( entries = z$entries)
    ## slot(kronxy, "entries", check=FALSE) <-  z$entries
    }else {
        z <- .C64("kroneckerf",
                 SIGNATURE = c(SS$signature, "double", SS$signature, SS$signature,
                               SS$signature, SS$signature, "double", SS$signature, SS$signature,
                               "double", "double", SS$signature, SS$signature),

                 Xdim[1],
                 Xentries,
                 Xcol,
                 Xrow,

                 Ydim[1],
                 Ydim[2],
                 Yentries,
                 Ycol,
                 Yrow,

                 ent1 = vector_dc( "double", kronlen),
                 ent2 = vector_dc( "double", kronlen),
                 colindices = vector_dc( SS$type, kronlen),
                 rowpointers = vector_dc( SS$type, Xdim[1]*Ydim[1]+1),

                 INTENT = c("r", "r", "r", "r",
                            "r", "r", "r", "r", "r",
                            "w", "w", "w", "w"),
                 NAOK=getOption("spam.NAOK"),
                 PACKAGE = SS$package)

        FUN <- match.fun(FUN)
        kronxy <- .newSpam( entries = FUN(z$ent1,z$ent2,...) )
    ## slot(kronxy, "entries", check=FALSE) <-  FUN(z$ent1,z$ent2,...)
    if (z$rowpointers[Xdim[1]*Ydim[1]+1]-1 < prod(Xdim,Ydim))
      warning("Sparseness structure of 'kronecker(X,Y)' preserved when applying 'FUN'.", call.=FALSE)
    }
    kronxy <- .newSpam(
        entries = kronxy@entries,
        colindices = z$colindices,
        rowpointers = z$rowpointers,
        dimension = Xdim*Ydim)
  ## slot(kronxy, "colindices", check=FALSE) <- z$colindices
  ## slot(kronxy, "rowpointers", check=FALSE) <- z$rowpointers
  ## slot(kronxy, "dimension", check=FALSE) <- Xdim*Ydim   # no need to use prod

  return(kronxy)
}


"kronecker.default" <- base::kronecker
setGeneric("kronecker")

setMethod("kronecker","spam", kronecker.spam)
setMethod("kronecker",signature(X="spam",Y="spam"), kronecker.spam)
setMethod("kronecker",signature(X="spam",Y="ANY"), kronecker.spam)
setMethod("kronecker",signature(X="ANY",Y="spam"), kronecker.spam)

---
title: "Illustrations and Examples"
author: "Reinhard Furrer, Roman Flury"
date: "`r Sys.Date()`"
output:
  rmarkdown::pdf_document:
    fig_caption: yes
    number_sections: true
    toc: true
    toc_depth: 2
    citation_package: natbib
header-includes:
   - \usepackage{bm}
   - \usepackage{setspace}\onehalfspacing
   - \usepackage[labelfont=bf]{caption}
   - \usepackage{natbib}
   - \usepackage{hyperref}
bibliography: spam.bib
vignette: >
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{`spam`, a SPArse Matrix package}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  # fig.width = 6, fig.height = 6,
  fig.align = "center"
)
options(digits = 3)
```

```{r loadpkgs, echo=FALSE, eval=TRUE, message=FALSE}
library("spam")
```

# Rational for spam

At the core of drawing multivariate normal random variables, calculating or maximizing multivariate normal log-likelihoods, calculating determinants, etc., we need to solve a large (huge) linear system involving a variance matrix, i.e., a symmetric, positive definite matrix.

Assume that we have such a symmetric, positive definite matrix $\boldsymbol{Q}$ of size $n\times n$ that contains many zeros (through tapering, @Furr:Gent:Nych:06, @Furr:Beng:07, or through a Markovian conditional construction).
Typically, $\boldsymbol{Q}$ contains only $\mathcal{O}(n)$ non-zero elements compared to $\mathcal{O}(n^2)$ for a regular, **full** matrix.
To take advantage of the few non-zero elements, special structures to represent the matrix are required, i.e., only the positions of the non-zeros and their values are kept in memory.
Further, new algorithms work with these structures are required.
The package `spam` provides this functionality, see @Furr:Sain:10 for a detailed exposition.


# A Simple Example

This first section illustrates with a simple example how to work with `spam`.
Within a running R we install and load the current `spam` version from CRAN.
```{r echo=TRUE, eval=FALSE, message=FALSE}
install.packages("spam")

library("spam")
```

We create a trivial matrix and "coerce" it to a sparse matrix.
```{r trivial}
Fmat <- matrix(c(3, 0, 1, 0, 2, 0, 1, 0, 3), nrow = 3, ncol = 3)
Smat <- as.spam(Fmat)
```

`spam` is conceptualized such that for many operations, the user proceeds as with ordinary full matrices.
For example:
```{r operations}
Fmat
Smat
Smat %*% t(Smat)
Fmat %*% t(Smat)
```

Hence, the user should not be worried which objects are sparse matrices and which are not.
Of course not all operations result in sparse objects again,
```{r nonspam}
rep(1, 3) %*% Smat
```

However, other operations yield to different results when applied to full or sparse matrices
```{r diffbehaviour}
range(Fmat)
range(Smat)
```


# Creating Sparse Matrices

The implementation of `spam` is designed as a trade-off between the following competing philosophical maxims.
It should be competitively fast compared to existing tools or approaches in R and it should be easy to use, modify and extend.
The former is imposed to assure that the package will be useful and used in practice.
The latter is necessary since statistical methods and approaches are often very specific and no single package could cover all potential tools.
Hence, the user needs to understand quickly the underlying structure of the implementation of `spam` and to be able to extend it without getting desperate.
(When faced with huge amounts of data, sub-sampling is one possibility; using `spam` is another.)
This philosophical approach also suggests trying to assure `S3` and `S4` compatibility, @Cham:93, see also @Luml:04.
`S4` has higher priority but there are only a handful cases of `S3` discrepancies, which do however not affect normal usage.

To store the non-zero elements, `spam` uses the "old Yale sparse format".
In this format, a (sparse) matrix is stored with four elements (vectors), which are (1) the nonzero values row by row, (2) the ordered column indices of nonzero values, (3) the position in the previous two vectors corresponding to new rows, given as pointers, and (4) the column dimension of the matrix.
We refer to this format as compressed sparse row (CSR) format.
Hence, to store a matrix with $z$ nonzero elements we thus need $z$ reals and $z+n+2$ integers compared to $n \times n$ reals. Section Representation describes the format in more details.

The package `spam` provides two classes, first, `spam` representing sparse matrices and, second, `spam.chol.NgPeyton` representing Cholesky factors.
A class definition specifies the objects belonging to the class, these objects are called slots in R and accessed with the `@` operator, see \citet{Cham:93} for a more thorough discussion.
The four vectors of the CSR representation are implemented as slots.
In `spam`, all operations can be performed without a detailed knowledge about the slots.
However, advanced users may want to work on the slots of the class `spam` directly because of computational savings, for example, changing only the contents of a matrix while maintaining its sparsity structure, see Section Tips.
The Cholesky factor requires additional information (e.g., the used permutation) hence the class `spam.chol.NgPeyton` contains more slots, which are less intuitive.
There are only very few, specific cases, where the user has to access these slots directly.
Therefore, user-visibility has been disregarded for the sake of speed.
The two classes are discussed in the more technical Section Representation.


# Displaying

As seen in Section A Simple Example, printing small matrices result in an expected output, i.e., the content of the matrix plus a line indicating the class of the object:
```{r displ1}
Smat
```

For larger objects, not all the elements are printed. For example:
```{r displ2}
diag.spam(100)
```

The size of the matrix when switching from the first printing format to the second is a `spam` option, see Section Options.
Naturally, we can also use the `str` command which gives us further insight into the individual slots of the `spam` object:
```{r disp3}
str(Smat)
```

Alternatively, calling `summary` gives additional information of the matrix.
```{r disp4}
summary(Smat)
```
`summary` itself prints on the standard output but also returns a list containing the number of non-zeros (`nnz`) and the density (`density`) (percentage of `nnz` over the total number of elements).

Quite often, it is interesting to look at the sparsity structure of a sparse matrix.
This is implemented with the command `display`.
Again, depending on the size, the structure is shown as proper rectangles or as points.
Figure \ref{fig:display_spam} is the result of the following code.
```{r disp5, echo=FALSE, warning=FALSE}
nz <- 2^12
Smat1 <- spam(0, nz, nz)
Smat1[cbind(sample(nz, nz), sample(nz, nz))] <- rnorm(nz)

tmp <- round(summary(Smat1)$density, 3)
```

```{r disp6, echo=FALSE, warning=FALSE, fig.cap = '\\label{fig:display_spam}Sparsity structure of sparse matrices.'}
par(mfcol=c(1,2),pty='s',mai=c(.8,.8,.2,.2))
display(Smat)
display(Smat1)
```

Additionally, compare Figures of this document.
Depending on the `cex` value, the `image` may issue a warning, meaning that the dot-size is probably not optimal.
In fact, visually the density of the matrix `Smat1` seems to be around some percentage whereas the actual density is `r tmp`.


The function `image` goes beyond the structure of the matrix by using a specified color scheme for the values.
Labels can be added manually or with `image.plot` from the package `fields`.


# Solving Linear Systems

To be more specific about one of `spam`'s main features, assume we need to calculate $\boldsymbol{A}^{-1}\boldsymbol{b}$ with $\boldsymbol{A}$ a symmetric positive definite matrix featuring some sparsity structure, which is usually accomplished by solving $\boldsymbol{A}\boldsymbol{x}=\boldsymbol{b}$.
We proceed by factorizing $\boldsymbol{A}$ into $\boldsymbol{R}^T\boldsymbol{R}$, where $\boldsymbol{R}$ is an upper triangular matrix, called the Cholesky factor or Cholesky triangle of $\boldsymbol{A}$, followed by solving $\boldsymbol{R}^T\boldsymbol{y}=\boldsymbol{b}$ and $\boldsymbol{R}\boldsymbol{x}=\boldsymbol{y}$, called forwardsolve and backsolve, respectively.
To reduce the fill-in of the Cholesky factor $\boldsymbol{R}$, we permute the columns and rows of $\boldsymbol{A}$ according to a (cleverly chosen) permutation $\boldsymbol{P}$, i.e., $\boldsymbol{U}^T\boldsymbol{U}=\boldsymbol{P}^T\boldsymbol{A}\boldsymbol{P}$, with $\boldsymbol{U}$ an upper triangular matrix.
There exist many different algorithms to find permutations which are optimal for specific matrices %tridiagonal matrices finite element/difference matrices defined on square grids or at least close to optimal with respect to different criteria.
Note that $\boldsymbol{R}$ and $\boldsymbol{U}$ cannot be linked through $\boldsymbol{P}$ alone.
The right panel of Figure \ref{fig:tree} illustrates the factorization with and without permutation.
For solving a linear system the two triangular solves are performed after the factorization.
The determinant of $\boldsymbol{A}$ is the squared product of the diagonal elements of its Cholesky factor $\boldsymbol{R}$.
Hence the same factorization can be used to calculate determinants (a necessary and computational bottleneck in the computation of the log-likelihood of a Gaussian model), illustrating that it is very important to have a very efficient integration (with respect to calculation time and storage capacity) of the Cholesky factorization.
In the case of GMRF, the off-diagonal non-zero elements correspond to the conditional dependence structure.
However, for the calculation of the Cholesky factor, the values themselves are less important than the sparsity structure, which is often represented using a graph with edges  representing the non-zero elements or using a "pixel" image of the zero/non-zero structure, see Figure \ref{fig:tree}.


```{r }
i <- c(2, 4, 4, 5, 5)
j <- c(1, 1, 2, 1, 3)

A <- spam(0, nrow = 5, ncol = 5)
A[cbind(i, j)] <- rep(0.5, length(i))
A <- t(A) + A + diag.spam(5)
A

U <- chol(x = A)
```

```{r tree, echo=FALSE, results = 'markup', out.width='40%', fig.show='hold', warning=FALSE, fig.cap='On the left side the associated graph to the matrix $\\boldsymbol{A}$ is visualized. The nodes of the graph are labeled according to $\\boldsymbol{A}$ (upright) and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ (italics). On the right side the sparsity structure of $\\boldsymbol{A}$ and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ (top row) and the Cholesky factors $\\boldsymbol{R}$ and $\\boldsymbol{U}$ of $\\boldsymbol{A}$ and $\\boldsymbol{P}^T\\boldsymbol{A}\\boldsymbol{P}$ respectively are given in the bottom row. The dashed lines in $\\boldsymbol{U}$ indicate the supernode partition.'}

knitr::include_graphics(c('figures/tree.png', 'figures/ill.png'))
```


The Cholesky factor of a banded matrix is again a banded matrix.
But arbitrary matrices may produce full Cholesky factors.
To reduce this so-called *fill-in* of the Cholesky factor $\boldsymbol{R}$, we permute the columns and rows of $\boldsymbol{A}$ according to a (cleverly chosen) permutation $\boldsymbol{P}$, i.e., $\boldsymbol{U}^T\boldsymbol{U}=\boldsymbol{P}^T\boldsymbol{A}\boldsymbol{P}$, with $\boldsymbol{U}$ an upper triangular matrix.
There exist many different algorithms to find permutations which are optimal for specific matrices or at least close to optimal with respect to different criteria.
The cost of finding a good permutation matrix $\boldsymbol{P}$ is at least of order $\mathcal{O}(n^{3/2})$.
However, there exist good, but suboptimal, approaches for $\mathcal{O}(n\log(n))$.

A typical Cholesky factorization of a sparse matrix consists of the steps illustrated in the following pseudo code algorithm.

Step | Description
---- | ---------------------------------------------------------------------------------------------------------------------------
[1]  | Determine permutation and permute the input matrix $\boldsymbol{A}$ to obtain $\boldsymbol{P}^T\boldsymbol{A}\boldsymbol{P}$
[2]  | Symbolic factorization, where the sparsity structure of $\boldsymbol{U}$ is constructed
[3]  | Numeric factorization, where the elements of $\boldsymbol{U}$ are computed


When factorizing matrices with the same sparsity structure Step 1 and 2 do not need to be repeated.
In MCMC algorithms, this is commonly the case, and exploiting this shortcut leads to very considerable gains in computational efficiency (also noticed by @Rue:Held:05, page 51).
However, none of the existing sparse matrix packages in R (`SparseM`, `Matrix`) provide the possibility to carry out Step 3 separately and `spam` fills this gap.

As for Step 1, there are many different algorithms to find a permutation, for example, the multiple minimum degree (MMD) algorithm, @Liu:85, and the reverse Cuthill-McKee (RCM) algorithm, @Geor:71.
The resulting sparsity structure in the permuted matrix determines the sparsity structure of the Cholesky factor.
As an illustration, Figure \ref{fig:fillin} shows the sparsity structure of the Cholesky factor resulting from an MMD, an RCM, and no permutation of a precision matrix induced by a second-order neighbor structure of the US counties.

How much fill-in with zeros is present depends on the permutation algorithm, in the example of Figure \ref{fig:fillin} there are $146735$, $256198$ and $689615$ non-zero elements in the Cholesky factors resulting from the MMD, RCM, and no permutation, respectively.
Note that the actual number of non-zero elements of the Cholesky factor may be smaller than what the constructed sparsity structure indicates.
Here, there are $14111$, $97565$ and $398353$ zero elements (up to machine precision) that are not exploited.

```{r fillin, echo=FALSE, fig.show='hold', out.width='\\textwidth', warning=FALSE, fig.cap='Sparsity structure of the Cholesky factor with MMD, RCM and no permutation of a precision matrix induced by a second-order neighbor structure of the US counties. The values *nnzR* and *fillin* are the number of non-zero elements in the sparsity structure of the factor and the fill-in, respectively.'}

knitr::include_graphics(c('figures/fig_ch2_factors.png'))
```


# More about Methods

For both sparse classes of `spam` standard methods like `plot`, `dim`, `backsolve`/`forwardsolve`, `determinant` (based on a Cholesky factor) are implemented and behave as in the case of full matrices.
Print methods display the sparse matrix as a full matrix in case of small matrices and display only the non-zero values otherwise.
The corresponding cutoff value as well as other parameters can be set and read via `spam.options`.


## Methods with Particular Behavior

For the `spam` class additional methods are defined, for examples `rbind`/`cbind`, `dim<-`, etc.
The group generic functions from `Math`, `Math2` and `Summary` are treated particularly since they operate only on the nonzero entries of the `spam` class.
For example, for the matrix `A` presented in the introduction `range(A)` is the vector `c(0.5, 1)`, i.e. the zeros are omitted from the calculation.
The help lists further available methods and highlights the (dis-)similarities compared to when applied to regular matrices or arrays.


## Particular Methods with Ordinary Behavior

Besides the two sparse classes mentioned above, `spam` does not maintain different classes for different types of sparse matrices, such as symmetric or diagonal matrices.
Doing so would result in some storage and computational gain for some matrix operations, at the cost of user visibility.
Instead of creating more classes we consider additional specific operators.
As an illustration, consider multiplying a diagonal matrix with a sparse matrix.
The operator `\%d*\%` uses standard matrix multiplication if both sides are matrices or multiplies each column according the diagonal entry if the left hand side is a diagonal matrix represented by vector.


## Importing Foreign Formats

`spam` is not the only R package for sparse matrix algebra.
The packages `SparseM` @Koen:Ng:03 and `Matrix` @Bates:06 contain similar functionalities for handling sparse matrices, however, both packages do not provide the possibility to split up the Cholesky factorization as discussed previously.
We briefly discuss the major differences with respect to `spam`; for a detailed description see their manual.

`SparseM` is also based on the Fortran Cholesky factorization of @Ng:Peyt:93 using the MMD permutation and almost exclusively on `SparseKit`.
It was originally designed for large least squares problems and later also ported to `S4` but is in a few cases inconsistent with existing R methods.
It supports different sparse storage systems.

`Matrix` incorporates many classes for sparse and full matrices and is based on C.
For sparse matrices, it uses different storage formats, defines classes for different types of matrices and uses a Cholesky factorization based on UMFPACK, @Davi:04.

`spam` has a few functions that allow to transform matrix formats of the different packages.

`spam` also contains functions that download matrices from MatrixMarket, a web side that stores many different sparse matrices.
The function `read.MM(file)`, very similar to the function `readMM` from `Matrix`, opens a connection, specified by the argument, and reads a matrix market file. However, as entries of `spam` matrices are of mode `double`, integers matrices are coerced to doubles, patterns lead to matrices containing ones and complex are coerced to the real part thereof.
In these aforementioned cases, a warning is issued.

MatrixMarket also defines an array format, in which case a (possibly) dense `spam` object is return (retaining only elements which are larger than `getOption('spam.eps')`), a warning is issued.

Similarly to `read.MM(file)`, the function `read.HB(file)` reads matrices in the Harwell-Boeing format.
Currently, only real assembled Harwell-Boeing can be read with `read.HB`.
Reading MatrixMarket formats is more flexible.

The functions are based on `readHB` and `readMM` from the library `Matrix` to build the connection and read the raw data.
At present, `read.MM(file)` is more flexible than `readMM`.

For many operations, `spam` is faster than `Matrix` and `SparseM`.
It would also be interesting to compare `spam` and the sparse matrix routines of `Matlab` (see Figure 6 of @Furr:Gent:Nych:06 for a comparison between `SparseM` and Matlab).


# Bibliography

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/complexity.Rd.                      %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{complexity}
\alias{complexity}
\alias{complexities}      
\title{Complexity for Sparse Matrices}
\description{A few results of computational complexities for selected
  sparse algoritms in \code{spam}
}
\details{A Cholesky factorization of an n-matrix requires n^3/3 flops. In case of
  banded matrices (bandwidth p, p<<n) a factorization requires about 2np^2
  flops. Forward- and backsolves for banded matrices require essentially
  2np flops.

  George and Liu (1981) proves that any reordering would require at
  least O(n^3/2) flops for the factorization and produce at least O(n
  log(n)) fill-ins for square lattices with a local neighbor hood.\cr
  They also show that algorithms based on nested dissection are optimal
  in the order of magnitude sense.

  More to follow.

  



%  It is also possible to furnish a specific
%permutation in which case \code{pivot} is a vector. For compatibility
%reasons, \code{pivot} can also take a logical in which for \code{FALSE}
%no permutation is done and for \code{TRUE} is equivalent to
%\code{MMD}.\cr%

%The exact operation count is often very difficult to determine. Easier
%bounds are obtained in the case of banded matrices.

}
\references{ 
%Ng, E. G. and B. W. Peyton (1993) Block sparse Cholesky algorithms
%on advanced uniprocessor computers, \emph{SIAM J. Sci. Comput.}, \bold{14},
%1034--1056.
%
George, A. and  Liu, J. (1981)
\emph{Computer Solution of Large Sparse Positive Definite Systems},
Prentice Hall.
}
\seealso{\code{\link{det}}, \code{\link{solve}}, 
  \code{\link{forwardsolve}}, \code{\link{backsolve}} and \code{\link{ordering}}.
}
%\examples{
%}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam.internal.Rd.                   %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
\name{spam internal}
\alias{printSize}
\title{
  Spam internal and auxiliary functions
}
\description{
  The functions listed below are auxiliary functions and are not exported by the NAMESPACE. 
  The user should not require to call these directly. 
}
\details{\code{printSize(x, size=8, digits=2L)} constructs a string for
  pretty printing the object size \code{x}. \cr

}
\keyword{internal}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/grid_zoom.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%







\name{grid_zoom}
\alias{grid_zoom}
\title{grid_zoom}
\description{This function takes a grob object (e.g. created with
  package grid) and adds a zoom window. }
\usage{
grid_zoom(inputGrob = pointsGrob(runif(200),runif(200)),
          inputViewport = viewport(name='main'),
          x = 'topleft', y, just,
          ratio = c(.3,.4), zoom_xlim, zoom_ylim,
          rect = TRUE, rect_lwd = 1, rect_fill = 'gray92',
          draw =TRUE, zoom_fill = 'white',
          zoom_frame_gp = gpar(lwd = 1),
          zoom_gp = NULL, zoom_xaxis = xaxisGrob(main = FALSE),
          zoom_yaxis = NULL)
}
\arguments{
  \item{inputGrob}{A grob object, e.g created with package grid.}
  \item{inputViewport}{Viewport related to \code{inputGrob}.}
  \item{x}{Specifies the \code{x} coordinate of the zoom
    window. Alternatively it can be set to 'topleft', 'topright',
    'bootmleft' or 'bootmright'}
  \item{y}{Specifies the \code{y} coordinate of the zoom
    window. }
  \item{just}{Specifies the justification of the zoom window. }
  \item{ratio}{Specifies size of the zoom window relative to the main window. }
  \item{zoom_xlim}{Specifies xlim value of the zoom window. }
  \item{zoom_ylim}{Specifies ylim value of the zoom window. }
  \item{rect}{Logical, if TRUE a rectangle of the zoom region is draw in the main window. }
  \item{rect_lwd}{lwd of the rectangle. }
  \item{rect_fill}{fill of the rectangle. }
  \item{draw}{logical, if TRUE the returned grob object is also drawn.}
  \item{zoom_fill}{fill color of the zoom window.}
  \item{zoom_frame_gp}{gpar() of the frame of the zoom window.}
  \item{zoom_gp}{gpar() of the inputGrob in the zoom viewport.}
  \item{zoom_xaxis}{xaxisGrob() to draw for the zoom window.}
  \item{zoom_yaxis}{yaxisGrob() to draw for the zoom window.}
}
\value{A grob object.}
\details{
A zoom plot does only make sense if all objects of the \code{inputGrob}
are specified in \code{native} units. Additional caution me be require
for certain grobs: e.g. a zoom of a circleGrob() is problematic if the x
and y axis are stretched by a different amount.

}
%\references{}
\seealso{grid_trace2}
\examples{
require(grid)
## -- Example 1 --
set.seed(133)
grid_zoom(inputGrob = pointsGrob(runif(200), runif(200)),
          inputViewport = viewport(name = 'main'),
          zoom_xlim = c(.2, .3), zoom_ylim = c(.2, .3))


## -- Example 2 --
## initial plot
grid.newpage()
vp <- viewport(width=.8, height=.8, clip='on')
gt <- gTree(children=gList(polylineGrob(x=c((0:4)/10, rep(.5, 5), (10:6)/10, rep(.5, 5)),
              y=c(rep(.5, 5), (10:6/10), rep(.5, 5), (0:4)/10),
              id=rep(1:5, 4), default.units='native',
              gp=gpar(col=1:5, lwd=3)),
              pointsGrob(runif(1000), runif(1000),pch='.', gp=gpar(cex=3)),
              rectGrob(gp=gpar(lwd=3))))
pushViewport(vp)
grid.draw(gt)

## plot with zoom window
grid.newpage()
grid_zoom(inputGrob = gt,
          inputViewport = vp,
          x='topright', zoom_xlim=c(.6,.73), zoom_ylim=c(.3,.43),ratio=.4,
          zoom_xaxis = NULL, zoom_gp = gpar(cex=3))
}
\author{Florian Gerber <florian.gerber@math.uzh.ch>}
\keyword{zoom_grid}
\keyword{zoom}
\keyword{plot}
\keyword{grid}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/xybind.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{cbind}
\alias{cbind.spam}
\alias{rbind.spam}
\alias{cbind}
\alias{rbind}
\alias{cbind,spam-method}
\alias{rbind,spam-method}
\title{Combine Sparse Matrices by Rows or Columns}
\description{Take a sequence of vector, matrix or \code{spam} object arguments and
     combine by \emph{c}olumns or \emph{r}ows, respectively.}
\usage{
# cbind(\dots, force64 = getOption("spam.force64"), deparse.level = 0)
# rbind(\dots, deparse.level = 0)
}
\arguments{
  \item{...}{vectors, matrices or \code{spam} objects. See \sQuote{Details} and \sQuote{Value}}
  \item{force64}{logical vector of length 1. If \code{TRUE}, a 64-bit
    spam matrix is returned in any case. If \code{FALSE}, a 32-bit
    matrix is returned when possible. }
  \item{deparse.level}{for compatibility reason here. Only \code{0} is implemented.}
}
\value{a \code{spam} object combining the \code{\dots} arguments
  column-wise or row-wise.  (Exception: if there are no inputs or all
  the inputs are \code{NULL}, the value is \code{NULL}.)}
\details{\code{rbind} and \code{cbind} are not exactly symmetric in how
  the objects are processed. \code{cbind} calls a
  Fortran routine after the input has been coerced to \code{spam}
  objects. Whereas \code{rbind} calls a Fortran routine only in the case
  of to \code{spam} matrices. Note that row binding is essentially an concatenation
  of the slots due to the sparse storage format.  \cr

  Only two objects at a time are processed. If more than two are
  present, a loop concatenates them successively. \cr

  A method is defined for a \code{spam} object as first argument. 
    }
%\references{}
%\seealso{\code{\link{cbind,spam-method}}.}
\examples{
x <- cbind.spam(1:5,6)

y <- cbind(x, 7)

rbind( x, x)
# for some large matrices   t( cbind( t(x), t(x)))
# might be slightly faster:


}
\author{Reinhard Furrer}
\keyword{array}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rdist.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{fields-wrapper}
\alias{spam_rdist}
\alias{rdist.dist}
\alias{spam_rdist.earth}
\title{Wrapper for Distance Matrix Computation}
\description{These functions are simple wrappers to \code{nearest.dist}
  to be used in \code{fields}.}
\usage{
spam_rdist( x1, x2, delta = 1)
spam_rdist.earth( x1, x2, delta = 1, miles=TRUE, R=NULL)
}
\arguments{
  \item{x1}{Matrix of first set of locations where each row gives the
          coordinates of a particular point. }
  \item{x2}{Matrix of second set of locations where each row gives the
          coordinates of a particular point.}
  \item{delta}{only distances smaller than \code{delta} are recorded,
    see Details.}
  \item{miles}{For great circle distance: If true distances are in statute miles if false distances in
          kilometers.}
  \item{R}{Radius to use for sphere to find spherical distances. If \code{NULL}
          the radius is either in miles or kilometers depending on the
          values of the miles argument. If \code{R=1} then distances are of
          course in radians.}
}
\value{A \code{spam} object containing the distances spanned between
  zero and \code{delta}. The sparse matrix may contain many zeros
  (e.g., collocated data). However, to calculate covariances, these zeros
  are essential.}
\details{These functions are wrappers to \code{rdist} and
  \code{rdist.earth} in \code{fields}. They are used
  to simplify the use of sparse matrices in functions like \code{mKrig}.
  \cr

  For great circle distance, the matrices \code{x1} and \code{x2}
  contain the degrees longitudes in the first and the degrees latitudes
  in the second column.  \code{delta} is in degrees. Hence to restrict
  to distances smaller than \code{delta.km}, one has to specify
  \code{delta=delta.km*360/(6378.388*2*pi)}.

}
%\references{}
\seealso{\code{\link{nearest.dist}}}
\examples{
\dontrun{
require(fields)
look <- mKrig(x,Y, Covariance="Wendland", dimension=2, k=1,
    cov.args=list( Distance='spam_rdist'))
}

}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/circulant.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{circulant}
\alias{circulant.spam}
\title{Create Circulant Matrices}
\description{Creates a circulant matrix in \code{spam} format.}
\usage{
circulant.spam(x, n = NULL, eps = getOption("spam.eps"))
}
\arguments{
  \item{x}{the first row to form the circulant matrix or a list
    containing the indices and the nonzero values.}
  \item{n}{if \code{x} is a list, the dimension of the matrix.}
  \item{eps}{A tolerance parameter:  elements of \code{x} such that \code{abs(x)
    <= eps} set to zero. Defaults to \code{eps = getOption("spam.eps")}}
}
\value{The circulant matrix in \code{spam} format.}
%\details{The vector \code{y} has to be of the same length as \code{x}
%  and its first element is discarded.
%    }
%\references{}
\seealso{\code{circulant} from package \pkg{magic}, \code{\link{toeplitz.spam}}}
\examples{
circulant.spam(c(1,.25,0,0,0))
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam-package.Rd.                    %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{spam-package}
\alias{spam-package}
\alias{SPAM}
\alias{Spam}
\alias{. SPAM .}
\docType{package}
\title{SPArse Matrix Package}
\description{\code{spam} is a collection of functions for sparse matrix
  algebra.
}
\section{Gereral overview}{What is spam and what is it not:\cr

While \code{Matrix} seems an overshoot of classes and \code{SparseM}
focuses mainly on regression type problem, we provide a minimal set of
sparse matrix functions fully functional for everyday spatial statistics
life. There is however some emphasize on Markov chain Monte Carlo type
calculations within the framework of (Gaussian) Markov random fields.
\cr

Emphasis is given on a comprehensive, simple, tutorial structure of the
code. The code is S4 based but (in a tutorial spirit) the functions are
in a S3 structure visible to the user (exported via \code{NAMESPACE}).\cr

There exist many methods for sparse matrices that work identically as in
the case of ordinary matrices. All the methods are discussed in the help
and can be accessed directly via a \code{*.spam} concatenation to the
function. For example, \code{help(chol.spam)} calls the help
directly. We deliberately avoided aliases according to analogue helps from the
base package.\cr

Sparseness is used when handling large matrices. Hence, care has been
used to provide efficient and fast routines. Essentially, the functions
do not transform the sparse structure into full matrices to use standard
(available) functionality, followed by a back transform. We agree, more
operators, functions, etc. should eventually be implemented.


The packages \code{fields} and \code{spam} are closely linked.  \cr

}
\references{Reinhard Furrer, Stephan R. Sain (2010).
  "spam: A Sparse Matrix R Package with Emphasis on MCMC
  Methods for Gaussian Markov Random Fields.",
        \emph{Journal of Statistical Software}, 36(10), 1-25,
        \doi{10.18637/jss.v036.i10}.\cr
Florian Gerber, Reinhard Furrer (2015).
        "Pitfalls in the Implementation of Bayesian Hierarchical Modeling of Areal Count Data: An Illustration Using BYM and Leroux Models.",
        \emph{Journal of Statistical Software}, Code Snippets, 63(1), 1-32,
        \doi{10.18637/jss.v063.c01}.\cr
F. Gerber, K. Moesinger, R. Furrer (2017),
        "Extending R packages to support 64-bit compiled code: An illustration with spam64 and GIMMS NDVI3g data.",
        \emph{Computer & Geoscience} 104, 109-119, \doi{10.1016/j.cageo.2016.11.015}."
}

\seealso{See \code{\link{spam.class}} for a detailed class description,
  \code{\link{spam}} and  \code{\link{spam.ops}} for creation,
  coercion and algebraic operations. \code{\link{options}}.%\cr%
%
%\code{demo(package='spam')} lists available demos.\cr%
%Related packages are \code{\link[fields]{fields}},
%  \code{\link[Matrix]{Matrix}} and
%  \code{SparseM.ontology}.
}
\examples{
## Citations:
## IGNORE_RDIFF_BEGIN
citation('spam')
citation('spam', auto=TRUE)
## IGNORE_RDIFF_END


## History of changes
\dontrun{
file.show(system.file("NEWS.md", package = "spam"))
}
}
\author{Reinhard Furrer, with the help of Florian Gerber, Kaspar
  Moesinger and many others. \cr
  Some Fortran routines were written by Youcef Saad, Esmond G. Ng, Barry W. Peyton, Joseph W.H. Liu, Alan D. George.
}
\keyword{documentation}
\keyword{package}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/version.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{version}
\alias{version}
\alias{spam.version}
\alias{spam.Version}
\title{Spam Version Information}
\description{
  \code{spam.version} is a variable (\code{list}) holding
  detailed information about the version of
  \code{spam} loaded.

  \code{spam.Version()} provides detailed information about the version of \code{spam}
     running.

  
}
\usage{
spam.version
}
\value{\code{spam.version} is a list with character-string components
  \item{status}{the status of the version (e.g., \code{"beta"})}

  \item{major}{the major version number}
  
  \item{minor}{the minor version number}
  
  \item{year}{the year the version was released}
  
  \item{month}{the month the version was released}
  
  \item{day}{the day the version was released}
  
  \item{version.string}{a \code{character} string concatenating
     the info  above, useful for plotting, etc.}

  \code{spam.version} is a list of class \code{"simple.list"} which
     has a \code{print} method.
}
% \references{}
\seealso{See the R counterparts  \code{\link[base]{R.version}}.
}
\author{Reinhard Furrer}
\examples{
spam.version$version.string
}
\keyword{environment}
\keyword{sysdata}
\keyword{programming}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/cleanup.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{cleanup}
\alias{cleanup}
\title{Cleaning up sparse matrices}
\description{
  Eliminates an zeros in a sparse matrix.}
\usage{
cleanup(x, eps = getOption("spam.eps"))
}
\arguments{
\item{x}{a sparse matrix of class \code{spam}.}
\item{eps}{numeric scalar > 0.  Smaller entries are coerced to zero.}
}
\details{A sparse matrix may still contain zeros. This function (aliased
  to \code{as.spam}) filters these values.\cr
  This often causes confusion when testing such matrices for symmetry
  or comparing apparently equal matrices with \code{all.equal}
  (see \sQuote{Examples} below.
}
\seealso{\code{\link{isSymmetric.spam}} and \code{\link{all.equal.spam}}.
}
\examples{
A <- diag.spam(2)
A[1,2] <- 0

all.equal(A, t(A))
isSymmetric.spam(A)
all.equal(cleanup(A), diag.spam(2))
}
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/toeplitz.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{toeplitz}
\alias{toeplitz.spam}
\title{Create Toeplitz Matrices}
\description{Creates symmetric and asymmetric Toeplitz matrices.}
\usage{
toeplitz.spam(x, y = NULL, eps = getOption("spam.eps"))
}
\arguments{
  \item{x}{the first row to form the Toeplitz matrix.}
  \item{y}{for asymmetric Toeplitz matrices, this contains the first
    column.}
  \item{eps}{A tolerance parameter:  elements of \code{x} such that \code{abs(x)
    <= eps} set to zero. Defaults to \code{eps = getOption("spam.eps")}.} 
}
\value{The Toeplitz matrix in \code{spam} format.}
\details{The vector \code{y} has to be of the same length as \code{x}
  and its first element is discarded. 
    }
%\references{}
\seealso{\code{\link{toeplitz}}, \code{\link{circulant.spam}}}
\examples{
toeplitz.spam(c(1,.25,0,0,0))

}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam.internal.Rd.                   %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{spam internal}
\alias{dcheck}
\alias{icheck}
\title{
  Spam internal and auxiliary functions
}
\description{
  The functions listed below are auxiliary functions but are 
  exported by the NAMESPACE. The user should not
  require to call these directly. 
}
\details{The functions are listed here for a better understanding of the
  code (to fulfill the tutorial style paradigm).\cr

  \code{validspamobject(object)} A few sanity checks if \code{object}
  is a proper \code{spam} object.

  \code{dcheck(x)}, \code{icheck(x)} testing and forcing of \code{x}
  to doubles and integers. These functions are used when calling Fortran
  routines.
}
%\note{The integers \code{int0}, \code{int1} and \code{int2} (0-2) are not
%  exported. 
%}
\keyword{internal}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/allequal.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





\name{allequal}
\alias{all.equal.spam}
\alias{all.equal,matrix,spam-method}
\alias{all.equal,spam,matrix-method}
\alias{all.equal,spam,spam-method}
\title{Test if Two Sparse Matrices are (Nearly) Equal}
\description{Utility to compare two \code{spam} objects
     testing 'near equality'.  Depending on the type of difference, comparison is
     still made to some extent, and a report of the differences is
     returned.}
\usage{

\S3method{all.equal}{spam}(target, current, tolerance = .Machine$double.eps^0.5,
    scale = NULL, check.attributes = FALSE,...)
}
\arguments{
  \item{target}{a \code{spam} object.}
  \item{current}{another \code{spam} object to be compared with \code{target}.}
  \item{tolerance}{numeric >= 0.  Differences smaller than
    \code{tolerance} are not considered.}
  \item{scale}{numeric scalar > 0 (or \code{NULL}). See \sQuote{Details}.}
  \item{check.attributes}{currently not yet implemented.}
  \item{...}{Further arguments for different methods.}
}
\value{Either \code{TRUE} or a vector of 'mode' \code{"character"} describing the
  differences between \code{target} and \code{current}.
}
\details{
   Numerical comparisons for \code{scale = NULL} (the default) are
   typically on a \emph{relative difference} scale unless the
   \code{target} values are close to zero or infinite. Specifically,
   the scale is computed as the average absolute value of \code{target}.
   If this scale is finite and exceeds \code{tolerance}, differences
   are expressed relative to it; otherwise, absolute differences are used.

   If \code{scale} is numeric (and positive), absolute comparisons are
   made after scaling (dividing) by \code{scale}. Note that if all of
   scale is sufficiently close to 1 (specifically, within \code{tolerance}),
   the difference is still reported as being on an absolute scale.

   Do not use \code{all.equal.spam} directly in \code{if}
   expressions: either use \code{isTRUE( all.equal.spam(...))} or
   \code{identical} if appropriate.


  Cholesky decomposition routines use this function to test for
  symmetry.

  A method for \code{matrix-spam} objects is defined as well.


  There is the additional catch of a zero matrix being represented by
  one zero element, see \sQuote{Examples} below.
 }
\seealso{\code{\link{isSymmetric.spam}} and \code{\link{cleanup}}.}
\examples{
obj <- diag.spam(2)
obj[1,2] <- .Machine$double.eps

all.equal( diag.spam(2), obj)

all.equal( t(obj), obj)

all.equal( t(obj), obj*1.1)

# We can compare a spam to a matrix
all.equal(diag(2),diag.spam(2))

# the opposite does often not make sense,
# hence, it is not implemented.
all.equal(diag.spam(2),diag(2))


# A zero matrix contains one element:
str(spam(0))
# hence
all.equal.spam(spam(0,3,3), diag.spam(0,3) )
norm(spam(0,3,3) - diag.spam(0,3) )
}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/coercion.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{Coercion}
\docType{class}
\alias{as.vector}
\alias{as.vector.spam}
\alias{as.vector,spam-method}
\alias{as.vector,spam.chol.NgPeyton-method}
\title{Coercion to a Vector}
\description{Coercion of \code{spam} matrices to proper vector objects
}
\usage{\S4method{as.vector}{spam}(x, mode = "any")
}
\arguments{\item{x}{spam object.}
\item{mode}{character string naming an atomic mode or \code{"any"}/\code{"list"}/\code{"expression"}.}
}
\value{If \code{structurebased=TRUE}, the vector \code{x@entries}.\cr
  Conversely, if \code{structurebased=FALSE}, the result is identical to
  one with \code{as.vector(as.matrix(x))}.
}
\details{This coercion allows smooth transitions between different
  matrix formats, see example below.\cr
The Cholesky factors are first transformed to a \code{spam} object.
}
%\references{
%}
\seealso{\code{\link{spam.options}}}
\examples{
x <- diag(2)
ifelse( x, x, 1-x)
ifelse( x, as.vector(x), 1-as.vector(x))

x <- diag.spam(2)
options(spam.structurebased=FALSE)
ifelse( x, as.vector(x), 1-as.vector(x))
options(spam.structurebased=TRUE)
ifelse( x, as.vector(x), 1-as.vector(x))
}
\author{Reinhard Furrer}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/grid_trace2.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{grid_trace2}
\alias{grid_trace2}
\title{Two trace plots and a scatter plot.}
\description{For two (MCMC) chains of the same length trace plots
  and a scatter plot are drawn. }
\usage{
grid_trace2(chain1, chain2 = NULL,
            xlim = NULL, ylim1 = NULL, ylim2=NULL,
            chain1_lab = "", chain2_lab = "", main = "",
            chain1_yaxis_at = NULL, chain2_yaxis_at = NULL,
            log = FALSE,
            cex_points = unit(0.2, "mm"),
            cex_main = unit(1.2, "mm"),
            lwd_lines = unit(0.1, "mm"),
            lwd_frame = unit(0.8, "mm"),
            draw = TRUE)
}
\arguments{
\item{chain1}{Numeric vector or a matrix with two columns.}
\item{chain2}{Numeric vector of same length as \code{chain1}. (Only used if \code{chain1} is specified as vector).}
\item{xlim}{x axis limits of both chains (numeric vector of length 2). }
\item{ylim1}{y axis limits of chain 1 (numeric vector of length 2).}
\item{ylim2}{y axis limits of chain 2 (numeric vector of length 2).}
\item{chain1_lab}{Label of chain 1 (character of length 1). }
\item{chain2_lab}{Label of chain 2 (character of length 1). }
\item{main}{Title (character of length 1). }
\item{chain1_yaxis_at}{Points at which tick-marks are drawn for chain 1.}
\item{chain2_yaxis_at}{Points at which tick-marks are drawn for chain 2. }
\item{log}{Logical, should the date be log transformed?}
\item{cex_points}{Size of points in scatter plot. }
\item{cex_main}{Size of the title font. }
\item{lwd_lines}{Line width of trace plots. }
\item{lwd_frame}{Line width of frames. }
\item{draw}{Logical, should the returned grob object be drawn?}
}
\value{A grob object.}
\details{The figure is optimized for a plot of the format
  x11(width = 10, height = 3).}
%\references{}
\seealso{grid_zoom}
\examples{
grid_trace2(runif(500), runif(500),
            chain1_yaxis_at = seq(.2, 1, by = .2),
            chain1_lab = "chain1", chain2_lab = "chain2",
            main = "Uniform",
            lwd_lines = grid::unit(.5, "mm"))
}
\author{Florian Gerber <florian.gerber@math.uzh.ch>}
\keyword{plot}
\keyword{trace plot}
\keyword{grid}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/coerce.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{coerce-methods}
\docType{methods}
\alias{coerce.spam}
\alias{coerce,spam,logical-method}
\alias{coerce,spam,matrix-method}
\alias{coerce,spam,vector-method}
\alias{coerce,spam,list-method}
\alias{coerce,spam,integer-method}
\title{Force a \code{spam} Object to Belong to a Class}
\description{
  These functions manage the relations that allow coercing a \code{spam} object
  to a given class.
}
\section{Methods}{
\describe{
\item{\code{signature(from = "spam", to = "matrix")}}{
  this is essentially equivalent to \code{as.matrix(object)}.
}
\item{\code{signature(from = "spam", to = "list")}}{
  this is essentially equivalent to \code{triplet(object)}.
}
\item{\code{signature(from = "spam", to = "vector")}}{
  this is essentially equivalent to \code{object@entries}
  (\code{structurebased=TRUE}) or \code{c(object)}.
}
\item{\code{signature(from = "spam", to = "logical")}}{
  the entries are forced to logicals (nonzeros only in case of
  \code{structurebased=TRUE}).
}
\item{\code{signature(from = "spam", to = "integer")}}{
  the entries are forced to integers (nonzeros only in case of
  \code{structurebased=TRUE}).
}
}}
\examples{
ifelse( diag.spam(2)*c(0,1), TRUE, FALSE)
}
\keyword{methods}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/lu.tri.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{lower.tri}
\alias{lower.tri}
\alias{upper.tri}
\alias{lower.tri.spam}
\alias{upper.tri.spam}
\title{Lower and Upper Triangular Part of a Sparse Matrix}
\description{Returns the lower or upper triangular structure or
  entries of  a sparse matrix.
}
\usage{
lower.tri(x, diag = FALSE)
upper.tri(x, diag = FALSE)
}
\arguments{
  \item{x}{a sparse matrix of class \code{spam}}
  \item{diag}{logical. Should the diagonal be included?}
}
\details{Often not only the structure of the matrix is required but the
  entries as well. For compatibility, the default is only a structure
  consisting of ones (representing \code{TRUE}s). Setting the flag
  \code{getOption( "spam.trivalues")} to \code{TRUE},
  the function returns the actual
  values.
  }
\seealso{\code{\link{spam.options}} and \code{\link{diag}}}
\examples{
smat <- spam( c( 1,2,0,3,0,0,0,4,5),3)
upper.tri( smat)
upper.tri( smat, diag=TRUE)

options(spam.trivalues=TRUE)
upper.tri( smat)

}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/large_matrix.Rd.                    %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{large_matrix}
\alias{large_matrix}
\alias{large matrix}
\alias{large_matries}
\alias{large matries}
\alias{64bit}
\alias{64bits}
\alias{long vector}
\alias{long vectors}
\alias{spam64}

\title{Large 64-bit matrices require the R package \pkg{spam64}}
\description{The R package \pkg{spam} can handle sparse matrices with up
  to 2^31-1 non-zero elements. For matrices with more non-zero elements
  it is necessary to load the \pkg{spam64} package in addition.
}
\details{With the help of the R package \pkg{dotCall64} spam interfaces
  either the compiled code with 32-bit integers provided in
  \pkg{spam} or the compiled code with 64-bit integers provided in
  \pkg{spam64}.
  \cr
  To mimick 64-bit behavior, set \code{options(spam.force64 = TRUE)}. The subsequent matrix indices are then stored in double format.
}
\references{
F. Gerber, K. Moesinger, R. Furrer (2017),
Extending R packages to support 64-bit compiled code: An illustration with spam64 and GIMMS NDVI3g data,
Computer & Geoscience 104, 109-119, https://doi.org/10.1016/j.cageo.2016.11.015.
}
\seealso{
 \code{\link[spam64]{spam64-package}},  \code{\link[dotCall64]{dotCall64}}.
}
\examples{
\dontrun{
## the following matrices are very large, and hence,
## require much memory and cpu time.
library("spam64")
s1 <- spam(1, ncol=2^30)        # 32-bit matrix
s1

s2 <- cbind(s1, s1)             # 64-bit matrix
s2

s3 <- spam(1, ncol=2^31)        # 64-bit matrix
s3
}
}
\author{Reinhard Furrer, Florian Gerber, Kaspar Moesinger, Daniel Gerber}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/nearestdist.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{nearestdist}
\alias{dist.spam}
\alias{nearest.dist}
\alias{distance}
\title{Distance Matrix Computation}
\description{This function computes and returns specific elements of distance
     matrix computed by
     using the specified distance measure.}
\usage{
nearest.dist( x, y=NULL, method = "euclidean",
             delta = 1, upper = if (is.null(y)) FALSE else NULL,
             p = 2, miles = TRUE, R = NULL, fortran = FALSE)
}
\arguments{
  \item{x}{Matrix of first set of locations where each row gives the
          coordinates of a particular point. See also \sQuote{Details}.}
  \item{y}{Matrix of second set of locations where each row gives the
          coordinates of a particular point. If this is missing \code{x} is
          used. See also \sQuote{Details}.}
  \item{method}{the distance measure to be used. This must be one of
          \code{"euclidean"}, \code{"maximum"}, \code{"minkowski"} or
                \code{"greatcircle"}. Any unambiguous substring can be
          given.}
  \item{delta}{only distances smaller than \code{delta} are recorded,
    see Details.}
  \item{upper}{Should the entire matrix (\code{NULL}) or only the upper-triagonal (\code{TRUE})
               or lower-triagonal (\code{FALSE}) values be calculated.}
  \item{p}{The power of the Minkowski distance.}
  \item{miles}{For great circle distance: If true distances are in statute miles if false distances in
          kilometers.}
  \item{R}{For great circle distance: Radius to use for sphere to find spherical distances. If \code{NULL}
          the radius is either in miles or kilometers depending on the
          values of the miles argument. If \code{R=1} then distances are of
          course in radians.}
  \item{fortran}{Should the C++ (\code{FALSE}) or the Fortran code (\code{TRUE}) be used. If 64-bit matrices are needed, the argument is set to (\code{TRUE}).}
%  \item{eps}{deprecated. Left for backwards consistency.}
%  \item{diag}{deprecated. Left for backwards consistency. See \sQuote{Details}.}a
}
\value{A \code{spam} object containing the distances spanned between
  zero and \code{delta}. The sparse matrix may contain many zeros
  (e.g., collocated data). However, to calculate covariances, these zeros
  are essential.}

\details{For great circle distance, the matrices \code{x} and \code{y}
  contain the degrees longitudes in the first and the degrees latitudes
  in the second column. \code{delta} is in
  degrees. Hence to restrict to distances smaller than \code{delta.km},
  one has to specify \code{delta=delta.km*360/(6378.388*2*pi)}.

  The distances are calculated based on spherical law of cosines.
  Care is needed for `zero' distances due to the final acosin:
  \code{acos(1-1e-16)}, especially with an actual radius.

%  The distance is
%  in single precision (I am still not sure where I lose the double precision in
%  the Fortran code) and if calculating the entire matrix
%  \code{upper=NULL} (instead of adding its transpose) it may not
%  pass the symmetry checks, for example.\cr

  Default value of Earth's radius is 3963.34miles (6378.388km).\cr

  There are many other packages providing distance functions. Especially
  for great circle distances there are considerable differences between
  the implementations. For high precision results, \code{sp::spDists} is
  a good candidate and distances of large amount of locations can be
  processed in parallel with the \code{parallelDist} package.\cr

  The formerly depreciated arguments \code{eps} and \code{diag} are now
  eliminated.

  \code{x} and \code{y} can be any object with an existing
  \code{as.matrix} method.\cr

%  A quick scan revealed distance functions in at least 7 packages
%  (around 2008). The argument names should be as general as possible and
%  be coherent with many (but not all) available distance functions.\cr

  The Fortran code is based on a idea of Doug Nychka.
    }
%\references{}
\seealso{\code{\link{spam_rdist}}}
\examples{
# Note that upper=T and using t(X)+X is quicker than upper=NULL;
#     upper=T marginally slower than upper=F.

# To compare nearest.dist with dist, use as.dist(...)
nx <- 4
x <- expand.grid(as.double(1:nx),as.double(1:nx))
sum( ( as.dist(nearest.dist( x, delta=nx*2))-
          dist(x)                            )^2)

# Create nearest neighbor structures:
par(mfcol=c(1,2))
x <- expand.grid(1:nx,1:(2*nx))
display( nearest.dist( x, delta=1))
x <- expand.grid(1:(2*nx),1:nx)
display( nearest.dist( x, delta=1))

}
\author{Annina Cincera (C++ code), Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/cov.Rd.                             %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{covmat}
\alias{covmat}
\alias{cov.exp}
\alias{cov.sph}
\alias{cor.sph}
\alias{cov.nug}
\alias{cov.wu1}
\alias{cov.wu2}
\alias{cov.wu3}
\alias{cov.wend1}
\alias{cov.wend2}
\alias{cov.mat}
\alias{cov.mat12}
\alias{cov.mat32}
\alias{cov.mat52}
\alias{cov.finnmat}
\title{Covariance Functions}
\description{
  Evaluate a covariance function.}
\usage{
covmat(h, theta, ... , type="sph")

cov.exp(h, theta, ... , eps= getOption("spam.eps"))
cov.sph(h, theta, ... , eps= getOption("spam.eps"))
cov.nug(h, theta, ... , eps= getOption("spam.eps"))
cov.wend1(h, theta, ... , eps= getOption("spam.eps"))
cov.wend2(h, theta, ... , eps= getOption("spam.eps"))
cov.wu1(h, theta, ... , eps= getOption("spam.eps"))
cov.wu2(h, theta, ... , eps= getOption("spam.eps"))
cov.wu3(h, theta, ... , eps= getOption("spam.eps"))
cov.mat(h, theta, ... , eps= getOption("spam.eps"))
cov.finnmat(h, theta, ... , eps= getOption("spam.eps"))
cov.mat12(h, theta, ... , eps= getOption("spam.eps"))
cov.mat32(h, theta, ... , eps= getOption("spam.eps"))
cov.mat52(h, theta, ... , eps= getOption("spam.eps"))

cor.sph(h, range, ... , eps= getOption("spam.eps"))

}
\arguments{
\item{h}{object containing the lags.}
\item{theta}{parameter of the covariance function, see
  \sQuote{Details}.}
\item{range}{parameter defining the compact support.}
\item{type}{covariance function specification.}
\item{\dots}{arguments passed from other methods.}
\item{eps}{tolerance level, see
  \sQuote{Details}.}
}
\details{\code{covmat} is a wrapper that calls the other functions
  according to the argument \code{type}. The nomenclature is similar to
  \code{precmat}.
  \cr
  The parametrization is (range, [partial-sill = 1], [smoothness = 1], [nugget = 0]), where
  only the range needs to be specified.
  In case of negative parameter values, a warning is issued and the
  absolute value is retained.
  Although more cryptic, having all arguments as a single vector
  simplifies optimization with \code{optim}.
  \cr
  The parameters are and locations are up to precision \code{epsilon}.
  That means that all distances smaller than \code{eps} are considered zero; a nugget smaller than \code{eps} is ignored; a range smaller than \code{eps} represents a nugget model; etc.
  \cr
  \code{cov.finnmat()} is similar to \code{cov.mat()} but with the
  \code{sqrt(8*smoothness)/range} argument in the Bessel function
  (instead of \code{1/range}).
  \code{cov.mat12()} is a wrapper to \code{cov.exp()}
  \code{cov.mat32()}, and \code{cov.mat52()} are fast version of
  \code{cov.mat()} with smoothness 3/2 and 5/2, respectively (factor 10).
  \cr
  \code{cor.sph(,range)} is a fast version of \code{cov.sph(,c(range,1,0))}.
  \cr
  Currently, the functions distinguish between a sparse \code{spam}
  object \code{h} and any other numeric type. In the future, this might
  change and appropriate methods will be implemented.
}
\value{Covariance function evaluated on \code{h}.}
%\note{There is intentionally no \acronym{S3} distinction between the classes
%  \code{spam} and \code{spam.chol.}\emph{method}.}
\references{Any classical book about geostatistics.}
\seealso{\code{\link{precmat}}.}
\examples{
set.seed(123)
n <- 200
locs <- cbind(runif(n),runif(n))
h <- nearest.dist(locs, delta=sqrt(2), upper = NULL)
Sigma <- cov.sph(h, c(.3, 1, .1))

iidsample <- rnorm(n)
cholS <- chol.spam(as.spam(Sigma))
iorder <- iord <- ordering(cholS, inv = TRUE)
sample <- (iidsample \%*\% as.spam(cholS))[iorder]
plot(locs, col = fields::tim.colors(n = 256)[cut(sample, n)], pch = 20)

\dontrun{
h <- seq(0, to=1, length.out=100)
plot( h, cov.exp(h, c(1/3,1)), type='l', ylim=c(0,1))
type <- c("sph","wendland1","wendland2","wu1","wu2","wu3")
for (i in 1:6)
  lines( h, covmat(h, 1, type=type[i]), col=i+1)
legend('topright',legend=type, col=2:7, lty=1)

}
}

\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/math2.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{Math2}
\alias{Math2}
\alias{Math2.spam}
\alias{Math2,spam-method}
\alias{Math2,spam,numeric-method}
\alias{round,spam-method}
\alias{signif,spam-method}
\alias{round.spam}
\alias{signif.spam}
\title{Rounding of Numbers}
\description{Applies the \code{Math2} group functions to '\code{spam}' objects
}
\usage{\S4method{round}{spam}(x, digits = 0)
\S4method{signif}{spam}(x, digits = 6)
}
\arguments{\item{x}{spam object.}
\item{digits}{integer indicating the precision to be used.}
}
\value{All functions operate on the vector \code{x@entries} and return the
  result thereof.
}
%\details{%  Is implemented for R>=2.3.x only. However,
%  it would be possible to use Martin's proposed workaround:
%    \url{http://tolstoy.newcastle.edu.au/R/help/05/12/18192.html}
%\references{
%}
\seealso{\code{\link{Ops.spam}} and \code{\link{Math.spam}}}
\examples{
getGroupMembers("Math2")

set.seed(12)
smat <- diag.spam( rnorm(15))
round(smat, 3)
}
\author{Reinhard Furrer}
\keyword{manip}
% "round"  "signif"

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/s3only.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{s3only}
\alias{var.spam}
\title{Wappers for Sparse Matrices}
\description{
  These functions are convenient wrappers for  \code{spam} objects
  to classical matrix operations.}
\usage{
var.spam(x, \dots)

\S3method{var}{spam}(x, \dots)
}
\arguments{
\item{x}{matrix of class \code{spam}.}
\item{\dots}{further arguments passed to or from other methods.}
}
\value{Depends on function\dots
}
\details{There is probably no point in fully defining methods
  here. Typically, these functions do not exploit sparsity
  structures. Hence, for very large matrices, warnings may be posted.}
%\references{}
%\note{}
\seealso{Option \code{"inefficiencywarning"} in \code{\link{spam.options}}.
}
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/germanydata.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{germany.data}
\alias{germany.data}
\alias{germany.info}
\alias{germany.poly}
\alias{germany.spam}
\alias{germany}
\docType{data}
\title{Meta-data About Administrative Districts of Germany}
\description{Supplementary data used for the display of data
  from the administrative districts of Germany}
\format{\code{germany.info} is a list with elements
  \describe{
    \item{n}{544 (number of districts around 1990).}
    \item{xrep,yrep}{representative coordinates of the districts (vectors
      of length 544)}
    \item{xlim,ylim}{2-vectors defining the limits of the districts.}
    \item{polyid}{linking the polygons to the districts (599 vector).}
    \item{id}{linking the districts to Community Identification Number.}}
  \code{germany.poly} defines the polygons. It is a 17965 by two matrix,
  each polygon separated by a row of \code{NA}s, each district by two rows.\cr
  \code{germany} defines the polygons in form of a list (backwards compatibility). }


\references{The meta-data has been constructed based on (essentially)
  files from the   package \pkg{INLA}, see \code{demo(Bym)}.\cr

  See also \url{https://de.wikipedia.org/wiki/Amtlicher_Gemeindeschl\%C3\%BCssel}
and \url{https://en.wikipedia.org/wiki/Districts_of_Germany}}
\details{The representative coordinates are calculated based on the mean
  value of the polygon coordinates. This creates sometimes strange
  values, e.g., district Leer.}
\seealso{\code{\link{germany.plot}} \code{\link{Oral}}.}
\examples{
# Plot the Bundeslaender:
germany.plot(germany.info$id\%/\%1000,col=rep(2:8,3), legend=FALSE)
}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/summary.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{Summary}
\alias{Summary}
\alias{Summary.spam}
\alias{Summary,spam-method}
\alias{all.spam}
\alias{any.spam}
\alias{max.spam}
\alias{min.spam}
\alias{prod.spam}
\alias{range.spam}
\alias{sum.spam}
\alias{all,spam-method}
\alias{any,spam-method}
\alias{max,spam-method}
\alias{min,spam-method}
\alias{prod,spam-method}
\alias{range,spam-method}
\alias{sum,spam-method}
\title{Rounding of Numbers}
\description{Applies the \code{Math2} group functions to \code{spam} objects
}
\usage{# max(x,..., na.rm = FALSE)
}
\arguments{\item{x}{spam object.}
   \item{na.rm}{a logical indicating whether missing values should be
          removed.}
}
\value{If \code{structurebased=TRUE}, all functions operate on the vector \code{x@entries} and return the
  result thereof.\cr
  Conversely, if \code{structurebased=FALSE}, the result is identical to
  one with \code{as.matrix(x)} input.
}
\details{The \code{na.rm} argument is only meaninful if \code{NAOK=TRUE}.
}
%\references{
%}
\seealso{\code{\link{Math.spam}} and \code{\link{Math2}}.}
\examples{
getGroupMembers("Summary")

smat <- diag.spam( runif(15))
range(smat)
options(spam.structurebased=FALSE)
range(smat)

\dontrun{
max( log(spam(c(1,-1))), na.rm=TRUE)
}
# allow 'NA's first:
# TODO
# options(spam.NAOK=TRUE)
# max( log(spam(c(1,-1))), na.rm=TRUE)
}
\author{Reinhard Furrer}
\keyword{manip}
% all any max min prod range sum

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/det.Rd.                             %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{det}
\alias{det,spam-method}
\alias{det,spam.chol.NgPeyton-method}
\alias{det.spam}
\alias{det}
\alias{determinant}
%\alias{determinant,spam.chol.NgPeyton-method}
\alias{determinant,spam-method}
\alias{determinant.spam}
\alias{determinant.spam.chol}
\alias{determinant.spam.chol.NgPeyton}
\title{Determinant of a Symmetric Positive Definite Sparse Matrix}
\description{\code{det} and \code{determinant} calculate the determinant of a
  symmetric, positive definite sparse matrix. \code{determinant} returns
  separately the modulus of the determinant, optionally on the logarithm scale,
  and the sign of the determinant.

}
\usage{
det(x, ...)
determinant(x, logarithm = TRUE, ...)
}
\arguments{
  \item{x}{sparse matrix of class \code{spam} or a Cholesky factor  of
    class \code{spam.chol.NgPeyton}.}
  \item{logarithm}{logical; if \code{TRUE} (default) return the logarithm of the
    modulus of the determinant.}
  \item{...}{Optional arguments. Examples include \code{method} argument
    and additional parameters used by the method.}
}
\value{For \code{det}, the determinant of \code{x}.  For \code{determinant}, a
   list with components

  \item{modulus}{a numeric value.  The modulus (absolute value) of the
          determinant if \code{logarithm} is \code{FALSE}; otherwise the
          logarithm of the modulus.}

  \item{sign}{+1, as only symmetric positive definite matrices are considered.}

}
\details{If the matrix is not positive definite, the function issues a
  warning and returns \code{NA}.

  The determinant is based on the product of the diagonal entries of a
  Cholesky factor, i.e. internally, a Cholesky decomposition is
  performed. By default, the NgPeyton algorithm with minimal degree
  ordering us used. To change the methods or supply additonal parameters
  to the Cholesky factorization function, it is possible to pass via
  \code{\link{chol}}.

  The determinant of a Cholesky factor is also defined.
}
\references{
Ng, E. G. and B. W. Peyton (1993) Block sparse Cholesky algorithms
on advanced uniprocessor computers, \emph{SIAM J. Sci. Comput.}, \bold{14},
1034--1056.
}
\seealso{\code{\link{chol.spam}}
}
\examples{
x <- spam( c(4,3,0,3,5,1,0,1,4), 3)
det( x)
determinant( x)

det( chol( x))
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rowcolstats.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{rowSums}
\alias{rowSums.spam}
\alias{colSums.spam}
\alias{rowMeans.spam}
\alias{colMeans.spam}
\alias{rowSums}
\alias{colSums}
\alias{rowMeans}
\alias{colMeans}
\alias{rowSums,spam-method}
\alias{colSums,spam-method}
\alias{rowMeans,spam-method}
\alias{colMeans,spam-method}
\title{Form Row and Column Sums and Means}
\description{
Form row and column sums and means for sparse \code{\linkS4class{spam}} matrices
}
\usage{
rowSums(x, na.rm = FALSE, dims = 1, \dots)
colSums(x, na.rm = FALSE, dims = 1, \dots)
rowMeans(x, na.rm = FALSE, dims = 1, \dots)
colMeans(x, na.rm = FALSE, dims = 1, \dots)
}
\arguments{
  \item{x}{a \code{\linkS4class{spam}} object}
  \item{na.rm}{currently ignored}
  \item{dims}{ignored as we have only two dimensions.}
  \item{\dots}{potentially further arguments from other methods.}
}
\details{
Depending on the flag \code{}.
}
\value{
Vector of appropriate length.
}
%\references{}
\author{
Reinhard Furrer
}
%\note{}
\seealso{
\code{\link{apply.spam}}, \code{\link{spam.options}}.
}
\examples{
x <- spam( rnorm(20), 5, 4)
rowSums( x)
c( x \%*\% rep(1,4))
}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/methods.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   


\name{spam methods}
\alias{methods.spam}
\title{Methods for Sparse Matrices}
\description{Methods without any additional parameters for sparse matrices.
}
\details{If a method for \code{spam} objects takes the same arguments,
  produces the intuitive output. We do not provide additional help
  pages.

  However, such methods are usually linked to a \code{xzy.spam}
  function, that could also be called directly.
	    }
\seealso{Corresponding base help functions.
}
\author{Reinhard Furrer}
\keyword{internal}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/diff.Rd.                            %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{diff}
\alias{diff}
\alias{diff.spam}
\alias{diff,spam-method}
\title{Lagged Differences}
\description{Returns suitably lagged and iterated differences.}
\usage{
# diff.spam(x, lag = 1, differences = 1, ...)
\S4method{diff}{spam}(x, lag = 1, differences = 1, ...)
}
\arguments{
  \item{x}{a \code{spam} matrix containing the values to be
    differenced.}
  \item{lag}{an integer indicating which lag to use.}
  \item{differences}{an integer indicating the order of the difference.}
  \item{...}{further arguments to be passed to or from methods.}
}
\value{A \code{spam} matrix with elements similar to
  \code{as.spam(diff(as.matrix(x), ...))}.
}
%\details{    }
%\references{}
\seealso{\code{\link[base]{diff}} in \code{base}, \code{\link{precmat}}.}
\examples{
# incidence matrix for a RW(3) model
D <- diff.spam(diag.spam(10), lag=1, differences=3)
t(D)\%*\%D
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/UScounties.Rd.                      %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{UScounties}
\alias{UScounties}
\alias{UScounties.storder}
\alias{UScounties.ndorder}
\docType{data}
\title{Adjacency Structure of the Counties in the Contiguous United States}
\description{
  First and second order adjacency structure of the counties
  in the contiguous United States. We consider that two counties
  are neighbors if they share at
  least one edge of their polygon description in \code{maps}.
}

\format{
  Two matrices of class \code{spam}

\describe{
\item{UScounties.storder}{
  Contains a one in the \code{i} and \code{j} element if  county
  \code{i} is a neighbor of county \code{j}.
}
\item{UScounties.ndorder}{
  Contains a one in the \code{i} and \code{j} element if  counties
  \code{i} and \code{j} are a neighbors of county \code{k} and counties
  \code{i} and \code{j}  are not  neighbors.
}
}
}
%\source{\url{ www.to somethin??}}
\seealso{\code{map}, from \pkg{maps}.}
%\references{}
\examples{
# number of counties:
n  <- nrow( UScounties.storder)

\dontrun{
# make a precision matrix
Q <- diag.spam( n) + .2 * UScounties.storder + .1 * UScounties.ndorder
display( as.spam( chol( Q)))
}
}
\keyword{datasets}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/import.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{import}
\alias{import}
\alias{read.MM}
\alias{read.HB}
\title{Read External Matrix Formats}
\description{
    Read matrices stored in the Harwell-Boeing or MatrixMarket formats.
}
\usage{
read.HB(file)
read.MM(file)
}
\arguments{
 \item{file}{the name of the file to read, as a character scalar.  

   Alternatively, \code{read.HB} and \code{read.MM} accept connection
   objects.}
}
\value{A sparse matrix of class  \code{spam}.}
\details{The names of files storing matrices in the
  Harwell-Boeing format usually end in \code{".rua"} or \code{".rsa"}.
  Those storing matrices in the MatrixMarket format usually end in
  \code{".mtx"}.

  Currently, only  real assembled Harwell-Boeing can be read with
  \code{read.HB}. Reading MatrixMarket formats is more flexible.
  However, as entries of \code{spam} matrices are of mode \code{double},
  integers matrices are coerced to doubles, patterns lead to matrices
  containing ones and complex are coerced to the real part thereof. In
  these aforementioned cases, a warning is issued.\cr

  MatrixMarket also defines an array format, in which case a (possibly)
  dense \code{spam} object is return (retaining only elements which are
  larger than \code{options('spam.eps')}. A warning is issued. }
\note{
  The functions are based on \code{readHB} and \code{readMM} from
  the library \code{Matrix} to build the connection and read the raw
  data.
  At present, \code{read.MM} is more flexible than  \code{readMM}.
}
\references{
  \url{https://math.nist.gov/MatrixMarket/}
  
  \url{https://sparse.tamu.edu/}
}
\examples{

\dontrun{
image(read.MM(gzcon(url(
  "ftp://math.nist.gov/pub/MatrixMarket2/Harwell-Boeing/bcspwr/bcspwr01.mtx.gz"))))
}

\dontrun{
## Datasets supplied within Matrix
str(read.MM(system.file("external/pores_1.mtx",package = "Matrix")))
str(read.HB(system.file("external/utm300.rua", package = "Matrix")))
str(read.MM(system.file("external/lund_a.mtx", package = "Matrix")))
str(read.HB(system.file("external/lund_a.rsa", package = "Matrix")))
}
}
\author{Reinhard Furrer based on \code{Matrix} functions by
  Douglas Bates \email{bates@stat.wisc.edu} and Martin Maechler \email{maechler@stat.math.ethz.ch}}
\keyword{IO}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/mle.Rd.                             %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{mle}
\alias{neg2loglikelihood.spam}
\alias{neg2loglikelihood.nomean}
\alias{mle.spam}
\alias{mle.nomean.spam}
\alias{neg2loglikelihood}
\alias{mle}
\alias{mle.nomean}
\title{Maximum likelihood estimates}
\description{Maximum likelihood estimates of a simple spatial model}
\usage{
neg2loglikelihood.spam(y, X, distmat, Covariance,
                 beta, theta, Rstruct = NULL, cov.args = NULL, ...)

neg2loglikelihood(y, X, distmat, Covariance,
                 beta, theta, cov.args = NULL, ...)

neg2loglikelihood.nomean(y, distmat, Covariance,
                 theta, cov.args = NULL, ...)

mle.spam(y, X, distmat, Covariance,
     beta0, theta0, thetalower, thetaupper, optim.control=NULL,
     Rstruct = NULL, hessian = FALSE, cov.args = NULL, ...)

mle(y, X, distmat, Covariance,
     beta0, theta0, thetalower, thetaupper, optim.control=NULL,
     hessian = FALSE, cov.args = NULL, ...)

mle.nomean.spam(y, distmat, Covariance,
     theta0, thetalower, thetaupper, optim.control=NULL,
     Rstruct = NULL, hessian = FALSE, cov.args = NULL, ...) 

mle.nomean(y, distmat, Covariance,
     theta0, thetalower, thetaupper, optim.control=NULL,
     hessian = FALSE, cov.args = NULL, ...) 
}
\arguments{
  \item{y}{data vector of length n.}
  \item{X}{the design matrix of dimension n x p.}
  \item{distmat}{a distance matrix. Usually the result of a call
     to \code{nearest.dist}.}
  \item{Covariance}{function defining the covariance. See example.}
  \item{beta}{parameters of the trend (fixed effects).}
  \item{theta}{parameters of the covariance structure.}
  \item{Rstruct}{the Cholesky structure of the covariance matrix.}
  \item{beta0,theta0}{inital values.}
  \item{thetalower,thetaupper}{lower and upper bounds of the parameter
    \code{theta}.}
  \item{optim.control}{arguments passed to \code{optim}.}
  \item{hessian}{Logical. Should a numerically differentiated Hessian matrix
    be returned?}
  \item{cov.args}{additional arguments passed to \code{Covariance}.}
  \item{...}{additional arguments passed to \code{chol}.}
}
\value{The negative-2-loglikelihood or the output from the function
     \code{optim}.
}
\details{
We provide functions to calculate the
negative-2-log-likelihood and maximum likelihood estimates  for the model

y ~ N_n( X beta, Sigma(h;theta) )

in the case of a sparse or ordinary covariance matrices.

In the case of the \code{*.spam} versions, the covariance function has
to return a \code{spam} object. In the other case, the methods are
correctly overloaded and work either way, slightly slower than the
\code{*.spam} counterparts though.


When working on the sphere, the distance matrix has to be transformed by

h ->   R / 2 sin(h/2)

where R is the radius of the sphere.


The covariance function requires that the first argument is the distance
matrix and the second the parameters. One can image cases in which the
covariance function does not take the entire distance matrix but only
some partial information thereof. (An example is the use of a kronecker
type covariance structure.)  In case of a sparse covariance construction
where the argument \code{Rstruct} is not given, the first parameter
element needs to be the range parameter. (This results from the fact,
that a sparse structure is constructed that is independent of the
parameter values to exploit the fast Choleski decomposition.)


In the zero-mean case, the \code{neg2loglikelihood} is calculated by setting
the parameters \code{X} or \code{beta} to zero.
}
%\references{}
\seealso{\code{\link{covmat}}, \code{\link{rmvnorm.spam}}
}
\examples{
# True parameter values:
truebeta <- c(1,2,.2)    # beta = (intercept, linear in x, linear in y)
truetheta <- c(.5,2,.02) # theta = (range, sill, nugget)



# We now define a grid, distance matrix, and a sample:
x <- seq(0,1,l=5)
locs <- expand.grid( x, x)
X <- as.matrix( cbind(1,locs))  # design matrix

distmat <- nearest.dist( locs, upper=NULL) # distance matrix
Sigma <- cov.sph( distmat, truetheta)    # true covariance matrix


set.seed(15)
y <- c(rmvnorm.spam(1,X \%*\% truebeta,Sigma)) # construct sample

# Here is the negative 2 log likelihood:
neg2loglikelihood.spam( y, X, distmat, cov.sph,
                       truebeta, truetheta)

# We pass now to the mle:
res <- mle.spam(y, X, distmat, cov.sph,
         truebeta, truetheta,thetalower=c(0,0,0),thetaupper=c(1,Inf,Inf))

# Similar parameter estimates here, of course:
mle.nomean.spam(y-X\%*\%res$par[1:3], distmat, cov.sph,
         truetheta, thetalower=c(0,0,0), thetaupper=c(1,Inf,Inf))
}
\author{Reinhard Furrer }
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/kronecker.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{kronecker}
\alias{kronecker.default}
\alias{kronecker.spam}
\alias{kronecker,spam,ANY-method}
\alias{kronecker,ANY,spam-method}
\alias{kronecker,spam,spam-method}
\title{Kronecker Products on Sparse Matrices}
\description{Computes the generalised kronecker product of two arrays, \code{X} and
     \code{Y}.}
\usage{
kronecker.spam(X, Y, FUN = "*", make.dimnames = FALSE, ...)
}
\arguments{
  \item{X}{sparse matrix of class \code{spam}, a vector or a matrix.}
  \item{Y}{sparse matrix of class \code{spam}, a vector or a matrix.}
  \item{FUN}{a function; it may be a quoted string. See details}
  \item{make.dimnames}{Provide dimnames that are the product of the dimnames of
    \code{X} and \code{Y}. }
    \item{ ...}{optional arguments to be passed to \code{FUN}.}
}
\value{An array \code{A} with dimensions
     \code{dim(X) * dim(Y)}.}
\details{The sparsity structure is determined by the ordinary
  \code{\%x\%}. Hence, the result of \code{kronecker(X, Y, FUN = "+")} is
    different depending on the input.
    }
%\references{}
%\seealso{\code{\link{chol}}}
\examples{
# Starting with non-spam objects, we get a spam matrix
kronecker.spam( diag(2), array(1:4, c(2, 2)))

kronecker( diag.spam(2), array(1:4, c(2, 2)))

# Notice the preservation of sparsity structure:
kronecker( diag.spam(2), array(1:4, c(2, 2)), FUN="+")


}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam.chol.NgPeyton-class.Rd.        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   


\name{spam.chol.NgPeyton-class}
\docType{class}
\alias{spam.chol.NgPeyton-class}
\alias{as.matrix,spam.chol.NgPeyton-method}
\alias{as.spam,spam.chol.NgPeyton-method}
\alias{backsolve,spam.chol.NgPeyton-method}
\alias{c,spam.chol.NgPeyton-method}
\alias{determinant,spam.chol.NgPeyton-method}
\alias{diag,spam.chol.NgPeyton-method}
\alias{diag<-,spam.chol.NgPeyton-method}
\alias{dim<-,spam.chol.NgPeyton-method}
\alias{dim,spam.chol.NgPeyton-method}
\alias{display,spam.chol.NgPeyton-method}
\alias{forwardsolve,spam.chol.NgPeyton-method}
\alias{image,spam.chol.NgPeyton-method}
\alias{length<-,spam.chol.NgPeyton-method}
\alias{length,spam.chol.NgPeyton-method}
\alias{ordering,spam.chol.NgPeyton-method}
\alias{print,spam.chol.NgPeyton-method}
\alias{show,spam.chol.NgPeyton-method}
\alias{summary,spam.chol.NgPeyton-method}
\alias{t,spam.chol.NgPeyton-method}
\alias{chol,spam.chol.NgPeyton-method}

\title{Class "spam.chol.NgPeyton"}
\description{Result of a Cholesky decomposition with the \code{NgPeyton}
method}
\section{Objects from the Class}{
Objects are created by calls of the form \code{chol(x,method="NgPeyton", ...)}
and should not be created directly with a
\code{new("spam.chol.NgPeyton", ...)} call.\cr
At present, no proper print method is defined. However, the factor can
be transformed into a \code{spam} object.
}
\section{Methods}{
  \describe{
    \item{as.matrix}{\code{signature(x = "spam.chol.NgPeyton")}: Transform the factor
      into a regular matrix. }
    \item{as.spam}{\code{signature(x = "spam.chol.NgPeyton")}:  Transform the factor
      into a \code{spam} object.}
    \item{backsolve}{\code{signature(r = "spam.chol.NgPeyton")}: solving
    a triangular system, see \code{\link{solve}}. }
    \item{forwardsolve}{\code{signature(l = "spam.chol.NgPeyton")}: solving
    a triangular system, see \code{\link{solve}}. }
    \item{c}{\code{signature(x = "spam.chol.NgPeyton")}: Coerce the factor into a vector. }
    \item{determinant}{\code{signature(x = "spam.chol.NgPeyton")}:
      Calculates the determinant from the factor, see also \code{\link{det}}. }
    \item{diag}{\code{signature(x = "spam.chol.NgPeyton")}: Extracts the
    diagonal entries.}
%    \item{dim<-}{\code{signature(x = "spam.chol.NgPeyton")}: ... }
    \item{dim}{\code{signature(x = "spam.chol.NgPeyton")}: Retrieve the
      dimension. Note that \code{"dim<-"} is not implemented.}
    \item{display}{\code{signature(x = "spam.chol.NgPeyton")}: Transformation
      to a \code{spam} object and display, see also \code{\link{display}}. }
    \item{image}{\code{signature(x = "spam.chol.NgPeyton")}: Transformation
      to a \code{spam} object and display, see also \code{\link{image}}. }
%    \item{length<-}{\code{signature(x = "spam.chol.NgPeyton")}: ... }
    \item{length}{\code{signature(x = "spam.chol.NgPeyton")}: Retrieve the
      dimension. Note that \code{"length<-"} is not implemented. }
    \item{ordering}{\code{signature(x = "spam.chol.NgPeyton")}:
      Retrieves the ordering, in \code{\link{ordering}}. }
    \item{print}{\code{signature(x = "spam.chol.NgPeyton")}: Short description. }
    \item{show}{\code{signature(object = "spam.chol.NgPeyton")}: Short description. }
    \item{summary}{\code{signature(object = "spam.chol.NgPeyton")}:
      Description of the factor, returns (as a list) \code{nnzR}, \code{nnzcolindices},
      the density of the factor \code{density}, and fill-in ratio
      \code{fillin}. For the use of the first two, see \sQuote{Examples}
    in \code{\link{chol}}.}
    \item{t}{\code{signature(x = "spam.chol.NgPeyton")}: Transformation
      to a \code{spam} object and transposition. }
    \item{chol}{\code{signature(x = "spam.chol.NgPeyton")}: Returns
      \code{x} unchanged. }
  }
}
\details{
It is not possible to directly change the length, dimension and the
diagonal entries of a \code{"spam.chol.NgPeyton"} object. 
}
\references{Ng, E. G. and B. W. Peyton (1993), "Block sparse Cholesky algorithms
on advanced uniprocessor computers", \emph{SIAM J. Sci. Comput.}, \bold{14},
pp. 1034-1056.

}
\author{Reinhard Furrer}
\seealso{\code{\link{print.spam}}
  \code{\link{ordering}} and \code{\link{chol}}}
\examples{
x <- spam( c(4,3,0,3,5,1,0,1,4),3)
cf <- chol( x)
cf
as.spam( cf)


# Modify at own risk...
slotNames(cf)

}
\keyword{classes}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/makeprec.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{makeprec}
\alias{precmat.GMRFreglat}
\title{Create Precision Matrices}
\description{Creates precision matrices for gridded GMRF.}
\usage{
precmat.GMRFreglat(n,m, par, model = "m1p1",  eps = getOption("spam.eps"))
}
\arguments{
  \item{n}{first dimension of the grid.}
  \item{m}{second dimension of the grid.}
  \item{par}{parameters used to construct the matrix.}
  \item{model}{see details and examples.}
  \item{eps}{A tolerance parameter:  elements of \code{x} such that \code{abs(x)
    <= eps} set to zero. Defaults to \code{eps = getOption("spam.eps")}} 
}
\value{A \code{spam} matrix of dimension \code{prod(dims)}x\code{prod(dims)}.}
\details{The function should be illustrative on how to create precision
  matrices for gridded GMRF. Hence, no testing (positive definiteness is
  done).

  The model specification \code{"m"} determines the complexity and
  \code{"p"} the number of parameters. 
  
  Please see the examples on the meaning of the different models. 
    }
%\references{}
\seealso{\code{\link{precmat}}, \code{\link{toeplitz.spam}}, \code{\link{kronecker.spam}}}
\examples{
as.matrix(precmat.GMRFreglat(4, 3, c(.4),         'm1p1'))
as.matrix(precmat.GMRFreglat(4, 3, c(.4,.3),      'm1p2'))
as.matrix(precmat.GMRFreglat(4, 3, c(.4,.3,.2),   'm2p3'))
as.matrix(precmat.GMRFreglat(4, 3, c(.4,.3,.2,.1),'m2p4'))

# up to the diagonal, the following are equivalent:
cleanup( precmat.IGMRFreglat(3,4) -
             precmat.GMRFreglat(3,4,1, 'm1p1'))

}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/precmat.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{precmat}
\alias{precmat}
\alias{precmat.RW1}
\alias{precmat.RW2}
\alias{precmat.RWn}
\alias{precmat.season}
\alias{precmat.IGMRFreglat}
\alias{precmat.IGMRFirreglat}
\title{IGMRF Precision Matrices}
\description{
  Fast ways to create sparse precision matrices for various IGMRF.}
\usage{
precmat(n, season=12, m=n, A=NULL, order=1, ... , type="RW1")
precmat.RW1(n)
precmat.RW2(n)
precmat.RWn(n, order=3)
precmat.season(n, season=12)
precmat.IGMRFreglat(n, m, order=1, anisotropy=1)
precmat.IGMRFirreglat(A, eps=getOption("spam.eps"))
}
\arguments{
\item{n}{dimension of the field.}
\item{type}{the type of the IGMRF.}
\item{season}{length of season.}
\item{m}{second dimension (in case of a regular lattice).}
\item{A}{adjacency matrix (see below).}
\item{order}{order for higher order RWs.}
\item{anisotropy}{anisotropy factor, between 0 and 2.}
\item{eps}{tolerance level.}
\item{\dots}{arguments passed to individual functions.}
}
\details{\code{precmat} is a wrapper that calls the other functions
  according to the argument \code{type}.
  \cr
  Implements many of the precision matrices discussed in Chapter
  3 of Rue and Held (2005). For example, \code{precmat.RW1},
  \code{precmat.RW2} and \code{precmat.season} are given in
  equations (3.22), (3.40) and (3.59); \code{precmat.IGMRFreglat} on
  page 107. Note that for the latter we reverse the order of the dimension here!
  \cr
  If adjacency matrix is a regular matrix, it is coerced to a
  \code{spam} object. Only the structure is used. Make sure, that the
  diagonal is empty.
}
\value{A sparse precision matrix.}
%\note{There is intentionally no \acronym{S3} distinction between the classes
%  \code{spam} and \code{spam.chol.}\emph{method}.}
\references{Rue and Held (2005). 
}
\seealso{\code{\link{precmat.GMRFreglat}}, \code{\link{rmvnorm.prec}}, \code{\link{adjacency.landkreis}}.
}
\examples{
n <- 10
Q <- precmat.RW2( n)
# rmvnorm.prec(1, Q=Q) # does not work, because the matrix is singular.

Q\%*\%cbind(1,1:n)
}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/powerboost.Rd.                      %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{powerboost}
\alias{powerboost}
\title{Specific Options Setting}
\description{
  Sets several options for speed-up.
}
\usage{
powerboost(flag)
}
\arguments{
  \item{flag}{on or off}
}
\details{The options turn checking off (\code{"safemode"},
  \code{"cholsymmetrycheck"}  and \code{"cholpivotcheck"}) and switch to
  single precision for \code{"eps"}.}
\value{
  \code{NULL} in any case.

}
\seealso{\code{\link{spam.options}}.
}
\author{Reinhard Furrer, after receiving too much C.mc.st adds.}
\keyword{environment}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/chol.Rd.                            %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{chol}
\alias{chol}
\alias{chol.spam}
\alias{chol,ANY-method}
\alias{chol,matrix-method}
\alias{chol,spam-method}
\alias{update.spam}
\alias{update,spam.chol.NgPeyton-method}
\alias{update.spam.chol.NgPeyton}
\title{Cholesky Factorization for Sparse Matrices}
\description{
\code{chol} performs a Cholesky
decomposition of a symmetric positive definite sparse matrix \code{x}
of class \code{spam}.}
\usage{
# chol(x, \dots)

\S4method{chol}{spam}(x, pivot = "MMD", method = "NgPeyton",
       memory = list(), eps = getOption("spam.eps"), Rstruct=NULL,
       \dots, verbose=FALSE)
%force64 =  getOption("spam.force64"),\dots)

# update.spam.chol.NgPeyton(object, x,...)
\S4method{update}{spam.chol.NgPeyton}(object, x,...)
%chol(x, method, ordering, memory, \dots)
}
\arguments{
\item{x}{symmetric positive definite matrix of class \code{spam}.}
\item{pivot}{should the matrix be permuted, and if, with what
  algorithm, see \sQuote{Details} below.}
\item{method}{Currently, only \code{NgPeyton} is implemented.}
\item{memory}{Parameters specific to the method, see \sQuote{Details} below.}
\item{eps}{threshold to test symmetry. Defaults to \code{getOption("spam.eps")}.}
%\item{force64}{logical vector of length 1. If \code{TRUE}, a 64-bit
%    spam matrix is returned in any case. If \code{FALSE}, a 32-bit
%    matrix is returned when possible. }
\item{Rstruct}{sparsity structure  of the factor, see \sQuote{Details} below.}
\item{\dots}{further arguments passed to or from other methods.}
\item{object}{an object from a previous call to \code{chol}, i.e.,
  sparsity structure of the factor.}
\item{verbose}{provides more details about the decomposition. Useful
  when working with  huge matrices.}
}
\value{The function returns the Cholesky factor in an object of class
  \code{spam.chol.}\emph{method}. Recall that the latter is the Cholesky
  factor of a reordered matrix \code{x}, see also \code{\link{ordering}}.
}

\details{\code{chol} performs a Cholesky decomposition of a symmetric
positive definite sparse matrix \code{x} of class
\code{spam}. Currently, there is only the block sparse Cholesky
algorithm of Ng and Peyton (1993) implemented (\code{method="NgPeyton"}).

To pivot/permute the matrix, you can choose between the multiple minimum
degree (\code{pivot="MMD"}) or reverse Cuthill-Mckee (\code{pivot="RCM"})
from George and Lui (1981). It is also possible to furnish a specific
permutation in which case \code{pivot} is a vector. For compatibility
reasons, \code{pivot} can also take a logical in which for \code{FALSE}
no permutation is done and for \code{TRUE} is equivalent to
\code{MMD}.\cr

Often the sparsity structure is fixed and does not change, but the
entries do. In those cases, we can update the Cholesky factor with
\code{update.spam.chol.NgPeyton} by suppling a Cholesky factor and the
updated matrix. For \code{U <- chol(A)},  \code{update(U,
  Anew)} and \code{chol(Anew, Rstruct=U)} are equivalent.

The option \code{cholupdatesingular} determines how singular matrices
are handled by \code{update}. The function hands back an error
(\code{"error"}), a warning (\code{"warning"}) or the value \code{NULL}
(\code{"null"}).\cr



The Cholesky decompositions requires parameters, linked to memory
allocation.  If the default values are too small the Fortran routine
returns an error to \R, which allocates more space and calls the Fortran
routine again. The user can also pass better estimates of the allocation
sizes to \code{chol} with the argument \code{memory=list(nnzR=...,
  nnzcolindices=...)}. The minimal sizes for a fixed sparsity
structure can be obtained from a \code{summary} call, see \sQuote{Examples}.\cr


The output of \code{chol} can be used with \code{forwardsolve} and
\code{backsolve} to solve a system of linear equations.\cr


Notice that the Cholesky factorization of the package \code{SparseM} is also
based on the algorithm of Ng and Peyton (1993).  Whereas the Cholesky
routine of the package \code{Matrix} are based on
\code{CHOLMOD} by Timothy A. Davis (\code{C} code).
}
\references{
Ng, E. G. and Peyton, B. W.  (1993) Block sparse Cholesky algorithms on
advanced uniprocessor computers, \emph{SIAM J. Sci. Comput.}, \bold{14},
1034--1056.


Gilbert, J. R., Ng, E. G. and Peyton, B. W. (1994) An efficient
   algorithm to compute row and column counts for sparse Cholesky
   factorization, \emph{SIAM J. Matrix Anal. Appl.}, \bold{15},
   1075--1091.


George, A. and  Liu, J. (1981)
\emph{Computer Solution of Large Sparse Positive Definite Systems},
Prentice Hall.
}
\note{
Although the symmetric structure of \code{x} is needed, only the upper
diagonal entries are used. By default, the code does check for
symmetry (contrarily to \code{base:::chol}). However,
depending on the matrix size, this is a time consuming test.
A test is ignored if
\code{options("spam.cholsymmetrycheck")} is set to \code{FALSE}.


If a permutation is supplied with \code{pivot},
\code{options("spam.cholpivotcheck")} determines if the permutation is
tested for validity (defaults to \code{TRUE}).
}
\seealso{\code{\link{det.spam}}, \code{\link{solve.spam}},
  \code{\link{forwardsolve.spam}}, \code{\link{backsolve.spam}} and \code{\link{ordering}}.
}
\examples{
# generate multivariate normals:
set.seed(13)
n <- 25    # dimension
N <- 1000  # sample size
Sigma <- .25^abs(outer(1:n,1:n,"-"))
Sigma <- as.spam( Sigma, eps=1e-4)

cholS <- chol( Sigma)
# cholS is the upper triangular part of the permutated matrix Sigma
iord <- ordering(cholS, inv=TRUE)

R <- as.spam(cholS)
mvsample <- ( array(rnorm(N*n),c(N,n)) \%*\% R)[,iord]
# It is often better to order the sample than the matrix
# R itself.

# 'mvsample' is of class 'spam'. We need to transform it to a
# regular matrix, as there is no method 'var' for 'spam' (should there?).
norm( var( as.matrix( mvsample)) - Sigma, type='m')
norm( t(R) \%*\% R - Sigma)


# To speed up factorizations, memory allocations can be optimized:
opt <- summary(cholS)
# here, some elements of Sigma may be changed...
cholS <- chol( Sigma, memory=list(nnzR=opt$nnzR,nnzcolindices=opt$nnzc))

}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer, based on Ng and Peyton (1993) Fortran routines}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/isSymmetric.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   


\name{isSymmetric}
\alias{isSymmetric.spam}
\alias{isSymmetric,spam-method}
\title{Test if a Sparse Matrix is Symmetric}
\description{Efficient function to test if 'object' is symmetric or not.}
\usage{
# isSymmetric.spam(object, ...)
\S3method{isSymmetric}{spam}(object, tol = 100 * .Machine$double.eps, ...)}
\arguments{
  \item{object}{a \code{spam} matrix.}
\item{tol}{numeric scalar >= 0.  Smaller differences are not considered,
          see \code{all.equal.spam}.}
\item{...}{further arguments passed to \code{all.equal.spam}.}
 }
 \details{symmetry is assessed by comparing the sparsity structure of
   \code{object} and \code{t(object)} via the function
   \code{all.equal.spam}. If a difference is detected, the matrix is
   cleaned with \code{cleanup} and compared again.}
 \value{
   logical indicating if \code{object} is symmetric or not.}
 \seealso{\code{\link{all.equal.spam}}, \code{\link{cleanup}}.}
 \examples{
obj <- diag.spam(2)
isSymmetric(obj)

obj[1,2] <- .Machine$double.eps
isSymmetric(obj)
all.equal(obj, t(obj))
}
 \author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/USprecip.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{USprecip}
\alias{USprecip}
\docType{data}
\title{Monthly Total Precipitation (mm) for April 1948 in the Contiguous United States}
\description{
This is a useful spatial data set of moderate to large size consisting of 11918
locations. See \url{https://www.image.ucar.edu/GSP/Data/US.monthly.met/} for the source of these data.
}

\format{
This data set is an array containing the following columns:

\describe{
\item{lon,lat}{
  Longitude-latitude position of monitoring stations.
}
\item{raw}{
  Monthly total precipitation in millimeters for April 1948.
}
\item{anomaly}{
  Preipitation anomaly for April 1948.
}
\item{infill}{
Indicator, which station values were observed (5906 out of the 11918)
compared to which were estimated.
}
}
}
\source{\url{https://www.image.ucar.edu/GSP/Data/US.monthly.met/}}
\seealso{\code{\link[fields]{RMprecip}}}
\references{
Johns, C., Nychka, D., Kittel, T., and Daly, C. (2003)
Infilling sparse records of spatial fields.
\emph{Journal of the American Statistical Association},
98, 796--806.
}
\examples{
# plot
\dontrun{
library(fields)

data(USprecip)
par(mfcol=c(2,1))
quilt.plot(USprecip[,1:2],USprecip[,3])
US( add=TRUE, col=2, lty=2)
quilt.plot(USprecip[,1:2],USprecip[,4])
US( add=TRUE, col=2, lty=2)
}
}
\keyword{datasets}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/bdiag.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{bdiag}
\alias{bdiag}
\alias{bdiag.spam}
\title{Binds Arrays Corner-to-Corner}
\description{Creates a sparse block-diagonal matrix.
}
\usage{
bdiag.spam(...)
}
\arguments{
  \item{...}{Arrays to be binded together}
 }

\details{
This is a small helper function to create block diagonal sparse matrices. In the two matrix case,
\code{bdiag.spam(A,B)}, this is equivalent to a complicated \code{rbind(cbind(A, null), cbind(B, t(null)))},
where \code{null} is a null matrix of appropriate dimension.\cr

It is recursively defined.

The arrays are coerced to sparse matrices first.\cr


This function is similar to the function \code{bdiag} from the package
\code{Matrix}. It is also similar to the function \code{adiag} from the package
\code{magic}.  However, here no padding is done and all the dimnames are
stripped.


	    }
\value{
    Returns a \code{spam} matrix as described above.
}
%\references{}
\seealso{
	\code{\link{diag.spam}}.
}
\examples{
A <- diag.spam(2, 4)           # 2*I4
B <- matrix(1,3,3)
AB <- bdiag.spam(A,B)

# equivalent to:
ABalt <- rbind(cbind( A, matrix(0,nrow(A),ncol(B))),
               cbind( matrix(0,nrow(B),ncol(A)), B))
         
norm(AB-ABalt)


# Matrices do not need to be square:
bdiag.spam(1,2:5,6)
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/crossprod.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{crossprod}
\alias{crossprod.default}
\alias{crossprod.spam}
\alias{tcrossprod.spam}
\alias{crossprod,spam,missing-method}
\alias{tcrossprod,spam,missing-method}
\alias{crossprod,ANY,spam-method}
\alias{tcrossprod,ANY,spam-method}
\alias{crossprod,spam,spam-method}
\alias{tcrossprod,spam,spam-method}
\alias{crossprod,spam,ANY-method}
\alias{tcrossprod,spam,ANY-method}
\title{Spam Matrix Crossproduct}
\description{
     Given matrices \code{x} and \code{y} as arguments, return a matrix
     cross-product.  This is formally equivalent to (but usually
     slightly faster than) the call \code{t(x) \%*\% y} (\code{crossprod.spam}) or \code{x
     \%*\% t(y)} (\code{tcrossprod.spam}).}

\usage{

     crossprod.spam(x, y = NULL, \dots)
     
     tcrossprod.spam(x, y = NULL, \dots)
}
\arguments{
  \item{x, y}{matrices: \code{y = NULL} is taken to be the
          same matrix as \code{x}.  Vectors are promoted to single-column or
          single-row matrices, depending on the context.}
  \item{\dots}{potentially further arguments from other methods.}
}
\value{A double matrix}
\note{
     When \code{x} or \code{y} are not matrices, they are treated as column or
     row matrices.
   }
%\references{}
%\seealso{\code{\link{chol}}}
\examples{
crossprod.spam(diag.spam(2),1:2)

}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/germany.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   





\name{germany.plot}
\alias{germany.plot}
\title{Plot Administrative Districts of Germany}
\description{Displaying data
  over the administrative districts of Germany}
\usage{
germany.plot(vect,  col=NULL, zlim=range(vect), legend=TRUE, 
             main=NULL, cex.axis=1, cex.main=1.5, add=FALSE, ... )
}
\arguments{
\item{vect}{vector of length 544}
\item{col}{color scheme to be used. By default uses \code{colorRampPalette(brewer.pal(9,"Blues"))(100)}.}
\item{zlim}{the minimum and maximum values for which colors should be
  plotted, defaulting to the range of \code{data}.}
\item{legend}{Should the legend be added, see \sQuote{Details}.}
\item{main}{an overall title for the plot.}
\item{cex.axis}{label size of legend.}
\item{cex.main}{label size of overall plot title.} 
\item{add}{logical, if true adds to current plot.}
\item{\dots}{additional arguments passed to \code{polygon}.}
}
\references{See also \url{https://de.wikipedia.org/wiki/Amtlicher_Gemeindeschl\%C3\%BCssel}
and \url{https://de.wikipedia.org/wiki/Liste_der_Landkreise_in_Deutschland}
%The code of \code{map.landkreis} is very similar to
%\code{germany.map} from the package \pkg{INLA}.}
}
\details{The legend is only added, provided (a) function
  \code{image.plot} is available.\cr

  The perfect position of the legend is an art per se and depends on
  various \code{par} parameters. One possiblity for finer control is to
  not plot it and to manually call the function
\code{image.plot} of \pkg{fields}.}
\seealso{\code{\link{Oral}}.}
\examples{
data( Oral)
germany.plot( Oral$Y/Oral$E)


# Plot the Bundeslaender:
germany.plot(germany.info$id\%/\%1000,col=rep(2:8,3), legend=FALSE)
}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/constructors.Rd.                    %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{constructors}
\alias{rowpointers}
\alias{rowpointers<-}
\alias{colindices}
\alias{colindices<-}
\alias{entries}
\alias{entries<-}
\alias{dimension<-}
\alias{constructors}
\title{Slot Modification}
\description{Modify slots of \code{spam} objects
}
\usage{
rowpointers( x) <- value
colindices( x) <- value
entries( x) <- value}
\arguments{
  \item{x}{a \code{spam} matrix}
  \item{value}{vector of appropriate length.}
  }
\value{Modified \code{spam} object.}
\details{Various tests are performed. Thus much slower than direct
  assignment.\cr
  Slot \code{dimension} should be changed through \code{pad} or \code{dim}
}
  \examples{
x <- diag.spam( 2)  
rowpointers( x) <- c(1,1,3)

# The last line is equivalent to 
x@rowpointers <- as.integer( c(1,1,3))
}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rmvnorm.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{rgrf}
\alias{rgrf}
\alias{rgrf.spam}
\title{Draw From a Gaussian Random Field}
\description{
  Fast and intuitive ways to draw from a Gaussian random field.}
\usage{
rgrf( n,
   locs, nx, ny=nx, xlim=c(0,1), ylim=c(0,1), tau=0,
   Covariance, theta, beta=0, X, 
   method=c('chol'),  method.args=list(sparse=FALSE), 
   eps = getOption("spam.eps"), drop=TRUE, attributes=TRUE, ...)
}
\arguments{
\item{n}{number of observations.}
\item{locs}{locations, the result of \code{as.matrix(locs)} will be
  used.}
\item{nx,ny}{if no locations are specified, at least one of these to
  specify the grid dimension.}
\item{xlim,ylim}{Domain, see \sQuote{Details}.}
\item{tau}{perturbation degree, see \sQuote{Details}.}
\item{Covariance}{covariance function name.}
\item{theta}{covariance parameter.}
\item{beta}{mean or vector for regression-type mean.}
\item{X}{design matrix for regression-type mean.}
 \item{method}{based on Choleski factorization.}
 \item{method.args}{list of arguments that can be passed to the corresponding approach.
  For \code{"chol"} it can be, e.g., \code{RStruct}, \code{chol.args}, \code{cov.args}.}
\item{eps}{small value, anything smaller is considered a collocation.}
\item{drop}{logical, if a single realization should be returned as a vector.}
\item{attributes}{logical, if should attributes be passed back.}

\item{\dots}{currently not used.}
}
\details{If no locations are given, the function constructs these
  according a regular or a regular perturbed grid. The perturbation is
  determined by \code{tau}, which has to be greater than zero (no
  perturbation) and strictly smaller than 1/2 (max perturbation).

  The regular grid has spacing (here for x) \code{dx=diff(xlim)/nx} and runs
  from \code{xlim[1]+dx/2} to \code{xlim[2]-dx/2}.
  The locations are at least  \code{(1/nx-2*tau*dx)} separated.
  \cr
  
  Currently, the only method implemented is a Cholesky factorization
  routine,  (much as in \code{\link{rmvnorm}}).

  The \code{rdist()} from the \code{fields} package is awefully
  fast. Unless one has very sparse covariance matrices, a sparse
  approach is not bringing a lot of improvements. 

  
  The methods may use different covariance construction approaches and
  thus the nesting of \code{cov.args} in \code{method.args}.
}
\seealso{\code{\link{rgrf}}, \code{\link{chol}} and \code{\link{ordering}}.
}
\examples{
require(fields)
# Regular grid with constant mean:
nx <- 10
field <- rgrf(1, nx=nx,  Covariance="cov.wend2", theta=c(.5, 1), beta=5)
quilt.plot(cbind(attr(field,"locs"),z=field), nx=nx, ny=nx)
points(attr(field,"locs"))

# Irregluar grid:
field <- rgrf(1, nx=10, tau=0.3, Covariance="cov.mat", theta=c(.2, 1, 1.5))
fields::quilt.plot(attr(field,"locs"), field)

}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/adjacency.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{adjacency}
\alias{adjacency}
\alias{adjacency.spam}
\alias{adjacency.landkreis}
\alias{germany.graph}
\title{Administrative Districts of Germany}
\description{Constructing the adjacency graph
  of the administrative districts of Germany}
\usage{
adjacency.landkreis(loc)
}
\arguments{
\item{loc}{location of the graph structure, can be an URL.}
}
\details{The function is included as an example on how to
construct adjacency matrices form a (common) adjacency structure.
For the particular example, note that the nodes are not numbered
consecutively and that they start from zero.}
\value{a sparse matrix in \code{spam} format.}
\references{The adjacency data has been provided by Havard Rue and is
  also available in \pkg{INLA}.}
\seealso{\code{\link{germany.plot}} super-seeding \code{map.landkreis}
  for plotting.\cr
  \code{\link{Oral}}.}
\examples{
\dontrun{
loc <- system.file("demodata/germany.adjacency", package="spam")
display( adjacency.landkreis( loc))
}
}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/display.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{display}
\alias{display}
\alias{display.spam}
\alias{display,spam-method}
%\alias{display,spam.chol.NgPeyton-method}
\title{Graphially Represent the Nonzero Entries}
\description{The function represents the nonzero entries in a simple
  bicolor plot.
}
\usage{
display(x, ...)
}
\arguments{\item{x}{matrix of class \code{spam} or \code{spam.chol.NgPeyton}.}
  \item{...}{any other arguments passed
    to \code{image.default}/\code{plot}.}
}
%\value{}
\details{\code{spam.getOption("imagesize")} determines  if the sparse matrix is
  coerced into a matrix and the plotted with \code{image.default} or if
  the matrix is simply represented as a scatterplot with
  \code{pch="."}.  The points are scaled according to
  \code{cex*getOption("spam.cex")/(nrow + ncol)}.
  For some devices or for non-square matrices,
  \code{cex} needs probably some adjustment.
}
%\references{}
\seealso{\code{\link{image}}, \code{\link{spam.options}}}
\examples{
set.seed(13)

smat <- spam_random(8)
par(mfcol=c(1,2), pty='s')
options(spam.imagesize = 1000)
display(smat)
options(spam.imagesize = 10)
display(smat, cex=.25)


# very large but very sparse matrix
smat <- spam_random(2^14, distribution=rnorm, density=1e-5, verbose=TRUE)
par(mfcol=c(1, 1), mai=c(.4,.4,.1,.1), pty='s')
display(smat)
}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/math.Rd.                            %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




\name{Math}
\alias{Math}
\alias{Math.spam}
\alias{Math,spam-method}
\alias{ceiling.spam}
\alias{ceiling,spam-method}
\alias{floor.spam}
\alias{floor,spam-method}
\alias{trunc.spam}
\alias{trunc,spam-method}
\alias{exp.spam}
\alias{exp,spam-method}
\alias{log.spam}
\alias{log,spam-method}
\alias{log2.spam}
\alias{log2,spam-method}
\alias{log10.spam}
\alias{log10,spam-method}
\alias{sqrt.spam}
\alias{sqrt,spam-method}
\alias{gamma.spam}
\alias{gamma,spam-method}
\alias{digamma.spam}
\alias{digamma,spam-method}
\alias{trigamma.spam}
\alias{trigamma,spam-method}
\alias{lgamma.spam}
\alias{lgamma,spam-method}
\alias{abs.spam}
\alias{abs,spam-method}
\alias{cumprod.spam}
\alias{cumprod,spam-method}
\alias{cumsum.spam}
\alias{cumsum,spam-method}
\alias{cummax.spam}
\alias{cummax,spam-method}
\alias{cummin.spam}
\alias{cummin,spam-method}
\alias{cos.spam}
\alias{cos,spam-method}
%\alias{cospi.spam}
%\alias{cospi,spam-method}
\alias{cosh.spam}
\alias{cosh,spam-method}
\alias{acos.spam}
\alias{acos,spam-method}
\alias{acosh.spam}
\alias{acosh,spam-method}
\alias{sin.spam}
\alias{sin,spam-method}
\alias{asin.spam}
\alias{asin,spam-method}
\alias{asinh.spam}
\alias{asinh,spam-method}
\alias{tan.spam}
\alias{tan,spam-method}
\alias{atan.spam}
\alias{atan,spam-method}
\alias{atanh.spam}
\alias{atanh,spam-method}
\title{Mathematical Functions}
\description{Applies the \code{Math} group functions to \code{spam} objects
}
\usage{# ceiling(x)
# floor(x)
# exp(x, base = exp(1))
# log(x, base = exp(1))
# sqrt(x)

# abs(x)
# cumprod(x)
# cumsum(x)

# cos(x)
# sin(x)
# tan(x)

# acosh(x)

}
\arguments{\item{x}{spam object.}
\item{base}{positive number. The base with respect to which logarithms
are computed. Defaults to \code{e=exp(1)}.}
}
\value{If the option \code{spam.structurebased=TRUE}, all functions operate on the vector \code{x@entries} and return the
result thereof.\cr
Conversely, if \code{structurebased=FALSE}, the result is identical to
one with \code{as.matrix(x)} input and an \code{as.spam} purger.
}
\details{
It is important to note that the zero entries do not enter the
evaluation when \code{structurebased=FALSE}. The operations are performed on the stored non-zero
elements. This may lead to differences if compared with the same
operation on a full matrix.
}
%\references{
%}
\seealso{\code{\link{Summary.spam}}, \code{\link{Ops.spam}} and \code{\link{Math2.spam}}
}
\examples{
getGroupMembers("Math")

mat <- matrix(c( 1,2,0,3,0,0,0,4,5),3)
smat <- as.spam( mat)
cos( mat)
cos( smat)

options(spam.structurebased=FALSE)
cos( smat)

sqrt( smat)
}
\author{Reinhard Furrer}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/random.Rd.                          %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{random}
\alias{spam_random}
\title{Create Random Sparse Matrices}
\description{Creates random spam matrix given the dimension and other parameters.}
\usage{
spam_random(nrow = 1L, ncol = nrow, density = 0.5, distribution = NULL, digits = NULL,
            sym = FALSE, spd = FALSE, verbose = FALSE, ...)
}
\arguments{
  \item{nrow}{integer value for the number of rows for the \code{spam} matrix to create.}
  \item{ncol}{integer value for the number of columns. The default value is the same as \code{nrow}.}
  \item{density}{A numeric value between 0 and 1 specifying the approximate density of matrix.
    If equal to zero the \code{spam} matrix contains only zeros and if equal to 1 the \code{spam} matrix is full.}
  \item{distribution}{a random number generating distribution function to sample the entries of the \code{spam} matrix.
    The function must have an argument with the name \code{n}, possible
    candidates are \code{rnorm}, \code{rexp}, \code{rpois},
    \code{rweibull}, etc. Default (\code{NULL}) fills with ones.}
  \item{...}{possible additional arguments for the distribution function if specified with \code{distribution}.}
  \item{digits}{an integer value for the number of digits the entries should be rounded.}
  \item{sym}{logical value to specify symmetry of the \code{spam} matrix.}
  \item{spd}{logical value to specify positive definitness of the \code{spam} matrix, via diagonal dominace criteria.
    Note, if \code{spd} TRUE, then \code{sym} is overwritten to \code{TRUE} in any case.}
  \item{verbose}{logical value to specify verbose statments of the function.}
}
\value{A random matrix in \code{spam} format.}
\details{To create a random spam64 matrix, set \code{options(spam.force64 = TRUE)}.}
%\references{}
\seealso{\code{\link{spam-class}} and \code{\link{display.spam}}}
\examples{
set.seed(42)
rspam <- spam_random(500, digits = 2, distribution = rnorm, sd = 2, mean = 10, density = .01)
display.spam(rspam, cex = 2)
}
\author{Florian Gerber, Roman Flury, Reinhard Furrer}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rmvnorm.const.Rd.                   %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{rmvnorm.const}
\alias{rmvnorm.const}
\alias{rmvnorm.prec.const}
\alias{rmvnorm.canonical.const}
\title{Draw Constrainted Multivariate Normals}
\description{
  Fast ways to draw multivariate normals with linear constrains when the variance or precision matrix
  is sparse.}
\usage{
rmvnorm.const(n, mu = rep.int(0, dim(Sigma)[1]), Sigma, Rstruct = NULL,
              A = array(1, c(1,dim(Sigma)[1])), a=0, U=NULL,  ...)
rmvnorm.prec.const(n, mu = rep.int(0, dim(Q)[1]), Q, Rstruct = NULL,
              A = array(1, c(1,dim(Q)[1])), a=0, U=NULL,  ...)
rmvnorm.canonical.const(n, b, Q, Rstruct = NULL,
              A = array(1, c(1,dim(Q)[1])), a=0, U=NULL, ...)
}
\arguments{
\item{n}{number of observations.}
\item{mu}{mean vector.}
\item{Sigma}{covariance matrix of class \code{spam}.}
\item{Q}{precision matrix.}
\item{b}{vector determining the mean.}
\item{Rstruct}{the Cholesky structure of \code{Sigma} or \code{Q}.}
\item{A}{Constrain matrix.}
\item{a}{Constrain vector.}
\item{U}{see below.}
\item{\dots}{arguments passed to \code{chol}.}
}
\details{The functions \code{rmvnorm.prec} and \code{rmvnorm.canonical}
  do not requrie sparse precision matrices.
  For \code{rmvnorm.spam}, the differences between regular and sparse
  covariance matrices are too significant to be implemented here.  \cr
  Often (e.g., in a Gibbs sampler setting), the sparsity structure of
  the covariance/precision does not change. In such setting, the
  Cholesky factor can be passed via \code{Rstruct} in which only updates
  are performed (i.e., \code{update.spam.chol.NgPeyton} instead of a
  full \code{chol}).
}
%\note{There is intentionally no \acronym{S3} distinction between the classes
%  \code{spam} and \code{spam.chol.}\emph{method}.}
\references{See references in \code{\link{chol}}.
}
\seealso{\code{\link{rmvnorm.spam}}.
}
\examples{
# to be filled in

}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/dim.Rd.                             %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   


\name{dim}
\alias{dim.spam}
\alias{dim<-.spam} 
\alias{dim<-,spam-method}
\title{Dimensions of an Object}
\description{
  Retrieve or set the dimension of an \code{spam} object.
}
\usage{
# dim(x)
# dim(x) <- value
%"dim<-.spam"(x,value)
}
\arguments{
  \item{x}{a \code{spam} matrix}
  \item{value}{A numeric two-vector, which is coerced to integer (by truncation).}
 }
\value{
  \code{dim} retrieves
  the \code{dimension} slot of the object.  It is a vector
  of mode \code{integer}.

  The replacemnt method changes the dimension of the object by rearranging.
  
}
\details{
  In older version of \code{spam}, the behavior of the replacement
  method was different and is now implemented in \code{\link{pad.spam}}.

}
%\references{}
\seealso{\code{\link{pad.spam}}.
}
\examples{
x <- diag(4)
dim(x)<-c(2,8)
x

s <- diag.spam(4)
dim(s) <- c(2,8)  # result is different than x

s <- diag.spam(4)
pad(s) <- c(7,3)  # any positive value can be used

}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/Oral.Rd.                            %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





\name{Oral}
\alias{Oral}
\alias{Oral.spam}
\alias{oral.spam}
\docType{data}
\title{Oral Cavity Cancer}
\description{Oral cavity cancer counts in 544 districts in Germany over 1986-1990.}
\format{\code{Oral} is a dataframe with 3 columns.
\describe{
  \item{Y}{observed counts}
  \item{E}{expected counts}
  \item{SMR}{standardized mortality ratios}}
\code{germany} is a
list of 544 elements, each describing an individual polygon of the district.
}
\details{The expected counts depend on the number of people in the
  region and their age distribution.\cr
The regions are ordered according the supplied polygon description and
adjacency graph.\cr

There is a similar dataset \code{data(Germany)}  with larynx cancer cases from the package
  \pkg{INLA} as well, with an additional smoking covariate.
}
\source{The data is available from the package \pkg{INLA}
  distributed from \url{https://www.r-inla.org}.
}
\references{
Knorr-Held, L. and Rasser, G. (2000) Bayesian Detection of Clusters and Discontinuities in Disease Maps,
     \emph{Biometrics},
     56,
     13--21.
}
\seealso{\code{\link{germany.plot}}.}
\keyword{datasets}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/image.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{image}
\alias{image}
\alias{image.spam}
\alias{image.spam.chol}
\alias{image,spam-method}
%\alias{image,spam.chol.NgPeyton-method}
\title{Display a Sparse Matrix as Color Image}
\description{The function creates a grid of colored rectangles with
  colors corresponding to the values of the \code{spam} matrix. }
\usage{
\S4method{image}{spam}(x, cex = NULL, ...)
}
\arguments{\item{x}{matrix of class \code{spam} or \code{spam.chol.NgPeyton}.}
  \item{cex}{for very large matrices, the dot size may need to be scaled.}
  \item{...}{any other arguments passed
    to \code{image.default} and \code{plot}.}
}
%\value{}
\details{\code{getOption("spam.imagesize")} determines  if the sparse matrix is
  coerced into a matrix and the plotted similarly to \code{image.default} or if
  the matrix is simply represented as a scatterplot with
  \code{pch="."}. The points are scaled according to
  \code{cex*getOption("spam.cex")/(nrow+ncol)}.
  For some devices or for non-square matrices,
  \code{cex} needs probably some adjustment.\cr
  The a zero matrix in \code{spam} format has as (1,1) entry the value
  zero and only missing entries are interpreted as \code{NA} in the
  scatter plot.
}
%\references{}
\seealso{\code{\link{display}} and  \code{\link{spam.options}}.\cr
  The code is based on  \code{\link[graphics]{image}} of \code{graphics}.
}
\examples{
set.seed(13)

smat <- spam_random(8)

par(mfcol=c(1,2),pty='s')
options(spam.imagesize=1000)
image(smat) # or use better color schemes
options(spam.imagesize=10)
image(smat, cex=.25)

smat <- spam_random(2^14, distribution=rnorm, density=1e-5, verbose=TRUE)
par(mfcol=c(1,1), mai=c(.4,.4,.1,.1), pty='s')
image(smat)

}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rmvnorm.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{rmvt}
\alias{rmvt}
\alias{rmvt.spam}
\title{Draw From a Multivariate t-Distribution}
\description{
  Fast ways to draw from a multivariate t-distribution the scale (covariance) matrix
  is sparse.}
\usage{
rmvt(n, Sigma, df = 1, delta = rep(0, nrow(Sigma)),
    type = c("shifted", "Kshirsagar"), ..., sigma)
rmvt.spam(n, Sigma, df = 1, delta = rep(0, nrow(Sigma)),
    type = c("shifted", "Kshirsagar"), ..., sigma)
}
\arguments{
\item{n}{number of observations.}
\item{Sigma}{scale matrix (of class \code{spam}).}
\item{df}{degrees of freedom.}
\item{delta}{vector of noncentrality parameters.}
\item{type}{type of the noncentral multivariate t distribution.}
\item{\dots}{arguments passed to \code{rmvnorm.spam}.}
\item{sigma}{similar to \code{Sigma}. Here for portability with \code{mvtnorm::rmvt()}}
}
\details{This function is very much like \code{rmvt()} from the package
  \pkg{mvtnorm}. We refer to the help of the afore mentioned.
}
\references{See references in \code{mvtnorm::rmvt()}.
}
\seealso{\code{\link{rmvnorm}}.
}
%\examples{}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/solve.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{spam-solve}
\alias{solve}
\alias{backsolve}
\alias{forwardsolve}
\alias{backsolve-methods}
\alias{backsolve,ANY-method}
\alias{backsolve,spam-method}
\alias{backsolve,matrix-method}
\alias{backsolve.spam}
\alias{forwardsolve-methods}
\alias{forwardsolve,ANY-method}
\alias{forwardsolve,spam-method}
\alias{forwardsolve,matrix-method}
\alias{forwardsolve.spam}
\alias{chol2inv}
\alias{chol2inv.spam}
\alias{chol2inv,spam-method}
\alias{chol2inv,spam.chol.NgPeyton-method}
\alias{solve.spam}
\alias{solve,ANY-method}
\alias{solve,spam-method}
\title{Linear Equation Solving for Sparse Matrices}
\description{
  \code{backsolve} and \code{forwardsolve} solve a system
  of linear equations where the coefficient matrix
   is upper or lower triangular.
\cr
\code{solve} solves a linear system or computes the inverse
of a matrix if the right-hand-side is missing.
}
\usage{
\S4method{solve}{spam}(a, b, Rstruct=NULL, \dots)
\S4method{backsolve}{spam}(r, x, \dots)
\S4method{forwardsolve}{spam}(l, x, \dots)
\S4method{chol2inv}{spam}(x, \dots)
}
\arguments{
\item{a}{symmetric positive definite matrix of class \code{spam} or a Cholesky factor
		  as the result of a \code{chol} call.}
\item{l,r}{object of class \code{spam} or \code{spam.chol.}\emph{method} returned by the function
\code{chol}.}
\item{x,b}{vector or regular matrix of right-hand-side(s) of a system of linear equations.}
\item{Rstruct}{the Cholesky structure of \code{a}.}
\item{\dots}{further arguments passed to or from other methods, see
  \sQuote{Details} below.}
}
\details{
  We can solve \code{A \%*\% x = b} by first computing the Cholesky decomposition \code{A =
  t(R)\%*\%R)}, then solving \code{t(R)\%*\%y = b} for \code{y}, and
  finally solving \code{R\%*\%x = y} for \code{x}.
   \code{solve} combines \code{chol}, a Cholesky decomposition of a
symmetric positive definite sparse matrix, with \code{forwardsolve} and
then \code{backsolve}.\cr

In case \code{a} is from a \code{chol} call, then \code{solve} is an
   efficient way to calculate \code{backsolve(a, forwardsolve( t(a), b))}.


   However, for \code{a.spam} and \code{a.mat} from a \code{chol} call
   with a sparse and ordinary matrix,  note that
   \code{forwardsolve( a.mat, b, transpose=T, upper.tri=T)}
   is equivalent to
   \code{forwardsolve( t(a.mat), b)}
   and \code{backsolve(a.spam, forwardsolve(a.spam, b, transpose=T, upper.tri=T))}
   yields the desired result. But
          \code{backsolve(a.spam,forwardsolve(t(a.spam), resid))} is
   wrong because  \code{t(a.spam)} is a \code{spam} and not a
   \code{spam.chol.NgPeyton} object.


\code{forwardsolve} and \code{backsolve} solve a system of linear
equations where the coefficient matrix is lower (\code{forwardsolve}) or
upper (\code{backsolve}) triangular.  Usually, the triangular matrix is
result from a \code{chol} call and it is not required to transpose it
for \code{forwardsolve}.  Note that arguments of the default
methods \code{k}, \code{upper.tri} and \code{transpose} do not have any
effects here.

Notice that it is more efficient to solve successively the linear
equations (both triangular solves) than to implement these in the
Fortran code.

If the right-hand-side in \code{solve} is missing it will compute
the inverse of a matrix. For details about the specific Cholsesky
decomposition, see \code{\link{chol}}.

Recall that the Cholesky factors are from ordered matrices.


\code{chol2inv(x)} is a faster way to \code{solve(x)}.
}
\note{There is intentionally no \acronym{S3} distinction between the classes
  \code{spam} and \code{spam.chol.}\emph{method}.}
\references{See references in \code{\link{chol}}.
}
\seealso{\code{\link{chol.spam}} and \code{\link{ordering}}.
}
\examples{
# Generate multivariate form a covariance inverse:
# (usefull for GRMF)
set.seed(13)
n <- 25    # dimension
N <- 1000  # sample size
Sigmainv <- .25^abs(outer(1:n,1:n,"-"))
Sigmainv <- as.spam( Sigmainv, eps=1e-4)


Sigma <- solve( Sigmainv)  # for verification
iidsample <- array(rnorm(N*n),c(n,N))

mvsample <- backsolve( chol(Sigmainv), iidsample)
norm( var(t(mvsample)) - Sigma)

# compare with:
mvsample <- backsolve( chol(as.matrix( Sigmainv)), iidsample, n)
   #### ,n as patch
norm( var(t(mvsample)) - Sigma)



# 'solve' step by step:
b <- rnorm( n)
R <- chol(Sigmainv)
norm( backsolve( R, forwardsolve( R, b))-
      solve( Sigmainv, b) )
norm( backsolve( R, forwardsolve( R, diag(n)))- Sigma )


# 'update':
R1 <- update( R, Sigmainv + diag.spam( n))



}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer, based on Ng and Peyton (1993) Fortran routines}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/triplet.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{triplet}
\alias{triplet}
\title{Transform a "spam" Format to Triplets}
\description{Returns a list containing the indices and elements of a  \code{spam} object.}
\usage{
triplet(x, tri=FALSE)
}
\arguments{
  \item{x}{sparse matrix of class \code{spam} or a matrix.}
  \item{tri}{Boolean indicating whether to create individual row and
    column  indices vectors.}
}
\value{A list with elements
  \item{indices}{a by two matrix containing the indices if
    \code{tri=FALSE}.}
  \item{i,j}{vectors  containing the row and column indices if \code{tri=TRUE}.}
  \item{values}{a vector containing the matrix elements.}
}
\details{
  The elements are row (column) first if \code{x} is a \code{spam} object (matrix).\cr
}
%\references{}
\seealso{ \code{\link{spam.creation}} for the inverse operation and
  \code{foreign} for other transformations.}
\examples{
x <- diag.spam(1:4)
x[2,3] <- 5
triplet(x)
all.equal( spam( triplet(x, tri=TRUE)), x)

}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/foreign.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{foreign}
\alias{foreign}
\alias{as.spam.matrix.csr}
%\alias{as.matrix.csr.spam}
\alias{as.dgRMatrix.spam}
\alias{as.dgCMatrix.spam}
\alias{as.spam.dgRMatrix}
\alias{as.spam.dgCMatrix}
\title{Transformation to Other Sparse Formats}
\description{Transform between the \code{spam} sparse format to the
  \code{matrix.csr} format of \code{SparseM} and  \code{dgRMatrix} format of \code{Matrix}}
\usage{
as.spam.matrix.csr(x)
as.dgRMatrix.spam(x)
as.dgCMatrix.spam(x)
as.spam.dgRMatrix(x)
as.spam.dgCMatrix(x)
}
\arguments{
  \item{x}{sparse matrix of class \code{spam}, \code{matrix.csr},
    \code{dgRMatrix} or \code{dgCMatrix}.}
}
\value{According to the call, a sparse matrix of class  \code{spam},
  \code{matrix.csr}, \code{dgRMatrix} or \code{dgCMatrix}.}
\details{
  We do not provide any \code{S4} methods and because of the existing
  mechanism a standard \code{S3} does not work.\cr

  The functions are based on \code{require}.\cr

  Notice that \code{as.matrix.csr.spam} should read as \code{as."matrix.csr".spam}.
}
%\references{}
\seealso{\code{\link{triplet}}, \code{\link[Matrix]{Matrix}} or \code{matrix.csr} from package \pkg{SparseM}.}
\examples{

\dontrun{
S <- diag.spam(4)
R <- as.dgRMatrix.spam( S)
C <- as.dgCMatrix.spam( S)
as.spam.dgCMatrix(C)
slotNames(C)
slotNames(R)
# For column oriented sparse formats a transpose does not the job,
# as the slot names change.


# as.spam(R) does not work.
}

\dontrun{
# for transformations between SparseM and spam:

as.matrix.csr.spam <- function(x,...) {
  if (new("matrix.csr")) {
    newx <- new("matrix.csr")
    slot(newx,"ra",check=FALSE) <- x@entries
    slot(newx,"ja",check=FALSE) <- x@colindices
    slot(newx,"ia",check=FALSE) <- x@rowpointers
    slot(newx,"dimension",check=FALSE) <- x@dimension
    return(newx)
    }
  stop("function requires 'SparseM' package")
}
# then with `SparseM`:  as.matrix.csr.spam( spamobject )
}

\dontrun{
# a dataset contained in Matrix
data(KNex, package='Matrix')
summary( KN <- as.spam.dgCMatrix(KNex$mm) )

}
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{manip}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam.creation.Rd.                   %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{spam}
\alias{spam.creation}
\alias{initialize,spam-method}
\alias{spam}
\alias{spam.list}
\alias{spam.numeric}
\alias{spam,list-method}
\alias{spam,numeric-method}
%\alias{spam.spam}
%\alias{spam,spam-method}
\alias{as.spam,matrix-method}
\alias{as.spam,numeric-method}
\alias{as.spam,spam-method}
\alias{as.spam,dist-method}
\alias{as.spam,list-method}
\alias{as.spam}
\alias{as.spam.spam}
\alias{as.spam.numeric}
\alias{as.spam.matrix}
\alias{as.spam.chol.NgPeyton}
\alias{as.spam.dist}
\alias{as.spam.list}
\alias{is.spam}
\title{Sparse Matrix Class}
\description{
This group of functions evaluates and coerces changes in class structure.
}
\usage{
spam(x, nrow = 1, ncol = 1, eps = getOption("spam.eps"))
%     force64 = getOption("spam.force64"))

as.spam(x, eps = getOption("spam.eps"))
%, force64 = getOption("spam.force64"))

is.spam(x)
}
\value{
  A valid \code{spam} object.\cr
  \code{is.spam} returns \code{TRUE} if \code{x} is a \code{spam} object.}
\arguments{
  \item{x}{is a matrix (of either dense or sparse form), a list, vector
    object or a distance object}
\item{nrow}{number of rows of matrix }
\item{ncol}{number of columns of matrix }
%\item{force64}{logical vector of length 1. If \code{TRUE}, a 64-bit
%  spam matrix is returned in any case. If \code{FALSE}, a 32-bit
%  matrix is returned when possible. }
\item{eps}{A tolerance parameter:  elements of \code{x} such that
  \code{abs(x) < eps} set to zero. Defaults to \code{eps =
    getOption("spam.eps")} }
}
\details{
  The functions \code{spam} and \code{as.spam} act like \code{matrix}
  and \code{as.matrix}
  to coerce an object to a sparse matrix object of class \code{spam}.


  If \code{x} is a list, it should contain either two or three elements.
  In case of the former, the list should contain a \code{n} by two
  matrix of indicies (called \code{ind}) and the values.
  In case of the latter, the list should contain three vectors
  containing the row, column indices (called \code{i} and
  \code{j}) and the values. In both cases partial matching is done.
  In case there are several triplets with the same \code{i}, \code{j},
  the values are added.

  \code{eps} should be at least as large as \code{.Machine$double.eps}.


  If \code{getOption("spam.force64")} is \code{TRUE}, a 64-bit
  spam matrix is returned in any case. If \code{FALSE}, a 32-bit
  matrix is returned when possible.
}
\note{The zero matrix has the element zero stored in (1,1).\cr

  The functions do not test the presence of \code{NA/NaN/Inf}. Virtually
  all call a Fortran routine with the \code{NAOK=NAOK}
  argument, which defaults to \code{FALSE} resulting in an error.
  Hence, the \code{NaN} do not always properly propagate through (i.e.
  \code{spam} is not IEEE-754 compliant).

}
\references{Reinhard Furrer, Stephan R. Sain (2010).
  "spam: A Sparse Matrix R Package with Emphasis on MCMC
  Methods for Gaussian Markov Random Fields.",
        \emph{Journal of Statistical Software}, 36(10), 1-25,
        \doi{10.18637/jss.v036.i10}.
}
\seealso{
  \code{\link{SPAM}} for a general overview of the package;
  \code{\link{spam_random}} to create matrices with a random sparsity pattern;
  \code{\link{cleanup}} to purge a sparse matrix;
  \code{\link{spam.options}} for details about the \code{safemode} flag;
    \code{\link{read.MM}} and \code{\link{foreign}} to create \code{spam}
	matrices from MatrixMarket
    files and from certain \pkg{Matrix} or \pkg{SparseM} formats.
}
\examples{
# old message, do not loop, when you create a large sparse matrix
set.seed(13)
nz <- 128
ln <- nz^2
smat <- spam(0,ln,ln)
is <- sample(ln,nz)
js <- sample(ln,nz)
## IGNORE_RDIFF_BEGIN
system.time(for (i in 1:nz) smat[is[i], js[i]] <- i)
system.time(smat[cbind(is,js)] <- 1:nz)
## IGNORE_RDIFF_END

getClass("spam")


options(spam.NAOK=TRUE)
as.spam(c(1, NA))
}
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/pad.Rd.                             %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   




\name{pad}
\alias{pad} 
\alias{pad.spam}
\alias{pad<-} 
\alias{pad<-.spam} 
\alias{pad<-,spam-method}
\alias{pad<-,matrix-method}
\title{Padding a Sparse Matrix}
\description{
  Resets the dimension of a \code{spam} matrix by truncation or zero padding.
}
\usage{
pad(x) <- value
%"pad<-.spam"(x,value)
}
\arguments{
  \item{x}{a \code{spam} matrix}
  \item{value}{A numeric two-vector.}
}
\value{
  A (\code{spam}) matrix of dimension \code{value} where trunction or
  padding has
  been used.
 
}
\details{
  It is important to notice the different behavior of the replacement
  method for ordinary arrays and \code{spam} objects (see
  \sQuote{Examples}). Here, the elements are not simply rearranged but
  an entirely new matrix is constructed. If the new column dimension is
  smaller than the original, the matrix is also cleaned (with
  \code{option("spam.eps")} as filter).
}
%\references{}
\seealso{\code{\link{dim.spam}}.
}
\examples{
x <- diag(4)
dim(x)<-c(2,8)
x

s <- diag.spam(4)
pad(s) <- c(7,3)  # any positive value can be used

s <- diag.spam(4)
pad(s) <- c(2,8)  # result is different than x
}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/gmult.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{gmult}
\alias{gmult}
\title{Generalized Multiplication}
\usage{
gmult(x, splits, fact)
}
\description{
Multiplying specific submatrices of a spam matrix with different factors.
}
\arguments{
  \item{x}{a spam matrix.}
  \item{splits}{vector of how to split the matrix into submatrices.
                It starts with \code{1} and ends with \code{max(dim(X))+1}.}
  \item{fact}{matrix of factors to multiply submatrices defined by splits.
              Dimensions of \code{fact} must correspond to the \code{length(splits)-1}.}
}
\value{Spam matrix, where each specified submatrix is multiplied with a factor.}

\examples{
x <- spam(1, 15, 15)
print(x, minimal=FALSE)

splits <- c(1, 2, 8, ncol(x)+1) # divide matrix into 9 submatrices
fact <- array(1:9, c(3, 3))     # multiply each submatrix with a different factor
oF <- gmult(x, splits, fact)
print(oF, minimal=FALSE)
}

\author{Florian Gerber, Roman Flury}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/operations.Rd.                      %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





\name{spam-operations}
\alias{Ops.spam}
\alias{spam.ops}
\alias{\%*\%-methods}
\alias{\%*\%,ANY,ANY-method}
\alias{\%*\%,spam,spam-method}
\alias{\%*\%,spam,matrix-method}
\alias{\%*\%,spam,numeric-method}
\alias{\%*\%,matrix,spam-method}
\alias{\%*\%,numeric,spam-method}
\alias{\%d*\%}
\alias{\%d*\%,spam,spam-method}
\alias{\%d*\%,spam,ANY-method}
\alias{\%d*\%,matrix,spam-method}
\alias{\%d*\%,matrix,ANY-method}
\alias{\%d*\%,spam,numeric-method}
\alias{\%d*\%,numeric,spam-method}
\alias{\%d*\%,numeric,matrix-method}
\alias{\%d*\%,numeric,numeric-method}
\alias{\%d+\%}
\alias{\%d+\%,spam,spam-method}
\alias{\%d+\%,spam,ANY-method}
\alias{\%d+\%,matrix,spam-method}
\alias{\%d+\%,matrix,ANY-method}
\alias{\%d+\%,spam,numeric-method}
\alias{\%d+\%,numeric,matrix-method}
\alias{\%d+\%,numeric,spam-method}
\alias{\%d+\%,numeric,numeric-method}
\alias{+,spam,spam-method}
\alias{+,matrix,spam-method}
\alias{+,spam,matrix-method}
\alias{+,ANY,spam-method}
\alias{+,spam,ANY-method}
\alias{-,spam,spam-method}
\alias{-,matrix,spam-method}
\alias{-,spam,matrix-method}
\alias{-,ANY,spam-method}
\alias{-,spam,ANY-method}
\alias{*,spam,spam-method}
\alias{*,ANY,spam-method}
\alias{*,spam,ANY-method}
\alias{/,spam,spam-method}
\alias{/,ANY,spam-method}
\alias{/,spam,ANY-method}
\alias{/,matrix,spam-method}
\alias{^,spam,spam-method}
\alias{^,ANY,spam-method}
\alias{^,spam,ANY-method}
\alias{&,spam,ANY-method}
\alias{&,spam,spam-method}
\alias{&,ANY,spam-method}
\alias{|,spam,ANY-method}
\alias{|,ANY,spam-method}
\alias{|,spam,spam-method}
\alias{^,spam-method}
\alias{\%\%,spam-method}
\alias{\%/\%,spam-method}
\alias{>,spam-method}
\alias{>=,spam-method}
\alias{<,spam-method}
\alias{<=,spam-method}
\alias{==,spam-method}
\alias{!=,spam-method}
\alias{norm}
\alias{norm.spam}
\alias{norm,ANY-method}
\alias{norm,spam,character-method}
\alias{norm,spam,missing-method}
\alias{norm,numeric,missing-method}
\alias{norm,numeric,character-method}
\alias{norm,matrix,missing-method}
\alias{norm,matrix,character-method}
%\alias{t,ANY-method}
\alias{t.spam}
\alias{ncol,spam-method}
\alias{nrow,spam-method}
\alias{dim,ANY-method}
\alias{dim,spam-method}
\alias{[.spam}
\alias{[<-.spam}
\alias{[<-,spam,missing,missing-method}
\alias{[<-,spam,missing,vector-method}
\alias{[<-,spam,vector,missing-method}
\alias{[<-,spam,vector,vector-method}
\alias{[<-,spam,matrix,missing-method}
\alias{[<-,spam,matrix,matrix-method}
\alias{[<-,spam,spam,missing-method}
\alias{[<-,spam,ANY,ANY-method}
\alias{plot.spam}
\alias{subset.spam}
\alias{assign.spam}
\title{Basic Linear Algebra for Sparse Matrices}
\description{Basic linear algebra operations for sparse matrices
of class \code{spam}.
}
%\usage{x \%*\% y
%y \%d*\% x
%y \%d+\% x
%x[i,]
%...
%}
%\arguments{
%\item{x}{matrix of class \code{spam}.}
%\item{y}{matrix of class \code{spam} or a dense matrix or vector.}
%\item{value}{replacement values.}
%\item{i,j}{vectors of elements to extract or replace.}
%\item{nrow}{optional number of rows for the result.}
%}
\details{Linear algebra operations for matrices of class
\code{spam} are designed to behave exactly as for
regular matrices.  In particular, matrix multiplication, transpose, addition,
subtraction and various logical operations should work as with the conventional
dense form of matrix storage, as does indexing, rbind, cbind, and diagonal
assignment and extraction (see for example \code{\link{diag}}).
Further functions with identical behavior are \code{dim} and thus \code{nrow}, \code{ncol}.




The function \code{norm} calculates the (matrix-)norm of the argument.
The argument \code{type} specifies the \code{l1} norm, \code{sup} or max
norm (default), or the Frobenius or Hilbert-Schmidt
(\code{frobenius/hs}) norm. Partial matching can be used. For example,
\code{norm} is used to check for symmetry in the function \code{chol} by
computing the norm of the difference between the matrix and its
transpose

The operator \code{\%d*\%} efficiently multiplies a diagonal matrix (in
vector form) and a sparse matrix and is used for compatibility with the
package fields. More specifically, this method is used in the internal
functions of \code{Krig} to make the code more readable. It avoids
having a branch in the source code to handle the diagonal or nondiagonal
cases.  Note that this operator is not symmetric: a vector in
the left argument is interpreted as a diagonal matrix and a vector in
the right argument is kept as a column vector.

The operator \code{\%d+\%} efficiently adds a diagonal matrix (in vector
form) and a sparse matrix, similarly to the operator \code{\%d+\%}.

}
\references{Some Fortran functions are based on
  \url{https://github.com/johannesgerer/jburkardt-f/blob/master/sparsekit/sparsekit.html}
}
\seealso{
\code{\link{spam}} for coercion and other class relations involving the
sparse matrix classes.
}
\examples{
# create a weight matrix and scale it:
\dontrun{
wij <- distmat
# with distmat from a nearest.dist(..., upper=TRUE) call

n <- dim(wij)[1]

wij@entries <- kernel( wij@entries, h) # for some function kernel
wij <- wij + t(wij) + diag.spam(n)     # adjust from diag=FALSE, upper=TRUE

sumwij <- wij \%*\% rep(1,n)
    # row scaling:
    #   wij@entries <- wij@entries/sumwij[ wij@colindices]
    # col scaling:
wij@entries <- wij@entries/sumwij[ rep(1:n, diff(wij@rowpointers))]
}
}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/headtail.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{head}
\alias{head.spam}
\alias{head,spam-method}
\alias{head,spam.chol.NgPeyton-method}
\alias{tail.spam}
\alias{tail,spam-method}
\alias{tail,spam.chol.NgPeyton-method}
\title{
  Return the First or Last Part of an Object
}
\description{
  Returns the upper left or lower right part of a
  \code{\linkS4class{spam}} object.
}
\usage{
\S4method{head}{spam}(x, n = 6L, m = n, \dots)
\S4method{tail}{spam}(x, n = 6L, m = n, addrownums = TRUE, \dots)
}
\arguments{
  \item{x}{a \code{\linkS4class{spam}} object}
  \item{n}{a single integer. If positive, size for the resulting
    object: number of elements for a vector (including lists), rows for
    a matrix or data frame or lines for a function. If negative, all but
    the \code{n} last/first number of elements of \code{x}.}
  \item{m}{similar to \code{n} but for the number of columns.}
  \item{addrownums}{create row and column namves them from the selected
    elements.}
  \item{\dots}{arguments to be passed to or from other methods.}
}
\details{
  For matrices, 2-dim tables and data frames, \code{head()} (\code{tail()}) returns
  the first (last) \code{n} rows and \code{m} columns when \code{n > 0} or all but the
  last (first) \code{n} rows when \code{n < 0} (with similar behavior
  for \code{m}).

  \code{tail()} will add row and column names of
  the form \code{"[n,]"} and \code{"[,n]"} to the result, so that it looks similar to the
  last lines and columns of \code{x} when printed.  Setting \code{addrownums =
    FALSE} suppresses this behaviour.


  A method for \code{\linkS4class{spam.chol.NgPeyton}} objects is
  exported as well.
}
\value{
  An regular matrix.
}
\author{
  Reinhard Furrer
}
\examples{
head( precmat.RW2( 10))
tail( precmat.season(n=10, season=3), n=4, m=10)
}
\keyword{ manip }

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/print.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{print}
\docType{methods}
\alias{print}
\alias{print.spam}
\alias{print,spam-method}
%\alias{print,spam.chol.NgPeyton-method}
\alias{print.spam.chol}
\alias{print.spam.chol.NgPeyton}
\alias{summary.spam}
\alias{summary,spam-method}
%\alias{summary,spam.chol.NgPeyton-method}
\alias{summary.spam.chol}
\alias{summary.spam.chol.NgPeyton}
\alias{print_nnzpos}
\title{Printing and Summarizing Sparse Matrices}
\description{Printing (non-zero elements) of sparse matrices
  and summarizing the sparsity structure thereof.
}
\usage{
  \S4method{print}{spam}(x, ...)
\S4method{summary}{spam}(object, ...)
}
\arguments{\item{x}{matrix of class \code{spam} or \code{spam.chol.}\emph{method}.}
  \item{object}{matrix of class \code{spam} or \code{spam.chol.}\emph{method}.}
  \item{...}{any other arguments passed to \code{print.default}.
             If the non-standard argument \code{minimal} is set to \code{FALSE},
             an extended spam print is available with logical argument \code{rowpointer}
             to print rowpointers, and \code{zerosymbol} defining the character
             to display the zero element.}
}
\value{\code{NULL} for \code{print}, because the information is printed with \code{cat} there
  is no real need to pass any object back. \cr
  %
  A list containing the non-zero elements and the density for
  \code{summary}
  for class \code{spam}.\cr
  %
  A list containing the non-zero elements of the factor, the density and
  the fill-in for  \code{summary} for class \code{spam.chol.NgPeyton}.}
\details{\code{getOption('spam.printsize')} determines  if the sparse matrix is
  coerced into a matrix and the printed as an array or if
  only the non-zero elements of the matrix are given.
}
%\references{}
\seealso{\code{\link{display}} or \code{\link{image}} for a graphical visualization;
   \code{\link{spam.options}}}
\examples{
set.seed(13)
smat <- spam_random(8)
par(mfcol=c(1,2),pty='s')
options(spam.printsize=1000)
print(smat)
options(spam.printsize=10)
print(smat)
summary(smat)
summary(smat)$nnz

smat@entries[1:5] <- 0
print(smat, minimal = FALSE)
print(smat, minimal = FALSE, rowpointer = TRUE)
smat@rowpointers
print_nnzpos(smat)

}
\author{Reinhard Furrer}
\keyword{hplot}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/ordering.Rd.                        %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

\name{ordering}
\docType{methods}
\alias{ordering}
\alias{ordering.spam}
\alias{ordering-methods}
%\alias{ordering,spam.chol.NgPeyton-method}
\alias{ordering.spam.chol}
\alias{ordering.spam.chol.NgPeyton}
\alias{ordering,spam-method}
\alias{ordering,matrix-method}
\title{Extract the permutation}
\description{Extract the (inverse) permutation used by the Cholesky
  decomposition}
\usage{
ordering( x, inv=FALSE)
}
\arguments{
  
\item{x}{object of class \code{spam.chol.}\emph{method} returned by the function
\code{chol}.}
\item{inv}{Return the permutation (default) or inverse thereof.}
}
\details{
  Recall that calculating a Cholesky factor from a sparse matrix
  consists of finding a permutation first, then calculating the factors
  of the permuted matrix. The ordering is important when working with
  the factors themselves.\cr

  The ordering from a full/regular matrix is \code{1:n}.\cr
  
  Note that there exists many different algorithms to find
  orderings.  
  \cr
  
  See the examples, they speak more than 10 lines.
}
\seealso{\code{\link{chol.spam}}, \code{\link{solve.spam}}.
}
\examples{
# Construct a pd matrix S to work with (size n)
n <- 100    # dimension
S <- .25^abs(outer(1:n,1:n,"-"))
S <- as.spam( S, eps=1e-4)
I <- diag(n)  # Identity matrix

cholS <- chol( S)
ord <- ordering(cholS)
iord <- ordering(cholS, inv=TRUE)

R <- as.spam( cholS ) # R'R = P S P', with P=I[ord,],
  # a permutation matrix (rows permuted).
RtR <- t(R) \%*\% R

# the following are equivalent:
as.spam( RtR -            S[ord,ord],    eps=1e-15)
as.spam( RtR[iord,iord] - S,             eps=1e-15)
as.spam( t(R[,iord]) \%*\% R[,iord] - S, eps=1e-15)
# we use 'eps' to avoid issues close to machine precision

# trivially:
as.spam( t(I[iord,]) - I[ord,])  # (P^-1)' = P  
as.spam( t(I[ord,]) - I[,ord])  # 
as.spam( I[iord,] - I[,ord])
as.spam( I[ord,]\%*\%S\%*\%I[,ord] - S[ord,ord] )
   # pre and post multiplication with P and P' is ordering
}
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rmvnorm.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{rmvnorm.conditional}
\alias{rmvnorm.cond}
\alias{rmvnorm.conditional}
\alias{rmvnorm.conditional.spam}
\title{Draw Conditional Multivariate Normals}
\description{
  Fast way to draw conditional multivariate normals when the covariance
  matrix is sparse.}
\usage{
rmvnorm.conditional(n, y, mu = rep.int(0, dim(SigmaXX)[1]+dim(SigmaYY)[1]),
                    SigmaXX, SigmaYY, SigmaXY, noise, RstructYY = NULL, ...)
}
\arguments{
\item{n}{number of observations.}
\item{y}{observed vector.}
\item{mu}{mean vector.}
\item{SigmaXX}{covariance of X, required (of class \code{spam}).}
\item{SigmaXY}{cross-covariance of X-Y, optional (of class \code{spam}).}
\item{SigmaYY}{covariance of Y, required (of class \code{spam}).}
\item{noise}{observational noice of Y, optional. See \sQuote{Details}. }
\item{RstructYY}{the Cholesky structure of \code{SigmaYY}.}
\item{\dots}{arguments passed to \code{chol}.}
}
\details{Quite often, we want to draw condional observations \eqn{X|y}
  from the model \eqn{Y=X+e}, where \eqn{X} has covariance matrix
  \code{SigmaXX} and \eqn{e} has white noise.\cr

  Covariance of \eqn{Y} can be specified by \code{SigmaYY} or
  \code{SigmaXX+diag(noise,)}. If \eqn{Y} and \eqn{X} do not have the
  same dimensions, \code{SigmaXY} needs to be specified. \cr



  The function also implmements a general multivariate model, where the
  we only observe part of the vector.  The components are first
  \eqn{X} then \eqn{Y}.


  The function \code{rmvnorm.cond()} is a wrapper to
  \code{rmvnorm.conditional()} and included to increase similarities
  with other packages.
}
\seealso{\code{\link{rmvnorm.spam}}.
}
\examples{
set.seed(12)
N <- 300
y <- c(5, -5, -5, 5)
SigmaXX <- as.spam(.95^abs(outer(1:N, 1:N, "-")), eps=1e-4)
sel <- c(10, 100, 120, 300)        # where we observe y
SigmaXY <- SigmaXX[, sel]
SigmaYY <- SigmaXX[sel,sel] + diag.spam(.01, length(y)) # some noise
x <- rmvnorm.conditional(3, y, SigmaXX=SigmaXX, SigmaXY=SigmaXY,
                         SigmaYY=SigmaYY)
# unconditional sample:
ux <- rmvnorm(1, Sigma=SigmaXX)
matplot(t(rbind(x, ux)), type='l', lty=1)
points(sel, y, pch=19)
}
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/diag.Rd.                            %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{diag}
\alias{diag}
\alias{diag.of.spam}
\alias{diag.spam}
\alias{spam_diag}
\alias{diag,spam-method}
\alias{diag,ANY-method} 
\alias{diag<-} 
\alias{diag<-,ANY-method} 
\alias{diag<-,spam-method}
\alias{diag<-.spam}
\alias{diag.spam<-}
\alias{diag.assign,spam-method} 
\title{Sparse Matrix diagonals}
\description{Extract or replace the diagonal of a matrix, or construct a
     diagonal matrix.
}
\usage{
## diag(x)
## diag(x=1, nrow, ncol, names = TRUE)
diag(x) <- value

diag.spam(x=1, nrow, ncol)
spam_diag(x=1, nrow, ncol)
diag.spam(x) <- value
}
\arguments{
  \item{x}{a \code{spam} matrix, a vector or a scalar.}
\item{nrow, ncol}{Optional dimensions for the result.}
\item{value}{either a single value or a vector of length equal to that of
   the current diagonal.}
 }
\value{
   If \code{x} is a spam matrix then \code{diag(x)} returns the diagonal of \code{x}.

   The assignment form sets the diagonal of the sparse matrix \code{x} to the
   given value(s).\cr

   \code{diag.spam} works as \code{diag} for spam matrices:
   If \code{x} is a vector (or 1D array) of length two or more, then
   \code{diag.spam(x)} returns a diagonal matrix whose diagonal is
   \code{x}.
   \code{spam_diag} is an alias for  \code{diag.spam} and in the spirit
   of the result of \code{diag} is a \code{spam} object.

   If \code{x} is a vector of length one then \code{diag.spam(x)} returns an
   identity matrix of order the nearest integer to \code{x}.  The
   dimension of the returned matrix can be specified by \code{nrow} and
   \code{ncol} (the default is square).

   The assignment form sets the diagonal of the matrix \code{x} to the
   given value(s).

  
}
\details{Using \code{diag(x)} can have unexpected effects if \code{x} is a vector
     that could be of length one. Use \code{diag(x, nrow = length(x))} for
     consistent behaviour.

}
%\references{}
\seealso{\code{\link{upper.tri}}, \code{\link{lower.tri}}.
}
\examples{
diag.spam(2, 4)           # 2*I4
smat <- diag.spam(1:5)
diag( smat)
diag( smat) <- 5:1

# The last line is equivalent to 
diag.spam( smat) <- 5:1

# Note that diag.spam( 1:5) <- 5:1 not work of course.


}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/landkreis.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{map.landkreis}
\alias{map.landkreis}
\title{Administrative districts of Germany}
\description{Displaying data
  over the administrative districts of Germany}
\usage{
map.landkreis(data, col=NULL, zlim=range(data), add=FALSE,
              legendpos=c( 0.88,0.9,0.05,0.4))
}
\arguments{
\item{data}{vector of length 544}
\item{col}{color scheme to be used. By default uses \code{tim.colors} if
available or a generic gray scale.}
\item{zlim}{the minimum and maximum values for which colors should be
          plotted, defaulting to the range of \code{data}.}
\item{add}{logical, if true adds to current plot.}
\item{legendpos}{if package \pkg{fields} is loaded, puts a legend at that position.}
}
\references{The code of \code{map.landkreis} is very similar to
\code{germany.map} from the package \pkg{INLA}.}
\details{The function \code{\link{germany.plot}} super-seeds
  \code{map.landkreis} (it is several factors faster). \cr

The perfect position of the legend is an art per se and depends on
various \code{par} parameters. See also the source code of the function
\code{image.plot} of \pkg{fields}.}
\seealso{\code{\link{germany.plot}} super-seeding \code{map.landkreis}.}
\examples{
\dontrun{
data( Oral)
par( mfcol=c(1,2))
germany.plot( log( Oral$Y), legend=TRUE)
map.landkreis( log( Oral$Y))
}
}
\author{Reinhard Furrer}
\keyword{hplot}

% dev.off()
% dev.off();system.time( for (i in 1:20) map.landkreis(1:544))
% dev.off();system.time( for (i in 1:20) germany.plot(1:544))

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/rmvnorm.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{rmvnorm}
\alias{rmvnorm}
\alias{rmvnorm.spam}
\alias{rmvnorm.prec}
\alias{rmvnorm.canonical}
\title{Draw Multivariate Normals}
\description{
  Fast ways to draw multivariate normals when the variance or precision matrix
  is sparse.}
\usage{
rmvnorm(n, mu=rep.int(0, dim(Sigma)[1]), Sigma, ..., mean, sigma)
rmvnorm.spam(n, mu=rep.int(0, dim(Sigma)[1]), Sigma, Rstruct=NULL, ..., mean, sigma)
rmvnorm.prec(n, mu=rep.int(0, dim(Q)[1]), Q, Rstruct=NULL, ...)
rmvnorm.canonical(n, b, Q, Rstruct=NULL, ...)
}
\arguments{
\item{n}{number of observations.}
\item{mu}{mean vector.}
\item{Sigma}{covariance matrix (of class \code{spam}).}
\item{Q}{precision matrix.}
\item{b}{vector determining the mean.}
\item{Rstruct}{the Cholesky structure of \code{Sigma} or \code{Q}.}
\item{\dots}{arguments passed to \code{chol}.}
\item{mean,sigma}{similar to \code{mu} and  \code{Sigma}. Here for portability with \code{mvtnorm::rmvnorm()}}
}
\details{All functions rely on a Cholesky factorization of the
  covariance or precision matrix.
  \cr
  The functions \code{rmvnorm.prec} and \code{rmvnorm.canonical}
  do not require sparse precision matrices
  Depending on the the covariance matrix \code{Sigma},  \code{rmvnorm}
  or \code{rmvnorm.spam} is used. If wrongly specified, dispatching to
  the other function is done.
  \cr
  Default mean is zero. Side note: mean is added via \code{sweep()} and
  no gain is accieved by distinguishing this case.
  \cr
  Often (e.g., in a Gibbs sampler setting), the sparsity structure of
  the covariance/precision does not change. In such setting, the
  Cholesky factor can be passed via \code{Rstruct} in which only updates
  are performed (i.e., \code{update.spam.chol.NgPeyton} instead of a
  full \code{chol}).
}
\references{See references in \code{\link{chol}}.
}
\seealso{\code{\link{rgrf}}, \code{\link{chol}} and \code{\link{ordering}}.
}
\examples{
# Generate multivariate from a covariance inverse:
# (usefull for GRMF)
set.seed(13)
n <- 25    # dimension
N <- 1000  # sample size
Sigmainv <- .25^abs(outer(1:n,1:n,"-"))
Sigmainv <- as.spam( Sigmainv, eps=1e-4)


Sigma <- solve( Sigmainv)  # for verification
iidsample <- array(rnorm(N*n),c(n,N))

mvsample <- backsolve( chol(Sigmainv), iidsample)
norm( var(t(mvsample)) - Sigma, type="m")

# compare with:
mvsample <- backsolve( chol(as.matrix( Sigmainv)), iidsample, n)
   #### ,n as patch
norm( var(t(mvsample)) - Sigma, type="m")


# 'solve' step by step:
b <- rnorm( n)
R <- chol(Sigmainv)
norm( backsolve( R, forwardsolve( R, b))-
      solve( Sigmainv, b) )
norm( backsolve( R, forwardsolve( R, diag(n)))- Sigma )


}
% backsolve( chol(as.matrix(V)[ord,ord]),iidsample)[iord,]
%
\author{Reinhard Furrer}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/bandwidth.Rd.                       %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{bandwidth}
\alias{bandwidth}
\title{Bandwidth of a Sparse Matrix}
\description{Returns the lower and upper bandwidth of a sparse matrix}
\usage{
bandwidth(A)
}
\arguments{
  \item{A}{spam object}
}
\details{The matrix does not need to be diagonal. Values can be negative
  indicating the the matrix contains a band cinfined in the
  upper or lower triangular part.
	}
\value{Integer vector containing the lower and upper bandwidth}
%\references{}
\seealso{
	\code{\link{diag.spam}}.
}
\examples{
bandwidth(spam(c(0, 1), 3, 2))

bandwidth(spam(c(0, 0, 1, rep(0, 9)), 4, 3))

}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/validate_spam.Rd.                   %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   



\name{validate_spam}
\alias{validate_spam}
\title{Validate a Sparse Matrix}
\description{Checks if the sparse matrix has the correct structure.}
\usage{
validate_spam(object)
}
\arguments{
  \item{object}{a spam matrix.}
}
\value{Returns \code{TRUE} if \code{object} is a valid spam objects. }

\seealso{\code{\link{spam.creation}}}
\examples{
validate_spam(spam(1, 20))
}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/defunct.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{spam-defunct}
\alias{spam-defunct}
\alias{validspamobject}
\title{Defunct Objects in Package \pkg{spam}}
\description{
  The functions or variables listed here are defunct, i.e. thorw an error when used.
}
\usage{
  validspamobject(...)
}
\arguments{
  \item{...}{some arguments}
}
\seealso{
  \code{\link{Deprecated}},
  \code{\link{Defunct}}
}
\keyword{misc}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/spam-class.Rd.                      %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%






\name{spam-class}
\docType{class}
\alias{spam.class}
\alias{spam-class}
\alias{as.matrix,spam-method}
\alias{as.matrix.spam}
\alias{[<-,spam,ANY,ANY,ANY-method}
\alias{[<-,spam,matrix,matrix,ANY-method}
\alias{[<-,spam,matrix,missing,ANY-method}
\alias{[<-,spam,missing,missing,ANY-method}
\alias{[<-,spam,missing,vector,ANY-method}
\alias{[<-,spam,missing,vector,spam-method}
\alias{[<-,spam,spam,missing,ANY-method}
\alias{[<-,spam,vector,missing,ANY-method}
\alias{[<-,spam,vector,missing,spam-method}
\alias{[<-,spam,vector,vector,ANY-method}
\alias{[<-,spam,vector,vector,spam-method}
\alias{[<-,spam,missing,missing,numeric-method}
\alias{[<-,spam,missing,vector,numeric-method}
\alias{[<-,spam,vector,missing,numeric-method}
\alias{[<-,spam,vector,vector,numeric-method}
\alias{[<-,spam,matrix,missing,numeric-method}
\alias{[<-,spam,matrix,matrix,numeric-method}
\alias{[<-,spam,spam,missing,numeric-method}
\alias{[<-,spam,ANY-method}
\alias{[,spam,missing,missing,ANY-method}
\alias{[,spam,missing,vector,ANY-method}
\alias{[,spam,vector,missing,logical-method}
\alias{[,spam,vector,missing,missing-method}
\alias{[,spam,vector,vector,ANY-method}
\alias{[,spam,matrix,missing,missing-method}
\alias{[,spam,matrix,missing,logical-method}
\alias{[,spam,matrix,matrix,ANY-method}
\alias{[,spam,spam,missing,ANY-method}
\alias{[,spam,ANY,ANY,ANY-method}
\alias{Arith,spam,spam-method}
\alias{Arith,spam,ANY-method}
\alias{Arith,ANY,spam-method}
\alias{Compare,ANY,spam-method}
\alias{Compare,spam,ANY-method}
\alias{Compare,spam,spam-method}
\alias{!,spam-method}
\alias{+,spam,missing-method}
\alias{-,spam,missing-method}
\alias{c,spam-method}
\alias{t,spam-method}
\alias{length<-,spam-method}
\alias{length,spam-method}
\alias{lower.tri,spam-method}
\alias{plot,spam,missing-method}
\alias{plot,spam,spam-method}
\alias{show,spam-method}
\alias{upper.tri,spam-method}

\title{Class "spam"}
\description{The \code{spam} class is a representation of sparse matrices.}
\section{Objects from the Class}{
  Objects can be created by calls of the form \code{new("spam", entries, colindices,
    rowpointes, dimension)}.
  The standard "old Yale sparse format" is used to store sparse matrices.\cr
  The matrix
\code{x} is stored in row form.  The first element of row \code{i} is
\code{x@rowpointers[i]}. The length of row \code{i} is determined by
\code{x@rowpointers[i+1]-x@rowpointers[i]}. The column indices of \code{x} are stored in
the \code{x@colindices} vector. The column index for element \code{x@entries[k]} is
\code{x@colindices[k]}.}
\section{Slots}{
  \describe{
    \item{\code{entries}:}{Object of class \code{"numeric"} contains the
      nonzero values. }
    \item{\code{colindices}:}{Object of class \code{"integer"} ordered indices
      of the nonzero values. }
    \item{\code{rowpointers}:}{Object of class \code{"integer"} pointer to the beginning
    of each row in the arrays \code{entries} and \code{colindices}.}
    \item{\code{dimension}:}{Object of class \code{"integer"} specifying
    the dimension of the matrix.}
  }
}
\section{Methods}{
  \describe{
    \item{as.matrix}{\code{signature(x = "spam")}:
      transforming a sparse
      matrix into a regular matrix.}
    \item{as.vector}{\code{signature(x = "spam")}:
      transforming a sparse
      matrix into a vector (dependings on \code{structurebased}) see \code{\link{as.vector.spam}} for details.}
    \item{as.spam}{\code{signature(x = "spam")}:
      cleaning of a sparse matrix.}
    \item{[<-}{\code{signature(x = "spam", i,j, value)}:
      assigning a
      sparse matrix. The negative vectors are not implemented yet. }
    \item{[}{\code{signature(x = "spam", i, j)}:
      subsetting a
      sparse matrix. The negative vectors are not implemented yet. }
    \item{\%*\%}{\code{signature(x, y)}:
      matrix multiplication, all combinations of sparse with full
      matrices or vectors are implemented.}
    \item{c}{\code{signature(x = "spam")}:
      vectorizes the sparse matrix and takes account of the zeros. Hence
    the lenght of the result is \code{prod(dim(x))}.}
    \item{cbind}{\code{signature(x = "spam")}: binds sparse matrices, see \code{\link{cbind}} for details.}
    \item{chol}{\code{signature(x = "spam")}:
    see \code{\link{chol}} for details.}
    \item{diag}{\code{signature(x = "spam")}:
    see \code{\link{diag}} for details.}
  \item{dim<-}{\code{signature(x = "spam")}: rearranges
    the matrix to reflect a new dimension.}
    \item{dim}{\code{signature(x = "spam")}: gives the dimension of the
      sparse matrix.}
  \item{pad<-}{\code{signature(x = "spam")}: truncates or augments
    the matrix see \code{\link{dim}} for details.}
    \item{image}{\code{signature(x = "spam")}:
      see \code{\link{image}} for details.}
    \item{display}{\code{signature(x = "spam")}:
      see \code{\link{display}} for details.}
%    \item{initialize}{\code{signature(.Object = "spam")}: ... }
    \item{length<-}{\code{signature(x = "spam")}: Is not implemented and
      causes an error.}
    \item{length}{\code{signature(x = "spam")}: gives the number of
      non-zero elements.}
    \item{lower.tri}{\code{signature(x = "spam")}: see \code{\link{lower.tri}} for details. }
    \item{Math}{\code{signature(x = "spam")}: see \code{\link{Math}} for details.}
    \item{Math2}{\code{signature(x = "spam")}: see \code{\link{Math2}} for details. }
    \item{norm}{\code{signature(x = "spam")}: calculates the norm of a matrix.}
    \item{plot}{\code{signature(x = "spam", y)}: same functionality as
      the ordinary \code{\link{plot}}. }
    \item{print}{\code{signature(x = "spam")}: see \code{\link{print}} for details.}
    \item{rbind}{\code{signature(x = "spam")}: binds sparse
      matrices, see \code{\link{cbind}} for details.}
%    \item{show}{\code{signature(object = "spam")}: ... }
    \item{solve}{\code{signature(a = "spam")}: see \code{\link{solve}} for details.}
    \item{summary}{\code{signature(object = "spam")}: small summary
      statement of the sparse matrix.}
    \item{Summary}{\code{signature(x = "spam")}:
      All functions of the \code{Summary} class (like \code{min}, \code{max},
      \code{range}...) operate on the vector \code{x@entries} and return the
      result thereof. See Examples or \code{\link{Summary}} for details.
    }
    \item{t}{\code{signature(x = "spam")}: transpose of a sparse matrix.}
    \item{upper.tri}{\code{signature(x = "spam")}: see \code{\link{lower.tri}} for details.}
  }
}
\section{Details}{
  The compressed sparse row (CSR) format is often described with the
  vectors \code{a}, \code{ia}, \code{ja}. To be a bit more
  comprehensive, we have chosen longer slot names.
}
\section{Note}{The slots \code{colindices} and \code{rowpointers} are
  tested for proper integer assignments. This is not true for
  \code{entries}.
}
%\references{}
\author{Reinhard Furrer, some of the Fortran code is based on A. George,
  J. Liu, E. S. Ng, B.W Peyton and Y. Saad (alphabetical)}
%\note{}
%\seealso{}
\examples{
showMethods("as.spam")


smat <- diag.spam(runif(15))
range(smat)
cos(smat)
}
\keyword{classes}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/options.Rd.                         %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



\name{options}
\alias{spam.options}
\alias{options.spam}
\alias{getOption.spam}
\title{Options Settings}
\description{
  Allow the user to set and examine a variety of \emph{options}
  which affect the way in which \R computes and displays sparse
  matrix results.
}
\details{
  Invoking \code{options()} with no arguments returns a list with the
  current values of the options. To access the value of a single option, one should
  use \code{getOption("spam.eps")}, e.g., rather than
  \code{options("spam.eps")} which is a \emph{list} of length one.\cr

  Of course, printing is still subordinate to
  \code{getOption("max.print")} or similar options.
}
\value{
  For \code{getOption}, the current value set for option \code{x}, or
  \code{NULL} if the option is unset.

  For \code{options()}, a list of all set options sorted by category.  For
  \code{options(name)}, a list of length one containing the set value,
  or \code{NULL} if it is unset.  For uses setting one or more options,
  a list with the previous values of the options changed (returned
  invisibly).
}
\section{Options used for the package \code{spam}}{
  A short description with the default values follows.
  \describe{
\item{\code{spam.eps=.Machine$double.eps}:}{values smaller than this are
  considered as zero. This is only used when creating spam objects.}

\item{\code{spam.drop=FALSE}:}{default parameter for \code{drop} when subsetting}

\item{\code{spam.printsize=100}:}{the max number of elements of a matrix which we
  display as regular matrix.}

\item{\code{spam.imagesize=10000}:}{the max number of elements of a matrix we display
  as regular matrix with \code{image} or \code{display}. Larger matrices are represented
  as dots only.}

\item{\code{spam.cex=1200}:}{default dot size for  \code{image} or \code{display}.}

\item{\code{spam.structurebased=FALSE}:}{should operations be carried out on
  the nonzero entries (the structure) or including the zeros.}

\item{\code{spam.inefficiencywarning=1e6}:}{issue a warning when inefficient
  operations are performed and the matrix exceeds the specified size.
  Valid value is a postive integer or a logical. \code{TRUE} corresponds
  to 1 (always), \code{FALSE} to \code{Inf}.}

\item{\code{spam.trivalues=FALSE}:}{a flag whether to return the structure
  (\code{FALSE}) or the values themselves (\code{TRUE}) when returning the
  upper and lower triangular part of a matrix.}

\item{\code{spam.listmethod="PE"}:}{algorithm for \code{spam.list}. Default
  is suggestion by Paul Eilers (thanks). Any other specification uses a
  bubble sort algorithm which is only slightly faster for very sparse matrices. }

\item{\code{spam.dopivoting=TRUE}:}{default parameter for "\code{solve}" routines. \code{FALSE}
  would solve the system without using the permutation.}

\item{\code{spam.NAOK=FALSE}:}{logical determines if \code{NA}, \code{NaN} and \code{Inf} are
  allowed to Fortan. Setting to \code{TRUE} allows to work with these but
  full functionality has not been tested.}

\item{\code{spam.safemodevalidity=TRUE}:}{logical determines if sanity check
  is peformed when constructing sparse matrices.
  Default is safer but somewhat slower.}

\item{\code{spam.cholsymmetrycheck=TRUE}:}{for the Cholesky factorization,
  verify if the matrix is symmetric.}
\item{\code{spam.cholpivotcheck=TRUE}:}{for the Cholesky factorization,
  when passing a permutation, should a minimum set of checks be performed?}
\item{\code{spam.cholupdatesingular="warning"}:}{for a Cholesky update, what
  happens if the matrix is singular: \code{"warning"} only and
  returning the not updated factor, \code{"error"}  or  return simply \code{"NULL"}.}
\item{\code{spam.cholincreasefactor=c(1.25,1.25)}:}{If not enought memory
  could be allocated, these are the steps to increase it.}

\item{\code{spam.nnznearestdistnnz=c(400^2,400)}:}{Memory allocation
  parameters for \code{nearest.dist}.}
\item{\code{spam.nearestdistincreasefactor=1.25}:}{If not enought memory
  could be allocated, this is the step to increase it. }
}
}
\seealso{Functions influenced by these options include: \code{\link{print.spam}},
  \code{\link{display.spam}},  \code{\link{image.spam}}, \code{\link{upper.tri.spam}},
  \code{\link{chol.spam}}, \code{\link{nearest.dist}}, etc.\cr
  \code{\link{powerboost}}\cr
}

\examples{
smat <- diag.spam(1:8)
smat
options(spam.printsize=49)
smat

# All spam options:
op <- options()[grep("spam", names(options()))]

# Reset to default values:
options(spam.eps=.Machine$double.eps,
        spam.drop=FALSE,
        spam.printsize=100,
        spam.imagesize=10000,
        spam.cex=1200,
        spam.structurebased=FALSE,
        spam.inefficiencywarning=1e6,
        spam.trivalues=FALSE,
        spam.listmethod="PE",
        spam.NAOK=FALSE,
        spam.safemodevalidity=TRUE,
        spam.dopivoting=TRUE,
        spam.cholsymmetrycheck=TRUE,
        spam.cholpivotcheck=TRUE,
        spam.cholupdatesingular="warning",
        spam.cholincreasefactor=c(1.25,1.25),
        spam.nearestdistincreasefactor=1.25,
        spam.nearestdistnnz=c(400^2,400))

}
\keyword{IO}
\keyword{environment}
\keyword{error}
\keyword{print}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/permutation.Rd.                     %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   


\name{permutation}
\alias{permutation}
\alias{permutation.spam}
\alias{permutation,spam-method}
\alias{permutation,matrix-method}
\title{Permute a Matrix}
\description{Row and/or column permutes a (sparse) matrix.
}
\usage{
permutation.spam(A, P=NULL, Q=NULL, ind=FALSE, check=TRUE)
}
\arguments{
\item{A}{sparse matrix}
\item{P}{vector giving the row permutation.}
\item{Q}{vector giving the column permutation.}
\item{ind}{are the indices given. See examples.}
\item{check}{Should rudimentary checks be performed.}
}
\value{A permuted matrix.
}
\details{If P and Q are permutation matrices, the result is
  PAQ. However, it is also possible to specify the indices
  and to perform in a very efficient way
  \code{A[rowind, colind]}, see examples.


  A row permutation is much faster than a colum permutation.
  For very large matrices, a double transpose might be faster.

  The spam option \code{spam.checkpivot} determines if the permutation
  is verified.
}
%\references{
%}
\seealso{\code{\link{ordering}}, \code{\link{spam.options}}.}
\examples{
A <- spam(1:12,3)
P <- c(3,1,2)
Q <- c(2,3,1,4)

permutation(A,P,Q)-A[order(P),order(Q)]

permutation(A,P,Q,ind=TRUE)-A[P,Q]

}
\author{Reinhard Furrer}
\keyword{array}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/apply.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   










\name{apply}
\alias{apply.spam}
\title{Apply Functions Over Sparse Matrix Margins}
\description{Returns a vector or array or list of values obtained by applying a
     function to margins of a sparse matrix.}
\usage{apply.spam(X, MARGIN=NULL, FUN, ...)
}
\arguments{
  \item{X}{the \code{spam} matrix to be used.}
  \item{MARGIN}{a vector giving the subscripts which the function will be
          applied over. \code{1} indicates rows, \code{2} indicates columns,
          \code{NULL} or \code{c(1,2)} indicates rows and columns.}
  \item{FUN}{the function to be applied.}
  \item{...}{optional arguments to \code{FUN}.}
}

\details{This is a handy wrapper to apply a function to the (nonzero)
  elements of a sparse matrix. 
  For example, it is possible to apply a covariance matrix to a distance
  matrix obtained by \code{nearest.dist}, see Examples.\cr

  A call to \code{apply} only coerces the sparse matrix to a regular one.\cr
  
  The basic principle is applying the function to \code{@entries}, or to
  the extracted columns or rows (\code{[,i,drop=F]} or
  \code{[i,,drop=F]}). It is important to note that an empty column
  contains at least one zero value and may lead to non intuitive
  results.\cr

  This function may evolve over the next few releases. 
}
\value{Similar as a call to \code{apply} with a regular matrix. The most
  important cases are as follows. The
  result is a vector (\code{MARGIN} is length 1 and \code{FUN} is
  scalar) or a matrix (\code{MARGIN} is length 1 and \code{FUN} returns
  fixed length vectors, or \code{MARGIN} is length 2 and \code{FUN} is
  scalar) or a list (if \code{FUN} returns vectors of different lengths).}
%\references{}
\seealso{\code{base:apply} for more details on Value.}
\examples{
S <- as.spam(dist(1:5))
S <- apply.spam(S/2, NULL, exp)
# instead of 
# S@entries <- exp( S@entries/2) 

# Technical detail, a null matrix consists
# of one zero element.
apply.spam(S,c(1,2),pmax)
apply.spam(S,1,range)


# A similar example as for the base apply.
# However, no dimnames else we would get warnings. 
x <- as.spam(cbind(x1 = 3, x2 = c(0,0,0, 5:2)))
apply.spam(x, 2, mean, trim = .2)
col.sums <- apply.spam(x, 2, sum)
row.sums <- apply.spam(x, 1, sum)
rbind(cbind(x, row.sums), c(col.sums, sum(col.sums)))

apply.spam(x, 2, is.vector)

# Sort the columns of a matrix
# Notice that the result is a list due to the different
# lengths induced by the nonzero elements
apply.spam(x, 2, sort)

# Function with extra args:
cave <- function(x, c1, c2) c(mean(x[c1]), mean(x[c2]))
apply(x,1, cave,  c1=1, c2=c(1,2))

ma <- spam(c(1:4, 0, 0,0, 6), nrow = 2)
ma
apply.spam(ma, 1, table)  #--> a list of length 2
apply.spam(ma, 1, stats::quantile)# 5 x n matrix with rownames
}
\author{Reinhard Furrer}
\keyword{array}
\keyword{algebra}

% HEADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is file spam/man/eigen.Rd.                           %
% It is part of the R package spam,                         %
%  --> https://CRAN.R-project.org/package=spam              %
%  --> https://CRAN.R-project.org/package=spam64            %
%  --> https://git.math.uzh.ch/reinhard.furrer/spam         %
% by Reinhard Furrer [aut, cre], Florian Gerber [aut],      %
%    Roman Flury [aut], Daniel Gerber [ctb],                %
%    Kaspar Moesinger [ctb]                                 %
% HEADER END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


\name{eigen}
\alias{eigen.spam}
\alias{eigen_approx}
\title{Eigenvalues for Sparse Matrices}
\description{
  Functions to calculate eigenvalues and eigenvectors of \code{sparse} matrices.
  It uses the value of \code{spam.options("inefficiencywarning")} to dispatch between \code{base::eigen()} or the Implicitly Restarted Arnoldi Process, using 'ARPACK'.

  \code{eigen.spam} is a wrapper function of \code{eigen_approx} and transforms its output to \code{base::eigen} like.
}
\note{
The user is advised to choose the \code{control} options carefully, see \sQuote{Details} for more information.
}
\usage{
eigen.spam(x, nev = 10, symmetric, only.values = FALSE, control = list())
eigen_approx(x, nev, ncv, nitr, mode, only.values = FALSE, verbose = FALSE, f_routine)
}
\arguments{
\item{x}{a matrix of class \code{spam} whose \code{nev} eigenvalues and eigenvectors are to be computed.}
\item{nev}{number of eigenvalues to calculate.}
\item{symmetric}{if TRUE, the matrix is assumed to be symmetric.}
\item{only.values}{if TRUE, only \code{nev} eigenvalues are computed and returned, otherwise \code{nev} eigenvalues and eigenvectors are returned.}
\item{control}{additional options, see \sQuote{Details}.}
\item{ncv}{see \sQuote{Details}, use the \code{control} option for \code{eigen.spam}.}
\item{nitr}{see \sQuote{Details}, use the \code{control} option for \code{eigen.spam}.}
\item{mode}{see \sQuote{Details}, use the \code{control} option for \code{eigen.spam}.}
\item{verbose}{see \sQuote{Details}, use the \code{control} option for \code{eigen.spam}.}
\item{f_routine}{only for \code{eigen_approx}, to call the Fortran routine for symmetric matrices set this option to "ds_eigen_f" and for non symmetric to "dn_eigen_f".}
}
\value{
  A vector of the length corresponding to the dimension of the input matrix.
  Containing the required \code{nev} eigenvalues.
  If requested also the corresponding eigenvectors.
  In the non symmetric case, the eigenvalues are returned in a matrix with a column containing the real parts and a column containing the imaginary parts of the eigenvalues.
  The eigenvectors are then returned in two matrices.}

\details{
  \describe{
    \item{\code{mode = "  "}:}{
      there are different modes available for this function, each mode returns a different range of eigenvalues.
      Also the available modes are dependent, whether the input matrix is symmetric or not:
      \describe{
        \item{\code{"LM"}:}{Eigenvalues with largest magnitude (sym, non sym), that is, largest eigenvalues in the Euclidean norm of complex numbers.}
        \item{\code{"SM"}:}{Eigenvalues with smallest magnitude (sym, non sym), that is, smallest eigenvalues in the Euclidean norm of complex numbers.}
        \item{\code{"LR"}:}{Eigenvalues with largest real part (non sym).}
        \item{\code{"SR"}:}{Eigenvalues with smallest real part (non sym).}
        \item{\code{"LI"}:}{Eigenvalues with largest imaginary part (non sym).}
        \item{\code{"SI"}:}{Eigenvalues with smallest imaginary part (non sym).}
        \item{\code{"LA"}:}{Eigenvalues with largest algebraic value (sym), that is, largest eigenvalues inclusive of any negative sign.}
        \item{\code{"SA"}:}{Eigenvalues with smallest algebraic value (syn), that is, smallest eigenvalues inclusive of any negative sign.}
      }% describe
    }% item
    \item{\code{ncv}:}{
      the largest number of basis vectors that will be used in the Implicitly Restarted Arnoldi Process.
      Work per major iteration is proportional to x@dimension[1]*ncv*ncv.
      The default is set if \code{symmetric} to min(x@dimension[1] + 1, max(2 * nev + 1, 200)) or else to min(x@dimension[1] - 1, max(2 * nev + 1, 100)).
      Note, this value should not be chosen arbitrary large, but slightly larger than \code{nev}.
      Otherwise it could lead to memory allocation problems.}
    \item{\code{nitr}:}{
      the maximum number of iterations.
      The default is set to \code{ncv + 1000}}
    \item{\code{spamflag = FALSE}:}{
      if TRUE, the Implicitly Restarted Arnoldi Process is used,
  independent of the dimension of the respective matrix (provided matrix
  is larger than 10x10).}
    \item{\code{verbose = FALSE}:}{
      print additional information.}
    \item{\code{cmplxeps}:}{
      threshold to determine whether a double value is zero, while transforming the ARPACK output to R class complex.
      The default is set to \code{.Machine$double.eps}.}
}% describe
}% details


\references{Lehoucq, R. B. and Sorensen, D. C. and Yang, C. (1997) \emph{ARPACK Users Guide: Solution of Large Scale Eigenvalue Problems by Implicitly Restarted Arnoldi Methods}.}
\seealso{Option \code{"inefficiencywarning"} in \code{\link{spam.options}} and \code{\link{spam_random}}.
}
\examples{
set.seed(81)
rspam <- spam_random(42^2, density = .0001, spd = TRUE)
SPD <- eigen.spam(rspam, nev = 18, control = list(mode = "SM"),
                  only.values = TRUE)

any(SPD$values <= 0, na.rm = TRUE)
isSymmetric(rspam)
# hence the matrix is symmetric positiv definit

rspam2 <- spam_random(50^2, density = .0001, spd = FALSE, sym = TRUE,
                      distribution = rpois, lambda = 2)
SNPD <- eigen.spam(rspam2, nev = 18, control = list(mode = "SM"),
                    only.values = TRUE)

any(SNPD$values <= 0, na.rm = TRUE)
isSymmetric(rspam2)
# hence the matrix is symmetric but not positiv definit
}
\author{Roman Flury, Reinhard Furrer}
\keyword{algebra}
